SystemOrganization addCategory: #'Glorp-Expressions'!
SystemOrganization addCategory: #'Glorp-Database'!
SystemOrganization addCategory: #'Glorp-UnitOfWork'!
SystemOrganization addCategory: #'Glorp-Queries'!
SystemOrganization addCategory: #'Glorp-Mappings'!
SystemOrganization addCategory: #'Glorp-Core'!
SystemOrganization addCategory: #'Glorp-DatabaseType'!
SystemOrganization addCategory: #'Glorp-DatabaseTypes'!
SystemOrganization addCategory: #'Glorp-Misc'!
SystemOrganization addCategory: #'Glorp-Licence'!

!String methodsFor: '*Glorp-Core' stamp: ''!
glorpIsCollection
	"For our purposes, these aren't collections, but rather a simple database type"
	^false.! !

!String methodsFor: '*Glorp-Core' stamp: ''!
glorpPrintSQLOn: aCommand
	| platform requiresEscape |
	platform := [aCommand platform] on: MessageNotUnderstood do: [:ex | ex return: nil].
	aCommand nextPut: $'.
	1
		to: self size
		do:
			[:i | 
			| char |
			char := self at: i.
			requiresEscape := platform isNil ifTrue: [false] ifFalse: [platform requiresEscapeFor: char].
			requiresEscape
				ifTrue: [aCommand nextPutAll: (aCommand platform escapeFor: char)]
				ifFalse: [aCommand nextPut: char]].
	aCommand nextPut: $'.! !

!String methodsFor: '*Glorp-Core' stamp: ''!
like: wildcardString
	" 'abcde' like: 'abc%'  "
	| newString |
	newString := wildcardString copy.
	newString replaceAll: $% with: $*.
	newString replaceAll: $_ with: $#.
	^newString match: self.! !

!Date methodsFor: '*Glorp-Core' stamp: ''!
glorpPrintSQLOn: aCommand
	"Print the date in ISO format. 'yyyy-mm-dd'  Don't rely on any dialect-specific formatting or padding mechanisms"
	| monthString dayString |
	aCommand 
		nextPut: $';
		print: self year;
		nextPut: $-.
	monthString := self monthIndex printString.
	monthString size = 1 ifTrue: [aCommand nextPut: $0 ].
	aCommand nextPutAll: monthString.
	aCommand nextPut: $-.
	dayString := self dayOfMonth printString.
	dayString size = 1 ifTrue: [aCommand nextPut: $0 ].
	aCommand nextPutAll: dayString.
	aCommand nextPut: $'.! !

!ReadStream methodsFor: '*Glorp-Core' stamp: ''!
actuallyAtEnd
	"For working around a VA cursor problem"
	^self atEnd! !

!ReadStream methodsFor: '*Glorp-Core' stamp: ''!
collect: aBlock
	| newStream |
	newStream := collection species new writeStream.
	[self atEnd] whileFalse: [newStream nextPut: (aBlock value: self next)].
	^newStream contents

"Use #writeStream, not WriteStream on: collection species, for portability;  VA cannot stream over all AdditiveSequenceableCollections."! !

!ReadStream methodsFor: '*Glorp-Core' stamp: ''!
collectionTypeFor: aClass
	^aClass.! !

!ReadStream methodsFor: '*Glorp-Core' stamp: ''!
detect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.
	Answer the first element for which aBlock evaluates to true."

	^self detect: aBlock ifNone: [self error: 'not found']! !

!ReadStream methodsFor: '*Glorp-Core' stamp: ''!
detect: aBlock ifNone: exceptionBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.
	Answer the first element for which aBlock evaluates to true."

	self do: [:each | (aBlock value: each) ifTrue: [^each]].
	^exceptionBlock value! !

!ReadStream methodsFor: '*Glorp-Core' stamp: ''!
includes: anObject
	self do: [:each | each = anObject ifTrue: [^true]].
	^false.! !

!ReadStream methodsFor: '*Glorp-Core' stamp: ''!
newCollectionOfSize: aSize for: aCollectionType

	^aCollectionType isBehavior ifTrue: [aCollectionType new: aSize] ifFalse: [aCollectionType copyEmpty: aSize].! !

!ReadStream methodsFor: '*Glorp-Core' stamp: ''!
select: aBlock
	| newStream |
	newStream := collection species new writeStream.
	[self atEnd] whileFalse: [
		| current |
		current := self next.
		(aBlock value: current) ifTrue: [newStream nextPut: current]].
	^newStream contents

"Use #writeStream, not WriteStream on: collection species, for portability;  VA cannot stream over all AdditiveSequenceableCollections."! !

!ReadStream methodsFor: '*Glorp-Core' stamp: ''!
statement: aGlorpPreparedStatement
	"This is the statement used to create us. If it's coming back to a simple read stream, then we must have already used all its data, and can thus reuse it"
	aGlorpPreparedStatement busy: false.! !

!Collection methodsFor: '*Glorp-Core' stamp: ''!
glorpAddAll: aCollection
	^self addAll: aCollection.! !

!Collection methodsFor: '*Glorp-Core' stamp: ''!
glorpAddAllToDictionary: aDictionary
	"Assumes that we are a collection of associations"
	self do: [:each | aDictionary add: each].! !

!Collection methodsFor: '*Glorp-Core' stamp: ''!
glorpGroupedBy: aBlock
	"Same as VW groupedBy:. This is too useful not to be able to use, but we don't want to conflict with other dialects/extensions that might implement, so make our own prefixed version"

	"Return a dictionary whose keys are the result of evaluating aBlock for all elements in
	 the collection, and the value for each key is the collection of elements that evaluated
	 to that key. e.g.
	     #(1 2 3 4 5) groupedBy: [:each | each odd]
	   a Dictionary  
	     true ---> #( 1 3 5)
	     false --> #(2 4)"

	| result |
	result := Dictionary new.
	self do:
		[:each | | key collection |
		key := aBlock value: each.
		collection := result at: key ifAbsentPut: [OrderedCollection new].
		collection add: each].
	self species ~~ OrderedCollection ifTrue:
		["Convert the result collections to be the right type.
		  Note that it should be safe to modify the dictionary
		  while iterating because we only replace values for existing keys"
		result keysAndValuesDo:
			[:key :value | result at: key put: (self species withAll: value)]].

	^result! !

!Collection methodsFor: '*Glorp-Core' stamp: ''!
glorpIsCollection

	^true.! !

!Collection methodsFor: '*Glorp-Core' stamp: ''!
glorpPrintSQLOn: aCommand
	aCommand nextPut: $(.
	self isEmpty
		ifTrue: [aCommand nextPutAll: 'NULL']
		ifFalse:
			[GlorpHelper
				do: [:each | each glorpPrintSQLOn: aCommand]
				for: self
				separatedBy: [aCommand nextPutAll: ', ']].
	aCommand nextPut: $).! !

!Collection methodsFor: '*Glorp-Core' stamp: ''!
glorpPrintSQLOn: aStream for: aType
	aStream nextPut: $(.
	self isEmpty
		ifTrue: [aStream nextPutAll: 'NULL']
		ifFalse:
			[GlorpHelper
				do: [:each | aType print: each on: aStream]
				for: self
				separatedBy: [aStream nextPutAll: ', ']].
	aStream nextPut: $).! !

!Collection methodsFor: '*Glorp-Core' stamp: ''!
glorpRegisterCollectionInternalsIn: anObjectTransaction 
	"Explicitly register any internal structures (e.g. a VW identity dictionary's valueArray) with the transaction. Assume we can safely register everything inside the collection reflectively. The obvious exceptions would be dependents and sortblocks. This is a cheat, and for peculiar cases you'll need to override this in the subclass"

	| names |
	names := self class allInstVarNames.
	(1 to: names size) do: 
			[:index | 
			(#('dependents' 'sortBlock') includes: (names at: index)) 
				ifFalse: [anObjectTransaction register: (self instVarAt: index)]]! !

!Collection methodsFor: '*Glorp-Core' stamp: ''!
writeStream
	^AddingWriteStream on: self! !

Collection subclass: #GlorpVirtualCollection
	instanceVariableNames: 'query session realObjects parameters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!GlorpVirtualCollection commentStamp: '<historical>' prior: 0!
This represents a virtual collection, i.e. one that we haven't really read into memory yet. It responds to a reasonable subset of collection protocol, and will read the elements into memory only when necessary. So, e.g. a select: operation takes a query block, and is equivalent to AND:ing that query block to the main query.

To create a virtual collection, ask the session for one. e.g. session virtualCollectionOf: AClass.

This is an initial version which will read in the objects fairly eagerly. An optimization might be to defer certain types of operations depending on whether the block can be evaluated into SQL or not. e.g.
  collect: [:each | each name]
can be turned into a retrieve: operation. But 
  collect: [:each | each printString]
cannot. We could try to check the block for operations like collect: and detect:, deferring the point at which the objects will be read in.

Handling of ordering is also a little bit funny. The blocks we like for ordering aren't compatible with sortedCollection type blocks. It'd be nice to be more compatible.

!

!GlorpVirtualCollection classSide methodsFor: 'instance creation' stamp: ''!
on: aClass in: aSession
	^self new on: aClass in: aSession.! !

!GlorpVirtualCollection methodsFor: 'private' stamp: ''!
AND: aBlock
	query AND: aBlock.! !

!GlorpVirtualCollection methodsFor: 'adding' stamp: ''!
add: newObject
	self readOnlyError.! !

!GlorpVirtualCollection methodsFor: 'enumerating' stamp: ''!
collect: aBlock
	^self realObjects collect: aBlock! !

!GlorpVirtualCollection methodsFor: 'copying' stamp: ''!
copy
	^self shallowCopy postCopy.! !

!GlorpVirtualCollection methodsFor: 'copying' stamp: ''!
copyAndGrowBy: anInteger
	"For VA compatibility"
	^super glorpShallowCopy.! !

!GlorpVirtualCollection methodsFor: 'copying' stamp: ''!
copyFrom: start to: stop
	^self realObjects copyFrom: start to: stop.! !

!GlorpVirtualCollection methodsFor: 'enumerating' stamp: ''!
do: aBlock
	self realObjects do: aBlock.! !

!GlorpVirtualCollection methodsFor: 'accessing' stamp: ''!
first
	^self realObjects first.! !

!GlorpVirtualCollection methodsFor: 'private' stamp: ''!
getQuery
	^query.! !

!GlorpVirtualCollection methodsFor: 'testing' stamp: ''!
isEmpty
	^self realObjects isEmpty.! !

!GlorpVirtualCollection methodsFor: 'testing' stamp: ''!
isInstantiated
	^realObjects notNil.! !

!GlorpVirtualCollection methodsFor: 'accessing' stamp: ''!
last
	^self realObjects last.! !

!GlorpVirtualCollection methodsFor: 'initialize-release' stamp: ''!
on: aClass in: aSession
	query := Query returningManyOf: aClass.
	session := aSession.! !

!GlorpVirtualCollection methodsFor: 'accessing' stamp: ''!
orderBy: aBlockOrExpression
	query orderBy: aBlockOrExpression.! !

!GlorpVirtualCollection methodsFor: 'accessing' stamp: ''!
parameters: aCollection
	parameters := aCollection! !

!GlorpVirtualCollection methodsFor: 'copying' stamp: ''!
postCopy
	query := query copy.
	realObjects := nil.! !

!GlorpVirtualCollection methodsFor: 'printing' stamp: ''!
printOn: aStream
	self isInstantiated
		ifTrue: [super printOn: aStream]
		ifFalse:
			[aStream nextPutAll: 'a virtual collection of '.
			query notNil ifTrue: [aStream nextPutAll: query resultClass name]].! !

!GlorpVirtualCollection methodsFor: 'private' stamp: ''!
readOnlyError
	self error: 'Virtual collections are read-only'.! !

!GlorpVirtualCollection methodsFor: 'private' stamp: ''!
realObjects
	realObjects isNil ifTrue: [	realObjects := query executeWithParameters: parameters in: session].
	^realObjects.! !

!GlorpVirtualCollection methodsFor: 'enumerating' stamp: ''!
reject: aBlock
	^self copy AND: [:each | (aBlock value: each) not].! !

!GlorpVirtualCollection methodsFor: 'removing' stamp: ''!
remove: oldObject ifAbsent: anExceptionBlock
	self readOnlyError.! !

!GlorpVirtualCollection methodsFor: 'enumerating' stamp: ''!
select: aBlock
	^self isInstantiated ifTrue: [self realObjects select: aBlock] ifFalse: [self copy AND: aBlock].! !

!GlorpVirtualCollection methodsFor: 'accessing' stamp: ''!
setQuery: aQuery
	"Semi-private. Set the internal query"
	query := aQuery.! !

!GlorpVirtualCollection methodsFor: 'accessing' stamp: ''!
setSession: aSession
	"Semi-private. Set the session"
	session := aSession.! !

!GlorpVirtualCollection methodsFor: 'accessing' stamp: ''!
size
	^self realObjects size.! !

ProtoObject subclass: #AbstractProxy
	instanceVariableNames: 'value isInstantiated'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!AbstractProxy commentStamp: '<historical>' prior: 0!
This is an abstract superclass for proxy objects. It inherits from ProtoObject, which is part of the platform specific layer, but implements most of the proxy functionality.

Instance Variables:
	isInstantiated	<Boolean>	Has this proxy been instantiated or not.
	value	<Object>	What we are proxying for.

!

!AbstractProxy classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new proxyInitialize.! !

!AbstractProxy methodsFor: 'accessing' stamp: ' 22/6/07 12:54'!
= anObject
	"In VA we inherit this as identity. Need to override"
	^self getValue = anObject.! !

!AbstractProxy methodsFor: 'printing' stamp: ''!
basicPrintOn: aStream 
	aStream nextPut: ${.
	isInstantiated 
		ifTrue: [self getValue basicPrintOn: aStream]
		ifFalse: [aStream nextPutAll: 'uninstantiated proxy'].
	aStream nextPut: $}! !

!AbstractProxy methodsFor: 'printing' stamp: ''!
basicPrintString
	| aStream |
	aStream := WriteStream on: String new.
	self basicPrintOn: aStream.
	^aStream contents! !

!AbstractProxy methodsFor: 'overrides' stamp: ''!
basicSize
	^self getValue basicSize.! !

!AbstractProxy methodsFor: 'dnu handling' stamp: ''!
doesNotUnderstand: aMessage

	^self getValue perform: aMessage selector withArguments: aMessage arguments.! !

!AbstractProxy methodsFor: 'dnu handling' stamp: ''!
doesNotUnderstand: sel args: args 
	"This is the variant for ObjectStudio"
	^self getValue perform: sel withArguments: args.! !

!AbstractProxy methodsFor: 'api' stamp: ''!
getValue

	self subclassesMustImplement! !

!AbstractProxy methodsFor: 'notification' stamp: ''!
glorpPostFetch: aSession! !

!AbstractProxy methodsFor: 'notification' stamp: ''!
glorpPostFetchValidate: aSession! !

!AbstractProxy methodsFor: 'overrides' stamp: ''!
hash
	^self getValue hash.! !

!AbstractProxy methodsFor: 'overrides' stamp: ''!
inspect
	^self getValue inspect.! !

!AbstractProxy methodsFor: 'testing' stamp: ''!
isGlorpProxy
	^true.! !

!AbstractProxy methodsFor: 'overrides' stamp: ''!
isImmutable
	^self getValue isImmutable.! !

!AbstractProxy methodsFor: 'testing' stamp: ''!
isInstantiated
	isInstantiated isNil ifTrue: [isInstantiated := false].
	^isInstantiated! !

!AbstractProxy methodsFor: 'accessing' stamp: ''!
isKindOf: aClass
	self isInstantiated ifTrue: [^value isKindOf: aClass].
	^false! !

!AbstractProxy methodsFor: 'printing' stamp: ''!
printOn: aStream 
	aStream nextPut: ${.
	isInstantiated 
		ifTrue: [self getValue printOn: aStream]
		ifFalse: [aStream nextPutAll: 'uninstantiated proxy'].
	aStream nextPut: $}! !

!AbstractProxy methodsFor: 'printing' stamp: ''!
printString

	| aStream |
	aStream := WriteStream on: (String new: 16).
	self printOn: aStream.
	^aStream contents! !

!AbstractProxy methodsFor: 'initialize' stamp: ''!
proxyInitialize

	isInstantiated := false.! !

!AbstractProxy methodsFor: 'overrides' stamp: ''!
store
	"For Object Studio compatibility"
	^self inspectorString.! !

!AbstractProxy methodsFor: 'accessing' stamp: ''!
uninstantiate

	value := nil.
	isInstantiated := false.! !

!AbstractProxy methodsFor: 'accessing' stamp: ''!
~= anObject 

	^(self = anObject) not! !

AbstractProxy subclass: #AdHocProxy
	instanceVariableNames: 'readBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!AdHocProxy commentStamp: '<historical>' prior: 0!
This represents a proxy that gets its value in a user-defined way, by evaluating a block.

Instance Variables:
	readBlock	<BlockClosure>	The zero-argument block we evaluate to get the proxied object.

!

!AdHocProxy classSide methodsFor: 'instance creation' stamp: ''!
readBlock: aZeroArgumentBlock

	^self new glorpReadBlock: aZeroArgumentBlock! !

!AdHocProxy methodsFor: 'accessing' stamp: ''!
class

	^AdHocProxy! !

!AdHocProxy methodsFor: 'api' stamp: ''!
getValue

	self isInstantiated ifTrue: [^value].
	[value := readBlock value] ensure: [isInstantiated := true].
	^value! !

!AdHocProxy methodsFor: 'api' stamp: ''!
glorpReadBlock: aZeroArgumentBlock
	readBlock := aZeroArgumentBlock! !

!AdHocProxy methodsFor: 'printing' stamp: ''!
printOn: aStream 
	aStream nextPut: ${.
	isInstantiated 
		ifTrue: [self getValue printOn: aStream]
		ifFalse: [aStream nextPutAll: 'uninstantiated AdHocProxy'].
	aStream nextPut: $}! !

!AdHocProxy methodsFor: 'initialize' stamp: ''!
proxyInitialize

	super proxyInitialize.
	readBlock := [nil]! !

AbstractProxy subclass: #Proxy
	instanceVariableNames: 'session query parameters action'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!Proxy commentStamp: '<historical>' prior: 0!
This implements the Glorp database proxy behaviour, for proxies that get their result by evaluating a query using a particular set of parameters, in the context of a particular Glorp query.

Instance Variables:
	parameters	<Dictionary | SequenceableCollection>	Our parameters, keyed either by name or by index.
	query	<Query>	The query we will evaluate. Note that for a mapping, all proxies will share the same query instance.
	session	<GlorpSession>	The session in which this proxy exists, and against which the query will be executed to find our value.

!

!Proxy classSide methodsFor: 'instance creation' stamp: ''!
returningManyOf: aClass where: aBlock 
	^self new query: (Query returningManyOf: aClass where: aBlock)! !

!Proxy classSide methodsFor: 'instance creation' stamp: ''!
returningManyOf: aClass where: aBlock in: aSession
	
	^self new 
		query: (Query returningManyOf: aClass where: aBlock);
		session: aSession! !

!Proxy classSide methodsFor: 'instance creation' stamp: ''!
returningOneOf: aClass where: aBlock 
	^self new query: (Query returningOneOf: aClass where: aBlock)! !

!Proxy classSide methodsFor: 'instance creation' stamp: ''!
returningOneOf: aClass where: aBlock in: aSession
	
	^self new 
		query: (Query returningOneOf: aClass where: aBlock);
		session: aSession.! !

!Proxy methodsFor: 'accessing' stamp: ''!
class

	^Proxy
	"self isInstantiated ifTrue: [^value class].
	query isNil ifTrue: [^Proxy].
	query resultClass isNil ifTrue: [^Proxy].
	^query resultClass."! !

!Proxy methodsFor: 'api' stamp: ''!
getValue
	self isInstantiated ifTrue: [^value].
	parameters isNil ifTrue: [parameters := Dictionary new: 0].
	[value := query isNil
		ifTrue: [nil]
		ifFalse: [query executeWithParameters: parameters in: session]]
		ensure: [isInstantiated := true].
	self glorpTriggerProxyAction: value.
	^value.! !

!Proxy methodsFor: 'proxy actions' stamp: ''!
glorpOwner
	"Return the object for which this proxy was created. Note that this is not kept track of unless we have a proxy action."
	action isNil ifTrue: [^nil].
	^action key.! !

!Proxy methodsFor: 'proxy actions' stamp: ''!
glorpOwner: anObject
	"If we don't have an action, the owner is unnecessary. Ignore it. If we do have one, put it in as the key of an association with the action block following it."
	action isNil ifTrue: [^self].
	action key: anObject.! !

!Proxy methodsFor: 'proxy actions' stamp: ''!
glorpProxyAction
	"Return the block we expect to evaluate when the proxy is fired."
	action isNil ifTrue: [^nil].
	^action value.! !

!Proxy methodsFor: 'proxy actions' stamp: ''!
glorpProxyAction: aTwoArgumentBlock
	"Set the action we'll take when the proxy fires. This is a two argument block that takes the owning object and the value of the proxy as arguments."
	aTwoArgumentBlock isNil ifTrue: [^action := nil].
	action := nil -> aTwoArgumentBlock.! !

!Proxy methodsFor: 'proxy actions' stamp: ''!
glorpTriggerProxyAction: target
	"Evaluate our action with the owner and the newly retrieved target."
	| owner block |
	action isNil ifTrue: [^self].
	owner := action key.
	block := action value.
	block value: owner value: target.! !

!Proxy methodsFor: 'accessing' stamp: ''!
isKindOf: aClass
	self isInstantiated ifTrue: [^value isKindOf: aClass].
	query isNil ifTrue: [^false].
	query resultClass isNil ifTrue: [^false].
	query resultClass isBehavior ifFalse: [^false].
	^query resultClass includesBehavior: aClass.! !

!Proxy methodsFor: 'accessing' stamp: ''!
parameters
	^parameters! !

!Proxy methodsFor: 'accessing' stamp: ''!
parameters: aDictionary
	parameters := aDictionary.! !

!Proxy methodsFor: 'printing' stamp: ''!
printOn: aStream 
	aStream nextPut: ${.
	isInstantiated 
		ifTrue: [self getValue printOn: aStream]
		ifFalse: 
			[aStream nextPutAll: 'uninstantiated '.
			query isNil
				ifTrue: [aStream nextPutAll: '???']
				ifFalse: [
					query readsOneObject ifFalse: [aStream nextPutAll: 'collection of '].
					aStream nextPutAll: query resultClass name]].
	aStream nextPut: $}! !

!Proxy methodsFor: 'accessing' stamp: ''!
query
	^query! !

!Proxy methodsFor: 'accessing' stamp: ''!
query: aQuery 
	query := aQuery! !

!Proxy methodsFor: 'accessing' stamp: ''!
session
	^session! !

!Proxy methodsFor: 'accessing' stamp: ''!
session: aSession 
	session := aSession! !

Proxy subclass: #SQLite3Proxy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!SQLite3Proxy commentStamp: '<historical>' prior: 0!
SQLiteProxy is different from a normal proxy in that its >>glorpTriggerProxyAction: method passes back the result of the block, which the >>getValue method reassigns to the originally retrieved #value inst var. The idea is to have a proxy that represents a DatabaseField, when there is no such entity in any SQLite schema table, so it must be constructed from a variety of special SQLite3 function calls (called pragmas).
!

!SQLite3Proxy methodsFor: 'accessing' stamp: ''!
class
	^SQLite3Proxy! !

!SQLite3Proxy methodsFor: 'api' stamp: ''!
getValue
	"This method further modifies the retrieved value using the action block."

	self isInstantiated ifTrue: [^value].
	parameters isNil ifTrue: [parameters := Dictionary new: 0].
	
	[value := query isNil
				ifTrue: [nil]
				ifFalse: [query executeWithParameters: parameters in: session]]
			ensure: [isInstantiated := true].
	value := self glorpTriggerProxyAction: value.
	^value! !

!SQLite3Proxy methodsFor: 'proxy actions' stamp: ''!
glorpTriggerProxyAction: target
	"Evaluate our action with the owner and the newly retrieved target.
	Return the result of this evaluation, or self if there is no action.
	Since the return value has ambiguous semantics, this needs work."

	| owner block |
	action isNil ifTrue: [^self].
	owner := action key.
	block := action value.
	^block value: owner value: target.! !

Proxy subclass: #TimedProxy
	instanceVariableNames: 'id reaper secondsToLive timeToDie association'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!TimedProxy commentStamp: '<historical>' prior: 0!
This is a proxy which only holds onto its value for a fixed amount of time, and then uninstantiates. This is useful in highly-connected object graphs, where we might like to get rid of parts of a large tree that we are no longer using, but where simple weak references won't give enough information. For one to one relationships, we can cut the proxy relation, the object will remain in cache if it is registered or being otherwised referenced, and if we ask for it again, we won't have to go to the database because it will be found in cache. It's probably not such a good idea to use these kind of proxies for collections, where we will have to go to the database again if we need the value. See also TimedProxyReaper.

Instance Variables:
	association	<Association> The association in the TimedProxyReaper which holds onto our identifier plus a time after which we should expire.
	id	<Integer>	An ID, assigned to us by the reaper, so that it doesn't hold strong references to us.
	reaper	<TimedProxyReaper>	The reaper which controls this proxy.
	secondsToLive	<ArithmeticValue>	How much time, in seconds, we should wait after we were last touched before expiring.
	timeToDie	<Integer>	The millisecond clock value at which we expire. Is this a problem if there are dialects/versions for which the millisecond clock wraps around?

!

!TimedProxy methodsFor: 'accessing' stamp: ''!
class
	^TimedProxy.! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
doesNotUnderstand: aMessage

	self glorpTouch.
	^super doesNotUnderstand: aMessage! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
glorpAssociation
	^association! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
glorpAssociation: anAssociation
	association := anAssociation! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
glorpId
	^id! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
glorpId: uniqueInteger
	id := uniqueInteger! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
glorpReaper: aTimedProxyReaper
	reaper := aTimedProxyReaper! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
glorpSecondsToLive
	"Answer the number of seconds the receiver should live after having been instantiated or sent a message."

	^secondsToLive! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
glorpSecondsToLive: seconds

	secondsToLive := seconds.
	self updateExpiryTime.
	reaper notNil ifTrue: [self glorpTouch]! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
glorpTimeToDie
	"Answer the time at which the receiver is scheduled to die."

	^timeToDie! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
glorpTouch
	self updateExpiryTime.
	reaper isNil ifFalse: [reaper touch: self].! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
proxyInitialize
	super proxyInitialize.
	secondsToLive := 30.
	self updateExpiryTime.! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
uninstantiate

	super uninstantiate.
	association := nil.! !

!TimedProxy methodsFor: 'accessing' stamp: ''!
updateExpiryTime
	timeToDie := Time millisecondClockValue + (secondsToLive * 1000).! !

ProtoObject subclass: #MessageArchiver
	instanceVariableNames: 'mySelector myArguments myReceiver query'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!MessageArchiver commentStamp: '<historical>' prior: 0!
This is what we use to construct GlorpExpression trees out of blocks. An instance of MessageArchiver is passed into the block. It has a doesNotUnderstand: handler, and in response to almost any message, will just return a new MessageArchiver which knows it is descended from this one. The result (assuming no inlined messages or other tricky stuff happened) is a tree of MessageArchivers recording all the sent messages. We can then traverse that and construct the appropriate expression tree.

Instance Variables:
	myArguments	<(Collection of: (Object | ProtoObject))>	The arguments to the message that caused me to be constructed.
	myReceiver	<MessageArchiver | nil>	The object that received the message that caused me to be constructed. For the initial MessageArchiver that's passed in, this will be nil. Otherwise it is a MessageArchiver.
	mySelector	<Symbol>	The selector of the message that caused me to be constructed.
	query	<AbstractReadQuery>	The query for which I am constructing the expression. This is here so that we can construct the final expression tree directly onto its BaseExpression.

!

!MessageArchiver classSide methodsFor: 'instance creation' stamp: ''!
receiver: aMessageCollector selector: aSymbol arguments: aCollection

	^self new
		receiver: aMessageCollector
		selector: aSymbol
		arguments: aCollection.! !

!MessageArchiver methodsFor: 'doesNotUnderstand' stamp: ''!
= anObject
	"Needed because VA's abtObservableWrapper implements =. Should be portable."
	^MessageArchiver 
		receiver: self
		selector: #= 
		arguments: (Array with: anObject).! !

!MessageArchiver methodsFor: 'expression creation' stamp: ''!
asGlorpExpression
	^self asGlorpExpressionOn: BaseExpression new! !

!MessageArchiver methodsFor: 'expression creation' stamp: ''!
asGlorpExpressionOn: aBaseExpression 
	| arguments |
	myReceiver == nil ifTrue: [^query isNil ifTrue: [aBaseExpression] ifFalse: [query baseExpression]].
	arguments := myArguments collect: [:each | each asGlorpExpressionOn: aBaseExpression].
	^self sendMessageTo: (myReceiver asGlorpExpressionOn: aBaseExpression) withArguments: arguments.! !

!MessageArchiver methodsFor: 'expression creation' stamp: ''!
asGlorpExpressionOn: aBaseExpression basedOn: aMessageArchiver withUltimateBase: ultimateBaseExpression
	"We know we are built from aMessageArchiver, but parts of us may refer to an outer scope. Those parts should be built on ultimateBaseExpression rather than aBaseExpression. e.g. consider 
 [:each | each amount amount = ((Query readManyOf: GlorpBankTransaction
			where: [:eachTrans | eachTrans owner = each owner]) 
				retrieve: [:eachTrans | eachTrans amount amount max])]."

	| arguments |
	self == aMessageArchiver ifTrue: [^aBaseExpression].
	myReceiver == nil ifTrue: [^query isNil ifTrue: [ultimateBaseExpression] ifFalse: [query baseExpression]].
	arguments := myArguments
		collect:
			[:each | 
			each
				asGlorpExpressionOn: aBaseExpression
				basedOn: aMessageArchiver
				withUltimateBase: ultimateBaseExpression].
	^self 
		sendMessageTo: (myReceiver
			asGlorpExpressionOn: aBaseExpression
			basedOn: aMessageArchiver
			withUltimateBase: ultimateBaseExpression)
		withArguments: arguments.! !

!MessageArchiver methodsFor: 'debugging' stamp: ''!
asText
	^self basicPrintString asText.! !

!MessageArchiver methodsFor: 'doesNotUnderstand' stamp: ''!
basicDoesNotUnderstand: aMessage 
	"Invoke this to avoid infinite recursion in the case of internal errors. We want a dialect-independent way of getting a walkback window, so we'll invoke it against a different object"


	(Array with: self) doesNotUnderstand: aMessage.! !

!MessageArchiver methodsFor: 'doesNotUnderstand' stamp: ''!
basicDoesNotUnderstand: sel args: args 
	"Invoke this to avoid infinite recursion in the case of internal errors. We want a dialect-independent way of getting a walkback window, so we'll invoke it against a different object"

	(Array with: self) doesNotUnderstand: sel args: args.! !

!MessageArchiver methodsFor: 'debugging' stamp: ''!
basicPrintString

	^self printString.! !

!MessageArchiver methodsFor: 'debugging' stamp: ''!
class
	^MessageArchiver! !

!MessageArchiver methodsFor: 'doesNotUnderstand' stamp: ''!
doesNotUnderstand: aMessage 
	| sel |
	sel := aMessage selector.
	sel == #doesNotUnderstand: ifTrue: [self basicDoesNotUnderstand: aMessage].
	(sel size >= 8 and: [(sel copyFrom: 1 to: 8) = 'perform:']) 
		ifTrue: 
			[^self get: aMessage arguments first
				withArguments: (aMessage arguments copyFrom: 2 to: aMessage arguments size)].
	^self get: sel withArguments: aMessage arguments.! !

!MessageArchiver methodsFor: 'doesNotUnderstand' stamp: ''!
doesNotUnderstand: sel args: args 
	"This is the variant for ObjectStudio"
	sel == #doesNotUnderstand: ifTrue: [self basicDoesNotUnderstand: sel args: args].
	(sel size >= 8 and: [(sel copyFrom: 1 to: 8) = 'perform:']) 
		ifTrue: 
			[^self get: args first
				withArguments: (args copyFrom: 2 to: args size)].
	^self get: sel withArguments: args.! !

!MessageArchiver methodsFor: 'expression protocol' stamp: ''!
get: aSymbol 
	^MessageArchiver receiver: self
		selector: aSymbol
		arguments: #().! !

!MessageArchiver methodsFor: 'expression protocol' stamp: ''!
get: aSymbol withArguments: anArray 
	| arguments |
	arguments := anArray collect: [:each | each glorpIsBlock 
		ifTrue: [(Dialect argumentCountFor: each) = 1
			ifTrue: [each value: self]
			ifFalse: [each value]]
		ifFalse: [each]].
	^MessageArchiver receiver: self
		selector: aSymbol
		arguments: arguments.! !

!MessageArchiver methodsFor: 'expression protocol' stamp: ''!
getMapping: aMapping named: aString
	"In this case, the name may actually be a string rather than a symbol."
	^MessageArchiverForGetMapping
		receiver: self
		selector: aString
		arguments: (Array with: aMapping with: aString).! !

!MessageArchiver methodsFor: 'testing' stamp: ''!
glorpIsBlock
	^false.! !

!MessageArchiver methodsFor: 'debugging' stamp: ''!
halt
	"Support this so that we can debug inside query blocks. For portability, send it to a different object so that we don't have to care how halt is implemented"

	(Array with: self) halt.! !

!MessageArchiver methodsFor: 'debugging' stamp: ''!
inspect

	"Not exactly the intended semantics, but should be portable"
	(Array with: self) inspect.! !

!MessageArchiver methodsFor: 'As yet unclassified' stamp: ''!
inspectorClasses
	^Array with: (Dialect smalltalkAt: 'Tools.Trippy.BasicInspector')! !

!MessageArchiver methodsFor: 'debugging' stamp: ''!
inspectorSize
  ^2! !

!MessageArchiver methodsFor: 'debugging' stamp: ''!
mustBeBoolean
	nil error: 'mustBeBoolean - this probably indicates you are trying to use an optimized message like and: or ifTrue: inside a Glorp expression block. Use AND: or & instead.'! !

!MessageArchiver methodsFor: 'doesNotUnderstand' stamp: ''!
os_Equal: anObject
	^self = anObject! !

!MessageArchiver methodsFor: 'debugging' stamp: ''!
printOn: aStream

	aStream nextPutAll: self printString.! !

!MessageArchiver methodsFor: 'debugging' stamp: ''!
printString

	"Hard-code this for maximum dialect portability"
	^'a MessageArchiver'.! !

!MessageArchiver methodsFor: 'private/accessing' stamp: ''!
privateGlorpQuery
	^query.! !

!MessageArchiver methodsFor: 'private/accessing' stamp: ''!
privateGlorpQuery: aQuery
	query := aQuery.! !

!MessageArchiver methodsFor: 'private/accessing' stamp: ''!
privateGlorpReceiver

	^myReceiver! !

!MessageArchiver methodsFor: 'private/accessing' stamp: ''!
privateGlorpSelector

	^mySelector! !

!MessageArchiver methodsFor: 'initialize' stamp: ''!
receiver: aMessageCollector selector: aSymbol arguments: aCollection

	myReceiver := aMessageCollector.
	mySelector := aSymbol.
	myArguments := aCollection.! !

!MessageArchiver methodsFor: 'expression creation' stamp: ''!
sendMessageTo: anExpression withArguments: arguments
	"Build our node by sending our message to anExpression"

	^anExpression get: mySelector withArguments: arguments.! !

MessageArchiver subclass: #MessageArchiverForGetMapping
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!MessageArchiverForGetMapping commentStamp: '<historical>' prior: 0!
This is a special subclass of MessageArchiver which constructs its expression using getMapping:named: rather than just get:. It's possible it ought to be generalized to one that's parameterized by what message it sends, but right now I only need the one...
!

!MessageArchiverForGetMapping methodsFor: 'expression creation' stamp: ''!
sendMessageTo: anExpression withArguments: arguments
	"Send our own particular message. Note that our arguments will have been wrapped as constant expressions, and we don't want that, so unwrap them"
	^anExpression getMapping: arguments first value named: (arguments at: 2) value.! !

!ProtoObject methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:20'!
_isProxy
	^true.! !

!ProtoObject methodsFor: '*Glorp-Core' stamp: ''!
asGlorpExpression
	^GlorpHelper glorpConstantExpressionClass for: self.! !

!ProtoObject methodsFor: '*Glorp-Core' stamp: ''!
asGlorpExpressionOn: anExpression

	^self asGlorpExpression.! !

!ProtoObject methodsFor: '*Glorp-Core' stamp: ''!
asGlorpExpressionOn: anExpression basedOn: aMessageArchiver withUltimateBase: aBaseExpression

	^self asGlorpExpression.! !

!ProtoObject methodsFor: '*Glorp-Core' stamp: ''!
isGlorpExpression
	^false.! !

!ProtoObject methodsFor: '*Glorp-Core' stamp: ''!
isGlorpProxy
	^false.! !

!ProtoObject methodsFor: '*Glorp-Core' stamp: ''!
isKindOf: aClass
	^self class includesBehavior: aClass.! !

!Association methodsFor: '*Glorp-Core' stamp: ''!
glorpAtIdentity: aKey ifAbsent: aBlock
	"A cheat to let use use this like an identity dictionary"
	^key == aKey ifTrue: [value] ifFalse: aBlock.! !

!Association methodsFor: '*Glorp-Core' stamp: ''!
keysAndValuesDo: aBlock
	"A memory allocation hack. Let us use a single association in place of a dictionary"
	aBlock value: key value: value.! !

!ByteArray methodsFor: '*Glorp-Core' stamp: ''!
glorpIsCollection
	"For our purposes, these aren't collections, but rather a simple database type"
	^false.! !

!ByteArray methodsFor: '*Glorp-Core' stamp: ''!
glorpPrintSQLOn: aCommand
	|strm|
	strm := WriteStream on: String new.
	strm nextPutAll: '0x'.
	strm nextPutAll: self asHexString.
	aCommand nextPutAll: strm contents! !

!UndefinedObject methodsFor: '*Glorp-Core' stamp: ''!
glorpPrintSQLOn: aCommand

	aCommand nextPutAll: 'NULL'.! !

Stream subclass: #GlorpCursoredStream
	instanceVariableNames: 'command cursor mightHaveDuplicateRows builders buildersThatReturnResults databaseAccessor peekBuffer rawPeekBuffer statement'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!GlorpCursoredStream commentStamp: '<historical>' prior: 0!
This is a Glorp wrapper for a database cursor. It has stream-like behaviour, which is implemented by getting rows from the underlying cursor and constructing the objects as required. All queries go through one of these streams, although in the default case they will get all the results out of it (at which point the cursor will be closed) and then return them in another kind of collection. But it's possible to use one of these as the return type, in which case you can make use of the open cursor to incrementally get results.

Instance Variables:
	builders	<(Collection of: ObjectBuilder)>	The ObjectBuilders used to construct the objects we return out of the rows.
	buildersThatReturnResults	<(SequenceableCollection of: ElementBuilder)>	
	command	<DatabaseCommand>	The command that was executed to give us this cursor.
	cursor	<>	The underlying database cursor. Class varies by dialect. We deal with the lack of polymorphism by indirecting all cursor operations through the accessor.
	databaseAccessor	<DatabaseAccessor>	The accessor that the command was executed against.
	mightHaveDuplicateRows	<Boolean>	May this stream have duplicate rows? This can happen, for example, if we do a join that causes some sub-section of the row to be duplicated. If so, we need to check for this and remove duplicates from the object stream as we go. This forces us to peek ahead.
	peekBuffer	<Object>	The peek buffer (holding an object) that we use when peeking ahead.
	rawPeekBuffer	<>	The peek buffer (holding a row) that we use when peeking ahead. We need to buffer at both levels.
	statement	<GlorpPreparedStatement>	The actual prepared statement that the command ran in order to get our results.

!

!GlorpCursoredStream classSide methodsFor: 'instance creation' stamp: ''!
on: aStream
	^self basicNew cursor: aStream.! !

!GlorpCursoredStream classSide methodsFor: 'instance creation' stamp: ''!
on: aStream accessor: aDatabaseAccessor
	^(self on: aStream)
		databaseAccessor: aDatabaseAccessor.! !

!GlorpCursoredStream classSide methodsFor: 'instance creation' stamp: ''!
on: aStream for: aCommand
	^(self on: aStream)
		command: aCommand.! !

!GlorpCursoredStream classSide methodsFor: 'instance creation' stamp: ''!
on: aStream for: aCommand accessor: aDatabaseAccessor
	^(self on: aStream)
		command: aCommand;
		databaseAccessor: aDatabaseAccessor.! !

!GlorpCursoredStream methodsFor: 'testing' stamp: ''!
accessor
	^databaseAccessor.! !

!GlorpCursoredStream methodsFor: 'testing' stamp: ''!
atEnd
	"Release ourselves fairly aggressively when we hit the end"
	| accessor |
	peekBuffer isNil ifFalse: [^false].
	rawPeekBuffer isNil ifFalse: [^false].
	cursor isNil ifTrue: [^true].
	"For test purposes, fabricate an accessor if we don't have one, it's not really needed for a non-db stream"
	accessor := self accessor isNil ifTrue: [DatabaseAccessor new] ifFalse: [self accessor].
	^([accessor isCursorAtEnd: cursor] on: Dialect error do: [:ex | databaseAccessor handleError: ex for: command])
		ifTrue:
			[self release.
			true]
		ifFalse: [	"We have to do this, because VA won't answer true to atEnd until we've read past end"
			self rawPeek isNil].! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
buildObjectsForRow: aRow
	self query isNil ifTrue: [^aRow].
	builders
		do:
			[:each | each findInstanceForRow: aRow proxyType: self query proxyType].
	builders do: [:each | each buildObjectFrom: aRow].
	builders
		do:
			[:each | 
			each
				sendPostFetchValidateIn: self session 
				OnFailureDo: [each instance: nil]].
	builders
		do: [:each | each instance == nil ifFalse: [each knitResultIn: self]].
	builders do: [:each | self session sendPostFetchEventTo: each instance].
	^self buildersThatReturnResults size = 1
		ifTrue: [self buildersThatReturnResults first returnValueIn: self]
		ifFalse:
			[self buildersThatReturnResults collect: [:each | each returnValueIn: self]].! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
buildersThatReturnResults
	buildersThatReturnResults isNil ifTrue: [
		buildersThatReturnResults := OrderedCollection new.
		builders
			do: [:each | 	
				| expression shouldBeRetrieved |
				"For some bizarre reason, includes: fails on ObjectStudio for this"
				expression := each expression.
				shouldBeRetrieved := (self query tracing retrievalExpressions detect: [:foo |
					foo == expression] ifNone: [nil]) notNil.
				(shouldBeRetrieved and: [
					(buildersThatReturnResults anySatisfy: [:eachBuilder | eachBuilder expression == expression]) not])
						ifTrue: [buildersThatReturnResults add: each]].
		buildersThatReturnResults := buildersThatReturnResults asArray].
	^buildersThatReturnResults.! !

!GlorpCursoredStream methodsFor: 'iterating' stamp: ''!
collect: aBlock 
	| newStream |
	newStream := WriteStream on: (Array new: 100).
	[self atEnd] whileFalse: [newStream nextPut: (aBlock value: self next)].
	^newStream contents! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
collectionTypeFor: aCollectionType
	"Make sure that we can reasonably return aCollectionType. If not, default to something we can"

	aCollectionType class == GlorpDictionaryDescription ifFalse: [^aCollectionType].
	self buildersThatReturnResults size = 1 ifTrue: [^aCollectionType].
	^Array.! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
command
	^command! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
command: aDatabaseCommand
	command := aDatabaseCommand.
	aDatabaseCommand isNil ifTrue: [^self].
	aDatabaseCommand query isNil ifTrue: [^self].
	databaseAccessor := aDatabaseCommand query session accessor.
	builders := self query copyOfBuilders.
	builders do: [:each | each cursor: self].! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
contents
	self shouldNotImplement.! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
cursor
	^cursor! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
cursor: aStream
	cursor := aStream! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
databaseAccessor: aDatabaseAccessor
	databaseAccessor := aDatabaseAccessor! !

!GlorpCursoredStream methodsFor: 'iterating' stamp: ''!
detect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.
	Answer the first element for which aBlock evaluates to true."

	^self detect: aBlock ifNone: [self error: 'not found']! !

!GlorpCursoredStream methodsFor: 'iterating' stamp: ''!
detect: aBlock ifNone: exceptionBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.
	Answer the first element for which aBlock evaluates to true."

	self do: [:each | (aBlock value: each) ifTrue: [^each]].
	^exceptionBlock value! !

!GlorpCursoredStream methodsFor: 'iterating' stamp: ''!
do: aBlock 
	"Evaluate aBlock for each of the elements of the receiver."

	[self atEnd]
		whileFalse: [aBlock value: self next]! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
elementBuilderFor: anExpression 
	^builders detect: [:each | each expression == anExpression] ifNone: [nil].! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
flush
	^self.! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
hasCompleteAnswerFor: objects
	| nextAnswer |
	self atEnd ifTrue: [^true].
	self mightHaveDuplicateRows ifFalse: [^true].
	nextAnswer := self peek.
	^objects ~= nextAnswer! !

!GlorpCursoredStream methodsFor: 'initializing' stamp: ''!
initialize! !

!GlorpCursoredStream methodsFor: 'testing' stamp: ''!
isReadable
	^true.! !

!GlorpCursoredStream methodsFor: 'testing' stamp: ''!
isWritable
	^false.! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
mightHaveDuplicateRows
	mightHaveDuplicateRows isNil ifTrue: [mightHaveDuplicateRows := self query isNil ifTrue: [false] ifFalse: [self query mightHaveDuplicateRows]].
	^mightHaveDuplicateRows.! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
newCollectionOfSize: aSize for: aCollectionType
	"Make sure that we can reasonably return aCollectionType. If not, default to something we can"
	| theType |
	theType := self collectionTypeFor: aCollectionType.
	^theType isBehavior ifTrue: [theType new: aSize] ifFalse: [theType copyEmpty: aSize].! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
next
	| answer session |
	self atEnd ifTrue: [^nil].
	session := self session.

	[self atEnd ifTrue: [^nil].
	answer := self nextAnswer.
	answer == nil] 
		whileTrue.

	[self hasCompleteAnswerFor: answer] whileFalse: [self nextAnswer].
	session == nil ifFalse: [session privateRegisterAsOld: answer].
	^answer.! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
next: anInteger
	| result |
	Dialect isObjectStudio ifFalse: [^super next: anInteger].
	result := OrderedCollection new: anInteger.
	anInteger timesRepeat: [result add: self next].
	^result.! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
nextAnswer
	| result |
	peekBuffer isNil ifFalse: [
		result := peekBuffer.
		peekBuffer := nil.
		^result].
	^self nextAnswerFromUnderlyingCursor.! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
nextAnswerFromUnderlyingCursor
	| objects |
	rawPeekBuffer isNil ifTrue: [rawPeekBuffer := cursor next].
	rawPeekBuffer isNil ifTrue: [^nil].
	objects := self buildObjectsForRow: rawPeekBuffer.
	rawPeekBuffer := nil.
	^objects.! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
nextObjectFrom: aRow
	| objects |
	objects := self buildObjectsForRow: aRow.
	self mightHaveDuplicateRows ifFalse: [
		self registerInstances.
		^objects].! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
nextPut: anObject
	self shouldNotImplement.! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
objectBuilderFor: anExpression 
	"Return the builder for this expression, but only those that directly build compound objects. Most importantly, exclude dictionary builders"
	^builders detect: [:each | 
		each buildsObjects and: [each expression == anExpression]] ifNone: [nil].! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
peek
	peekBuffer = nil ifFalse: [^peekBuffer].
	peekBuffer := self nextAnswer.
	^peekBuffer.! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
privateGetBuilders
	^builders.! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
query
	command isNil ifTrue: [^nil].
	^command query.! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
query: x! !

!GlorpCursoredStream methodsFor: 'private' stamp: ''!
rawPeek
	"Peek at the underlying rows, but don't build objects"
	rawPeekBuffer isNil ifFalse: [^rawPeekBuffer].
	rawPeekBuffer := cursor next.
	^rawPeekBuffer.! !

!GlorpCursoredStream methodsFor: 'building objects' stamp: ''!
registerInstances
	self session isNil ifTrue: [^self].
	builders do: [:each | self session register: each instance].! !

!GlorpCursoredStream methodsFor: 'initializing' stamp: ''!
release
	super release.
	self releaseCursor.
	cursor := nil.
	rawPeekBuffer := nil.
	peekBuffer := nil.
	statement isNil
		ifFalse:
			[databaseAccessor reusePreparedStatements
				ifTrue: [statement busy: false]
				ifFalse: [statement release].
			statement := nil].
	databaseAccessor := nil.! !

!GlorpCursoredStream methodsFor: 'initializing' stamp: ''!
releaseCursor
	
	self accessor isNil
		ifFalse:
			[self accessor releaseCursor: cursor in: statement.
			self accessor completedCommand: command].! !

!GlorpCursoredStream methodsFor: 'debugging' stamp: ''!
rows
	"This will extract all the rows and replace the internal stream with a regular stream on the rows. This is useful for debugging, but in real use rather destroys the purpose of the cursored stream"
	| rows |
	rows := cursor upToEnd.
	self releaseCursor.
	cursor := rows readStream.
	^rows.! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
session
	^self query isNil ifTrue: [nil] ifFalse: [self query session].! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
statement
	^statement.! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
statement: aGlorpPreparedStatement
	statement := aGlorpPreparedStatement.! !

!GlorpCursoredStream methodsFor: 'accessing' stamp: ''!
upToEnd
	| result |
	result := WriteStream on: (Array new: 100).
	[self atEnd] whileFalse: [
			| nextValue |
			"Stupid VA's cursors don't deal with atEnd or isEmpty nicely, so check for nil"
			nextValue := self next.
			nextValue == nil ifFalse: [	result nextPut: nextValue]].
	^result contents! !

!Dictionary methodsFor: '*Glorp-Core' stamp: ''!
glorpAddAll: aCollection
	"An addAll: that can work for either a dictionary argument or a collection of associations"
	aCollection glorpAddAllToDictionary: self.! !

!Time methodsFor: '*Glorp-Core' stamp: ''!
glorpPadToTwoDigits: anInteger
	| string |
	string := anInteger truncated printString.
	^string size = 1 ifTrue: ['0', string] ifFalse: [string].! !

!Time methodsFor: '*Glorp-Core' stamp: ''!
glorpPrintSQLOn: aCommand
	"Print as 24 hour time"

	aCommand 
		nextPut: $';
		nextPutAll: 	(self glorpPadToTwoDigits: self hours);
		nextPut: $:;
		nextPutAll: (self glorpPadToTwoDigits: self minutes);
		nextPut: $:;
		nextPutAll: (self glorpPadToTwoDigits: self seconds);
		nextPut: $'.! !

!BlockClosure methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:15'!
asGlorpExpression

	^self asGlorpExpressionOn: BaseExpression new.! !

!BlockClosure methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:15'!
asGlorpExpressionForDescriptor: aDescriptor

	| base |
	base := BaseExpression new.
	base descriptor: aDescriptor.
	^self asGlorpExpressionOn: base.! !

!BlockClosure methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:15'!
asGlorpExpressionOn: anExpression 
	^(self value: MessageArchiver new) asGlorpExpressionOn: anExpression! !

!BlockClosure methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:15'!
asGlorpExpressionOn: anExpression in: aQuery 
	^(self value: (MessageArchiver new privateGlorpQuery: aQuery)) asGlorpExpressionOn: anExpression.! !

!BlockClosure methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:16'!
asGlorpExpressionOn: anExpression withUltimateBase: aBaseExpression
	"Build the expression, knowing that some parts of it may refer to an outer context based on aBaseExpression. This is primarily useful for subselect blocks"
	| archiver tree |
	archiver := MessageArchiver new.
	tree := self value: archiver.
	^tree asGlorpExpressionOn: anExpression basedOn: archiver withUltimateBase: aBaseExpression.! !

!BlockClosure methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:16'!
glorpIsBlock
	^true.! !

Object subclass: #AddingWriteStream
	instanceVariableNames: 'target'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!AddingWriteStream commentStamp: '<historical>' prior: 0!
Why *can't* you stream onto a set? Or a bag, or a SortedCollection? No good reason that I can see. This implements only a subset of stream behaviour, that which is necessary to let us build up collections where we have to "append" elements using #add: rather than #at:put: and explicit grows.

Instance Variables:
	target	<Collection>	The thing we're streaming onto.

!

!AddingWriteStream classSide methodsFor: 'instance creation' stamp: ''!
on: aCollection
	^self new target: aCollection.! !

!AddingWriteStream methodsFor: 'accessing' stamp: ''!
contents
	^target.! !

!AddingWriteStream methodsFor: 'accessing' stamp: ''!
nextPut: anObject
	target add: anObject.! !

!AddingWriteStream methodsFor: 'accessing' stamp: ''!
nextPutAll: aCollection
	target glorpAddAll: aCollection.! !

!AddingWriteStream methodsFor: 'accessing' stamp: ''!
on: aSet
	target := aSet.! !

!AddingWriteStream methodsFor: 'accessing' stamp: ''!
target
	^target! !

!AddingWriteStream methodsFor: 'accessing' stamp: ''!
target: aCollection
	target := aCollection! !

Object subclass: #Cache
	instanceVariableNames: 'items policy mainCache extraReferences'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!Cache commentStamp: '<historical>' prior: 0!
This is the per-class cache of instances read from the database.

Instance Variables:
	items	<Dictionary from: Object to: Object>	The cached items, keyed by their primary key values
	policy	<CachePolicy>	The settings for this cache.


!

!Cache classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!Cache classSide methodsFor: 'instance creation' stamp: ''!
newFor: aClass in: aCacheManager 
	| newCache descriptor |
	descriptor := aCacheManager session isNil ifFalse: [aCacheManager system descriptorFor: aClass].
	newCache := Cache new.
	newCache mainCache: aCacheManager.
	descriptor isNil 
		ifTrue: [newCache cachePolicy: CachePolicy default]
		ifFalse: [newCache cachePolicy: descriptor cachePolicy].
	^newCache.! !

!Cache methodsFor: 'lookup' stamp: ''!
at: key ifAbsent: aBlock 
	| item value |
	item := self basicAt: key ifAbsent: [^aBlock value].
	value := policy contentsOf: item.
	(self hasItemExpired: item) 
		ifTrue: 
			[policy takeExpiryActionForKey: key withValue: value in: self. 
			(items includesKey: key) ifFalse: [^aBlock value]].
	^value.! !

!Cache methodsFor: 'lookup' stamp: ''!
at: key ifAbsentPut: aBlock 
	| item |
	item := self at: key ifAbsent: [nil].
	^item isNil 
		ifTrue: [ | newItem|
			newItem := policy cacheEntryFor: aBlock value.
			self markEntryAsCurrent: newItem.
			items at: key put: newItem]
		ifFalse: [self markEntryAsCurrent: item. item].! !

!Cache methodsFor: 'private' stamp: ''!
basicAt: anObject ifAbsent: aBlock

	^items at: anObject ifAbsent: aBlock.! !

!Cache methodsFor: 'accessing' stamp: ''!
cachePolicy
	^policy.! !

!Cache methodsFor: 'accessing' stamp: ''!
cachePolicy: aCachePolicy
	policy := aCachePolicy.
	self initializeCache.! !

!Cache methodsFor: 'private' stamp: ''!
do: aBlock
	items do: aBlock.! !

!Cache methodsFor: 'lookup' stamp: ''!
expireAll
	items do: [:each | policy expire: each].! !

!Cache methodsFor: 'lookup' stamp: ''!
expireKey: key ifAbsent: aBlock
	| item |
	item := self basicAt: key ifAbsent: [^aBlock value].
	policy expire: item.! !

!Cache methodsFor: 'private' stamp: ''!
expiredInstanceFor: key
	"Return the expired instance. Used for refreshing so that we don't recursively try and refresh when we get the instance to be refreshed"
	| item value |
	item := self basicAt: key ifAbsent: [self error: 'No expired instance found'].
	value := policy contentsOf: item.
	(self hasItemExpired: item) ifFalse: [self error: 'No expired instance found'].
	^value.! !

!Cache methodsFor: 'lookup' stamp: ''!
hasExpired: key
	| item |
	item := self basicAt: key ifAbsent: [^false].
	^self hasItemExpired: item.! !

!Cache methodsFor: 'lookup' stamp: ''!
hasItemExpired: anItem
	^(policy hasExpired: anItem) and: [(mainCache session isRegistered: (policy contentsOf: anItem)) not].! !

!Cache methodsFor: 'lookup' stamp: ''!
includesKey: key
	"Return true if we include the object. Don't listen to any expiry policy"
	self basicAt: key ifAbsent: [^false].
	^true.! !

!Cache methodsFor: 'lookup' stamp: ''!
includesKey: key as: anObject
	"Return true if we include the object, and it matches the given object. If we include a different object with the same key, raise an exception. Don't listen to any expiry policy"
	| item value |
	item := self basicAt: key ifAbsent: [^false].
	value := policy contentsOf: item.
	value == anObject ifFalse: [
		(DuplicatePrimaryKeyException new: anObject existing: value) signal].
	^true.! !

!Cache methodsFor: 'lookup' stamp: ''!
includesKey: key withClass: aClass
	"Return true if we include the object, and it matches our class. Don't listen to any expiry policy"
	| item value |
	item := self basicAt: key ifAbsent: [^false].
	value := policy contentsOf: item.
	^value isKindOf: aClass.! !

!Cache methodsFor: 'initialize' stamp: ''!
initialize! !

!Cache methodsFor: 'initialize' stamp: ''!
initializeCache

	items := policy newItemsIn: self.
	extraReferences := policy collectionForExtraReferences.! !

!Cache methodsFor: 'accessing' stamp: ''!
mainCache
	^mainCache! !

!Cache methodsFor: 'accessing' stamp: ''!
mainCache: aCacheManager
	mainCache := aCacheManager! !

!Cache methodsFor: 'private' stamp: ''!
markAsCurrentAtKey: key
	| item |
	item := self basicAt: key ifAbsent: [^false].
	^policy markEntryAsCurrent: item in: self.! !

!Cache methodsFor: 'private' stamp: ''!
markEntryAsCurrent: anItem
	"The policy has told us to mark an item as current. This is only really useful for weak policies, which tell us to keep an additional pointer to the object in a (presumably) fixed-size collection"
	extraReferences isNil ifFalse: [extraReferences add: anItem].! !

!Cache methodsFor: 'finalization' stamp: ''!
mournKeyOf: anEphemeron
	policy takeExpiryActionForKey: anEphemeron key withValue: anEphemeron value in: self.! !

!Cache methodsFor: 'accessing' stamp: ''!
numberOfElements
	^items size.! !

!Cache methodsFor: 'initialize' stamp: ''!
release
	policy release: self.
	extraReferences := nil.! !

!Cache methodsFor: 'lookup' stamp: ''!
removeKey: key ifAbsent: aBlock

	^items removeKey: key ifAbsent: aBlock.! !

!Cache methodsFor: 'accessing' stamp: ''!
session
	^mainCache session.! !

!Cache methodsFor: 'accessing' stamp: ''!
willRemoveItems
	^policy willRemoveItems.! !

Object subclass: #CacheManager
	instanceVariableNames: 'subCaches session'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!CacheManager commentStamp: '<historical>' prior: 0!
This is the entire cache for a session, consisting of multiple sub-caches, one per class.

Instance Variables:
	session	<Session>	The containing session.
	subCaches	<Dictionary from: Class to: Cache>	The per-class caches.


!

!CacheManager classSide methodsFor: 'instance creation' stamp: ''!
forSession: aSession 
	^self new session: aSession.! !

!CacheManager classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!CacheManager methodsFor: 'adding' stamp: ''!
at: aKey insert: anObject

	| subCache |
	subCache := self cacheForClass: anObject class.
	subCache at: aKey ifAbsentPut: [anObject].! !

!CacheManager methodsFor: 'private/caching' stamp: ''!
cacheFor: anObject
	"Get the cache for a particular object. Since this could conceivably be passed a proxy, check for that. The cache for an uninstantiated proxy is kind of ambiguous, treat it as nil.  This could also be a class"

	| nonMetaClass |
	nonMetaClass := anObject isBehavior ifTrue: [anObject] ifFalse: [anObject class].
	^nonMetaClass == Proxy
		ifTrue: [anObject isInstantiated ifTrue: [self cacheFor: anObject getValue] ifFalse: [nil]]
		ifFalse: [self cacheForClass: nonMetaClass].! !

!CacheManager methodsFor: 'initialize/release' stamp: ''!
cacheForClass: aClass
	^subCaches at: aClass
		ifAbsentPut: [self makeCacheFor: aClass]! !

!CacheManager methodsFor: 'querying' stamp: ''!
containsObject: anObject key: aKey

	| cache |
	cache := self cacheForClass: anObject class.
	^cache includesKey: aKey as: anObject.! !

!CacheManager methodsFor: 'querying' stamp: ''!
containsObjectForClass: aClass key: aKey

	| cache |
	cache := self cacheForClass: aClass.
	^cache includesKey: aKey.! !

!CacheManager methodsFor: 'querying' stamp: ''!
expireAll
	subCaches do: [:each | each expireAll].! !

!CacheManager methodsFor: 'querying' stamp: ''!
expireClass: aClass key: aKey

	^self expireClass: aClass key: aKey ifAbsent: [self error: 'Object not in cache'].! !

!CacheManager methodsFor: 'querying' stamp: ''!
expireClass: aClass key: aKey ifAbsent: failBlock 
	| cache |
	cache := self cacheForClass: aClass.
	(cache includesKey: aKey withClass: aClass) ifFalse: [^failBlock value].
	cache expireKey: aKey ifAbsent: [ failBlock value ].! !

!CacheManager methodsFor: 'private/caching' stamp: ''!
expiredInstanceOf: aClass key: aKey

	^(self cacheForClass: aClass) expiredInstanceFor: aKey.! !

!CacheManager methodsFor: 'querying' stamp: ''!
hasExpired: anObject
	| key cache |
	key := (session descriptorFor: anObject) primaryKeyFor: anObject.
	cache := self cacheFor: anObject.
	cache isNil ifTrue: [^false].  "We have an uninstantiated proxy."
	^cache hasExpired: key.! !

!CacheManager methodsFor: 'querying' stamp: ''!
hasExpired: aClass key: key
	| cache |
	cache := self cacheFor: aClass.
	^cache hasExpired: key.! !

!CacheManager methodsFor: 'querying' stamp: ''!
hasObjectExpiredOfClass: aClass withKey: key
	| cache |
	cache := self cacheForClass: aClass.
	^cache hasExpired: key.! !

!CacheManager methodsFor: 'initialize/release' stamp: ''!
initialize

	subCaches := IdentityDictionary new: 100.! !

!CacheManager methodsFor: 'querying' stamp: ''!
lookupClass: aClass key: aKey

	^self lookupClass: aClass key: aKey ifAbsent: [self error: 'cache miss'].! !

!CacheManager methodsFor: 'querying' stamp: ''!
lookupClass: aClass key: aKey ifAbsent: failBlock

	| object |
	object := (self cacheForClass: aClass) at: aKey ifAbsent: failBlock.
	^(object isKindOf: aClass)
		ifTrue: [ object ]
		ifFalse: [ failBlock value ].! !

!CacheManager methodsFor: 'querying' stamp: ''!
lookupClass: aClass key: aKey ifAbsentPut: failBlock

	^(self cacheForClass: aClass) at: aKey ifAbsentPut: failBlock.! !

!CacheManager methodsFor: 'initialize/release' stamp: ''!
makeCacheFor: aClass 
	| rootClass cache |
	rootClass := session isNil ifTrue: [aClass] ifFalse: [session lookupRootClassFor: aClass].
	cache := subCaches at: rootClass ifAbsentPut: [ Cache newFor: rootClass in: self ].
	subCaches at: aClass put: cache.
	^cache! !

!CacheManager methodsFor: 'querying' stamp: ''!
markAsCurrentOfClass: aClass key: key
	| cache |
	aClass == Proxy ifTrue: [^self].
	cache := self cacheForClass: aClass.
	cache markAsCurrentAtKey: key.! !

!CacheManager methodsFor: 'accessing' stamp: ''!
numberOfElements
	^subCaches inject: 0 into: [:sum :each |
		sum + each numberOfElements].! !

!CacheManager methodsFor: 'initialize/release' stamp: ''!
release
	subCaches do: [:each | each release].! !

!CacheManager methodsFor: 'querying' stamp: ''!
removeClass: aClass key: aKey

	^self removeClass: aClass key: aKey ifAbsent: [self error: 'Object not in cache'].! !

!CacheManager methodsFor: 'querying' stamp: ''!
removeClass: aClass key: aKey ifAbsent: failBlock 
	| cache |
	cache := self cacheForClass: aClass.
	(cache includesKey: aKey withClass: aClass) ifFalse: [^failBlock value].
	cache removeKey: aKey ifAbsent: [ failBlock value ].! !

!CacheManager methodsFor: 'accessing' stamp: ''!
session

	^session.! !

!CacheManager methodsFor: 'accessing' stamp: ''!
session: aSession 
	session := aSession.! !

!CacheManager methodsFor: 'accessing' stamp: ''!
system

	^self session system.! !

Object subclass: #CachePolicy
	instanceVariableNames: 'expiryAction numberOfElements'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!CachePolicy commentStamp: '<historical>' prior: 0!
A CachePolicy implements the different possible policies we might use for caching. The superclass implements the trivial policy of keeping all objects forever.

The policy also controls what we store in the cache. In general, it's assumed to be a cache entry of some sort, and the policy is responsible for wrapping and unwrapping objects going to and from the cache. The default policy is that the objects themselves are the cache entry (saving one object per cached object in overhead).

Instance Variables
	size	<Number>	The minimum cache size we want to use.
	expiryAction <Symbol> What to do when an object has expired. Currently hard-coded as one of #remove, #notify, #refresh, #notifyAndRemove.
!

!CachePolicy classSide methodsFor: 'instance creation' stamp: ''!
default
	"Use of the WeakVWCache is tied in with Ephemera which we do not expect to get working in VW3."

	Dialect isVWWithNameSpaces ifTrue: [^WeakVWCachePolicy new].
	^self new! !

!CachePolicy classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!CachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
cacheEntryFor: anObject

	^anObject.! !

!CachePolicy methodsFor: 'initialize' stamp: ''!
collectionForExtraReferences
	^nil.! !

!CachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
contentsOf: aCacheEntry

	^aCacheEntry.! !

!CachePolicy methodsFor: 'accessing' stamp: ''!
dictionaryClass
	^Dictionary.! !

!CachePolicy methodsFor: 'expiry' stamp: ''!
expire: anItem
	"Force anItem to be expired. This may be ignored if the policy doesn't respect this, and is ignored by default"! !

!CachePolicy methodsFor: 'accessing' stamp: ''!
expiryAction
	^expiryAction! !

!CachePolicy methodsFor: 'accessing' stamp: ''!
expiryAction: aSymbol
	"See class comment for possible values"
	expiryAction := aSymbol! !

!CachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
hasExpired: aCacheEntry

	^false.! !

!CachePolicy methodsFor: 'initialize' stamp: ''!
initialize

	numberOfElements := 100.
	expiryAction := #remove.! !

!CachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
markEntryAsCurrent: aCacheEntry in: aCache

	^self.! !

!CachePolicy methodsFor: 'initialize' stamp: ''!
newItemsIn: aCache
	^self dictionaryClass new: 20.! !

!CachePolicy methodsFor: 'expiry' stamp: ''!
notifyOfExpiry: anObject in: aCache 
	anObject glorpNoticeOfExpiryIn: aCache session.! !

!CachePolicy methodsFor: 'accessing' stamp: ''!
numberOfElements
	^numberOfElements! !

!CachePolicy methodsFor: 'accessing' stamp: ''!
numberOfElements: anInteger
	numberOfElements := anInteger! !

!CachePolicy methodsFor: 'expiry' stamp: ''!
release: aCache
	(expiryAction == #notify or: [expiryAction == #notifyAndRemove])
		ifTrue: [
			aCache do: [:each |
				(self contentsOf: each)  glorpNoticeOfExpiryIn: aCache session]].! !

!CachePolicy methodsFor: 'expiry' stamp: ''!
takeExpiryActionForKey: key withValue: anObject in: aCache
	expiryAction == #refresh
		ifTrue: [aCache session refresh: anObject].
	(#(#notify #notifyAndRemove) includes: expiryAction) ifTrue: [
		self notifyOfExpiry: anObject in: aCache].
	(#(#remove #notifyAndRemove) includes: expiryAction) ifTrue: [
		aCache removeKey: key ifAbsent: []].! !

!CachePolicy methodsFor: 'expiry' stamp: ''!
willRemoveItems
	^(expiryAction == #notifyAndRemove or: [expiryAction == #remove])! !

CachePolicy subclass: #TimedExpiryCachePolicy
	instanceVariableNames: 'timeout'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!TimedExpiryCachePolicy commentStamp: '<historical>' prior: 0!
This implements a cache that notes that an object is stale after some amount of time since it has been read.

Instance Variables:
	timeout	<Integer>	The time in seconds until we note an object as needing refreshing.


!

!TimedExpiryCachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
cacheEntryFor: anObject

	^Array with: self totalSeconds with: anObject.! !

!TimedExpiryCachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
contentsOf: aCacheEntry

	^aCacheEntry at: 2.! !

!TimedExpiryCachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
expire: aCacheEntry

	^aCacheEntry at: 1 put: 0.! !

!TimedExpiryCachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
hasExpired: aCacheEntry

	^(self totalSeconds - (aCacheEntry at: 1)) >= timeout.! !

!TimedExpiryCachePolicy methodsFor: 'initialize' stamp: ''!
initialize

	super initialize.
	timeout := 300.! !

!TimedExpiryCachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
markEntryAsCurrent: aCacheEntry in: aCache

	aCacheEntry at: 1 put: self totalSeconds.! !

!TimedExpiryCachePolicy methodsFor: 'accessing' stamp: ''!
timeout
	^timeout! !

!TimedExpiryCachePolicy methodsFor: 'accessing' stamp: ''!
timeout: seconds
	timeout := seconds! !

!TimedExpiryCachePolicy methodsFor: 'accessing' stamp: ''!
timeoutSeconds
	^timeout! !

!TimedExpiryCachePolicy methodsFor: 'accessing' stamp: ''!
timeoutSeconds: anInteger
	timeout := anInteger.! !

!TimedExpiryCachePolicy methodsFor: 'utility' stamp: ''!
totalSeconds

	^Dialect totalSeconds.! !

CachePolicy subclass: #TimedStrongCachePolicy
	instanceVariableNames: 'timeout'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!TimedStrongCachePolicy commentStamp: '<historical>' prior: 0!
This implements a cache that notes that an object is stale after some amount of time since it has been read.

Instance Variables:
	timeout	<Integer>	The time in seconds until we note an object as needing refreshing.


!

!TimedStrongCachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
cacheEntryFor: anObject

	^Array with: self totalSeconds with: anObject.! !

!TimedStrongCachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
contentsOf: aCacheEntry

	^aCacheEntry at: 2.! !

!TimedStrongCachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
expire: aCacheEntry

	^aCacheEntry at: 1 put: 0.! !

!TimedStrongCachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
hasExpired: aCacheEntry

	^(self totalSeconds - (aCacheEntry at: 1)) >= timeout.! !

!TimedStrongCachePolicy methodsFor: 'initialize' stamp: ''!
initialize

	super initialize.
	timeout := 300.! !

!TimedStrongCachePolicy methodsFor: 'wrap/unwrap' stamp: ''!
markEntryAsCurrent: aCacheEntry in: aCache

	aCacheEntry at: 1 put: self totalSeconds.! !

!TimedStrongCachePolicy methodsFor: 'accessing' stamp: ''!
timeout
	^timeout! !

!TimedStrongCachePolicy methodsFor: 'accessing' stamp: ''!
timeout: seconds
	timeout := seconds! !

!TimedStrongCachePolicy methodsFor: 'accessing' stamp: ''!
timeoutSeconds
	^timeout! !

!TimedStrongCachePolicy methodsFor: 'accessing' stamp: ''!
timeoutSeconds: anInteger
	timeout := anInteger.! !

!TimedStrongCachePolicy methodsFor: 'utility' stamp: ''!
totalSeconds

	^Dialect totalSeconds.! !

CachePolicy subclass: #WeakVWCachePolicy
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!WeakVWCachePolicy commentStamp: '<historical>' prior: 0!
This is a cache policy that uses VisualWorks 7.x weak references (ephemerons) to store references to objects, letting them vanish if not referenced. It uses the numberOfElements inst var as an indicator of how many objects to keep hard references to, preventing objects from disappearing too quickly.

Instance Variables:!

!WeakVWCachePolicy methodsFor: 'expiry' stamp: ''!
collectionForExtraReferences

	^FixedSizeQueue maximumSize: self numberOfReferencesToKeepAround.! !

!WeakVWCachePolicy methodsFor: 'accessing' stamp: ''!
dictionaryClass
	^Dialect smalltalkAt: #EphemeralValueDictionary ifAbsent: [Dictionary].! !

!WeakVWCachePolicy methodsFor: 'expiry' stamp: ''!
markEntryAsCurrent: item in: aCache
	aCache markEntryAsCurrent: item.! !

!WeakVWCachePolicy methodsFor: 'initialize-release' stamp: ''!
newItemsIn: aCache
	| items |
	items := super newItemsIn: aCache.
	items manager: aCache.
	^items.! !

!WeakVWCachePolicy methodsFor: 'accessing' stamp: ''!
numberOfReferencesToKeepAround
	^numberOfElements.! !

Object subclass: #DatabaseAccessor
	instanceVariableNames: 'connection currentLogin logging reusePreparedStatements deniedCommands mutex dependents'
	classVariableNames: 'LoggingEnabled'
	poolDictionaries: ''
	category: 'Glorp-Database'!

!DatabaseAccessor commentStamp: '<historical>' prior: 0!
This translates between the the general, portable Glorp protocol and whatever needs to be done to access a database in a particular smalltalk dialect. Most of the real work is done in its subclasses.

Subclasses must implement the following messages:
	accessing
		connectionClassForLogin:
	executing
		basicExecuteSQLString:
		externalDatabaseErrorSignal
		rowCount
	login
		loginIfError:
		logout
		showDialog:

Instance Variables:
	connection	<varies>	The underlying database connection
	currentLogin	<Login> The glorp-level login we are using
	logging	<Boolean>	Is logging on or off
	permittedCommands	<Collection of: DatabaseCommand class>	The list of commands we are permitted to use. This allows us to filter commands, indicating that a particular session should only be doing inserts right now, not updates, deletes, or queries. Or that it should only be doing writes, not reads.
	reusePreparedStatements	<Boolean>	Should we try to reuse prepared statements,that have the same SQL, or just re-prepare each time.

!

!DatabaseAccessor classSide methodsFor: 'instance creation' stamp: 'GuillermoPolito 4/16/2011 22:42'!
classForThisPlatform

	Dialect isGNU ifTrue: [^Dialect smalltalkAt: #JdmDatabaseAccessor].
	Dialect isSqueak ifTrue: [^Dialect smalltalkAt: #SqueakDatabaseAccessor].
	Dialect isPharo ifTrue: [^Dialect smalltalkAt: #PharoDatabaseAccessor].
	Dialect isVisualWorks ifTrue: [^Dialect smalltalkAt: #'Glorp.VWDatabaseAccessor'].
	Dialect isVisualAge ifTrue: [^Dialect smalltalkAt: #VADatabaseAccessor].
	Dialect isDolphin ifTrue: [^Dialect smalltalkAt: #DolphinDatabaseAccessor].
	Dialect isObjectStudio ifTrue: [^Dialect smalltalkAt: #ObjectStudioDatabaseAccessor].
	self error: 'unknown dialect'.
	^self! !

!DatabaseAccessor classSide methodsFor: 'instance creation' stamp: ''!
forLogin: aLogin 
	^self classForThisPlatform new currentLogin: aLogin! !

!DatabaseAccessor classSide methodsFor: 'accessing' stamp: ''!
loggingEnabled
	LoggingEnabled isNil ifTrue: [LoggingEnabled := false].
	^LoggingEnabled! !

!DatabaseAccessor classSide methodsFor: 'accessing' stamp: ''!
loggingEnabled: aBoolean 
	LoggingEnabled := aBoolean! !

!DatabaseAccessor classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
addConstraint: aConstraint 
	
	^self 
		addConstraint: aConstraint
		ifError: [ :ex | 	| msg|
				msg := ex messageText.
				msg isEmpty ifTrue: [msg := 'No error message provided for top level exception.'].
				Transcript cr; show: ('Add constraint <1s> ERROR: (<2s>) ' expandMacrosWith: aConstraint name with: msg).
				ex return: nil].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
addConstraint: aConstraint ifError: aBlock
	| addString |
	addString := self platform createConstraintAddStatement: aConstraint.
	addString isEmpty ifTrue: [^self].
	^self
		executeInDDLOperation: [ self executeSQLStringNoResult: addString ]
		ifError: aBlock! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
addFields: aCollection 

	self 
		addFields: aCollection 
		ifError: [ :ex  | 	| msg |
				msg := ex messageText.
				msg isEmpty ifTrue: [msg := 'No error message provided for top level exception.'].
				Transcript cr; show: ('Add field ERROR: (<2s>) ' expandMacrosWith: msg).
				ex return: nil]! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
addFields: aCollection ifError: aBlock

	aCollection
		do: 	[:aField | 
			self execute: [self executeSQLStringNoResult: (self platform createAddColumnStatementFor: aField)]
				ifError: aBlock ].! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
allowCommand: aCommand during: aBlock
	| oldCommands |
	oldCommands := self deniedCommands.
	((self deniedCommands ifNil: [#()]) includes: aCommand)
		ifTrue: [deniedCommands := deniedCommands copyWithout: aCommand].
	aBlock ensure: [deniedCommands := oldCommands].! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
alsoDenyCommand: aCommand
	self deniedCommands isNil
		ifTrue: [self denyCommand: aCommand]
		ifFalse:
			[self deniedCommands: self deniedCommands , (Array with: aCommand)].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
alterField: aField newType: aType 

	^self 
		alterField: aField 
		newType: aType 
		ifError: [ :ex | 	| msg|
				msg := ex messageText.
				msg isEmpty ifTrue: [msg := 'No error message provided for top level exception.'].
				Transcript cr; show: ('Alter column <1s> to set new type <2s> ERROR: (<3s>) ' expandMacrosWith: aField name with: aType sqlString with: msg).
				ex return: nil].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
alterField: aField newType: aType ifError: aBlock

	^self doDDLOperation: [
		self 
			execute: [self executeSQLStringNoResult: 
						(self platform createAlterColumnStatement: aField newType: aType)]
			ifError: aBlock ]! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
alterField: aField setDefault: aType

	self 
		alterField: aField 
		setDefault: aType 
		ifError: [ :ex | 	| msg|
				msg := ex messageText.
				msg isEmpty ifTrue: [msg := 'No error message provided for top level exception.'].
				Transcript cr; show: ('Alter column  <1s> to set default <2s>  ERROR: (<2s>) ' expandMacrosWith: aField name with: aType sqlString with: msg).
				ex return: nil].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
alterField: aField setDefault: anObject ifError: aBlock

	^self doDDLOperation: [
		self 
			execute: [self executeSQLStringNoResult: 
						(self platform createAlterColumnStatement: aField setDefault: anObject )]
			ifError: aBlock ]! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
alterFieldDropDefault: aField

	self 
		alterFieldDropDefault: aField 
		ifError: [ :ex | 	| msg|
				msg := ex messageText.
				msg isEmpty ifTrue: [msg := 'No error message provided for top level exception.'].
				Transcript cr; show: ('Alter column  <1s> to drop default ERROR: (<2s>) ' expandMacrosWith: aField name  with: msg).
				ex return: nil].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
alterFieldDropDefault: aField ifError: aBlock

	^self doDDLOperation: [
		self 
			execute: [self executeSQLStringNoResult: 
						(self platform createAlterColumnStatementDropDefault: aField)]
			ifError: aBlock ]! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
alterFieldDropNotNull: aField

	self 
		alterFieldDropNotNull: aField 
		ifError: [ :ex | 	| msg|
				msg := ex messageText.
				msg isEmpty ifTrue: [msg := 'No error message provided for top level exception.'].
				Transcript cr; show: ('Alter column  <1s> to drop not null ERROR: (<2s>) ' expandMacrosWith: aField name  with: msg).
				ex return: nil].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
alterFieldDropNotNull: aField ifError: aBlock

	^self doDDLOperation: [
		self 
			execute: [self executeSQLStringNoResult: 
						(self platform createAlterColumnStatementDropNotNull: aField)]
			ifError: aBlock ]! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
alterFieldSetNotNull: aField	

	self 
		alterFieldSetNotNull: aField
		ifError: [ :ex | 	| msg|
				msg := ex messageText.
				msg isEmpty ifTrue: [msg := 'No error message provided for top level exception.'].
				Transcript cr; show: ('Alter column  <1s> to set not null ERROR: (<2s>) ' expandMacrosWith: aField name  with: msg).
				ex return: nil].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
alterFieldSetNotNull: aField ifError: aBlock

	^self doDDLOperation: [
		self 
			execute: [self executeSQLStringNoResult: 
						(self platform createAlterColumnStatementSetNotNull: aField)]
			ifError: aBlock ]! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString
	self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString bindingArray: aBindingArray
	self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString returnResult: returnResult binding: aBindingArray doing: aBlock
	^self subclassResponsibility! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLStringNoResult: aString doing: aBlock
	^self basicExecuteSQLString: aString returnResult: false binding: nil doing: aBlock.! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
beReadOnly
	self denyWrites.! !

!DatabaseAccessor methodsFor: 'transactions' stamp: ''!
beginTransaction
	^self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
checkPermissionFor: command
	deniedCommands isNil ifTrue: [^self].
	(deniedCommands includes: command class)
		ifTrue: [(GlorpIllegalCommand new command: command) signal].! !

!DatabaseAccessor methodsFor: 'transactions' stamp: ''!
commitTransaction
	^self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'cursors' stamp: ''!
completedCommand: aCommand
	"Indicate that we've completed the command. Release associated resources."
	^self.! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
connection
	^connection! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
connection: aPlatformDatabaseConnection
	connection := aPlatformDatabaseConnection! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
connectionClass
	^self connectionClassForLogin: currentLogin! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
connectionClassForLogin: aLogin
	^self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'copying' stamp: ''!
copy

	^self shallowCopy postCopy.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
copyDataFromField: fromField toField: toField ifError: errorBlock

	self
		executeInDDLOperation: [ self executeSQLStringNoResult: (self platform sqlStatementToCopyDataFromColumn: fromField to: toField )]
		ifError: errorBlock! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
copyDataFromTable: fromTable toTable: anotherTable ifError: errorBlock

	self
		executeInDDLOperation: [ self executeSQLStringNoResult: (self platform sqlStatementToCopyDataFromTable: fromTable to: anotherTable) ]
		ifError: errorBlock! !

!DatabaseAccessor methodsFor: 'obsolete' stamp: ''!
createConstraint: aConstraint
	^self addConstraint: aConstraint! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
createSequence: aDatabaseSequence ifError: aBlock 

	self
		executeInDDLOperation: [ self executeSQLStringNoResult: aDatabaseSequence creationString]
		ifError: aBlock! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
createTable: aGlorpDatabaseTable ifError: aBlock 
	"This method should be used to create a database table from aTable"
	aGlorpDatabaseTable isImaginary ifTrue: [^self].
	self
		executeInDDLOperation: [
			self executeSQLStringNoResult: (self platform createTableStatementStringFor: aGlorpDatabaseTable)]
		ifError: aBlock! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
createTableFKConstraints: aGLORBDatabaseTable ifError: aBlock 
	"This method should be used to define foreign key constraints for a database table from aTable"

	aGLORBDatabaseTable foreignKeyConstraintsForCreation
		do: [ :each | self addConstraint: each ifError: aBlock ].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
createTableIndexes: aGlorpDatabaseTable ifError: aBlock 
	"Create all indexes for this table"

	self execute: 
			[(self platform createTableIndexStatementStringsFor: aGlorpDatabaseTable) 
				do: [:ea | self executeSQLStringNoResult: ea]]
		ifError: aBlock.! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
currentLogin
	^currentLogin! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
currentLogin: aLogin 
	currentLogin := aLogin! !

!DatabaseAccessor methodsFor: 'cursors' stamp: ''!
cursorFor: aStream
	^GlorpCursoredStream on: aStream accessor: self.! !

!DatabaseAccessor methodsFor: 'cursors' stamp: ''!
cursorFor: answerStream command: aCommand
	^aCommand isNil
		ifTrue: [self cursorFor: answerStream]
		ifFalse: [GlorpCursoredStream on: answerStream for: aCommand accessor: self].! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
deniedCommands
	^deniedCommands.! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
deniedCommands: aCollection
	deniedCommands := aCollection.! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
denyCommand: aCommand
	(aCommand isBehavior) ifFalse: [self error: 'invalid command type'].
	self denyCommands: (Array with: aCommand).! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
denyCommands: aCollection
	deniedCommands := aCollection asArray.! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
denyReads
	deniedCommands := SelectCommand glorpAllSubclasses.! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
denyWrites
	deniedCommands := RowBasedCommand glorpAllSubclasses.! !

!DatabaseAccessor methodsFor: 'transactions' stamp: ''!
doDDLOperation: aBlock
	"Execute DDL SQL - something which creates/drops/modifies tables or other schema information. The main reason for this method is that some databases require a transaction, while others don't require it, and still others don't allow it. Try to do the right thing."
	
	^self platform requiresTransactionForTableOperations
		ifTrue: [self inTransactionDo: aBlock]
		ifFalse: aBlock.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropConstraint: aConstraint 
	
	^self 
		dropConstraint: aConstraint 
		ifError:  [ :ex | 	| msg|
				msg := ex messageText.
				msg isEmpty ifTrue: [msg := 'no error message provided for top level exception'].
				Transcript cr; show: ('Drop constraint <1s> ERROR: (<2s>'  expandMacrosWith: aConstraint name with: msg).
				ex return: nil].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropConstraint: aConstraint ifError: aBlock
	| dropString |
	self platform supportsConstraints ifFalse: [^self].
	dropString := self platform createConstraintDropStatement: aConstraint.
	dropString isEmpty ifTrue: [^self].
	self
		executeInDDLOperation: [ self executeSQLStringNoResult: dropString ]
		ifError:aBlock! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropFields: aCollection 

	self 
		dropFields: aCollection 
		ifError: [ :ex  | 	| msg |
				msg := ex messageText.
				msg isEmpty ifTrue: [msg := 'No error message provided for top level exception.'].
				Transcript cr; show: ('Drop field ERROR: (<2s>) ' expandMacrosWith: msg).
				ex return: nil]! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropFields: aCollection ifError: aBlock

	aCollection
		do: 	[:aField | 
			self execute: [self executeSQLStringNoResult: (self platform createDropColumnStatementFor: aField)]
				ifError: aBlock ].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropForeignKeyConstraintsForTable: aTable ifError: aBlock 
	aTable isImaginary ifTrue: [^self].
	aTable  foreignKeyConstraints do: [:each | self dropConstraint: each ifError: [ :ex | aBlock value: ex]].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropSequence: aDatabaseSequence ifAbsent: aBlock 
	self
		executeInDDLOperation: [ self executeSQLStringNoResult: aDatabaseSequence dropString ]
		ifError: aBlock! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropSequences: anArray
	anArray
		do:
			[:each | 
			self
				dropSequence: each
				ifAbsent:
					[:ex | 
					Transcript show: (ex messageText ifNil: [ex printString]).
					ex return: nil]].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropTable: aTable ifError: aBlock 
	aTable isImaginary ifTrue: [^self].
	self dropForeignKeyConstraintsForTable: aTable ifError: aBlock.
	self dropTableDefinition: aTable ifError: aBlock.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropTableDefinition: aTable ifError: aBlock
	"The dropping of the primary key constraints (a singular object not a collection, despite its name) is probably not needed before dropping the table by any database platform.  If 'platform supportsConstraints' is false then it will not be done, since dropConstraint:ifError: assumes it is called on a foreign key constraint (the case in all other callers) and does nothing if the platform does not support them."

	aTable isImaginary ifTrue: [^self].
	aTable hasPrimaryKeyConstraints ifTrue: [self dropConstraint: aTable  primaryKeyConstraints ifError: [:ex | aBlock value: ex]].
	self dropTableNamed: aTable qualifiedName ifAbsent: aBlock.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropTableNamed: aString 
	self 
		dropTableNamed: aString 
		ifAbsent: [:ex | 	| msg|
				msg := ex messageText.
				msg isEmpty ifTrue: [msg := 'No error message provided for top level exception.'].
				Transcript cr; show: ('Drop table <1s> ERROR: (<2s>) ' expandMacrosWith: aString with: msg).
				ex return: nil]! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropTableNamed: aString ifAbsent: aBlock 
	self doDDLOperation: [
		self 
			execute: [self executeSQLStringNoResult: (self platform createDropTableStatementFor: aString )]
			ifError: aBlock].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
dropTables: anArray 
	anArray do: [:eachTable |
				eachTable foreignKeyConstraints 
					do: [:eachConstraint | self dropConstraint: eachConstraint]].

	anArray do: [:each | 
				self dropTableDefinition: each 
					ifError: [ :ex | 	| msg|
							msg := ex messageText.
							msg isEmpty ifTrue: [msg := 'no error message provided for top level exception'].
							Transcript cr; show: 'ERROR : (', msg, ' )'.
							ex return: nil]].! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
encoding
	"The connection is specific to the dialect;  some dialects will need to override this method."

	^connection encoding! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
encoding: aSymbol
	"The platform is a Glorp databasePlatform, so its setter code is generic across all dialects.  The connection is specific to the dialect;  some dialects will need to override how its encoding is set.  In the hope that most dialects will find this works for them, and that coders porting to dialects in which it does not will benefit from seeing the walkback and knowing they have something to do, we attempt to set both at top level."

	self platform characterEncoding: aSymbol.
	connection encoding: aSymbol.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
errorsToCatch
	^Dialect error.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
execute: aBlock

	^self execute: aBlock ifError: [:ex | ex pass].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
execute: aBlock ifError: errorBlock

	^aBlock on: self errorsToCatch do: errorBlock.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
executeCommand: command
	^self executeCommand: command returnCursor: false.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
executeCommand: command returnCursor: aBoolean
	^mutex
		critical:
			[| result |
			self checkPermissionFor: command.
			result := [(self useBinding and: [command useBinding])
				ifTrue: [command executeBoundIn: self]
				ifFalse: [command executeUnboundIn: self]]
				on: Dialect error
				do: [:ex | self handleError: ex for: command].
			aBoolean ifTrue: [result] ifFalse: [result upToEnd]].! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
executeCommandBound: command
	"By default, we don't handle binding"
	^command executeUnboundIn: self.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
executeCommandUnbound: aCommand
	| cursor |
	self logCommand: aCommand sqlString bound: false.
	cursor := self basicExecuteSQLString: aCommand sqlString.
	cursor command: aCommand.
	^cursor.! !

!DatabaseAccessor methodsFor: 'transactions' stamp: ''!
executeInDDLOperation: aBlock ifError: errorBlock
	"Execute DDL SQL - something which creates/drops/modifies tables or other schema information. The main reason for this method is that some databases require a transaction, while others don't require it, and still others don't allow it. Try to do the right thing."

	^ self doDDLOperation: [
		self
			execute: aBlock
			ifError: errorBlock ]! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
executeSQLString: aString
	| result block time |
	self log: aString.
	block := [result := (self basicExecuteSQLString: aString) upToEnd].
	time := self logging
		ifTrue: [Time millisecondsToRun: block]
		ifFalse:
			[block value.
			0].
	self logTime: time.
	^result.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
executeSQLString: aString binding: aBindingArray
	| result block time |
	self log: aString.
	block := [ 
		result := (self basicExecuteSQLString: aString binding: aBindingArray) upToEnd].
	time := self logging
		ifTrue: [Time millisecondsToRun: block]
		ifFalse:
			[block value.
			0].
	self logTime: time.
	^result.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
executeSQLStringNoResult: aString
	^self
		executeSQLStringNoResult: aString
		doing: []
		forCommand: nil.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
executeSQLStringNoResult: aString doing: aBlock forCommand: aCommand
	"Execute the string and discard the result. Evaluate aBlock before releasing the statement's answer stream"

	| block time |
	aCommand isNil ifTrue: [self log: aString] ifFalse: [self logCommand: aCommand bound: false].
	block := [self basicExecuteSQLStringNoResult: aString doing: aBlock].
	time := self logging
		ifTrue: [Time millisecondsToRun: block]
		ifFalse:
			[block value.
			0].
	self logTime: time.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
externalDatabaseErrorSignal

	self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'executing' stamp: 'GuillermoPolito 4/17/2011 18:16'!
handleError: exception for: command
	| errorClass error |
	errorClass := command isReadCommand
		ifTrue: [GlorpDatabaseReadError]
		ifFalse: [GlorpDatabaseWriteError].
	error := errorClass new.
	error command: command.
	error databaseError: (self innerExceptionFor: exception).
	error accessor: self.
	error signal = true
		ifTrue:
			[self reset.
			self logout.
			self reLogin.
			exception restart]
		ifFalse: [error return: nil].! !

!DatabaseAccessor methodsFor: 'transactions' stamp: ''!
inTransactionDo: aBlock
	"Execute aBlock in a database transaction. If one is already in progress, just continue. If not, start a transaction."

	| alreadyInTransaction result |
	[ 
	alreadyInTransaction := self isInTransaction.
	alreadyInTransaction ifFalse: [self beginTransaction].
	result := aBlock numArgs = 1 ifTrue: [aBlock value: self] ifFalse: [aBlock value].
	alreadyInTransaction ifFalse: [self commitTransaction]]
		ifCurtailed:
			[alreadyInTransaction ifFalse: [self rollbackTransaction]].
	^result.! !

!DatabaseAccessor methodsFor: 'initialize-release' stamp: ''!
initialize
	deniedCommands := nil.
	mutex := Semaphore forMutualExclusion.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
innerExceptionFor: exception
	^exception parameter.! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
insertOnly
	self denyCommands: (Array with: UpdateCommand with: DeleteCommand).! !

!DatabaseAccessor methodsFor: 'cursors' stamp: ''!
isCursorAtEnd: aCursor
	"Return true if aCursor is at its end. This implementation may not work for all, but it seems like a reasonable default"

	^aCursor atEnd.! !

!DatabaseAccessor methodsFor: 'transactions' stamp: ''!
isInTransaction
	^self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'logging' stamp: ''!
log: aStringOrBlock
	"Log an arbitrary string"
	| string |
	self logging ifFalse: [^self].
	string := aStringOrBlock isString
		ifTrue: [aStringOrBlock]
		ifFalse: [aStringOrBlock value].
	string size > 4000 ifTrue: [string := string copyFrom: 1 to: 4000].
	Transcript
		cr;
		show: string.! !

!DatabaseAccessor methodsFor: 'logging' stamp: ''!
logCommand: aDatabaseCommand bound: aBoolean

	"Log the execution of a command"
	| string |
	self changed: #command: with: aDatabaseCommand.
	self logging ifFalse: [^self].
	string := aDatabaseCommand sqlString.
 	aBoolean ifTrue: [string := string , '  ' , aDatabaseCommand bindings printString].
	string size > 4000 ifTrue: [string := string copyFrom: 1 to: 4000].
	Transcript
		cr;
		show: string.! !

!DatabaseAccessor methodsFor: 'logging' stamp: ''!
logError: anErrorObject 
	self log: anErrorObject printString! !

!DatabaseAccessor methodsFor: 'logging' stamp: ''!
logTime: aTimeOrBlock
	| time |
	self changed: #time: with: aTimeOrBlock.
	self logging ifFalse: [^self].
	time := (Dialect isBlock: aTimeOrBlock)
		ifTrue: [aTimeOrBlock value]
		ifFalse: [aTimeOrBlock].
	Transcript
		cr;
		show: '(' , (time / 1000.0) printString , ' s)'.! !

!DatabaseAccessor methodsFor: 'logging' stamp: ''!
logging
	logging isNil ifTrue: [^self class loggingEnabled].
	^logging.! !

!DatabaseAccessor methodsFor: 'logging' stamp: ''!
logging: aBoolean 
	logging := aBoolean! !

!DatabaseAccessor methodsFor: 'login' stamp: ''!
login
	
	| |
	self loginIfError: [:ex | 
		ex pass].
	"Just to help avoid confusion if someone thinks they're getting a login object back from this"
	^nil.! !

!DatabaseAccessor methodsFor: 'login' stamp: ''!
loginIfError: aBlock

	self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'login' stamp: ''!
logout
	^self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'notification' stamp: ''!
myDependents
	^dependents.! !

!DatabaseAccessor methodsFor: 'notification' stamp: ''!
myDependents: aCollection
	dependents := aCollection.! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
numberOfPreparedStatements
	^0! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
permitEverything
	"Note that nil is treated as a flag indicating that all commands are allowed"
	deniedCommands := nil.! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
permitNothing
	deniedCommands := DatabaseCommand glorpAllSubclasses asArray.! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
permitNothingDuring: aBlock
	| oldCommands |
	oldCommands := self deniedCommands.
	self permitNothing.
	aBlock ensure: [deniedCommands := oldCommands].! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
permitsInserts
	deniedCommands isNil ifTrue: [^true].
	^(deniedCommands includes: InsertCommand) not.! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
permitsUpdates
	deniedCommands isNil ifTrue: [^true].
	^(deniedCommands includes: UpdateCommand) not.! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
platform
	^currentLogin database.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
populateField: aDatabaseField with: aValue ifError: errorBlock

	self
		executeInDDLOperation: [
			self executeSQLStringNoResult: (self platform sqlStatementToPopulateColumn: aDatabaseField with: aValue) ]
		ifError: errorBlock! !

!DatabaseAccessor methodsFor: 'copying' stamp: ''!
postCopy! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
renameField: aDatabaseField to: nameString ifError: errorBlock

	self
		executeInDDLOperation: [ self executeSQLStringNoResult: (self platform sqlStatementToRenameColumn: aDatabaseField to: nameString) ]
		ifError: errorBlock! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
renameTable: aDatabaseTable to: newTableName inPlatform: aDatabasePlatform ifError: errorBlock

	| sqlStatement |
	sqlStatement := aDatabasePlatform sqlStatementToRenameTable: aDatabaseTable to: newTableName.	
	self
		executeInDDLOperation: [ self executeSQLStringNoResult: sqlStatement ]
		ifError: errorBlock! !

!DatabaseAccessor methodsFor: 'initialize-release' stamp: ''!
reset
	"Don't reset the permissions, because the replicator wants to reset sessions without changing their characteristics"
"	self permitEverything."! !

!DatabaseAccessor methodsFor: 'permissions' stamp: ''!
resetPermissions
	self permitEverything.! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
returnRowCount: aBoolean
	^self.! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
reusePreparedStatements
	^reusePreparedStatements and: [self useBinding].! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
reusePreparedStatements: aBoolean
	reusePreparedStatements := aBoolean.! !

!DatabaseAccessor methodsFor: 'transactions' stamp: ''!
rollbackTransaction
	^self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'executing' stamp: ''!
rowCount
	^self subclassResponsibility! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
schema

	^ self currentLogin schema! !

!DatabaseAccessor methodsFor: 'login' stamp: ''!
showDialog: aString

	self subclassResponsibility.! !

!DatabaseAccessor methodsFor: 'accessing' stamp: ''!
useBinding
	^self platform useBinding.! !

DatabaseAccessor subclass: #DolphinDatabaseAccessor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!DolphinDatabaseAccessor commentStamp: '<historical>' prior: 0!
The database accessor subclass for Dolphin Smalltalk.!

!DolphinDatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString
	| resultSet array |
	self log: aString.
	resultSet := connection query: aString.	"Write statements don't like asArray if there's not result. DDL statements don't even like the isEmpty test"
	array := [resultSet isEmpty ifTrue: [#()] ifFalse: [resultSet asArray]]
		on: Dialect error
		do: [:ex | ex return: #()].
	resultSet
		close;
		free.
	^ReadStream on: array read.! !

!DolphinDatabaseAccessor methodsFor: 'transactions' stamp: ''!
beginTransaction

	connection beginRWTxn.! !

!DolphinDatabaseAccessor methodsFor: 'transactions' stamp: ''!
commitTransaction
	self log: 'Commit Transaction'.
	connection commitTxn! !

!DolphinDatabaseAccessor methodsFor: 'executing' stamp: ''!
externalDatabaseErrorSignal

	^Smalltalk at: #DBError! !

!DolphinDatabaseAccessor methodsFor: 'transactions' stamp: ''!
isInTransaction

	^connection isInTransaction! !

!DolphinDatabaseAccessor methodsFor: 'accessing' stamp: ''!
isLoggedIn

	^connection notNil.! !

!DolphinDatabaseAccessor methodsFor: 'login' stamp: ''!
loginIfError: aBlock

	self
		execute:
			[connection := ((Smalltalk at: #DBConnection) new)
				dsn: currentLogin connectString;
				uid: currentLogin username;
				pwd: currentLogin password;
				connect.]
		ifError: aBlock! !

!DolphinDatabaseAccessor methodsFor: 'login' stamp: ''!
logout

	connection notNil
		ifTrue:
			[[connection close.] ensure: [connection := nil]].! !

!DolphinDatabaseAccessor methodsFor: 'transactions' stamp: ''!
rollbackTransaction
	self log: 'Rollback Transaction'.
	connection rollbackTxn! !

!DolphinDatabaseAccessor methodsFor: 'login' stamp: ''!
showDialog: aString

	(Smalltalk at: #MessageBox) warning: aString.! !

DatabaseAccessor subclass: #ObjectStudioDatabaseAccessor
	instanceVariableNames: 'inTransaction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!ObjectStudioDatabaseAccessor commentStamp: '<historical>' prior: 0!
An accessor for ObjectStudio.

Instance Variables:
	inTransaction	<Boolean>	Are we in a transaction. We don't have an indicator for this from the database layer, so keep track of it here.

!

!ObjectStudioDatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString
	| result cursor records |
	result := connection execSql: aString.
	records := result records.
	"Fix up nulls"
	records do: [:each |
		1 to: each size do: [:i |
			(each at: i) == #'(NULL)' ifTrue: [each at: i put: nil]]].
	cursor := self cursorFor: (ReadStream on: result records).
	cursor statement: result.
	^cursor.! !

!ObjectStudioDatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString bindingArray: aBindingArray
	" *** This method was defined by Glorp.DatabaseAccessor as a subclass responsibility.
	Replace its body with a proper implementation. *** "

	self error: 'Subclass responsibility stub not reimplemented'! !

!ObjectStudioDatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString returnResult: returnResult binding: aBindingArray doing: aBlock

	^connection execSql: aString.! !

!ObjectStudioDatabaseAccessor methodsFor: 'transactions' stamp: ''!
beginTransaction
	inTransaction := true.
	^connection beginTran.! !

!ObjectStudioDatabaseAccessor methodsFor: 'transactions' stamp: ''!
commitTransaction
	inTransaction := false.
	^connection commit.! !

!ObjectStudioDatabaseAccessor methodsFor: 'accessing' stamp: ''!
connectionClassForLogin: aLogin
	aLogin database class == OraclePlatform
		ifTrue: [^Smalltalk at: #OracleOCIDatabase ifAbsent: [self error: 'no such driver']].
	aLogin database isODBCPlatform ifTrue: [^Smalltalk at: #ODBCDatabase].
	self error: 'Unknown database: ' , aLogin database name.! !

!ObjectStudioDatabaseAccessor methodsFor: 'executing' stamp: ''!
externalDatabaseErrorSignal
	^Error.! !

!ObjectStudioDatabaseAccessor methodsFor: 'initialize-release' stamp: ''!
initialize
	inTransaction := false.! !

!ObjectStudioDatabaseAccessor methodsFor: 'executing' stamp: ''!
isInTransaction
	^inTransaction.! !

!ObjectStudioDatabaseAccessor methodsFor: 'login' stamp: ''!
isLoggedIn

	connection isNil ifTrue: [^false].
	^connection isActive! !

!ObjectStudioDatabaseAccessor methodsFor: 'login' stamp: ''!
loginIfError: aBlock
	| |
"	connection :=self connectionClass 
		name: currentLogin connectString
		alias: currentLogin connectString
		password: currentLogin password
		drive: 'c'
		comment: currentLogin connectString
		create: false."
	connection := self connectionClass
		logOnServer: currentLogin connectString
		 user: currentLogin username
		 password: currentLogin password
		 alias: currentLogin connectString.
	(connection isKindOf: Message)
		ifTrue: [^self errorBoxMessage: connection asString].
	connection start.! !

!ObjectStudioDatabaseAccessor methodsFor: 'login' stamp: ''!
logout
	connection class logOff.! !

!ObjectStudioDatabaseAccessor methodsFor: 'executing' stamp: ''!
releaseCursor: aCursor in: aPreparedStatement! !

!ObjectStudioDatabaseAccessor methodsFor: 'login' stamp: ''!
rollbackTransaction
	inTransaction := false.
	^connection rollback.! !

!ObjectStudioDatabaseAccessor methodsFor: 'executing' stamp: ''!
rowCount
	^connection rowsProcessed.! !

!ObjectStudioDatabaseAccessor methodsFor: 'initialize-release' stamp: ''!
showDialog: aString
	" *** This method was defined by Glorp.DatabaseAccessor as a subclass responsibility.
	Replace its body with a proper implementation. *** "

	self error: 'Subclass responsibility stub not reimplemented'! !

DatabaseAccessor subclass: #PharoDatabaseAccessor
	instanceVariableNames: 'isInTransaction databaseDriver'
	classVariableNames: 'DefaultDriver'
	poolDictionaries: ''
	category: 'Glorp-Database'!

!PharoDatabaseAccessor classSide methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/16/2011 23:13'!
DefaultDriver
	^DefaultDriver! !

!PharoDatabaseAccessor classSide methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/16/2011 23:13'!
DefaultDriver: aDriver
	DefaultDriver := aDriver! !

!PharoDatabaseAccessor methodsFor: 'executing' stamp: 'GuillermoPolito 4/16/2011 23:17'!
basicExecuteSQLString: aString
	| result rowCollection |
	self log: aString.
	result := self connection execute: aString.
	result errorResponse notNil
		ifTrue:
			[self externalDatabaseErrorSignal signal: result errorResponse value].
	rowCollection := OrderedCollection new.
	result rows do: [:ea | rowCollection add: ea data asArray].
 	^rowCollection asArray! !

!PharoDatabaseAccessor methodsFor: 'executing' stamp: 'GuillermoPolito 4/16/2011 22:38'!
basicExecuteSQLStringNoResult: aString doing: aBlock

	| a |
	a := self basicExecuteSQLString: aString.
	aBlock value.
	^a! !

!PharoDatabaseAccessor methodsFor: 'transactions' stamp: 'GuillermoPolito 4/16/2011 23:14'!
beginTransaction
	self log: 'Begin Transaction'.
	isInTransaction := true.
	self databaseDriver beginTransaction! !

!PharoDatabaseAccessor methodsFor: 'transactions' stamp: 'GuillermoPolito 4/16/2011 23:23'!
commitTransaction
	self log: 'Commit Transaction'.
	isInTransaction := false.
	self databaseDriver commitTransaction! !

!PharoDatabaseAccessor methodsFor: 'executing' stamp: 'GuillermoPolito 4/16/2011 23:15'!
connection
	self databaseDriver ifNil: [ ^nil ].
	^self databaseDriver connection! !

!PharoDatabaseAccessor methodsFor: 'login' stamp: 'GuillermoPolito 4/16/2011 22:38'!
connectionArgsFromCurrentLogin

	"connect string looks like  hostname:portNumber_db-name. :portNumber is optional"
	| str portIndex index dbIndex |
	str := self currentLogin connectString.
	portIndex := index := str indexOf: $:.
	dbIndex := str indexOf: $_.
	index = 0 ifTrue: [index := dbIndex].
	^(Dialect smalltalkAt: #PGConnectionArgs)
		hostname: (str copyFrom: 1 to: index - 1)
		portno: (portIndex = 0 ifTrue: [5432] ifFalse: [(str copyFrom: portIndex + 1 to: dbIndex - 1) asNumber])
		databaseName: (str copyFrom: dbIndex + 1 to: str size)
		userName: self currentLogin username
		password: self currentLogin password! !

!PharoDatabaseAccessor methodsFor: 'login' stamp: 'GuillermoPolito 4/16/2011 23:19'!
connectionClassForLogin: aLogin 
	^ self class DefaultDriver.! !

!PharoDatabaseAccessor methodsFor: 'login' stamp: 'GuillermoPolito 4/16/2011 23:20'!
databaseDriver
	^ databaseDriver! !

!PharoDatabaseAccessor methodsFor: 'login' stamp: 'GuillermoPolito 4/16/2011 23:20'!
databaseDriver: aDriver
	databaseDriver := aDriver! !

!PharoDatabaseAccessor methodsFor: 'executing' stamp: 'GuillermoPolito 4/16/2011 23:16'!
disconnect
	^ self connection disconnect! !

!PharoDatabaseAccessor methodsFor: 'executing' stamp: 'GuillermoPolito 4/16/2011 22:38'!
executeCommandUnbound: aCommand
	| answerStream |
	self log: aCommand sqlString.
	answerStream := ReadStream on: (self basicExecuteSQLString: aCommand sqlString).
	aCommand needsRowCount
		ifTrue:
			[aCommand
				rowCount:
					(self platform supportsRowCount
						ifTrue: [answerStream rowCount]
						ifFalse: [1])].
	^self cursorFor: answerStream command: aCommand.! !

!PharoDatabaseAccessor methodsFor: 'executing' stamp: 'GuillermoPolito 4/16/2011 22:38'!
externalDatabaseErrorSignal
	^ Dialect error! !

!PharoDatabaseAccessor methodsFor: 'initialize' stamp: 'GuillermoPolito 4/16/2011 22:38'!
initialize
	super initialize.
	isInTransaction := false! !

!PharoDatabaseAccessor methodsFor: 'executing' stamp: 'GuillermoPolito 4/16/2011 22:38'!
innerExceptionFor: exception
	"I don't know how to get the inner error in Squeak"
	^exception messageText.! !

!PharoDatabaseAccessor methodsFor: 'transactions' stamp: 'GuillermoPolito 4/16/2011 22:38'!
isInTransaction
	^ isInTransaction! !

!PharoDatabaseAccessor methodsFor: 'login' stamp: 'GuillermoPolito 4/16/2011 23:21'!
isLoggedIn
	self databaseDriver ifNil: [^ false].
	^ self databaseDriver isConnected! !

!PharoDatabaseAccessor methodsFor: 'login' stamp: 'GuillermoPolito 4/16/2011 23:21'!
loginIfError: aBlock 
	self log: 'Login'.
	self databaseDriver: self connectionClass new.
	self
		execute: [self databaseDriver connect: currentLogin]
		ifError: aBlock.
	self log: 'Login finished'! !

!PharoDatabaseAccessor methodsFor: 'login' stamp: 'GuillermoPolito 4/16/2011 23:21'!
logout
	self isLoggedIn ifFalse: [^self].
	self log: 'Logout'.
	self databaseDriver logout.
	self log: 'Logout finished'! !

!PharoDatabaseAccessor methodsFor: 'executing' stamp: 'GuillermoPolito 4/16/2011 22:38'!
releaseCursor: aCursor in: aPreparedStatement
	"Release the cursor. We're trapping exceptions, so return true or false to indicate success/failure"
	aCursor isNil ifTrue: [^self].
	^[aCursor close. true] on: Dialect error do: [:ex | ex return: false].! !

!PharoDatabaseAccessor methodsFor: 'transactions' stamp: 'GuillermoPolito 4/16/2011 23:23'!
rollbackTransaction
	self log: 'Rollback Transaction'.
	isInTransaction := false.
	self databaseDriver rollbackTransaction! !

!PharoDatabaseAccessor methodsFor: 'executing' stamp: 'GuillermoPolito 4/16/2011 23:18'!
rowCount
	^ self databaseDriver rowCount.! !

!PharoDatabaseAccessor methodsFor: 'login' stamp: 'GuillermoPolito 4/16/2011 23:22'!
showDialog: aString 
	UIManager default inform: aString! !

DatabaseAccessor subclass: #SqueakDatabaseAccessor
	instanceVariableNames: 'isInTransaction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!SqueakDatabaseAccessor commentStamp: '<historical>' prior: 0!
A database accessor for Squeak.

Instance Variables:
	isInTransaction	<Boolean>	Are we in a transaction. We don't have an indicator for that from the database layer, so keep track of it here.

!

!SqueakDatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString
	| result rowCollection |
	self log: aString.
	result := connection execute: aString.
	result errorResponse notNil
		ifTrue:
			[self externalDatabaseErrorSignal signal: result errorResponse value].
	rowCollection := OrderedCollection new.
	result rows do: [:ea | rowCollection add: ea data asArray].
 	^rowCollection asArray! !

!SqueakDatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLStringNoResult: aString doing: aBlock

	| a |
	a := self basicExecuteSQLString: aString.
	aBlock value.
	^a! !

!SqueakDatabaseAccessor methodsFor: 'transactions' stamp: ''!
beginTransaction
	self log: 'Begin Transaction'.
	isInTransaction := true.
	connection execute: 'BEGIN TRANSACTION'! !

!SqueakDatabaseAccessor methodsFor: 'transactions' stamp: ''!
commitTransaction
	self log: 'Commit Transaction'.
	isInTransaction := false.
	connection execute: 'COMMIT TRANSACTION'! !

!SqueakDatabaseAccessor methodsFor: 'login' stamp: ''!
connectionArgsFromCurrentLogin

	"connect string looks like  hostname:portNumber_db-name. :portNumber is optional"
	| str portIndex index dbIndex |
	str := self currentLogin connectString.
	portIndex := index := str indexOf: $:.
	dbIndex := str indexOf: $_.
	index = 0 ifTrue: [index := dbIndex].
	^(Dialect smalltalkAt: #PGConnectionArgs)
		hostname: (str copyFrom: 1 to: index - 1)
		portno: (portIndex = 0 ifTrue: [5432] ifFalse: [(str copyFrom: portIndex + 1 to: dbIndex - 1) asNumber])
		databaseName: (str copyFrom: dbIndex + 1 to: str size)
		userName: self currentLogin username
		password: self currentLogin password! !

!SqueakDatabaseAccessor methodsFor: 'login' stamp: ''!
connectionClassForLogin: aLogin 
	aLogin database class == PostgreSQLPlatform
		ifTrue: [^ Smalltalk at: #PGConnection].
	self error: 'Unknown database: ' , aLogin database! !

!SqueakDatabaseAccessor methodsFor: 'executing' stamp: ''!
disconnect
	^ connection disconnect! !

!SqueakDatabaseAccessor methodsFor: 'executing' stamp: ''!
executeCommandUnbound: aCommand
	| answerStream |
	self log: aCommand sqlString.
	answerStream := ReadStream on: (self basicExecuteSQLString: aCommand sqlString).
	aCommand needsRowCount
		ifTrue:
			[aCommand
				rowCount:
					(self platform supportsRowCount
						ifTrue: [answerStream rowCount]
						ifFalse: [1])].
	^self cursorFor: answerStream command: aCommand.! !

!SqueakDatabaseAccessor methodsFor: 'executing' stamp: ''!
externalDatabaseErrorSignal
	^ Dialect error! !

!SqueakDatabaseAccessor methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	isInTransaction := false! !

!SqueakDatabaseAccessor methodsFor: 'executing' stamp: ''!
innerExceptionFor: exception
	"I don't know how to get the inner error in Squeak"
	^exception messageText.! !

!SqueakDatabaseAccessor methodsFor: 'transactions' stamp: ''!
isInTransaction
	^ isInTransaction! !

!SqueakDatabaseAccessor methodsFor: 'login' stamp: ''!
isLoggedIn
	connection isNil
		ifTrue: [^ false].
	^ connection isConnected! !

!SqueakDatabaseAccessor methodsFor: 'login' stamp: ''!
loginIfError: aBlock 
	self log: 'Login'.
	connection := self connectionClass new.
	connection connectionArgs: self connectionArgsFromCurrentLogin.
	self
		execute: [connection startup]
		ifError: aBlock.
	self log: 'Login finished'! !

!SqueakDatabaseAccessor methodsFor: 'login' stamp: ''!
logout
	self isLoggedIn ifFalse: [^self].
	self log: 'Logout'.
	self execute: [connection terminate].
	self log: 'Logout finished'! !

!SqueakDatabaseAccessor methodsFor: 'executing' stamp: ''!
releaseCursor: aCursor in: aPreparedStatement
	"Release the cursor. We're trapping exceptions, so return true or false to indicate success/failure"
	aCursor isNil ifTrue: [^self].
	^[aCursor close. true] on: Dialect error do: [:ex | ex return: false].! !

!SqueakDatabaseAccessor methodsFor: 'transactions' stamp: ''!
rollbackTransaction
	self log: 'Rollback Transaction'.
	isInTransaction := false.
	connection execute: 'ROLLBACK TRANSACTION'! !

!SqueakDatabaseAccessor methodsFor: 'executing' stamp: ''!
rowCount

	|count|
	count := 0.
	connection result resultSets
		do: [:each| count := count + 
			(each completedResponse value subStrings: (Array with: Character space)) last asInteger].
	^count! !

!SqueakDatabaseAccessor methodsFor: 'login' stamp: ''!
showDialog: aString 
	"Do 'Smalltalk at: #PopUpMenu' so that the code will compile in other dialects, even if it won't run"

	(Smalltalk at: #PopUpMenu) inform: aString! !

DatabaseAccessor subclass: #VADatabaseAccessor
	instanceVariableNames: 'isInTransaction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

VADatabaseAccessor class
	instanceVariableNames: 'databaseErrorSignal'!

!VADatabaseAccessor commentStamp: '<historical>' prior: 0!
An accessor for VisualAge 8.0 and higher. 

Instance Variables
	isInTransaction	<Boolean>	Are we in a transaction. We don't have an indicator for this from the database layer, so keep track of it here.


!

VADatabaseAccessor class
	instanceVariableNames: 'databaseErrorSignal'!

!VADatabaseAccessor classSide methodsFor: 'accessing' stamp: ''!
externalDatabaseErrorSignal

	databaseErrorSignal isNil ifTrue: [self initializeDatabaseErrorSignal ].
	^databaseErrorSignal! !

!VADatabaseAccessor classSide methodsFor: 'initialization' stamp: ''!
initializeDatabaseErrorSignal

	databaseErrorSignal := ((Smalltalk at: #SystemExceptions) at: 'ExError') newChild.
	databaseErrorSignal markReadOnly: false;
		description: 'Database error';
		resumable: true;
		defaultHandler: nil;
		markReadOnly: true! !

!VADatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString
	^self basicExecuteSQLString: aString binding: nil! !

!VADatabaseAccessor methodsFor: 'As yet unclassified' stamp: ''!
basicExecuteSQLString: aString binding: aBindingArray
	"VA does not return results for as many statements as VW.  Some calls that take returnResult: true in VW
do not want to return a result in the basic VA RDB framework that we call.  We therefore check the String and
only ask for a return if it is a select statement.  (This is experimental and could be reviewed.)"

	^self
		basicExecuteSQLString: aString
		returnResult: (aString copyFrom: 1 to: (6 min: aString size)) asUppercase = 'SELECT'
		binding: aBindingArray
		doing: []! !

!VADatabaseAccessor methodsFor: 'As yet unclassified' stamp: ''!
basicExecuteSQLString: aString returnResult: returnResult binding: aBindingArray doing: aBlock
	"^<GlorpCursoredStream | empty readStream> This method executes a general SQL command against the connection.
If a result is requested, it returns a GlorpCursoredStream wrapping an AbtResultTable that (should) hold an AbtCursor.  If
not, it returns self to cursorFor: which returns an empty readStream since along one calling route VA does not provide a
result if the string is not a select whereas VW does in general, and the caller sends upToEnd.
	Later, develop to use GlorpPreparedStatements (c.f. this method in VW) and resultTableFromQuerySpec:withValues:...
VW creates an SQLStringCommand to put in the cursor."

	^self cursorFor:  (returnResult		"was check if aString was SELECT statement, now moved to a caller"
		ifFalse:
			[connection
				executeSQLStatement: aString
				ifError: [:err | GlorpError signalWith: err].
			aBlock value.
			self]					"must return self to be handled in cursorFor:"
		ifTrue:
			[self basicExecuteSQLStringWithResult: aString doing: aBlock])! !

!VADatabaseAccessor methodsFor: 'As yet unclassified' stamp: ''!
basicExecuteSQLStringNoResult: aString doing: aBlock
	^self basicExecuteSQLString: aString returnResult: false binding: nil doing: aBlock.! !

!VADatabaseAccessor methodsFor: 'As yet unclassified' stamp: ''!
basicExecuteSQLStringWithResult: aString doing: aBlock
	"^<AbtResultTable | self> This VA-specific method returns an AbtResultTable unless it fails.  Later,
add a binding: parameter and call resultTableFromQuerySpec:withValues:ifError:"

	| resultTable |
	resultTable := connection
		resultTableFromQuerySpec: ((Smalltalk at: #AbtQuerySpec) new statement: aString)
		ifError: [:err | GlorpError signalWith: err].
	^resultTable isAbtError
		ifTrue: 
			[GlorpError signalWith: resultTable.
			#() readStream]
		ifFalse:
			[aBlock value.
			resultTable]! !

!VADatabaseAccessor methodsFor: 'transactions' stamp: ''!
beginTransaction
	self log: 'Begin Transaction'.
	isInTransaction := true.
	self connection beginUnitOfWork.! !

!VADatabaseAccessor methodsFor: 'transactions' stamp: ''!
commitTransaction
	self log: 'Commit Transaction'.
	isInTransaction := false.
	self connection commitUnitOfWorkIfError: 
			[:err |
			self log: 'Commit Transaction failed.'. 
			self connection autoCommit: true.
			self todo: 'Need to handle this error somehow....'.
			"errorBlock value: err"].! !

!VADatabaseAccessor methodsFor: 'login' stamp: ''!
connectionClassForLogin: aLogin 
	aLogin database isDB2Platform ifTrue: [^ Smalltalk at: #AbtIbmCliDatabaseManager].
	aLogin database class == OraclePlatform 
		ifTrue: [^Smalltalk at: #AbtOracle10DatabaseManager].
	aLogin database isODBCPlatform 
		ifTrue: [^Smalltalk at: #AbtOdbcDatabaseManager].
	aLogin database isPostgreSQLPlatform 
		ifTrue: [self error: 'PostgreSQL is not yet supported under VA'].
	self error: 'Unknown database platform' , aLogin database printString! !

!VADatabaseAccessor methodsFor: 'cursors' stamp: ''!
cursorFor: aStream
	"basicExecuteSQLString: binding: sets returnResult: to false if aString does not begin with 'SELECT' but
its non-VA-specific caller executeSQLString:binding: sends upToEnd to what it returns, expecting the more
usual behaviour of returnResult always being true on that branch.  Thu we return an empty readStream, like
the VW implementation of this method in other no-return cases."

	aStream == self ifTrue: [^#() readStream].
	aStream onFinalizeDo: #close.
	"In V8Beta, the finalize cycle has not always been run by the time the database' databaseMgr has been nilled,
	causing a DNU when trying to get the error block (that it will not run).  If this persists, replace #close above with
		Message selector: #closeIfError: arguments: (Array with: aStream errorBlock)
	so we get the block when we know we can."
	^super cursorFor: aStream.! !

!VADatabaseAccessor methodsFor: 'As yet unclassified' stamp: ''!
databaseName
	"connect string looks like  hostname:portNumber_db-name. :portNumber is optional.
	If the string is zero, return nil."

	| str dbIndex |
	str := self currentLogin connectString.
	dbIndex := str indexOf: $_.
	^str copyFrom: dbIndex + 1 to: str size! !

!VADatabaseAccessor methodsFor: 'executing' stamp: ''!
executeCommandUnbound: aCommand
	| answerStream |
	self log: aCommand sqlString.	"change to self logCommand: aCommand sqlString bound: false." 
	answerStream := self basicExecuteSQLStringWithResult: aCommand sqlString doing: [].
	aCommand needsRowCount
		ifTrue:
			[aCommand
				rowCount:
					(self platform supportsRowCount
						ifTrue: [answerStream rowCount]
						ifFalse: [1])].
	^self cursorFor: answerStream command: aCommand.! !

!VADatabaseAccessor methodsFor: 'executing' stamp: ''!
externalDatabaseErrorSignal

	^self class externalDatabaseErrorSignal.! !

!VADatabaseAccessor methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	isInTransaction := false.! !

!VADatabaseAccessor methodsFor: 'executing' stamp: ''!
innerExceptionFor: exception
	^exception exception.! !

!VADatabaseAccessor methodsFor: 'cursors' stamp: ''!
isCursorAtEnd: aCursor
	"Return true if aCursor is at its end."

	^aCursor actuallyAtEnd.! !

!VADatabaseAccessor methodsFor: 'transactions' stamp: ''!
isInTransaction
	^isInTransaction! !

!VADatabaseAccessor methodsFor: 'login' stamp: ''!
isLoggedIn

	^connection notNil! !

!VADatabaseAccessor methodsFor: 'login' stamp: ''!
logStream
	^Transcript! !

!VADatabaseAccessor methodsFor: 'login' stamp: ''!
loginIfError: aBlock 
	"
		This method creates a NEW database connection.
	"

	| dbMgrClass anAbtDatabaseLogonSpec anAbtDatabaseConnectionSpec aliasNameString baseAliasName |
	self log: 'Login'.
	aliasNameString := baseAliasName := currentLogin connectString.

	"
		We try to generate a unique alias name to get a new connection and to 
		prevent the normal VAST behaviour to reuse a connection
	"
	
	[((Smalltalk at: #AbtDbmSystem) 
		activeDatabaseConnectionWithAlias: aliasNameString) notNil] 
			whileTrue: 
				[aliasNameString := baseAliasName , Time millisecondClockValue printString].


	"
		Get the native VA class responsible for doing the work against the
		three special connections available in VA
	"
	dbMgrClass := self connectionClass.

	"
		Get the logon specification for the database ... do not use
		any server information
	"
	anAbtDatabaseLogonSpec := (Smalltalk at: #AbtDatabaseLogonSpec) 
				id: currentLogin username
				password: currentLogin password
				server: self serverName.

	"	
		Get the specification for the connection ... 
   "
	anAbtDatabaseConnectionSpec := (Smalltalk at: #AbtDatabaseConnectionSpec) 
				forDbmClass: dbMgrClass
				dataSourceName: self databaseName.
	connection := anAbtDatabaseConnectionSpec 
				connectUsingAlias: aliasNameString
				logonSpec: anAbtDatabaseLogonSpec
				ifError: 
					[:error | 
					"throw away the connection ..."
					connection := nil.
					aBlock value: error].
	connection isNil ifFalse: [
		connection databaseMgr errorBlock: [:err | self externalDatabaseErrorSignal signalWith: err]].! !

!VADatabaseAccessor methodsFor: 'login' stamp: ''!
logout
	self isLoggedIn ifFalse: [^self].
	self log: 'Logout'.
	"disconnectIfError: doesn't properly catch errors."
	[
	[connection disconnectIfError: [:ex | self logError: ex]]
		ensure:	[connection := nil]]
		  on: Dialect error
		  do: [:ex | ex return: nil].
	self log: 'Logout finished'! !

!VADatabaseAccessor methodsFor: 'cursors' stamp: ''!
releaseCursor: aCursor in: aPreparedStatement
	aCursor close.! !

!VADatabaseAccessor methodsFor: 'accessing' stamp: ''!
returnRowCount: aBoolean

	connection isNil ifFalse: [connection queryRowCount: true].! !

!VADatabaseAccessor methodsFor: 'transactions' stamp: ''!
rollbackTransaction
	self log: 'Rollback Transaction'.
	isInTransaction := false.
	self connection rollbackUnitOfWork.! !

!VADatabaseAccessor methodsFor: 'accessing' stamp: ''!
rowCount
	^connection rowCount.! !

!VADatabaseAccessor methodsFor: 'As yet unclassified' stamp: ''!
serverName
	"connect string looks like  hostname:portNumber_db-name. The :portNumber is optional;
	for the moment, we ignore it.  If the hostName is empty, return nil."

	| str dbIndex "portIndex index" |
	str := self currentLogin connectString.
"	portIndex := index := str indexOf: $:."
	dbIndex := str indexOf: $_.
"	index = 0 ifTrue: [index := dbIndex]."
	^dbIndex = 0 ifFalse: [str copyFrom: 1 to: dbIndex - 1]! !

!VADatabaseAccessor methodsFor: 'login' stamp: ''!
showDialog: aString

	(Smalltalk at: #CwMessagePrompter) warn: aString.! !

!VADatabaseAccessor methodsFor: 'development' stamp: ''!
todo: aString

	^self! !

DatabaseAccessor subclass: #VWDatabaseAccessor
	instanceVariableNames: 'driverSession preparedStatements'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!VWDatabaseAccessor commentStamp: '<historical>' prior: 0!
An accessor for VisualWorks EXDI connections.

Instance Variables:
	driverSession	<ExternalDatabaseSession>	If we aren't re-using prepared statements, then we work directly with a single EXDI session that we keep hold of and reuse for each command.
	preparedStatements	<CacheManager>	A cache of prepared statements that we can re-use, if the session is set appropriately.

!

!VWDatabaseAccessor classSide methodsFor: 'exception handling' stamp: ''!
errorsToCatch
	| pgError |
	pgError := Smalltalk at: #PostgreSQLEXDIException ifAbsent: [^Error].
	^Error, pgError.! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString
	^self basicExecuteSQLString: aString binding: nil.! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString binding: aBindingArray 
	^self basicExecuteSQLString: aString returnResult: true binding: aBindingArray doing: [].! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLString: aString returnResult: returnResult binding: aBindingArray doing: aBlock
	"Duplicate some of the logic of command execution, primarily because of the need to do special treatment depending if we want a result or not, and to execute the block. Don't cache the prepared statement, but do put it into the driverSesssion (ugh) so that we can check the rowCount if necessary"
	| answerStream command statement session |
	answerStream := nil.
	command := SQLStringCommand new.
	command setSQLString: aString.
	command parameters: aBindingArray.
	(aBindingArray isNil or: [aBindingArray class ~= Association and: [aBindingArray isEmpty]]) ifTrue: [command useBinding: false].
	statement := self newPreparedStatementFor: command.
	session := statement statement.
	driverSession := session.
	command useBinding ifTrue: [
		session bindInput: aBindingArray].
	session execute.
	"Always do this, because at least with ODBC, if we don't ask for the answer, we don't get errors, and subsequent statements get them"
	answerStream := session answer.
	answerStream == #noAnswerStream ifTrue: [answerStream := ReadStream on: #()].
	answerStream == #noMoreAnswers ifTrue: [answerStream := ReadStream on: #()].
	aBlock value.
	returnResult 
		ifTrue: [ | cursor |
			cursor := self cursorFor: answerStream command: command.
			cursor statement: statement.
			^cursor]
		ifFalse: [session disconnect].! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
basicExecuteSQLStringNoResult: aString doing: aBlock
	^self basicExecuteSQLString: aString returnResult: false binding: nil doing: aBlock.! !

!VWDatabaseAccessor methodsFor: 'transactions' stamp: ''!
beginTransaction
	self log: 'Begin Transaction'.
	connection begin! !

!VWDatabaseAccessor methodsFor: 'transactions' stamp: ''!
commitTransaction
	self log: 'Commit Transaction'.
	connection commit! !

!VWDatabaseAccessor methodsFor: 'login' stamp: ''!
connectionClassForLogin: aLogin 
	^aLogin database vwConnectionClass! !

!VWDatabaseAccessor methodsFor: 'accessing' stamp: ''!
createSession

	| session |
	session := connection getSession.
	(session respondsTo: #defaultDisplayLobSize:) ifTrue: [
		session defaultDisplayLobSize: DatabaseCommand new maximumLobSizeToRetrieveDirectly]. 
	session blockFactor: 100.  "We don't know, so make it big"
	^session.! !

!VWDatabaseAccessor methodsFor: 'accessing' stamp: ''!
createSessionFor: aCommand

	| session types |
	session := connection getSession.
	(session respondsTo: #defaultDisplayLobSize:) ifTrue: [
		session defaultDisplayLobSize: aCommand maximumLobSizeToRetrieveDirectly]. 

	session blockFactor: aCommand blockFactor.
	types := aCommand bindingTypesForEXDI.
	types isEmpty ifFalse: [session bindTemplate: types].
	^session.! !

!VWDatabaseAccessor methodsFor: 'accessing' stamp: ''!
currentLogin: aLogin
	self reusePreparedStatements: aLogin database reusePreparedStatements.
	super currentLogin: aLogin.! !

!VWDatabaseAccessor methodsFor: 'cursors' stamp: ''!
cursorFor: answerStream

	answerStream == #noAnswerStream ifTrue: [^#() readStream].
	answerStream == #noMoreAnswers ifTrue: [^#() readStream].

	^ super cursorFor: answerStream! !

!VWDatabaseAccessor methodsFor: 'cursors' stamp: ''!
cursorFor: answerStream command: aCommand

	| stream |
	stream := answerStream.
	answerStream == #noAnswerStream ifTrue: [
		stream := #() readStream].
	answerStream == #noMoreAnswers ifTrue: [
		stream := #() readStream].

	^ super cursorFor: stream command: aCommand! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
disconnect

	self reset.
	^connection disconnect! !

!VWDatabaseAccessor methodsFor: 'accessing' stamp: ''!
driverSession
	driverSession isNil ifTrue: [driverSession := self createSession].
	^driverSession.! !

!VWDatabaseAccessor methodsFor: 'accessing' stamp: ''!
encoding
	^connection encoding.! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
errorsToCatch
	^self class errorsToCatch.! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
executeCommand: aCommand inDBSession: aSession
	"Execute the command, given a database session which has already been prepared for this command"
	| answerStream cursor |
	aSession execute.
	answerStream := aSession answer.
	aCommand needsRowCount
		ifTrue:
			[aCommand
				rowCount:
					(self platform supportsRowCount
						ifTrue: [aSession rowCount]
						ifFalse: [1])].
	cursor := self cursorFor: answerStream command: aCommand.
	cursor statement: (GlorpPreparedStatement new statement: aSession).
	^cursor.! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
executeCommandBound: aCommand
	| statement cursor time block |
	self logCommand: aCommand bound: true.
	block := [
		statement := self statementFor: aCommand.
		statement statement bindInput: aCommand bindings.
		cursor := self executeCommand: aCommand inDBSession: statement statement.
		cursor statement: statement].
	time := Time millisecondsToRun: block.
	self logTime: time.
	^cursor.! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
executeCommandUnbound: aCommand
	| statement cursor block time |
	self logCommand: aCommand bound: false.
	block := [
		statement := self statementFor: aCommand.
		cursor := self executeCommand: aCommand inDBSession: statement statement.
		cursor statement: statement].
	time := Time millisecondsToRun: block.
	self logTime: time.
	^cursor.! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
externalDatabaseErrorSignal

	^connection class externalDatabaseErrorSignal.! !

!VWDatabaseAccessor methodsFor: 'initialize-release' stamp: ''!
initialize
	super initialize.
	reusePreparedStatements := true.
	self initializePreparedStatementCache.! !

!VWDatabaseAccessor methodsFor: 'initialize-release' stamp: ''!
initializePreparedStatementCache

	| policy cache |
	policy := CachePolicy default.
	policy numberOfElements: 20.
	policy expiryAction: #notifyAndRemove.
	preparedStatements := CacheManager new.
	cache := preparedStatements cacheForClass:GlorpPreparedStatement.
	cache cachePolicy: policy.! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
innerExceptionFor: exception
	(exception parameter isKindOf: GenericException) ifTrue: [^exception parameter].
	^exception.! !

!VWDatabaseAccessor methodsFor: 'transactions' stamp: ''!
isInTransaction

	^connection inTransactionMode! !

!VWDatabaseAccessor methodsFor: 'login' stamp: ''!
isLoggedIn

	connection isNil ifTrue: [^false].
	^connection isConnected! !

!VWDatabaseAccessor methodsFor: 'login' stamp: 'GuillermoPolito 4/17/2011 18:17'!
loginIfError: aBlock 
	self log: 'Login'.
	connection := self connectionClass new.
	connection username: currentLogin username.
	connection environment: currentLogin connectString.
	self execute: [connection connect: currentLogin password] ifError: aBlock.
	self log: 'Login finished'! !

!VWDatabaseAccessor methodsFor: 'login' stamp: ''!
logout
	self isLoggedIn ifFalse: [^self].
	self log: 'Logout'.
	self execute: [connection disconnect. self reset] ifError: [:ex | ex return: nil].
	self log: 'Logout finished'! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
newPreparedStatementFor: aCommand
	| newSession |
	newSession := self createSessionFor: aCommand.
	aCommand useBinding
		ifTrue: [newSession prepare: aCommand sqlString]
		ifFalse: [newSession prepareWithoutBinding: aCommand sqlString].
	^(GlorpPreparedStatement new)
		signature: aCommand signature;
		statement: newSession;
		busy: true.! !

!VWDatabaseAccessor methodsFor: 'accessing' stamp: ''!
numberOfPreparedStatements
	^preparedStatements numberOfElements.! !

!VWDatabaseAccessor methodsFor: 'copying' stamp: ''!
postCopy
	super postCopy.
	self initializePreparedStatementCache.! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
preparedStatementFor: aCommand
	| wasMarkedBusyForUs statement |
	wasMarkedBusyForUs := false.
	statement := preparedStatements
		lookupClass: GlorpPreparedStatement
		key: aCommand signature
		ifAbsentPut:
			[wasMarkedBusyForUs := true.
			self newPreparedStatementFor: aCommand].
	(wasMarkedBusyForUs not and: [statement busy])
		ifTrue: [statement := self newPreparedStatementFor: aCommand].
	^statement.! !

!VWDatabaseAccessor methodsFor: 'login' stamp: ''!
reLogin
	"Attempt to re-login to a database that's not responding. Give it a maximum of 5 attempts"
	| attempts |
	attempts := 0.
	[attempts < 5 and: [self isLoggedIn not]] whileTrue: [
		attempts := attempts + 1.
		self loginIfError: [:ex | nil]].! !

!VWDatabaseAccessor methodsFor: 'cursors' stamp: ''!
releaseCursor: aCursor in: aPreparedStatement
	"Release the cursor. We're trapping exceptions, so return true or false to indicate success/failure"
	aCursor isNil ifTrue: [^self].
	aPreparedStatement isNil ifTrue: [
		 ^(aCursor class ~~ ReadStream) ifTrue: [self halt] ifFalse: [self]].
	[aPreparedStatement statement isNil or: [aPreparedStatement statement answer == #noMoreAnswers]] whileFalse.
	^[aCursor close. true] on: Dialect error do: [:ex | ex return: false].! !

!VWDatabaseAccessor methodsFor: 'initialize-release' stamp: ''!
reset
	super reset.	
	preparedStatements release.
	self initializePreparedStatementCache.
	driverSession := nil.! !

!VWDatabaseAccessor methodsFor: 'transactions' stamp: ''!
rollbackTransaction
	self log: 'Rollback Transaction'.
	connection rollback! !

!VWDatabaseAccessor methodsFor: 'accessing' stamp: ''!
rowCount
	"GACK. Kind of assumes we know that we just prepared the last statement. And if the driver doesn't support it, just fake it. Very ugly. "
	self platform supportsRowCount ifFalse: [^1].
	^self driverSession rowCount.! !

!VWDatabaseAccessor methodsFor: 'login' stamp: ''!
showDialog: aString

	(Smalltalk at: #Dialog) warn: aString.! !

!VWDatabaseAccessor methodsFor: 'executing' stamp: ''!
statementFor: aCommand
	^self reusePreparedStatements
		ifTrue: [self preparedStatementFor: aCommand]
		ifFalse: [self newPreparedStatementFor: aCommand].! !

Object subclass: #DatabaseCommand
	instanceVariableNames: 'useBinding stream sqlString session'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!DatabaseCommand commentStamp: '<historical>' prior: 0!
A DatabaseCommand is what a query uses to actually generate its SQL and to execute it. A command wraps a stream, into which the SQL is written, but with knowledge of the platform for which it's being written, and potentially details about how it should execute.

Subclasses must implement the following messages:
	accessing
		arrayBoundFields
		bindings
		bindingTypes
		parameterTypeSignature
	executing
		printSQL
	testing
		succeeded

Instance Variables:
	platform	<DatabasePlatform>	The database for which we're to generate/execute SQL.
	sqlString	<String>	The resulting SQL.
	stream	<Stream>	The stream into which we generate the SQL. Once we've finished and got sqlString this isn't necessary any more.
	useBinding	<Boolean>	Should we use binding, or print parameters directly into the string.

!

!DatabaseCommand classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
arrayBoundFields
	self subclassResponsibility! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
bindingTypes
	self subclassResponsibility.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
bindingTypesForEXDI
	^self bindingTypes collect: [:each | each exdiType].! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
bindings
	self subclassResponsibility.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
blockFactor
	^5  "A reasonable default if we don't know"! !

!DatabaseCommand methodsFor: 'testing' stamp: ''!
canBind: aValue to: aType
	useBinding ifFalse: [^false].
	^self platform canBind: aValue to: aType underlyingType.! !

!DatabaseCommand methodsFor: 'stream behaviour' stamp: ''!
contents
	^stream contents.! !

!DatabaseCommand methodsFor: 'stream behaviour' stamp: ''!
cr
	stream cr.! !

!DatabaseCommand methodsFor: 'executing' stamp: ''!
executeBoundIn: anAccessor
	^anAccessor executeCommandBound: self.! !

!DatabaseCommand methodsFor: 'executing' stamp: ''!
executeUnboundIn: anAccessor
	^anAccessor executeCommandUnbound: self.! !

!DatabaseCommand methodsFor: 'testing' stamp: ''!
hasMultipleRows
	^false.! !

!DatabaseCommand methodsFor: 'initializing' stamp: ''!
initialize
	useBinding := true.
	stream := WriteStream on: (String new: 100).! !

!DatabaseCommand methodsFor: 'testing' stamp: ''!
isReadCommand
	^false.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
maximumLobSizeToRetrieveDirectly
	^32768.! !

!DatabaseCommand methodsFor: 'testing' stamp: ''!
needsRowCount 

	^false.! !

!DatabaseCommand methodsFor: 'stream behaviour' stamp: ''!
nextPut: aCharacter
	^stream nextPut: aCharacter.! !

!DatabaseCommand methodsFor: 'stream behaviour' stamp: ''!
nextPutAll: aString
	^stream nextPutAll: aString.! !

!DatabaseCommand methodsFor: 'stream behaviour' stamp: ''!
nextPutString: aString
	^self nextPutAll: aString.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
parameterTypeSignature
	self subclassResponsibility.! !

!DatabaseCommand methodsFor: 'As yet unclassified' stamp: ''!
platform
	self session isNil ifTrue: [^nil].
	^self session platform.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
platform: aDatabasePlatform
	self error: 'This API is obsolete. Pass the session instead'.! !

!DatabaseCommand methodsFor: 'stream behaviour' stamp: ''!
policy
	^stream policy.! !

!DatabaseCommand methodsFor: 'stream behaviour' stamp: ''!
print: anObject

	anObject printOn: self! !

!DatabaseCommand methodsFor: 'executing' stamp: ''!
printSQL
	self subclassResponsibility.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
query
	^nil.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
session
	^session.! !

!DatabaseCommand methodsFor: 'As yet unclassified' stamp: ''!
session: aGlorpSession
	session := aGlorpSession.! !

!DatabaseCommand methodsFor: 'testing' stamp: ''!
shouldUseGroupWriting
	^(self hasMultipleRows and: [self platform supportsGroupWritingFor: self])! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
signature
	^self sqlString, self parameterTypeSignature.! !

!DatabaseCommand methodsFor: 'stream behaviour' stamp: ''!
space
	stream space.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
sqlString
	sqlString isNil ifTrue: [
		stream isNil ifTrue: [stream := WriteStream on: (String new: 100)].
		self printSQL.
		sqlString := stream contents.
		stream := nil].
	^sqlString.! !

!DatabaseCommand methodsFor: 'testing' stamp: ''!
succeeded

	^self subclassResponsibility.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
system
	self session isNil ifTrue: [^nil].
	^self session system.! !

!DatabaseCommand methodsFor: 'reporting' stamp: ''!
tables
	"Return a list of the tables we will affect"
	^self subclassResponsibility.! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
useBinding
	^useBinding! !

!DatabaseCommand methodsFor: 'accessing' stamp: ''!
useBinding: aBoolean
	useBinding := aBoolean! !

DatabaseCommand subclass: #RowBasedCommand
	instanceVariableNames: 'row allRows rowCount failed'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!RowBasedCommand commentStamp: '<historical>' prior: 0!
This is a command based on one or more in-memory rows. The purposes is to write, or delete that row or rows. This knows enough to be able to use features like array binding to do the row operations more efficiently.

Subclasses must implement the following messages:
	bound values
		arrayBoundFields
		singleRowBoundFields

Instance Variables:
	allRows	<(Collection of: (DatabaseRow))> A collection of rows. May be nil. If we have multiples, then we will be array bound or otherwise bulk-written under certain circumstances.
	row	<DatabaseRow>	The primary row, either the single row we're writing, or the one we're writing right now out of the allRows collection.
	rowCount	<Integer>	The row count from the last statement we executed.
	succeeded	<Boolean>	Has the statement overall succeded or failed.

!

RowBasedCommand subclass: #DeleteCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!DeleteCommand commentStamp: '<historical>' prior: 0!
A command to delete our set of rows.!

!DeleteCommand methodsFor: 'accessing' stamp: ''!
arrayBoundFields
	"For a delete, it makes no difference, we only ever include the primary key fields"
	^self singleRowBoundFields.! !

!DeleteCommand methodsFor: 'accessing' stamp: ''!
blockFactor
	^1.! !

!DeleteCommand methodsFor: 'accessing' stamp: ''!
printSQL
	self nextPutAll: 'DELETE FROM '.
	row table printSQLOn: self withParameters: #().
	self nextPutAll: ' WHERE '.
	row printPrimaryKeyTemplateOn: self.! !

!DeleteCommand methodsFor: 'accessing' stamp: ''!
singleRowBoundFields
	| fields |
	fields := row table primaryKeyFields.
	fields isEmpty ifTrue: [fields := row fields].
	^fields asArray! !

RowBasedCommand subclass: #InsertCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!InsertCommand commentStamp: '<historical>' prior: 0!
A command to insert our set of rows.!

!InsertCommand methodsFor: 'accessing' stamp: ''!
arrayBoundFields
	"For array binding, include all the fields, except if we're inserting into a database with identity columns, where we must omit those columns"
	^self platform usesIdentityColumns
		ifTrue: [row table fields reject: [:each | each type isIdentityColumn]]
		ifFalse: [row table fields].! !

!InsertCommand methodsFor: 'accessing' stamp: ''!
basicPrintSQLFor: aRow
	| fields |
	self nextPutAll: 'INSERT INTO '.
	aRow table printSQLOn: self withParameters: #().
	fields := self shouldUseGroupWriting
		ifTrue: [self arrayBoundFields]
		ifFalse: [aRow nonGeneratedFieldsWithValues].
	self nextPutAll: ' ('.
	GlorpHelper
		do: [:each | self nextPutAll: (self platform nameForColumn: each name)]
		for: fields
		separatedBy: [self nextPutAll: ','].
	self nextPutAll: ') '.
	self nextPutAll: ' VALUES ('.
	GlorpHelper
		do:
			[:each | 
			(self canBind: (aRow at: each ifAbsent: [nil]) to: each type)
				ifTrue: [self nextPut: $?]
				ifFalse: [aRow printValueOfField: each on: self]]
		for: fields
		separatedBy: [self nextPutAll: ','].
	self nextPutAll: ')'.! !

!InsertCommand methodsFor: 'accessing' stamp: ''!
printSQL
	| iterations |
	iterations := self shouldUseGroupWriting
		ifTrue:
			[self platform usesArrayBindingRatherThanGrouping
				ifTrue: [1]
				ifFalse: [allRows size]]
		ifFalse: [1].
	GlorpHelper
		do:
			[:i | self basicPrintSQLFor: (i = 1 ifTrue: [row] ifFalse: [allRows at: i])]
		for: (1 to: iterations)
		separatedBy: [self nextPutAll: '; '].! !

!InsertCommand methodsFor: 'accessing' stamp: ''!
singleRowBoundFields
	"Include only the fields that have values in our particular row"
	^row nonGeneratedFieldsWithValues! !

!InsertCommand methodsFor: 'testing' stamp: ''!
supportsGroupWriting
	^true.! !

!RowBasedCommand classSide methodsFor: 'instance creation' stamp: ''!
forRow: aDatabaseRow useBinding: aBoolean session: aSession
	^(self new)
		row: aDatabaseRow;
		useBinding: aBoolean;
		session: aSession;
		yourself.! !

!RowBasedCommand classSide methodsFor: 'instance creation' stamp: ''!
forRows: aCollectionOfDatabaseRows useBinding: aBoolean session: aSession
	^(self new)
		allRows: aCollectionOfDatabaseRows;
		row: aCollectionOfDatabaseRows first;
		useBinding: aBoolean;
		session: aSession;
		yourself.! !

!RowBasedCommand methodsFor: 'accessing' stamp: ''!
allRows
	^allRows.! !

!RowBasedCommand methodsFor: 'accessing' stamp: ''!
allRows: aCollectionOfDatabaseRows
	allRows := aCollectionOfDatabaseRows.! !

!RowBasedCommand methodsFor: 'bound values' stamp: ''!
arrayBindings
	"Return bindings in a form suitable for Oracle array binding, i.e. a list of arguments, but each argument is itself a list of multiple values"
	| allBindings fields |
	fields := self arrayBoundFields.
	allBindings := (1 to: fields size)
		collect: [:each | Array new: allRows size].
	1
		to: allRows size
		do:
			[:rowNumber | 
			| inputRow |
			inputRow := allRows at: rowNumber.
			1
				to: fields size
				do:
					[:columnNumber | 
					(allBindings at: columnNumber)
						at: rowNumber
						put: (inputRow at: (fields at: columnNumber) ifAbsent: [nil])]].
	^allBindings asArray.! !

!RowBasedCommand methodsFor: 'bound values' stamp: ''!
arrayBoundFields
	"Return the list of fields to be bound in the SQL string we will use for array binding. For array binding, we need to include all possible fields, rather than those that just have values in a particular row"
	self subclassResponsibility.! !

!RowBasedCommand methodsFor: 'bound values' stamp: ''!
batchStatementBindings
	"Return our bindings, which we assume are for multiple rows, as one big array, one statement after the other"

	| bound fields |
	fields := self arrayBoundFields.
	bound := OrderedCollection new: allRows size * fields size.
	allRows do: [:eachRow | 
		fields do: [:each | 
			| fieldValue |
			fieldValue := eachRow at: each ifAbsent: [nil].
			(self canBind: fieldValue to: each type) ifTrue: [bound add: fieldValue]]].
	^bound asArray.! !

!RowBasedCommand methodsFor: 'bound values' stamp: ''!
bindingTypes
	| all basic size |
	^(self shouldUseGroupWriting)
		ifTrue: [
			basic := self row nonGeneratedFields collect: [:each | each type].
			size := basic size.
			all := Array new: (allRows size * size).
			0 to: allRows size - 1 do: [:i |
				| index |
				index := i * size.
				all replaceElementsFrom: index + 1 to: index + size withArray: basic startingAt: 1].
			all]
		ifFalse: [self singleRowBoundFields collect: [:each | each type]].! !

!RowBasedCommand methodsFor: 'bound values' stamp: ''!
bindings
	^(self shouldUseGroupWriting)
		ifTrue: [self platform bindingsForGroupWritingFor: self]
		ifFalse: [self singleRowBindings].! !

!RowBasedCommand methodsFor: 'executing' stamp: ''!
executeBoundIn: anAccessor
	| result |
	(self hasMultipleRows and: [self shouldUseGroupWriting not])
		ifTrue:
			[self allRows
				do:
					[:each | 
					self row: each.
					result := anAccessor executeCommandBound: self.
					result release.
					failed := failed or: [self needsRowCount and: [rowCount ~= 1]].
					each postWriteAssignSequencesUsing: anAccessor]]
		ifFalse:
			[result := anAccessor executeCommandBound: self.
			result release.
			failed := self needsRowCount and: [rowCount ~= self expectedNumberOfModifiedRows].
			row postWriteAssignSequencesUsing: anAccessor].
	"Results aren't interesting"
	^ anAccessor cursorFor: #() readStream! !

!RowBasedCommand methodsFor: 'bound values' stamp: ''!
executeUnboundIn: anAccessor
	anAccessor returnRowCount: self needsRowCount.
	(self hasMultipleRows and: [self shouldUseGroupWriting not])
		ifTrue:
			[self allRows
				do:
					[:each | 
					self row: each.
					anAccessor
						executeSQLStringNoResult: self sqlString
						doing:
							[failed := failed or: [self needsRowCount and: [anAccessor rowCount ~= 1]]]
						forCommand: self.
					each postWriteAssignSequencesUsing: anAccessor]]
		ifFalse:
			[self tooBigForSingleWrite
				ifTrue:
					[self subCommands do: [:each | each executeUnboundIn: anAccessor]]
				ifFalse:
					[anAccessor
						executeSQLStringNoResult: self sqlString
						doing:
							[failed := self needsRowCount
								and: [anAccessor rowCount ~= self expectedNumberOfModifiedRows]]
						forCommand: self].
			row postWriteAssignSequencesUsing: anAccessor].
	^#() readStream.! !

!RowBasedCommand methodsFor: 'executing' stamp: ''!
expectedNumberOfModifiedRows
	^self hasMultipleRows ifTrue: [allRows size] ifFalse: [1].! !

!RowBasedCommand methodsFor: 'testing' stamp: ''!
hasMultipleRows
	^allRows notNil and: [allRows size > 1].! !

!RowBasedCommand methodsFor: 'initializing' stamp: ''!
initialize
	super initialize.
	failed := false.! !

!RowBasedCommand methodsFor: 'testing' stamp: ''!
maximumSingleWriteSize
	"How many values will write at once. This is only really applicable when not doing binding and concatenating many statements together for one trip to the DB"
	^1000.! !

!RowBasedCommand methodsFor: 'accessing' stamp: ''!
parameterTypeSignature
	"Return a string indicating the parameter types. Complicated by the possibility of multiple rows, and of null values"
	| result |
	result := WriteStream on: String new.
	self hasMultipleRows
		ifTrue:
			[row
				keysAndValuesDo:
					[:eachKey :eachValInFirstRow| 
					| nextValue |
					eachValInFirstRow notNil
						ifTrue: [result nextPutAll: eachValInFirstRow class name]
						ifFalse:
							[allRows
								do:
									[:eachRow | 
									(nextValue := eachRow at: eachKey ifAbsent: [nil]) notNil
										ifTrue: [result nextPutAll: nextValue class name]]]]]
		ifFalse:
			[row
				keysAndValuesDo:
					[:eachKey :eachValue | result nextPutAll: eachValue class name]].
	^result contents.! !

!RowBasedCommand methodsFor: 'accessing' stamp: ''!
row
	^row.! !

!RowBasedCommand methodsFor: 'accessing' stamp: ''!
row: anObject
	row := anObject.
	sqlString := nil.! !

!RowBasedCommand methodsFor: 'bound values' stamp: ''!
rowAtATimeBindings
	"Return multiple sets of bindings in a form suitable for iterating over"
	| allBindings fields rowTemplate |
	fields := self arrayBoundFields.
	rowTemplate := Array new: fields size.
	allBindings := Array new: allRows size.
	1 to: allBindings size do: [:i | | currentRow input |
		currentRow := rowTemplate copy.
		input := allRows at: i.
		allBindings at: i put: currentRow.
		1 to: fields size do: [:j |
			currentRow at: j put: (input at: (fields at: j) ifAbsent: [nil])]].
	^allBindings asArray.! !

!RowBasedCommand methodsFor: 'accessing' stamp: ''!
rowCount
	^rowCount.! !

!RowBasedCommand methodsFor: 'accessing' stamp: ''!
rowCount: anInteger
	rowCount := anInteger.! !

!RowBasedCommand methodsFor: 'bound values' stamp: ''!
singleRowBindings
	| bound |
	bound := OrderedCollection new.
	self singleRowBoundFields
		do:
			[:each | 
			| fieldValue |
			fieldValue := row at: each ifAbsent: [nil].
			(self canBind: fieldValue to: each type) ifTrue: [bound add: fieldValue]].
	^bound asArray.! !

!RowBasedCommand methodsFor: 'bound values' stamp: ''!
singleRowBoundFields
	"Return the list of fields to be bound in the SQL string we will use for a single operation. We only need to include those fields that actually have values in the particular row we're operating on"
	self subclassResponsibility.! !

!RowBasedCommand methodsFor: 'executing' stamp: ''!
subCommands
	| startPosition commands endPosition subRows newCommand |
	startPosition := 1.
	commands := OrderedCollection new.
	[startPosition > allRows size] whileFalse: [
		endPosition := (startPosition + self maximumSingleWriteSize - 1) min: allRows size.
		subRows := allRows copyFrom: startPosition to: endPosition.
		newCommand := self class forRows: subRows useBinding: self useBinding session: self session.
		commands add: newCommand.
		startPosition := endPosition + 1].
	^commands.! !

!RowBasedCommand methodsFor: 'testing' stamp: ''!
succeeded
	^failed not.! !

!RowBasedCommand methodsFor: 'testing' stamp: ''!
supportsGroupWriting
	"Return true if array binding or a similar mechanism can be used to execute us with multiple commands at once. I think in general this can only ever make sense for inserts (and possibly selects under some circumstances). Updates may want to update only part of the fields."
	^false.! !

!RowBasedCommand methodsFor: 'reporting' stamp: ''!
tables
	^Array with: self row table.! !

!RowBasedCommand methodsFor: 'testing' stamp: ''!
tooBigForSingleWrite
	^self hasMultipleRows and: [allRows size > self maximumSingleWriteSize].! !

RowBasedCommand subclass: #UpdateCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!UpdateCommand commentStamp: '<historical>' prior: 0!
A command to update our set of rows.!

!UpdateCommand methodsFor: 'accessing' stamp: ''!
arrayBindings
	"I don't think we can array bind updates. There may be values in the row that we haven't read, and if we attempt to write all the fields we'll be overwriting them with nulls. It's possible that there's a set of rows with all the same fields and we could update them all the same way, but that isn't done yet"
	self error: 'cannot array bind an update'.! !

!UpdateCommand methodsFor: 'accessing' stamp: ''!
arrayBoundFields
	^row table nonPrimaryKeyFields asArray , row table primaryKeyFields asArray.! !

!UpdateCommand methodsFor: 'bound values' stamp: ''!
bindingTypes
	^super bindingTypes, (row lockKeyFields collect: [:each | each type]).! !

!UpdateCommand methodsFor: 'accessing' stamp: ''!
blockFactor
	^1.! !

!UpdateCommand methodsFor: 'accessing' stamp: ''!
fieldsNeedingUpdate
	^self shouldUpdateOnlyChangedFields 
		ifTrue: [row fieldsNeedingUpdate]
		ifFalse: [row nonPrimaryKeyFields]! !

!UpdateCommand methodsFor: 'accessing' stamp: ''!
needsRowCount
	^row lockKeyFields isEmpty not.! !

!UpdateCommand methodsFor: 'printing' stamp: ''!
printSQL
	| bodyFields |
	self nextPutAll: 'UPDATE '.
	row table printSQLOn: self withParameters: #().
	self nextPutAll: ' SET '.
	bodyFields := self fieldsNeedingUpdate.
	GlorpHelper 
		do: [:field | row printEqualityTemplateForField: field on: self]
		for: bodyFields
		separatedBy: [self nextPut: $,].
	self nextPutAll: ' WHERE '.
	row printPrimaryKeyTemplateOn: self.
	row lockKeyFields notEmpty
		ifTrue: 
			[self nextPutAll: ' AND '.
			GlorpHelper
				do: [:field | row printLockEqualityForField: field on: self]
				for: row lockKeyFields
				separatedBy: [self nextPutAll: ' AND ']]! !

!UpdateCommand methodsFor: 'testing' stamp: ''!
shouldUpdateOnlyChangedFields
	"Return true if we should write only changed fields out in an update statement, or if we should write all of them. The tradeoff is that writing only the changed ones will produce less SQL, but make more variations of the statement, and require us to keep more prepared statements around if we're using binding. By default, use this option only if we're not using binding by default. It would be nice if we could delegate this decision to somewhere the user has control over, e.g. the descriptor system, but we don't actually know about it"
	^self system shouldUpdateOnlyChangedFieldsFor: self.! !

!UpdateCommand methodsFor: 'bound values' stamp: ''!
singleRowBindings
	| basic lockFields lockValues |
	basic := super singleRowBindings.
	lockFields := row lockKeyFields.
	lockFields isEmpty ifTrue: [^basic].
	lockValues := OrderedCollection new: 2.
	row lockKeyFields do: [:each | 	
			| fieldValue |
			fieldValue := row oldVersionAt: each ifAbsent: [nil].
			(self canBind: fieldValue to: each type) ifTrue: [lockValues add: fieldValue]].
	^basic, lockValues.! !

!UpdateCommand methodsFor: 'bound values' stamp: ''!
singleRowBoundFields
	"Note: We will actually have to bind one more thing than this, the optimistic lock fields (if present). That's handled in the caller"
	| basicFields |
	basicFields := self fieldsNeedingUpdate.
	^basicFields , row table primaryKeyFields asArray! !

DatabaseCommand subclass: #SQLStringCommand
	instanceVariableNames: 'parameters maximumBlobSizeToRetrieveDirectly'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!SQLStringCommand commentStamp: '<historical>' prior: 0!
This represents a command that executes arbitrary SQL and is not a select. For selects use SQLStringSelectCommand.!

!SQLStringCommand classSide methodsFor: 'instance creation' stamp: ''!
for: aString
	^self new setSQLString: aString.! !

!SQLStringCommand methodsFor: 'accessing' stamp: ''!
arrayBoundFields
	"Don't have enough information to do this, I don't think"
	^#()! !

!SQLStringCommand methodsFor: 'accessing' stamp: ''!
bindingTypes
	^#()! !

!SQLStringCommand methodsFor: 'accessing' stamp: ''!
bindings
	^parameters.! !

!SQLStringCommand methodsFor: 'initialize-release' stamp: ''!
initialize
	super initialize.
	maximumBlobSizeToRetrieveDirectly := super maximumLobSizeToRetrieveDirectly.! !

!SQLStringCommand methodsFor: 'accessing' stamp: ''!
maximumLobSizeToRetrieveDirectly
	^maximumBlobSizeToRetrieveDirectly! !

!SQLStringCommand methodsFor: 'accessing' stamp: ''!
maximumLobSizeToRetrieveDirectly: anInteger
	maximumBlobSizeToRetrieveDirectly := anInteger! !

!SQLStringCommand methodsFor: 'accessing' stamp: ''!
parameterTypeSignature
	^''! !

!SQLStringCommand methodsFor: 'accessing' stamp: ''!
parameters
	^parameters.! !

!SQLStringCommand methodsFor: 'accessing' stamp: ''!
parameters: anArray
	parameters := anArray! !

!SQLStringCommand methodsFor: 'executing' stamp: ''!
printSQL
	" *** This method was defined by Glorp.DatabaseCommand as a subclass responsibility.
	Replace its body with a proper implementation. *** "

	self error: 'Subclass responsibility stub not reimplemented'! !

!SQLStringCommand methodsFor: 'accessing' stamp: ''!
setSQLString: aString
	sqlString := aString.! !

!SQLStringCommand methodsFor: 'testing' stamp: ''!
succeeded
	" *** This method was defined by Glorp.DatabaseCommand as a subclass responsibility.
	Replace its body with a proper implementation. *** "

	self error: 'Subclass responsibility stub not reimplemented'! !

DatabaseCommand subclass: #SelectCommand
	instanceVariableNames: 'parameters blockFactor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!SelectCommand commentStamp: '<historical>' prior: 0!
This represents a SELECT query to the database. This can either be based on a GLORP query, or on a hard-coded SQL string.

Subclasses must implement the following messages:
	accessing
		printSQL

Instance Variables:
	blockFactor	<SmallInteger>	The blocking factor to use for this query. This is most likely to make a difference for these kind of operations, and so is directly settable here.
	parameters	<Dictionary from: DatabaseField to: Object>	The parameters to the query.

!

SelectCommand subclass: #QuerySelectCommand
	instanceVariableNames: 'query boundExpressions'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!QuerySelectCommand commentStamp: '<historical>' prior: 0!
This is the normal class to use for select queries. It generates SQL for a select operation returning a set of rows.

Instance Variables:
	boundExpressions	<(Collection of: GlorpExpression)>	The expressions for which we need bind parameters. We expect that these are things that resolve to fields, or more generally, to something the database will accept as a valid thing to bind ofr.
	query	<Query>	The query in which we're executing.

!

QuerySelectCommand subclass: #CompoundSelectCommand
	instanceVariableNames: 'commands'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!CompoundSelectCommand commentStamp: '<historical>' prior: 0!
This is a select command that uses a number of sub-commands to print multiple queries, and combine them into one for execution. Used in conjunction with compound queries, e.g. using UNION ALL. The query knows what the combination operation is.

Instance Variables:
	commands	<(Collection of: DatabaseCommand)>	The commands we are to print.!

!CompoundSelectCommand methodsFor: 'accessing' stamp: ''!
arrayBoundFields
	" *** This method was defined by Glorp.DatabaseCommand as a subclass responsibility.
	Replace its body with a proper implementation. *** "

	self error: 'Subclass responsibility stub not reimplemented'! !

!CompoundSelectCommand methodsFor: 'accessing' stamp: ''!
bindingTypes
	^self commands inject: #() into: [:sum :each | sum, each bindingTypes].! !

!CompoundSelectCommand methodsFor: 'accessing' stamp: ''!
bindings
	^commands inject: #() into: [:sum :each | sum, each bindings].! !

!CompoundSelectCommand methodsFor: 'accessing' stamp: ''!
commands
	commands isNil ifTrue: [commands := query queries collect: [:each | each sqlWith: parameters]].
	^commands.! !

!CompoundSelectCommand methodsFor: 'accessing' stamp: ''!
parameters: aDictionary
	parameters := aDictionary.
	commands := nil.! !

!CompoundSelectCommand methodsFor: 'accessing' stamp: ''!
printSQL
	self nextPutAll: 'SELECT '.
	self requiresDistinct ifTrue: [self nextPutAll: 'DISTINCT '].
	self nextPutAll: '* FROM '.
	self platform supportsAliasInCompoundQuery ifTrue: [self nextPut: $(].
	GlorpHelper 
		do: [:each | 
			self platform parenthesizeCommandsInCompoundQuery ifTrue: [self nextPut: $(].
			self nextPutAll: each sqlString.
			self platform parenthesizeCommandsInCompoundQuery ifTrue: [self nextPut: $)]]
		for: self commands
		separatedBy: [query printOperationOn: self].
	self platform supportsAliasInCompoundQuery ifTrue: [self nextPut: $)].
	query printCompoundSelectAliasesOn: self.
	query printOrderingOn: self.
	query printGroupByOn: self.! !

!CompoundSelectCommand methodsFor: 'accessing' stamp: ''!
requiresDistinct
	^self query requiresDistinct.! !

!QuerySelectCommand methodsFor: 'accessing' stamp: ''!
bindingTypes
	self useBinding ifFalse: [^#()].
	boundExpressions isNil ifTrue: [self findBoundExpressions].
	^boundExpressions collect: [:each | each bindingType].! !

!QuerySelectCommand methodsFor: 'accessing' stamp: ''!
bindings
	self useBinding ifFalse: [^#()].
	boundExpressions isNil ifTrue: [self findBoundExpressions].
	^boundExpressions collect: [:each | each bindingIn: self].! !

!QuerySelectCommand methodsFor: 'accessing' stamp: ''!
blockFactor
	blockFactor isNil ifTrue: [blockFactor := query expectedRows].
	^blockFactor.! !

!QuerySelectCommand methodsFor: 'accessing' stamp: ''!
boundExpressions
	^boundExpressions! !

!QuerySelectCommand methodsFor: 'accessing' stamp: ''!
boundExpressions: anObject
	boundExpressions := anObject! !

!QuerySelectCommand methodsFor: 'accessing' stamp: ''!
findBoundExpressions
	self useBinding ifFalse: [^nil].
	boundExpressions := OrderedCollection new.
	query addBoundExpressionsIn: self to: boundExpressions.! !

!QuerySelectCommand methodsFor: 'accessing' stamp: ''!
maximumLobSizeToRetrieveDirectly
	^query isNil
		ifTrue: [super maximumLobSizeToRetrieveDirectly]
		ifFalse: [query maximumLobSizeToRetrieveDirectly].! !

!QuerySelectCommand methodsFor: 'printing' stamp: ''!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(.
	query printOn: aStream.
	aStream nextPut: $).! !

!QuerySelectCommand methodsFor: 'accessing' stamp: ''!
printSQL
	stream nextPutAll: 'SELECT '.
	query printSelectFieldsOn: self.
	self findBoundExpressions.
	query printTablesOn: self.
	query printWhereClauseOn: self.
	query printJoinsOn: self.
	query printOrderingOn: self.
	query printGroupByOn: self.
	query printPostLimitOn: self.
	query printOffsetOn: self.! !

!QuerySelectCommand methodsFor: 'accessing' stamp: ''!
query
	^query.! !

!QuerySelectCommand methodsFor: 'accessing' stamp: ''!
query: aQuery
	query := aQuery.! !

!QuerySelectCommand methodsFor: 'reporting' stamp: ''!
tables
	^(self query tablesToPrint collect: [:each | each parent]) asSet.! !

QuerySelectCommand subclass: #SQLStringSelectCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!SQLStringSelectCommand commentStamp: '<historical>' prior: 0!
This is a select command whose SQL is given, rather than being generated from a query.

!

!SQLStringSelectCommand classSide methodsFor: 'As yet unclassified' stamp: ''!
sqlString: aString parameters: aDictionaryOrArray useBinding: aBoolean session: aSession
	^(self new)
		setSQLString: aString;
		parameters: aDictionaryOrArray;
		useBinding: aBoolean;
		session: aSession.! !

!SQLStringSelectCommand methodsFor: 'accessing' stamp: ''!
bindingTypes
	^query isNil ifTrue: [#()] ifFalse: [super bindingTypes].! !

!SQLStringSelectCommand methodsFor: 'accessing' stamp: ''!
bindings
	"We're not actually sure if we're part of a query with a where clause, or being used in a more ad hoc way. So if we have a query, use either our pre-set parameters, or the query's bindings, whichever produces more. In a tie, pick the query's bindings."
	| queryBindings |
	query isNil ifTrue: [^parameters].
	queryBindings := super bindings.
	^queryBindings size >= parameters size 
		ifTrue: [queryBindings]
		ifFalse: [parameters].! !

!SQLStringSelectCommand methodsFor: 'accessing' stamp: ''!
blockFactor
	blockFactor isNil ifTrue: [
		blockFactor := query isNil ifTrue: [5] ifFalse: [super blockFactor]].
	^blockFactor.! !

!SQLStringSelectCommand methodsFor: 'sql' stamp: ''!
printSQL
	"We assume the SQL is provided."
	^self.! !

!SQLStringSelectCommand methodsFor: 'accessing' stamp: ''!
setSQLString: aString
	sqlString := aString.! !

!SQLStringSelectCommand methodsFor: 'reporting' stamp: ''!
tables
	^query isNil ifTrue: [Array with: #Unknown] ifFalse: [super tables].! !

SQLStringSelectCommand subclass: #SQLite3PragmaCommand
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!SQLite3PragmaCommand methodsFor: 'sql' stamp: ''!
printSQL

	stream nextPutAll: 'PRAGMA '.
	query resultClass = Glorp.DatabaseField
		ifTrue: [stream nextPutAll: 'table_info ('].
	query resultClass = Glorp.SQLite3ForeignKeyData
		ifTrue: [stream nextPutAll: 'foreign_key_list('].
	query resultClass = Glorp.SQLite3IndexData
		ifTrue: [stream nextPutAll: 'index_list('].
	query resultClass = Glorp.SQLite3IndexColumnData
		ifTrue: [stream nextPutAll: 'index_info('].
	"query printSelectFieldsOn: self."
	self findBoundExpressions.
	"query printTablesOn: self."
	query printWhereClauseOn: self.
	stream nextPutAll: ')'
	"query printJoinsOn: self.
	query printOrderingOn: self.
	query printGroupByOn: self.
	query printPostLimitOn: self.
	query printOffsetOn: self."! !

!SelectCommand classSide methodsFor: 'instance creation' stamp: ''!
forQuery: aQuery parameters: aDictionary
	^(self new)
		query: aQuery;
		parameters: aDictionary;
		yourself.! !

!SelectCommand classSide methodsFor: 'instance creation' stamp: ''!
forQuery: aQuery parameters: aDictionary useBinding: aBoolean session: aSession
	^(self new)
		query: aQuery;
		parameters: aDictionary;
		useBinding: aBoolean;
		session: aSession;
		yourself.! !

!SelectCommand methodsFor: 'accessing' stamp: ''!
blockFactor: anInteger
	blockFactor := anInteger.! !

!SelectCommand methodsFor: 'testing' stamp: ''!
canBind: aValue to: aType
	aValue isNil ifTrue: [^false].
	^super canBind: aValue to: aType.! !

!SelectCommand methodsFor: 'testing' stamp: ''!
isReadCommand
	^true.! !

!SelectCommand methodsFor: 'accessing' stamp: ''!
parameterTypeSignature
	| result |
	result := WriteStream on: String new.
	self bindings do: [:each | result nextPutAll: each class name].
	^result contents.! !

!SelectCommand methodsFor: 'accessing' stamp: ''!
parameters
	^parameters! !

!SelectCommand methodsFor: 'accessing' stamp: ''!
parameters: aDictionary
	parameters := aDictionary! !

!SelectCommand methodsFor: 'accessing' stamp: ''!
printSQL
	self subclassResponsibility.! !

!SelectCommand methodsFor: 'testing' stamp: ''!
succeeded

	^true.! !

Object subclass: #DatabaseConverter
	instanceVariableNames: 'name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!DatabaseConverter commentStamp: '<historical>' prior: 0!
This is a superclass for objects that convert back and forth between database representations and object representations.

Subclasses must implement the following messages:
	converting
		convert:fromDatabaseRepresentationAs:
		convert:toDatabaseRepresentationAs:

Instance Variables:
	name	<ByteSymbol>	The name for this converter. The name is used, concatenated with #Converter, to create a symbol that can be performed against the DatabasePlatform to give this converter. So, e.g. booleanToIntegerConverter.

!

!DatabaseConverter classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!DatabaseConverter methodsFor: 'converting' stamp: ''!
convert: anObject fromDatabaseRepresentationAs: aDatabaseType 
	self subclassResponsibility! !

!DatabaseConverter methodsFor: 'converting' stamp: ''!
convert: anObject toDatabaseRepresentationAs: aDatabaseType 
	self subclassResponsibility! !

!DatabaseConverter methodsFor: 'initialize' stamp: ''!
initialize
	name := #unnamed.! !

!DatabaseConverter methodsFor: 'accessing' stamp: ''!
name
	^name.! !

!DatabaseConverter methodsFor: 'accessing' stamp: ''!
name: aString
	name := aString.! !

!DatabaseConverter methodsFor: 'printing' stamp: ''!
printOn: aString
	aString nextPutAll: 'DatabaseConverter(', name, ')'.! !

DatabaseConverter subclass: #DelegatingDatabaseConverter
	instanceVariableNames: 'host stToDbSelector dbToStSelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!DelegatingDatabaseConverter commentStamp: '<historical>' prior: 0!
This is a converter which delegates the conversion to methods on some other object, typically the database platform.

Instance Variables:
	dbToStSelector	<Symbol>	The symbol we perform when converter from database to object representation. Takes two arguments, the value, and the database type.
	host	<Object>	The object that holds the methods specified by the selectors.
	stToDbSelector	<Symbol>	The symbol we perform when converter from object to database representation. Takes two arguments, the value, and the database type.

!

!DelegatingDatabaseConverter classSide methodsFor: 'instance creation' stamp: ''!
hostedBy: anObject fromStToDb: stDbSelector fromDbToSt: dbStSelector

	^super new
		hostedBy: anObject fromStToDb: stDbSelector fromDbToSt: dbStSelector! !

!DelegatingDatabaseConverter classSide methodsFor: 'instance creation' stamp: ''!
named: aSymbol hostedBy: anObject fromStToDb: stDbSelector fromDbToSt: dbStSelector
	^super new
		named: aSymbol
		hostedBy: anObject
		fromStToDb: stDbSelector
		fromDbToSt: dbStSelector.! !

!DelegatingDatabaseConverter methodsFor: 'converting' stamp: ''!
convert: anObject fromDatabaseRepresentationAs: aDatabaseType 
	^host perform: dbToStSelector with: anObject with: aDatabaseType! !

!DelegatingDatabaseConverter methodsFor: 'converting' stamp: ''!
convert: anObject toDatabaseRepresentationAs: aDatabaseType 
	^host perform: stToDbSelector with: anObject with: aDatabaseType.! !

!DelegatingDatabaseConverter methodsFor: 'initialize-release' stamp: ''!
hostedBy: anObject fromStToDb: stDbSelector fromDbToSt: dbStSelector

	host := anObject.
	stToDbSelector := stDbSelector.
	dbToStSelector := dbStSelector! !

!DelegatingDatabaseConverter methodsFor: 'initialize-release' stamp: ''!
named: aSymbol hostedBy: anObject fromStToDb: stDbSelector fromDbToSt: dbStSelector

	name := aSymbol.
	host := anObject.
	stToDbSelector := stDbSelector.
	dbToStSelector := dbStSelector! !

DatabaseConverter subclass: #NullConverter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!NullConverter commentStamp: '<historical>' prior: 0!
This is a no-op converter. When converting in either direction, it just returns the value.!

!NullConverter methodsFor: 'converting' stamp: ''!
convert: anObject fromDatabaseRepresentationAs: aDatabaseType
	^anObject.! !

!NullConverter methodsFor: 'converting' stamp: ''!
convert: anObject toDatabaseRepresentationAs: aDatabaseType
	^anObject.! !

DatabaseConverter subclass: #PluggableDatabaseConverter
	instanceVariableNames: 'stToDb dbToSt'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!PluggableDatabaseConverter commentStamp: '<historical>' prior: 0!
This is a converter where the conversion is specified by two blocks.

Instance Variables:
	dbToSt	<BlockClosur>	The block to run to convert from the database representation to Smalltalk representation. Takes one argument, the value to be converted.
	stToDb	<BlockClosure>	The block to run to convert from Smalltalk representation to the database representation. Takes one argument, the value to be converted.

!

!PluggableDatabaseConverter methodsFor: 'converting' stamp: ''!
convert: anObject fromDatabaseRepresentationAs: aDatabaseType 
	^dbToSt isNil ifTrue: [anObject] ifFalse: [dbToSt value: anObject]! !

!PluggableDatabaseConverter methodsFor: 'converting' stamp: ''!
convert: anObject toDatabaseRepresentationAs: aDatabaseType 
	^stToDb isNil ifTrue: [anObject] ifFalse: [stToDb value: anObject]! !

!PluggableDatabaseConverter methodsFor: 'accessing' stamp: ''!
dbToStConverter: aBlock

	dbToSt := aBlock.! !

!PluggableDatabaseConverter methodsFor: 'accessing' stamp: ''!
stToDbConverter: aBlock

	stToDb := aBlock.! !

Object subclass: #DatabaseDriver
	instanceVariableNames: 'connection'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!DatabaseDriver methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/17/2011 17:55'!
basicExecuteSQLString: aString
	^self subclassResponsibility.! !

!DatabaseDriver methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/17/2011 17:55'!
beginTransaction
self subclassResponsibility.! !

!DatabaseDriver methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/17/2011 17:55'!
commitTransaction
	^self subclassResponsibility.! !

!DatabaseDriver methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/17/2011 17:55'!
connect: aLogin
	^self subclassResponsibility.! !

!DatabaseDriver methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/17/2011 17:56'!
connection
	^ connection! !

!DatabaseDriver methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/17/2011 17:56'!
connection: anObject
	connection := anObject! !

!DatabaseDriver methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/17/2011 17:56'!
connectionClass
	^self subclassResponsibility.! !

!DatabaseDriver methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/17/2011 17:56'!
isConnected
	^ self subclassResponsibility ! !

!DatabaseDriver methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/17/2011 17:56'!
logout
^self subclassResponsibility.! !

!DatabaseDriver methodsFor: 'as yet unclassified' stamp: 'GuillermoPolito 4/17/2011 17:56'!
rollbackTransaction
	^self subclassResponsibility.! !

Object subclass: #DatabaseField
	instanceVariableNames: 'table name isPrimaryKey position type isNullable isUnique isLockKey defaultValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!DatabaseField commentStamp: '<historical>' prior: 0!
This represents a field in a database table. See also DatabaseTable. Glorp relies heavily on identity of fields for comparison purposes, so be careful when copying these.

Instance Variables:
	defaultValue	<Object>	What is the default value of this field, if unspecified in the creation statement.
	isLockKey	<Boolean>	Is this field part of an optimistic lock key. If so, then on update we will generate an update statement that checks that this value hasn't changed since the object wasn't read.
	isNullable	<Boolean>	Is this field allowed to be null?
	isPrimaryKey	<Boolean>	Is this field part of the primary key?
	isUnique	<Boolean>	Does this field have a unique constraint in the database?
	name	<String>	The name of the field.
	position	<Object | ProtoObject>	description of position
	table	<DatabaseTable>	The table that contains this field.
	type	<GlorpDatabaseType>	The type of  data that this field contains.

!

!DatabaseField classSide methodsFor: 'instance creation' stamp: ''!
named: aString

	^self error: 'type needed'! !

!DatabaseField classSide methodsFor: 'instance creation' stamp: ''!
named: aString type: dbType

	^super new initialize
		name: aString;
		type: dbType! !

!DatabaseField classSide methodsFor: 'instance creation' stamp: ''!
new

	^self error: 'dbType needed'! !

!DatabaseField methodsFor: 'comparing' stamp: ''!
<= aField
	^self name <= aField name! !

!DatabaseField methodsFor: 'obsolete' stamp: ''!
asConstraintReferenceString

	^table name, ' (', self name, ')'.! !

!DatabaseField methodsFor: 'converting' stamp: ''!
asDistinctFieldOn: anExpression

	^(anExpression getField: self) asDistinctFieldOn: anExpression.! !

!DatabaseField methodsFor: 'converting' stamp: ''!
asField
	^self.! !

!DatabaseField methodsFor: 'converting' stamp: ''!
asGlorpExpression

	^ParameterExpression forField: self basedOn: nil.! !

!DatabaseField methodsFor: 'converting' stamp: ''!
asGlorpExpressionOn: anExpression

	^ParameterExpression forField: self basedOn: anExpression.! !

!DatabaseField methodsFor: 'configuring' stamp: ''!
beIndexed
	self table addIndex: (DatabaseIndex forField: self).! !

!DatabaseField methodsFor: 'configuring' stamp: ''!
beLockKey

	isLockKey := true.
	self table isNil ifFalse: [self table addAsLockKeyField: self]! !

!DatabaseField methodsFor: 'configuring' stamp: ''!
beNullable: aBoolean

	self isPrimaryKey ifFalse: [ isNullable := aBoolean ]! !

!DatabaseField methodsFor: 'configuring' stamp: ''!
bePrimaryKey
	isPrimaryKey := true.
	isNullable := false.
	self table isNil ifFalse: [self table addAsPrimaryKeyField: self]! !

!DatabaseField methodsFor: 'converting' stamp: ''!
converterForStType: aClass

	^self type converterForStType: (aClass isBehavior ifTrue: [aClass] ifFalse: [aClass class]).! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
defaultValue
	^defaultValue! !

!DatabaseField methodsFor: 'configuring' stamp: ''!
defaultValue: anObject
	defaultValue := anObject! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
foreignKeyConstraints
	^self table foreignKeyConstraints select: [:each | each sourceFields includes: self].! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
impliedSmalltalkType
	"Return the default Smalltalk type corresponding to our database type"

	^self type impliedSmalltalkType.! !

!DatabaseField methodsFor: 'initializing' stamp: ''!
initialize
	isPrimaryKey := false.
	isNullable := true.
	isUnique := false.
	isLockKey := false.! !

!DatabaseField methodsFor: 'testing' stamp: ''!
isDistinct
	^false.! !

!DatabaseField methodsFor: 'testing' stamp: ''!
isGenerated

	^type isGenerated or: [defaultValue notNil].! !

!DatabaseField methodsFor: 'testing' stamp: ''!
isIndexed
	^self table hasIndexForField: self.! !

!DatabaseField methodsFor: 'testing' stamp: ''!
isLockKey
	isLockKey isNil ifTrue: [isLockKey := false].
	^isLockKey! !

!DatabaseField methodsFor: 'testing' stamp: ''!
isMappable
	"Return true if we're something that can be returned in a select statement. Always true for fields"
	^true.! !

!DatabaseField methodsFor: 'testing' stamp: ''!
isNullable
	
	^isNullable.! !

!DatabaseField methodsFor: 'testing' stamp: ''!
isPrimaryKey
	"Private - Answer the value of the receiver's ''isPrimaryKey'' instance variable."

	^isPrimaryKey yourSelf.! !

!DatabaseField methodsFor: 'copying' stamp: ''!
isSameAliasedFieldAs: aField
	"Return true if we are the same as another aliased field. For non-aliased fields, identity is assured, so this is redundant. Note that we == the string. This will work because we know these must both be shallow copies of the same thing"
	self class == aField class ifFalse: [^false].
	^self table == aField table and: [self name == aField name].! !

!DatabaseField methodsFor: 'comparing' stamp: 'GuillermoPolito 4/17/2011 17:39'!
isTheSameAs: aDatabaseField
	"Validate that two fields are exactly the same. Most useful for testing purposes, rather than normal usage, which is why we don't just use the #= method."
	self name asUppercase = aDatabaseField name asUppercase ifFalse: [^false].
	self isPrimaryKey = aDatabaseField isPrimaryKey ifFalse: [^false].
	(self isTheSameTypeAs: aDatabaseField) ifFalse: [^false].
	self isNullable = aDatabaseField isNullable ifFalse: [^false].
	self isUnique = aDatabaseField isUnique ifFalse: [^false].
	self isLockKey = aDatabaseField isLockKey ifFalse: [^false].
	self defaultValue = aDatabaseField defaultValue ifFalse: [^false].
	^true.! !

!DatabaseField methodsFor: 'comparing' stamp: ''!
isTheSameTypeAs: aDatabaseField
	| myType aType |
	self isPrimaryKey
		ifTrue:
		   [myType := self type.
		   aType := aDatabaseField type.
		   ^(myType = aType 
				or: [(myType isSerial or: [aType isSerial ]) 
					and: [myType impliedSmalltalkType = aType impliedSmalltalkType ]]) ].
		
	^self type = aDatabaseField type! !

!DatabaseField methodsFor: 'testing' stamp: ''!
isUnique
	isUnique isNil ifTrue: [isUnique := false].
	^isUnique.! !

!DatabaseField methodsFor: 'configuring' stamp: ''!
isUnique: aBoolean
	isUnique := aBoolean.! !

!DatabaseField methodsFor: 'testing' stamp: ''!
matchesField: aField
	table name = aField table name ifFalse: [^false].
	name = aField name ifFalse: [^false].
	isNullable = aField isNullable ifFalse: [^false].
	isPrimaryKey = aField isPrimaryKey ifFalse: [^false].
	isUnique = aField isUnique ifFalse: [^false].
	^true! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
name
	"Private - Answer the value of the receiver's ''name'' instance variable."

	^name! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
name: aString
	"Private - Set the value of the receiver's ''name'' instance variable to the argument, anObject."

	name := aString asString.! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
platform
	^self type platform.! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
position
	^position! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
position: anObject
	position := anObject! !

!DatabaseField methodsFor: 'initializing' stamp: ''!
postInitializeIn: aDescriptorSystem
	"Any initialization that has to be delayed until we're in the table"
	type initializeForField: self in: aDescriptorSystem.! !

!DatabaseField methodsFor: 'obsolete' stamp: ''!
printForConstraintNameOn: aStream maxLength: maxLength 
	| constraintName |
	constraintName := table name , '_' , name.
	constraintName size > maxLength 
		ifTrue: [constraintName := constraintName copyFrom: 1 to: maxLength].
	aStream nextPutAll: constraintName! !

!DatabaseField methodsFor: 'obsolete' stamp: ''!
printNameOn: aStream withParameters: anArray 
	aStream nextPutAll: self name! !

!DatabaseField methodsFor: 'printing' stamp: ''!
printOn: aStream
	
	aStream nextPutAll: 'Field'.
	aStream 
		nextPutAll: '(';
		nextPutAll: (table isNil ifTrue: [''] ifFalse: [table name]);
		nextPutAll: '.';
		nextPutAll: name;
		nextPutAll: ')'.! !

!DatabaseField methodsFor: 'obsolete' stamp: ''!
printQualifiedSQLOn: aStream withParameters: aDictionary 
	aStream nextPutAll: self qualifiedName! !

!DatabaseField methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: anArray 
	aStream nextPutAll: self quotedQualifiedName.! !

!DatabaseField methodsFor: 'printing' stamp: ''!
printSelectSQLOn: aStream withParameters: anArray 
	self printSQLOn: aStream withParameters: anArray.! !

!DatabaseField methodsFor: 'printing' stamp: ''!
printUnqualifiedSQLOn: aStream withParameters: anArray 
	aStream nextPutAll: self name.! !

!DatabaseField methodsFor: 'querying' stamp: ''!
qualifiedName

	^table isNil 
		ifTrue: [self name]
		ifFalse: [self table qualifiedName, '.', self name].! !

!DatabaseField methodsFor: 'querying' stamp: ''!
quotedQualifiedName

	| basicName |
	basicName := self platform nameForColumn: self name.
	^table isNil 
		ifTrue: [basicName]
		ifFalse: [self table qualifiedName, '.', basicName].! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
table
	^table! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
table: anObject

	table := anObject! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
tablesToPrint
	^Array with: table! !

!DatabaseField methodsFor: 'accessing' stamp: ''!
type

	^type! !

!DatabaseField methodsFor: 'configuring' stamp: ''!
type: aDatabaseType

	type := aDatabaseType.
	(type notNil and: [type isSerial]) ifTrue: [self bePrimaryKey].! !

!DatabaseField methodsFor: 'database' stamp: ''!
typeString

	^type typeString! !

!DatabaseField methodsFor: 'copying' stamp: ''!
withTable: aTable
	"Return a copy of ourselves, with the table set to aTable. Presumably aTable is an aliased version of our table"
	^self copy table: aTable.! !

Object subclass: #DatabaseIndex
	instanceVariableNames: 'name fields'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!DatabaseIndex commentStamp: '<historical>' prior: 0!
This describes an index in the database. Glorp doesn't use it internally, other than when creating tables in the database.

Instance Variables:
	fields	<(SequenceableCollection of: DatabaseField)>	The set of fields that are to be indexed.
	name	<String>	The name of the index. We will try and generate one based on the table if one is not supplied.

!

!DatabaseIndex classSide methodsFor: 'instance creation' stamp: ''!
forField: aField
	^(self new initialize)
		addField: aField;
		yourself! !

!DatabaseIndex classSide methodsFor: 'instance creation' stamp: ''!
forFields: aCollection
	^(self new)
		addFields: aCollection;
		yourself! !

!DatabaseIndex classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize.! !

!DatabaseIndex methodsFor: 'initializing' stamp: ''!
addField: aField
	fields add: aField! !

!DatabaseIndex methodsFor: 'initializing' stamp: ''!
addFields: aCollection
	fields addAll: aCollection! !

!DatabaseIndex methodsFor: 'accessing' stamp: ''!
fields
	^fields.! !

!DatabaseIndex methodsFor: 'printing' stamp: ''!
generateName
	"Generate a name for this index that will fit within the size limits of the database. This only takes into account the first field being used. If there's more than one that has multiple columns and the same first one, then the user will have to manually provide a name"

	| stream maxLength tableLength roomToBeVerbose minTableLength maxTableLength |
	maxLength := fields first table platform maximumLengthOfTableName.
	minTableLength := ((maxLength - 8) // 2) min: 10.
	maxTableLength := (maxLength // 2) - 1.
	roomToBeVerbose := minTableLength >= 10.
	tableLength := roomToBeVerbose ifTrue: [maxTableLength] ifFalse: [minTableLength].

	stream := WriteStream on: (String new: 100).
	stream nextPutAll: (fields first table name copyFrom: 1 to: ((fields first table name size) min: tableLength)).
	stream nextPutAll: '_'.
	stream nextPutAll: (fields first name copyFrom: 1 to: ((fields first name size) min: tableLength)).
	stream nextPutAll: (maxLength - stream position >= 4 ifTrue: ['_IDX'] ifFalse: ['X']).
	^stream contents.! !

!DatabaseIndex methodsFor: 'testing' stamp: ''!
indexesField: aField
	^fields includes: aField.! !

!DatabaseIndex methodsFor: 'initializing' stamp: ''!
initialize
	fields := OrderedCollection new: 2.! !

!DatabaseIndex methodsFor: 'comparing' stamp: ''!
isTheSameAs: aDatabaseIndex
	"Validate that two indexes are exactly the same. Most useful for testing purposes, rather than normal usage, which is why we don't just use the #= method."
	self name asUppercase = aDatabaseIndex name asUppercase ifFalse: [^false].
	self fields with: aDatabaseIndex fields  do: [:a :b | (a isTheSameAs: b) ifFalse: [^false]].	
	^true.! !

!DatabaseIndex methodsFor: 'accessing' stamp: ''!
name
	name isNil ifTrue: [name := self generateName].
	^name.! !

!DatabaseIndex methodsFor: 'accessing' stamp: ''!
name: aString
	name := aString.! !

!DatabaseIndex methodsFor: 'printing' stamp: ''!
printStatementOn: aStream
	aStream nextPut: $(.
	fields do: [ :f | aStream nextPutAll: f name; nextPut: $, ].
	aStream skip: -1; nextPut: $)! !

Object subclass: #DatabasePlatform
	instanceVariableNames: 'types converters useBinding reservedWords functions characterEncoding'
	classVariableNames: 'UseBindingIfSupported'
	poolDictionaries: ''
	category: 'Glorp-Database'!

DatabasePlatform class
	instanceVariableNames: 'converterRepository'!

!DatabasePlatform commentStamp: '<historical>' prior: 0!
This is an abstract superclass for classes that provide a compatibility layer for different databases. They encapsulate different SQL syntaxes, binding mechanism, types that are available, and general limits.

Subclasses must implement the following messages:
	accessing
		vwEXDIConnectionClass
	binding
		bindingsForGroupWritingFor:
	constants
		areSequencesExplicitlyCreated
		supportsMillisecondsInTimes
	sequences
		databaseSequenceClass
	testing
		usesArrayBindingRatherThanGrouping
	types
		int4
		sequence
		serial
		timestamp
		varbinary
		varchar

Instance Variables:
	characterEncoding	<ByteSymbol>	What character encoding should we use for the database connection.
	converters	<Dictionary from: Symbol to: DatabaseConverter>	The available type converters. So, for example, if there is a boolean to integer conversion, we expect it in this dictionary associated with the symbol #booleanToInteger. If that symbol isn't there, we will look up the converter by performing #booleanToIntegerConverter and caching the result.
	functions	<FunctionExpression>	Which functions are available for this database. There is a set of common functions created by FunctionExpression, plus we can define our own additions or overrides that are specific to this database.
	reservedWords	<Collection of: String>	Which names are reserved words in this database. Any that we use as table or column names will need to be quoted.
	types	<Dictionary from: Symbol to: GlorpDatabaseType>	A cache of available types, mapped by name to the type instance.
	useBinding	<Boolean>	By default, should we use binding for this database.

!

DatabasePlatform class
	instanceVariableNames: 'converterRepository'!

DatabasePlatform subclass: #AdabasLikePlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!AdabasLikePlatform commentStamp: '<historical>' prior: 0!
This is a platform for Adabas and related databases.!

AdabasLikePlatform subclass: #AdabasDPlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!AdabasDPlatform commentStamp: '<historical>' prior: 0!
This is a platform for the Adabas database.!

!AdabasLikePlatform methodsFor: 'SQL' stamp: ''!
supportsConstraints
	"^<Boolean> This method returns true, if the database support a contraint concept"

	^false! !

AdabasLikePlatform subclass: #SAPDBPlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!SAPDBPlatform commentStamp: '<historical>' prior: 0!
Ths is a platform for SAPDB.!

!DatabasePlatform classSide methodsFor: 'eoglorp' stamp: ''!
createLoginFromConnectionDictionary: aDict 
	self subclassResponsibility! !

!DatabasePlatform classSide methodsFor: 'eoglorp' stamp: ''!
loginWithConnectionDictionary: aDict 
	| platformClass |
	platformClass := self glorpAllSubclasses
				detect: [:cls | cls understandsConnectionDictionary: aDict].
	^ platformClass isNil
		ifFalse: [platformClass createLoginFromConnectionDictionary: aDict] ifTrue: [nil]! !

!DatabasePlatform classSide methodsFor: 'eoglorp' stamp: ''!
understandsConnectionDictionary: aDict

	^false! !

!DatabasePlatform classSide methodsFor: 'testing' stamp: ''!
useBindingByDefault
	^true! !

!DatabasePlatform classSide methodsFor: 'testing' stamp: ''!
useBindingIfSupported
	"If I am nil (the default), each platform that supportsBinding lazily initializes to its defaultUseBinding value.  If I have a value, I force lazy initialization of each platform that supportsBinding to my value.  Normally, a value is set only for specific testing purposes."

	UseBindingIfSupported isNil ifTrue: [^self useBindingByDefault].
	^UseBindingIfSupported! !

!DatabasePlatform classSide methodsFor: 'testing' stamp: ''!
useBindingIfSupported: aBooleanOrNil
	"If I am nil (the default), each platform that supportsBinding lazily initializes to its defaultUseBinding value.  If I have a value, I force lazy initialization of each platform that supportsBinding to my value.  Normally, a value is set only for specific testing purposes."

	UseBindingIfSupported := aBooleanOrNil.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
areSequencesExplicitlyCreated

	self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
asTimestamp: anObject for: aType
	anObject isNil ifTrue: [^nil].
	^anObject class = Time
		ifTrue: [Dialect newTimestampFromDate: Date today time: anObject]
		ifFalse: [anObject asTimestamp].! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
batchWriteStatementTerminatorString
	"^<String> This statement return the string to be used to devide several statement during batch write ..."

	^';'! !

!DatabasePlatform methodsFor: 'binding' stamp: ''!
bindingsForGroupWritingFor: aCommand
	"Return the bindings array for a group write. This can be in different formats, depending on the database and perhaps the mechanism in place."
	self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
blob
	^self typeNamed: #blob ifAbsentPut: [GlorpBlobType new].! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
boolean
	"By default, we represent booleans as small integers."
	^self smallint.! !

!DatabasePlatform methodsFor: 'types' stamp: 'GP 11/3/2009 00:20'!
booleanFalse
^ 0.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
booleanToBooleanConverter
	^DelegatingDatabaseConverter
		named: #booleanToBoolean
		hostedBy: self
		fromStToDb: #convertBooleanToDBBoolean:for:
		fromDbToSt: #convertDBBooleanToBoolean:for:.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
booleanToIntegerConverter
	^DelegatingDatabaseConverter
		named: #booleanToInteger
		hostedBy: self
		fromStToDb: #convertBooleanToInteger:for:
		fromDbToSt: #convertIntegerToBoolean:for:.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
booleanToStringTFConverter
	^DelegatingDatabaseConverter
		named: #booleanToString
		hostedBy: self
		fromStToDb: #convertBooleanToTF:for:
		fromDbToSt: #convertTFToBoolean:for:.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
booleanToStringYNConverter
	^DelegatingDatabaseConverter
		named: #booleanToString
		hostedBy: self
		fromStToDb: #convertBooleanToYN:for:
		fromDbToSt: #convertYNToBoolean:for:.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
booleanToStringYesNoConverter
	^DelegatingDatabaseConverter
		named: #booleanToString
		hostedBy: self
		fromStToDb: #convertBooleanToYesNo:for:
		fromDbToSt: #convertYesNoToBoolean:for:.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
byteArrayToString: aByteArray for: aType
	^aByteArray isNil ifTrue: [nil] ifFalse: [Dialect convert: aByteArray toStringWithEncoding: self characterEncoding].! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
byteArrayToStringConverter
	^DelegatingDatabaseConverter
		named: #byteArrayToString
		hostedBy: self
		fromStToDb: #convertStringToByteArray:for:
		fromDbToSt: #convertByteArrayToString:for:.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
canBind: aValue to: aType
	(Dialect unbindableClassNames includes: aValue class name) ifTrue: [^false].
	^true.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
canUseInWhereClause: aValue type: aType
	(Dialect isFloatingPoint: aValue) ifTrue: [^false].
	^true.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
capitalWritingOfColumnName
	"^<Boolean> This method returns true, if the dbms wants to have column 
	names written in capital letters"

	^false! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
capitalWritingOfCreatorName
	"^<Boolean> This method returns true, if the dbms wants to have column 
	names written in capital letters"

	^true! !

!DatabasePlatform methodsFor: 'constants' stamp: ' 22/6/07 12:54'!
capitalWritingOfDatabaseName
	"^<Boolean>"

	^true! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
capitalWritingOfSQLCommands
	"^<Boolean>"

	^true! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
capitalWritingOfTableName
	"^<Boolean>"

	^true! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
char
	^self typeNamed: #char ifAbsentPut: [GlorpCharType new].! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
char: anInteger
	^self char width: anInteger.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
character
	^self char! !

!DatabasePlatform methodsFor: 'accessing' stamp: ''!
characterEncoding
	"We hope that this will be set appropriately for the connection, but make sure the default is fairly harmless, and will in most cases complain rather than write characters incorrectly"
	characterEncoding isNil ifTrue: [characterEncoding := #ascii].
	^characterEncoding! !

!DatabasePlatform methodsFor: 'accessing' stamp: ''!
characterEncoding: anEncodingName
	characterEncoding := anEncodingName.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
charactersThatNeedEscaping
	^#($' )! !

!DatabasePlatform methodsFor: 'types' stamp: ' 22/6/07 12:54'!
clob
	^self typeNamed: #blob ifAbsentPut: [GlorpClobType new].! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
columnNameSeparatorString
	"^<String> This statement return the string to be used to devide several columns ..."

	^','! !

!DatabasePlatform methodsFor: 'database-specific' stamp: ''!
compoundOperationFor: aSymbol
	"Return the platform specific version of a compound statement symbol"
	^aSymbol.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
convertBooleanToDBBoolean: aBoolean for: aType
	(self usesNullForFalse and: [aBoolean isNil]) ifTrue: [^false].
	^aBoolean.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
convertBooleanToInteger: aBoolean for: aType
	aBoolean isNil ifTrue: [^nil].
	^aBoolean ifTrue: [1] ifFalse: [0].! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
convertBooleanToTF: aBoolean for: aType
	aBoolean isNil ifTrue: [^aBoolean].
	^aBoolean ifTrue: ['T'] ifFalse: ['F'].! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
convertBooleanToYN: aBoolean for: aType
	aBoolean isNil ifTrue: [^aBoolean].
	^aBoolean ifTrue: ['Y'] ifFalse: ['N'].! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
convertBooleanToYesNo: aBoolean for: aType
	aBoolean isNil ifTrue: [^aBoolean].
	^aBoolean ifTrue: ['YES'] ifFalse: ['NO'].! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
convertByteArrayToSQLString: aByteArray for: aType
	"This is a slightly misleading name. Really what we want is conversion to a Blob-type of field. So if it's a bytearray, we can just let the driver take care of that conversion, but if we're given a string, we should convert it."

	^aByteArray isString
		ifTrue:
			[Dialect
				convert: aByteArray
				toByteArrayWithEncoding: self characterEncoding]
		ifFalse: [aByteArray].! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
convertByteArrayToString: aByteArray for: aType
	^aByteArray asString.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
convertDBBooleanToBoolean: aBoolean for: aType
	^aBoolean.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
convertIntegerToBoolean: anInteger for: aType
	anInteger isNil ifTrue: [^anInteger].
	anInteger = 1 ifTrue: [^true].
	anInteger = 0 ifTrue: [^false].
	self error: 'invalid boolean conversion'.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
convertStringToByteArray: aString for: aType
	^aString asByteArray.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
convertTFToBoolean: aString for: aType
	aString isNil ifTrue: [^aString].
	aString = 'T' ifTrue: [^true].
	aString = 'F' ifTrue: [^false].
	self error: 'invalid boolean conversion'.! !

!DatabasePlatform methodsFor: 'conversion-numbers' stamp: ''!
convertToDouble: aNumber for: type 

	aNumber isNil ifTrue: [^nil].
	^Dialect coerceToDoublePrecisionFloat: aNumber! !

!DatabasePlatform methodsFor: 'conversion-numbers' stamp: ''!
convertToFloat: aNumber for: aType 
	aNumber isNil ifTrue: [^nil].
	^aNumber asFloat.! !

!DatabasePlatform methodsFor: 'conversion-numbers' stamp: ''!
convertToInteger: aNumber for: aType 
	^aNumber isNil ifTrue: [aNumber] ifFalse: [aNumber asNumber asInteger].! !

!DatabasePlatform methodsFor: 'conversion-numbers' stamp: ''!
convertToNumber: aNumber for: aType
	"Convert to an arbitrary numeric type. Make sure our precision matches. *Hate* floating point"

	| exponent result |
	aNumber isNil ifTrue: [^nil].
	aType precision isNil ifTrue: [^aNumber].
	(Dialect doesPrecisionOf: aNumber equal: aType precision) ifTrue: [^aNumber].
	exponent := Dialect
		coerceToDoublePrecisionFloat: (10 raisedToInteger: aType precision).
	result := (aNumber * exponent) rounded / exponent.
	^aNumber isInteger ifTrue: [result rounded] ifFalse: [result].! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
convertYNToBoolean: aString for: aType
	aString isNil ifTrue: [^aString].
	aString = 'Y' ifTrue: [^true].
	aString = 'N' ifTrue: [^false].
	self error: 'invalid boolean conversion'.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
convertYesNoToBoolean: aString for: aType
	aString isNil ifTrue: [^aString].
	aString asUppercase = 'YES' ifTrue: [^true].
	aString asUppercase = 'NO' ifTrue: [^false].
	self error: 'invalid boolean conversion'.! !

!DatabasePlatform methodsFor: 'type converters' stamp: ''!
converterNamed: aSymbol

	^self converters at: aSymbol ifAbsentPut: [
		| converter |
		converter := self perform: (aSymbol, 'Converter') asSymbol.
		converter name: aSymbol]! !

!DatabasePlatform methodsFor: 'type converters' stamp: ''!
converters
	converters isNil ifTrue: [converters := IdentityDictionary new].
	^converters.! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
createAddColumnStatementFor: aField
	| stream |
	stream := WriteStream on: String new.
	stream 
		nextPutAll: 'ALTER TABLE';
		space.
	self printDDLTableNameFor: aField table on: stream.
	stream
		space;
		nextPutAll: ' ADD ';
		nextPutAll: self startColumnAddString.
	self  printColumn: aField on: stream.
	stream nextPutAll: self  endColumnAddString.
		
	^stream contents! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatement: aField newType: aType

	^self createAlterColumnStatement: aField newType: aType usingExpression: nil! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatement: aField newType: aType usingExpression: expression

	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
	self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' ALTER COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  ' TYPE ';
		nextPutAll: aType typeString.
	expression ifNotNil: [ sqlStatementStream nextPutAll: ' USING ', expression ].

	^sqlStatementStream contents! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatement: aField setDefault: anObject
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
	self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' ALTER COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  'SET DEFAULT';
		space.
	anObject printOn: sqlStatementStream.
	
	^sqlStatementStream contents! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatementDropDefault: aField
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
		self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' ALTER COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  'DROP DEFAULT '.

	^sqlStatementStream contents! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatementDropNotNull: aField
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
		self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' ALTER COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  'DROP';
		space;
		nextPutAll:  self sqlTextForNOTNULLAttributeConstraint.

	
	^sqlStatementStream contents! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatementSetNotNull: aField
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
		self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' ALTER COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  'SET';
		space;
		nextPutAll:  self sqlTextForNOTNULLAttributeConstraint.

	
	^sqlStatementStream contents! !

!DatabasePlatform methodsFor: 'services constraints' stamp: ''!
createConstraintAddStatement:  aConstraint
	| sqlStatementStream |
	self supportsAlterTableForConstraints ifFalse: [^''].
	aConstraint shouldCreateInDatabase ifFalse: [^''].
	self supportsConstraints
			ifTrue: [sqlStatementStream := WriteStream on: String new.
					sqlStatementStream
						nextPutAll:  'ALTER TABLE';
						space.
					self printDDLTableNameFor: aConstraint table on: sqlStatementStream.
					(aConstraint creationStringFor: self)
						do: [ :each | sqlStatementStream nextPutAll: ' ADD ', each.]
						separatedBy: [sqlStatementStream nextPut: $,].
					^sqlStatementStream contents].
	^String new! !

!DatabasePlatform methodsFor: 'services constraints' stamp: ''!
createConstraintDropStatement: aConstraint
	| aStream |
	self supportsAlterTableForConstraints ifFalse: [^''].
	aConstraint shouldCreateInDatabase ifFalse: [^''].
	aStream := WriteStream on: (String new: 50).
	aStream nextPutAll: 'ALTER TABLE '.
	self printDDLTableNameFor: aConstraint table on: aStream.
	aStream nextPutAll: ' DROP '.
	self writeConstraintToDrop: aConstraint onStream: aStream.
	^aStream contents.! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
createDropColumnStatementFor: aField
	| stream |
	stream := WriteStream on: String new.
	stream 
		nextPutAll: 'ALTER TABLE';
		space.
	self printDDLTableNameFor: aField table  on: stream.
	stream
		space;
		nextPutAll: 'DROP COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name).
		
	^stream contents! !

!DatabasePlatform methodsFor: 'services tables' stamp: ''!
createDropTableStatementFor: aTableNameString
	"^<String> This method returns a string which can be used to drop a database table ..."

	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll: 'DROP TABLE';
		space;
		nextPutAll: aTableNameString.

	^sqlStatementStream contents! !

!DatabasePlatform methodsFor: 'services constraints' stamp: ''!
createForeignKeyConstraintAddStatement: aForeignKeyConstraint
	"Callback from the ForeignKeyConstraint. Returns collection of statements "
	| ws |
	aForeignKeyConstraint shouldCreateInDatabase ifFalse: [^#('')].
	ws := WriteStream on: (String new: 50).
	ws
		nextPutAll: 'CONSTRAINT ';
		nextPutAll: aForeignKeyConstraint name;
		nextPutAll: ' FOREIGN KEY ('.
	GlorpHelper print: [:each | each name] on: ws for: aForeignKeyConstraint sourceFields separatedBy: ','.
	ws 	nextPutAll: ') REFERENCES ';
		nextPutAll: aForeignKeyConstraint targetTable qualifiedName;
		nextPutAll: ' ('.
	GlorpHelper print: [:each | each name] on: ws for: aForeignKeyConstraint targetFields separatedBy: ','.
	ws nextPutAll: ')'.
	aForeignKeyConstraint suffixExpression isNil
		ifFalse:
			[ws
				space;
				nextPutAll: aForeignKeyConstraint suffixExpression].

	^OrderedCollection with: ws contents.! !

!DatabasePlatform methodsFor: 'services constraints' stamp: ''!
createPrimaryKeyConstraintAddStatement: aPrimaryKeyConstraint
	"Callback from the PrimaryKeyConstraint. Returns collection of statements "
	| aStream statements |
	aPrimaryKeyConstraint shouldCreateInDatabase ifFalse: [^''].
	statements := OrderedCollection new.
	aStream := WriteStream on: String new.
	aStream nextPutAll: 'CONSTRAINT '.
	aStream nextPutAll: aPrimaryKeyConstraint name.
	aStream nextPutAll: ' PRIMARY KEY  ('.
	GlorpHelper 
		print: [:each | each name]
		on: aStream
		for: aPrimaryKeyConstraint primaryKeyFields
		separatedBy: ','.
	aStream nextPut: $).
	statements add: aStream contents.
	self primaryKeysAreAutomaticallyUnique ifTrue: [^statements].

	aStream reset.
	aStream nextPutAll: ' CONSTRAINT '.
	aStream nextPutAll: (self usesPrimaryKeyUniqueConstraintNames 
							ifTrue: [aPrimaryKeyConstraint table name, '_UNIQ'] 
							ifFalse: ['']).
	aStream nextPutAll: ' UNIQUE  ('.
	GlorpHelper 
		print: [:each | each name]
		on: aStream
		for: aPrimaryKeyConstraint  primaryKeyFields
		separatedBy: ','.
	aStream nextPut: $).
	statements add: aStream contents.

	^statements! !

!DatabasePlatform methodsFor: 'obsolete' stamp: ''!
createTableFKContraintsStatementStringsFor: aGLORPDatabaseTable

	^aGLORPDatabaseTable foreignKeyConstraintsForCreation
		collect: [:each |	self createConstraintAddStatement: each ]! !

!DatabasePlatform methodsFor: 'services tables' stamp: ''!
createTableIndexStatementStringsFor: aGLORPDatabaseTable
	^aGLORPDatabaseTable indexes
		collect:
			[:idx | 
			| sqlStatementStream |
			sqlStatementStream := WriteStream on: String new.
			sqlStatementStream nextPutAll: 'CREATE INDEX '.
			sqlStatementStream nextPutAll: idx name.
			sqlStatementStream nextPutAll: ' ON '.
			self printDDLTableNameFor: aGLORPDatabaseTable on: sqlStatementStream.
			idx printStatementOn: sqlStatementStream.
			sqlStatementStream contents].! !

!DatabasePlatform methodsFor: 'services tables' stamp: ''!
createTableStatementStringFor: aGLORPDatabaseTable 
	"^<String> This method returns a string which can be used to create a database table ..."

	| sqlStatementStream tmpString |
	tmpString := 'create table'.
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll: (self capitalWritingOfSQLCommands 
			ifTrue: [tmpString asUppercase]
			ifFalse: [tmpString]);
		space.
	self printDDLTableNameFor: aGLORPDatabaseTable on: sqlStatementStream.

	"Now print the columns specification for each field in the table ..."
	self 
		printColumnsSpecificationFor: aGLORPDatabaseTable
		on: sqlStatementStream.
	self supportsConstraints ifTrue:
		[aGLORPDatabaseTable hasPrimaryKeyConstraints ifTrue: 
			[sqlStatementStream nextPutAll: ', '.
			(aGLORPDatabaseTable primaryKeyConstraints creationStringFor: self)
				do: [ :each | sqlStatementStream nextPutAll: each.]
				separatedBy: [sqlStatementStream nextPut: $,]].

		aGLORPDatabaseTable hasUniqueConstraints ifTrue:
			[sqlStatementStream nextPutAll: ', '.
			aGLORPDatabaseTable uniqueConstraints do:
				[:each | sqlStatementStream nextPutAll: (each creationString)]
				separatedBy: [sqlStatementStream nextPut: $,]]].

	sqlStatementStream
		nextPut: $).
	^sqlStatementStream contents! !

!DatabasePlatform methodsFor: 'services constraints' stamp: ''!
createUniqueConstraintAddStatement: aUniqueConstraint
	"Callback from the UniqueConstraint. Returns collection of statements "
	
	| stream |
	aUniqueConstraint shouldCreateInDatabase ifFalse: [^#('')].
	stream := WriteStream on: (String new: 50).
	stream
		nextPutAll: 'CONSTRAINT ';
		nextPutAll: aUniqueConstraint name;
		nextPutAll: ' UNIQUE ('.
	GlorpHelper 
		print: [:each | each name] 
		on: stream 
		for: aUniqueConstraint fields 
		separatedBy: ','.
	stream nextPutAll: ')'.
	^OrderedCollection with: stream contents! !

!DatabasePlatform methodsFor: 'sequences' stamp: ''!
databaseSequenceClass
	self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
date
	"Return the type we use to represent dates. By default, we use the same type as for timestamps."
	^self timestamp.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
dateConverter
	^DelegatingDatabaseConverter
		named: #date
		hostedBy: self
		fromStToDb: #toDate:for:
		fromDbToSt: #readDate:for:.	"#printDate:for:"! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
datetime
	^self timestamp! !

!DatabasePlatform methodsFor: 'services tables' stamp: ''!
ddlTableNameFor: aTableOrString
	"
	Either a databaseTable or a string can be passed in the argument.
	"

	^ aTableOrString isString
		ifTrue: [ self capitalWritingOfTableName 
			ifTrue: [ aTableOrString asUppercase ] 
			ifFalse:[ aTableOrString ] ]
		ifFalse: [ self ddlTableNameFor: aTableOrString qualifiedName ]! !

!DatabasePlatform methodsFor: 'as yet unclassified' stamp: 'DiogenesMoreira 1/5/2011 11:39'!
defaultPort
	"self subclassResponsibility.
	for this instance i'll return 5432, but all platforms must be return you own port"
	^ '5432'.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
deleteViewWithTableSyntax

	^false! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
double
	^self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
endColumnAddString
	"The string we use as a suffix when adding a colum"
	^''.! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
endColumnQuote
	"Return the character used to stop quoting the name of a column so that we can avoid conflicts with reserved names. ANSI says to use double quotes, so that's the default."
	^'"'.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
escapeFor: aCharacter
	aCharacter = $' ifTrue: [^''''''].
	^String with: $\ with: aCharacter.! !

!DatabasePlatform methodsFor: 'exdi specific' stamp: ''!
exdiTypeForDates
	^#Date.! !

!DatabasePlatform methodsFor: 'functions' stamp: ''!
functions
	functions isNil ifTrue: [self initializeFunctions].
	^functions.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
generatorFor: aType withBlock: aBlock
	"Don't cache this, because we want to return different instances depending on aType"
	^VersionType underlyingType: aType generator: (AdHocVersionGenerator block: aBlock).! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
hasSubtransaction
	"^<Boolean> This method returns true, if the used dbms is able to execute multiple sql-statements
	transferred via a command line transmitted from client to server - otherwise I return false"

	^true! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
inMemorySequence
	^self typeNamed: #inMemorySequence ifAbsentPut: [
		GlorpInMemorySequenceDatabaseType representedBy: self int4].! !

!DatabasePlatform methodsFor: 'functions' stamp: ''!
initializeFunctions
	"Subclasses should override to add database-specific functions"

	functions := FunctionExpression createBasicFunctionsFor: self.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
initializeReservedWords
	reservedWords := Set new.! !

!DatabasePlatform methodsFor: 'constants' stamp: 'TestRunner 12/9/2009 08:53'!
initializeTypes
	types := nil.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
int
	^self integer.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
int2

	^self subclassResponsibility! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
int4

	^self subclassResponsibility! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
int8

	^self subclassResponsibility! !

!DatabasePlatform methodsFor: 'types' stamp: 'TestRunner 12/2/2009 09:03'!
integer
	^self typeNamed: #integer ifAbsentPut: [GlorpIntegerType new].! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
integerToString: anInteger for: aType
	anInteger isNil ifTrue: [^nil].
	^self padString: anInteger printString for: aType.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
integerToStringConverter
	^DelegatingDatabaseConverter
		named: #integerToString
		hostedBy: self
		fromStToDb: #integerToString:for:
		fromDbToSt: #stringToInteger:for:.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
isAccessPlatform

	^false! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
isDB2Platform

	^false! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
isMySQLPlatform
	^false.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
isODBCPlatform

	^false! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
isOcelotPlatform

	^false! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
isOraclePlatform

	^false! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
isPostgreSQLPlatform

	^false! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
isSQLServerPlatform

	^false! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
isSQLite3Platform

	^false! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
maxLengthOfDatabaseName
	"^<Integer>I return the max. length of a database name"

	^8! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
maxSQLBufferLength
	"^<Integer> I return the maximum length of a sql command stream"

	^8192! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfColumnName
	"^<Integer> I return the maximum length of a column name.  (I am not sent at present;  I am here so that if I and #maximumLengthOfTableName (which is sent) should ever differ on some database platform, I can be used in revised truncation logic.)"

	^18! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfTableName
	"^<Integer> I return the max. length of a table name"

	^18! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
maximumQueryValueSize
	"If there's a limit on how big a bound input parameter can be for a query, return it here. By default, no maximum. "
	^536870911 "VisualWorks SmallInteger maxVal. Lacking infinity, this will do as a default."! !

!DatabasePlatform methodsFor: 'binding' stamp: ''!
maximumSizeToGroupWriteFor: aCollectionOfDatabaseRows
	"If we are going to group write, how many rows of this collection should we do it for at once"
	^aCollectionOfDatabaseRows size.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
measuresStringFieldWidthsInBytes
	"Return true if we measure the width of a string in bytes rather than characters. So, a 255-character field can hold 255 bytes worth of string, not 255 characters. This is important with encodings"
	^false.! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
nameForColumn: aColumnString
	| upper |
	upper := self capitalWritingOfColumnName ifTrue: [aColumnString asUppercase] ifFalse: [aColumnString].
	(self reservedWords includes: aColumnString asLowercase) ifFalse: [^upper].
	^self startColumnQuote, upper, self endColumnQuote.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
nameForPrimaryKeyConstraintsWhenDropping
	"Return the name we use to describe a constraint when dropping them. This is required because MySQL wants you to use 'DROP PRIMARY KEY' rather than the ANSI 'DROP CONSTRAINT'."
	^'CONSTRAINT'.! !

!DatabasePlatform methodsFor: 'sequences' stamp: ''!
newDatabaseSequenceNamed: aString
	"Return a sequence of the type we use, with the given name"
	^self databaseSequenceClass named: aString.! !

!DatabasePlatform methodsFor: 'conversion-null' stamp: ''!
nullConversion: anObject for: aType
	^anObject.! !

!DatabasePlatform methodsFor: 'conversion-null' stamp: ''!
nullConverter
	^NullConverter new
		name: #null.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
number

	^self typeNamed: #number ifAbsentPut: [GlorpNumericType new].! !

!DatabasePlatform methodsFor: 'conversion-numbers' stamp: ''!
numberToDoubleConverter
	^DelegatingDatabaseConverter
		named: #numberToDouble
		hostedBy: self
		fromStToDb: #convertToDouble:for:
		fromDbToSt: #convertToDouble:for:.! !

!DatabasePlatform methodsFor: 'conversion-numbers' stamp: ''!
numberToFloatConverter
	^DelegatingDatabaseConverter
		named: #numberToFloat
		hostedBy: self
		fromStToDb: #convertToFloat:for:
		fromDbToSt: #convertToFloat:for:.! !

!DatabasePlatform methodsFor: 'conversion-numbers' stamp: ''!
numberToIntegerConverter
	^DelegatingDatabaseConverter
		named: #numberToInteger
		hostedBy: self
		fromStToDb: #convertToInteger:for:
		fromDbToSt: #convertToInteger:for:.! !

!DatabasePlatform methodsFor: 'conversion-numbers' stamp: ''!
numberToNumberConverter
	^DelegatingDatabaseConverter
		named: #numberToNumber
		hostedBy: self
		fromStToDb: #convertToNumber:for:
		fromDbToSt: #convertToNumber:for:.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
numeric
	^self number.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
nvarchar
	^self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
padString: aString for: aType 
	| padding trimmed |
	aString isNil ifTrue: [^nil].
	aString isString ifFalse: [self error: 'Invalid data type'].
	(self usesNullForEmptyStrings and: [aString isEmpty]) ifTrue: [^nil].
	trimmed := self trimString: aString for: aType.
				
	(aType isVariableWidth or: [trimmed size = aType width]) ifTrue: [^trimmed].
	padding := String new: aType width - aString size.
	padding atAllPut: self paddingCharacter.
	^aString , padding! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
paddingCharacter
	^1 asCharacter.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
parenthesizeCommandsInCompoundQuery
	^true! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
postfixTableNameBeforeDeleting

	^false! !

!DatabasePlatform methodsFor: 'general services' stamp: ''!
predefinedKeywords
	"
		^<OrderdCollection of: String> This method returns a list of preserved keyword, which should
		not be used in database-, table or column names or any othe names in the platform system
	"

	^OrderedCollection new! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
prefixQualifierBeforeCreatingAndDeleting

	^true! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
prefixQualifierSeparatorString
	"^<String> This statement return the string to be used to separate the qualifier and the table/column name"

	^'.'! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
prefixTableNameBeforeDeleting

	^false! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
primaryKeysAreAutomaticallyUnique
	"Return false if, in addition to specifying something as a primary key, we must separately specify it as unique"

	^false.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
printBlob: aByteArray on: aStream for: aType
	aByteArray isNil ifTrue: [^aStream nextPutAll: 'NULL'].
	aStream nextPutAll: 'X'''.
	aByteArray do: [:each |
		each printOn: aStream paddedWith: $0 to: 2 base: 16].
	aStream nextPut: $'.! !

!DatabasePlatform methodsFor: 'conversion-boolean' stamp: ''!
printBoolean: aBoolean for: aType
	"If the platform is MS SQL Server, a boolean object in a SELECT needs to be converted to  BIT type, since T, Y, or true make SQLServer complain. "
	| stream |
	stream := WriteStream on: String new.
	aBoolean glorpPrintSQLOn: stream.
	^stream contents.! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
printColumn: eachGlorpDatabaseField on: sqlStatementStream
	"Print the column specification for the CREATE TABLE statement.
	Don't show a default value for serial columns, since their semanitics are different."

	sqlStatementStream
		nextPutAll: (self nameForColumn: eachGlorpDatabaseField name);
		space;
		nextPutAll: eachGlorpDatabaseField typeString;
		space;
		nextPutAll: 
				(eachGlorpDatabaseField type isSerial
					ifTrue: ['']
					ifFalse: [self sqlTextForDEFAULT: eachGlorpDatabaseField defaultValue]);
		space;
		nextPutAll:
				(eachGlorpDatabaseField isNullable
						ifTrue: [self sqlTextForNULLAttributeConstraint]
						ifFalse: [self sqlTextForNOTNULLAttributeConstraint]);
		space;
		nextPutAll:
				(eachGlorpDatabaseField isUnique
						ifTrue: [self sqlTextForUNIQUEAttributeConstraint]
						ifFalse: [self sqlTextForNOTUNIQUEAttributeConstraint]).! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
printColumnsSpecificationFor: aGlorpDatabaseTable on: sqlStatementStream
	aGlorpDatabaseTable fields isEmpty not
		ifTrue:
			[| sepFlag |
			sqlStatementStream
				space;
				nextPut: $(.
			sepFlag := false.
			aGlorpDatabaseTable fields
				do:
					[:eachGlorpDatabaseField | 
					sepFlag
						ifTrue:
							[sqlStatementStream nextPutAll: self columnNameSeparatorString].
					self printColumn: eachGlorpDatabaseField on: sqlStatementStream.
					sepFlag := true]].! !

!DatabasePlatform methodsFor: 'services tables' stamp: ''!
printDDLTableNameFor: aGLORBDatabaseTable on: sqlStatementStream
	"This method just writes the name of a table to a stream, in a way appropriate for DDL statements"
	
	sqlStatementStream
		nextPutAll: (self capitalWritingOfTableName 
				ifTrue:[ aGLORBDatabaseTable qualifiedName asUppercase ] 
				ifFalse:[ aGLORBDatabaseTable qualifiedName ]).! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
printDate: aTimestamp isoFormatOn: stream
	"Print the date as yyyy-mm-dd"
	| monthNumber dayOfMonth |
	aTimestamp isNil ifTrue: [^'NULL'].
	aTimestamp year printOn: stream.
	stream nextPut: $-.
	monthNumber := aTimestamp monthIndex.
	stream nextPutAll: (GlorpDatabaseType padToTwoDigits: monthNumber).
	stream nextPut: $-.
	dayOfMonth := aTimestamp dayOfMonth.
	stream nextPutAll: (GlorpDatabaseType padToTwoDigits: dayOfMonth).! !

!DatabasePlatform methodsFor: 'database-specific' stamp: ''!
printOffset: anInteger on: aCommand
	"By default, do nothing"! !

!DatabasePlatform methodsFor: 'database-specific' stamp: 'TestRunner 12/9/2009 18:58'!
printOffset: anInteger on: aCommand withLimit: aBoolean
	self printOffset: anInteger  on: aCommand. ! !

!DatabasePlatform methodsFor: 'database-specific' stamp: ''!
printPostLimit: anInteger on: aCommand
	"By default, do nothing"! !

!DatabasePlatform methodsFor: 'database-specific' stamp: ''!
printPreLimit: anInteger on: aCommand
	"By default, do nothing"! !

!DatabasePlatform methodsFor: 'obsolete' stamp: ''!
printPrimaryKeyConstraintsOn: sqlStatementStream for: aTable 
	"This method print the constraint specification on sqlStatementStream"

	aTable primaryKeyFields isEmpty ifTrue: [^self].
	(self createPrimaryKeyConstraintAddStatement: aTable primaryKeyConstraints)
		do: [ :each | sqlStatementStream nextPutAll: each].! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
printSqlStatementToCopyDataFromColumn: fromColumn to: toColumn on: aStream
	"
	Example:
	UPDATE test_bank_acct
	SET close_balance = balance
	WHERE close_balance IS NULL OR close_balance <> balance
	"
	| toColumnName fromColumnName |
	toColumnName := self nameForColumn: toColumn name.
	fromColumnName := self nameForColumn: fromColumn name.
	aStream
		nextPutAll: 'UPDATE ';
		nextPutAll: (self ddlTableNameFor: fromColumn table);
		nextPutAll: ' SET ';
		nextPutAll: toColumnName;
		nextPutAll: ' = ';
		nextPutAll: fromColumnName;
		nextPutAll: ' WHERE ';
		nextPutAll: toColumnName;
		nextPutAll: ' IS NULL OR ';
		nextPutAll: toColumnName;
		nextPutAll: ' <> ';
		nextPutAll: fromColumnName! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
printSqlStatementToCopyDataFromTable: fromTable to: toTable on: aStream

	aStream
		nextPutAll: 'INSERT INTO ';
		nextPutAll: (self ddlTableNameFor: toTable);
		nextPutAll: ' ( '.

	GlorpHelper 
		print: [:each | self nameForColumn: each name]
		on: aStream
		for: fromTable fields
		separatedBy: ','.


	aStream
		nextPutAll: ' ) ';
		nextPutAll: ' SELECT '.

	GlorpHelper 
		print: [:each | self nameForColumn: each name]
		on: aStream
		for: fromTable fields
		separatedBy: ','.

	aStream
		nextPutAll: ' FROM ';
		nextPutAll: (self ddlTableNameFor: fromTable)! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
printSqlStatementToPopulateColumn: aDatabaseField with: aValue on: aStream

	| converter dbValue |

	converter := aDatabaseField converterForStType: aValue class.
	dbValue := converter
		convert: aValue
		toDatabaseRepresentationAs: aDatabaseField type.

	aStream
		nextPutAll: 'UPDATE ';
		nextPutAll: (self ddlTableNameFor: aDatabaseField table);
		nextPutAll: ' SET ';
		nextPutAll: (self nameForColumn: aDatabaseField name);
		nextPutAll: ' = ';
		nextPutAll: dbValue printString;
		nextPutAll: ' WHERE (';
		nextPutAll: ( self nameForColumn: aDatabaseField name);
		nextPutAll: ' IS NULL OR ';
		nextPutAll: (self nameForColumn: aDatabaseField name);
		nextPutAll: ' <> ';
		nextPutAll: dbValue printString;
		nextPutAll: ')'! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
printSqlStatementToRenameColumn: oldColumn to: newColumnName on: aStream

	aStream
		nextPutAll: 'ALTER TABLE ';
		nextPutAll: (self ddlTableNameFor: oldColumn table);
		nextPutAll: ' RENAME ';
		nextPutAll:  (self nameForColumn:  oldColumn name);
		nextPutAll: ' TO ';
		nextPutAll:  (self nameForColumn: newColumnName).! !

!DatabasePlatform methodsFor: 'services tables' stamp: ''!
printSqlStatementToRenameTable: aDatabaseTable to: newTableName on: aStream
	" Tje default SQL statement is
	ALTER TABLE <oldTableName> RENAME TO <newTableName>.
	Different platforms may use another form. Override in subclasses if necessary.
	"

	aStream
		nextPutAll: 'ALTER TABLE ';
		nextPutAll: (self ddlTableNameFor: aDatabaseTable);
		nextPutAll: ' RENAME TO ';
		nextPutAll: newTableName! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
printTime: aTime isoFormatOn: aStream 

	self
		printTime: aTime
		isoFormatOn: aStream
		milliseconds: self supportsMillisecondsInTimes.
	^self! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
printTime: aTimestamp isoFormatOn: stream milliseconds: aBoolean
	"Print the time as hh:mm:ss.mmm"
	| ms |
	aTimestamp isNil ifTrue: [^nil].
	stream nextPutAll: (GlorpDatabaseType padToTwoDigits: aTimestamp hours).
	stream nextPut: $:.
	stream nextPutAll: (GlorpDatabaseType padToTwoDigits: aTimestamp minutes).
	stream nextPut: $:.
	stream nextPutAll: (GlorpDatabaseType padToTwoDigits: aTimestamp seconds).
	aBoolean ifFalse: [^self].
	ms := aTimestamp milliseconds.
	ms = 0 ifTrue: [^self].
	stream nextPut: $..
	stream nextPutAll: (GlorpDatabaseType padToThreeDigits: ms).! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
printTimestamp: aTimestamp on: aStream for: aType
	aTimestamp isNil ifTrue: [aTimestamp glorpPrintSQLOn: aStream. ^self].

	aStream nextPut: $'; nextPutAll: aTimestamp printString; nextPut: $'.! !

!DatabasePlatform methodsFor: 'database-specific' stamp: ''!
queryWithUnsupportedOperationsEliminatedFrom: aQuery do: aBlock
	"If aQuery has operations that we don't support, rewrite it to do them in terms of lower level operations. In particular, rewrite INTERSECT/EXCEPT operations into EXISTS clauses in a single query. Pass the new query to aBlock."
	^self.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
raw
	^self blob.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
readDate: anObject for: aType
	"format '2003-03-13"
	anObject isNil ifTrue: [^nil].
	anObject class == Date ifTrue: [^anObject].
	anObject isString ifTrue: [
		^self readDateFromStream: (ReadStream on: anObject) for: aType].
	^anObject asDate.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
readDateFromStream: aStream for: aType
	"Seems like we get to do this ourselves, in a lowest common denominator kind of way. Translate into GMT if we've got a timezone."
	"assumes ISO format.
	self readTimestamp: '2003-03-03 15:29:28.337-05' for: nil.
	self readTimestamp: '2003-03-03 19:29:28.337-05' for: nil

"
	|  years months days |
	years := (aStream upTo: $-) asNumber.
	months := (aStream upTo: $-) asNumber.
	days := (aStream upTo: $ ) asNumber.
	^Dialect newDateWithYears: years months: months days: days.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
readTime: anObject for: aType
	"format 15:29:28.337-05  (timezone optional)"
	anObject isNil ifTrue: [^nil].
	anObject class == Time ifTrue: [^anObject].
	anObject isString ifTrue: [
		^self readTimeFromStream: (ReadStream on: anObject) for: aType].
	^anObject asTime.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
readTimeFromStream: aStream for: aType
	"Seems like we get to do this ourselves, in a lowest common denominator kind of way. Ignore timezones right now"
	"assumes ISO format.
	self readTimestamp: '2003-03-03 15:29:28.337-05' for: nil.
	self readTimestamp: '2003-03-03 19:29:28.337-05' for: nil

"
	| hours minutes seconds milliseconds timeZoneOffset millisecondAccumulator |
	hours := (aStream upTo: $:) asNumber.
	minutes := (aStream upTo: $:) asNumber.
	seconds := (aStream next: 2) asNumber.
	(aStream peek = $.) 
		ifTrue: [
			aStream next.
			millisecondAccumulator := WriteStream on: String new.
			[aStream atEnd not and: [aStream peek isDigit]] whileTrue: [
				millisecondAccumulator nextPut: aStream next].
			milliseconds := millisecondAccumulator contents asNumber]
		ifFalse: [milliseconds := 0].
	timeZoneOffset := aStream upToEnd asNumber.
	^Dialect newTimeWithHours: hours minutes: minutes seconds: seconds milliseconds: milliseconds.
	"^Dialect addSeconds: (timeZoneOffset * -1* 60 * 60) to: aTime."! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
readTimestamp: anObject for: aType
	"Seems like we get to do this ourselves, in a lowest common denominator kind of way. Translate into GMT if we've got a timezone."

	"assumes ISO format.
	self readTimestamp: '2003-03-03 15:29:28.337-05' for: nil.
	self readTimestamp: '2003-03-03 19:29:28.337-05' for: nil"

	anObject isNil ifTrue: [^nil].
	anObject class == Dialect timestampClass ifTrue: [^anObject].
	anObject isString
		ifTrue:
			[| stream |
			stream := ReadStream on: anObject.
			^self readTimestampFromStream: stream for: aType].
	^anObject asTimestamp.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
readTimestampFromStream: aStream for: aType

	| years months days hours minutes seconds millisecondAccumulator milliseconds timeZoneOffset |
	years := (aStream upTo: $-) asNumber.
	months := (aStream upTo: $-) asNumber.
	days := (aStream upTo: $ ) asNumber.
	hours := (aStream upTo: $:) asNumber.
	minutes := (aStream upTo: $:) asNumber.
	seconds := (aStream next: 2) asNumber.
	(aStream peek = $.) 
		ifTrue: [
			aStream next.
			millisecondAccumulator := WriteStream on: String new.
			[aStream atEnd not and: [aStream peek isDigit]] whileTrue: [
				millisecondAccumulator nextPut: aStream next].
			milliseconds := millisecondAccumulator contents asNumber]
		ifFalse: [milliseconds := 0].
	timeZoneOffset := aStream upToEnd asNumber.
	timeZoneOffset isNil ifTrue: [timeZoneOffset := 0].
	timeZoneOffset := timeZoneOffset * 60 * 60.
	^Dialect 
		newTimestampWithYears: years
		months: months
		days: days
		hours: hours
		minutes: minutes
		seconds: seconds
		milliseconds: milliseconds
		offset:timeZoneOffset.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
requiresCastsForNullsInUnions
	"If we have explicit nulls in a query which is unioned, do we need to cast them to the appropriate type in order to keep the database happy."
	^false.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
requiresEscapeFor: aCharacter
	^self charactersThatNeedEscaping includes: aCharacter.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
requiresTransactionForTableOperations
	"Return true if we need a transaction in order to do table creation, deletion, and so forth. Some databases require it. Others don't like it"
	^true.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
reservedWords
	reservedWords isNil ifTrue: [self initializeReservedWords].
	^reservedWords.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
reusePreparedStatements
	"The ordinary choice is to initialize the accessor to reuse statements whenever binding is being used."

	^true! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
sequence

	^self subclassResponsibility! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
serial
	self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
smallint

	^self typeNamed: #smallint ifAbsentPut: [GlorpSmallIntType new].! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
sqlStatementToCopyDataFromColumn: fromDatabaseField to: toDatabaseField

	| stream |

	stream := WriteStream on: String new.

	self
		printSqlStatementToCopyDataFromColumn: fromDatabaseField
		to: toDatabaseField
		on: stream.

	^ stream contents! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
sqlStatementToCopyDataFromTable: fromDatabaseTable to: toDatabaseTable

	| stream |

	stream := WriteStream on: String new.

	self
		printSqlStatementToCopyDataFromTable: fromDatabaseTable
		to: toDatabaseTable
		on: stream.

	^ stream contents! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
sqlStatementToPopulateColumn: aDatabaseField with: aValue

	| stream |

	stream := WriteStream on: String new.

	self
		printSqlStatementToPopulateColumn: aDatabaseField
		with: aValue
		on: stream.

	^ stream contents! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
sqlStatementToRenameColumn: oldColumn to: newColumnName

	| stream |

	stream := WriteStream on: String new.

	self
		printSqlStatementToRenameColumn: oldColumn
		to: newColumnName
		on: stream.

	^ stream contents! !

!DatabasePlatform methodsFor: 'services tables' stamp: ''!
sqlStatementToRenameTable: aDatabaseTable to: newTableName

	| stream |

	stream := WriteStream on: String new.

	self printSqlStatementToRenameTable: aDatabaseTable to: newTableName on: stream.

	^ stream contents! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
sqlTextCurrentServerUTCTimestamp

	self subclassResponsibility! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
sqlTextForBeginTransaction
	"comment"

	^'BEGIN'! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
sqlTextForDEFAULT: defaultValue
	"^<String>. Answer the column DEFAULT clause as a string with the given value.
	This is used in CREATE TABLE statements as part of each column specification."
	
	| strm |
	^defaultValue
		ifNotNil:
			[strm := String new writeStream.
			strm nextPutAll: 'DEFAULT '.
			defaultValue glorpPrintSQLOn: strm.
			strm contents]
		ifNil: ['']! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
sqlTextForNOTNULLAttributeConstraint
	"^<String>"

	^'NOT NULL'! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
sqlTextForNOTNULLWithDefaultAttributeConstraint
	"^<String>"

	^'NOT NULL WITH DEFAULT'! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
sqlTextForNOTUNIQUEAttributeConstraint
	^''.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
sqlTextForNULLAttributeConstraint
	"^<String>"

	^'NULL'! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
sqlTextForUNIQUEAttributeConstraint
	^'UNIQUE'.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
sqlWildcardForMultipleCharacters
	"^<String> This method returns the used wildcard string for multiple characters"
	
	^'%'! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
sqlWildcardForSingleCharacter
	"^<String> This method returns the used wildcard string for single characters"
	
	^'_'! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
startColumnAddString
	"The string we use as a prefix when adding a colum"
	^' COLUMN '.! !

!DatabasePlatform methodsFor: 'services columns' stamp: ''!
startColumnQuote
	"Return the character used to start quoting the name of a column so that we can avoid conflicts with reserved names. ANSI says to use double quotes, so that's the default."
	^'"'.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
stringToByteArray: aString for: aType
	| string |
	string := self unpadString: aString for: aType.
	^string isNil ifTrue: [nil] ifFalse: [Dialect convert: string toByteArrayWithEncoding: self characterEncoding].! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
stringToInteger: aString for: aType
	| string |
	string := self unpadString: aString for: aType.
	^string isNil ifTrue: [nil] ifFalse: [string asNumber truncated].! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
stringToStringConverter
	^DelegatingDatabaseConverter
		named: #stringToString
		hostedBy: self
		fromStToDb: #padString:for:
		fromDbToSt: #unpadString:for:.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
stringToSymbol: aString for: aType
	| string |
	string := self unpadString: aString for: aType.
	^string isNil ifTrue: [nil] ifFalse: [string asSymbol].! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsANSIJoins
	"Do we support the JOIN <tableName> USING <criteria> syntax. Currently hard-coded, but may also vary by database version"
	^false.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsAliasInCompoundQuery
	"Return true if, when we have a subselect, we can (and/or must) alias the result with a table name. Access doesn't like the table name"
	^true.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsAlterTableForConstraints

	^true! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
supportsAuthentication
	^true! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsBinding
	"Return true if this platform supports binding parameters rather than printing them as strings into the SQL statement"
	^false.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsCaseInsensitiveLike
	^false.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
supportsConstraints
	"Return true if we support integrity constraints. That is, foreign key, primary key or unique constraints."
	^true.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsDecimalsOnAllNumerics
	"Return true if a general 'numeric' type will allow numbers after the decimal place"
	^true.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsGroupWritingFor: aCommand
	"Return true if this platform can do array binding or an equivalent for this command"
	^false.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsInformationSchema
	^true.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsIntersect
	"Do we support the INTERSECT or MINUS operations."
	^true.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsLimit
	"Do we support anything analogous to the postgresql LIMIT, returning only the first N rows"
	^false.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
supportsMillisecondsInTimes
	self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsMultipleOpenCursors
	"Can this database support multiple open cursors at once"
	^true.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsOffset
	"Do we support anything analogous to the postgresql OFFSET, skipping the first N rows"
	^false.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsRowCount
	^true.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsSchemas
	"Return true if this platform supports schemas. Access and SQLite 3 do not."

	^true! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsStringQueryOnBlobs
	^true.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
supportsTableOwners
	"Return true if this platform supports table owners, i.e. expects table names of the form Bern.TW_* rather than just TW_* in its SQL."
"Access, Firebird and PostGreSQL do not, Oracle does, others I know not."

	^false! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
supportsVariableSizedNumerics
	"Return true if this platform can support numbers with a varying size and number of decimal places. Access, notably, doesn't seem to be able to"
	^true.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
symbolToString: aSymbol for: aType
	aSymbol isNil ifTrue: [^nil].
	^self padString: aSymbol asString for: aType.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
symbolToStringConverter
	^DelegatingDatabaseConverter
		named: #symbolToString
		hostedBy: self
		fromStToDb: #symbolToString:for:
		fromDbToSt: #stringToSymbol:for:.! !

!DatabasePlatform methodsFor: 'services tables' stamp: ''!
tableNameFor: aStringRepresentingATableName
	"Return the table name, as modified to be used in a database statement."
	^aStringRepresentingATableName.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
text
	^self clob.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
time
	self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
timeConverter
	^DelegatingDatabaseConverter
		named: #time
		hostedBy: self
		fromStToDb: #toTime:for:
		fromDbToSt: #readTime:for:.	"#printTime:for:"! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
timestamp
	self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
timestampConverter
	^DelegatingDatabaseConverter
		named: #timestamp
		hostedBy: self
		fromStToDb: #toTimestamp:for:
		fromDbToSt: #readTimestamp:for:.	"#printTimestamp:for:"! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
timestampToDateConverter
	^DelegatingDatabaseConverter
		named: #timestampToDate
		hostedBy: self
		fromStToDb: #toTimestamp:for:
		fromDbToSt: #readDate:for:.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
toDate: anObject for: aType
	anObject isNil ifTrue: [^nil].
	anObject class = Date ifTrue: [^anObject].
	^anObject asDate.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
toGlorpType: aGlorpType fromString: aString
	"Return a Smalltalk instance of the type indicated by aGlorpType."
	
	| aClass |
	^aString
		ifNotNil:
			[aClass := aGlorpType impliedSmalltalkType.
			aClass readFrom: aString readStream]! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: ''!
toTime: anObject for: aType
	anObject isNil ifTrue: [^nil].
	anObject class = Time ifTrue: [^anObject].
	^anObject asTime.! !

!DatabasePlatform methodsFor: 'conversion-times' stamp: 'noha 11/20/2007 22:36'!
toTimestamp: anObject for: aType	
	anObject isNil ifTrue: [^nil].
	anObject class = Dialect timestampClass ifTrue: [^anObject].

	^anObject asTimeStamp.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
trimString: aString for: aType 
	| trimmedString length excess |
	aType width isNil ifTrue: [^aString].
	self measuresStringFieldWidthsInBytes ifFalse: [
		^aString size > aType width ifTrue: [aString copyFrom: 1 to: aType width] ifFalse: [aString]].
	"If our characters are less than half the number of bytes it can hold, assume it will fit and don't try to do this"
	(aString size < (aType width // 2)) ifTrue: [^aString].
	"If we're doing defaults that don't map characters to multi-bytes, don't do this"
	(#(ascii default) includes: self characterEncoding) ifTrue: [
		^aString size > aType width ifTrue: [aString copyFrom: 1 to: aType width] ifFalse: [aString]].
	trimmedString := aString.
	[length := (Dialect convert: trimmedString toByteArrayWithEncoding: self characterEncoding) size.
	excess := length - aType width.
	excess > 0 
		"Try to be careful and trim as little as possible, while still keeping the number of steps reasonable for large excesses"
		ifTrue: [trimmedString := trimmedString copyFrom: 1 to: trimmedString size - ((excess // 2) max: 1)]
		ifFalse: [^trimmedString].
	true] whileTrue.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
typeNamed: aSymbol ifAbsentPut: aBlock

	| type |
	type := self types at: aSymbol ifAbsentPut: [
		| newType |
		newType := aBlock value.
		newType platform: self.
		newType selector: aSymbol].
	type hasParameters ifTrue: [type := type copy].
	^type.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
typeWithDatabaseName: aString characterWidth: aNumber
	"Return our type which matches the database name for the type. Assumes all variable width fields are strings right now."
	| baseType matchingTypes simpleTypes |
	matchingTypes := self typesWithValues values asArray select: [:each | each typeName asLowercase = aString asLowercase].
	matchingTypes isEmpty ifTrue: [
		Transcript cr; show: 'Cannot find type ', aString, ' using integer instead.'. matchingTypes := Array with: self integer].
	"We may get generated and non-generated both matching, e.g. inMemorySequence and int. Prefer the non-generated."
	simpleTypes := matchingTypes reject: [:each | each isGenerated].
	baseType := simpleTypes isEmpty ifTrue: [matchingTypes first] ifFalse: [simpleTypes first].
	^baseType hasWidth ifTrue: [baseType copy width: aNumber] ifFalse: [baseType].! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
typeWithDatabaseName: aString characterWidth: aNumber extraInformation: anythingElse
	"Return our type which matches the database name for the type. Assumes all variable width fields are strings right now."
	^self typeWithDatabaseName: aString characterWidth: aNumber.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
types
	types == nil ifTrue: [types := IdentityDictionary new].
	^types.! !

!DatabasePlatform methodsFor: 'type helpers' stamp: ''!
typesWithValues
	"Return our types list, but makes sure it has at least basic types in it"
	self varchar.
	self int4.
	self integer.
	self int.
	self number.
	self timestamp.
	self char.
	self blob.
	self clob.
	self raw.
	self serial.
	self text.
	^self types.! !

!DatabasePlatform methodsFor: 'conversion-strings' stamp: ''!
unpadString: aString for: aType

	aString isNil ifTrue: [^nil].
	^aType isVariableWidth 
		ifTrue: [aString]
		ifFalse: [(ReadStream on: aString) upTo: 1 asCharacter].! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
useBinding
	"Return true if we should use binding"

	self supportsBinding ifFalse: [^false].
	useBinding isNil ifTrue: [useBinding := self class useBindingIfSupported].
	^useBinding! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
useBinding: aBoolean
	useBinding := aBoolean.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
useMicrosoftOuterJoins
	"Return true if we use the Microsoft x *= y syntax for outer joins"
	^false.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
useOracleOuterJoins
	"Return true if we use the old Oracle x = y (+) syntax for outer joins"
	^false.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
usesArrayBindingRatherThanGrouping
	"Return true if we use array binding for grouped writes rather than printing the sql multiple times. Only applies if we support grouped writes"
	^self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
usesIdentityColumns
	^false.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
usesNullForEmptyStrings
	"Return true if this database is likely to use nil as an empty string value"
	^false.! !

!DatabasePlatform methodsFor: 'constants' stamp: ''!
usesNullForFalse
	"Return true if this database is likely to use nil as an empty string value"
	^false.! !

!DatabasePlatform methodsFor: 'testing' stamp: ''!
usesPrimaryKeyUniqueConstraintNames
	"When dropping/creating primary key constraints, do they have distinguished names."
	^true.! !

!DatabasePlatform methodsFor: 'services tables' stamp: ''!
validateTableName: tableNameString
	" <Boolean> I return true, if the choosen tableNameString is valid for the platform"

 	^( tableNameString size <= self maximumLengthOfTableName ) 
			and: [ (self predefinedKeywords includes: tableNameString asLowercase) not ]! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
varChar
	^self varchar.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
varChar: anInt

	^self varchar width: anInt.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
varbinary
	^self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
varbinary: anInteger
	^self varbinary width: anInteger.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
varchar
	^self subclassResponsibility.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
varchar: anInt

	^self varchar width: anInt.! !

!DatabasePlatform methodsFor: 'types' stamp: ''!
versionFieldFor: aType
	"Don't cache this, because we want to return different instances depending on aType"
	^(VersionType underlyingType: aType) generatesOverExistingValues: true; platform: self.! !

!DatabasePlatform methodsFor: 'accessing' stamp: ''!
vwConnectionClass
	^self isODBCPlatform 
			ifTrue: [self vwODBCConnectionClass]
			ifFalse: [self vwEXDIConnectionClass]! !

!DatabasePlatform methodsFor: 'accessing' stamp: ''!
vwEXDIConnectionClass
	^self subclassResponsibility! !

!DatabasePlatform methodsFor: 'accessing' stamp: ''!
vwODBCConnectionClass
	^'ODBCThreadedConnection' asQualifiedReference valueIfUndefinedLoadFrom: 'ODBCThapiEXDI'! !

!DatabasePlatform methodsFor: 'services constraints' stamp: ''!
writeConstraintToDrop: aConstraint onStream: aStream
	aStream
		nextPutAll: 'CONSTRAINT ';
		nextPutAll: aConstraint name.! !

DatabasePlatform subclass: #MySQLPlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!MySQLPlatform commentStamp: '<historical>' prior: 0!
This is a definition for MySQL databases. It's not done yet, just here as a placeholder.!

MySQLPlatform subclass: #MySQLODBCPlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!MySQLODBCPlatform methodsFor: 'testing' stamp: ''!
isODBCPlatform

	^true.! !

!MySQLPlatform classSide methodsFor: 'eoglorp' stamp: ''!
createLoginFromConnectionDictionary: aDict
	" *** This method was defined by Glorp.DatabasePlatform class as a subclass responsibility.
	Replace its body with a proper implementation. *** "

	self error: 'Subclass responsibility stub not reimplemented'! !

!MySQLPlatform methodsFor: 'constants' stamp: ''!
areSequencesExplicitlyCreated
	^false.! !

!MySQLPlatform methodsFor: 'binding' stamp: ''!
bindingsForGroupWritingFor: aCommand
	"Return the bindings array for a group write. This can be in different formats, depending on the database and perhaps the mechanism in place."
	^aCommand batchStatementBindings.! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
blob
	"Return the MySQL type for long blob."
	^self typeNamed: #blob ifAbsentPut: [GlorpBlobType new typeString: 'LONGBLOB'].! !

!MySQLPlatform methodsFor: 'testing' stamp: ''!
canUseInWhereClause: aValue type: aType
	"This is mostly for test support purposes. If we know we can't use a particular kind of value in the where clause on some database, don't try."

	(super canUseInWhereClause: aValue type: aType) ifFalse: [^false].
	"On MySQL we don't seem to be able to compare constants against blobs."
	aType = self blob ifTrue: [^false].
	^true.! !

!MySQLPlatform methodsFor: 'constants' stamp: ''!
charactersThatNeedEscaping
	^#($' $\)! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
clob
	"Return the MySQL type for long text."
	^self typeNamed: #clob ifAbsentPut: [GlorpMySQLClobType new typeString: 'LONGTEXT'].! !

!MySQLPlatform methodsFor: 'sequences' stamp: ''!
databaseSequenceClass
	^MySQLSequence.! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
date
	"Return the type we use to represent dates."
	^self typeNamed: #date ifAbsentPut: [GlorpDateType new].! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
double
	^self typeNamed: #double ifAbsentPut: [GlorpDoubleType new typeString: 'double' ].! !

!MySQLPlatform methodsFor: 'services tables' stamp: ''!
endColumnQuote
	"Return the character used to start quoting the name of a column so that we can avoid conflicts with reserved names. The default MySQL quoting is to use backquotes rather than the ANSI style."
	^'`'.! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
float
	^self typeNamed: #float ifAbsentPut: [GlorpFloatType new typeString: 'float' ].! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
float4
	^self float.! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
float8
	^self double.! !

!MySQLPlatform methodsFor: 'constants' stamp: ''!
initializeReservedWords
	reservedWords := Set new.
	reservedWords 
		add: 'key'.! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
int2
	^self smallint! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
int4
	^self typeNamed: #int4 ifAbsentPut: [GlorpIntegerType new typeString: 'int'].! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
int8
	^self typeNamed: #int8 ifAbsentPut: [GlorpIntegerType new typeString: 'bigint'].! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
integer

	^self typeNamed: #integer ifAbsentPut: [GlorpIntegerType new typeString: 'int'].! !

!MySQLPlatform methodsFor: 'testing' stamp: ''!
isMySQLPlatform
	^true.! !

!MySQLPlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfColumnName
	"^<Integer> I return the maximum length of a column name.  (I am not sent at present;  I am here so that if I and #maximumLengthOfTableName (which is sent) should ever differ on some database platform, I can be used in revised truncation logic.)
	Don't know the answer for this platform;  make something large up."

	^64! !

!MySQLPlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfTableName
	"^<Integer> I return the max. length of a table name"
	"Don't know, make something large up."
	^64! !

!MySQLPlatform methodsFor: 'constants' stamp: ''!
nameForPrimaryKeyConstraintsWhenDropping
	"Return the name we use to describe a constraint when dropping them. This is required because MySQL wants you to use 'DROP PRIMARY KEY' rather than the ANSI 'DROP CONSTRAINT'."
	^'PRIMARY KEY'.! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
number

	^self typeNamed: #number ifAbsentPut: [GlorpNumericType new typeString: 'decimal'].! !

!MySQLPlatform methodsFor: 'conversion-times' stamp: ''!
printDate: aDate for: aType
	| stream |
	aDate isNil ifTrue: [^'NULL'].
	stream := WriteStream on: String new.
	stream nextPut: $".
	self
		printDate: aDate
		isoFormatOn: stream.
	stream nextPut: $".
	^stream contents.! !

!MySQLPlatform methodsFor: 'conversion-times' stamp: ''!
printTime: aTime for: aType
	| stream |
	aTime isNil ifTrue: [^'NULL'].
	stream := WriteStream on: String new.
	stream nextPut: $".
	self
		printTime: aTime
		isoFormatOn: stream
		milliseconds: false.
	stream nextPut: $".
	^stream contents.! !

!MySQLPlatform methodsFor: 'conversion-times' stamp: ''!
printTimestamp: aTimestamp on: stream for: aType
	aTimestamp isNil ifTrue: [stream nextPutAll: 'NULL'. ^self].
	stream nextPut: $".
	self
		printDate: aTimestamp
		isoFormatOn: stream.
	stream nextPutAll: ' '.
	self
		printTime: aTimestamp
		isoFormatOn: stream
		milliseconds: false.
	stream nextPut: $".! !

!MySQLPlatform methodsFor: 'database-specific' stamp: ''!
queryWithUnsupportedOperationsEliminatedFrom: aQuery do: aBlock
	"If aQuery has operations that we don't support, rewrite it to do them in terms of lower level operations. In particular, rewrite INTERSECT/EXCEPT operations into EXISTS clauses in a single query. Pass the new query to aBlock."
	| newQuery |
	newQuery := aQuery rewriteIntersect.
	newQuery := newQuery rewriteExcept.
	newQuery == aQuery ifFalse: [aBlock value: newQuery].! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
sequence
	^self serial.! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
serial

	^self typeNamed: #serial ifAbsentPut: [GlorpSerialType new typeString: 'int AUTO_INCREMENT'].! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
smallint

	^self typeNamed: #smallint ifAbsentPut: [GlorpSmallIntType new typeString: 'SMALLINT'].! !

!MySQLPlatform methodsFor: 'services tables' stamp: ''!
startColumnQuote
	"Return the character used to start quoting the name of a column so that we can avoid conflicts with reserved names. The default MySQL quoting is to use backquotes rather than the ANSI style."
	^'`'.! !

!MySQLPlatform methodsFor: 'conversion-strings' stamp: ''!
stringToByteArrayOneWayConverter
	"This is a special converter for MySQL that is a null conversion on the way out, but converts the byte arrays that we get back from blobs into strings."
	^DelegatingDatabaseConverter
		named: #stringToByteArrayOneWay
		hostedBy: self
		fromStToDb: #nullConversion:for:
		fromDbToSt: #byteArrayToString:for:.! !

!MySQLPlatform methodsFor: 'testing' stamp: ''!
supportsANSIJoins
	"Do we support the JOIN <tableName> USING <criteria> syntax. Currently hard-coded, but may also vary by database version"
	^true.! !

!MySQLPlatform methodsFor: 'testing' stamp: ''!
supportsDecimalsOnAllNumerics
	"Return true if a general 'numeric' type will allow numbers after the decimal place"
	^false.! !

!MySQLPlatform methodsFor: 'testing' stamp: ''!
supportsIntersect
	"Do we support operations like UNION ALL And INTERSECT."
	^false.! !

!MySQLPlatform methodsFor: 'constants' stamp: ''!
supportsMillisecondsInTimes
	^false.! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
time
	^self typeNamed: #time ifAbsentPut: [GlorpTimeType new typeString: 'time'].! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
timestamp
	"Although MySQL has a type called 'timestamp', 'datetime' is more suitable."
	^self typeNamed: #timestamp ifAbsentPut: [GlorpTimeStampType new typeString: 'DATETIME'].! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
typeWithDatabaseName: aString characterWidth: aNumber extraInformation: extraValue
	"Return our type which matches the database name for the type. Assumes all variable width fields are strings right now."
	extraValue = 'auto_increment' ifTrue: [^self serial].
	^super typeWithDatabaseName: aString characterWidth: aNumber.! !

!MySQLPlatform methodsFor: 'testing' stamp: ''!
usesArrayBindingRatherThanGrouping
	^false.! !

!MySQLPlatform methodsFor: 'testing' stamp: ''!
usesIdentityColumns
	^true.! !

!MySQLPlatform methodsFor: 'testing' stamp: ''!
usesPrimaryKeyUniqueConstraintNames
	"When dropping/creating primary key constraints, do they have distinguished names."
	^false.! !

!MySQLPlatform methodsFor: 'types' stamp: ''!
varchar
	^self typeNamed: #varchar ifAbsentPut: [GlorpVarCharType new].! !

!MySQLPlatform methodsFor: 'accessing' stamp: ''!
vwEXDIConnectionClass
	^'MySQLConnection' asQualifiedReference valueIfUndefinedLoadFrom: 'MySQLEXDI'! !

!MySQLPlatform methodsFor: 'services constraints' stamp: ''!
writeConstraintToDrop: aConstraint onStream: aStream
	"Return the term we use to describe a constraint when dropping them.  This is required because MySQL wants you to use
		'DROP PRIMARY KEY' for primary key constraints (and then says don't name the constraint name in the drop statement!!)
		'DROP FOREIGN KEY'  for foreign key constraints
		'DROP KEY' for unique constraints (it also supports DROP INDEX for this)
	rather than the ANSI 'DROP CONSTRAINT'."

	aConstraint isForeignKeyConstraint ifTrue: [aStream nextPutAll: 'FOREIGN KEY'].
	aConstraint isPrimaryKeyConstraint ifTrue: [aStream nextPutAll: 'PRIMARY KEY'].
	aConstraint isUniqueConstraint ifTrue: [aStream nextPutAll: 'KEY'].
	aConstraint isPrimaryKeyConstraint ifFalse:
		[aStream
			nextPut: $ ;
			nextPutAll: aConstraint name].! !

DatabasePlatform subclass: #OcelotPlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!OcelotPlatform commentStamp: '<historical>' prior: 0!
This is a platform for the open-source Ocelot database.!

OcelotPlatform subclass: #FirebirdPlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!FirebirdPlatform commentStamp: '<historical>' prior: 0!
This is a Glorp platform for the InterBase and Firebird databases.  It is currently subclassed from Ocelot because both claim to be highly standard-compliant.  It would probably be better if both inherited from some sort of abstract SQL92Platform.

Firebird/InterBase supports binding in all statements (indeed its EXDI appears to require it for blobs) and can also reuse SELECT statements, but reuse of RowBasedCommands (INSERT, DELETE, UPDATE) for blobs is unreliable although binding of such statements is essential.  Set
	reusePreparedStatements: false
on Logins that have an instance of this class as their platform.
!

!FirebirdPlatform methodsFor: 'constants' stamp: ''!
areSequencesExplicitlyCreated
	^true! !

!FirebirdPlatform methodsFor: 'binding' stamp: ''!
bindingsForGroupWritingFor: aCommand
	"Return the bindings array for a group write. This can be in different formats, depending on the database and perhaps the mechanism in place."
	^aCommand batchStatementBindings.! !

!FirebirdPlatform methodsFor: 'conversion' stamp: ''!
blobDataToStreamConverter
	^DelegatingDatabaseConverter
		named: #blobDataToStream
		hostedBy: self
		fromStToDb: #convertBlobData:for:
		fromDbToSt: #nullConversion:for:.! !

!FirebirdPlatform methodsFor: 'types' stamp: ''!
clob

	^self typeNamed: #clob ifAbsentPut: [GlorpClobType new typeString: 'blob sub_type text'].! !

!FirebirdPlatform methodsFor: 'conversion' stamp: ''!
convertBlobData: aData for: aType
	"From standard Store getBlobData called in writeSession;  Firebird seems to need it.  Glorp needs to think that two such streams on the same data are equal."

	^EqualOnContentReadStream on: aData! !

!FirebirdPlatform methodsFor: 'sequences' stamp: ''!
databaseSequenceClass
	^FirebirdSequence.! !

!FirebirdPlatform methodsFor: 'constants' stamp: ''!
initializeReservedWords
	reservedWords := Set new.
	reservedWords 
		add: 'type';
		add: 'value';
		add: 'password';
		add: 'unique';
		add: 'date';
		add: 'current';
		add: 'order';
		add: 'active';
		add: 'action';
		add: 'database';
		add: 'table';
		add: 'timestamp';
		add: 'position'.! !

!FirebirdPlatform methodsFor: 'testing' stamp: ''!
isODBCPlatform
	"IBEXDI defines an EXDI connection class, not an ODBC one."
 
	^false! !

!FirebirdPlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfColumnName
	"^<Integer> I return the maximum length of a column name.  (I am not sent at present;  I am here so that if I and #maximumLengthOfTableName (which is sent) should ever differ on some database platform, I can be used in revised truncation logic.)"

	^31! !

!FirebirdPlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfTableName
	"^<Integer> I return the max. length of a table name"

	^31! !

!FirebirdPlatform methodsFor: 'constants' stamp: ''!
requiresTransactionForTableOperations
	"Return true if we need a transaction in order to do table creation, deletion, and so forth. Some databases require it. Others don't like it"
	^false.! !

!FirebirdPlatform methodsFor: 'testing' stamp: ''!
reusePreparedStatements
	"Firebird/InterBase supports binding of parameters (i.e. instead of printing them as strings within the SQL statement);  indeed, when handling Blobs, the current IBEXDI implementation demands it.  However if RowBasedCommands (INSERT, DELETE, UPDATE) attempt to reuse bound statements, there are failures;  every 79th or 358th or whatever reuse of such a statement returns bad data.  (SELECT statements are fine.)  This was hidden when failure to release the GlorpCursor in RowBasedCommands left statements busy, so unreusable;  the fix (first made in Glorp 0.3.181) revealed it.  We return false here so that reusePreparedStatements is initialized to false for any accessor whose login uses this platform."

	^false

"To study further:  is it that prepared-statement-signature-matching needs to distingush the types of the contents of an EqualOnContentsReadStream, or does the database need a bug fix, or does the IBEXDI need to handle Blobs differently?  If the problems with RowBaseCommands cannot be resolved, can we / should we distinguish between reusing them and reusing SELECT statements?"! !

!FirebirdPlatform methodsFor: 'types' stamp: ''!
sequence

	^self typeNamed: #sequence ifAbsentPut: [GlorpSerialType new typeString: (self integer typeString)].! !

!FirebirdPlatform methodsFor: 'testing' stamp: ''!
supportsBinding
	"Firebird/InterBase supports binding in all statements (indeed its EXDI appears to require it for blobs) and can also reuse SELECT statements, but reuse of RowBasedCommands (INSERT, DELETE, UPDATE) for Blobs is unreliable although binding of such statements is essential (see my #reusePreparedStatementsIfSupported method comment)."

	^true! !

!FirebirdPlatform methodsFor: 'constants' stamp: ''!
supportsMillisecondsInTimes
	^false.! !

!FirebirdPlatform methodsFor: 'accessing' stamp: ''!
vwEXDIConnectionClass
	^'InterBaseConnection' asQualifiedReference valueIfUndefinedLoadFrom: 'IBEXDI'! !

!OcelotPlatform methodsFor: 'constants' stamp: ''!
areSequencesExplicitlyCreated

^false! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
boolean
	^self typeNamed: #boolean ifAbsentPut: [GlorpBooleanType new typeString: 'boolean'].! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
clob

	^self typeNamed: #clob ifAbsentPut: [GlorpClobType new typeString: 'clob'].! !

!OcelotPlatform methodsFor: 'constants' stamp: ''!
databaseSequenceClass
	^InMemorySequence.! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
date
	^self typeNamed: #date ifAbsentPut: [GlorpDateType new].! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
double
	^self float8! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
float
	^self float4.! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
float4
	^self typeNamed: #float4 ifAbsentPut: [GlorpFloatType new typeString: 'float'].! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
float8
	^self typeNamed: #float8 ifAbsentPut: [GlorpDoubleType new typeString: 'double precision'].! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
int2
	^self smallint.! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
int4

	^self integer.! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
int8

	^self integer! !

!OcelotPlatform methodsFor: 'testing' stamp: ''!
isODBCPlatform

	^true! !

!OcelotPlatform methodsFor: 'testing' stamp: ''!
isOcelotPlatform

	^true! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
numeric
	^self typeNamed: #numeric ifAbsentPut: [GlorpNumericType new].! !

!OcelotPlatform methodsFor: 'conversion-strings' stamp: ''!
padString: aString for: aType 
	| padding |
	aString isNil ifTrue: [^nil].
	(self usesNullForEmptyStrings and: [aString isEmpty]) ifTrue: [^nil].
	aString size > aType width 
				ifTrue: [^aString copyFrom: 1 to: aType width].
				
	aType isVariableWidth ifTrue: [^aString].
	padding := String new: aType width - aString size.
	padding atAllPut: 32 asCharacter.
	^aString , padding! !

!OcelotPlatform methodsFor: 'conversion-strings' stamp: ''!
paddingCharacter
	^32 asCharacter.! !

!OcelotPlatform methodsFor: 'constants' stamp: ''!
primaryKeysAreAutomaticallyUnique

	^true.! !

!OcelotPlatform methodsFor: 'conversion times' stamp: ''!
printDate: aDate for: aType
	| stream |
	aDate isNil ifTrue: [^nil].
	stream := WriteStream on: String new.
	stream nextPutAll: 'date '''.
	self
		printDate: aDate
		isoFormatOn: stream.
	stream nextPutAll: ''''.
	^stream contents.! !

!OcelotPlatform methodsFor: 'conversion times' stamp: ''!
printTime: aTime for: aType
	| stream |
	aTime isNil ifTrue: [^nil].
	stream := WriteStream on: String new.
	stream nextPutAll: 'time '''.
	self
		printTime: aTime
		isoFormatOn: stream.
	stream nextPutAll: ''''.
	^stream contents.! !

!OcelotPlatform methodsFor: 'conversion times' stamp: ''!
printTimestamp: aTimestamp on: stream for: aType

	aTimestamp isNil ifTrue: [stream nextPutAll: 'NULL'. ^self].
	stream nextPutAll: 'timestamp '''.
	self
		printDate: aTimestamp
		isoFormatOn: stream.
	stream nextPut: $ .
	self
		printTime: aTimestamp
		isoFormatOn: stream.
	"	milliseconds: true."
	stream nextPutAll: ''''.! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
sequence

	^self inMemorySequence! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
serial

	^self sequence! !

!OcelotPlatform methodsFor: 'constants' stamp: ''!
sqlTextForNULLAttributeConstraint
	"^<String>"

	^''! !

!OcelotPlatform methodsFor: 'testing' stamp: ''!
supportsANSIJoins
	"Do we support the JOIN <tableName> USING <criteria> syntax. Currently hard-coded, but may also vary by database version"

	^true! !

!OcelotPlatform methodsFor: 'constants' stamp: ''!
supportsMillisecondsInTimes
	"I'm guessing here"
	^true.! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
text
	^self clob.! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
time
	^self typeNamed: #time ifAbsentPut: [GlorpTimeType new typeString: 'time'].! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
timestamp
	^self typeNamed: #timestamp ifAbsentPut: [GlorpTimeStampType new typeString: 'timestamp'].! !

!OcelotPlatform methodsFor: 'conversion-strings' stamp: ''!
unpadString: aString for: aType

	aString isNil ifTrue: [^nil].
	^aType isVariableWidth 
		ifTrue: [aString]
		ifFalse: [(ReadStream on: aString) upTo: 32 asCharacter].! !

!OcelotPlatform methodsFor: 'testing' stamp: ''!
usesArrayBindingRatherThanGrouping
	^false.! !

!OcelotPlatform methodsFor: 'constants' stamp: ''!
usesNullForEmptyStrings
	"Return true if this database is likely to use nil as an empty string value"
	^false.! !

!OcelotPlatform methodsFor: 'types' stamp: ''!
varchar
	^self typeNamed: #varchar ifAbsentPut: [GlorpVarCharType new].! !

DatabasePlatform subclass: #OraclePlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!OraclePlatform commentStamp: '<historical>' prior: 0!
This is a platform for Oracle databases. Currently there's only one, but it might be better for some purposes to separate out by versions, as some older versions don't support all features.!

OraclePlatform subclass: #OracleODBCPlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!OracleODBCPlatform commentStamp: '<historical>' prior: 0!
This is a platform for Oracle, as accessed through ODBC. The database doesn't change, but at least in VW we determine which connection type to use based on the platform object, so we need to use an ODBC connection rather than an OCI connection.!

!OracleODBCPlatform methodsFor: 'testing' stamp: ''!
isODBCPlatform

	^true.! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
areSequencesExplicitlyCreated
	^true.! !

!OraclePlatform methodsFor: 'binding' stamp: ''!
bindingsForGroupWritingFor: aCommand
	"Return the bindings array for a group write. This can be in different formats, depending on the database and perhaps the mechanism in place."
	^aCommand arrayBindings.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
boolean
	^self int2.! !

!OraclePlatform methodsFor: 'testing' stamp: ''!
canUseInWhereClause: aValue type: aType
	"This is mostly for test support purposes. If we know we can't use a particular kind of value in the where clause on some database, don't try."

	(super canUseInWhereClause: aValue type: aType) ifFalse: [^false].
	"On Oracle we effectively can't compare constants against blobs."
	aType = self blob ifFalse: [^true].
	aValue isNil ifTrue: [^true].
	^false.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
char
	^self typeNamed: #char ifAbsentPut: [GlorpCharType new].! !

!OraclePlatform methodsFor: 'types' stamp: ''!
clob

	^self typeNamed: #clob ifAbsentPut: [GlorpClobType new typeString: 'clob'].! !

!OraclePlatform methodsFor: 'database-specific' stamp: ''!
compoundOperationFor: aSymbol
	"Return the platform specific version of a compound statement symbol"
	aSymbol == #EXCEPT ifTrue: [^'MINUS'].
	^aSymbol.! !

!OraclePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatement: aField newType: aType usingExpression: expression

	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
	self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' MODIFY (';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll: aType typeString, ')'.

	^sqlStatementStream contents! !

!OraclePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatement: aField setDefault: anObject
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
	self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' MODIFY (';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  'DEFAULT';
		space.
	anObject printOn: sqlStatementStream.
	sqlStatementStream
		nextPutAll:  ')'.
	
	^sqlStatementStream contents! !

!OraclePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatementDropDefault: aField
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
		self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' MODIFY (';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  'DEFAULT NULL)'.

	^sqlStatementStream contents! !

!OraclePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatementDropNotNull: aField
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
		self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' MODIFY (';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  'NULL)'.

	
	^sqlStatementStream contents! !

!OraclePlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatementSetNotNull: aField
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
		self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' MODIFY (';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  self sqlTextForNOTNULLAttributeConstraint;
		nextPutAll: ')'.

	
	^sqlStatementStream contents! !

!OraclePlatform methodsFor: 'sequences' stamp: ''!
databaseSequenceClass
	^OracleSequence.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
date
	^self typeNamed: #date ifAbsentPut: [GlorpTimeStampType new typeString: 'date'].! !

!OraclePlatform methodsFor: 'types' stamp: ''!
double
	^self float8! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
endColumnAddString
	"The string we use as a suffix when adding a colum"
	^' ) '.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
float
	^self float4.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
float4
	^self typeNamed: #float4 ifAbsentPut: [GlorpFloatType new typeString: 'float'].! !

!OraclePlatform methodsFor: 'types' stamp: ''!
float8
	^self typeNamed: #float8 ifAbsentPut: [GlorpDoubleType new typeString: 'double precision'].! !

!OraclePlatform methodsFor: 'functions' stamp: ''!
initializeFunctions
	| baseExists dbSpecificExists selectList baseNotExists dbSpecificNotExists |
	super initializeFunctions.

	"Oracle doesn't allow functions like EXISTS in a field list, but will allow a select there as an argument to a function, such as DECODE. So print it that way if we're in the select list."
	baseExists := functions at: #exists:.
	dbSpecificExists := DualRoleFunction new.
	dbSpecificExists function: '#exists:'.
	dbSpecificExists whereClauseVersion: baseExists.
	selectList := PrefixFunction named: 'DECODE'.
	selectList argumentModifierBlock: [:function :args | 
		function base: args first.
		Array with: 'x' asGlorpExpression with: 1 asGlorpExpression with: 0 asGlorpExpression].
	selectList type: baseExists type.
	dbSpecificExists selectListVersion: selectList.
	dbSpecificExists type: baseExists type.
	functions at: #exists: put: dbSpecificExists.

	"And similarly NOT EXISTS"
	baseNotExists := functions at: #notExists:.
	dbSpecificNotExists := DualRoleFunction new.
	dbSpecificNotExists function: '#notExists:'.
	dbSpecificNotExists whereClauseVersion: baseNotExists.
	selectList := PrefixFunction named: 'DECODE'.
	selectList argumentModifierBlock: [:function :args | 
		function base: args first.
		Array with: 'x' asGlorpExpression with: 0 asGlorpExpression with: 1 asGlorpExpression].
	selectList type: baseNotExists type.
	dbSpecificNotExists selectListVersion: selectList.
	dbSpecificNotExists type: baseExists type.
	functions at: #notExists: put: dbSpecificNotExists.! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
initializeReservedWords
	super initializeReservedWords.
	reservedWords add: 'number'.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
int

	^self integer! !

!OraclePlatform methodsFor: 'types' stamp: ''!
int2

	^self integer! !

!OraclePlatform methodsFor: 'types' stamp: ''!
int4

	^self integer! !

!OraclePlatform methodsFor: 'types' stamp: ''!
int8

	^self integer! !

!OraclePlatform methodsFor: 'SQL' stamp: ''!
isOraclePlatform

	^true! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfColumnName
	"^<Integer> I return the maximum length of a column name.  (I am not sent at present;  I am here so that if I and #maximumLengthOfTableName (which is sent) should ever differ on some database platform, I can be used in revised truncation logic.)"

	^30! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfTableName
	"^<Integer> I return the max. length of a table name"

	^30! !

!OraclePlatform methodsFor: 'testing' stamp: ''!
maximumQueryValueSize
	"If there's a limit on how big a bound input parameter can be for a query, return it here. By default, no maximum. "
	^4001! !

!OraclePlatform methodsFor: 'testing' stamp: ''!
measuresStringFieldWidthsInBytes
	"Return true if we measure the width of a string in bytes rather than characters. So, a 255-character field can hold 255 bytes worth of string, not 255 characters. This is important with encodings."

	^true! !

!OraclePlatform methodsFor: 'types' stamp: ''!
number

	^self typeNamed: #number ifAbsentPut: [GlorpNumericType new typeString: 'number'].! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
primaryKeysAreAutomaticallyUnique
	"Return false if, in addition to specifying something as a primary key, we must separately specify it as unique"

	^true.! !

!OraclePlatform methodsFor: 'conversion-strings' stamp: ''!
printBlob: aByteArray on: aStream for: aType
	aByteArray isNil ifTrue: [^aStream nextPutAll: 'NULL'].
	aStream nextPutAll: 'HEXTORAW('''.
	aByteArray do: [:each |
		each printOn: aStream paddedWith: $0 to: 2 base: 16].
	aStream nextPutAll: ''')'.! !

!OraclePlatform methodsFor: 'conversion-times' stamp: ''!
printDate: aDate for: aType
	| stream |
	aDate isNil ifTrue: [^'NULL'].
	stream := WriteStream on: String new.
	stream nextPutAll: 'to_date('''.
	self
		printDate: aDate
		isoFormatOn: stream.
	stream nextPutAll: ''','''.
	stream nextPutAll: 'YYYY-MM-DD'.
	stream nextPutAll: ''')'.
	^stream contents.! !

!OraclePlatform methodsFor: 'table creation' stamp: ''!
printNumericFieldOfSize: anInteger on: aStream

	aStream 
		nextPutAll: 'number('.
	anInteger printOn: aStream.
	aStream nextPutAll: ')'.! !

!OraclePlatform methodsFor: 'table creation' stamp: ''!
printStringFieldOfSize: anInteger on: aStream

	aStream 
		nextPutAll: 'varchar2('.
	anInteger printOn: aStream.
	aStream nextPutAll: ')'.! !

!OraclePlatform methodsFor: 'conversion-times' stamp: ''!
printTime: aTime for: aType
	| stream |
	aTime isNil ifTrue: [^'NULL'].
	stream := WriteStream on: String new.
	stream nextPutAll: 'to_date('''.
	self
		printTime: aTime
		isoFormatOn: stream.
	stream nextPutAll: ''','''.
	stream nextPutAll: 'HH24:MI:SS'.
	stream nextPutAll: ''')'.
	^stream contents.! !

!OraclePlatform methodsFor: 'conversion-times' stamp: ''!
printTimestamp: aTimestamp on: stream for: aType

	aTimestamp isNil ifTrue: [stream nextPutAll: 'NULL'. ^self].
	stream nextPutAll: 'to_date('''.
	self
		printDate: aTimestamp
		isoFormatOn: stream.
	stream nextPut: $ .
	self
		printTime: aTimestamp
		isoFormatOn: stream
		milliseconds: false.
	stream nextPutAll: ''','''.
	stream nextPutAll: 'YYYY-MM-DD HH24:MI:SS'.
	stream nextPutAll: ''')'.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
raw
	^self typeNamed: #raw ifAbsentPut: [GlorpBlobType new typeString: 'raw'].! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
requiresTransactionForTableOperations
	"Return true if we need a transaction in order to do table creation, deletion, and so forth. Some databases require it. Others don't like it"
	^false.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
sequence

	^self serial.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
serial

	^self typeNamed: #serial ifAbsentPut: [GlorpSerialType new typeString: (self integer typeString)].! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
sqlTextCurrentServerUTCTimestamp

	^'SELECT SYS_EXTRACT_UTC(CURRENT_TIMESTAMP) FROM DUAL'! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
startColumnAddString
	"The string we use as a prefix when adding a colum"
	^' ( '.! !

!OraclePlatform methodsFor: 'testing' stamp: ''!
supportsANSIJoins
	"Do we support the JOIN <tableName> USING <criteria> syntax. Currently hard-coded, but may also vary by database version"
	^false.! !

!OraclePlatform methodsFor: 'testing' stamp: ''!
supportsBinding
	"Binding works only with VW EXDI so far"
	^Dialect isVisualWorks.! !

!OraclePlatform methodsFor: 'testing' stamp: ''!
supportsGroupWritingFor: aCommand
	^self supportsBinding and: [self useBinding and: [aCommand supportsGroupWriting]].! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
supportsMillisecondsInTimes
	^false.! !

!OraclePlatform methodsFor: 'testing' stamp: ''!
supportsTableOwners
	^true! !

!OraclePlatform methodsFor: 'types' stamp: ''!
text
	^self clob.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
time
	^self timestamp.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
timestamp
	^self typeNamed: #timestamp ifAbsentPut: [GlorpTimeStampType new typeString: 'date'].! !

!OraclePlatform methodsFor: 'conversion-times' stamp: ''!
timestampConverter
	^DelegatingDatabaseConverter
		named: #timestamp
		hostedBy: self
		fromStToDb: #asTimestamp:for:
		fromDbToSt: #readTimestamp:for:.! !

!OraclePlatform methodsFor: 'testing' stamp: ''!
useOracleOuterJoins
	"Return true if we use the old Oracle x = y (+) syntax for outer joins"
	^true.! !

!OraclePlatform methodsFor: 'testing' stamp: ''!
usesArrayBindingRatherThanGrouping
	"Return true if we use array binding for grouped writes rather than printing the sql multiple times. Only applies if we support grouped writes"
	^true.! !

!OraclePlatform methodsFor: 'constants' stamp: ''!
usesNullForEmptyStrings
	"Return true if this database is likely to use nil as an empty string value"
	^true.! !

!OraclePlatform methodsFor: 'types' stamp: ''!
varchar
	^self typeNamed: #varchar ifAbsentPut: [GlorpVarCharType new typeName: 'varchar2'].! !

!OraclePlatform methodsFor: 'accessing' stamp: ''!
vwEXDIConnectionClass
	^'OracleThreadedConnection' asQualifiedReference valueIfUndefinedLoadFrom: 'OracleThapiEXDI'! !

DatabasePlatform subclass: #PostgreSQLPlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!PostgreSQLPlatform commentStamp: '<historical>' prior: 0!
This is a platform for the open-source PostgreSQL database.
!

!PostgreSQLPlatform classSide methodsFor: 'testing' stamp: ''!
useBindingByDefault
	^false! !

!PostgreSQLPlatform methodsFor: 'constants' stamp: ''!
areSequencesExplicitlyCreated
	^true.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
bigint
	^self typeNamed: #bigint ifAbsentPut: [GlorpIntegerType new typeString: 'bigint'].! !

!PostgreSQLPlatform methodsFor: 'binding' stamp: ''!
bindingsForGroupWritingFor: aCommand
	"Return the bindings array for a group write. This can be in different formats, depending on the database and perhaps the mechanism in place."
	^aCommand batchStatementBindings.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
blob

	^self typeNamed: #blob ifAbsentPut: [GlorpBlobType new typeString: 'bytea'].! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
boolean
	^self typeNamed: #boolean ifAbsentPut: [GlorpBooleanType new typeString: 'boolean'].! !

!PostgreSQLPlatform methodsFor: 'conversion-strings' stamp: ''!
byteArrayToSQLStringConverter

	^DelegatingDatabaseConverter
		named: #byteArrayToSQLString
		hostedBy: self
		fromStToDb: #convertByteArrayToSQLString:for:
		fromDbToSt: #convertSQLStringToByteArray:for:.! !

!PostgreSQLPlatform methodsFor: 'testing' stamp: ''!
canBind: aValue to: aType
	"The Postgres EXDI seems to have trouble binding date formats"
	aType = self date ifTrue: [^false].
	aType = self time ifTrue: [^false].
	aType = self timestamp ifTrue: [^false].
	aType = self timetz ifTrue: [^false].
	(Dialect isVisualWorks and: [aValue class == Dialect doublePrecisionFloatClass]) ifTrue: [^false].
	(Dialect isVisualWorks and: [aValue class == Dialect fixedPointClass]) ifTrue: [^false].
	^super canBind: aValue to: aType.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
char
	^self typeNamed: #char ifAbsentPut: [GlorpCharType new typeName: 'character'].! !

!PostgreSQLPlatform methodsFor: 'constants' stamp: ''!
charactersThatNeedEscaping
	^#($' $\)! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
clob

	^self typeNamed: #clob ifAbsentPut: [GlorpClobType new typeString: 'text'].! !

!PostgreSQLPlatform methodsFor: 'conversion-strings' stamp: ''!
convertSQLStringToByteArray: aString for: aType

	| aStream str | 
	aString isNil ifTrue: [^nil].
	aStream := (ByteArray new: aString size // 4) writeStream.
	str := aString readStream.
	[str atEnd] whileFalse: [ |nextChar|
		nextChar := str next.
		aStream nextPut: (nextChar = $\
								ifTrue: [str peek = $\
											ifTrue: [str next asInteger]
											ifFalse: [Number readFrom: (str next: 3) base: 8]]
								ifFalse: [nextChar asInteger])
	].
	aStream close.
	^aStream contents! !

!PostgreSQLPlatform methodsFor: 'sequences' stamp: ''!
databaseSequenceClass
	^PGSequence.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
date
	^self typeNamed: #date ifAbsentPut: [GlorpDateType new typeString: 'date'].! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
decimal
	^self numeric.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
double
	^self float8! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
float
	^self float4.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
float4
	^self typeNamed: #float4 ifAbsentPut: [GlorpFloatType new].! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
float8
	^self typeNamed: #float8 ifAbsentPut: [GlorpDoubleType new].! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
int
	^self int4! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
int2
	^self typeNamed: #int2 ifAbsentPut: [GlorpIntegerType new typeString: 'int2'].! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
int4
	^self typeNamed: #int4 ifAbsentPut: [GlorpIntegerType new typeString: 'int4'].! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
int8
	^self typeNamed: #int8 ifAbsentPut: [GlorpIntegerType new typeString: 'int8'].! !

!PostgreSQLPlatform methodsFor: 'SQL' stamp: ''!
isPostgreSQLPlatform
	^true.! !

!PostgreSQLPlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfColumnName
	"^<Integer> I return the maximum length of a column name.  (I am not sent at present;  I am here so that if I and #maximumLengthOfTableName (which is sent) should ever differ on some database platform, I can be used in revised truncation logic.)"

	^31! !

!PostgreSQLPlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfTableName
	"^<Integer> I return the max. length of a table name"

	^31! !

!PostgreSQLPlatform methodsFor: 'testing' stamp: ''!
measuresStringFieldWidthsInBytes
	"Return true if we measure the width of a string in bytes rather than characters. So, a 255-character field can hold 255 bytes worth of string, not 255 characters. This is important with encodings"
	^true.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
numeric
	^self typeNamed: #numeric ifAbsentPut: [GlorpNumericType new].! !

!PostgreSQLPlatform methodsFor: 'conversion-strings' stamp: 'GuillermoPolito 4/16/2011 23:05'!
printBlob: aByteArray on: aStream for: aType
	| base64Stream |
	aByteArray isNil ifTrue: [^aStream nextPutAll: 'NULL'].
	(Dialect isSqueak or: [Dialect isPharo]) ifTrue: [^self squeakPrintBlob: aByteArray on: aStream for: aType].
	aStream nextPut: $'.
	base64Stream := (Dialect smalltalkAt: #Base64EncodingWriteStream) onStream: aStream.
	base64Stream lineBreaksEnabled: false.
	base64Stream nextPutAll: aByteArray.
	base64Stream emitClosingCharacters.
	aStream nextPut: $'.! !

!PostgreSQLPlatform methodsFor: 'conversion-times' stamp: ''!
printDate: aDate for: aType
	"Print a date (or timestamp) as yyyy-mm-dd"
	| stream |
	aDate isNil ifTrue: [^'NULL'].
	stream := WriteStream on: String new.
	stream nextPut: $'.
	self
		printDate: aDate
		isoFormatOn: stream.
	stream nextPut: $'.
	^stream contents.! !

!PostgreSQLPlatform methodsFor: 'database-specific' stamp: ''!
printOffset: anInteger on: aCommand
	aCommand
		nextPutAll: ' OFFSET '.
	anInteger printOn: aCommand.! !

!PostgreSQLPlatform methodsFor: 'database-specific' stamp: ''!
printPostLimit: anInteger on: aCommand
	aCommand
		nextPutAll: ' LIMIT '.
	anInteger printOn: aCommand.! !

!PostgreSQLPlatform methodsFor: 'conversion-times' stamp: ''!
printTime: aTime for: aType
	"Print a time (or timestamp) as hh:mm:ss.fff"
	| stream |
	aTime isNil ifTrue: [^'NULL'].
	stream := WriteStream on: String new.
	stream nextPut: $'.
	self
		printTime: aTime
		isoFormatOn: stream.
	stream nextPut: $'.
	^stream contents.! !

!PostgreSQLPlatform methodsFor: 'conversion-times' stamp: ''!
printTimestamp: aTimestamp on: stream for: aType
	aTimestamp isNil ifTrue: [stream nextPutAll: 'NULL'. ^self].
	stream nextPut: $'.
	self
		printDate: aTimestamp
		isoFormatOn: stream.
	stream nextPutAll: ' '.
	self
		printTime: aTimestamp
		isoFormatOn: stream.
	stream nextPut: $'.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
real
	^self float4! !

!PostgreSQLPlatform methodsFor: 'testing' stamp: ''!
requiresCastsForNullsInUnions
	"If we have explicit nulls in a query which is unioned, do we need to cast them to the appropriate type in order to keep the database happy."
	^true.! !

!PostgreSQLPlatform methodsFor: 'constants' stamp: ''!
requiresTransactionForTableOperations
	"Return true if we need a transaction in order to do table creation, deletion, and so forth. Some databases require it. Others don't like it"
	^false.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
sequence

	^self serial! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
serial
	"For postgresql, we use sequences, and explicitly get the values ourselves, so just tell the database that they're integers."
	^self typeNamed: #serial ifAbsentPut: [GlorpSerialType new typeString: 'integer'].! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
smallint
	^self int2! !

!PostgreSQLPlatform methodsFor: 'constants' stamp: ''!
sqlTextCurrentServerUTCTimestamp

	^'SELECT CURRENT_TIMESTAMP AT TIME ZONE ''UTC'' '! !

!PostgreSQLPlatform methodsFor: 'conversion-strings' stamp: ''!
squeakPrintBlob: aByteArray on: aStream for: aType

	aStream nextPut: $'.
	aByteArray do: [:each |
		aStream nextPutAll: '\\'.
		each asInteger printOn: aStream paddedWith: $0 to: 3 base: 8].
	aStream nextPutAll: '''::bytea'.! !

!PostgreSQLPlatform methodsFor: 'testing' stamp: ''!
supportsANSIJoins
	"Do we support the JOIN <tableName> USING <criteria> syntax."
	^true.! !

!PostgreSQLPlatform methodsFor: 'testing' stamp: ''!
supportsBinding
	"Binding works only with VW EXDI so far"
	^Dialect isVisualWorks.! !

!PostgreSQLPlatform methodsFor: 'testing' stamp: ''!
supportsCaseInsensitiveLike
	^true.! !

!PostgreSQLPlatform methodsFor: 'testing' stamp: ''!
supportsGroupWritingFor: aCommand
	^aCommand supportsGroupWriting.! !

!PostgreSQLPlatform methodsFor: 'testing' stamp: ''!
supportsLimit
	"Do we support anything analogous to the postgresql LIMIT, returning only the first N rows"
	^true.! !

!PostgreSQLPlatform methodsFor: 'constants' stamp: ''!
supportsMillisecondsInTimes
	"I think this depends on settings, but it's easier right now to just turn it off"
	^false.! !

!PostgreSQLPlatform methodsFor: 'testing' stamp: ''!
supportsOffset
	"Do we support anything analogous to the postgresql OFFSET, skipping the first N rows"
	^true.! !

!PostgreSQLPlatform methodsFor: 'services tables' stamp: ''!
tableNameFor: aStringRepresentingATableName
	"Return the table name, as modified to be used in a database statement."
	
	^(aStringRepresentingATableName
		detect: [:each | each isUppercase]
		ifNone: [nil]) isNil
	ifTrue: [aStringRepresentingATableName]
	ifFalse: ['"' , aStringRepresentingATableName , '"'].! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
time
	^self typeNamed: #time ifAbsentPut: [GlorpTimeType new typeString: 'time'].! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
timeStampTypeString
	^'timestamp'.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
timestamp
	^self typeNamed: #timestamp ifAbsentPut: [GlorpTimeStampType new typeString: 'timestamp without time zone'].! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
timetz
	^self typeNamed: #timetz ifAbsentPut: [GlorpPGTimeWithTimeZoneType new].! !

!PostgreSQLPlatform methodsFor: 'type helpers' stamp: ''!
typesWithValues
	super typesWithValues.
	self boolean.
	self bigint.
	self date.
	self decimal.
	self double.
	self float.
	self float4.
	self float8.
	self int2.
	self int.
	self integer.
	self numeric.
	self real.
	self sequence.
	self serial.
	self timetz.
	^self types.! !

!PostgreSQLPlatform methodsFor: 'testing' stamp: ''!
usesArrayBindingRatherThanGrouping
	"Return true if we use array binding for grouped writes rather than printing the sql multiple times. Only applies if we support grouped writes"
	^false.! !

!PostgreSQLPlatform methodsFor: 'types' stamp: ''!
varchar
	^self typeNamed: #varchar ifAbsentPut: [GlorpVarCharType new typeName: 'character varying'].! !

!PostgreSQLPlatform methodsFor: 'accessing' stamp: ''!
vwEXDIConnectionClass
	^'PostgreSQLEXDIConnection' asQualifiedReference valueIfUndefinedLoadFrom: 'PostgreSQLEXDI'! !

DatabasePlatform subclass: #SQLServerPlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!SQLServerPlatform commentStamp: '<historical>' prior: 0!
This is a platform for the Microsoft SQL Server database.!

SQLServerPlatform subclass: #AccessPlatform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!AccessPlatform commentStamp: '<historical>' prior: 0!
This is a platform for the Microsoft Access database. It subclasses from SQL Server, because the two have a number of characteristics in common.!

!AccessPlatform methodsFor: 'constants' stamp: ''!
areSequencesExplicitlyCreated

	^false.! !

!AccessPlatform methodsFor: 'types' stamp: ''!
blob

	^self typeNamed: #blob ifAbsentPut: [GlorpBlobType new typeString: 'longbinary'].! !

!AccessPlatform methodsFor: 'conversion-boolean' stamp: ''!
booleanToBooleanConverter
	^DelegatingDatabaseConverter
		named: #booleanToBoolean
		hostedBy: self
		fromStToDb: #convertBooleanToDBBoolean:for:
		fromDbToSt: #convertDBBooleanToBoolean:for:.! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
canBind: aValue to: aType
	(Dialect isVisualWorks and: [aValue class == Dialect fixedPointClass]) ifTrue: [^false].
	^super canBind: aValue to: aType.! !

!AccessPlatform methodsFor: 'types' stamp: ''!
clob

	^self typeNamed: #clob ifAbsentPut: [GlorpClobType new typeString: 'memo'].! !

!AccessPlatform methodsFor: 'services tables' stamp: ''!
endColumnQuote
	^']'.! !

!AccessPlatform methodsFor: 'functions' stamp: ''!
initializeFunctions
	super initializeFunctions.
	functions
		at: #copyFrom:to: put: (SubstringFunction named: 'MID')! !

!AccessPlatform methodsFor: 'constants' stamp: ''!
initializeReservedWords
	super initializeReservedWords.
	reservedWords add: 'timestamp'.! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
isAccessPlatform

	^true! !

!AccessPlatform methodsFor: 'types' stamp: ''!
longText

	^self typeNamed: #longText ifAbsentPut: [GlorpClobType new typeString: 'memo'].! !

!AccessPlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfColumnName
	"^<Integer> I return the maximum length of a column name.  (I am not sent at present;  I am here so that if I and #maximumLengthOfTableName (which is sent) should ever differ on some database platform, I can be used in revised truncation logic.)"

	^64! !

!AccessPlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfTableName
	"^<Integer> I return the max. length of a table name"

	^64! !

!AccessPlatform methodsFor: 'types' stamp: ''!
memo

	^self typeNamed: #memo ifAbsentPut: [GlorpClobType new typeString: 'memo'].! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
requiresCastsForNullsInUnions
	"If we have explicit nulls in a query which is unioned, do we need to cast them to the appropriate type in order to keep the database happy."
	^true.! !

!AccessPlatform methodsFor: 'types' stamp: ''!
serial

	^self typeNamed: #serial ifAbsentPut: [GlorpSerialType new typeString: 'autoincrement'].! !

!AccessPlatform methodsFor: 'services tables' stamp: ''!
startColumnQuote
	^'['.! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
supportsAliasInCompoundQuery
	"Return true if, when we have a subselect, we can (and/or must) alias the result with a table name. Access doesn't like the table name"
	^false.! !

!AccessPlatform methodsFor: 'constants' stamp: ''!
supportsAuthentication
	^false! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
supportsDecimalsOnAllNumerics
	"Return true if a general 'numeric' type will allow numbers after the decimal place"
	^true.! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
supportsGroupWritingFor: aCommand
	^false.! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
supportsInformationSchema
	^false.! !

!AccessPlatform methodsFor: 'constants' stamp: ''!
supportsMillisecondsInTimes
	^false.! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
supportsSchemas
	^false! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
supportsStringQueryOnBlobs
	"Return true if we can do string queries on parts of blobs."
	^false.! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
supportsTableOwners
	"Return true if this platform supports table owners, i.e. expects table names of the form Bern.TW_* rather than just TW_* in its SQL."
"Access, Firebird and PostGreSQL do not, Oracle does, others I know not."

	^false! !

!AccessPlatform methodsFor: 'SQL' stamp: ''!
supportsVariableSizedNumerics
	"Return true if this platform can support numbers with a varying size and number of decimal places. Access, notably, doesn't seem to be able to"
	^false.! !

!AccessPlatform methodsFor: 'types' stamp: ''!
text

	^self typeNamed: #text ifAbsentPut: [GlorpClobType new typeString: 'longtext'].! !

!AccessPlatform methodsFor: 'testing' stamp: ''!
useMicrosoftOuterJoins
	"Return true if we use the Microsoft x *= y syntax for outer joins. Oddly enough, false for access 2k"
	^false.! !

!AccessPlatform methodsFor: 'constants' stamp: ''!
usesNullForFalse
	"Return true if this database is likely to use nil as an empty string value"
	^true.! !

!SQLServerPlatform methodsFor: 'constants' stamp: ''!
areSequencesExplicitlyCreated

^false! !

!SQLServerPlatform methodsFor: 'binding' stamp: ''!
bindingsForGroupWritingFor: aCommand
	"Return the bindings array for a group write. This can be in different formats, depending on the database and perhaps the mechanism in place."
	^aCommand batchStatementBindings.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
bit
	^self typeNamed: #bit ifAbsentPut: [GlorpBooleanType new typeString: 'bit'].! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
blob

	^self typeNamed: #blob ifAbsentPut: [GlorpBlobType new typeString: 'varbinary(max)'; queryType: (self varbinary)].! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
boolean
	^self bit.! !

!SQLServerPlatform methodsFor: 'conversion-boolean' stamp: ''!
booleanToBooleanConverter
	^DelegatingDatabaseConverter
		named: #booleanToBoolean
		hostedBy: self
		fromStToDb: #convertBooleanToInteger:for:
		fromDbToSt: #convertDBBooleanToBoolean:for:.! !

!SQLServerPlatform methodsFor: 'constants' stamp: ''!
capitalWritingOfTableName
	"^<Boolean>"

	^false! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
char
	^self typeNamed: #char ifAbsentPut: [GlorpCharType new].! !

!SQLServerPlatform methodsFor: 'conversion-strings' stamp: ''!
charactersThatNeedEscaping
	"There seem to be all kind of contradictory bits of information about what sql server does/requires for escaped characters, all of which differ from standard sql. Empirically the only thing that requires escaping appears to be single quote"
	^#($' ).! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
clob

	^self typeNamed: #clob ifAbsentPut: [GlorpClobType new typeString: 'varchar(max)'].! !

!SQLServerPlatform methodsFor: 'database-specific' stamp: ''!
compoundOperationFor: aSymbol
	"Return the platform specific version of a compound statement symbol"
	aSymbol == #INTERSECT ifTrue: [^'WHERE EXISTS'].
	aSymbol == #MINUS ifTrue: [^'WHERE NOT EXISTS'].
	^aSymbol.! !

!SQLServerPlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatement: aField newType: aType usingExpression: expression

	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
	self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' ALTER COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll: aType typeString.
	expression ifNotNil: [ sqlStatementStream nextPutAll: ' USING ', expression ].

	^sqlStatementStream contents! !

!SQLServerPlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatement: aField setDefault: anObject
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
	self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' ALTER COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  'DEFAULT ';
		space.
	anObject printOn: sqlStatementStream.
	sqlStatementStream
		nextPutAll:  ' WITH VALUES'.
	
	^sqlStatementStream contents! !

!SQLServerPlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatementDropDefault: aField
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
		self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' ALTER COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll:  'DEFAULT NULL'.

	^sqlStatementStream contents! !

!SQLServerPlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatementDropNotNull: aField
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
		self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' ALTER COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll: aField type typeString;
		space;
		nextPutAll:  'NULL'.

	
	^sqlStatementStream contents! !

!SQLServerPlatform methodsFor: 'services columns' stamp: ''!
createAlterColumnStatementSetNotNull: aField
	| sqlStatementStream |
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll:  'ALTER TABLE';
		space.
		self printDDLTableNameFor: aField table on: sqlStatementStream.

	sqlStatementStream
		nextPutAll:  ' ALTER COLUMN';
		space;
		nextPutAll: (self nameForColumn: aField name);
		space;
		nextPutAll: aField type typeString;
		space;
		nextPutAll:  self sqlTextForNOTNULLAttributeConstraint.

	
	^sqlStatementStream contents! !

!SQLServerPlatform methodsFor: 'constants' stamp: ''!
databaseSequenceClass
	"My AccessPlatform subclass has a hardcoded check in StoreDescriptorSystem>>sequenceFor: to ignore this method in favour of calling #justSelectTheMaximumCurrentValueAndAddOneSequenceFor: for AccessPlatform.  Can we retire that?"

	^SQLServerSequence! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
date
	"SQL Server does not compare a timestamp that's cast to a date as being equal to a date, which is the behaviour we expect. Its dates are really timestamps internally, so just treat date and timestamp as the same."
	^self timestamp.! !

!SQLServerPlatform methodsFor: 'conversion-times' stamp: ''!
dateConverter
	"SQL server doesn't have plain dates, and doesn't accept them"
	^DelegatingDatabaseConverter
		named: #date
		hostedBy: self
		fromStToDb: #dateToTimestampConversion:for:
		fromDbToSt: #readDate:for:.	"#printDate:for:"! !

!SQLServerPlatform methodsFor: 'conversion-times' stamp: ''!
dateToTimestampConversion: aDate for: aType
	aDate isNil ifTrue: [^aDate].
	^aDate asTimestamp.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
decimal
	^self numeric.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
double
	^self float.! !

!SQLServerPlatform methodsFor: 'conversion-strings' stamp: ''!
escapeFor: aCharacter
	^String with: $' with: aCharacter.
"	^'\', (aCharacter asInteger printStringRadix: 16)."! !

!SQLServerPlatform methodsFor: 'exdi specific' stamp: ''!
exdiTypeForDates
	^#Timestamp.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
float

	^self typeNamed: #float ifAbsentPut: [GlorpMSSQLFloatType new].! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
float4
	^self float.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
float8
	^self float.! !

!SQLServerPlatform methodsFor: 'functions' stamp: ''!
initializeFunctions
	| baseExists dbSpecificExists selectList baseNotExists dbSpecificNotExists |
	super initializeFunctions.
	functions
		at: #, put: (InfixFunction named: '+');
		at: #copyFrom:to: put: (SubstringFunction named: 'SUBSTRING');
		at: #objectId put: (PrefixFunction named: 'OBJECT_ID');
		at: #columnPropertySchema:property: put: ((PrefixFunction named: 'COLUMNPROPERTY') type: self integer);
		at: #convertTo: put: ((PrefixFunction named: 'CONVERT')). "CONVERT(toType, aThing)"

	"Like Oracle, SQL Server doesn't allow functions like EXISTS in a field list,
	but will allow a select there as an argument to a function, such as CASE.
	So print it that way if we're in the select list."

	baseExists := functions at: #exists:.
	dbSpecificExists := DualRoleFunction new.
	dbSpecificExists function: '#exists:'.
	dbSpecificExists whereClauseVersion: baseExists.
	selectList := Case new argumentTypes: (Array with: (self varchar: 1) with: self integer with: self integer).
	selectList argumentModifierBlock: [:function :args | 
		function base: args first.
		Array with: 'x' asGlorpExpression with: 1 asGlorpExpression with: 0 asGlorpExpression].
	dbSpecificExists selectListVersion: selectList.
	dbSpecificExists type: self integer.
	functions at: #exists: put: dbSpecificExists.

	"And similarly NOT EXISTS"
	baseNotExists := functions at: #notExists:.
	dbSpecificNotExists := DualRoleFunction new.
	dbSpecificNotExists function: '#notExists:'.
	dbSpecificNotExists whereClauseVersion: baseNotExists.
	selectList := Case new argumentTypes: (Array with: (self varchar: 1) with: self integer with: self integer).
	selectList argumentModifierBlock: [:function :args | 
		function base: args first.
		Array with: 'x' asGlorpExpression with: 0 asGlorpExpression with: 1 asGlorpExpression].
	dbSpecificNotExists selectListVersion: selectList.
	dbSpecificNotExists type: self integer.
	functions at: #notExists: put: dbSpecificNotExists.! !

!SQLServerPlatform methodsFor: 'constants' stamp: ''!
initializeReservedWords
	super initializeReservedWords.
	reservedWords add: 'key'! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
int
	^self integer.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
int2
	^self smallint.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
int4

	^self integer.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
int8

	^self numeric.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
integer

	^self typeNamed: #integer ifAbsentPut: [GlorpIntegerType new typeString: 'int'].! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
isODBCPlatform

	^true! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
isSQLServerPlatform

	^true! !

!SQLServerPlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfColumnName
	"^<Integer> I return the maximum length of a column name.  (I am not sent at present;  I am here so that if I and #maximumLengthOfTableName (which is sent) should ever differ on some database platform, I can be used in revised truncation logic.)"

	^128! !

!SQLServerPlatform methodsFor: 'constants' stamp: ''!
maximumLengthOfTableName
	"^<Integer> I return the max. length of a table name"

	^128! !

!SQLServerPlatform methodsFor: 'binding' stamp: ''!
maximumSizeToGroupWriteFor: aCollectionOfDatabaseRows
	"If we are going to group write, how many rows of this collection should we do it for at once"
	^aCollectionOfDatabaseRows size min: 250.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
nclob

	^self typeNamed: #nclob ifAbsentPut: [GlorpClobType new typeString: 'nvarchar(max)'].! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
numeric
	^self typeNamed: #numeric ifAbsentPut: [GlorpNumericType new].! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
nvarchar
	^self typeNamed: #nvarchar ifAbsentPut: [GlorpVarCharType new typeString: 'nvarchar'].! !

!SQLServerPlatform methodsFor: 'conversion-strings' stamp: ''!
printBlob: aByteArray on: aStream for: aType
	aByteArray isNil ifTrue: [^aStream nextPutAll: 'NULL'].
	aStream nextPutAll: '0x'.
	aByteArray do: [:each |
		each printOn: aStream paddedWith: $0 to: 2 base: 16].! !

!SQLServerPlatform methodsFor: 'conversion-boolean' stamp: ''!
printBoolean: aBoolean for: aType
	"If the platform is MS SQL Server, a boolean object in a SELECT needs to be converted to  BIT type, since T, Y, or true make SQLServer complain. "
	
	| string |
	aBoolean isNil ifTrue: [^'NULL'].
	"We may be getting passed something that's already 1 or 0, or something else instead of a boolean. That probably shouldn't be happening, but let's handle it gracefully. We define gracefully to mean 1 is true, and everything else is false"
	string := (aBoolean == 1 or: [aBoolean = true]) 
		ifTrue: ['1'] 
		ifFalse: ['0'].
	^'CONVERT( BIT, ' , string , ' )'.! !

!SQLServerPlatform methodsFor: 'conversion-times' stamp: ''!
printDate: aDate for: aType
	"Print a date (or timestamp) as yyyy-mm-dd"
	| stream |
	aDate isNil ifTrue: [^'NULL'].
	stream := WriteStream on: String new.
	stream nextPutAll: '{ d '''.
	self
		printDate: aDate
		isoFormatOn: stream.
	stream nextPutAll: ''' }'.
	^stream contents.! !

!SQLServerPlatform methodsFor: 'database-specific' stamp: ''!
printPreLimit: anInteger on: aCommand
	aCommand
		nextPutAll: ' TOP '.
	anInteger printOn: aCommand.
	aCommand nextPutAll: ' '.! !

!SQLServerPlatform methodsFor: 'conversion-times' stamp: ''!
printTime: aTime for: aType
	"Print a time (or timestamp) as hh:mm:ss.fff"
	| stream |
	aTime isNil ifTrue: [^'NULL'].
	stream := WriteStream on: String new.
	stream nextPutAll: '{ t '''.
	self
		printTime: aTime
		isoFormatOn: stream
		milliseconds: self supportsMillisecondsInTimes.
	stream nextPutAll: ''' }'.
	^stream contents.! !

!SQLServerPlatform methodsFor: 'conversion-times' stamp: ''!
printTimestamp: aTimestamp on: stream for: aType
	aTimestamp isNil ifTrue: [stream nextPutAll: 'NULL'. ^self].
	stream nextPutAll: '{ ts '''.
	self
		printDate: aTimestamp
		isoFormatOn: stream.
	stream nextPutAll: ' '.
	self
		printTime: aTimestamp
		isoFormatOn: stream.
	stream nextPutAll: ''' }'.! !

!SQLServerPlatform methodsFor: 'database-specific' stamp: ''!
queryWithUnsupportedOperationsEliminatedFrom: aQuery do: aBlock
	"If aQuery has operations that we don't support, rewrite it to do them in terms of lower level operations. In particular, rewrite INTERSECT/EXCEPT operations into EXISTS clauses in a single query. Pass the new query to aBlock."
	| newQuery |
	newQuery := aQuery rewriteIntersect.
	newQuery := newQuery rewriteExcept.
	newQuery == aQuery ifFalse: [aBlock value: newQuery].! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
real
	^self float.! !

!SQLServerPlatform methodsFor: 'conversion-strings' stamp: ''!
removeNestedParensFromString: aString
	"Return a Smalltalk instance of the type indicated by aGlorpType.  SQL-Server wraps aString in parentheses, so uwrap it first.  If aString is nil, ensure we return nil."

	| debracket |
	aString size < 2 ifTrue: [^aString].	"relies on nil size = 0"
	debracket := aString copy.
	[debracket first = $( and: [debracket last = $)]]
		whileTrue: [debracket := debracket copyFrom: 2 to: debracket size - 1].
	^debracket! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
reusePreparedStatements
	^false! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
sequence
	^self serial.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
serial

	^self typeNamed: #serial ifAbsentPut: [GlorpSerialType new typeString: 'int  IDENTITY (1, 1) '].! !

!SQLServerPlatform methodsFor: 'constants' stamp: ''!
sqlTextCurrentServerUTCTimestamp

	^'SELECT GETUTCDATE()'! !

!SQLServerPlatform methodsFor: 'constants' stamp: ''!
startColumnAddString
	"The string we use as a prefix when adding a colum.
	Override the superclass default because MS doesn't use COLUMN."
	^' '.! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
supportsANSIJoins
	"Do we support the JOIN <tableName> USING <criteria> syntax. Currently hard-coded, but may also vary by database version"
	^true.! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
supportsBinding
	"Binding works only with VW EXDI so far.  Reuse of prepared statements is set to false by default (see my class-side)."

	^Dialect isVisualWorks.! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
supportsDecimalsOnAllNumerics
	"Return true if a general 'numeric' type will allow numbers after the decimal place"
	^false.! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
supportsGroupWritingFor: aCommand
	^aCommand supportsGroupWriting.! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
supportsLimit
	"Do we support anything analogous to the postgresql LIMIT, returning only the first N rows"
	^true.! !

!SQLServerPlatform methodsFor: 'constants' stamp: ''!
supportsMillisecondsInTimes
	"I'm guessing here"
	^true.! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
supportsMultipleOpenCursors
	"Can this database support multiple open cursors at once"
	^false.! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
supportsTableOwners
	"Return true if this platform supports table owners, i.e. expects table names of the form Bern.TW_* rather than just TW_* in its SQL."
"Access, Firebird and PostGreSQL do not, Oracle does, others I know not."

	^true! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
text
	^super text queryType: self varchar.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
time
	^self typeNamed: #time ifAbsentPut: [GlorpTimeType new typeString: 'datetime'].! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
timeStampTypeString
	^'datetime'.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
timestamp
	^self typeNamed: #timestamp ifAbsentPut: [GlorpTimeStampType new typeString: 'datetime'].! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
tinyint
	^self typeNamed: #tinyInt ifAbsentPut: [GlorpIntegerType new typeString: 'tinyint'].! !

!SQLServerPlatform methodsFor: 'conversion-strings' stamp: ''!
toGlorpType: aGlorpType fromString: aString
	"Since SQL-Server wraps the string in parentheses, uwrap it first."

	^super
		toGlorpType: aGlorpType
		fromString: (self removeNestedParensFromString: aString)! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
typeWithDatabaseName: aString characterWidth: aNumber extraInformation: extraValue
	"Return our type which matches the database name for the type. Assumes all variable width fields are strings right now."
	extraValue = 'IDENTITY' ifTrue: [^self serial].
	^super typeWithDatabaseName: aString characterWidth: aNumber.! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
useMicrosoftOuterJoins
	"Return true if we use the Microsoft x *= y syntax for outer joins"
	^false.! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
usesArrayBindingRatherThanGrouping
	"Return true if we use array binding for grouped writes rather than printing the sql multiple times. Only applies if we support grouped writes"
	^false.! !

!SQLServerPlatform methodsFor: 'testing' stamp: ''!
usesIdentityColumns
	^true.! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
varbinary
	^self typeNamed: #varbinary ifAbsentPut: [GlorpVarBinaryType new].! !

!SQLServerPlatform methodsFor: 'types' stamp: ''!
varchar
	^self typeNamed: #varchar ifAbsentPut: [GlorpVarCharType new].! !

DatabasePlatform subclass: #SQLite3Platform
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!SQLite3Platform commentStamp: '<historical>' prior: 0!
I support Glorp's use of the SQLite database;  see http://www.sqlite.org.  Some notes on the implementation:

SQLite provides no specific support for time-related data other than column-default-value UTC functions:
	CURRENT_DATE format YYYY-MM-DD
	CURRENT_TIME format HH:MM:SS
	CURRENT_TIMESTAMP format YYYY-MM-DD HH:MM:SS
I therefore map Date, Time and Timestamp to and from strings via #to... and (inherited) #read... methods whose formats agree with these functions.  (I also return false to #supportsMillisecondsInTimes, again in agreement with the above functions which do not include millliseconds.)

My #maximumLengthOfTableName value is limited purely for convenience;  it can probably be much larger.

In VW, this version uses the non-threaded SQLite3Connection EXDI class.  Make my #vwEXDIConnectionClass return SQLite3ThreadedConnection to make Glorp use the threaded interface and, if using this with Store, also change its string reference in #asGlorpDatabase, #databaseFor:.  Alternatively, create an SQLlite3ThreadedPlatform subclass of me and add an extra line, instead of changing a line, to the Store methods.  (Note that in order to be thread-safe, your SQLite installation must be compiled with the SQLITE_THREADSAFE preprocessor macro set to 1.  In SQLite3, both the Windows and Linux precompiled binaries in the distribution are compiled this way.  If you are unsure if the SQLite library you are linking against is compiled to be threadsafe you can call the sqlite3_threadsafe()  interface to find out.)

SQLite parses FOREIGN KEY constraints but does not enforce them.

SQLite's approach to data typing is unusual for SQL.  Its data values are:
	NULL
	INTEGER. The value is a signed integer, stored in 1, 2, 3, 4, 6, or 8 bytes depending on the magnitude of the value.
	REAL. The value is a floating point value, stored as an 8-byte IEEE floating point number.
	TEXT. The value is a text string, stored using the database encoding (UTF-8, UTF-16BE or UTF-16-LE).
	BLOB. The value is a blob of data, stored exactly as it was input
but its typing is dynamic, so any column can in principle store any value, and its column affinities are
	TEXT - can store NULL, TEXT or BLOB
	NUMERIC - can store all five; text values that converts to reals or integers without loss of information are so stored
	INTEGER - as NUMERIC plus reals that convert to integer without loss of information are so stored
	REAL - as NUMERIC but integer values are coerced to reals
	NONE - no coercion is performed.
My int* and float*/double type methods that define Glorp types for SQLite conform to the expressibility of the above data values.

!

!SQLite3Platform methodsFor: 'constants' stamp: ''!
areSequencesExplicitlyCreated
	^false! !

!SQLite3Platform methodsFor: 'binding' stamp: ''!
bindingsForGroupWritingFor: aCommand
	"Return the bindings array for a group write. This can be in different formats, depending on the database and perhaps the mechanism in place."

	^aCommand batchStatementBindings! !

!SQLite3Platform methodsFor: 'constants' stamp: ''!
canBind: aValue to: aType
	"See #supportsVariableSizedNumerics comment."

	(Dialect isVisualWorks and: [aValue class == Dialect fixedPointClass]) ifTrue: [^false].
	^super canBind: aValue to: aType! !

!SQLite3Platform methodsFor: 'conversion-boolean' stamp: ''!
convertIntegerToBoolean: anInteger for: aType
	"This method accomodates an error in earlier SQLite 3 dlls. For the table_info() PRAGMA, 99 means True."

	anInteger isNil ifTrue: [^anInteger].
	anInteger >= 1 ifTrue: [^true].
	anInteger = 0 ifTrue: [^false].
	self error: 'invalid boolean conversion'! !

!SQLite3Platform methodsFor: 'services tables' stamp: ''!
createTableStatementStringFor: aGLORPDatabaseTable 
	"^<String> This method returns a string which can be used to create a database table ..."

	| sqlStatementStream tmpString fkeys |
	tmpString := 'create table'.
	sqlStatementStream := WriteStream on: String new.
	sqlStatementStream
		nextPutAll: (self capitalWritingOfSQLCommands 
					ifTrue: [tmpString asUppercase]
					ifFalse: [tmpString]);
		space.
	self printDDLTableNameFor: aGLORPDatabaseTable on: sqlStatementStream.

	"Now print the columns specification for each field in the table ..."
	self printColumnsSpecificationFor: aGLORPDatabaseTable
		on: sqlStatementStream.
	"Print the primary key constraint, unless using serial pkeys, since it would be a redundant declaration."
	(aGLORPDatabaseTable hasPrimaryKeyConstraints and: [aGLORPDatabaseTable primaryKeyFields first type isSerial not])
			ifTrue: [sqlStatementStream nextPutAll: ', '.
					(aGLORPDatabaseTable primaryKeyConstraints creationStringFor: self)
						do: [ :each | sqlStatementStream nextPutAll: each.]
						separatedBy: [sqlStatementStream nextPut: $,]].

	fkeys := aGLORPDatabaseTable foreignKeyConstraintsForCreation.
	fkeys notEmpty
			ifTrue: [sqlStatementStream nextPutAll: ', '.
					(fkeys collect: [:each| self createForeignKeyConstraintAddStatement: each])
						do: [ :each | sqlStatementStream nextPutAll: each first]
						separatedBy: [sqlStatementStream nextPut: $,]].

	sqlStatementStream
		nextPut: $).
	^sqlStatementStream contents! !

!SQLite3Platform methodsFor: 'sequences' stamp: ''!
databaseSequenceClass
	^SQLite3Sequence! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
date
	"SQLite does not support any time-related types;  use text as the type and convert in readDate.../toDate..."

	^self typeNamed: #date ifAbsentPut: [GlorpDateType new typeString: 'date']! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
double
	^self typeNamed: #double ifAbsentPut: [GlorpDoubleType new typeString: 'double' ].! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
float4
	"The only kind of REAL that SQLite supports is an 8-byte IEEE floating point number, so we just map float4 to float8."

	^self double! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
float8
	"The only kind of REAL that SQLite supports is an 8-byte IEEE floating point number."

	^self double! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
int2
	^self smallint.! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
int4
	^self typeNamed: #int4 ifAbsentPut: [GlorpIntegerType new typeString: 'int'].! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
int8
	^self int4.! !

!SQLite3Platform methodsFor: 'testing' stamp: ''!
isSQLite3Platform
	^true! !

!SQLite3Platform methodsFor: 'constants' stamp: ''!
maximumLengthOfTableName
	"It is the boast of SQLite that their limits are only for stability purposes:  if it fits in 32 bits then they will handle it.  Since the maximum length of an SQL statement is limited to SQLITE_MAX_LENGTH (defaults to 1000000, cannot be higher than 1073741824), table names had better be shorter and in normal use should be much shorter.  I chose the number below at random;  change it if it causes you any inconvenience."

	^128! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
nvarchar
	^self text! !

!SQLite3Platform methodsFor: 'constants' stamp: ''!
parenthesizeCommandsInCompoundQuery
	^false! !

!SQLite3Platform methodsFor: 'conversion-times' stamp: ''!
printDate: aDateString for: aType
	"SQLite does not suppport any time-related types.  We convert them all to strings via the to... methods, so parameters to this method should be Strings of the form '1959-09-30'."

	aDateString isNil ifTrue: [^'NULL'].
	"The next line is checking code and can be dropped when SQLite is sufficiently tested."
	aDateString class = Date ifTrue: [self error: 'SQLite does not support date types;  this should have been a String'].
	^aDateString printString! !

!SQLite3Platform methodsFor: 'conversion-times' stamp: ''!
printTime: aTimeString for: aType
	"SQLite does not suppport any time-related types.  We convert them all to strings via the to... methods, so parameters to this method should be Strings of the form 10:20:23."

	aTimeString isNil ifTrue: [^'NULL'].
	"The next line is checking code and can be dropped when SQLite is sufficiently tested."
	aTimeString class = Time ifTrue: [self error: 'SQLite does not support time types;  this should have been a String'].
	^aTimeString printString! !

!SQLite3Platform methodsFor: 'conversion-times' stamp: ''!
printTimestamp: aTimestampString on: stream for: aType
	"SQLite does not suppport any time-related types.  We convert them all to strings via the to... methods, so parameters to this method should be Strings of the form '1959-03-27 10:20:23'."

	aTimestampString isNil ifTrue: [stream nextPutAll: 'NULL'. ^self].
	"The next line is checking code and can be dropped when SQLite is sufficiently tested."
	aTimestampString class = Timestamp ifTrue: [self error: 'SQLite does not support timestamp types;  this should have been a String'].
	stream nextPut: $'.
	stream nextPutAll: aTimestampString.
	stream nextPut: $'.! !

!SQLite3Platform methodsFor: 'testing' stamp: ''!
requiresTransactionForTableOperations
	^false! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
sequence
	^self serial! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
serial
	"In SQLLite, the column declared as INTEGER PRIMARY KEY will default to ROWID, unless the inserted value is not null. But in Sqlite3 a declaration like: INTEGER PRIMARY KEY AUTOINCREMENT will cause ROWID to be fetched from the sqlite_sequence table, which provides always increasing values."

	^self typeNamed: #serial ifAbsentPut: [GlorpSerialType new typeString: 'INTEGER PRIMARY KEY AUTOINCREMENT'].! !

!SQLite3Platform methodsFor: 'testing' stamp: ''!
supportsANSIJoins
	"Do we support the JOIN <tableName> USING <criteria> syntax."
	^true.! !

!SQLite3Platform methodsFor: 'testing' stamp: ''!
supportsAlterTableForConstraints

	^false! !

!SQLite3Platform methodsFor: 'constants' stamp: ''!
supportsAuthentication
	"SQLite does not support authentication:  anyone can login with any username and password.  The only security is that provided by the OS for the file being used as the database."

	^false! !

!SQLite3Platform methodsFor: 'testing' stamp: ''!
supportsBinding
	"Return true if this platform supports binding parameters rather than printing them as strings into the SQL statement"
	^true.! !

!SQLite3Platform methodsFor: 'constants' stamp: ''!
supportsConstraints
	"Return false, because this platform does not support constraint add or drop statements.
	However, foreign key constraints can be created as part of the create table statement.
	Also, a foreign key constraint can be created in the statement that adds that column to
	the table. (Eg., ALTER TABLE ADD COLUMN ...)."

	^false! !

!SQLite3Platform methodsFor: 'constants' stamp: ''!
supportsMillisecondsInTimes
	"The only time-type-related functions in SQLite are CURRENT_TIME format HH:MM:SS and CURRENT_TIMESTAMP format YYYY-MM-DD HH:MM:SS, neither of which support milliseconds.  As against that, since we map to and from strings, we can easily include milliseconds in these strings.  For now, I go with the SQLite function choice and say no, but I have verified that tests work fine when I set this to true."

	^false! !

!SQLite3Platform methodsFor: 'testing' stamp: ''!
supportsSchemas
	^false! !

!SQLite3Platform methodsFor: 'testing' stamp: ''!
supportsStringQueryOnBlobs
	"This is something of a guess but I see no evidence that SQLite does and the relevant test certainly fails."

	^false! !

!SQLite3Platform methodsFor: 'constants' stamp: ''!
supportsVariableSizedNumerics
	"SQLite supports a single 8-byte IEEE floating point type:  e.g. give it 321.12s and it will return 321.12d.  Perhaps clever Glorp converters could arrange for Glorp to know which fields are really FixedPoint numbers, as we do for Date, Time and Timestamp (which SQLite also does not support) but I'll leave that until there's demand for it (ee also #canBind:to:)."

	^false! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
text
	^self typeNamed: #nvarchar ifAbsentPut: [GlorpTextType new typeString: 'text']! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
time
	"SQLite does not support any time-related types;  use text as the type and convert in readTime.../toTime..."

	^self typeNamed: #time ifAbsentPut: [GlorpTimeType new typeString: 'time']! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
timestamp
	"SQLite does not support any time-related types;  use text as the type and convert in readTimestamp.../toTimestamp..."

	^self typeNamed: #timestamp ifAbsentPut: [GlorpTimeStampType new typeString: 'timestamp']! !

!SQLite3Platform methodsFor: 'conversion-times' stamp: ''!
toDate: anObject for: aType
	"SQLite does not handle timestamps;  all such must be treated as strings and the to... and read... formats must agree.  SQLite does use column-default-value UTC functions:
	CURRENT_DATE format YYYY-MM-DD
	CURRENT_TIME format HH:MM:SS
	CURRENT_TIMESTAMP format YYYY-MM-DD HH:MM:SS
so our string conversions must agree with these formats."

	| stream |
	anObject isNil ifTrue: [^nil].
	stream := String new writeStream.
	self
		printDate: (super toDate: anObject for: aType)
		isoFormatOn: stream.
	^stream contents! !

!SQLite3Platform methodsFor: 'conversion-times' stamp: ''!
toTime: anObject for: aType
	"SQLite does not handle timestamps;  all such must be treated as strings and the to... and read... formats must agree.  SQLite does use column-default-value UTC functions:
	CURRENT_DATE format YYYY-MM-DD
	CURRENT_TIME format HH:MM:SS
	CURRENT_TIMESTAMP format YYYY-MM-DD HH:MM:SS
so our string conversions must agree with these formats."

	| stream |
	anObject isNil ifTrue: [^nil].
	stream := String new writeStream.
	self
		printTime: (super toTime: anObject for: aType)
		isoFormatOn: stream.
	^stream contents! !

!SQLite3Platform methodsFor: 'conversion-times' stamp: ''!
toTimestamp: anObject for: aType
	"SQLite does not handle timestamps;  all such must be treated as strings and the to... and read... formats must agree.  SQLite does use column-default-value UTC functions:
	CURRENT_DATE format YYYY-MM-DD
	CURRENT_TIME format HH:MM:SS
	CURRENT_TIMESTAMP format YYYY-MM-DD HH:MM:SS
so our string conversions must agree with these formats."

	| stream timestampObject |
	anObject isNil ifTrue: [^nil].
	stream := String new writeStream.
	timestampObject := super toTimestamp: anObject for: aType.
	self
		printDate: timestampObject
		isoFormatOn: stream.
	stream space.
	self
		printTime: timestampObject
		isoFormatOn: stream.
	^stream contents! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
typeWithDatabaseName: aString characterWidth: aNumber
	"Since this is SQLite, let the type names be more flexible."

	| aTypeName total block |
	block := [:sum :name | sum + (aString asLowercase findString: name startingAt: 1)].
	aTypeName := aString.
	total := #('char' 'clob' 'text') inject: 0 into: block.
	total > 0
		ifTrue: [aTypeName := 'text']
		ifFalse: 
			[total := #('num' 'double' 'float') inject: 0 into: block.
			total > 0
				ifTrue: [aTypeName := 'numeric']
				ifFalse: 
					[total := #('int') inject: 0 into: block.
					total > 0
						ifTrue: [aTypeName := 'integer']
						ifFalse: 
							[total := #('blob') inject: 0 into: block.
							total > 0 ifTrue: [aTypeName := 'blob']]]].
	aString isEmpty ifTrue: [aTypeName := 'blob'].
	^super typeWithDatabaseName: aTypeName characterWidth: aNumber! !

!SQLite3Platform methodsFor: 'testing' stamp: ''!
usesArrayBindingRatherThanGrouping
	^false.! !

!SQLite3Platform methodsFor: 'testing' stamp: ''!
usesIdentityColumns
	^true.! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
varbinary
	" *** This method was defined by Glorp.DatabasePlatform as a subclass responsibility.
	Replace its body with a proper implementation. *** "

	self error: 'Subclass responsibility stub not reimplemented'! !

!SQLite3Platform methodsFor: 'types' stamp: ''!
varchar
	^self text.! !

!SQLite3Platform methodsFor: 'accessing' stamp: ''!
vwEXDIConnectionClass
	^'SQLite3Connection' asQualifiedReference valueIfUndefinedLoadFrom: 'SQLite3EXDI'! !

Object subclass: #DatabaseRow
	instanceVariableNames: 'table contents shouldBeWritten owner forDeletion status oldVersion generated parent operation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

DatabaseRow class
	instanceVariableNames: 'missingFieldIndicator emptySlot'!

!DatabaseRow commentStamp: '<historical>' prior: 0!
This represents the data to be written out to a row. Database rows are normally stored in a rowmap, keyed according to their table and the object that did the primary writes to them. We expect that that's only one object, although embedded values are an exception to that.

Instance Variables:

	table	<DatabaseTable>	The table holding the data
	contents	<IdentityDictionary>	Holds the fields with their values, indirectly through FieldValueWrapper instances.
	shouldBeWritten	<Boolean>	Normally true, but can be set false to suppress writing of a particular row. Used with embedded value mappings, where we create their row, unify it with the parent row, and suppress writing of the original row.
	owner	<Object>	The primary object that wrote into this row, would also be the key into the rowmap.
	forDeletion <Boolean> Is this row to be deleted or written.
	status <Symbol> Used for topological sort to determine write order. We keep a symbol indicating the "visited" status of the row while doing the sort. This doesn't separate the sort from the row, using each row directly as a node in the graph. This is less general, but uses less memory than other alternatives.
	parent <DatabaseRow> Also used during sorting. Keep track of our parent node.
	oldVersion <DatabaseRow> The old version of us. Used if we increment a version field to keep track of what its value was before.
	generated <Boolean> True if we have already had field with generated types generated in the preWrite phase. This is needed because we generate all sequences before doing any writes. But if we use an explicit writeRow: operation, then we want to make sure that generated values are generated. And some generated types write over what was there before (e.g. optimistic lock fields). So we need a way of knowing if generation has already been done. Note that this doesn't apply if using post-write sequence generation (i.e. identity columns) because then it's all after the write. This makes certain things not work as well, but that's what you get for using identity columns.
	!

DatabaseRow class
	instanceVariableNames: 'missingFieldIndicator emptySlot'!

!DatabaseRow classSide methodsFor: 'private' stamp: ''!
emptySlot
	emptySlot == nil ifTrue: [emptySlot := Object new].
	^emptySlot.! !

!DatabaseRow classSide methodsFor: 'private' stamp: ''!
missingFieldIndicator
	missingFieldIndicator == nil ifTrue: [missingFieldIndicator := Object new].
	^missingFieldIndicator.! !

!DatabaseRow classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!DatabaseRow classSide methodsFor: 'instance creation' stamp: ''!
newForTable: aTable

	^self new table: aTable.! !

!DatabaseRow classSide methodsFor: 'instance creation' stamp: ''!
newForTable: aTable withOwner: anObject

	^self new table: aTable; owner: anObject.! !

!DatabaseRow methodsFor: 'comparing' stamp: ''!
= aRow
	self class == aRow class ifFalse: [^false].
	^self equals: aRow.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
at: aField

	^self at: aField ifAbsent: [self error: 'missing field'].! !

!DatabaseRow methodsFor: 'private' stamp: ''!
at: aField forcePut: aValue 
	"Write the value. If there's a previous value, overwrite it. Use with extreme caution."
	| wrapper |
	aValue isGlorpExpression ifTrue: [self error: 'cannot store expressions in rows'].
	aField table == self table ifFalse: [self error: 'Invalid table'].
	wrapper := self contentsAt: aField ifAbsentPut: [FieldValueWrapper new].
	wrapper privateSetContents: aValue.
	^wrapper.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
at: aField ifAbsent: absentBlock

	^(self wrapperAt: aField ifAbsent: [^absentBlock value]) contents.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
at: aField ifAbsentPut: absentBlock

	| old |
	old := contents at: aField position.
	old == self class emptySlot 
		ifTrue: absentBlock
		ifFalse: [^old contents].! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
at: aField put: aValue 
	
	"For generated fields, we expect the real value to be provided later by the database, so don't write a nil value"
	
	| wrapper |
"	aValue isGlorpExpression ifTrue: [self error: 'cannot store expressions in rows']."
	aField table == self table ifFalse: [self error: 'Invalid table'].
	wrapper := self contentsAt: aField ifAbsentPut: [FieldValueWrapper new].
	(aValue isNil and: [aField isGenerated]) ifFalse: [wrapper contents: aValue].
	wrapper isNowContainedBy: self and: aField.
	^wrapper.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
atFieldNamed: aString

	| field |
	field := table fieldNamed: aString.
	^self at: field! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
atFieldNamed: aString ifAbsent: aBlock

	| field |
	field := table fieldNamed: aString ifAbsent: aBlock.
	^self at: field ifAbsent: aBlock.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
atFieldNamed: aString put: anObject

	| field |
	field := table fieldNamed: aString.
	^self at: field put: anObject! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
contentsAt: aField ifAbsentPut: absentBlock

	| old newValue |
	old := contents at: aField position.
	^old == self class emptySlot 
		ifTrue: [
			newValue := absentBlock value.
			contents at: aField position put: newValue.
			newValue]
		ifFalse: [old].! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
doesNotUnderstand: aMessage
	^self atFieldNamed: aMessage selector asString ifAbsent: [super doesNotUnderstand: aMessage].! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
equalityStringForField: aDatabaseField

	| stream |
	stream := WriteStream on: (String new: 50).
	self printEqualityStringForField: aDatabaseField on: stream.
	^stream contents.! !

!DatabaseRow methodsFor: 'testing' stamp: ''!
equals: aRow 
	self forDeletion = aRow forDeletion ifFalse: [^false].
	self fieldsAndValuesDo: 
			[:eachField :eachWrapper | 
			| otherValue |
			otherValue := aRow at: eachField ifAbsent: [self class missingFieldIndicator].
			eachWrapper = otherValue ifFalse: [^false]].
	^true! !

!DatabaseRow methodsFor: 'querying' stamp: ''!
expectsToGetValueFor: aField
	"Are we constrained to a field which is presumably going to get a value, which for the moment must mean it's a serial type. This is primarily required for identity column databases, because in other cases we will pre-allocate sequence numbers, and this should never come up"
	| wrapper innerWrappers testBlock containing |
	wrapper := self wrapperAt: aField ifAbsent: [^false].
	innerWrappers := wrapper allInnerWrappers.
	testBlock := [:eachField | (eachField ~~ aField and: [eachField type isSerial]) ifTrue: [^true]].
	(Array with: wrapper), innerWrappers do: [:eachWrapper |
		containing := eachWrapper containedBy.
		containing glorpIsCollection 
			ifTrue: [containing keysDo: testBlock]
			ifFalse: [testBlock value: containing key]].
	^false.! !

!DatabaseRow methodsFor: 'enumerating' stamp: ''!
fieldValuesDo: aBlock 
	contents do: [:each | each == self class emptySlot ifFalse: [aBlock value: each]].! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
fields
	^self table fields select: [:each | (self includesField: each)].! !

!DatabaseRow methodsFor: 'initializing' stamp: ''!
fieldsAndValidValuesDo: aBlock 
	"If iterating over fields and values, we include wrappers with no value assigned yet. This might or might not be what we want. This one just iterates over ones with actual values"
	self isEmpty ifTrue: [^self].
	table fields do: [:each |  | value |
		value := self at: each ifAbsent: [self class missingFieldIndicator].
		value == self class missingFieldIndicator ifFalse: [
			aBlock value: each value: value]].! !

!DatabaseRow methodsFor: 'initializing' stamp: ''!
fieldsAndValuesDo: aBlock 
	table isNil ifTrue: [^self].
	table fields do: [:each | aBlock value: each value: (self at: each ifAbsent: [self class missingFieldIndicator])]! !

!DatabaseRow methodsFor: 'enumerating' stamp: ''!
fieldsDo: aBlock 
	1 to: contents size do: [:i |
		| value |
		value := contents at: i.
		(value == self class emptySlot) ifFalse: [aBlock value: (self table fields at: i)]].! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
fieldsNeedingUpdate
	"Return a list of the fields that are different from the previous version (if we know what that was)"

	| result |
	oldVersion isNil ifTrue: [^self nonPrimaryKeyFields].
	result := OrderedCollection new: contents size.
	self fieldsDo: 
			[:each | 
			each isPrimaryKey 
				ifFalse: 
					[(oldVersion at: each ifAbsent: [self class missingFieldIndicator]) 
						= (self at: each) ifFalse: [result add: each]]].
	^result! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
forDeletion
	^forDeletion.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
forDeletion: aBoolean
	forDeletion := aBoolean.! !

!DatabaseRow methodsFor: 'querying' stamp: ''!
hasValueFor: aField

	^(self wrapperAt: aField ifAbsent: [^false]) hasValue.! !

!DatabaseRow methodsFor: 'testing' stamp: ''!
hash
	| basic |
	basic := self table hash.
	self fieldsAndValuesDo: 
			[:eachField :eachValue | 
			basic := basic bitXor: eachField hash.
			basic := basic bitXor: eachValue hash].
	^basic.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
includesField: aField
	^(contents at: aField position) ~~ self class emptySlot.! !

!DatabaseRow methodsFor: 'initializing' stamp: ''!
initialize
	shouldBeWritten := true.
	forDeletion := false.
	generated := false.! !

!DatabaseRow methodsFor: 'testing' stamp: ''!
isEmpty
	| empty |
	contents isNil ifTrue: [^true].
	empty := self class emptySlot.
	contents do: [:each | each == empty ifFalse: [^false]].
	^true.! !

!DatabaseRow methodsFor: 'enumerating' stamp: ''!
keysAndValuesDo: aBlock
	1 to: contents size do: [:i |
		| value | 
		value := contents at: i.
		value == self class emptySlot ifFalse: [
			aBlock value: (self table fields at: i) value: value contents]].! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
lockKeyFields

	^table lockKeyFields! !

!DatabaseRow methodsFor: 'sequencing' stamp: ''!
needsValueFromSequence: aSequence 
	self table fields do: 
			[:each | 
			((self shouldGenerateValueFor: each) and: [each type sequence == aSequence]) 
				ifTrue: [^true]].
	^false! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
nonGeneratedFields
	"Return a list of our fields that are not generated"
	| result |
	result := OrderedCollection new: contents size.
	self table fields do: [:field | 
		field isGenerated not ifTrue: [result add: field]].
	^result.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
nonGeneratedFieldsWithValues
	"Return a list of our fields that a) are not generated or b) have values. That is, exclude values we expect the database to generate"
	| result |
	result := OrderedCollection new: contents size.
	self fieldsAndValidValuesDo: [:field :value | 
		(value notNil or: [field isGenerated not]) ifTrue: [result add: field]].
	^result.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
nonPrimaryKeyFields
	| result |
	result := OrderedCollection new: contents size.
	self fieldsDo: [:field | 
		field isPrimaryKey ifFalse: [result add: field]].
	^result.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
numberOfFields
	^contents
		inject: 0
		into:
			[:sum :each | sum + (each == self class emptySlot ifTrue: [0] ifFalse: [1])].! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
oldVersionAt: aField ifAbsent: aBlock
	oldVersion isNil ifTrue: [^aBlock value].
	^oldVersion at: aField ifAbsent: aBlock.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
owner

	^owner.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
owner: anObject

	owner := anObject.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
parent
	^parent! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
parent: aDatabaseRow
	parent := aDatabaseRow.! !

!DatabaseRow methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequencesUsing: anAccessor

	self table fields do: [:each |
		(self hasValueFor: each) ifFalse: [
			each type postWriteAssignSequenceValueFor: each in: self using: anAccessor]].! !

!DatabaseRow methodsFor: 'sequencing' stamp: ''!
preWriteAssignSequencesUsing: aSession
	"Skip if we've already been generated"
	generated ifTrue: [^self].
	self table fields do: 
			[:each | 
			(self shouldGenerateValueFor: each in: aSession) 
				ifTrue: 
					[each type 
						preWriteAssignSequenceValueFor: each
						in: self
						using: (aSession isNil ifTrue: [nil] ifFalse: [aSession accessor])]].
	generated := true.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
primaryKey

	| |
	self table primaryKeyFields isEmpty ifTrue: [^nil].
	^self table hasCompositePrimaryKey 
		ifTrue: [
			self table primaryKeyFields 
				collect: [:each | 	self at: each]]
		ifFalse: [self at: self table primaryKeyFields first ifAbsent: [nil]].! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
printEqualityStringForField: aDatabaseField on: aCommand 
	"Get around PostgreSQL bug.  Qualified names cannot appear in SET expression."
	aCommand nextPutAll: (aDatabaseField platform nameForColumn: aDatabaseField name).
	aCommand nextPutAll: ' = '.
	self printValueOfField: aDatabaseField on: aCommand! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
printEqualityTemplateForField: aDatabaseField on: aCommand
	"Get around PostgreSQL bug.  Qualified names cannot appear in SET expression."
	| bind |
	aCommand nextPutAll: (aDatabaseField platform nameForColumn: aDatabaseField name).
	aCommand nextPutAll: ' = '.
	bind := aCommand canBind: (self at: aDatabaseField ifAbsent: [nil]) to: aDatabaseField type.
	bind
		ifTrue: [aCommand nextPutAll: '?']
		ifFalse: [self printValueOfField: aDatabaseField on: aCommand].! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
printEqualityTemplateForField: aDatabaseField on: aCommand withBinding: aBoolean
	"Get around PostgreSQL bug.  Qualified names cannot appear in SET expression."
	aCommand nextPutAll: (aDatabaseField platform nameForColumn: aDatabaseField name).
	aCommand nextPutAll: ' = '.
	aBoolean 
		ifTrue: [aCommand nextPutAll: '?']
		ifFalse: [self printValueOfField: aDatabaseField on: aCommand].! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
printFieldNamesOn: aWriteStream 
	GlorpHelper 
		do: [:each | aWriteStream nextPutAll: each name]
		for: self table fields
		separatedBy: [aWriteStream nextPutAll: ','].! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
printLockEqualityForField: aDatabaseField on: aStream
	| expression |
	expression := ((BaseExpression new getTable: aDatabaseField table) getField: aDatabaseField) get: #= withArguments: (Array with: (oldVersion at: aDatabaseField)).
	expression printSQLOn: aStream withParameters: #().! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: '(' , ((table isNil or: [table name isNil]) ifTrue: [''] ifFalse: [table name]),  ')'.
	aStream cr.
	self fieldsAndValuesDo: [:eachField :eachWrapper |
		aStream nextPutAll: '    '.
		eachField printOn: aStream.
		aStream nextPutAll: '->'.
		eachWrapper printOn: aStream.
		aStream cr.].! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
printPrimaryKeyStringOn: aStream 
	"If there is no primary key (i.e. this is a link table) use all the values that we have"
	| fields |
	fields := table primaryKeyFields isEmpty 
		ifTrue: [self fields]
		ifFalse: [table primaryKeyFields].
	GlorpHelper
		do:  [:eachField |
			self printEqualityStringForField: eachField on: aStream]
		for: fields
		separatedBy: [aStream nextPutAll: ' AND '].! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
printPrimaryKeyTemplateOn: aStream
	"If there is no primary key (i.e. this is a link table) use all the values that we have"
	| fields |
	fields := table primaryKeyFields isEmpty 
		ifTrue: [self fields]
		ifFalse: [table primaryKeyFields].
	GlorpHelper
		do:  [:eachField |
			self printEqualityTemplateForField: eachField on: aStream]
		for: fields
		separatedBy: [aStream nextPutAll: ' AND '].! !

!DatabaseRow methodsFor: 'printing' stamp: ''!
printValueOfField: aDatabaseField on: aWriteStream 
	aDatabaseField type print: (self at: aDatabaseField ifAbsent: [nil]) on: aWriteStream.! !

!DatabaseRow methodsFor: 'relationships' stamp: ''!
relatedRowsIn: aRowMap
	| relatedRows |
	relatedRows := OrderedCollection new.
	self table foreignKeyConstraints
		do:
			[:eachConstraint | relatedRows addAll: (eachConstraint rowsRelatedTo: self)].
	^relatedRows.! !

!DatabaseRow methodsFor: 'querying' stamp: ''!
shouldBeWritten
	table isNil ifTrue: [^shouldBeWritten].
	^shouldBeWritten and: [table shouldBeWritten].! !

!DatabaseRow methodsFor: 'configuring' stamp: ''!
shouldBeWritten: aBoolean 
	shouldBeWritten := aBoolean! !

!DatabaseRow methodsFor: 'sequencing' stamp: ''!
shouldGenerateValueFor: aField
	^self shouldGenerateValueFor: aField in: nil.! !

!DatabaseRow methodsFor: 'sequencing' stamp: ''!
shouldGenerateValueFor: aField in: aSession
	| isInsert |
	aField isGenerated ifFalse: [^false].
	aField type generatesOverExistingValues ifTrue: [^true].
	"For those that don't write over existing values, only update them if this is an insert. Otherwise we can't tell the difference between an empty column and one which we're just not updating in this unit of work. This requires the session, but the session may be nil - if so, default to true, because that usage is just for reserving sequence numbers, and it's better to err on the side of too many."
	isInsert := aSession isNil ifTrue: [true] ifFalse: [aSession shouldInsert: self].
	^isInsert and: [(self hasValueFor: aField) not].! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
status
	^status! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
status: aSymbol
	status := aSymbol! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
table
	"Private - Answer the value of the receiver's ''table'' instance variable."

	^table! !

!DatabaseRow methodsFor: 'configuring' stamp: ''!
table: aDatabaseTable
	table := aDatabaseTable.
	contents := Array new: (aDatabaseTable fields size).
	contents atAllPut: self class emptySlot.! !

!DatabaseRow methodsFor: 'api' stamp: ''!
unifyField: eachSourceField withField: eachTargetField inRow: targetRow
	| sourceWrapper targetWrapper |
	sourceWrapper := self wrapperAt: eachSourceField ifAbsent: [nil].
	targetWrapper := targetRow wrapperAt: eachTargetField ifAbsent: [nil].
	sourceWrapper isNil
		ifTrue:
			[targetWrapper isNil
				ifTrue:
					[targetWrapper := FieldValueWrapper new.
					targetRow wrapperAt: eachTargetField put: targetWrapper].
			^self wrapperAt: eachSourceField put: targetWrapper].
	targetWrapper isNil ifTrue: [^targetRow wrapperAt: eachTargetField put: sourceWrapper].
	sourceWrapper unifyWith: targetWrapper.! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
values
	| result |
	result := OrderedCollection new.
	self keysAndValuesDo: [:key :value |
		result add: value].
	^result! !

!DatabaseRow methodsFor: 'configuring' stamp: ''!
withAllFieldsIn: aRow 
	"aRow represents our original state. Make sure that we have all the fields in aRow, using nil values for any that are missing. This is needed if, e.g. we have been removed from a 1-many relationship, so we don't get a value generated for our foreign key, but we should still write it as a nil. We have to distinguish this from the case of a value that simply hasn't changed."

	aRow isEmpty ifTrue: [^self].
	oldVersion := aRow.
	self numberOfFields = table fields size ifTrue: [^self].
	aRow fieldsAndValidValuesDo: 
			[:eachField :eachValue | 
			(self includesField: eachField) ifFalse: [self at: eachField put: nil]].
	^self! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
wrapperAt: aField

	^self wrapperAt: aField ifAbsent: [self error: 'Field not found'].! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
wrapperAt: aField ifAbsent: aBlock

	| thing |
	contents isNil ifTrue: [^aBlock value].
	thing := contents at: aField position.
	^thing == self class emptySlot ifTrue: aBlock ifFalse: [thing].! !

!DatabaseRow methodsFor: 'accessing' stamp: ''!
wrapperAt: aField put: aWrapper
	| old |
	old := contents at: aField position.
	old == aWrapper ifTrue: [^self].
	contents at: aField position put: aWrapper.
	aWrapper isNowContainedBy: self and: aField.! !

Object subclass: #DatabaseTable
	instanceVariableNames: 'name schema fields foreignKeyConstraints indexes parent lockKeyFields primaryKeyFields isImaginary replacementSubSelect uniqueConstraints'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!DatabaseTable commentStamp: '<historical>' prior: 0!
This is our internal representation of a table in the database. It holds onto enough information for us to both map to the table, and to create it in the database if required.

Instance Variables
	fields	<(Collection of: DatabaseField)>	The columns in the table.
	foreignKeyConstraints	<(Collection of: ForeignKeyConstraint)>	The foreign key constraints between this table and others.
	indexes	<(Collection of: DatabaseIndex)>	The indexes on this table.
	isImaginary	<Boolean>	We allow "imaginary" tables, which don't really exist in the database. This is most useful in connection with embedded mappings. For example, we can have a Currency object which maps to a CURRENCY table, but there is no currency table in the database, and instead we have a number of embedded mappings that put those fields inside another table.
	lockKeyFields	<Collection of: DatabaseField>	Which fields (which should also be in our field list) are checked when doing optimistic locking
	name	<String>	The name of the table.
	parent	<DatabaseTable | nil>	For a normal table, this is nil. When we create an aliased table we remember who the original table was using this field.
	primaryKeyFields	<(SequenceableCollection of: (DatabaseField))>	Which fields (which should also be in our field list) are primary keys.
	replacementSubSelect	<CompoundQuery | SimpleQuery>	I honestly don't remember at the moment what this does, but it seems to be used when we're traversing relationships that are in imaginary tables in order to figure out what the actual join has to be. And maybe the join needs to be done in a subselect rather than as a simple join in certain cases.
	schema	<String>	What is the name of the schema in which this table exists. This is used when printing the qualified name of the table. 
	uniqueConstraints	<(Collection of: UniqueConstraint)>	The unique constraints for this table
!

!DatabaseTable classSide methodsFor: 'instance creation' stamp: ''!
named: aString

	^self new name: aString.! !

!DatabaseTable classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!DatabaseTable methodsFor: 'comparing' stamp: ''!
<= aTable
	^self qualifiedName <= aTable qualifiedName.! !

!DatabaseTable methodsFor: 'private/fields' stamp: ''!
addAsLockKeyField: aField 

	lockKeyFields add: aField! !

!DatabaseTable methodsFor: 'private/fields' stamp: ''!
addAsPrimaryKeyField: aField

	(fields includes: aField) ifFalse: [self error: 'Primary key field not part of the table'].
	aField isPrimaryKey ifFalse: ["Somebody is calling this as API, since there's an example out on the net that does it that way - just do the right thing instead"
		^aField bePrimaryKey].
	(primaryKeyFields notNil and: [primaryKeyFields includes: aField]) ifTrue: [^self].
	primaryKeyFields := primaryKeyFields isNil ifTrue: [Array with: aField] ifFalse: [primaryKeyFields, (Array with: aField)].! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
addField: aField

	fields add: aField.
	aField isPrimaryKey ifTrue: [
		self addAsPrimaryKeyField: aField].
	aField table: self.
	aField position: fields size.
	^aField.! !

!DatabaseTable methodsFor: 'obsolete' stamp: ''!
addFields: aCollection fromAccessor: aDatabaseAccessor
	| stream |
	aCollection
		do:
			[:each | 
			aDatabaseAccessor
				execute:
					[stream := WriteStream on: String new.
					stream
						nextPutAll: 'ALTER TABLE ';
						nextPutAll: self name;
						nextPutAll: ' ADD ';
						nextPutAll: self platform startColumnAddString.
					self platform printColumn: each on: stream.
					stream nextPutAll: self platform endColumnAddString.
					aDatabaseAccessor executeSQLStringNoResult: stream contents]
				ifError: [:ex | ex pass]].! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
addForeignKeyFrom: sourceField to: targetField
	^self
		addForeignKeyFrom: sourceField
		to: targetField
		suffixExpression: nil.! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
addForeignKeyFrom: sourceField to: targetField from: sourceField2 to: targetField2
	^self
		addForeignKeyFromAll: (Array with: sourceField with: sourceField2)
		toAll: (Array with: targetField with: targetField2)
		suffixExpression: nil.! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
addForeignKeyFrom: sourceField to: targetField from: sourceField2 to: targetField2 from: sourceField3 to: targetField3
	^self
		addForeignKeyFromAll: (Array with: sourceField with: sourceField2 with: sourceField3)
		toAll: (Array with: targetField with: targetField2 with: targetField3)
		suffixExpression: nil.! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
addForeignKeyFrom: sourceField to: targetField suffixExpression: suffixExpression
	| newFK |
	newFK := ForeignKeyConstraint
					sourceField: sourceField
					targetField: targetField
					suffixExpression: suffixExpression.
	newFK name: (newFK name , (foreignKeyConstraints size + 1) printString).
	foreignKeyConstraints add: newFK.
	^newFK.! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
addForeignKeyFrom: sourceField toField: targetField

	^self addForeignKeyFrom: sourceField to: targetField! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
addForeignKeyFrom: sourceField toTable: targetTable

	^self addForeignKeyFrom: sourceField toField: (targetTable fieldNamed: 'ID' ifAbsent: [^nil])! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
addForeignKeyFromAll: sourceFields toAll: targetFields
	^self addForeignKeyFromAll: sourceFields toAll: targetFields suffixExpression: nil.! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
addForeignKeyFromAll: sourceFields toAll: targetFields suffixExpression: suffixExpression
	| newFK |
	newFK := ForeignKeyConstraint
					sourceFields: sourceFields
					targetFields: targetFields
					suffixExpression: suffixExpression.
	newFK name: (newFK name , (foreignKeyConstraints size + 1) printString).
	^foreignKeyConstraints add: newFK.! !

!DatabaseTable methodsFor: 'indexes' stamp: ''!
addIndex: aDatabaseIndex
	^indexes add: aDatabaseIndex.! !

!DatabaseTable methodsFor: 'indexes' stamp: ''!
addIndexForField: aField
	^indexes add: (DatabaseIndex forField: aField).! !

!DatabaseTable methodsFor: 'indexes' stamp: ''!
addIndexForField: aField and: anotherField
	^self addIndexForFields: (Array with: aField with: anotherField).! !

!DatabaseTable methodsFor: 'indexes' stamp: ''!
addIndexForField: aField and: anotherField and: thirdField
	^self addIndexForFields: (Array with: aField with: anotherField with: thirdField).! !

!DatabaseTable methodsFor: 'indexes' stamp: ''!
addIndexForFields: aFieldCollection
	^indexes add: (DatabaseIndex forFields: aFieldCollection).! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
addUniqueConstraintFor: sourceFields

	| newConstraint |
	newConstraint := UniqueConstraint
		forTable: self
		fields: sourceFields.
	^uniqueConstraints add: newConstraint! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
allSequences
	"Return a dictionary of all the sequences associated with this table, indexed by name."

	| result |
	result := Dictionary new.
	self fields do: [:eachField | 
 		eachField type hasSequence ifTrue:
			[result at: eachField type sequence name put: eachField type sequence]].
	^result.! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
createFieldNamed: aString type: dbType

	| existingField |
	existingField := fields detect: [:each | each name = aString] ifNone: [nil].
	(existingField notNil)
	ifTrue: [self error: 'field ', aString, ' already exists'].
	^self addField: (DatabaseField named: aString type: dbType)! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
creator
	^self schema! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
creator: aString
	"For backward-compatibility. Use schema: instead."
	self schema: aString.! !

!DatabaseTable methodsFor: 'obsolete' stamp: ''!
dropFields: aCollection fromAccessor: aDatabaseAccessor
	aCollection
		do:
			[:each | 
			aDatabaseAccessor
				execute:
					[aDatabaseAccessor
						executeSQLStringNoResult:
							'ALTER TABLE ' , self name , ' DROP COLUMN ' , each name]
				ifError: [:ex | ex pass]].! !

!DatabaseTable methodsFor: 'obsolete' stamp: ''!
dropForeignKeyConstraintsFromAccessor: aDatabaseAccessor 
	self foreignKeyConstraints 
		do: [:each | aDatabaseAccessor dropConstraint: each]! !

!DatabaseTable methodsFor: 'create/delete in db' stamp: ' 22/6/07 12:54'!
dropFromAccessor: aDatabaseAccessor
	aDatabaseAccessor platform supportsConstraints
		ifTrue: [self dropPrimaryKeyConstraintsFromAccessor: aDatabaseAccessor].
	aDatabaseAccessor dropTableNamed: self name.! !

!DatabaseTable methodsFor: 'obsolete' stamp: ''!
dropPrimaryKeyConstraintsFromAccessor: aDatabaseAccessor
	(self primaryKeyFields isEmpty not and: [self platform usesPrimaryKeyUniqueConstraintNames])
		ifTrue: [
			aDatabaseAccessor	execute:
				[aDatabaseAccessor
						executeSQLStringNoResult:
							'ALTER TABLE ' , self name , ' DROP '
								, self platform nameForPrimaryKeyConstraintsWhenDropping , ' '
								, self primaryKeyUniqueConstraintName]
				ifError:
					[:ex | 
					Transcript show: ex messageText.
					ex return: nil].
		aDatabaseAccessor execute: [
			aDatabaseAccessor executeSQLStringNoResult:
					'ALTER TABLE ' , self name , ' DROP '
						, self platform nameForPrimaryKeyConstraintsWhenDropping , ' '
						, self primaryKeyConstraintName]
			ifError: [:ex | 
				Transcript show: ex messageText.
				ex return: nil]].! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
fieldNamed: aString

	^self fieldNamed: aString ifAbsent: [self error: 'No field ', aString, ' in ', self name].! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
fieldNamed: aString ifAbsent: aBlock

	^self fieldNamed: aString ifAbsent: aBlock caseSensitive: false.! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
fieldNamed: aString ifAbsent: aBlock caseSensitive: aBoolean
	| fieldName |
	fieldName := aBoolean ifTrue: [aString asString] ifFalse: [aString asUppercase].
	^fields
		detect:
			[:each | (aBoolean ifTrue: [each name] ifFalse: [each name asUppercase]) = fieldName]
		ifNone: aBlock.! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
fields

	^fields! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
foreignKeyConstraints
	"Private - Answer the value of the receiver's ''foreignKeyConstraints'' instance variable."

	^foreignKeyConstraints! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
foreignKeyConstraintsForCreation
	"Return the foreign key constraints that should actually be created in the database. Omit those that refer to imaginary tables"
	^foreignKeyConstraints select: [:each |
		each isImaginary not and: [each shouldCreateInDatabase]].! !

!DatabaseTable methodsFor: 'constraints' stamp: ''!
foreignKeyConstraintsForField: aField
	^self foreignKeyConstraints select: [:each | each sourceFields includes: aField].! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasCompositePrimaryKey

	^primaryKeyFields size > 1.! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasConstraints
	^self hasForeignKeyConstraints or: [self hasPrimaryKeyConstraints or: [self hasUniqueConstraints]]! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasConstraintsTo: aTable
	self foreignKeyConstraints
		do: [:each | (each refersToTable: aTable) ifTrue: [^true]].
	^false.! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasConstraintsToAnyOf: aCollectionOfTables
	aCollectionOfTables do: [:each |
		(self hasConstraintsTo: each) ifTrue: [^true]].
	^false.! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasFieldNamed: aString

	^(self fieldNamed: aString ifAbsent: [nil] caseSensitive: false) notNil.! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasForeignKeyConstraintMatching: aConstraint
	^self foreignKeyConstraints
		anySatisfy: [:each | (each matchesForeignKeyConstraint: aConstraint)].! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasForeignKeyConstraintTheSameAs: aConstraint

	^self foreignKeyConstraints anySatisfy: [:each | (each isTheSameAs: aConstraint)].! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasForeignKeyConstraints
	^foreignKeyConstraints isEmpty not! !

!DatabaseTable methodsFor: 'constraints' stamp: ''!
hasForeignKeyConstraintsForField: aField
	^self foreignKeyConstraints anySatisfy: [:each | each sourceFields includes: aField].! !

!DatabaseTable methodsFor: 'constraints' stamp: ''!
hasForeignKeyConstraintsForTargetField: aField
	^self foreignKeyConstraints anySatisfy: [:each | each targetFields includes: aField].! !

!DatabaseTable methodsFor: 'indexes' stamp: ''!
hasIndexForField: aField
	^(indexes detect: [:each | each indexesField: aField] ifNone: [nil]) notNil.! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasIndexNamed: aString

	^indexes contains: [:each | each name = aString]! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasPrimaryKeyConstraints
	^self primaryKeyFields isEmpty not.! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasUniqueConstraintMatching: aConstraint
	
	^self uniqueConstraints anySatisfy: [:each | each matchesUniqueConstraint: aConstraint]! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
hasUniqueConstraints
	^uniqueConstraints isEmpty not! !

!DatabaseTable methodsFor: 'printing' stamp: ''!
imaginaryReplacementName
	| stream |
	stream := WriteStream on: (String new: 100).
	stream nextPut: $(.
	stream nextPutAll: (replacementSubSelect sqlWith: #()) sqlString.
	stream nextPutAll: ') '.
	^stream contents.! !

!DatabaseTable methodsFor: 'indexes' stamp: ''!
indexes
	^indexes.! !

!DatabaseTable methodsFor: 'initialize' stamp: ''!
initialize

	schema := ''.
	fields := OrderedCollection new.
	foreignKeyConstraints := OrderedCollection new: 4.
	uniqueConstraints := OrderedCollection new.
	lockKeyFields := OrderedCollection new.
	indexes := OrderedCollection new.
	isImaginary := false.! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
isAliased
	^parent notNil.! !

!DatabaseTable methodsFor: 'testing' stamp: ''!
isEmpty
	^fields isEmpty.! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
isImaginary
	^isImaginary.! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
isImaginary: aBoolean
	isImaginary := aBoolean.! !

!DatabaseTable methodsFor: 'comparing' stamp: ''!
isTheSameAs: aDatabaseTable
	"Validate that two tables are exactly the same. Most useful for testing purposes, rather than normal usage, which is why we don't just use the #= method."
	self name asUppercase = aDatabaseTable name asUppercase ifFalse: [^false].
	"For the test purposes, we won't worry about schema right now."
"	self schema = aDatabaseTable schema ifFalse: [^false]."
	self fields size = aDatabaseTable fields size ifFalse: [^false].
	self fields asSortedCollection with: aDatabaseTable fields asSortedCollection do: [:a :b | (a isTheSameAs: b) ifFalse: [^false]].
	self foreignKeyConstraints size = aDatabaseTable foreignKeyConstraints size ifFalse: [^false].
	self foreignKeyConstraints with: aDatabaseTable foreignKeyConstraints do: [:a :b | (a isTheSameAs: b) ifFalse: [^false]].	
	self indexes with: aDatabaseTable indexes do: [:a :b | (a isTheSameAs: b) ifFalse: [^false]].
	self lockKeyFields with: aDatabaseTable lockKeyFields do: [:a :b | (a isTheSameAs: b) ifFalse: [^false]].	
	self primaryKeyFields size = aDatabaseTable primaryKeyFields size ifFalse: [^false].
	self primaryKeyFields with: aDatabaseTable primaryKeyFields do: [:a :b | (a isTheSameAs: b) ifFalse: [^false]].
	self isImaginary = aDatabaseTable isImaginary ifFalse: [^false].
	^true.
	"Note, we ignore parent and replacementSubselect"! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
lockKeyFields

	^lockKeyFields! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
name

	^name! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
name: anObject

	name := anObject! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
newFieldNamed: aString

	^self error: 'use #createFieldNamed:type:'! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
nonPrimaryKeyFields

	^self fields reject: [:each | each isPrimaryKey].! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
parent

	^parent.! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
parent: aDatabaseTable

	parent := aDatabaseTable.! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
platform
	fields do: [:each | ^each type platform].
	^nil.! !

!DatabaseTable methodsFor: 'initialize' stamp: ''!
postInitializeIn: aDescriptorSystem
	"Any initialization that happens after all the fields have been added"
	1 to: fields size do: [:i |
		| each | 
		"If fields were read from the database, set their position now"
		each := fields at: i.
		each table: self.
		each position: i.
		each postInitializeIn: aDescriptorSystem].! !

!DatabaseTable methodsFor: 'obsolete' stamp: ''!
primaryKeyConstraintName

	^self name, '_PK'.! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
primaryKeyConstraints
	"Answer the value of the receiver's ''PrimaryKeyConstraints'' instance variable."
	^self hasPrimaryKeyConstraints
		ifTrue: [ PrimaryKeyConstraint forTable: self ]
		ifFalse: [ nil ]! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
primaryKeyFields

	primaryKeyFields isNil ifTrue: [primaryKeyFields := fields select: [:each | each isPrimaryKey yourSelf "might be a proxy"]].
	^primaryKeyFields.! !

!DatabaseTable methodsFor: 'obsolete' stamp: ''!
primaryKeyUniqueConstraintName
	^self platform usesPrimaryKeyUniqueConstraintNames ifTrue: [self name, '_UNIQ'] ifFalse: [''].! !

!DatabaseTable methodsFor: 'obsolete' stamp: ''!
printAddPrimaryKeyConstraintsOn: aStream for: aDatabaseAccessor 
	self primaryKeyFields isEmpty ifTrue: [^self].
	aStream nextPutAll: 'ALTER TABLE ' , self name.
	aStream nextPutAll: ' ADD CONSTRAINT '.
	aStream nextPutAll: self primaryKeyConstraintName.
	aStream nextPutAll: ' PRIMARY KEY  ('.
	GlorpHelper 
		print: [:each | each name]
		on: aStream
		for: self primaryKeyFields
		separatedBy: ','.
	aStream nextPut: $).

	aStream
		nextPutAll: ',';
		cr.

	aStream nextPutAll: ' ADD CONSTRAINT '.
	aStream nextPutAll: self primaryKeyUniqueConstraintName.
	aStream nextPutAll: ' UNIQUE  ('.
	GlorpHelper 
		print: [:each | each name]
		on: aStream
		for: self primaryKeyFields
		separatedBy: ','.
	aStream nextPut: $)! !

!DatabaseTable methodsFor: 'obsolete' stamp: ''!
printDelimiterOn: aStream
	
	aStream
		nextPut: $,;
		cr! !

!DatabaseTable methodsFor: 'printing' stamp: ''!
printOn: aStream

	super printOn: aStream.
	aStream 
		nextPutAll: '(';
		nextPutAll: ((schema isNil or: [schema isEmpty]) ifTrue: [''] ifFalse: [schema, '.']);
		nextPutAll: (name isNil ifTrue: [''] ifFalse: [name]);
		nextPutAll: ')'.! !

!DatabaseTable methodsFor: 'obsolete' stamp: ''!
printPrimaryKeyConstraintsOn: aStream for: aDatabaseAccessor 
	self primaryKeyFields isEmpty ifTrue: [^self].
	aStream nextPutAll: 'CONSTRAINT '.
	aStream nextPutAll: self primaryKeyConstraintName.
	aStream nextPutAll: ' PRIMARY KEY  ('.
	GlorpHelper 
		print: [:each | each name]
		on: aStream
		for: self primaryKeyFields
		separatedBy: ','.
	aStream nextPut: $).

	aStream
		nextPutAll: ',';
		cr.

	aStream nextPutAll: 'CONSTRAINT '.
	aStream nextPutAll: self primaryKeyUniqueConstraintName.
	aStream nextPutAll: ' UNIQUE  ('.
	GlorpHelper 
		print: [:each | each name]
		on: aStream
		for: self primaryKeyFields
		separatedBy: ','.
	aStream nextPut: $)! !

!DatabaseTable methodsFor: 'printing' stamp: ''!
printSQLOn: aWriteStream withParameters: aDictionary
	aWriteStream nextPutAll: self qualifiedName! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
qualifiedName

	^(schema isNil or: [schema isEmpty]) ifTrue: [name] ifFalse: [schema, '.', name].! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
removeFieldNamed: aString
	^self removeFieldNamed: aString ifAbsent: [self error: 'No such field'].! !

!DatabaseTable methodsFor: 'fields' stamp: ''!
removeFieldNamed: aString ifAbsent: aBlock
	
	fields remove: (self fieldNamed: aString ifAbsent: aBlock).
	1 to: fields size do: [:i | (fields at: i) position: i].! !

!DatabaseTable methodsFor: 'As yet unclassified' stamp: ''!
replacementSubSelect: aQuery
	replacementSubSelect := aQuery.! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
schema
	^schema! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
schema: aString

	schema := aString.
	parent isNil ifTrue: [
		self sequences do: [:each | each schema: aString]].! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
sequences
	"Return a collection (which is likely either empty or of size 1) of all the sequences we use."
	| sequences |
	sequences := OrderedCollection new: 2.
	self primaryKeyFields do: [:each | each type hasSequence ifTrue: [sequences add: each type sequence]].
	^sequences.! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
shouldBeWritten
	^isImaginary not.! !

!DatabaseTable methodsFor: 'printing' stamp: ''!
sqlString

	^name.! !

!DatabaseTable methodsFor: 'printing' stamp: ''!
sqlTableName
	"Our name, as appropriate for the list of tables in a SQL statement. Take into account aliasing"

	^parent isNil
		ifTrue:
			[self isImaginary
				ifTrue: [self imaginaryReplacementName]
				ifFalse: [self qualifiedName]]
		ifFalse: [parent sqlTableName , ' ' , self qualifiedName].! !

!DatabaseTable methodsFor: 'accessing' stamp: ''!
uniqueConstraints
	"Private - Answer the value of the receiver's ''uniqueConstraints'' instance variable."

	^uniqueConstraints! !

!DatabaseTable methodsFor: 'validation' stamp: ''!
validateIn: aDescriptorSystem
	"Do any tests we can to make sure that this definition makes sense in this context"! !

Object subclass: #DatabaseTableConstraint
	instanceVariableNames: 'name table shouldCreateInDatabase fields'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!DatabaseTableConstraint commentStamp: '<historical>' prior: 0!
Abstract superclass reifying a constraint on a value in a column or between values in columns.
!

!DatabaseTableConstraint classSide methodsFor: 'instance creation' stamp: ''!
forTable: aDatabaseTable
	^self new
		forTable: aDatabaseTable;
		yourself! !

!DatabaseTableConstraint classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize.! !

!DatabaseTableConstraint methodsFor: 'printing' stamp: ''!
creationString

	^(self creationStringFor: self table platform) first.! !

!DatabaseTableConstraint methodsFor: 'printing' stamp: ''!
creationStringFor: aPlatform

	^self subclassResponsibility! !

!DatabaseTableConstraint methodsFor: 'printing' stamp: ''!
dropString

	^self dropStringFor:  self table platform! !

!DatabaseTableConstraint methodsFor: 'printing' stamp: ''!
dropStringFor: aPlatform

	^aPlatform createConstraintDropStatement: self! !

!DatabaseTableConstraint methodsFor: 'accessing' stamp: ''!
fields
	^fields! !

!DatabaseTableConstraint methodsFor: 'accessing' stamp: ''!
fields: anObject
	fields := anObject! !

!DatabaseTableConstraint methodsFor: 'initializing' stamp: ''!
forTable: aTable

	table := aTable! !

!DatabaseTableConstraint methodsFor: 'accessing' stamp: ''!
generateName

	^self subclassResponsibility! !

!DatabaseTableConstraint methodsFor: 'initializing' stamp: ''!
initialize

	shouldCreateInDatabase := true.! !

!DatabaseTableConstraint methodsFor: 'testing' stamp: ''!
isForeignKeyConstraint

	^false! !

!DatabaseTableConstraint methodsFor: 'testing' stamp: ''!
isPrimaryKeyConstraint

	^false! !

!DatabaseTableConstraint methodsFor: 'testing' stamp: ''!
isUniqueConstraint

	^false! !

!DatabaseTableConstraint methodsFor: 'accessing' stamp: ''!
name
	name isNil ifTrue: [name := self generateName].
	^name.! !

!DatabaseTableConstraint methodsFor: 'accessing' stamp: ''!
name: aString
	name := aString.! !

!DatabaseTableConstraint methodsFor: 'printing' stamp: ''!
printOn: aStream
	super printOn: aStream.
	aStream
		nextPut: $(;
		nextPutAll: self name;
		nextPut: $).! !

!DatabaseTableConstraint methodsFor: 'accessing' stamp: ''!
shouldCreateInDatabase
	^shouldCreateInDatabase ifNil: [true]! !

!DatabaseTableConstraint methodsFor: 'accessing' stamp: ''!
shouldCreateInDatabase: aBoolean
	"If false, then this is a constraint that we use for in-memory ordering, but it should not be created in the database, possibly for performance, or because it might cause a cycle"
	shouldCreateInDatabase := aBoolean! !

!DatabaseTableConstraint methodsFor: 'accessing' stamp: ''!
table
	^table! !

!DatabaseTableConstraint methodsFor: 'accessing' stamp: ''!
table: anObject
	table := anObject! !

DatabaseTableConstraint subclass: #ForeignKeyConstraint
	instanceVariableNames: 'suffixExpression sourceFields targetFields'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!ForeignKeyConstraint commentStamp: '<historical>' prior: 0!
This represents a foreign key constraint in the database. It's used during table creation, but more important it's used when calculating the order in which we need to write rows. Also, if joins aren't specified on a relationship, we will try to infer them from the foreign key constraints.

Instance Variables:
	name	<Object>	The name of the constraint.
	shouldCreateInDatabase	<Boolean>	We may have table definitions that we don't want to create in the database, either because we know they're already there, or because they are for "imaginary" tables.
	sourceFields	<(SequenceableCollection of: DatabaseField)>	The fields we constrain from. The order is fairly arbitrary, but at least we use it consistently.
	suffixExpression	<String>	A bit of a hack to allow us to append some extra text to the constraint. In particular, added to suppor the ability to append "ON DELETE CASCADE".
	targetFields	<(SequenceableCollection of: DatabaseField)>	The fields we constraint to.

!

!ForeignKeyConstraint classSide methodsFor: 'instance creation' stamp: ''!
sourceField: aDatabaseField targetField: anotherDatabaseField
	^self
		sourceField: aDatabaseField
		targetField: anotherDatabaseField
		suffixExpression: nil.! !

!ForeignKeyConstraint classSide methodsFor: 'instance creation' stamp: ''!
sourceField: aDatabaseField targetField: anotherDatabaseField suffixExpression: suffixExpression
	^(self new)
		sourceField: aDatabaseField targetField: anotherDatabaseField suffixExpression: suffixExpression;
		yourself.! !

!ForeignKeyConstraint classSide methodsFor: 'instance creation' stamp: ''!
sourceFields: sourceFields targetFields: targetFields
	^self
		sourceFields: sourceFields
		targetFields: targetFields
		suffixExpression: nil.! !

!ForeignKeyConstraint classSide methodsFor: 'instance creation' stamp: ''!
sourceFields: sourceFields targetFields: targetFields suffixExpression: suffixExpression
	^(self new)
		sourceFields: sourceFields targetFields: targetFields suffixExpression: suffixExpression;
		yourself.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
addSourceField: aField
	(self sourceFields allSatisfy: [:each | aField table = each table]) ifFalse: [self error: 'Fields in foreign key constraints must all have the same source table'].
	self sourceFields add: aField.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
addTargetField: aField
	(self targetFields allSatisfy: [:each | aField table = each table]) ifFalse: [self error: 'Fields in foreign key constraints must all have the same target table'].
	self targetFields add: aField.! !

!ForeignKeyConstraint methodsFor: 'converting' stamp: ''!
asJoin
	^Join fromAll: self sourceFields toAll: self targetFields.! !

!ForeignKeyConstraint methodsFor: 'converting' stamp: ''!
asReverseJoin
	^Join fromAll: self targetFields toAll: self sourceFields.! !

!ForeignKeyConstraint methodsFor: 'printing' stamp: ''!
creationStringFor: aPlatform

	^aPlatform createForeignKeyConstraintAddStatement: self! !

!ForeignKeyConstraint methodsFor: 'printing' stamp: ''!
generateName
	| stream maxLength tableLength roomToBeVerbose shortTableLength longTableLength |
	self sourceFields isEmpty ifTrue: [^'INVALID CONSTRAINT'].
	"Figure out if we have room for verbose separators or if we should maximize the space for the table/field names"
	maxLength := self sourceFields first table platform maximumLengthOfTableName.
	shortTableLength := ((maxLength - 8) // 2) min: 10.
	longTableLength := (maxLength // 2) - 1.
	roomToBeVerbose := shortTableLength >= 10.
	tableLength := roomToBeVerbose ifTrue: [shortTableLength] ifFalse: [longTableLength].

	stream := WriteStream on: (String new: 100).
	self printForConstraintName: self sourceFields first on: stream maxLength: tableLength.
	roomToBeVerbose ifTrue: [stream nextPutAll: '_TO_'].
	self printForConstraintName: self targetFields first on: stream maxLength: tableLength.
	stream nextPutAll: (roomToBeVerbose ifTrue: ['_REF'] ifFalse: ['R']).
	^stream contents.! !

!ForeignKeyConstraint methodsFor: 'initializing' stamp: ''!
initialize

	self sourceFields: (OrderedCollection new: 2).
	targetFields := OrderedCollection new: 2.! !

!ForeignKeyConstraint methodsFor: 'testing' stamp: ''!
isForeignKeyConstraint

	^true! !

!ForeignKeyConstraint methodsFor: 'testing' stamp: ''!
isImaginary
	^self targetTable isImaginary or: [self sourceTable isImaginary].! !

!ForeignKeyConstraint methodsFor: 'comparing' stamp: ''!
isTheSameAs: aForeignKeyConstraint
	"Validate that two constraints are exactly the same. Most useful for testing purposes, rather than normal usage, which is why we don't just use the #= method."
	self name asUppercase = aForeignKeyConstraint name asUppercase ifFalse: [^false].
	self suffixExpression = aForeignKeyConstraint suffixExpression ifFalse: [^false].
	self sourceFields size ~= aForeignKeyConstraint sourceFields size ifTrue: [^false].
	self sourceFields with: aForeignKeyConstraint sourceFields  do: [:a :b | (a isTheSameAs: b) ifFalse: [^false]].	
	self targetFields size ~= aForeignKeyConstraint targetFields size ifTrue: [^false].
	self targetFields with: aForeignKeyConstraint targetFields  do: [:a :b | (a isTheSameAs: b) ifFalse: [^false]].	
	self shouldCreateInDatabase = aForeignKeyConstraint shouldCreateInDatabase ifFalse: [^false].
	^true.! !

!ForeignKeyConstraint methodsFor: 'testing' stamp: ''!
matchesForeignKeyConstraint: aConstraint
	name asUppercase = aConstraint name asUppercase ifFalse: [^false].
	self sourceFields size ~= aConstraint sourceFields size ifTrue: [^false].
	self sourceFields with: aConstraint sourceFields do: [:field1 :field2 | 
		(field1 matchesField: field2) ifFalse: [^false]].

	self targetFields size ~= aConstraint targetFields size ifTrue: [^false].
	self targetFields with: aConstraint targetFields do: [:field1 :field2 | 
		(field1 matchesField: field2) ifFalse: [^false]].
	^true! !

!ForeignKeyConstraint methodsFor: 'printing' stamp: ''!
printForConstraintName: aField on: aStream maxLength: maxLength 
	| constraintName |
	constraintName := aField table name , '_' , aField name.
	constraintName size > maxLength 
		ifTrue: [constraintName := constraintName copyFrom: 1 to: maxLength].
	aStream nextPutAll: constraintName! !

!ForeignKeyConstraint methodsFor: 'testing' stamp: ''!
refersToTable: aTable
	^self targetTable yourself == aTable.! !

!ForeignKeyConstraint methodsFor: 'relationships' stamp: ''!
rowsRelatedTo: aRow
	| myWrapper relatedRows sourceField targetField allWrappers |
	"Note that we don't care which field, as presumably all of them must point the same place for a composite key, so grab the first pair"
	sourceField := self sourceFields first.
	targetField := self targetFields first.
	myWrapper := aRow wrapperAt: sourceField ifAbsent: [nil].
	myWrapper isNil ifTrue: [^#()].
	relatedRows := Set new: 5.
	allWrappers := OrderedCollection with: myWrapper.
	allWrappers addAll: myWrapper allInnerWrappers.
	allWrappers
		do: [:eachWrapper |
			(eachWrapper containedBy glorpAtIdentity: targetField ifAbsent: [#()])
				do: [:eachRow | eachRow == aRow ifFalse: [relatedRows add: eachRow]]].
	^relatedRows.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
someSourceField
	"Return any one of our source fields"
	^self sourceFields first.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
sourceField: aField
	self sourceFields isEmpty ifFalse: [self error: 'Adding a field to a non-empty constraint'].
	self sourceFields add: aField.! !

!ForeignKeyConstraint methodsFor: 'initializing' stamp: ''!
sourceField: aDatabaseField targetField: anotherDatabaseField
	self
		sourceField: aDatabaseField
		targetField: anotherDatabaseField
		suffixExpression: nil.! !

!ForeignKeyConstraint methodsFor: 'initializing' stamp: ''!
sourceField: aDatabaseField targetField: anotherDatabaseField suffixExpression: suffixExpressionString
	self
		sourceField: aDatabaseField;
		targetField: anotherDatabaseField;
		suffixExpression: suffixExpressionString.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
sourceFields
	^sourceFields! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
sourceFields: aCollection
	
	sourceFields := aCollection.! !

!ForeignKeyConstraint methodsFor: 'initializing' stamp: ''!
sourceFields: sourceFieldCollection targetFields: targetFieldCollection suffixExpression: suffixExpressionString
	self
		sourceFields: sourceFieldCollection;
		targetFields: targetFieldCollection;
		suffixExpression: suffixExpressionString.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
sourceTable
	"All the source fields have to come from the same table, so just pick one"
	^self sourceFields first table.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
suffixExpression
	^suffixExpression! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
suffixExpression: anObject
	suffixExpression := anObject.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
table
	^self sourceTable.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
targetField: aField

	self targetFields isEmpty ifFalse: [self error: 'Adding a field to a non-empty constraint'].
	self targetFields add: aField.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
targetFields
	^targetFields! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
targetFields: aCollection
	targetFields := aCollection.! !

!ForeignKeyConstraint methodsFor: 'accessing' stamp: ''!
targetTable
	"All the target fields have to come from the same table, so just pick one"
	^self targetFields isEmpty ifTrue: [nil] ifFalse: [self targetFields first table].! !

DatabaseTableConstraint subclass: #PrimaryKeyConstraint
	instanceVariableNames: 'tableName schema type'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!PrimaryKeyConstraint commentStamp: '<historical>' prior: 0!
This is a class representing non-foreign key database constraints. It's not actually used for very much, but is mostly so we can model the metadata in the database's information_schema to get information we need about primary keys and other information that may be represented as constraints rather than field attributes.!

!PrimaryKeyConstraint classSide methodsFor: 'instance creation' stamp: ''!
forTable: aDatabaseTable fields: collDatabaseFields
	^self new
		forTable: aDatabaseTable  fields: collDatabaseFields;
		yourself! !

!PrimaryKeyConstraint methodsFor: 'printing' stamp: ''!
creationStringFor: aPlatform

	^aPlatform createPrimaryKeyConstraintAddStatement: self! !

!PrimaryKeyConstraint methodsFor: 'initializing' stamp: ''!
forTable: aTable

	super forTable: aTable.
	name := self primaryKeyConstraintName.
	self primaryKeyFields:  table primaryKeyFields! !

!PrimaryKeyConstraint methodsFor: 'initializing' stamp: ''!
forTable: aDatabaseTable fields: collDatabaseFields

	self forTable: aDatabaseTable.
	self primaryKeyFields: collDatabaseFields! !

!PrimaryKeyConstraint methodsFor: 'printing' stamp: ''!
generateName

	^self table name, '_PK'.! !

!PrimaryKeyConstraint methodsFor: 'testing' stamp: ''!
isPrimaryKeyConstraint

	^true! !

!PrimaryKeyConstraint methodsFor: 'comparing' stamp: ''!
isTheSameAs: aPrimaryKeyConstraint
	"Validate that two constraints are exactly the same. Most useful for testing purposes, rather than normal usage, which is why we don't just use the #= method."
	self name asUppercase = aPrimaryKeyConstraint name asUppercase ifFalse: [^false].
	self primaryKeyFields size = aPrimaryKeyConstraint primaryKeyFields size ifFalse: [^false].
	self primaryKeyFields with: aPrimaryKeyConstraint primaryKeyFields  do: [:a :b | (a isTheSameAs: b) ifFalse: [^false]].	
	self shouldCreateInDatabase = aPrimaryKeyConstraint shouldCreateInDatabase ifFalse: [^false].
	^true.! !

!PrimaryKeyConstraint methodsFor: 'testing' stamp: ''!
matchesPrimaryKeyConstraint: aConstraint
	name = aConstraint name ifFalse: [^false].
	self primaryKeyFields size ~= aConstraint primaryKeyFields size ifTrue: [^false].
	self primaryKeyFields with: aConstraint primaryKeyFields do: [ :field1 :field2 | 
		(field1 matchesField: field2) ifFalse: [^false]].
	^true! !

!PrimaryKeyConstraint methodsFor: 'accessing' stamp: ''!
primaryKeyConstraintName

	^self table name, '_PK'.! !

!PrimaryKeyConstraint methodsFor: 'accessing' stamp: ''!
primaryKeyFields

	^fields! !

!PrimaryKeyConstraint methodsFor: 'accessing' stamp: ''!
primaryKeyFields: aCollection

	fields := aCollection! !

DatabaseTableConstraint subclass: #UniqueConstraint
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!UniqueConstraint commentStamp: '<historical>' prior: 0!
UniqueConstraint implementation has not been finished

Instance Variables:
	isForPrimaryKey	<Boolean>	description of isForPrimaryKey

!

!UniqueConstraint classSide methodsFor: 'instance creation' stamp: ''!
forTable: aTable fields: aCollectionOfFields

	| instance |
	instance := self new.
	instance forTable: aTable fields: aCollectionOfFields.
	^instance! !

!UniqueConstraint methodsFor: 'printing' stamp: ''!
creationStringFor: aPlatform

	^aPlatform createUniqueConstraintAddStatement: self! !

!UniqueConstraint methodsFor: 'instance creation' stamp: ''!
forTable: aTable fields: aCollectionOfFields

	self table: aTable.
	self fields: aCollectionOfFields! !

!UniqueConstraint methodsFor: 'accessing' stamp: ''!
generateName
	| stream maxLength tableLength roomToBeVerbose shortTableLength longTableLength |
	(self fields isNil or: [self fields isEmpty]) ifTrue: [^'INVALID CONSTRAINT'].
	"Figure out if we have room for verbose separators or if we should maximize the space for the table/field names"
	maxLength := self fields first table platform maximumLengthOfTableName.
	shortTableLength := ((maxLength - 7) // 2) min: 10.
	longTableLength := (maxLength - 7).
	roomToBeVerbose := shortTableLength >= 10.
	tableLength := roomToBeVerbose ifTrue: [longTableLength] ifFalse: [shortTableLength].

	stream := WriteStream on: (String new: 100).
	self printForConstraintNames: self fields on: stream maxLength: tableLength.
	stream nextPutAll: (roomToBeVerbose ifTrue: ['_UNIQ'] ifFalse: ['U']).
	^stream contents.! !

!UniqueConstraint methodsFor: 'testing' stamp: ''!
isUniqueConstraint

	^true! !

!UniqueConstraint methodsFor: 'testing' stamp: ''!
matchesUniqueConstraint: aConstraint
	
	name asUppercase = aConstraint name asUppercase ifFalse: [^false].
	self fields size ~= aConstraint fields size ifTrue: [^false].
	self fields with: aConstraint fields do: [:each :other | (each matchesField: other) ifFalse: [^false]].
	^true! !

!UniqueConstraint methodsFor: 'printing' stamp: ''!
printForConstraintNames: aCollection on: aStream maxLength: maxLength 
	| constraintName |
	constraintName := self table name.
	aCollection do:
		[:each | constraintName := constraintName, each name].
	constraintName size > maxLength 
		ifTrue: [constraintName := constraintName copyFrom: 1 to: maxLength].
	aStream nextPutAll: constraintName! !

Object subclass: #Descriptor
	instanceVariableNames: 'tables mappings system mappedFields cachePolicy typeResolver mapsPrimaryKeys keyFields classModel multipleTableJoin imaginaryTableReplacements mappingsForFields'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!Descriptor commentStamp: '<historical>' prior: 0!
For each persistent class, there is a descriptor. This is what tells GLORP how to read and write objects of this class. The descriptor is not actually a very complicated object, because most of the information is in the mappings described below. But the descriptor does keep hold of a few per-class pieces of information, notably

Instance Variables:
	cachePolicy	<CachePolicy>	The cache policy to use. If left blank, default to what the system specifies.
	describedClass	<Class>	The persistent class that we map.
	mappedFields	<Collection of: DatabaseField>	The database fields we have mappings for (cached computed value)
	mappings	<(Collection of: Mapping)>	Our mappings.
	mapsPrimaryKeys	<Object>	Do we map the primary keys (cached computed value)
	multipleTableCriteria	<(Collection of: Join)>	description of multipleTableCriteria
	system	<DescriptorSystem>	The descriptor system that we belong to.
	tables	<(SequenceableCollection of: (DatabaseTable))>	The tables that we map
	typeResolver	<TypeResolver>	description of typeResolver

!

!Descriptor classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
addMapping: aMapping 
	self basicAddMapping: aMapping.
	aMapping hasAttribute ifTrue: [
		(aMapping classModel attributeNamed: aMapping attribute name) == aMapping attribute ifFalse: [self error: 'If you have a classModel you must use newMapping: rather than addMapping:']].! !

!Descriptor methodsFor: 'accessing' stamp: ''!
addMultipleTableJoin: anExpression 
	self multipleTableJoin add: anExpression! !

!Descriptor methodsFor: 'accessing' stamp: ''!
addTable: aDatabaseTable 
	tables add: aDatabaseTable.! !

!Descriptor methodsFor: 'type mapping' stamp: ''!
allDescribedConcreteClasses
	^ self typeMapping allDescribedConcreteClasses! !

!Descriptor methodsFor: 'accessing' stamp: ''!
allMappingsForField: aField 
	"Return all of the mappings that use this field"
	^mappings select: [:each | each mappedFields includes: aField]! !

!Descriptor methodsFor: 'accessing' stamp: ''!
allTables
	"Return all tables used by this descriptor or its subclasses. For a non-inheritance descriptor, this is the same as tables."
	^self typeResolver allTablesFor: self.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
basicAddMapping: aMapping 
	mappings := mappings, (Array with: aMapping).
	aMapping descriptor: self.
	self resetMappedFields.! !

!Descriptor methodsFor: 'internal' stamp: ''!
buildReplacementTableSelect
	"We have an imaginary table which is defined by selecting from our attributes instead of from the table. Build the query for that."

	| sourceMappings targetMappings mainQuery subQuery |
	mainQuery := nil.
	sourceMappings := imaginaryTableReplacements
		collect: [:each | self mappingForAttributeNamed: each first].
	targetMappings := imaginaryTableReplacements
		collect: [:each | self mappingForAttributeNamed: each last].
	sourceMappings
		with: targetMappings
		do:
			[:eachSourceMapping :eachTargetMapping | 
			subQuery := self
				buildReplacementTableSelectFrom: eachSourceMapping
				to: eachTargetMapping.
			mainQuery := mainQuery isNil
				ifTrue: [subQuery]
				ifFalse: [mainQuery unionAll: subQuery]].
	mainQuery session: self session.
	self primaryTable replacementSubSelect: mainQuery.! !

!Descriptor methodsFor: 'internal' stamp: ''!
buildReplacementTableSelectFrom: sourceMapping to: targetMapping
	"Build a select statement that combines the two mappings. We expect that both are relationships from a single object whose table is imaginary, so we need to knit together the information from the mappings to join the real tables directly together."

	| fullJoin targetJoin sourceJoin subQuery join |
	subQuery := Query returningManyOf: sourceMapping referenceClass.
	subQuery session: self session.	
	subQuery setUpExpressions.
	"Invert because these will both start with the imaginary one on the left and we want to merge them together"
	sourceJoin := (sourceMapping join
		replacing: self primaryTable
		withEquivalenceFrom: targetMapping join) inverseJoin.
	targetJoin := targetMapping join
		replacing: self primaryTable
		withEquivalenceFrom: sourceMapping join.
	fullJoin := sourceJoin = targetJoin
		ifTrue: [sourceJoin]
		ifFalse: [sourceJoin , targetJoin].
	join := fullJoin asGeneralGlorpExpression asExpressionJoiningSource: subQuery baseExpression toTarget: subQuery baseExpression.
	subQuery addJoin: join.
	sourceMapping usesLinkTable ifTrue: [subQuery addJoin: (sourceMapping  expressionFromLinkToReferenceTableWithBase: subQuery baseExpression)].
	targetMapping usesLinkTable ifTrue: [subQuery addJoin: (targetMapping  expressionFromLinkToReferenceTableWithBase: subQuery baseExpression)].
	self
		setRetrievalExpressionsInReplacementTableForQuery: subQuery
		fromMapping: sourceMapping
		toMapping: targetMapping.
	^subQuery.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
cachePolicy
	cachePolicy isNil ifTrue: [^system cachePolicy].
	^cachePolicy! !

!Descriptor methodsFor: 'accessing' stamp: ''!
cachePolicy: aCachePolicy
	cachePolicy:= aCachePolicy! !

!Descriptor methodsFor: 'accessing' stamp: ''!
classModel
	^classModel! !

!Descriptor methodsFor: 'accessing' stamp: ''!
classModel: aGlorpClassModel
	classModel := aGlorpClassModel! !

!Descriptor methodsFor: 'type resolution' stamp: ''!
classesRequiringIndependentQueries

	^ self typeResolver classesRequiringIndependentQueriesFor: self describedClass! !

!Descriptor methodsFor: 'testing' stamp: ''!
computeMapsPrimaryKeys
	| primaryKeyFields |
	self tables isEmpty ifTrue: [^false].
	primaryKeyFields := self primaryTable primaryKeyFields.
	primaryKeyFields isEmpty ifTrue: [^false].
	primaryKeyFields
		do: [:each | (self mappedFields anySatisfy: [ :eachMappedField |
			eachMappedField = each]) ifFalse: [^false]].
	^true.! !

!Descriptor methodsFor: 'deleting' stamp: ''!
createDeleteRowFor: anObject table: aTable in: aRowMap
	"Create records for rows that require deletion"

	(self primaryKeyMappingsFor: aTable)
		do: [:each | 
			each first mapFromObject: anObject intoRowsIn: aRowMap].! !

!Descriptor methodsFor: 'deleting' stamp: ''!
createDeleteRowsFor: anObject in: aRowMap
	"Create records for rows that require deletion"

	|  |
	anObject class == self describedClass ifFalse: [self error: 'wrong descriptor for this object'].
	"If we don't map primary keys, then we're something that gets embedded, and a delete isn't necessary"
	self mapsPrimaryKeys ifFalse: [^self]. 

	self tables do: [:eachTable |
		self createDeleteRowFor: anObject table: eachTable in: aRowMap.
		aRowMap rowForTable: eachTable withKey: anObject ifPresent: [:eachRow | eachRow forDeletion: true]].
	"It's possible that we might not have any direct mapping for a secondary table's primary keys, so 
allow the multiple table criteria to specify them if that's the only one. If they're not, then they don't do any harm"
	self multipleTableJoin do: [:each |
		each mapFromSource: anObject andTarget: anObject intoRowsIn: aRowMap.
		each allTables do: [:eachTable |
			aRowMap rowForTable: eachTable withKey: anObject ifPresent: [:eachRow | eachRow forDeletion: true]]].

	self mappings do: [:each | each  createDeleteRowFor: anObject in: aRowMap].! !

!Descriptor methodsFor: 'mapping' stamp: ''!
createRowsFor: anObject in: aRowMap
	| row |
	anObject class == self describedClass
		ifFalse: [self error: 'wrong descriptor for this object'].
	mappings do: [:each | 
			each stopIfDebuggingWrite.
			each mapFromObject: anObject intoRowsIn: aRowMap].

	"If we haven't written out our primary keys, then make sure we do so, but if absolutely none of our mappings have written anything, then just ignore it."
	row := aRowMap rowForTable: self primaryTable withKey: anObject ifAbsent: [^self].
	self primaryTable primaryKeyFields do: [:each | 
		self setFieldValue: each 	in: row from: anObject].

	"This is a bit of a hack/heuristic. If the multiple table join is an outer join, and the row doesn't exist yet, then assume we won't need it and don't do the join. It means none of our mappings wrote to it. Hope that in that case no one else's will. It seems reasonable that we wouldn't map an object to multiple tables unless we planned to map some field within it, but it's still a bit of a risk."
	self multipleTableJoin do: [:each | 
		(each isOuterJoin not or: [each allTables anySatisfy: [:eachTable | 
			aRowMap includesRowForTable: eachTable withKey: anObject]]) 
				ifTrue: [
					each
						mapFromSource: anObject
						andTarget: anObject
						intoRowsIn: aRowMap]].! !

!Descriptor methodsFor: 'tracing' stamp: ''!
defaultTracing
	^Tracing new.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
describedClass
	^classModel describedClass.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
describedClass: aClass
	classModel := self system classModelFor: aClass.! !

!Descriptor methodsFor: 'internal' stamp: ''!
describedConcreteClassFor: row withBuilder: builder
	"Lookup the class that is represented by the row when there is a possibility
		of this row representing any class within a hierarchy. "
	^self typeResolver describedConcreteClassFor: row withBuilder: builder descriptor: self! !

!Descriptor methodsFor: 'creating mappings' stamp: ''!
directMapping
	^self newMapping: DirectMapping.! !

!Descriptor methodsFor: 'creating mappings' stamp: ''!
directMappingFor: attributeName
	"Create a direct mapping from the attribute of this name to the field in the table with a matching name. For convenience when there's a good correspondence between the field and instance variable names."
	| field |
	self tables do: [:each |
		field := each fieldNamed: attributeName ifAbsent: [nil] caseSensitive: false].
	field isNil ifTrue: [self error: 'Cannot find field named ', attributeName printString].
	^self directMapping from: attributeName to: field.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
directMappingForField: aField 
	"Return a single, direct mapping for this field. There may conceivably be more than one, but they all have to agree, so it shouldn't matter as far as the value. There may also be none."

	^(self mappingsForFields at: aField ifAbsent: [^nil]) 
		detect: [:each | each isRelationship not]
		ifNone: [nil]! !

!Descriptor methodsFor: 'accessing' stamp: ''!
directMappingsForField: aField 
	"Return all the relationship mappings that use this field"

	^(self mappingsForFields at: aField ifAbsent: [^#()]) 
		reject: [:each | each isRelationship].! !

!Descriptor methodsFor: 'creating mappings' stamp: ''!
fieldFor: aFieldOrString
	"We allow either a field or a string naming the field, and return the appropriate field object."
	aFieldOrString isString ifFalse: [^aFieldOrString].
	self tables do: [:each |
		| field | 
		field := each fieldNamed: aFieldOrString.
		field isNil ifFalse: [^field]].
	self error: 'Cannot find field ', aFieldOrString, '. Please pass in the field object'.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
fieldsForSelectStatement
	| myFields inheritedFields |
	myFields := self mappedFields.
	inheritedFields := self typeResolver fieldsForSelectStatement.
	^inheritedFields isEmpty ifTrue: [myFields] ifFalse: [myFields, inheritedFields].! !

!Descriptor methodsFor: 'accessing' stamp: ''!
imaginaryTableIsDefinedByRelationFrom: sourceAttributeName to: targetAttributeName
	self imaginaryTableReplacements add: (Array with: sourceAttributeName with: targetAttributeName).! !

!Descriptor methodsFor: 'As yet unclassified' stamp: ''!
imaginaryTableReplacements
	imaginaryTableReplacements isNil ifTrue: [imaginaryTableReplacements := OrderedCollection new: 4].
	^imaginaryTableReplacements.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
initialize

	mappings := Array new.
	tables := OrderedCollection new: 1.! !

!Descriptor methodsFor: 'testing' stamp: ''!
isForAssociation
	^self describedClass == Association.! !

!Descriptor methodsFor: 'testing' stamp: ''!
isTypeMappingRoot
	^self typeResolver isTypeMappingRoot: self! !

!Descriptor methodsFor: 'accessing' stamp: ''!
keyFields
	"Return the fields we use as primary keys for caching purposes. Typically, these are just going to be the primary keys of our primary table. Objects intended to be embedded won't map the primary keys at all. Weird objects may supply their own keys (e.g. functions rather than fields directly)"

	^keyFields == nil
		ifTrue:
			[self mapsPrimaryKeys
				ifTrue: [self primaryTable primaryKeyFields]
				ifFalse: [nil]]
		ifFalse: [keyFields].! !

!Descriptor methodsFor: 'accessing' stamp: ''!
keyFields: aFieldCollection
	"Set the fields we use as primary keys for caching purposes. Typically, these are just going to be the primary keys of our primary table"
	keyFields := aFieldCollection.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
mappedFields
	"Return all the fields that are mapped, in the order that they occur in the table. This specifically omits fields that are not in the selected set of tables (so e.g. even if a mapping uses a link table field, if it's not in the list of tables to select from, we don't want it). We also include mapped, non-constant expressions, which we would normally expect to be functions"
	mappedFields isNil ifTrue: [
		| fieldSet |
		fieldSet := IdentitySet new: mappings size.
		mappings do: [:each | fieldSet addAll: each fieldsForDescriptorSelectStatement].
		mappedFields := OrderedCollection new.
		tables do: [:each |
			each fields do: [:eachField | (fieldSet includes: eachField) ifTrue: [mappedFields add: eachField]]].
		fieldSet do: [:each | 
 			(each isGlorpExpression and: [each isConstantExpression not]) ifTrue: [mappedFields add: each]]].
	^mappedFields.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
mappingForAttributeNamed: aSymbol

	^mappings detect: [:each | each attributeName == aSymbol] ifNone: [
		self isForAssociation ifTrue: [self valueDescriptor mappingForAttributeNamed: aSymbol]]! !

!Descriptor methodsFor: 'accessing' stamp: ''!
mappingForField: aField

	^(self mappingsForFields at: aField ifAbsent: [^nil]) first.! !

!Descriptor methodsFor: 'mapping' stamp: ''!
mappings
	^ReadStream on: mappings! !

!Descriptor methodsFor: 'accessing' stamp: ''!
mappingsForFields
	"Return a dictionary from fields to the mappings that contain them. This specifically omits fields that are not in the selected set of tables (so e.g. even if a mapping uses a link table field, if it's not in the list of tables to select from, we don't want it). We also include mapped, non-constant expressions, which we would normally expect to be functions"
	mappingsForFields isNil ifTrue: [
		mappingsForFields := IdentityDictionary new: mappings size.
		mappings do: [:eachMapping | 
			eachMapping mappedFields do: [:eachField | 
				(mappingsForFields at: eachField ifAbsentPut: [OrderedCollection new]) add: eachMapping]]].
	^mappingsForFields.! !

!Descriptor methodsFor: 'testing' stamp: ''!
mapsPrimaryKeys
	mapsPrimaryKeys isNil ifTrue: [mapsPrimaryKeys := self computeMapsPrimaryKeys].
	^mapsPrimaryKeys! !

!Descriptor methodsFor: 'accessing' stamp: ''!
multipleTableJoin
	multipleTableJoin isNil
		ifTrue: [multipleTableJoin := OrderedCollection new: 1].
	^multipleTableJoin.! !

!Descriptor methodsFor: 'creating mappings' stamp: ''!
newMapping: aClass
	| mapping |
	mapping := aClass new.
	self basicAddMapping: mapping.
	^mapping.! !

!Descriptor methodsFor: 'creating mappings' stamp: ''!
oneToOneMapping
	^self newMapping: OneToOneMapping.! !

!Descriptor methodsFor: 'mapping' stamp: ''!
populateObject: anObject inBuilder: anElementBuilder
	"Answer an object using the values for the specified fields."

	mappings
		do: [:each | 
			each stopIfDebuggingRead.
			each mapObject: anObject inElementBuilder: anElementBuilder].! !

!Descriptor methodsFor: 'mapping' stamp: ''!
primaryKeyCollectionFor: anObject
	"Find the primary key for anObject. Always return it as a collection."
	| fields |
	anObject yourSelf class == self describedClass
		ifFalse: [self error: 'Wrong descriptor for this object'].
	fields := self primaryTable primaryKeyFields.
	^fields collect: [:each | self valueOfField: each fromObject: anObject].! !

!Descriptor methodsFor: 'mapping' stamp: ''!
primaryKeyExpressionFor: anObject basedOn: anExpression relation: aSymbol
	"anObject can be either a real domain object, or a constant expression containing it, or an expression representing the object"

	| expression value pkMappings |
	value := anObject isGlorpExpression
		ifTrue:
			[anObject isConstantExpression ifTrue: [anObject value] ifFalse: [anObject]]
		ifFalse:
			[anObject class == self describedClass
				ifFalse: [self error: 'Wrong descriptor for this object'].
			anObject].
	self isForAssociation ifTrue: [
		"We just rewrite as self value = x"
		^((anExpression get: #value) get: aSymbol withArguments: (Array with:  value)) rewriteEquality].
	expression := nil.
	pkMappings := self primaryKeyMappings.
	(pkMappings isEmpty ifTrue: [self mappings] ifFalse: [pkMappings])
		do:
			[:each | 
			| clause |
			clause := each
				expressionFor: value
				basedOn: anExpression
				relation: aSymbol.
			expression := aSymbol = #<> ifTrue: [clause OR: expression] ifFalse: [clause AND: expression]].
	^expression.! !

!Descriptor methodsFor: 'mapping' stamp: ''!
primaryKeyFor: anObject
	"Find the primary key for anObject. If it's a single field, just return a single value. Otherwise, return a collection. Caller is expected to know what to expect."
	| fields |
	anObject class == self describedClass
		ifFalse: [self error: 'Wrong descriptor for this object'].
	fields := self primaryTable primaryKeyFields.
	^fields size = 1
		ifTrue: [self valueOfField: (fields at: 1) fromObject: anObject]
		ifFalse:
			[fields collect: [:each | self valueOfField: each fromObject: anObject]].! !

!Descriptor methodsFor: 'mapping' stamp: ''!
primaryKeyMappings
	"Return things that map to our primary key fields. Prefer direct mappings"
	| fields pkMappings |
	fields := self primaryTable primaryKeyFields.
	pkMappings := OrderedCollection new: fields size.
	fields do: [:each |  | mapping |
		mapping := self directMappingForField: each.
		mapping isNil ifTrue: [mapping := self mappingForField: each].
		mapping isNil ifFalse: [pkMappings add: mapping]].
	^pkMappings.! !

!Descriptor methodsFor: 'deleting' stamp: ''!
primaryKeyMappingsFor: aTable
	"Return our primary key mappings as a collection of mappings, one collection per primary key"

	| allMappings |
	allMappings := Dictionary new: aTable primaryKeyFields size.
	aTable primaryKeyFields
		do: [:each | 
			| mappingsForField |
			mappingsForField := self mappingsForFields at: each ifAbsent: [nil].
			mappingsForField isNil ifFalse: [allMappings at: each put: mappingsForField]].
	^allMappings.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
primaryTable
	tables isEmpty ifTrue: [^nil].
	^tables first.! !

!Descriptor methodsFor: 'printing' stamp: ''!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: '('.
	self describedClass printOn: aStream.
	aStream nextPutAll: ')'.! !

!Descriptor methodsFor: 'internal' stamp: ''!
readBackNewRowInformationFor: anObject! !

!Descriptor methodsFor: 'mapping' stamp: ''!
readBackNewRowInformationFor: anObject in: aRowMap 
	anObject yourSelf class == self describedClass ifFalse: [self error: 'wrong descriptor for this object'].
	mappings do: [:each | each readBackNewRowInformationFor: anObject fromRowsIn: aRowMap]! !

!Descriptor methodsFor: 'internal' stamp: ''!
referencedIndependentObjectsFrom: anObject do: aBlock

	mappings do: [:each |
		(each referencedIndependentObjectsFrom: anObject) do: [:eachReferencedObject |
			aBlock value: eachReferencedObject]].! !

!Descriptor methodsFor: 'internal' stamp: ''!
referencedIndependentObjectsWithMappingsFrom: anObject do: aBlock

	mappings do: [:each |
		(each referencedIndependentObjectsFrom: anObject) do: [:eachReferencedObject |
			aBlock value: eachReferencedObject value: each]].! !

!Descriptor methodsFor: 'type mapping' stamp: ''!
registerTypeResolver: aResolver
	self registerTypeResolver: aResolver abstract: false! !

!Descriptor methodsFor: 'type mapping' stamp: ''!
registerTypeResolver: aResolver abstract: shouldBeAbstract
	shouldBeAbstract ifFalse: [ self beAbstract ].
	aResolver register: self.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
relationshipMappingForField: aField 
	"Return a single, direct mapping for this field. There may conceivably be more than one, but they all have to agree, so it shouldn't matter as far as the value. There may also be none."

	^(self mappingsForFields at: aField ifAbsent: [^nil]) 
		detect: [:each | each isRelationship]
		ifNone: [nil]! !

!Descriptor methodsFor: 'accessing' stamp: ''!
relationshipMappingsForField: aField 
	"Return all the relationship mappings that use this field"

	^(self mappingsForFields at: aField ifAbsent: [^#()]) 
		select: [:each | each isRelationship].! !

!Descriptor methodsFor: 'accessing' stamp: ''!
removeMapping:  aMapping

	mappings := mappings copyWithout: aMapping! !

!Descriptor methodsFor: 'accessing' stamp: ''!
resetMappedFields
	mappedFields := nil.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
session

	^system session.! !

!Descriptor methodsFor: 'mapping' stamp: ''!
setFieldValue: each in: row from: anObject
	"Try to set the value of the field, which is normally a primary key field, in the row, based on data in the object. Useful for the case where we have only a relationship mapping to the object that holds our primary key, and it is uninstantiated"
	| value |
	(row wrapperAt: each ifAbsent: [FieldValueWrapper new]) hasValue
		ifFalse:
			[value := self valueOfField: each fromObject: anObject ifNone: [^self].
			"We're using this to get primary key fields. Nil is not a useful answer."
			value isNil ifTrue: [^self].
			row at: each put: value].! !

!Descriptor methodsFor: 'internal' stamp: ''!
setRetrievalExpressionsInReplacementTableForQuery: subQuery fromMapping: sourceMapping toMapping: targetMapping
	self primaryTable fields
		do:
			[:each | 
			| sourceReplacement replacement expression |
			sourceReplacement := sourceMapping join targetForSource: each.
			replacement := sourceReplacement isNil
				ifTrue: [targetMapping join targetForSource: each]
				ifFalse: [sourceReplacement].
			expression := replacement isGlorpExpression
				ifTrue: [replacement] 
				ifFalse: [(BaseExpression new getTable: replacement table) getField: replacement].
			expression alias: each name.
			subQuery retrieve: expression].! !

!Descriptor methodsFor: 'tracing' stamp: ''!
setupTracing: aTracing
	"Find all the other objects that need to be read when this one is read"

	self trace: aTracing context: aTracing base.! !

!Descriptor methodsFor: 'testing' stamp: ''!
supportsOrdering

	typeResolver isNil ifTrue: [^true].
	^typeResolver supportsOrderingFor: self.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
system
	^system! !

!Descriptor methodsFor: 'accessing' stamp: ''!
system: anObject
	system := anObject! !

!Descriptor methodsFor: 'accessing' stamp: ''!
table
	tables isEmpty ifTrue: [^nil].
	^tables first.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
table: aDatabaseTable
	
	| table |
	table := aDatabaseTable isString
		ifTrue: [table := self system tableNamed: aDatabaseTable]
		ifFalse: [aDatabaseTable].
	tables add: table.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
tables
	^tables! !

!Descriptor methodsFor: 'creating mappings' stamp: ''!
toManyMapping
	^self newMapping: ToManyMapping.! !

!Descriptor methodsFor: 'tracing' stamp: ''!
trace: aTracing context: anExpression
	"For each mapping, check if the relationship is involved in the set of things
to be read"
	mappings do: [:each |
		each trace: aTracing context: anExpression].! !

!Descriptor methodsFor: 'accessing' stamp: ''!
typeMapping
	^mappings detect: [ :each | each isTypeMapping ] ifNone: [
		| mapping |
		mapping := IdentityTypeMapping new.
		self addMapping: mapping.
		mapping]! !

!Descriptor methodsFor: 'accessing' stamp: ''!
typeMapping: aMapping
	self addMapping: aMapping! !

!Descriptor methodsFor: 'type mapping' stamp: ''!
typeMappingCriteriaIn: base
	| r l |
	r := ConstantExpression for: self typeMapping keys.
	l := FieldExpression forField: self typeMapping field basedOn: base.
	^l in: r! !

!Descriptor methodsFor: 'type resolution' stamp: ''!
typeMappingRoot
	^self typeResolver typeMappingRoot! !

!Descriptor methodsFor: 'type resolution' stamp: ''!
typeMappingRootDescriptor
	^self typeResolver typeMappingRootDescriptor! !

!Descriptor methodsFor: 'accessing' stamp: ''!
typeResolver
	typeResolver isNil
		ifTrue: [  IdentityTypeResolver new register: self  ].
	^typeResolver! !

!Descriptor methodsFor: 'accessing' stamp: ''!
typeResolver: anObject
	typeResolver := anObject! !

!Descriptor methodsFor: 'validation' stamp: ''!
validate
	"Do any tests we can to make sure that this definition makes sense in this context"
	| readableMappings attributeNames |
	self system validateDescriptor: self.
	self mappings do: [:eachMapping |
		eachMapping validate].
	"Validate that we have at most one readable mapping for an instance variable"
	readableMappings := self mappings select: [:each | each canRead and: [each isTypeMapping not and: [each attributeName isEmpty not]]].
	attributeNames := readableMappings collect: [:each | each attributeName].
	(attributeNames asSet size = readableMappings size) 
		ifFalse: [MultipleMappingsForInstanceVariable new
					model: self describedClass;
					readableMappings: readableMappings;
					raiseRequest].

	self typeResolver validate.! !

!Descriptor methodsFor: 'accessing' stamp: ''!
valueDescriptor
	| valueMapping |
	valueMapping := self mappingForAttributeNamed: #value.
	valueMapping isNil ifTrue: [^nil].
	^valueMapping referenceDescriptor.! !

!Descriptor methodsFor: 'mapping' stamp: ''!
valueOfField: aField fromObject: anObject
	^self valueOfField: aField fromObject: anObject ifNone: [nil].! !

!Descriptor methodsFor: 'mapping' stamp: ''!
valueOfField: aField fromObject: anObject ifNone: aBlock
	| directMapping notFound |
	directMapping := self directMappingForField: aField.
	directMapping isNil
		ifFalse: [^directMapping valueOfField: aField fromObject: anObject].
	notFound := Object new.
	(self relationshipMappingsForField: aField)
		do:
			[:each | 
			| candidate |
			(each hasValueThatCanProvideField: aField from: anObject)
				ifTrue: [
					candidate := each valueOfField: aField fromObject: anObject ifNone: [notFound].
					candidate == notFound ifFalse: [^candidate]]].
	^aBlock value.! !

Object subclass: #DescriptorSystem
	instanceVariableNames: 'session platform descriptors tables sequences typeResolvers cachePolicy allClasses useDirectAccessForMapping classModels useJoinsForAnySatisfy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!DescriptorSystem commentStamp: '<historical>' prior: 0!
This holds onto the Glorp metadata. Each session will have a DescriptorSystem associated with it, which describes the tables, descriptors, and class models that are to be made persistent.

Subclasses must implement the following messages:
	accessing
		allTableNames
		constructAllClasses

Instance Variables:
	allClasses	<Collection of: Behavior>	The classes we have metadata for.
	cachePolicy	<CachePolicy>	The default cache policy to use. Descriptors can override this.
	classModels	<Dictionary from: Behavior to: GlorpClassModel>	Our class models, indexed by class.
	descriptors	<Dictionary from: Behavior to: Descriptor>	Our descriptors, indexed by class.
	platform	<DatabasePlatform>	The platform for which we will be used. This is required because the table definitions will refer to the platform in order to get the appropriate types for fields.
	sequences	<Dictionary from: String to: DatabaseSequence>	The sequences that we have available, indexed by name asUppercase.
	session	<GlorpSession>	The session we provide metadata to.
	tables	<Dictionary from: String to: DatabaseTable>	Our database tables, indexed by name asUppercase.
	typeResolvers	<Dictionary from: Behavior to: TypeResolver>	The type resolvers that we will use for inheritance, indexed by class.
	useDirectAccessForMapping	<Boolean>	By default, should we use direct instance variable access (instVarAt:) or send get/set messages. Individual attributes can override this.
	useJoinsForAnySatisfy	<Boolean>	By default, should we use joins for anySatisfy and similar queries, or should we use EXISTS subselects.

!

!DescriptorSystem classSide methodsFor: 'utility' stamp: ''!
createTablesFor: aLogin
	"Create the tables for this descriptor system in the given database. This will fail if the tables already exist. Use recreateTables for that."

	| session |
	session := self sessionForLogin: aLogin.
	session login.
	[session createTables] ensure: 
		[session logout].! !

!DescriptorSystem classSide methodsFor: 'utility' stamp: ''!
dropTablesFor: aLogin
	"Drop the tables for this descriptor system in the given database."

	| session |
	session := self sessionForLogin: aLogin.
	session login.
	[session dropTables] ensure: [session logout].! !

!DescriptorSystem classSide methodsFor: 'instance creation' stamp: ''!
forPlatform: dbPlatform

	^super new initialize;
		platform: dbPlatform! !

!DescriptorSystem classSide methodsFor: 'configuration' stamp: ''!
includeAllClassesWithDescriptorMethods
	"If this is true, the descriptor system will automatically include all classes that define the #glorpSetupDescriptor:forSystem: method on the class side. If false, it will only include classes which have descriptor methods defined in this class, or those explicitly added in the #constructAllClasses method."
	^false.! !

!DescriptorSystem classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!DescriptorSystem classSide methodsFor: 'utility' stamp: ''!
recreateTablesFor: aLogin
	"Drop and recreate the tables for this descriptor system in the given database."

	| session |
	session := self sessionForLogin: aLogin.
	session login.
	[session recreateTables] ensure: [session logout].! !

!DescriptorSystem classSide methodsFor: 'instance creation' stamp: ''!
sessionForLogin: aGlorpLogin
	"Create a session associated with the given login, and using us as a descriptor system. This is the easiest way to get a full blown session going."
	| system session |
	system := self forPlatform: aGlorpLogin database.
	session := GlorpSession new.
	session accessor: (DatabaseAccessor forLogin: aGlorpLogin).
	session system: system.
	^session.! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
addClassModelFor: aClass
	^classModels
		at: aClass
		ifAbsentPut: [self newClassModelFor: aClass]! !

!DescriptorSystem methodsFor: 'initialization' stamp: ''!
addClassesFromDescriptorMethodsFor: aClass
	| match |
	match := 'descriptorFor'.
	Dialect selectorsFor: aClass do: 
			[:each | 
			(each copyFrom: 1 to: (match size min: each size)) = match
				ifTrue:
					[| className actualClass |
					className := each copyFrom: match size + 1 to: each size - 1.
					actualClass := Dialect
						lookupClassNamed: className
						inContextOf: self class
						ifAbsent: [nil].
					actualClass isNil ifFalse: [allClasses add: actualClass]]].! !

!DescriptorSystem methodsFor: 'initialization' stamp: ''!
addTablesFromInitializeMethodsFor: aClass
	| match |
	match := 'tableFor'. 
	Dialect selectorsFor: aClass do:
			[:each | 
			(each copyFrom: 1 to: (match size min: each size)) = match
				ifTrue:
					[| tableName |
					tableName := each copyFrom: match size + 1 to: each size - 1.
					self tableNamed: tableName]].! !

!DescriptorSystem methodsFor: 'callbacks' stamp: ''!
additionalRelatedRowsFor: aRow in: aRowMap
	"This method serves as a hook. It is called when sorting rows before writing to the database, and can be used to return a list of additional related rows that aren't found by the normal foreign key constraints. These rows will act as an ordering constraint, and will have to be written before aRow."
	^nil.! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
allClassModels
	^self allClasses collect: [:each | 
		self classModelFor: each].! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
allClasses
	allClasses isEmpty ifTrue: [allClasses := self constructAllClasses].
	^allClasses! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
allDescriptors

	^self allClasses collect: [:each | 
		self descriptorFor: each].! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
allSequences
	sequences isEmpty ifFalse: [^sequences].
	sequences := self allSequencesFor: self allTables.
	^sequences.! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
allSequencesFor: aCollection
	"Find all the sequences associated with aCollectionOfTables"
	| result |
	result := Dictionary new.
	aCollection do: [:each |
		each allSequences keysAndValuesDo: [:eachName :eachSequence |
				result at: eachName put: eachSequence]].
	^result.! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
allTableNames
	"Return a list of all the table names that this system uses."
	^#().! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
allTables
	tables isEmpty ifTrue: [self constructAllTables].
	^tables.! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
cachePolicy
	"Return the default cache policy that will be used for descriptors that don't specify their own policy"
	cachePolicy isNil ifTrue: [cachePolicy := CachePolicy default].
	^cachePolicy.! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
cachePolicy: aCachePolicy
	cachePolicy := aCachePolicy.! !

!DescriptorSystem methodsFor: 'testing' stamp: ''!
canUseGroupedWrites
	"We can use grouped writes as long as there are no identity columns in use"
	self platform usesIdentityColumns ifFalse: [^true].
	^(self allTables anySatisfy: [:each |
		each primaryKeyFields anySatisfy: [:eachField |
			eachField type isSerial and: [eachField type sequence isIdentityColumn]]]) not.! !

!DescriptorSystem methodsFor: 'testing' stamp: ''!
canUseGroupedWritesFor: aTable
	"We can use grouped writes as long as there are no identity columns in use"
	self platform usesIdentityColumns ifFalse: [^true].
	^(aTable primaryKeyFields anySatisfy: [:eachField |
			eachField type isSerial and: [eachField type sequence isIdentityColumn]]) not.! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
classModelFor: aClass
	(self allClasses includes: aClass) ifFalse: [^nil].
	^classModels
		at: aClass
		ifAbsentPut: [self newClassModelFor: aClass]! !

!DescriptorSystem methodsFor: 'initialization' stamp: ''!
constructAllClasses
	"Subclasses can override this method to add classes. But by default we also look for all descriptor methods in our system, and add them to the class list, so if the methods are written there, then we don't need to do anything further. If the descriptor methods are defined on the classes themselves, then we'll still have to construct the class list manually."

	self constructClassesFromDescriptorMethods.
	^allClasses.! !

!DescriptorSystem methodsFor: 'initialization' stamp: ''!
constructAllTables
	"Subclasses can override this method to add tables to the list. But by default we also look for all table methods in our system, and add them to the class list, so if the methods are written there, then we don't need to do anything further. If the table methods are defined on the classes themselves, then we'll still have to construct the list manually."
	self allTableNames do: [:each | 
		self tableNamed: each asString].
	self constructTablesFromInitializeMethods.
	^tables.! !

!DescriptorSystem methodsFor: 'initialization' stamp: ''!
constructClassesFromDescriptorMethods
	self withAllSuperclassesDo: [:aClass | self addClassesFromDescriptorMethodsFor: aClass].

	self class includeAllClassesWithDescriptorMethods
		ifTrue:
			[Object withAllSubclasses
				do:
					[:each | 
					(each class
						compiledMethodAt: #glorpSetupDescriptor:forSystem:
						ifAbsent: [nil]) isNil ifFalse: [allClasses add: each]]].! !

!DescriptorSystem methodsFor: 'initialization' stamp: ''!
constructTablesFromInitializeMethods
	self withAllSuperclassesDo: [:aClass | self addTablesFromInitializeMethodsFor: aClass].! !

!DescriptorSystem methodsFor: 'private' stamp: ''!
converterBetweenStType: aClass andField: aField 
	| selector |
	selector := ('converterBetween' , aClass name , 'AndField:') asSymbol.
	(self respondsTo: selector) ifTrue: [^self perform: selector with: aField].
	^aField converterForStType: aClass! !

!DescriptorSystem methodsFor: 'private' stamp: ''!
createMappingIn: aDescriptor fromClassModelForAttributeNamed: aSymbol 
	| attribute referenceModel |
	attribute := aDescriptor classModel attributeNamed: aSymbol.
	referenceModel := self classModelFor: attribute type.
	referenceModel isNil 
		ifTrue: 
			[^(aDescriptor newMapping: DirectMapping) from: aSymbol
				to: (aDescriptor table 
						fieldNamed: (self fieldNameForAttributeName: aSymbol))]
		ifFalse: 
			[attribute isCollectionAttribute 
				ifTrue: [^(aDescriptor newMapping: ToManyMapping) attributeName: aSymbol]
				ifFalse: [^(aDescriptor newMapping: OneToOneMapping) attributeName: aSymbol]]! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
createMappingsFromClassModelIn: aDescriptor 
	aDescriptor classModel attributesDo: 
			[:each | 
			(aDescriptor mappingForAttributeNamed: each name) isNil 
				ifTrue: 
					[self createMappingIn: aDescriptor
						fromClassModelForAttributeNamed: each name]]! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
databaseSchema

	^ self session schema! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
databaseSequenceNamed: aString

	^sequences at: aString asUppercase ifAbsentPut: [platform newDatabaseSequenceNamed: aString].! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
databaseSequenceNamed: aString ifAbsentPut: aBlock

	^sequences at: aString asUppercase ifAbsentPut: aBlock.! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
defaultTracing

	^Tracing new! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
descriptorFor: aClassOrObject 
	| theClass newDescriptor |
	aClassOrObject == Proxy ifTrue: [self error: 'Cannot find descriptor for the class Proxy. Pass in the instance'].
	theClass := aClassOrObject glorpCanHaveDescriptorDirectly 
				ifTrue: [aClassOrObject]
				ifFalse: [aClassOrObject isGlorpProxy 
							ifTrue: [aClassOrObject getValue class]
							ifFalse: [aClassOrObject class]].
	^descriptors
		at: theClass
		ifAbsent: [
			(self allClasses includes: theClass) ifTrue: [
				newDescriptor := self newDescriptorFor: theClass.
				descriptors at: theClass put: newDescriptor.
				self initializeDescriptor: newDescriptor.
				newDescriptor.
			]].! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
descriptorForTable: aTable
	"Return a descriptor that uses aTable. If there is more than one, the choice is random."
	^descriptors detect: [:each | each tables includes: aTable] ifNone: [nil].! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
descriptorsForTable: aTable
	"Return all descriptors that use aTable."
	^descriptors select: [:each | each tables includes: aTable].! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
existingTableNamed: aString

	^tables at: aString asUppercase ifAbsent: [self error: 'missing table'].! !

!DescriptorSystem methodsFor: 'callbacks' stamp: ''!
fieldNameForAttributeName: aSymbol 
	^aSymbol asString asUppercase! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
flushAllClasses
	allClasses := OrderedCollection new.! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
hasDescriptorFor: aClassOrObject 

	^(self descriptorFor: aClassOrObject) notNil.! !

!DescriptorSystem methodsFor: 'initialization' stamp: ''!
initialize
	descriptors := Dictionary new.
	tables := Dictionary new.
	classModels := Dictionary new.
	typeResolvers := Dictionary new.
	sequences := Dictionary new.
	useDirectAccessForMapping := true.
	allClasses := OrderedCollection new.! !

!DescriptorSystem methodsFor: 'private' stamp: ''!
initializeClassModel: aClassModel
	| selector |
	selector := ('classModelFor' , aClassModel describedClass name , ':')
		asSymbol.
	(self respondsTo: selector)
		ifTrue: [^self perform: selector with: aClassModel].
	(aClassModel describedClass respondsTo: #glorpSetupClassModel:forSystem:)
		ifTrue:
			[^aClassModel describedClass
				glorpSetupClassModel: aClassModel
				forSystem: self].
	"Otherwise, just initialize it dynamically, so that old code with only descriptors will still work"
	^aClassModel.! !

!DescriptorSystem methodsFor: 'private' stamp: ''!
initializeDescriptor: aDescriptor 
	| selector |
	selector := ('descriptorFor' , aDescriptor describedClass name , ':') asSymbol.
	(self respondsTo: selector) 
		ifTrue: [self perform: selector with: aDescriptor]
		ifFalse: [
			(aDescriptor describedClass respondsTo: #glorpSetupDescriptor:forSystem:)
				ifTrue: [aDescriptor describedClass
								glorpSetupDescriptor: aDescriptor
								forSystem: self]
				ifFalse: [self noDescriptorInitializerFound: aDescriptor]].! !

!DescriptorSystem methodsFor: 'private' stamp: ''!
initializeTable: newTable
	| initializerName |
	initializerName := ('tableFor' , (newTable name asUppercase) , ':') asSymbol.
	(self respondsTo: initializerName)
		ifTrue: [self perform: initializerName with: newTable]
		ifFalse: [self noTableInitializerFound: newTable].
	newTable postInitializeIn: self.! !

!DescriptorSystem methodsFor: 'callbacks' stamp: ''!
joinFor: aMapping toTables: toTables fromConstraints: fromConstraints toConstraints: toConstraints
	"Return the join to be used for aMapping. The set of tables we expect this to be related to, and the constraints from this table, and to this table are provided as possibly useful input."
	| numberOfConstraints |
	numberOfConstraints := (Set new addAll: toConstraints; addAll: fromConstraints; yourself) size.
	numberOfConstraints = 0 ifTrue: [
		CannotAutomaticallyDetermineJoin new 
			tables: toTables;
			constraints: (toConstraints, fromConstraints);
			signal: 'Cannot automatically determine join, no constraints found.'].
	numberOfConstraints > 1 ifTrue: [
		CannotAutomaticallyDetermineJoin new
			tables: toTables;
			constraints: (toConstraints, fromConstraints);
			signal: 'Cannot automatically determine join, too many constraints found.'].
	^toConstraints isEmpty ifTrue: [fromConstraints first asReverseJoin] ifFalse: [toConstraints first asJoin].! !

!DescriptorSystem methodsFor: 'optimization parameters' stamp: ''!
maximumSizeToGroupWriteFor: aCollectionOfDatabaseRows
	"If we are going to group write, how many rows of this collection should we do it for at once. Normally we delegate this to the platform, but we allow the system to intervene if desired"
	^platform maximumSizeToGroupWriteFor: aCollectionOfDatabaseRows.! !

!DescriptorSystem methodsFor: 'private' stamp: ''!
newClassModelFor: aClass

	| newClassModel |
	newClassModel := GlorpClassModel new.
	newClassModel system: self.
	newClassModel describedClass: aClass.
	self initializeClassModel: newClassModel.
	^newClassModel.! !

!DescriptorSystem methodsFor: 'private' stamp: ''!
newDescriptorFor: aClass

	| newDescriptor |
	(self allClasses includes: aClass) ifFalse: [^nil].
	newDescriptor := Descriptor new.
	newDescriptor system: self.
	newDescriptor classModel: (self classModelFor: aClass).
	^newDescriptor.! !

!DescriptorSystem methodsFor: 'private' stamp: ''!
newTypeResolverFor: aClass 
	| selector |
	(self allClasses includes: aClass) ifFalse: [^nil].
	selector := ('typeResolverFor' , aClass name) asSymbol.
	^(self respondsTo: selector) 
		ifTrue: [self perform: selector]
		ifFalse: [aClass glorpTypeResolver]! !

!DescriptorSystem methodsFor: 'validation' stamp: ''!
noDescriptorInitializerFound: aDescriptor
	self error: 'No descriptor initializer found for ', aDescriptor printString.! !

!DescriptorSystem methodsFor: 'validation' stamp: ''!
noTableInitializerFound: aDatabaseTable
	self error: 'No table initializer found for ', aDatabaseTable printString.! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
platform

	^platform! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
platform: dbPlatform

	platform := dbPlatform! !

!DescriptorSystem methodsFor: 'private' stamp: ''!
privateAndBasicAddDescriptorFor: anObject as: aDescriptor
	"Just add the descriptor and don't try to validate. anObject may not actually even be a class. Assume the descriptor is already properly initialized"
	descriptors at: anObject put: aDescriptor.
	aDescriptor system: self.! !

!DescriptorSystem methodsFor: 'sequences' stamp: ''!
sequenceNameForField: aDatabaseField
	"Return the name of the sequence we will use for a field. Subclasses can override to use different naming conventions."
	^(aDatabaseField table sqlString), '_', (aDatabaseField name), '_seq'! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
sequenceTypeNamed: aString
	"Return a sequence type whose sequence has the given name."
	^self platform sequence sequence: (self databaseSequenceNamed: aString).! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
session
	^session! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
session: anObject
	session := anObject! !

!DescriptorSystem methodsFor: 'session creation' stamp: ''!
sessionForLogin: aGlorpLogin
	"Create a session associated with the given login, and using us as a descriptor system. This is the easiest way to get a full blown session going."
	| newSession |
	platform isNil ifTrue: [self platform: aGlorpLogin database].
	aGlorpLogin database class = self platform class ifFalse: [self error: 'Login is incompatible with the database specified in the descriptor system'].
	newSession := GlorpSession new.
	newSession accessor: (DatabaseAccessor forLogin: aGlorpLogin).
	newSession system: self.
	^newSession.! !

!DescriptorSystem methodsFor: 'initialization' stamp: ''!
setUpDefaults
	"For systems that are configurable, set them up for testing configuration"! !

!DescriptorSystem methodsFor: 'settings' stamp: ''!
shouldForeignKeyValueBeTreatedAsNull: aValue
	"Return true if this foreign key value should be treated as null, meaning there's no relationship. Normally this means just nil, but descriptor systems can override if they have special values which they use (e.g. 0) that indicate a missing relationship. This can improve performance by avoiding unnecessary queries to the database for things we know will never return a result"
	^aValue isNil.! !

!DescriptorSystem methodsFor: 'settings' stamp: ''!
shouldResetExpiryTimeForObjectOfClass: aClass readFrom: aQuery
	"Return true if, when an object is read from a query, even if it's not refreshing. we 'touch' it, resetting its expiry time. By default, we do this if the cache policy for that type of object will remove from the cache. The rationale is that this means we're using the object. Removing an object in use is a very bad thing, and can cause loss of identity and other bad things. However, if expiry only indicates a need to refresh, we don't need to touch the object. Subclasses can override to provide different behaviour."

	aQuery shouldRefresh ifTrue: [^true].
	aQuery returnProxies ifTrue: [^false].
	^(self session cacheForClass: aClass)  willRemoveItems! !

!DescriptorSystem methodsFor: 'optimization parameters' stamp: ''!
shouldUpdateOnlyChangedFieldsFor: aRow
	"Return true if we should write only changed fields out in an update statement, or if we should write all of them. The tradeoff is that writing only the changed ones will produce less SQL, but make more variations of the statement, and require us to keep more prepared statements around if we're using binding. By default, use this option only if we're not using binding by default"
	^aRow useBinding not.! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
tableNamed: aString

	^self tableNamed: aString ifNone: [NoTableFound new parameter: aString; raiseRequest].! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
tableNamed: aString ifNone: aBlock

	^tables at: aString asString asUppercase ifAbsent: [
		| newTable|
		newTable := DatabaseTable new.
		newTable name: aString.
		tables at: aString asUppercase put: newTable.
		self initializeTable: newTable.
		newTable isEmpty ifTrue: [tables removeKey: aString asUppercase. ^aBlock value].
		newTable].! !

!DescriptorSystem methodsFor: 'internal/writing' stamp: ''!
tablesInCommitOrder

	^(TableSorter for: self allTables) sort.! !

!DescriptorSystem methodsFor: 'api' stamp: ''!
typeResolverFor: aClassOrObject 

	| theClass |
	aClassOrObject == Proxy ifTrue: [self error: 'Cannot find type resolver for the class Proxy. Pass in the instance'].
	theClass := aClassOrObject glorpCanHaveDescriptorDirectly 
				ifTrue: [aClassOrObject]
				ifFalse: [aClassOrObject isGlorpProxy ifTrue: [aClassOrObject getValue class] ifFalse: [aClassOrObject class]].
	^typeResolvers at: theClass ifAbsentPut: [self newTypeResolverFor: theClass]! !

!DescriptorSystem methodsFor: 'validation' stamp: ''!
unmappedInstanceVariables
	"Return a list, for all of our classes, of instance variables with no mappings. This can be used by inspection to see if we're missing mappings we think we ought to have."
	"self unmappedInstanceVariables"

	| unmapped |
	unmapped := Dictionary new.
	self allClasses do: [:eachClass |
		eachClass allInstVarNames do: [:eachInstVarName |
			((self descriptorFor: eachClass) mappingForAttributeNamed: eachInstVarName asSymbol)
				isNil ifTrue: [
					(unmapped at: eachClass ifAbsentPut: [OrderedCollection new])
						add: eachInstVarName]]].
	^unmapped.! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
useDirectAccessForMapping
	^useDirectAccessForMapping! !

!DescriptorSystem methodsFor: 'accessing' stamp: ''!
useDirectAccessForMapping: anObject
	useDirectAccessForMapping := anObject! !

!DescriptorSystem methodsFor: 'settings' stamp: ''!
useJoinsForAnySatisfy
	useJoinsForAnySatisfy isNil ifTrue: [useJoinsForAnySatisfy := false].
	^useJoinsForAnySatisfy.! !

!DescriptorSystem methodsFor: 'settings' stamp: ''!
useJoinsForAnySatisfy: aBoolean
	useJoinsForAnySatisfy := aBoolean.! !

!DescriptorSystem methodsFor: 'validation' stamp: ''!
validate
	session isNil ifFalse: [session system == self ifFalse: [InvalidSessionSetup  raiseErrorString:  'Invalid session setup']].
	self allTables do: [:each | each validateIn: self].
	self allClassModels do: [:each | each validateIn: self].
	self allDescriptors do: [:each | each validate].! !

!DescriptorSystem methodsFor: 'validation' stamp: ''!
validateDescriptor: aDescriptor
	"Validate aDescriptor in the context of this system, doing any specific validations that need the full environment information or rely on subclass-specific information here."! !

!DescriptorSystem methodsFor: 'validation' stamp: ''!
validateRows: aUnitOfWork
	"This is a hook to allow the system to do application-specific validation of the rows to be written. I suppose it's also possible to modify the rows here. Subclasses can override if they have useful work to do here. This will be called after the rows are generated, pre-write notification has been sent, rows have been sorted for writing, and and just before sequence numbers (if pre-write) are assigned, and writing starts"! !

!DescriptorSystem methodsFor: 'validation' stamp: ''!
validateRowsPostWrite: aUnitOfWork
	"Validate after writing. This is a bit late, but some things may not be catchable ahead of time"! !

!DescriptorSystem methodsFor: 'initialization' stamp: ''!
withAllSuperclassesDo: aBlock
	| currentClass |
	currentClass := self class.
	[currentClass == Object]
		whileFalse:
			[aBlock value: currentClass.
			currentClass := currentClass superclass].! !

DescriptorSystem subclass: #DynamicDescriptorSystem
	instanceVariableNames: 'attributeNameToFieldNameBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!DynamicDescriptorSystem commentStamp: '<historical>' prior: 0!
This is a descriptor system whose descriptors and tables are created dynamically rather than out of generated code. Note that identity is extremely important, so care is required to set these up properly.
!

!DynamicDescriptorSystem methodsFor: 'accessing' stamp: ''!
addEmptyDescriptorFor: aClass 
	| descriptor |
	descriptor := Descriptor new.
	descriptor system: self.
	self privateDescriptorAt: aClass put: descriptor.
	descriptor classModel: (self classModelFor: aClass).
	^descriptor! !

!DynamicDescriptorSystem methodsFor: 'accessing' stamp: ''!
addEmptyTableNamed: aString 
	| table |
	table := DatabaseTable named: aString.
	self privateTableAt: aString put: table.
	^table! !

!DynamicDescriptorSystem methodsFor: 'accessing' stamp: ''!
allClasses
	^descriptors keys! !

!DynamicDescriptorSystem methodsFor: 'accessing' stamp: ''!
allTableNames
	^tables keys! !

!DynamicDescriptorSystem methodsFor: 'accessing' stamp: ''!
attributeNameToFieldNameBlock: aBlock
	attributeNameToFieldNameBlock := aBlock! !

!DynamicDescriptorSystem methodsFor: 'hooks' stamp: ''!
fieldNameForAttributeName: aSymbol 
	attributeNameToFieldNameBlock isNil  
		ifTrue: [^super fieldNameForAttributeName: aSymbol].
	^attributeNameToFieldNameBlock value: aSymbol! !

!DynamicDescriptorSystem methodsFor: 'accessing' stamp: ''!
initializeClassModel: aClassModel
	"Assume we'll initialize it either before or after this"! !

!DynamicDescriptorSystem methodsFor: 'private' stamp: ''!
privateDescriptorAt: aClass put: aDescriptor
	"Normally you don't want to be setting tables explicitly, as it may defeat the identity management but it's here if needed"
	descriptors at: aClass put: aDescriptor.! !

!DynamicDescriptorSystem methodsFor: 'private' stamp: ''!
privateTableAt: aString put: aTable
	"Normally you don't want to be setting tables explicitly, as it may defeat the identity management but it's here if needed"
	tables at: aString asUppercase put: aTable.! !

Object subclass: #Dialect
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

Dialect class
	instanceVariableNames: 'dialectName timestampClass identitySetClass lookedUpNames isNamespaceAware'!

!Dialect commentStamp: '<historical>' prior: 0!
This is an encapsulation layer for dialect differences between Smalltalk. It has a variety of class messages that do different things depending which dialect we are running on.!

Dialect class
	instanceVariableNames: 'dialectName timestampClass identitySetClass lookedUpNames isNamespaceAware'!

!Dialect classSide methodsFor: 'dates' stamp: ''!
addSeconds: seconds to: aTime

	self isVisualWorks ifTrue: [^ Time fromSeconds: aTime asSeconds + seconds].
	self isDolphin ifTrue: [^ self addTimeForDolphin: aTime seconds: seconds].
	self isGNU ifTrue: [^ aTime addSeconds: seconds].
	self isVisualAge ifTrue: [^Time fromSeconds: aTime asSeconds + seconds].
	self error: 'not implemented'.
	^ self! !

!Dialect classSide methodsFor: 'dates' stamp: ''!
addTimeForDolphin: aTime seconds: seconds

	"Dolphin's time/date arithmetic is pretty weak, especially for timestamps. Hack around it. This is likely only to work for seconds <24 hours"

	| result |

	^ aTime class == Time
		ifTrue: [Time fromMilliseconds: (aTime asMilliseconds + (seconds * 1000)) \\ 86400000.]
		ifFalse: [
			result := self timestampClass date: aTime date
				time: (self addTimeForDolphin: aTime time seconds: seconds).
			(seconds > 0 and: [result time < aTime time.])
				ifTrue: [result date: (result date addDays: 1).].
			(seconds < 0 and: [result time > aTime time.])
				ifTrue: [result date: (result date addDays: -1).].
			^ result
		]! !

!Dialect classSide methodsFor: 'general portability' stamp: 'GuillermoPolito 4/16/2011 23:00'!
argumentCountFor: aBlock

	Dialect isGNU ifTrue: [^ self error: 'not implemented yet'].
	Dialect isVisualWorks ifTrue: [^ aBlock numArgs].
	Dialect isVisualAge ifTrue: [^ aBlock argumentCount].
	Dialect isDolphin ifTrue: [^ aBlock argumentCount].
	Dialect isSqueak ifTrue: [^self error: 'not implemented yet'].
	Dialect isPharo ifTrue: [ ^aBlock argumentCount ].
	Dialect isObjectStudio ifTrue: [^aBlock numArgs].
	self error: 'not implemented yet'.
	^ self! !

!Dialect classSide methodsFor: 'private' stamp: ''!
basicIsDolphin
	^Smalltalk includesKey: #DolphinSplash.! !

!Dialect classSide methodsFor: 'private' stamp: ''!
basicIsGNU
	^Smalltalk includesKey: #BindingDictionary.! !

!Dialect classSide methodsFor: 'private' stamp: ''!
basicIsObjectStudio
	^Smalltalk class name == #SmalltalkClass.! !

!Dialect classSide methodsFor: 'private' stamp: 'GuillermoPolito 4/16/2011 22:42'!
basicIsPharo

	| versionString |
	Smalltalk at: #SystemVersion ifPresent: [:cl |
		versionString := cl current version.
		^(versionString beginsWith: 'Pharo')
	].
	^false! !

!Dialect classSide methodsFor: 'private' stamp: 'GuillermoPolito 4/16/2011 22:42'!
basicIsSqueak

	| versionString |
	Smalltalk at: #SystemVersion ifPresent: [:cl |
		versionString := cl current version.
		^(versionString beginsWith: 'Squeak')
	].
	^false! !

!Dialect classSide methodsFor: 'private' stamp: ''!
basicIsVisualAge

	^ Smalltalk class name == #EsSmalltalkNamespace
	"	| sys |
	sys := Smalltalk at: #System ifAbsent: [^false].
	(sys respondsTo: #vmType) ifFalse: [^false].
	^sys vmType = 'ES'"! !

!Dialect classSide methodsFor: 'private' stamp: ''!
basicIsVisualWorks

	^ Smalltalk class name == #NameSpace
	"Smalltalk class selectors do: [ :s | 
		(s == #versionName and: [ (Smalltalk versionName copyFrom: 1 to: 11) = 'VisualWorks']) 
		    ifTrue: [^true]].
	^false"! !

!Dialect classSide methodsFor: 'private' stamp: ''!
basicIsVisualWorksWithoutNamespaces
	^Smalltalk class name == #SystemDictionary! !

!Dialect classSide methodsFor: 'numbers' stamp: 'GuillermoPolito 4/16/2011 22:46'!
coerceToDoublePrecisionFloat: aNumber

	self isGNU ifTrue: [^ aNumber asFloatD].
	self isVisualWorks ifTrue: [^ aNumber asDouble].
	self isVisualAge ifTrue: [^ aNumber asFloat].
	self isSqueak ifTrue: [^aNumber asFloat].
	self isPharo ifTrue: [^aNumber asFloat].
	^ aNumber! !

!Dialect classSide methodsFor: 'files' stamp: 'GuillermoPolito 4/16/2011 22:51'!
contentsOfFileNamed: aString 
	"Given a text file name, read its contents"

	| stream contents fileClass |
	Dialect isVisualWorks ifTrue: [^aString asFilename contentsOfEntireFile].
	Dialect isSqueak ifTrue: [
		stream := (Dialect smalltalkAt: #CrLfFileStream) oldFileNamed: aString.
		[contents := stream contents] ensure: [stream close].
		^contents].
	
	Dialect isPharo ifTrue: [
		stream := (Dialect smalltalkAt: #MultiByteFileStream) oldFileNamed: aString.
		[contents := stream contents] ensure: [stream close].
		^contents ].

	fileClass := Dialect isVisualAge 
				ifTrue: [self smalltalkAt: #CfsReadFileStream]
				ifFalse: 
					[Dialect isDolphin ifTrue: [self smalltalkAt: #File] ifFalse: [self halt]].
	stream := fileClass read: aString.
	[contents := stream contents] ensure: [stream close].
	^contents! !

!Dialect classSide methodsFor: 'strings' stamp: ''!
convert: aString toByteArrayWithEncoding: encodingName
	self isVisualWorks ifTrue: [^aString asByteArrayEncoding: encodingName].
	self error: 'Not yet implemented'.! !

!Dialect classSide methodsFor: 'strings' stamp: ''!
convert: aByteArray toStringWithEncoding: encodingName
	self isVisualWorks ifTrue: [^aByteArray asStringEncoding: encodingName].
	self error: 'Not yet implemented'.! !

!Dialect classSide methodsFor: 'private' stamp: 'GuillermoPolito 4/16/2011 22:41'!
determineDialect
	self basicIsDolphin ifTrue: [^ dialectName := #Dolphin].
	self basicIsGNU ifTrue: [^ dialectName := #GNU].
	self basicIsVisualAge ifTrue: [^ dialectName := #VisualAge].
	self basicIsVisualWorks ifTrue: [^ dialectName := #VisualWorks].
	self basicIsSqueak ifTrue: [^ dialectName := #Squeak].
	self basicIsPharo ifTrue: [^ dialectName := #Pharo].
	self basicIsObjectStudio ifTrue: [^dialectName := #ObjectStudio].
	self basicIsVisualWorksWithoutNamespaces ifTrue: [^dialectName := #VisualWorks].
	self error: 'I don''t know what dialect this is'.
	^ self! !

!Dialect classSide methodsFor: 'identifying' stamp: ''!
dialectName

	dialectName isNil ifTrue: [self determineDialect.].
	^ dialectName! !

!Dialect classSide methodsFor: 'numbers' stamp: ''!
doesPrecisionOf: aNumber equal: aPrecision
	^aNumber class == self fixedPointClass and: [aNumber scale = aPrecision].! !

!Dialect classSide methodsFor: 'numbers' stamp: ''!
doublePrecisionFloatClass

	self isGNU ifTrue: [^ self smalltalkAt: #FloatD].
	self isVisualWorks ifTrue: [^ self smalltalkAt: #Double].
	^ Float! !

!Dialect classSide methodsFor: 'general portability' stamp: ''!
error
	"If VASmalltalk 8 has switched entirely to ANSI exceptions, perhaps we can retire this.
Meanwhile, it is used in some Glorp on:do: calls."

	^Dialect isVisualAge
		ifTrue: [Error, (Dialect smalltalkAt: #ExError)]
		ifFalse: [Error]! !

!Dialect classSide methodsFor: 'general portability' stamp: ''!
fitHashIntoSmallInteger: aNumber
	"Truncate the number to fit into smallinteger range. Mostly useful for ObjectStudio where hashes have to be 16 bits or less"
	^Dialect isObjectStudio ifTrue: [aNumber hash] ifFalse: [aNumber].! !

!Dialect classSide methodsFor: 'numbers' stamp: ''!
fixedPointClass
	Dialect isVisualWorks ifTrue: [^self smalltalkAt: #FixedPoint].! !

!Dialect classSide methodsFor: 'general portability' stamp: 'GuillermoPolito 4/16/2011 22:54'!
garbageCollect

	Dialect isGNU ifTrue: [^ (self smalltalkAt: #ObjectMemory) globalGarbageCollect].
	Dialect isVisualWorks ifTrue: [^ (self smalltalkAt: #ObjectMemory) quickGC].
	Dialect isVisualAge ifTrue: [^ (self smalltalkAt: #System) globalGarbageCollect].
	(Dialect isSqueak or: [ Dialect isPharo ]) ifTrue: [^Smalltalk garbageCollect].
	Dialect isObjectStudio ifTrue: [^(self smalltalkAt: #System) garbageCollect].
	self error: 'not implemented yet'.
	^ self! !

!Dialect classSide methodsFor: 'forward references' stamp: ''!
glorpConstantExpressionClass
	^ConstantExpression.! !

!Dialect classSide methodsFor: 'general portability' stamp: ''!
identitySetClass
	identitySetClass == nil
		ifTrue:
			[Dialect isVisualAge
				ifTrue: [identitySetClass := self smalltalkAt: #EsIdentitySet]
				ifFalse: [identitySetClass := self smalltalkAt: #IdentitySet]].
	^identitySetClass.! !

!Dialect classSide methodsFor: 'general portability' stamp: ''!
instVarNameFor: aName

	Dialect isGNU ifTrue: [^ aName asSymbol].
	^ aName asString! !

!Dialect classSide methodsFor: 'general portability' stamp: ''!
isBlock: anObject
	"Return true if this is a block"
	self isVisualAge ifTrue: [^anObject isKindOf: (self smalltalkAt: #Block)].
	^anObject class == [] class.! !

!Dialect classSide methodsFor: 'identifying' stamp: ''!
isDolphin

	^ self dialectName = #Dolphin! !

!Dialect classSide methodsFor: 'numbers' stamp: ''!
isFloatingPoint: aNumber
	^aNumber class == self doublePrecisionFloatClass or: [aNumber class == Float].! !

!Dialect classSide methodsFor: 'identifying' stamp: ''!
isGNU

	^ self dialectName = #GNU! !

!Dialect classSide methodsFor: 'general portability' stamp: ''!
isHeadless
	"Dialect isHeadless"
	Dialect isVisualWorks ifTrue: [^ (Smalltalk at: 'WindowingSystem' ifAbsent: [^true]) isHeadless].
	self error: 'not implemented yet'.
	^ self! !

!Dialect classSide methodsFor: 'identifying' stamp: ''!
isNamespaceAware
	isNamespaceAware isNil ifTrue: [isNamespaceAware := self basicIsVisualWorks].
	^isNamespaceAware! !

!Dialect classSide methodsFor: 'identifying' stamp: ''!
isObjectStudio

	^ self dialectName = #ObjectStudio! !

!Dialect classSide methodsFor: 'identifying' stamp: 'GuillermoPolito 4/16/2011 22:41'!
isPharo

	^ self dialectName = #Pharo! !

!Dialect classSide methodsFor: 'identifying' stamp: 'GuillermoPolito 4/16/2011 22:41'!
isSqueak

	^ self dialectName = #Squeak! !

!Dialect classSide methodsFor: 'identifying' stamp: ''!
isVWWithNameSpaces
	"For cases where our hacked backports of asQualifiedReference and etc. to VW3 are not sufficient."

	^self isVisualWorks and: [self isNamespaceAware]! !

!Dialect classSide methodsFor: 'identifying' stamp: ''!
isVisualAge

	^ self dialectName = #VisualAge! !

!Dialect classSide methodsFor: 'identifying' stamp: ''!
isVisualWorks

	^ self dialectName = #VisualWorks! !

!Dialect classSide methodsFor: 'private' stamp: ''!
lookedUpNames
	lookedUpNames isNil ifTrue: [lookedUpNames := IdentityDictionary new].
	^lookedUpNames.! !

!Dialect classSide methodsFor: 'class lookup' stamp: ''!
lookupClassNamed: aString inContextOf: aClass
	"Look up the class by name, but use the namespace of the class. Mostly useful if we're trying to guess from unqualified names in dialects with namespaces."
	^self lookupClassNamed: aString inContextOf: aClass ifAbsent: [nil].! !

!Dialect classSide methodsFor: 'class lookup' stamp: ''!
lookupClassNamed: aString inContextOf: aClass ifAbsent: aBlock
	"Look up the class by name, but use the namespace of the class. Mostly useful if we're trying to guess from unqualified names in dialects with namespaces."
	self isVisualWorks ifFalse: [^self smalltalkAt: aString ifAbsent: aBlock].
	^aClass environment at: aString ifAbsent: aBlock.! !

!Dialect classSide methodsFor: 'collections' stamp: ''!
needsCollectionMementoFor: aCollection
	^true.! !

!Dialect classSide methodsFor: 'dates' stamp: 'GuillermoPolito 4/16/2011 22:44'!
newDateWithYears: years months: months days: days

	(self isVisualWorks or: [self isSqueak]) ifTrue: [^ Date newDay: days monthNumber: months year: years].
	self isPharo ifTrue: [ ^Date newDay: years month: months year: years ].
	self isGNU ifTrue: [^ Date newDay: days monthIndex: months year: years].
	self isVisualAge ifTrue: [^Date newDay: days monthIndex: months year: years].
	self isObjectStudio ifTrue: [^Date newDay: days imonth: months year: years].
	self error: 'not implemented'.
	^ self! !

!Dialect classSide methodsFor: 'general portability' stamp: ''!
newIdentitySet
	^self identitySetClass new.! !

!Dialect classSide methodsFor: 'dates' stamp: 'GuillermoPolito 4/16/2011 23:01'!
newTimeWithHours: hours minutes: minutes seconds: seconds milliseconds: milliseconds

	self isGNU ifTrue: [^ Time fromSeconds: hours * 60 * 60 + (minutes * 60) + seconds].
	(self isVisualWorks or: [self isSqueak or: [ self isPharo ]]) ifTrue: [^ Time fromSeconds: hours * 60 * 60 + (minutes * 60) + seconds].
	self isVisualAge
		ifTrue: [
			^ Time fromMilliseconds: ((hours * 60 * 60) + (minutes * 60) + seconds) * 1000 + milliseconds
		].
	self isObjectStudio ifTrue: [^Time hour: hours minute: minutes second: seconds millisecond: milliseconds].
	self error: 'Not implemented yet'.
	^ self! !

!Dialect classSide methodsFor: 'dates' stamp: ''!
newTimestampFromDate: aDate time: aTime

	self isVisualWorks ifTrue: [^self timestampClass fromDate: aDate andTime: aTime].
	self error: 'Not yet implemented'.! !

!Dialect classSide methodsFor: 'dates' stamp: 'GuillermoPolito 4/16/2011 22:57'!
newTimestampWithYears: years
	months: months
	days: days
	hours: hours
	minutes: minutes
	seconds: seconds
	milliseconds: millis
	offset: offset

	| date time ts |

	self isGNU
		ifTrue: [
			^ self timestampClass year: years
				month: months
				day: days
				hour: hours
				minute: minutes
				second: seconds
				offset: ((Dialect smalltalkAt: #Duration) fromSeconds: offset)
		].
	date := self newDateWithYears: years months: months days: days.
	time := self newTimeWithHours: hours minutes: minutes seconds: seconds milliseconds: millis.
	self isVisualWorks
		ifTrue: [^ (self timestampClass fromDate: date andTime: time) addMilliseconds: millis].
	(self isSqueak or: [ self isPharo])
		ifTrue: [ts := self timestampClass date: date time: time.
			ts time addSeconds: (millis / 1000) asInteger.
				^ts]. 
	self isVisualAge
		ifTrue: [^(Dialect smalltalkAt: #AbtTimestamp) date: date time: time].
	self isObjectStudio ifTrue: [^self timestampClass newDate: date time: time].
	self error: 'not implemented'.
	^ self! !

!Dialect classSide methodsFor: 'general portability' stamp: 'GuillermoPolito 4/16/2011 22:46'!
quitTheImage
	"Quit the entire image"

	self isGNU ifTrue: [^ self error: 'Not implemented yet'.].
	self isVisualWorks ifTrue: [^(self smalltalkAt: #ObjectMemory) quit].
	(self isSqueak or: [ self isPharo ]) 
		ifTrue: [^(self smalltalkAt: #SmalltalkImage) current snapshot: false andQuit: true].
	self isDolphin ifTrue: [^(self smalltalkAt: #SessionManager) current quit].
	self isVisualAge ifTrue: [^  self error: 'Not implemented yet'.].
	self isObjectStudio ifTrue: [^ self error: 'Not implemented yet'.].
	self error: 'not implemented yet'.
	^ self! !

!Dialect classSide methodsFor: 'dates' stamp: 'GuillermoPolito 4/16/2011 22:59'!
readFixedPointFrom: aString

	self isVisualWorks ifTrue: [^ (self smalltalkAt: #FixedPoint) readFrom: (ReadStream on: aString)].
	(self isDolphin or: [self isSqueak or: [self isPharo]]) ifTrue: [^ Number readFrom: (ReadStream on: aString , 's')].
	self isGNU
		ifTrue: [
			^ (Number readFrom: (ReadStream on: aString)) asScaledDecimal:
				aString size - (aString indexOf: $. ifAbsent: [aString size.])
		].
	self isVisualAge ifTrue: [ ^(self smalltalkAt: #Decimal) fromString: aString].
	self isObjectStudio ifTrue: [^(self smalltalkAt: #Decimal) newString: aString].
	self error: 'not implemented'.
	^ self! !

!Dialect classSide methodsFor: 'private' stamp: ''!
reset
	"self reset"
	lookedUpNames := nil.
	identitySetClass := nil.
	timestampClass := nil.
	dialectName := nil.! !

!Dialect classSide methodsFor: 'reflection' stamp: 'GuillermoPolito 4/16/2011 22:52'!
selectorsFor: aClass do: aBlock
	"Stupid dialects"
	self isVisualWorks ifTrue: [^aClass selectorsAndMethodsDo: [:key :value | aBlock value: key]].
	self isVisualAge ifTrue: [^ aClass methodDictionary keysDo: aBlock].
	(self isSqueak or: [ self isPharo ])ifTrue: [^aClass methodDict keysDo: aBlock].
	self error: 'Not yet implemented'.! !

!Dialect classSide methodsFor: 'numbers' stamp: ''!
singlePrecisionFloatClass
	^ Float! !

!Dialect classSide methodsFor: 'class lookup' stamp: ''!
smalltalkAssociationAt: aName

	^ self smalltalkAssociationAt: aName ifAbsent: [self error: 'element not found'.]! !

!Dialect classSide methodsFor: 'class lookup' stamp: ''!
smalltalkAssociationAt: aName ifAbsent: aBlock

	self isVisualWorks ifTrue: [^ aName asQualifiedReference].
	^ Smalltalk associationAt: aName asSymbol ifAbsent: aBlock! !

!Dialect classSide methodsFor: 'class lookup' stamp: ''!
smalltalkAt: aName

	^ self smalltalkAt: aName ifAbsent: [self error: 'element not found'].! !

!Dialect classSide methodsFor: 'class lookup' stamp: ''!
smalltalkAt: aName ifAbsent: aBlock
	"We may look these names up a lot, so cache them in a small, local dictionary"
	| cached value symbol |
	symbol := aName asSymbol.
	cached := self lookedUpNames at: symbol ifAbsent: [nil].
	cached isNil ifFalse: [
		(cached isBehavior and: [cached isObsolete]) ifFalse: [^cached]].
	value := self isVisualWorks
			ifTrue: [((symbol includes: $.)
				ifTrue: [symbol asStrictReference]
				ifFalse: [symbol asQualifiedReference]) valueOrDo: [^aBlock value]]
			ifFalse: [Smalltalk at: symbol ifAbsent: aBlock].
	self lookedUpNames
		at: symbol
		put: value.
	^value.! !

!Dialect classSide methodsFor: 'dates' stamp: ''!
supportsMillisecondsInTimeStamps

	self isGNU ifTrue: [^ false].
	self isVisualWorks ifTrue: [^ true].
	self isDolphin ifTrue: [^ true].
	self isVisualAge ifTrue: [^ true].
	self isObjectStudio ifTrue: [^true].
	self error: 'not yet implemented'.
	^ self! !

!Dialect classSide methodsFor: 'dates' stamp: 'GuillermoPolito 4/16/2011 22:45'!
supportsMillisecondsInTimes

	self isGNU ifTrue: [^ false].
	self isVisualWorks ifTrue: [^ false].
	self isSqueak ifTrue: [^false].
	self isPharo ifTrue: [^false].
	self isDolphin ifTrue: [^ true].
	self isVisualAge ifTrue: [^ true].
	self isObjectStudio ifTrue: [^true]. "Sort of. Supports hundredths of a second"
	^ self! !

!Dialect classSide methodsFor: 'dates' stamp: ''!
timeOffsetFromGMT

	self isGNU ifTrue: [Time timezoneBias / (60 * 60).].
	self isVisualWorks ifTrue: [^ (self smalltalkAt: #TimeZone) default secondsFromGMT / (60 * 60)].
	^ 0! !

!Dialect classSide methodsFor: 'dates' stamp: 'GuillermoPolito 4/16/2011 22:53'!
timestampClass

	timestampClass == nil ifFalse: [^ timestampClass].
	Dialect isGNU ifTrue: [^ timestampClass := self smalltalkAt: #DateTime].
	(Dialect isSqueak or: [Dialect isDolphin or: [ self isPharo ] ])
		ifTrue: [^ timestampClass := self smalltalkAt: #TimeStamp].
	Dialect isVisualWorks ifTrue: [^ timestampClass := self smalltalkAt: #Timestamp].
	Dialect isVisualAge ifTrue: [^ timestampClass := self smalltalkAt: #Timestamp].
	Dialect isObjectStudio ifTrue: [^timestampClass := self smalltalkAt: #Timestamp].
	self error: 'Not yet implemented'.
	^ self! !

!Dialect classSide methodsFor: 'dates' stamp: 'GuillermoPolito 4/16/2011 22:53'!
timestampNow

	Dialect isGNU ifTrue: [^ self timestampClass dateAndTimeNow].
	(Dialect isSqueak or: [ Dialect isPharo ]) ifTrue: [^ self timestampClass current].
	Dialect isVisualWorks ifTrue: [^ self timestampClass now].
	Dialect isDolphin ifTrue: [^ self timestampClass current].
	Dialect isVisualAge ifTrue: [^ self timestampClass now].
	Dialect isObjectStudio ifTrue: [^self timestampClass now].
	self error: 'Not yet implemented'.
	^ self! !

!Dialect classSide methodsFor: 'dates' stamp: ''!
timestampNowUTC

	Dialect isGNU ifTrue: [^ self timestampClass utcDateAndTimeNow].
	Dialect isVisualWorks ifTrue: [^ (self smalltalkAt: #Timestamp) fromSeconds: Time secondClock].
	Dialect isDolphin ifTrue: [self error: 'not supported'.].
	self error: 'Not yet implemented'.
	^ self! !

!Dialect classSide methodsFor: 'general portability' stamp: 'GuillermoPolito 4/16/2011 22:58'!
tokensBasedOn: aString in: stringToTokenize

	self isGNU ifTrue: [^ stringToTokenize subStrings: aString first].
	self isVisualWorks ifTrue: [^ stringToTokenize tokensBasedOn: aString first].
	(self isSqueak or: [self isPharo] ) ifTrue: [^ stringToTokenize findTokens: aString].
	self isDolphin ifTrue: [^ stringToTokenize subStrings: aString].
	self isVisualAge ifTrue: [^ (stringToTokenize subStrings: aString first)].
	self isObjectStudio ifTrue: [^(stringToTokenize asArrayOfSubstringsDelimiter: aString first)].
	self error: 'not implemented yet'.
	^ self! !

!Dialect classSide methodsFor: 'dates' stamp: ''!
totalSeconds

	self isGNU ifTrue: [^ Time utcSecondClock].
	self isVisualAge ifTrue: [^ (self smalltalkAt: #AbtTimestamp) now totalSeconds].
	^ Time totalSeconds! !

!Dialect classSide methodsFor: 'binding' stamp: ''!
unbindableClassNames

	self isVisualWorks ifTrue: [^ #()].
	^ #()! !

!Dialect classSide methodsFor: 'general portability' stamp: 'GuillermoPolito 4/16/2011 23:01'!
weakValueDictionaryClass

	Dialect isVisualWorks ifTrue: [^self smalltalkAt: #EphemeralValueDictionary].
	(Dialect isSqueak or: [Dialect isPharo]) ifTrue: [^self smalltalkAt: #WeakValueDictionary].
	self error: 'Not yet implemented for this dialect'.! !

Object subclass: #ElementBuilder
	instanceVariableNames: 'instance requiresPopulating key expression cursor fieldTranslations isExpired row query'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!ElementBuilder commentStamp: '<historical>' prior: 0!
This is the abstract superclass of builders. These assemble information, either primitive data or objects, from a database result set.

Subclasses must implement the following messages:
	building objects
		buildObjectFrom:
		findInstanceForRow:useProxy:
	selecting fields
		fieldsFromMyPerspective

Instance Variables:
	expression	<MappingExpression>	The expression we're mapping. e.g. if the query is reading people, this might be the expression corresponding to "each address", meaning that we build the address object related to the main Person instance by the given relationship. 
	fieldTranslations	<Array of: Integer>	 The translation of the field positions from where they are in the descriptor to where they are in the row we're reading. This is done so we can read the rows efficiently, by index, rather than doing lots of dictionary lookups by name. If we're doing a simple read, the translations will probably be a no-op, but if we read multiple objects, some of the fields will be in different positions than they are in our table definition.
	instance	<Object>	The thing we're constructing.
	isExpired	<Boolean>	If our instance is in cache, we use that instead. However, if the instance has expired, then we do something different (most likely force a refresh) than if it's still alive.
	key	<Object>	The key for this row. This is lazily computed, and "self" is used a special marker to indicate that it hasn't been computed yet.
	query	<AbstractReadQuery>	the query that we're building results for.
	requiresPopulating	<Boolean>	Do we need to populate the object. Will be false if the object was found in cache and hasn't expired.
	row	<Array>	The database results. May actually be a result set row of some sort rather than an array, depending on the dialect, but should always respond to indexing protocol.

!

ElementBuilder subclass: #DataElementBuilder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!DataElementBuilder commentStamp: '<historical>' prior: 0!
This builds raw data items rather than persistent objects with descriptors. Used if we do something like 
  aQuery retrieve: [:each | each address streetName].
giving us back simple data objects.
This makes building them quite simple.!

!DataElementBuilder methodsFor: 'building objects' stamp: ''!
buildObjectFrom: anArray
	self row: anArray.
	instance := self valueOf: expression.! !

!DataElementBuilder methodsFor: 'building objects' stamp: ''!
didValidationFail: validateResult

	"In other words, don't treat the result of the validation send to the object as a failure if the object was the boolean false, so returning self would normally be interpreted as the failure"
	validateResult == self instance ifTrue: [^false].

	^super didValidationFail: validateResult.! !

!DataElementBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsForSelectStatement
	^Array with: expression.! !

!DataElementBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsFromMyPerspective
	^expression mappedFields.! !

!DataElementBuilder methodsFor: 'building objects' stamp: ''!
findInstanceForRow: aRow proxyType: proxyType

	^self.! !

ElementBuilder subclass: #DictionaryBuilder
	instanceVariableNames: 'keyBuilder valueBuilder'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!DictionaryBuilder commentStamp: '<historical>' prior: 0!
This is a subclass of builders that builds dictionaries. Basically it needs to build both the key and the value and then assemble them into a dictionary.

Instance Variables:
	keyBuilder	<ElementBuilderr>	The builder for the key
	valueBuilder	<ElementBuilder>	The builder for the value.

!

!DictionaryBuilder methodsFor: 'building objects' stamp: ''!
buildObjectFrom: anArray
	"We just build an association out of the sub-builders "
	| |
	self row: anArray.! !

!DictionaryBuilder methodsFor: 'accessing' stamp: ''!
canCauseDuplicateRows
	^expression class == MappingExpression and: [expression mapping isToManyRelationship].! !

!DictionaryBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsForSelectStatement
	^#()! !

!DictionaryBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsFromMyPerspective
	^#()! !

!DictionaryBuilder methodsFor: 'building objects' stamp: ''!
findInstanceForRow: aRow proxyType: proxyType

	instance := Association new.! !

!DictionaryBuilder methodsFor: 'accessing' stamp: ''!
instanceValue
	^self instance value! !

!DictionaryBuilder methodsFor: 'building objects' stamp: ''!
knitResultIn: aCursoredStream
	"Connect up our built object with any other objects that reference it. Used if we retrieve more than one thing in the same query"
	| relatedBuilder |
	instance key: keyBuilder instance.
	instance value: valueBuilder instance.

	"Code from ObjectBuilder>>knitResultIn:"
	expression canKnit ifFalse: [^self].
	relatedBuilder := aCursoredStream objectBuilderFor: expression base.
	(relatedBuilder isNil or: [relatedBuilder instance isNil]) 
		ifFalse: 
			[expression mapping knitFromBuilder: relatedBuilder toBuilder: self]! !

!DictionaryBuilder methodsFor: 'copying' stamp: ''!
postCopy! !

!DictionaryBuilder methodsFor: 'copying' stamp: ''!
reconcileWithBuilders: oldCollection andCopy: newCollection
	"Fix up references to our internal sub-builders"
	| keyIndex valueIndex |
	keyIndex := oldCollection indexOf: keyBuilder.
	keyBuilder := newCollection at: keyIndex.
	valueIndex := oldCollection indexOf: valueBuilder.
	valueBuilder := newCollection at: valueIndex.! !

!DictionaryBuilder methodsFor: 'initialize-release' stamp: ''!
setup
	"Do any initialization once we've got our parameters"
	| base collectionType |
	base := query baseExpression.
	collectionType := expression == base ifTrue: [query collectionType] ifFalse: [expression mapping query collectionType].
	keyBuilder := ElementBuilder for: (collectionType keyExpression asGlorpExpressionOn: expression) in: query.
	valueBuilder := ElementBuilder basicFor: (collectionType valueExpression asGlorpExpressionOn: expression) in: query.! !

!DictionaryBuilder methodsFor: 'accessing' stamp: ''!
subBuilders
	^Array with: keyBuilder with: valueBuilder.! !

!DictionaryBuilder methodsFor: 'accessing' stamp: ''!
withAllSubBuilders
	^Array with: self with: keyBuilder with: valueBuilder.! !

!ElementBuilder classSide methodsFor: 'private' stamp: ''!
basicClassFor: anExpression in: aQuery

	^(anExpression hasDescriptor and: [anExpression mapsSimpleObject not])
		ifTrue: [ObjectBuilder] 
		ifFalse: [DataElementBuilder].! !

!ElementBuilder classSide methodsFor: 'instance creation' stamp: ''!
basicFor: anExpression in: aQuery

	^(self basicClassFor: anExpression in: aQuery) new
		expression: anExpression;
		query: aQuery;
		setup.! !

!ElementBuilder classSide methodsFor: 'private' stamp: ''!
classFor: anExpression in: aQuery
	
	((aQuery returnsDictionary and: [anExpression == aQuery baseExpression])
		or: [anExpression returnsDictionary])
			ifTrue: [^DictionaryBuilder].
	^self basicClassFor: anExpression in: aQuery.! !

!ElementBuilder classSide methodsFor: 'instance creation' stamp: ''!
for: anExpression in: aQuery

	^(self classFor: anExpression in: aQuery) new
		expression: anExpression;
		query: aQuery;
		setup.! !

!ElementBuilder classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!ElementBuilder methodsFor: 'building objects' stamp: ''!
buildObjectFrom: anArray 
	self subclassResponsibility.! !

!ElementBuilder methodsFor: 'testing' stamp: ''!
buildsObjects
	^false.! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
canCauseDuplicateRows
	^false.! !

!ElementBuilder methodsFor: 'copying' stamp: ''!
copy
	^self shallowCopy postCopy.! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
cursor
	^cursor! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
cursor: aGlorpCursoredStream
	cursor := aGlorpCursoredStream! !

!ElementBuilder methodsFor: 'building objects' stamp: ''!
didValidationFail: validateResult
	^validateResult == false.! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
expression
	^expression! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
expression: anExpression
	expression := anExpression.! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
fieldTranslations
	^fieldTranslations! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
fieldTranslations: aDictionary 
	fieldTranslations := aDictionary.! !

!ElementBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsForSelectStatement

	^self subclassResponsibility.! !

!ElementBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsFromMyPerspective
	self subclassResponsibility.! !

!ElementBuilder methodsFor: 'building objects' stamp: ''!
findInstanceForRow: aRow proxyType: proxyType
	self subclassResponsibility.! !

!ElementBuilder methodsFor: 'executing' stamp: ''!
hasFieldTranslations
	^self fieldTranslations notNil! !

!ElementBuilder methodsFor: 'initializing' stamp: ''!
initialize! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
instance
	^instance! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
instanceValue
	^instance! !

!ElementBuilder methodsFor: 'building objects' stamp: ''!
knitResultIn: aSimpleQuery
	"Connect up our built object with any other builders that use the same thing"

	^self.! !

!ElementBuilder methodsFor: 'copying' stamp: ''!
postCopy
	instance := nil.
	row := nil.
	requiresPopulating := false.
	isExpired := false.! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
query
	^query! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
query: aQuery 
	query := aQuery! !

!ElementBuilder methodsFor: 'copying' stamp: ''!
reconcileWithBuilders: oldCollection andCopy: newCollection! !

!ElementBuilder methodsFor: 'building objects' stamp: ''!
registerObjectInUnitOfWork
	"If there is a current unit of work, then we must register in it, after population because that way the state is already in place. The nil checks are mostly for safety during unit tests, as those conditions should never occur in real use"
	query isNil ifTrue: [^self].
	query session isNil ifTrue: [^self].
	query session register: instance.! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
requiresDistinct

	^expression requiresDistinct.! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
requiresPopulating
	^true! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
returnValueIn: aCursoredStream
	^self instance! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
row
	^row! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
row: anArray
	"Since nil is a possible key value, use self as a special marker to indicate we haven't found the key yet"
	row == anArray ifFalse: [
		key := self].
	row := anArray.! !

!ElementBuilder methodsFor: 'building objects' stamp: ''!
sendPostFetchValidateIn: aSession OnFailureDo: aBlock
	| validateResult |
	validateResult := aSession sendPostFetchValidateEventTo: self instance.
	(self didValidationFail: validateResult) ifTrue: aBlock.! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
session
	^expression descriptor session.! !

!ElementBuilder methodsFor: 'initializing' stamp: ''!
setup
	"Do any initialization required once we've got our parameters"
	^self.! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
subBuilders
	^#().! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
system
	^self session system.! !

!ElementBuilder methodsFor: 'translating fields' stamp: ''!
translateFieldPosition: aDatabaseField 
	fieldTranslations isNil ifTrue: [^aDatabaseField position].
	^fieldTranslations at: aDatabaseField.! !

!ElementBuilder methodsFor: 'translating fields' stamp: ''!
valueOf: anExpression
	^expression valueInBuilder: self.! !

!ElementBuilder methodsFor: 'translating fields' stamp: ''!
valueOfField: aField
	"aField is either a database field, or a constant expression containing a non-varying value that isn't derived from the row"
	(aField class == ConstantExpression) ifTrue: [^aField value].
	^self row atIndex: (self translateFieldPosition: aField).! !

!ElementBuilder methodsFor: 'translating fields' stamp: ''!
valueOfField: aField in: aRow
	"Since the elementBuilder now holds the row, #valueOfField: is preferred protocol, but some things (e.g. ad hoc mapping blocks) might still be using this, so left for compatibility"
	(aField class == ConstantExpression) ifTrue: [^aField value].
	^aRow atIndex: (self translateFieldPosition: aField).! !

!ElementBuilder methodsFor: 'accessing' stamp: ''!
withAllSubBuilders
	^Array with: self.! !

ElementBuilder subclass: #ObjectBuilder
	instanceVariableNames: 'queryCache'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!ObjectBuilder commentStamp: '<historical>' prior: 0!
This builds full-blown persistent objects with descriptors. This is the most common type of builder.!

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
buildObjectFrom: anArray
	self row: anArray.
	self requiresPopulating ifTrue: [self populateInstance].
	(self system shouldResetExpiryTimeForObjectOfClass: instance class readFrom: self query)
		ifTrue: [self session markAsCurrentOfClass: instance class key: self key].! !

!ObjectBuilder methodsFor: 'testing' stamp: ''!
buildsObjects
	^true.! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
cachedQueryFor: aMapping
	^self queryCache at: aMapping ifAbsent: [nil].! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
cachedQueryFor: aMapping is: aQuery
	"Save a query which we will use when constructing filtered queries"
	self queryCache at: aMapping put: aQuery.! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
canBuild
	"If we have a regular object with a nil primary key, or if we have an embedded object whose values are all nil, we can't build anything (probably due to an outer join)"
	^self descriptor mapsPrimaryKeys
		ifTrue: [self key notNil]
		ifFalse: [true].! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
canCache
	| keyFields |
	keyFields := self descriptor keyFields.
	^keyFields notNil and: [keyFields notEmpty].! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
canCauseDuplicateRows
	^expression class == MappingExpression and: [expression mapping isToManyRelationship].! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
command
	^cursor command.! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
createObjectOfClass: aClass
	"Try to create objects using new, so initialization can run. But if we can't do it, then just use basicNew."
	^[aClass new] on: Error do: [:ex | ex return: aClass basicNew].! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
descriptor
	
	^expression descriptor.! !

!ObjectBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsForSelectStatement
	^self fieldsFromThePerspectiveOfTheMainSelect: (self fieldsFromMyPerspective).! !

!ObjectBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsFromMyPerspective
	^query returnProxies 
		ifTrue: [self descriptor table primaryKeyFields]
		ifFalse: [self descriptor fieldsForSelectStatement]! !

!ObjectBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsFromThePerspectiveOfTheMainSelect: aCollection
	^expression translateFields: aCollection.! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
findInstanceForRow: aRow proxyType: proxyType
	instance := nil.
	self row: aRow.
	self canBuild ifFalse: [^self].
	(self shouldProxyInstanceFor: proxyType)
		ifTrue: [^self newProxyWithNewQueryForClass: proxyType].
	self lookupCachedObject.
	self instanceExistsButIsDeleted
		ifTrue:
			[instance := nil.
			^self].
	instance isNil
		ifFalse:
			[requiresPopulating := requiresPopulating | query shouldRefresh.
			^self].
	requiresPopulating := true.
	instance := self createObjectOfClass: (expression descriptor
		describedConcreteClassFor: self row
		withBuilder: self).
	self canCache ifTrue: [self session cacheAt: self key put: instance].! !

!ObjectBuilder methodsFor: 'initializing' stamp: ''!
initialize

	requiresPopulating := false.
	isExpired := false.! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
instance: anObject
	instance := anObject! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
instanceDescriptor
	"Return the descriptor for the instance we're going to build. Note that we have to look it up by instance in order to find subclass. "
	| basicDescriptor |
	instance isNil ifTrue: [^nil].
	basicDescriptor := self system descriptorFor: instance.
	basicDescriptor isNil ifFalse: [^basicDescriptor].
	^expression descriptor.! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
instanceExistsButIsDeleted
	instance isNil ifTrue: [^false].
	instance := self session filterDeletionFrom: instance.
	^instance isNil.! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
key
	^self primaryKey.! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
knitResultIn: aCursoredStream 
	"Connect up our built object with any other objects that reference it. Used if we retrieve more than one thing in the same query"

	| relatedBuilder |
	expression canKnit ifFalse: [^self].
	relatedBuilder := aCursoredStream objectBuilderFor: expression base.
	(relatedBuilder isNil or: [relatedBuilder instance isNil]) 
		ifFalse: 
			[expression mapping knitFromBuilder: relatedBuilder toBuilder: self]! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
lookupCachedObject
	| resultClass |
	self canBuild ifFalse: [^self].
	self canCache
		ifTrue:
			[
			resultClass := expression descriptor describedClass.
			(self session hasExpired: resultClass key: self key)
				ifTrue:
					[instance := self session expiredInstanceOf: resultClass key: self key.
					requiresPopulating := true.
					isExpired := true]
				ifFalse:
					[instance := self session cacheLookupForClass: resultClass key: self key.
					requiresPopulating := instance isNil]].! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
newProxyWithNewQueryForClass: aClass
	| parameters |
	parameters := IdentityDictionary new.
	self descriptor primaryTable primaryKeyFields
		do:
			[:eachField | parameters at: eachField put: (self valueOfField: eachField in: row)].
	instance := aClass new
		session: self session;
		parameters: parameters;
		query: self queryForNewProxy.
	self session addToReaper: instance.
	^self.! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
populateInstance
	| tracker |
	key isNil ifTrue: [^self].
	self instanceDescriptor populateObject: instance inBuilder: self.
	tracker := self session modificationTracker.
	tracker isNil ifFalse: [tracker track: instance].! !

!ObjectBuilder methodsFor: 'copying' stamp: ''!
postCopy
	super postCopy.
	queryCache := nil.! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
primaryKey
	"We use self as a special guard value to indicate that the value hasn't changed"
	key == self ifFalse: [^key].
	self canCache ifFalse: [^nil].
	key := self descriptor keyFields
		collect: [:each | self valueOfField: each].
	key size = 1 ifTrue: [key := key first].
	^key.! !

!ObjectBuilder methodsFor: 'printing' stamp: ''!
printOn: aStream
	super printOn: aStream.
	aStream nextPut: $(.
	expression printOn: aStream.
	aStream nextPut: $).! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
queryCache
	queryCache isNil ifTrue: [queryCache := IdentityDictionary new: 5].
	^queryCache.! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
queryForNewProxy
	"Create a proxy with a primary key query in which the parameters are the primary key fields. We can't just reuse a mapping's query, it has to be a new primary key one, because the mapping's query has just run and been told to return proxies. This kind of stinks"
	| whereExpression |
	whereExpression := Join new.
	self descriptor primaryTable primaryKeyFields do: [:eachField | 
		whereExpression addSource: eachField target: eachField].
	^Query
		returningOneOf: query resultClass 
		where: whereExpression! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
requiresPopulating
	^requiresPopulating and: [self returnProxies not].! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
requiresPopulating: aBoolean
	requiresPopulating := aBoolean! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
returnProxies
	^query returnProxies.! !

!ObjectBuilder methodsFor: 'accessing' stamp: ''!
returnValueIn: aCursoredStream
	"If we're a to-many relationship, what we retrieve: is the collection containing us, not the individual items"
	| relatedBuilder |
	expression canKnit ifFalse: [^self instance].
	expression mapping isToManyRelationship ifFalse: [^self instance].
	relatedBuilder := aCursoredStream elementBuilderFor: expression base.
	(relatedBuilder isNil or: [relatedBuilder instance isNil]) ifTrue: [^self instance].
	^expression mapping getValueFrom: relatedBuilder instance.! !

!ObjectBuilder methodsFor: 'building objects' stamp: ''!
shouldProxyInstanceFor: proxyType
	^proxyType notNil.! !

ElementBuilder subclass: #SelectingElementBuilder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!SelectingElementBuilder commentStamp: '<historical>' prior: 0!
This is a special subclass of ElementBuilder used when we execute a query that reads multiple different things, but in any row, we expect only one of them to have a value. This is useful if we're doing inheritance queries where each concrete subclass is in a separate table. We can join the queries together, but we expect each row to contain only one particular subclass instance.!

!SelectingElementBuilder methodsFor: 'building objects' stamp: ''!
buildObjectFrom: anArray
	^self.! !

!SelectingElementBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsForSelectStatement
	^#().! !

!SelectingElementBuilder methodsFor: 'selecting fields' stamp: ''!
fieldsFromMyPerspective
	^#().! !

!SelectingElementBuilder methodsFor: 'building objects' stamp: ''!
findInstanceForRow: aRow proxyType: proxyType
	^self.! !

!SelectingElementBuilder methodsFor: 'building objects' stamp: ''!
returnValueIn: aCursoredStream
	| builder |
	builder := aCursoredStream privateGetBuilders
		detect:
			[:each | 
			each instance notNil
				and: [each expression ~~ aCursoredStream command query baseExpression]]
		ifNone:
			[self error: 'Internal error in computing horizontal inheritance query.'].
	^builder instance.! !

Object subclass: #FieldValueWrapper
	instanceVariableNames: 'contents hasValue containedBy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-UnitOfWork'!

!FieldValueWrapper commentStamp: '<historical>' prior: 0!
When we are creating our rowmap, we need to be able to unify values. That is, we express the constraint that field A in table T1 must be equal to field B in table T2. And we have to be able to do this whether or not we have yet assigned a value to either of these fields, and whether or not other equality constraints have already been imposed on one or the other of these fields. We do this by holding the values in a wrapper. When we unify them, we replace one wrapper with the other, so that both rows now refer to the same value. In order to efficiently find where things are contained, we hold a collection of the containing rows.

Instance Variables:
	containedBy	<(Association key: DatabaseField value: (IdentitySet of: DatabaseRow)) | (IdentityDictionary of: Same)>	The rows that contain this wrapper. Using a trick similar to #dependents, we optimize the containedByCollection to be an association if we are only contained in one row (the most common case). Otherwise we have a dictionary. The containing rows are indexed by field, because when we're updating the row to replace the wrapper, we need to know which field to replace. Also, a single row could conceivably contain the same wrapper in two different fields, and both need to be updated.
	contents	<Object>	The value we contain
	hasValue	<Boolean>	Has our value been set or not. We can't just test for nil in contents, because nil is a valid value.

!

!FieldValueWrapper classSide methodsFor: 'public' stamp: ''!
new

	^super new initialize.! !

!FieldValueWrapper methodsFor: 'accessing' stamp: ''!
allInnerWrappers
	^#().! !

!FieldValueWrapper methodsFor: 'accessing' stamp: ''!
basicContents
	"Return the un-transformed contents"
	^self contents.! !

!FieldValueWrapper methodsFor: 'public' stamp: ''!
containedBy

	^containedBy.! !

!FieldValueWrapper methodsFor: 'unifying' stamp: ''!
containsWrapper: aFieldValueWrapper
	^self == aFieldValueWrapper.! !

!FieldValueWrapper methodsFor: 'public' stamp: ''!
contents

	^contents.! !

!FieldValueWrapper methodsFor: 'public' stamp: ''!
contents: anObject

	(hasValue and: [contents ~= anObject]) ifTrue: [self error: 'Inconsistent values in field'].
	self privateSetContents: anObject.! !

!FieldValueWrapper methodsFor: 'public' stamp: ''!
hasValue

	^hasValue.! !

!FieldValueWrapper methodsFor: 'public' stamp: ''!
initialize

	hasValue := false.! !

!FieldValueWrapper methodsFor: 'accessing' stamp: ''!
innerWrapper
	^self.! !

!FieldValueWrapper methodsFor: 'containing' stamp: ''!
isNowContainedBy: aRow and: aField
	"Optimize this for memory consumption with the common case of one container. If we're contained by nothing, make it a single association. Otherwise, use a dictionary"
	| shortCircuit thisFieldsEntries |
	shortCircuit := false.
	containedBy == nil ifTrue: [containedBy := aField -> ((IdentitySet new: 2) add: aRow; yourself). ^self].
	containedBy class == Association 
		ifTrue: [
		containedBy key == aField
			ifTrue: [thisFieldsEntries := containedBy value]
			ifFalse: [ | newContainedBy |
				newContainedBy := IdentityDictionary new. 
				newContainedBy at: containedBy key put: containedBy value.
				containedBy := newContainedBy]].
	thisFieldsEntries isNil ifTrue: [
		thisFieldsEntries := containedBy at: aField ifAbsentPut: [shortCircuit := true. (IdentitySet new: 2) add: aRow; yourself]].
	shortCircuit ifTrue: [^self].
	(thisFieldsEntries includes: aRow) ifFalse: [thisFieldsEntries add: aRow].! !

!FieldValueWrapper methodsFor: 'public' stamp: ''!
printOn: aStream

	aStream 
		nextPutAll: '<<'.
	self hasValue ifTrue: [aStream print: self contents].
	aStream
		nextPutAll: '>>'.! !

!FieldValueWrapper methodsFor: 'public' stamp: ''!
privateSetContents: anObject
	contents := anObject.
	hasValue := true.! !

!FieldValueWrapper methodsFor: 'unifying' stamp: ''!
replaceWith: aFieldValueWrapper
	containedBy isNil ifTrue: [^nil].
	containedBy
		keysAndValuesDo:
			[:eachField :eachListOfRows | eachListOfRows do: [:eachRow | eachRow wrapperAt: eachField put: aFieldValueWrapper]].! !

!FieldValueWrapper methodsFor: 'unifying' stamp: ''!
unifyWith: targetWrapper
	(self containsWrapper: targetWrapper) ifTrue: [^self].
	(targetWrapper containsWrapper: self) ifTrue: [^self].
	targetWrapper hasValue ifFalse: [^targetWrapper replaceWith: self].
	self hasValue ifFalse: [^self replaceWith: targetWrapper].
	self contents = targetWrapper contents
		ifFalse: [self error: 'Conflicting values in rows'].
	targetWrapper replaceWith: self.! !

FieldValueWrapper subclass: #TransformingFieldValueWrapper
	instanceVariableNames: 'forwardTransformation inverseTransformation innerWrapper'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-UnitOfWork'!

!TransformingFieldValueWrapper commentStamp: '<historical>' prior: 0!
This wraps another wrapper and transforms its value.

Instance Variables:
	forwardTransformation	<Block>	What to do to convert our value going out.
	innerWrapper	<FieldValueWrapper>	The wrapper that really has the value.
	inverseTransformation	<Block>	What to do to convert our value going in.

!

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
allInnerWrappers
	^innerWrapper allInnerWrappers, (Array with: innerWrapper).! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
basicContents
	"Return the un-transformed contents"
	^innerWrapper contents.! !

!TransformingFieldValueWrapper methodsFor: 'unifying' stamp: ''!
containsWrapper: aFieldValueWrapper
	^self == aFieldValueWrapper or: [innerWrapper == aFieldValueWrapper
		or: [aFieldValueWrapper innerWrapper == self
			or: [aFieldValueWrapper innerWrapper == self innerWrapper]]].! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
contents

	^forwardTransformation value: innerWrapper contents.! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
contents: anObject

	innerWrapper contents: (inverseTransformation value: anObject).! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
forwardTransformation
	^forwardTransformation! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
forwardTransformation: aBlock
	forwardTransformation := aBlock! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
hasValue

	^innerWrapper hasValue.! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
innerWrapper
	^innerWrapper! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
innerWrapper: aFieldValueWrapper
	innerWrapper := aFieldValueWrapper! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
inverseTransformation
	^inverseTransformation! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
inverseTransformation: aBlock
	inverseTransformation := aBlock! !

!TransformingFieldValueWrapper methodsFor: 'accessing' stamp: ''!
privateSetContents: anObject
	innerWrapper privateSetContents: (inverseTransformation value: anObject).! !

!TransformingFieldValueWrapper methodsFor: 'unifying' stamp: ''!
replaceWith: aFieldValueWrapper
	innerWrapper replaceWith: aFieldValueWrapper.! !

Object subclass: #FixedSizeQueue
	instanceVariableNames: 'maximumSize items'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!FixedSizeQueue commentStamp: '<historical>' prior: 0!
This is a fixed size queue of objects. It's intended for keeping around a fixed number of references to objects in a weak dictionary. As such its API is rather limited (one method), and it's write-only.

Instance Variables:
	items	<OrderedCollection>	The items in the queue
	maximumSize	<Integer>	How many items we're allowed!

!FixedSizeQueue classSide methodsFor: 'instance creation' stamp: ''!
maximumSize: anInteger
	^self basicNew maximumSize: anInteger.! !

!FixedSizeQueue classSide methodsFor: 'instance creation' stamp: ''!
new
	self error: 'must supply a size'.! !

!FixedSizeQueue classSide methodsFor: 'instance creation' stamp: ''!
new: anInteger
	^self maximumSize: anInteger.! !

!FixedSizeQueue methodsFor: 'api' stamp: ''!
add: anObject
	items add: anObject.
	items size > maximumSize ifTrue: [items removeFirst].! !

!FixedSizeQueue methodsFor: 'accessing' stamp: ''!
maximumSize
	^maximumSize! !

!FixedSizeQueue methodsFor: 'accessing' stamp: ''!
maximumSize: anInteger
	maximumSize := anInteger.
	items := OrderedCollection new: maximumSize + 1.! !

!FixedSizeQueue methodsFor: 'printing' stamp: ''!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: '('.
	aStream nextPutAll: items size printString.
	aStream nextPut: $/.
	aStream nextPutAll: maximumSize printString.
	aStream nextPutAll: ')'.! !

Object subclass: #GlorpAttributeModel
	instanceVariableNames: 'name type collectionType keyType classModel attributeIndex useDirectAccess setSelector canRead canWrite'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!GlorpAttributeModel commentStamp: '<historical>' prior: 0!
This is the metadata for an individual instance variable. It will be held in a GlorpClassModel, and used for accessing the variable when building or writing out objects.

Instance Variables:
	attributeIndex	<Integer>	What is the index in the class of the instance variable. Used if we are using instVarAt: to get values.
	classModel	<GlorpClassModel>	The class model that contains us.
	collectionType	<Class>	If we are an attribute that holds a collection, what is the class of the collection.
	keyType	<Class>	If we represent a dictionary attribute, what is the type of our key (type will be used for the type of the value).
	name	<Symbol>	The name of the attribute/instance variable.
	setSelector	<Symbol>	If we are using message access to get/set values, what is the selector to use for setting. Defaults to name, ':'.
	type	<Class>	The class containing this instance variable
	useDirectAccess	<Boolean>	If true (the default) we use instVarAt:/instVarAt:put: to access the attribute values.

!

!GlorpAttributeModel classSide methodsFor: 'instance creation' stamp: ''!
named: aSymbol
	^self new name: aSymbol.! !

!GlorpAttributeModel classSide methodsFor: 'instance creation' stamp: ''!
named: aSymbol collection: aCollectionClass of: aClass
	^(self new)
		name: aSymbol;
		type: aClass;
		collectionType: aCollectionClass.! !

!GlorpAttributeModel classSide methodsFor: 'instance creation' stamp: ''!
named: aSymbol collectionOf: aClass
	^self named: aSymbol collection: OrderedCollection of: aClass.! !

!GlorpAttributeModel classSide methodsFor: 'instance creation' stamp: ''!
named: aSymbol dictionary: dictionaryClass from: keyClass to: valueClass
	^(self new)
		name: aSymbol;
		type: valueClass;
		keyType: keyClass;
		collectionType: dictionaryClass.! !

!GlorpAttributeModel classSide methodsFor: 'instance creation' stamp: ''!
named: aSymbol dictionaryFrom: keyClass to: valueClass
	^self named: aSymbol dictionary: Dictionary from: keyClass to: valueClass.! !

!GlorpAttributeModel classSide methodsFor: 'instance creation' stamp: ''!
named: aSymbol type: aClass
	^self new name: aSymbol; type: aClass.! !

!GlorpAttributeModel classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize.! !

!GlorpAttributeModel methodsFor: 'knitting' stamp: ''!
add: anObject to: aCollection in: ourObject
	"If this is an array we can't just add to it, we must concatenate and re-set the value"
	| newCollection |
	aCollection class == Array 
		ifTrue: [	
			newCollection := aCollection, (Array with: anObject).
			self setValueIn: ourObject to: newCollection.
			^self].
	(anObject class == Association and: [anObject key == nil])
		ifFalse: [aCollection add: anObject].! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
beForPseudoVariable
	"Make this be a pseudo-variable - one which is neither read nor written, and doesn't have a corresponding instance variable, but is only used for queries."
	canRead := false.
	canWrite := false.! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
canRead
	^canRead! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
canRead: aBoolean
	canRead := aBoolean! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
canWrite
	^canWrite! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
canWrite: aBoolean
	canWrite := aBoolean! !

!GlorpAttributeModel methodsFor: 'get/set' stamp: ''!
checkTargetType: anObject
	"Check that anObject's type corresponds to something we know how to handle for this attribute. We expect this to be a real object, not a collection."
	self checkTargetType: anObject against: self type.! !

!GlorpAttributeModel methodsFor: 'get/set' stamp: ''!
checkTargetType: anObject against: aClass
	"Check that anObject's type corresponds to aClass, which we expect as our values. We expect this to be a real object, not a collection."
	| targetModel |
	anObject yourSelf isNil ifTrue: [^self].
	aClass isNil ifTrue: [self error: 'No type to check against in class model'].
	anObject glorpIsCollection ifTrue: [(GlorpInvalidTypeError new object: anObject; expectedType: self collectionType) signal].
	targetModel := self system classModelFor: aClass.
	targetModel notNil ifTrue: [
		(targetModel matches: anObject)
			ifFalse: [(GlorpInvalidTypeError new object: anObject; expectedType: aClass) signal]].! !

!GlorpAttributeModel methodsFor: 'get/set' stamp: ''!
checkTargetTypeForKey: aKey value: anObject
	"Check that anObject's type corresponds to something we know how to handle for this attribute. We expect this to be a real object, not a collection. This is the version used for checking a dictionary, where we check both key and value, but only if they correspond to mapped classes."
	(self system classModelFor: self keyType) isNil 
		ifFalse: [self checkTargetType: aKey against: self keyType].
	(self system classModelFor: self type) isNil
		ifFalse: [self checkTargetType: anObject against: type].! !

!GlorpAttributeModel methodsFor: 'get/set' stamp: ''!
checkType: anObject
	(self classModel matches: anObject)
		ifFalse: [self error: 'Wrong type of object'].! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
classModel
	^classModel! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
classModel: aGlorpClassModel
	classModel := aGlorpClassModel! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
collectionType
	^collectionType! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
collectionType: aClass
	collectionType := aClass! !

!GlorpAttributeModel methodsFor: 'copying' stamp: ''!
copyFrom: anAttributeModel
	"Copy over the attributes if it defines them and we don't"
	anAttributeModel isNil ifTrue: [^self].
	self type isNil ifTrue: [self type: anAttributeModel type].
	self collectionType isNil
		ifTrue: [self collectionType: anAttributeModel collectionType].
	self privateUseDirectAccess isNil
		ifTrue: [useDirectAccess := anAttributeModel privateUseDirectAccess].
	canRead := anAttributeModel canRead.
	canWrite := anAttributeModel canWrite.! !

!GlorpAttributeModel methodsFor: 'private' stamp: ''!
directGetValueFrom: anObject 
	| index |
	index := self instVarIndexIn: anObject.
	index = 0 ifTrue: [self raiseInvalidAttributeError].
	^anObject instVarAt: index! !

!GlorpAttributeModel methodsFor: 'private' stamp: ''!
directSetValueIn: anObject to: aValue 
	| index |
	index := self instVarIndexIn: anObject.
	index = 0 ifTrue: [self raiseInvalidAttributeError].
	^anObject instVarAt: index put: aValue! !

!GlorpAttributeModel methodsFor: 'get/set' stamp: ''!
getValueFrom: anObject
	| real |
	real := anObject yourSelf.
	self checkType: real.
	self useDirectAccess ifTrue: [^self directGetValueFrom: real].
	^real perform: self name.! !

!GlorpAttributeModel methodsFor: 'initialization' stamp: ''!
initialize
	canRead := true.
	canWrite := true.! !

!GlorpAttributeModel methodsFor: 'private' stamp: ''!
instVarIndex
	| soughtName |
	attributeIndex == nil ifFalse: [^attributeIndex].
	soughtName := Dialect instVarNameFor: name.
	^attributeIndex := self classModel describedClass allInstVarNames 
				indexOf: soughtName! !

!GlorpAttributeModel methodsFor: 'private' stamp: ''!
instVarIndexIn: anObject 
	^self instVarIndex! !

!GlorpAttributeModel methodsFor: 'validation' stamp: ''!
isCollectionAttribute
	^collectionType notNil.! !

!GlorpAttributeModel methodsFor: 'mapping' stamp: ''!
isDictionaryAttribute
	^collectionType isBehavior and: [collectionType includesBehavior: Dictionary].! !

!GlorpAttributeModel methodsFor: 'testing' stamp: ''!
isForPseudoVariable
	"Return true if this is a pseudo-variable - one which is neither read nor written, and doesn't have a corresponding instance variable, but is only used for queries."
	^(canRead not & canWrite not) or: [self isValid not].! !

!GlorpAttributeModel methodsFor: 'validation' stamp: ''!
isReadable
	self canRead ifFalse: [^false].
	^self useDirectAccess 
		ifTrue: [self instVarIndex > 0]
		ifFalse: [self classModel describedClass canUnderstand: self name]! !

!GlorpAttributeModel methodsFor: 'private' stamp: ''!
isValid
	^name notNil and: [classModel notNil].! !

!GlorpAttributeModel methodsFor: 'validation' stamp: ''!
isWritable
	self canWrite ifFalse: [^false].
	^self useDirectAccess 
		ifTrue: [self instVarIndex > 0]
		ifFalse: [self classModel describedClass canUnderstand: self setSelector]! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
keyType
	^keyType! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
keyType: aClass
	keyType := aClass! !

!GlorpAttributeModel methodsFor: 'knitting' stamp: ''!
knit: ourObject to: anotherObject 
	"Set up the relationship from our object to another one, indicated by our mapping."

	self isCollectionAttribute 
		ifTrue: 
			[| collection |
			collection := self getValueFrom: ourObject.
			(collection isGlorpProxy and: [collection isInstantiated not]) 
				ifTrue: 
					[collection := self newCollection.
					self setValueIn: ourObject to: collection.
					self 
						add: anotherObject
						to: collection
						in: ourObject.
					^self].
			(collection includes: anotherObject) 
				ifFalse: 
					[self 
						add: anotherObject
						to: collection
						in: ourObject]]
		ifFalse: [self setValueIn: ourObject to: anotherObject]! !

!GlorpAttributeModel methodsFor: 'mapping' stamp: ''!
loopOver: mightBeACollection doing: aBlock
	"It is possible that the value for a dictionary attribute is nil. If so, do not map it back"
	| index |
	self isCollectionAttribute ifFalse: [self checkTargetType: mightBeACollection. ^aBlock value: 1 value: mightBeACollection].
	mightBeACollection isNil ifTrue: [^nil].
	self isDictionaryAttribute ifTrue: [^mightBeACollection keysAndValuesDo: [:key :value |
		self checkTargetTypeForKey: key value: value. 
		value ifNotNil: [aBlock value: key value: value]]].
	index := 1.
	mightBeACollection glorpIsCollection ifFalse: [
		(GlorpInvalidTypeError new object: mightBeACollection; expectedType: self collectionType) signal].
	mightBeACollection
		do:
			[:each |
			| object | 
			object := each glorpRealObject. 
			self checkTargetType: object.
			aBlock value: index value: object.
			index := index + 1].! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
name
	^name isNil ifTrue: [''] ifFalse: [name]! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
name: aSymbol
	| symbol |
	symbol := aSymbol isNil ifTrue: [aSymbol] ifFalse: [aSymbol asSymbol].
	name := symbol.! !

!GlorpAttributeModel methodsFor: 'private' stamp: ''!
newCollection
	^collectionType isBehavior ifTrue: [collectionType new] ifFalse: [collectionType copyEmpty].! !

!GlorpAttributeModel methodsFor: 'printing' stamp: ''!
printOn: aStream
	aStream nextPutAll: self classModel printString, '>>', self name.! !

!GlorpAttributeModel methodsFor: 'private' stamp: ''!
privateUseDirectAccess
	^useDirectAccess.! !

!GlorpAttributeModel methodsFor: 'private' stamp: ''!
raiseInvalidAttributeError
	self error: 'Invalid attribute: ', classModel printString, '>>', name printString.! !

!GlorpAttributeModel methodsFor: 'private' stamp: ''!
setSelector
	setSelector  == nil ifTrue: [setSelector := (name, ':') asSymbol].
	^setSelector.! !

!GlorpAttributeModel methodsFor: 'get/set' stamp: ''!
setValueIn: anObject to: aValue 
	| real |
	self isValid ifFalse: [^self].
	real := anObject yourSelf.
	self checkType: real.
	self useDirectAccess ifTrue: [^self directSetValueIn: real to: aValue].
	^real perform: self setSelector with: aValue! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
system
	classModel isNil ifTrue: [^nil].
	^classModel system.! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
type
	^type! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
type: aClass
	type := aClass! !

!GlorpAttributeModel methodsFor: 'testing' stamp: ''!
useDirectAccess
	useDirectAccess isNil ifFalse: [^useDirectAccess].
	self system isNil ifTrue: [^true].  "A global default, although only applicable if we're not in a valid setup"
	^self system useDirectAccessForMapping.! !

!GlorpAttributeModel methodsFor: 'accessing' stamp: ''!
useDirectAccess: aBoolean
	useDirectAccess := aBoolean.! !

!GlorpAttributeModel methodsFor: 'validation' stamp: ''!
validate
	name isEmpty ifTrue: [self error: 'Invalid attribute name']! !

!GlorpAttributeModel methodsFor: 'validation' stamp: ''!
validateCanRead
	self isReadable ifFalse: [self raiseInvalidAttributeError]! !

!GlorpAttributeModel methodsFor: 'validation' stamp: ''!
validateCanWrite
	self isWritable ifFalse: [self raiseInvalidAttributeError]! !

Object subclass: #GlorpClassModel
	instanceVariableNames: 'system describedClass subclasses attributes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!GlorpClassModel commentStamp: '<historical>' prior: 0!
This holds the metadata describing a class. Since Smalltalk classes are fairly self-describing, this mostly consists of the names and types of the instance variables that we want to be persistent. If those variables are collections, we also want to know the types of the items in the collections.

Instance Variables:
	attributes	<(Collection of: GlorpAttributeModel)>	Our attributes, one for each instance variable.
	describedClass	<Class>	The class we hold metadata for.
	subclasses	<Array of: Class>	The collection of subclasses of our describedClass. Computed automatically, and used as a performance optimization for inheritance testing.
	system	<DescriptorSystem>	The system that contains us.

!

!GlorpClassModel classSide methodsFor: 'instance creation' stamp: ''!
for: aClass
	^self new describedClass: aClass.! !

!GlorpClassModel classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize.! !

!GlorpClassModel methodsFor: 'accessing' stamp: ''!
addAttribute: anAttributeDescription
	attributes add: anAttributeDescription.
	anAttributeDescription classModel: self.
	^anAttributeDescription.! !

!GlorpClassModel methodsFor: 'accessing' stamp: ''!
attributeNamed: aSymbol
	^attributes detect: [:each | each name = aSymbol] ifNone: [self newAttributeNamed: aSymbol].! !

!GlorpClassModel methodsFor: 'accessing' stamp: ''!
attributeNamed: aSymbol type: aClass
	^attributes
		detect:
			[:each | each name = aSymbol and: [each type isNil or: [each type = aClass]]]
		ifNone: [self newAttributeNamed: aSymbol type: aClass].! !

!GlorpClassModel methodsFor: 'accessing' stamp: ''!
attributeNamed: aSymbol type: aClass collectionType: collectionClass
	^attributes
		detect:
			[:each | 
			| collectionTypeMatches typeMatches nameMatches |
			nameMatches := each name = aSymbol.
			typeMatches := each type isNil or: [each type = aClass].
			collectionTypeMatches := each collectionType isNil
				or: [each collectionType = collectionClass].
			nameMatches & typeMatches & collectionTypeMatches]
		ifNone: [self newAttributeNamed: aSymbol type: aClass].! !

!GlorpClassModel methodsFor: 'accessing' stamp: ''!
attributes
	^attributes.! !

!GlorpClassModel methodsFor: 'attributes' stamp: ''!
attributesDo: aBlock
	attributes do: aBlock! !

!GlorpClassModel methodsFor: 'accessing' stamp: ''!
describedClass
	^describedClass! !

!GlorpClassModel methodsFor: 'accessing' stamp: ''!
describedClass: aClass
	describedClass := aClass.
	subclasses := aClass allSubclasses isEmpty ifTrue: [nil] ifFalse: [aClass allSubclasses asArray].! !

!GlorpClassModel methodsFor: 'accessing' stamp: ''!
hasAttributeNamed: aSymbol

	attributes do: [:attribute | attribute name = aSymbol ifTrue: [^true]].
	^false! !

!GlorpClassModel methodsFor: 'initializing' stamp: ''!
initialize
	attributes := OrderedCollection new.! !

!GlorpClassModel methodsFor: 'validation' stamp: ''!
matches: anObject
	"Try to optimize this check as much as possible. We cache the list of subclasses, and inline the check for isKindOf:"
	| theClass |
	theClass := anObject glorpRealObject class.
	describedClass == theClass ifTrue: [^true].
	nil == subclasses ifTrue: [^false].
	1 to: subclasses size do: [:i | theClass == (subclasses at: i)
		ifTrue: [^true]].
	^false.! !

!GlorpClassModel methodsFor: 'attributes' stamp: ''!
newAttributeNamed: aSymbol
	^self addAttribute: (GlorpAttributeModel named: aSymbol).! !

!GlorpClassModel methodsFor: 'attributes' stamp: ''!
newAttributeNamed: aSymbol collection: collectionClass of: aClass
	^self addAttribute: (GlorpAttributeModel named: aSymbol collection: collectionClass of: aClass).! !

!GlorpClassModel methodsFor: 'attributes' stamp: ''!
newAttributeNamed: aSymbol collectionOf: aClass
	^self addAttribute: (GlorpAttributeModel named: aSymbol collectionOf: aClass).! !

!GlorpClassModel methodsFor: 'attributes' stamp: ''!
newAttributeNamed: aSymbol dictionaryFrom: keyClass to: valueClass
	^self addAttribute: (GlorpAttributeModel named: aSymbol dictionaryFrom: keyClass to: valueClass).! !

!GlorpClassModel methodsFor: 'attributes' stamp: ''!
newAttributeNamed: aSymbol type: aClass
	^self addAttribute: (GlorpAttributeModel named: aSymbol type: aClass).! !

!GlorpClassModel methodsFor: 'printing' stamp: ''!
printOn: aStream
	aStream nextPutAll: '~~'.
	self describedClass isNil ifFalse: [aStream nextPutAll: self describedClass name].
	aStream nextPutAll: '~~'.! !

!GlorpClassModel methodsFor: 'accessing' stamp: ''!
system
	^system! !

!GlorpClassModel methodsFor: 'accessing' stamp: ''!
system: aDescriptorSystem
	system := aDescriptorSystem! !

!GlorpClassModel methodsFor: 'validation' stamp: ''!
validateIn: aDescriptorSystem 
	attributes do: [:each | each validate]! !

Object subclass: #GlorpCorePackage
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Licence'!

!GlorpCorePackage commentStamp: '<historical>' prior: 0!
This class represents package information for the GlorpCore package, for systems that don't have things like package properties to store things like copyright and license information.

LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!

!GlorpCorePackage classSide methodsFor: 'instance creation' stamp: ''!
LICENSE
"LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA"! !

Object subclass: #GlorpDatabasePackage
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Licence'!

!GlorpDatabasePackage commentStamp: '<historical>' prior: 0!
LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!

!GlorpDatabasePackage classSide methodsFor: 'instance creation' stamp: ''!
LICENSE
"LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA"! !

Object subclass: #GlorpDatabaseType
	instanceVariableNames: 'platform selector typeString'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpDatabaseType commentStamp: '<historical>' prior: 0!
This is an abstract superclass for objects representing types in the database. Each one will correspond to a particular simple type that we can store and retrieve from the database. e.g. int, varchar, and so on. For types that are variable sizes, different instances may represent, e.g. varchar(5) vs. varchar(20). While the classes may be reused across different databases, a database platform is expected to be able to return the types that it knows about as instances of these clasess. So you can send #varchar to OraclePlatform and get back one of these.

Subclasses must implement the following messages:
	exdi specific
		exdiType

Instance Variables
	platform	<DatabasePlatform>	The platform that owns us.
	typeString	<String>	Our name in the database.
	selector <Symbol> The name of the message we can send to the platform to get ourselves or an equivalent type back again.

!

GlorpDatabaseType subclass: #GlorpAbstractNumericType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

GlorpAbstractNumericType subclass: #GlorpAbstractIntegerType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpAbstractIntegerType methodsFor: 'converting' stamp: ''!
converterForStType: aClass
	(aClass includesBehavior: Boolean) ifTrue: [^self platform converterNamed: #booleanToInteger].
	^self platform converterNamed: #numberToInteger.! !

!GlorpAbstractIntegerType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#Integer.! !

!GlorpAbstractIntegerType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^Integer.! !

GlorpAbstractIntegerType subclass: #GlorpIntegerType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!GlorpIntegerType methodsFor: 'comparing' stamp: ''!
= aType
	(aType class == self platform numeric class and: [aType isIntegerType]) ifTrue: [^true].
	^super = aType! !

!GlorpIntegerType methodsFor: 'comparing' stamp: ''!
hash
	"Our hash must be identical to GlorpNumericType's hash when it is an #isInteger"
	^self platform hash bitXor: self impliedSmalltalkType hash! !

!GlorpIntegerType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'integer'.! !

GlorpAbstractIntegerType subclass: #GlorpSerialType
	instanceVariableNames: 'generated sequence'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpSerialType methodsFor: 'accessing' stamp: ''!
hasParameters
	^true.! !

!GlorpSerialType methodsFor: 'testing' stamp: ''!
hasSequence
	^true.! !

!GlorpSerialType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	generated := true.! !

!GlorpSerialType methodsFor: 'initialize' stamp: ''!
initializeForField: aDatabaseField in: aDescriptorSystem
	sequence isNil ifFalse: [^self].
	sequence := aDescriptorSystem databaseSequenceNamed: (aDescriptorSystem sequenceNameForField: aDatabaseField).
	sequence schema: aDatabaseField table schema.! !

!GlorpSerialType methodsFor: 'accessing' stamp: ''!
isGenerated
	"answer if we should autogenerate a value for this type, i.e. if it is a database-generated primary key column"

	^generated! !

!GlorpSerialType methodsFor: 'testing' stamp: ''!
isIdentityColumn
	^self sequence isIdentityColumn.! !

!GlorpSerialType methodsFor: 'testing' stamp: ''!
isSerial
	"answer if this is a database-generated primary key column"

	^true! !

!GlorpSerialType methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor

	^sequence postWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor.! !

!GlorpSerialType methodsFor: 'sequencing' stamp: ''!
preWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor

	^sequence preWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor.! !

!GlorpSerialType methodsFor: 'testing' stamp: ''!
printSelectSQLOn: aCommand withParameters: aDictionary
	"We record our type as serial, but that isn't really useful for a cast. We assume serials are integers, so just print a generic integer type instead."
	self platform integer printSelectSQLOn: aCommand withParameters: aDictionary.! !

!GlorpSerialType methodsFor: 'accessing' stamp: ''!
sequence
	^sequence.! !

!GlorpSerialType methodsFor: 'accessing' stamp: ''!
sequence: aDatabaseSequence
	sequence := aDatabaseSequence.! !

GlorpAbstractIntegerType subclass: #GlorpSmallIntType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!GlorpSmallIntType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'smallint'.! !

!GlorpAbstractNumericType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^Number.! !

GlorpAbstractNumericType subclass: #GlorpDoubleType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpDoubleType methodsFor: 'converting' stamp: ''!
converterForStType: aClass

	^self platform converterNamed: #numberToDouble.! !

!GlorpDoubleType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#Double.! !

!GlorpDoubleType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^Dialect doublePrecisionFloatClass.! !

!GlorpDoubleType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'double precision'.! !

GlorpAbstractNumericType subclass: #GlorpFloatType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpFloatType methodsFor: 'converting' stamp: ''!
converterForStType: aClass

	^self platform converterNamed: #numberToFloat.! !

!GlorpFloatType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#Float.! !

!GlorpFloatType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^Float.! !

!GlorpFloatType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'float4'.! !

GlorpAbstractNumericType subclass: #GlorpMSSQLFloatType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!GlorpMSSQLFloatType methodsFor: 'SQL' stamp: ''!
converterForStType: aClass

	^self platform converterNamed: #numberToDouble.! !

!GlorpMSSQLFloatType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#Float.! !

!GlorpMSSQLFloatType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^Float.! !

!GlorpMSSQLFloatType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'float'.! !

!GlorpMSSQLFloatType methodsFor: 'SQL' stamp: ''!
typeString
	^'float'.! !

GlorpAbstractNumericType subclass: #GlorpNumericType
	instanceVariableNames: 'precision scale'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseType'!

!GlorpNumericType methodsFor: 'comparing' stamp: ''!
= aType

	"We do a special case check here that something that reads back as numeric with no scale/precision counts as being the same as an integer type"
	((aType class == self platform integer class) and: [self isIntegerType]) ifTrue: [^true].
	super = aType ifFalse: [^false].
	self precision = aType precision ifFalse: [^false].
	self scale = aType scale ifFalse: [^false].
	^true.! !

!GlorpNumericType methodsFor: 'converting' stamp: ''!
converterForStType: aClass
	(aClass includesBehavior: Boolean) ifTrue: [^self platform converterNamed: #booleanToInteger].
	^self platform converterNamed: #numberToNumber.! !

!GlorpNumericType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"

	"Can we do better than this??"
	^#FixedPoint! !

!GlorpNumericType methodsFor: 'testing' stamp: ''!
hasParameters
	"Return true if this has modifiable parameters. That is, when we return one of these, should we return a copy rather than trying to save space be re-using instances"
	^true.! !

!GlorpNumericType methodsFor: 'comparing' stamp: ''!
hash
	"A very special case here, if we're effectively an Integer, then we need to hash in a way compatible with GlorpIntegerType"
	self isIntegerType ifTrue: [^self platform hash bitXor: self impliedSmalltalkType hash].
	^(super hash bitXor: self precision hash) bitXor: self scale hash! !

!GlorpNumericType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	self isIntegerType ifTrue: [^Integer].
	^super impliedSmalltalkType! !

!GlorpNumericType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'numeric'.! !

!GlorpNumericType methodsFor: 'testing' stamp: ''!
isIntegerType
	"If we have no precision and no scale, we're effectively an Integer"
	^self precision isNil and: [self scale isNil]! !

!GlorpNumericType methodsFor: 'accessing' stamp: ''!
precision
	^precision.! !

!GlorpNumericType methodsFor: 'accessing' stamp: ''!
precision: anInteger

	precision := anInteger.! !

!GlorpNumericType methodsFor: 'accessing' stamp: ''!
scale
	^scale.! !

!GlorpNumericType methodsFor: 'accessing' stamp: ''!
scale: anInteger

	scale := anInteger.! !

!GlorpNumericType methodsFor: 'SQL' stamp: ''!
typeString
	| w |
	platform supportsVariableSizedNumerics ifFalse: [^typeString].
	w := WriteStream on: String new.
	w nextPutAll: typeString.
	precision isNil 
		ifFalse: 
			[w nextPutAll: '(' , precision printString.
			scale isNil ifFalse: [w nextPutAll: ',' , scale printString].
			w nextPutAll: ')'].
	^w contents! !

GlorpDatabaseType subclass: #GlorpAbstractStringType
	instanceVariableNames: 'width queryType'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpAbstractStringType methodsFor: 'comparing' stamp: ''!
= aType
	super = aType ifFalse: [^false].
	self width = aType width ifFalse: [^false].
	self queryType == self ifFalse: [
		self queryType = aType queryType ifFalse: [^false]].
	^true.! !

!GlorpAbstractStringType methodsFor: 'converting' stamp: ''!
converterForStType: aClass
	(aClass includesBehavior: Boolean) ifTrue: [^self platform converterNamed: #booleanToStringTF].
	(aClass includesBehavior: Symbol) ifTrue: [^self platform converterNamed: #symbolToString].
	width isNil ifFalse: [
		(aClass includesBehavior: String) ifTrue: [^self platform converterNamed: #stringToString]].
	(aClass includesBehavior: Integer) ifTrue: [^self platform converterNamed: #integerToString].
	^super converterForStType: aClass.! !

!GlorpAbstractStringType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#String.! !

!GlorpAbstractStringType methodsFor: 'testing' stamp: ''!
hasParameters
	"Return true if this has modifiable parameters. That is, when we return one of these, should we return a copy rather than trying to save space be re-using instances"
	^true.! !

!GlorpAbstractStringType methodsFor: 'testing' stamp: ''!
hasWidth
	"Return true if this type has a width attribute - basically meaning it's a string or LOB type. Note that this is different from the isVariableWidth test, which tells us if a field has fixed vs. varying width. So a CHAR field hasWidth=>true, but isVariableWidth=>false, while a VARCHAR field hasWidth=>true isVariableWidth=>false, and a NUMBER has both false."
	^true.! !

!GlorpAbstractStringType methodsFor: 'comparing' stamp: ''!
hash
	^self queryType == self
		ifTrue: [super hash bitXor: self width hash]
		ifFalse: [(super hash bitXor: self width hash) bitXor: self queryType hash]! !

!GlorpAbstractStringType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^String.! !

!GlorpAbstractStringType methodsFor: 'testing' stamp: ''!
isStringType
	^true.! !

!GlorpAbstractStringType methodsFor: 'testing' stamp: ''!
isVariableWidth
	self subclassResponsibility.! !

!GlorpAbstractStringType methodsFor: 'accessing' stamp: ''!
queryType
	queryType isNil ifTrue: [^self].
	^queryType.! !

!GlorpAbstractStringType methodsFor: 'accessing' stamp: ''!
queryType: aType
	queryType := aType! !

!GlorpAbstractStringType methodsFor: 'private' stamp: ''!
typeName
	^typeString.! !

!GlorpAbstractStringType methodsFor: 'private' stamp: ''!
typeName: aString
	typeString := aString.! !

!GlorpAbstractStringType methodsFor: 'SQL' stamp: ''!
typeString
	typeString isNil
		ifTrue:
			[typeString := self typeName.
			width isNil
				ifFalse: [typeString := typeString , '(' , width printString , ')']].
	^typeString.! !

!GlorpAbstractStringType methodsFor: 'SQL' stamp: ''!
typeString: aString
	"If we set the typeString, we are presumably setting it to the non-parameterized name as well"
	typeString := aString.
	super typeString: aString.! !

!GlorpAbstractStringType methodsFor: 'accessing' stamp: ''!
width
	^width.! !

!GlorpAbstractStringType methodsFor: 'accessing' stamp: ''!
width: anInteger

	width := anInteger = 0 ifTrue: [nil] ifFalse: [anInteger]! !

GlorpAbstractStringType subclass: #GlorpCharType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpCharType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'char'.! !

!GlorpCharType methodsFor: 'testing' stamp: ''!
isVariableWidth
	^false.! !

!GlorpCharType methodsFor: 'accessing' stamp: ''!
typeString

	^self typeName, '(', width printString, ')'.! !

GlorpAbstractStringType subclass: #GlorpClobType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpClobType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#LargeString.! !

!GlorpClobType methodsFor: 'testing' stamp: ''!
hasParameters
	"Return true if this has modifiable parameters. That is, when we return one of these, should we return a copy rather than trying to save space be re-using instances"
	^false.! !

!GlorpClobType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'text'.
	width := nil.! !

!GlorpClobType methodsFor: 'testing' stamp: ''!
isVariableWidth
	^true.! !

!GlorpClobType methodsFor: 'testing' stamp: ''!
requiresCast
	^(self platform class = PostgreSQLPlatform) or: [self platform class = SQLServerPlatform]! !

GlorpClobType subclass: #GlorpMySQLClobType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpMySQLClobType commentStamp: '<historical>' prior: 0!
MySQL clobs come back from the driver as binary data, so we need to make a subclass that knows it needs a converter for its incoming data.!

GlorpAbstractStringType subclass: #GlorpTextType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpTextType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#LargeString.! !

!GlorpTextType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'text'.! !

!GlorpTextType methodsFor: 'testing' stamp: ''!
isVariableWidth
	^true.! !

GlorpAbstractStringType subclass: #GlorpVarCharType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpVarCharType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'varchar'.! !

!GlorpVarCharType methodsFor: 'testing' stamp: ''!
isVariableWidth
	"Return true if this type allows varying length data within a particular instance. e.g., this is true for a varchar, but false for a fixed size character field"
	^true.! !

!GlorpVarCharType methodsFor: 'accessing' stamp: ''!
typeString
	
	^self typeName,   ( width ifNil: [''] ifNotNil: [ :w |'(', w printString, ')'])! !

GlorpDatabaseType subclass: #GlorpBlobType
	instanceVariableNames: 'queryType'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpBlobType methodsFor: 'comparing' stamp: ''!
= aType
	super = aType ifFalse: [^false].
	self queryType == self ifTrue: [^true].
	self queryType = aType queryType
		 ifFalse: [^false].
	^true.! !

!GlorpBlobType methodsFor: 'types' stamp: ''!
converterForStType: aClass
	(aClass includesBehavior: String) ifTrue: [^self platform converterNamed: #byteArrayToString].
	self platform isPostgreSQLPlatform ifTrue: [^self platform converterNamed: #byteArrayToSQLString].
	self platform isOcelotPlatform ifTrue: [^self platform converterNamed: #blobDataToStream].
	^self platform nullConverter! !

!GlorpBlobType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#ByteArray.! !

!GlorpBlobType methodsFor: 'comparing' stamp: ''!
hash
	^self queryType == self
		ifTrue: [super hash]
		ifFalse: [super hash bitXor: self queryType hash]! !

!GlorpBlobType methodsFor: 'types' stamp: ''!
impliedSmalltalkType
	^ByteArray.! !

!GlorpBlobType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'blob'! !

!GlorpBlobType methodsFor: 'SQL' stamp: ''!
print: aValue on: aStream
	self platform printBlob: (aValue isNil ifTrue: [nil] ifFalse: [aValue asByteArray]) on: aStream for: self.! !

!GlorpBlobType methodsFor: 'types' stamp: ''!
queryType
	"Return the type that we can cast an argument to in order to match against this"
	queryType isNil ifTrue: [^self].
	^queryType.! !

!GlorpBlobType methodsFor: 'types' stamp: ''!
queryType: aType
	"Set the type that we can cast an argument to in order to match against this"
	queryType := aType.! !

!GlorpBlobType methodsFor: 'testing' stamp: ''!
requiresConvertToVarbinary
	"This is a hack for testing against SQL Server"
	^true.! !

GlorpDatabaseType subclass: #GlorpBooleanType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpBooleanType methodsFor: 'converting' stamp: ''!
converterForStType: aClass
	(aClass includesBehavior: Boolean) ifTrue: [^self platform converterNamed: #booleanToBoolean].
	^self platform nullConverter.! !

!GlorpBooleanType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#Boolean.! !

!GlorpBooleanType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^Boolean.! !

!GlorpBooleanType methodsFor: 'SQL' stamp: ''!
print: aValue on: aStream
	aStream nextPutAll: (self platform printBoolean: aValue for: self).! !

!GlorpDatabaseType classSide methodsFor: 'instance creation' stamp: ''!
instance

	^super new! !

!GlorpDatabaseType classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize.! !

!GlorpDatabaseType classSide methodsFor: 'printing' stamp: ''!
padToThreeDigits: anInteger

	| string |
	string := anInteger truncated printString.
	string size = 3 ifTrue: [^string].
	string size = 2 ifTrue: [^'0', string].
	string size = 1 ifTrue: [^'00', string].! !

!GlorpDatabaseType classSide methodsFor: 'printing' stamp: ''!
padToTwoDigits: anInteger

	| string |
	string := anInteger truncated printString.
	^string size = 1 ifTrue: ['0', string] ifFalse: [string].! !

!GlorpDatabaseType methodsFor: 'comparing' stamp: ''!
= aType
	^self basicEquals: aType.! !

!GlorpDatabaseType methodsFor: 'comparing' stamp: ''!
basicEquals: aType
	self class == aType class ifFalse: [^false].
	self platform == aType platform ifFalse: [^false].
	self typeString = aType typeString ifFalse: [^false].
	^true.! !

!GlorpDatabaseType methodsFor: 'converting' stamp: ''!
converterForStType: aClass

	^self platform nullConverter.! !

!GlorpDatabaseType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	self subclassResponsibility.! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
generatesOverExistingValues
	"Do we generate our value even if there's alread one present in the field"

	^false.! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
glorpPrintSQLOn: aCommand
	aCommand nextPutAll: self typeString.! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
glorpPrintSelectSQLOn: aCommand
	self printSelectSQLOn: aCommand withParameters: #().! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
hasParameters
	"Return true if this has modifiable parameters. That is, when we return one of these, should we return a copy rather than trying to save space be re-using instances"
	^false.! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
hasSequence
	^false.! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
hasWidth
	"Return true if this type has a width attribute - basically meaning it's a string or LOB type. Note that this is different from the isVariableWidth test, which tells us if a field has fixed vs. varying width. So a CHAR field hasWidth=>true, but isVariableWidth=>false, while a VARCHAR field hasWidth=>true isVariableWidth=>false, and a NUMBER has both false."
	^false.! !

!GlorpDatabaseType methodsFor: 'comparing' stamp: ''!
hash
	^self platform hash bitXor: self typeString hash! !

!GlorpDatabaseType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	"Return the Smalltalk type which most commonly corresponds to our database type. By default, Object if we don't have any more specific information."
	^Object.! !

!GlorpDatabaseType methodsFor: 'initialize' stamp: ''!
initialize! !

!GlorpDatabaseType methodsFor: 'initialize' stamp: ''!
initializeForField: aDatabaseField in: aDescriptorSystem! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
isGenerated

	^false! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
isIdentityColumn
	^false.! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
isSerial
	"answer if this is a database-generated primary key column"

	^false! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
isStringType
	"Return true if the type of values this stores are strings"
	^false.! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
isVariable
	^false.! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
isVariableWidth
	"Return true if this type allows varying length data within a particular instance. e.g., this is true for a varchar, but false for a fixed size character field"
	^false.! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
platform

	^platform! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
platform: aDatabasePlatform

	platform := aDatabasePlatform.! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
postWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
postWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: aSession! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
preWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor
	aDatabaseField defaultValue isNil ifTrue: [^self].
	(aDatabaseRow hasValueFor: aDatabaseField) ifTrue: [^self].
	"Default values will not write if this field is constrained to a primary key generator"
	(aDatabaseRow expectsToGetValueFor: aDatabaseField) ifTrue: [^self].
	aDatabaseRow at: aDatabaseField put: (aDatabaseField defaultValue).! !

!GlorpDatabaseType methodsFor: 'accessing' stamp: ''!
precision: anInteger

	^self error: ((self class name asString), ' is not a variable precision type.')! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
print: aValue on: aStream
	aValue glorpPrintSQLOn: aStream.! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
printCollection: aCollection on: aStream
	aCollection glorpPrintSQLOn: aStream for: self.! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
printSelectSQLOn: aCommand withParameters: aDictionary
	aCommand nextPutAll: self typeString.! !

!GlorpDatabaseType methodsFor: 'accessing' stamp: ''!
queryType
	"If we can't pass this type in as an argument to a query (e.g. a blob) then return a type we can use instead"
	^self.! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
requiresCast
	^false.! !

!GlorpDatabaseType methodsFor: 'testing' stamp: ''!
requiresConvertToVarbinary
	"This is a hack for testing against SQL Server"
	^false.! !

!GlorpDatabaseType methodsFor: 'accessing' stamp: ''!
scale: anInteger

	^self error: ((self class name asString), ' is not a variable scale type.')! !

!GlorpDatabaseType methodsFor: 'accessing' stamp: ''!
selector
	^selector! !

!GlorpDatabaseType methodsFor: 'accessing' stamp: ''!
selector: aSymbol
	selector := aSymbol! !

!GlorpDatabaseType methodsFor: 'accessing' stamp: ''!
sequence
	^nil! !

!GlorpDatabaseType methodsFor: 'accessing' stamp: ''!
size: anInteger

	^self error: ((self class name asString), ' is not a variable sized type.')! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
typeName
	"An alias that can be used to print the non-parameterized name"
	^typeString.! !

!GlorpDatabaseType methodsFor: 'SQL' stamp: ''!
typeString

	^typeString.! !

!GlorpDatabaseType methodsFor: 'accessing' stamp: ''!
typeString: aString

	typeString := aString! !

!GlorpDatabaseType methodsFor: 'accessing' stamp: ''!
underlyingType
	^self.! !

GlorpDatabaseType subclass: #GlorpDateType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpDateType methodsFor: 'converting' stamp: ''!
converterForStType: aClass

	^self platform converterNamed: #date.! !

!GlorpDateType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^self platform exdiTypeForDates.! !

!GlorpDateType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^Date.! !

!GlorpDateType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'date'! !

!GlorpDateType methodsFor: 'SQL' stamp: ''!
print: aValue on: aStream
	aStream nextPutAll: (self platform printDate: aValue for: self).! !

GlorpDatabaseType subclass: #GlorpInMemorySequenceDatabaseType
	instanceVariableNames: 'representationType'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

GlorpInMemorySequenceDatabaseType class
	instanceVariableNames: 'count'!

GlorpInMemorySequenceDatabaseType class
	instanceVariableNames: 'count'!

!GlorpInMemorySequenceDatabaseType classSide methodsFor: 'accessing' stamp: ''!
next

	(count isNil)
	ifTrue: [count := 0].
	count := count + 1.
	^count! !

!GlorpInMemorySequenceDatabaseType classSide methodsFor: 'instance creation' stamp: ''!
representedBy: dbType

	^super new
		representedBy: dbType! !

!GlorpInMemorySequenceDatabaseType classSide methodsFor: 'accessing' stamp: ''!
reset

	count := 0! !

!GlorpInMemorySequenceDatabaseType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^representationType exdiType.! !

!GlorpInMemorySequenceDatabaseType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^Integer.! !

!GlorpInMemorySequenceDatabaseType methodsFor: 'testing' stamp: ''!
isGenerated

	^true! !

!GlorpInMemorySequenceDatabaseType methodsFor: 'SQL' stamp: ''!
preWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor

	aDatabaseRow at: aDatabaseField put: (self class next)! !

!GlorpInMemorySequenceDatabaseType methodsFor: 'initialize-release' stamp: ''!
representedBy: dbType

	representationType := dbType! !

!GlorpInMemorySequenceDatabaseType methodsFor: 'SQL' stamp: ''!
typeName

	^representationType typeName.! !

!GlorpInMemorySequenceDatabaseType methodsFor: 'SQL' stamp: ''!
typeString

	^representationType typeString! !

GlorpDatabaseType subclass: #GlorpMSSQLSysnameType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!GlorpMSSQLSysnameType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#String.! !

!GlorpMSSQLSysnameType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^String.! !

!GlorpMSSQLSysnameType methodsFor: 'testing' stamp: ''!
isStringType
	^true.! !

!GlorpMSSQLSysnameType methodsFor: 'SQL' stamp: ''!
typeString

	^'sysname'.! !

GlorpDatabaseType subclass: #GlorpPGTimeWithTimeZoneType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpPGTimeWithTimeZoneType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#Timestamp.! !

!GlorpPGTimeWithTimeZoneType methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	typeString := 'TIME WITH TIME ZONE'.! !

GlorpDatabaseType subclass: #GlorpTimeStampType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpTimeStampType methodsFor: 'converting' stamp: ''!
converterForStType: aClass
	aClass = Date ifTrue: [^self platform converterNamed: #timestampToDate].
	^self platform converterNamed: #timestamp.! !

!GlorpTimeStampType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#Timestamp.! !

!GlorpTimeStampType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^Dialect timestampClass.! !

!GlorpTimeStampType methodsFor: 'SQL' stamp: ''!
print: aValue on: aStream
	self platform printTimestamp: aValue on: aStream for: self.! !

GlorpDatabaseType subclass: #GlorpTimeType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpTimeType methodsFor: 'conversion-times' stamp: ''!
converterForStType: aClass

	^self platform converterNamed: #time.! !

!GlorpTimeType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^#Time.! !

!GlorpTimeType methodsFor: 'converting' stamp: ''!
impliedSmalltalkType
	^Time.! !

!GlorpTimeType methodsFor: 'SQL' stamp: ''!
print: aValue on: aStream
	aStream nextPutAll: (self platform printTime: aValue for: self).! !

GlorpDatabaseType subclass: #GlorpVarBinaryType
	instanceVariableNames: 'width'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!GlorpVarBinaryType methodsFor: 'exdi specific' stamp: ''!
exdiType
	^#ByteArray.! !

!GlorpVarBinaryType methodsFor: 'testing' stamp: ''!
hasParameters
	"Return true if this has modifiable parameters. That is, when we return one of these, should we return a copy rather than trying to save space be re-using instances"
	^true.! !

!GlorpVarBinaryType methodsFor: 'testing' stamp: ''!
impliedSmalltalkType
	^ByteArray.! !

!GlorpVarBinaryType methodsFor: 'initialize-release' stamp: ''!
initialize
	super initialize.
	typeString := 'varbinary'! !

!GlorpVarBinaryType methodsFor: 'testing' stamp: ''!
isVariableWidth
	"Return true if this type allows varying length data within a particular instance. e.g., this is true for a varchar, but false for a fixed size character field"
	^true.! !

!GlorpVarBinaryType methodsFor: 'accessing' stamp: ''!
width
	^width.! !

!GlorpVarBinaryType methodsFor: 'accessing' stamp: ''!
width: anInteger

	width := anInteger! !

GlorpDatabaseType subclass: #VersionType
	instanceVariableNames: 'underlyingType generator generatesOverExistingValues'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-DatabaseTypes'!

!VersionType commentStamp: '<historical>' prior: 0!
This represents the type of something used for a version field.

Instance Variables:
	underlyingType	<DatabaseType>	The underlying type of our field.
	sequence <> The generator used to generate the next value.

!

!VersionType classSide methodsFor: 'instance creation' stamp: ''!
underlyingType: aDatabaseType
	^self new underlyingType: aDatabaseType.! !

!VersionType classSide methodsFor: 'instance creation' stamp: ''!
underlyingType: aDatabaseType generator: aGenerator
	^self new underlyingType: aDatabaseType generator: aGenerator.! !

!VersionType methodsFor: 'converting' stamp: ''!
converterForStType: aClass

	^underlyingType converterForStType: aClass.! !

!VersionType methodsFor: 'exdi specific' stamp: ''!
exdiType
	"Return a symbol that the VisualWorks EXDI uses to indicate the database type -- this is a symbol which is often the name of a Smalltalk class, but with variations for multiple db types that correspond to the same Smalltalk symbol. It also varies by Database. This is equivalent to the values returned by methods like #odbcConversion or #oracleConversion"
	^self underlyingType exdiType.! !

!VersionType methodsFor: 'testing' stamp: ''!
generateOverExistingValues
	generatesOverExistingValues := true.! !

!VersionType methodsFor: 'testing' stamp: ''!
generatesOverExistingValues
	"Do we generate our value even if there's alread one present in the field"
	^generatesOverExistingValues.! !

!VersionType methodsFor: 'testing' stamp: ''!
generatesOverExistingValues: aBoolean
	"Do we generate our value even if there's alread one present in the field"
	generatesOverExistingValues := aBoolean.! !

!VersionType methodsFor: 'accessing' stamp: ''!
generator
	^generator! !

!VersionType methodsFor: 'accessing' stamp: ''!
generator: aDatabaseValueGenerator
	generator := aDatabaseValueGenerator! !

!VersionType methodsFor: 'initializing' stamp: ''!
generatorFor: aDatabaseType
	aDatabaseType impliedSmalltalkType == Integer ifTrue: [^VersionNumberVersionGenerator new].
	aDatabaseType impliedSmalltalkType == Dialect timestampClass ifTrue: [^TimestampVersionGenerator new].
	self error: 'unknown version type'.! !

!VersionType methodsFor: 'initializing' stamp: ''!
initialize
	super initialize.
	generatesOverExistingValues := false.! !

!VersionType methodsFor: 'testing' stamp: ''!
isGenerated
	^true.! !

!VersionType methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor

	^self.! !

!VersionType methodsFor: 'sequencing' stamp: ''!
preWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor

	^generator preWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor.! !

!VersionType methodsFor: 'SQL' stamp: ''!
print: aValue on: aStream
	self underlyingType print: aValue on: aStream.! !

!VersionType methodsFor: 'SQL' stamp: ''!
typeString
	^underlyingType typeString.! !

!VersionType methodsFor: 'accessing' stamp: ''!
underlyingType
	^underlyingType! !

!VersionType methodsFor: 'accessing' stamp: ''!
underlyingType: aDatabaseType
	underlyingType := aDatabaseType.
	platform := aDatabaseType platform.
	generator := self generatorFor: aDatabaseType.! !

!VersionType methodsFor: 'accessing' stamp: ''!
underlyingType: aDatabaseType generator: aGenerator
	underlyingType := aDatabaseType.
	platform := aDatabaseType platform.
	generator := aGenerator.! !

!VersionType methodsFor: 'accessing' stamp: ''!
width
	^self underlyingType width.! !

Object subclass: #GlorpDatabaseTypesPackage
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Licence'!

!GlorpDatabaseTypesPackage commentStamp: '<historical>' prior: 0!
This class represents package information for the GlorpDatabaseTypes package, for systems that don't have things like package properties to store things like copyright and license information.

LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!

!GlorpDatabaseTypesPackage classSide methodsFor: 'instance creation' stamp: ''!
LICENSE
"LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA"! !

Object subclass: #GlorpDatabaseValueGenerator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!GlorpDatabaseValueGenerator commentStamp: '<historical>' prior: 0!
This is an abstract superclass for thing that generate values when objects are written to the database. The most common usage is the various types of sequences, identity columns and other primary key generating mechanism. But it can also be used for optimistic locking via version numbers or timestamps. It doesn't actually have any code at the moment, so it might just be deletable, or maybe there's some refactoring to be done.!

GlorpDatabaseValueGenerator subclass: #DatabaseSequence
	instanceVariableNames: 'reservedNumbers'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!DatabaseSequence commentStamp: '<historical>' prior: 0!
This is an abstract superclass for identifier generation mechanisms in the database. For example, Oracle sequences, SQL Server identity columns, and so on.

Subclasses must implement the following messages:
	sequencing
		getSequenceValueFromDatabaseFor:in:using:
		postWriteAssignSequenceValueFor:in:using:

Instance Variables:
	reservedNumbers	<OrderedCollection>	Numbers that we have already reserved and have available. This is only useful if the sequence can generate more than one identifier in a single request, and we have configured it (and ourselves) to do so and use those numbers. This is useful for efficiency when inserting lots of data.

!

!DatabaseSequence classSide methodsFor: 'instance creation' stamp: ''!
named: aString
	^self new name: aString.! !

!DatabaseSequence classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!DatabaseSequence methodsFor: 'sequencing' stamp: ''!
creationString

	^'Create sequence ', self qualifiedName.! !

!DatabaseSequence methodsFor: 'sequencing' stamp: ''!
dropString

	^'Drop string unspecified for this type of sequence'.! !

!DatabaseSequence methodsFor: 'sequencing' stamp: ''!
getSequenceValueFromDatabaseFor: aField in: aRow using: aSession 
	self subclassResponsibility.! !

!DatabaseSequence methodsFor: 'sequencing' stamp: ''!
hasReservedNumbers
	^reservedNumbers notNil and: [reservedNumbers isEmpty not].! !

!DatabaseSequence methodsFor: 'initialize/release' stamp: ''!
initialize! !

!DatabaseSequence methodsFor: 'testing' stamp: ''!
isIdentityColumn
	^false.! !

!DatabaseSequence methodsFor: 'sequencing' stamp: ''!
nextReservedNumber
	^reservedNumbers removeFirst.! !

!DatabaseSequence methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aField in: aRow using: aSession

	self subclassResponsibility.! !

!DatabaseSequence methodsFor: 'sequencing' stamp: ''!
preWriteAssignSequenceValueFor: aField in: aRow using: anAccessor
	self hasReservedNumbers
		ifTrue: [aRow at: aField put: self nextReservedNumber]
		ifFalse:
			[self
				getSequenceValueFromDatabaseFor: aField
				in: aRow
				using: anAccessor].! !

!DatabaseSequence methodsFor: 'accessing' stamp: ''!
reservedNumbers
	reservedNumbers isNil ifTrue: [reservedNumbers := OrderedCollection new: 50].
	^reservedNumbers.! !

DatabaseSequence subclass: #InMemorySequence
	instanceVariableNames: 'count'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!InMemorySequence commentStamp: '<historical>' prior: 0!
This is a sequence generator that just generates numbers in memory. Mostly useful for testing, because the numbers aren't guaranteed unique in the database unless we are the only user, and our initial value was appropriately set.

Instance Variables:
	count	<Integer>	The last used identifier.
!

!InMemorySequence classSide methodsFor: 'defaults' stamp: ''!
default

	^self new.! !

!InMemorySequence methodsFor: 'public' stamp: ''!
getSequenceValueFromDatabaseFor: aField in: aRow using: aSession

	aRow at: aField put: (count := count + 1).! !

!InMemorySequence methodsFor: 'initialize/release' stamp: ''!
initialize

	super initialize.
	count := 0.! !

!InMemorySequence methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aField in: aRow! !

!InMemorySequence methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aField in: aRow using: aSession! !

!InMemorySequence methodsFor: 'sequencing' stamp: ''!
preWriteAssignSequenceValueFor: aField in: aRow

	aRow at: aField put: (count := count + 1).! !

DatabaseSequence subclass: #NamedSequence
	instanceVariableNames: 'name tableSelectCommand schema'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!NamedSequence commentStamp: '<historical>' prior: 0!
This is a database sequence that is identified by name.

Subclasses must implement the following messages:
	SQL
		tableSelectCommandFor:in:reserving:

Instance Variables:
	name	<String>	The name of the sequence
	schema	<String>	The schema that contains the sequence. Subclasses should set this up appropriately.
	tableSelectCommand	<SQLStringSelectCommand>	The command to execute to get the next value. 

!

NamedSequence subclass: #FirebirdSequence
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!FirebirdSequence commentStamp: '<historical>' prior: 0!
Implements firebird sequence generation by calling out to gen_id()

!

!FirebirdSequence methodsFor: 'public' stamp: ''!
creationString

	^'create generator ' , self qualifiedName.! !

!FirebirdSequence methodsFor: 'sequencing' stamp: ''!
dropString
	^'DELETE FROM RDB$GENERATORS
        WHERE RDB$GENERATOR_NAME=''', self qualifiedName, ''''.! !

!FirebirdSequence methodsFor: 'sequencing' stamp: ''!
getSequenceValueFromDatabaseFor: aDatabaseField in: aDatabaseRow using: anAccessor 
	| statement rows pk |
	statement := 'select gen_id(' , self qualifiedName 
				, ',1) from rdb$database'.
	rows := anAccessor executeSQLString: statement.
	pk := rows first first.
	aDatabaseRow at: aDatabaseField put: pk! !

!FirebirdSequence methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aField in: aRow using: aSession! !

!FirebirdSequence methodsFor: 'As yet unclassified' stamp: ''!
reserveViaBackupMechanism: anInteger in: aSession
	"For interbase, the generator mechanism should always work"
	self error: 'This should never happen'.! !

!FirebirdSequence methodsFor: 'sequencing' stamp: ''!
reserveViaTableSelect: anInteger in: aSession for: aTable
	"Interbase returns the top of the generated range in a single row. Do the iteration ourselves"
	| rows command max |
	command := self
		tableSelectCommandFor: aTable
		in: aSession
		reserving: anInteger.
	command parameters: (Array with: anInteger).
	command useBinding: false.
	rows := aSession accessor executeCommand: command.
	max := rows first first.
	max - anInteger + 1 to: max do: [:i | self reservedNumbers add: i].! !

!FirebirdSequence methodsFor: 'SQL' stamp: ''!
tableSelectCommandFor: aTable in: aSession reserving: anInteger
	"Since tableSelectCommand was never nilled, a second pundle version reusing this sequence used to get an SQL string embedding the first's integer value.  If it was a version of a different pundle, its integer probably differed;  when a call to reserve e.g. 1233 returns a maximum got by reusing the prior version's call to reserve e.g. 85, the result is 1148 duplicate keys.  If #insertOnly was set,  replication failed on an illegal command.  If we reached one of the fix-up states where it was not, then our cache assured us we were updating and we happily overwrote e.g. the blob for a class definition with that for a method definition (as I experienced :-/).  As sequences are set in the descriptor system and reused, this persisted until the session was discarded or its system was revised.  We now check the integer."

	| stream useBinding |
	"Binding into this doesn't seem to work"
	useBinding := false.
	(tableSelectCommand notNil and: [tableSelectCommand parameters includes: anInteger]) ifFalse:
		[stream := WriteStream on: (String new: 200).
		stream nextPutAll: 'select gen_id('.
		stream nextPutAll: self qualifiedName.
		stream nextPut: $,.
		stream nextPutAll: (useBinding ifTrue: ['?'] ifFalse: [anInteger printString]).
		stream nextPutAll: ') from rdb$database'.
		tableSelectCommand := SQLStringSelectCommand new.
		tableSelectCommand parameters: (Array with: anInteger).
		tableSelectCommand setSQLString: stream contents.
		tableSelectCommand session: aSession].
	^tableSelectCommand! !

NamedSequence subclass: #JustSelectTheMaximumSequenceValueAndAddOne
	instanceVariableNames: 'field table'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!JustSelectTheMaximumSequenceValueAndAddOne commentStamp: '<historical>' prior: 0!
This is a sequence that just does a select max(primaryKeyFieldName) for the table in question and adds one to it. This is, um, less-than-perfectly efficient, and I'm not at all clear that it'll work for a multi-user system. But it's what Store used to do on SQL Server, so we'd like to be able to mimic it. And it's actually easier to preallocate for than identity columns, which is nice. This isn't really named in the same sense as the others, but it's useful to inherit the named preallocation behaviour, and we store the field to select against in the name.

Instance Variables:
	tableName	<DatabaseTable>	the table we sequence.
	field <DatabaseField> the field we sequence.!

!JustSelectTheMaximumSequenceValueAndAddOne methodsFor: 'accessing' stamp: ''!
field
	field isNil ifTrue: [field := table primaryKeyFields first].
	^field.! !

!JustSelectTheMaximumSequenceValueAndAddOne methodsFor: 'accessing' stamp: ''!
field: aDatabaseField
	field := aDatabaseField.! !

!JustSelectTheMaximumSequenceValueAndAddOne methodsFor: 'other' stamp: ''!
getSequenceValueFromDatabaseFor: aField in: aRow using: aSession

	| stream rows value |
	"I repeat: ick"
	stream := WriteStream on: (String new: 50).
	stream nextPutAll: 'SELECT MAX('.
	aField printSQLOn: stream withParameters: #().
	stream nextPutAll: ') FROM '.
	aRow table printSQLOn: stream withParameters: #().
	rows := aSession accessor executeSQLString: stream contents.
	value := rows first first isNil ifTrue: [1] ifFalse: [rows first first + 1].
	aRow at: aField put: value.
	^value.! !

!JustSelectTheMaximumSequenceValueAndAddOne methodsFor: 'other' stamp: ''!
reserveSequenceNumbers: anInteger in: anAccessor for: aTable
	| numberNeeded firstNumber fakeRow |
	numberNeeded := anInteger - self availableReservedNumbers.
	numberNeeded <= 0 ifTrue: [^self].
	fakeRow := DatabaseRow newForTable: self field table.
	firstNumber := self
		getSequenceValueFromDatabaseFor: self field
		in: fakeRow
		using: anAccessor.
	self reservedNumbers addAll: (firstNumber to: firstNumber + numberNeeded - 1).! !

!JustSelectTheMaximumSequenceValueAndAddOne methodsFor: 'accessing' stamp: ''!
table: aTable
	table := aTable.! !

NamedSequence subclass: #MySQLSequence
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!MySQLSequence commentStamp: '<historical>' prior: 0!
This is a sequence type for MySQL AutoIncrement fields.!

!MySQLSequence methodsFor: 'sequencing' stamp: ''!
getSequenceValueFromDatabaseFor: aField in: aRow using: aSession! !

!MySQLSequence methodsFor: 'testing' stamp: ''!
isIdentityColumn
	^true.! !

!MySQLSequence methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor 
	aDatabaseRow at: aDatabaseField
		put: ((anAccessor 
				executeSQLString: 'SELECT LAST_INSERT_ID()') first atIndex: 1).! !

!MySQLSequence methodsFor: 'sequencing' stamp: ''!
reserveSequenceNumbers: anInteger in: aSession for: aTable
	"No real sequences here, just identity columns, which we can't pre-allocate"! !

!NamedSequence methodsFor: 'SQL' stamp: ''!
availableReservedNumbers
	^self reservedNumbers size.! !

!NamedSequence methodsFor: 'sequencing' stamp: ''!
dropString

	^'Drop sequence ', self qualifiedName.! !

!NamedSequence methodsFor: 'SQL' stamp: ''!
maximumToReserveAtOnce
		^1000.! !

!NamedSequence methodsFor: 'accessing' stamp: ''!
name
	^name.! !

!NamedSequence methodsFor: 'accessing' stamp: ''!
name: aString
	name := aString.! !

!NamedSequence methodsFor: 'printing' stamp: ''!
printOn: aStream
	super printOn: aStream.
	aStream
		nextPut: $(;
		nextPutAll: (name isNil ifTrue: ['unnamed'] ifFalse: [name]);
		nextPut: $).! !

!NamedSequence methodsFor: 'accessing' stamp: ''!
qualifiedName
	^(schema isNil or: [schema isEmpty]) ifTrue: [name] ifFalse: [schema, '.', name].! !

!NamedSequence methodsFor: 'SQL' stamp: ''!
reserveSequenceNumbers: anInteger in: aSession for: aTable
	| numberNeeded numberToReserve |
	numberNeeded := anInteger - self availableReservedNumbers.
	numberNeeded <= 0 ifTrue: [^self].
	self
		reserveViaTableSelect: numberNeeded
		in: aSession
		for: aTable.
	[anInteger - self availableReservedNumbers >0 ] whileTrue: [
		numberToReserve := (anInteger - self availableReservedNumbers) min: (self maximumToReserveAtOnce).
		self reserveViaBackupMechanism: numberToReserve in: aSession].! !

!NamedSequence methodsFor: 'SQL' stamp: ''!
reserveViaTableSelect: anInteger in: aSession for: aTable
	| rows command |
	command := self
		tableSelectCommandFor: (self tableToSelectFromFor: aTable)
		in: aSession
		reserving: anInteger.
	command parameters: (Array with: anInteger).
	rows := aSession accessor executeCommand: command.
	rows do: [:each | self reservedNumbers add: each first].! !

!NamedSequence methodsFor: 'accessing' stamp: ''!
schema
	^schema! !

!NamedSequence methodsFor: 'accessing' stamp: ''!
schema: aString
	schema := aString! !

!NamedSequence methodsFor: 'SQL' stamp: ''!
tableSelectCommandFor: aTable in: aSession reserving: anObject
	self subclassResponsibility.! !

!NamedSequence methodsFor: 'SQL' stamp: ''!
tableToSelectFromFor: aTable
	"Sometimes it's inefficient to select sequence values against our own table and we can use a database-specific system table instead to be much faster. Subclasses can override where this is the case"
	^aTable.! !

NamedSequence subclass: #OracleSequence
	instanceVariableNames: 'count sequenceIncrement accessProtect'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!OracleSequence commentStamp: '<historical>' prior: 0!
This supports Oracle sequence using select nextval from sys.allObjects. It also supports getting multiple sequence numbers at once. Going against AllObjects seemed to work much better than going against any real table.

Instance Variables:
	accessProtect	<RecursionLock | Semaphore>	Mutual exclusion on our sequence allocation, since we may need to protect our cache of already-allocated numbers, and prevent two different processes from going to the database simultaneously. This might want to be pushed up to a superclass so it can be shared.
	count	<Number>	How many identifiers have we gotten in a row without going to the database.
	sequenceIncrement	<SmallInteger>	By how much does the sequence in the database increment on each nextVal call.

!

!OracleSequence methodsFor: 'testing' stamp: ''!
alwaysGetFromDb
	^sequenceIncrement = 1! !

!OracleSequence methodsFor: 'As yet unclassified' stamp: ''!
backupReserveCommand: anInteger in: aSession
	| command stream |
	stream := WriteStream on: (String new: 200).
	stream nextPutAll: 'select '.
	stream nextPutAll: self qualifiedName.
	stream nextPutAll: '.nextval from SYS.ALL_OBJECTS '.
	stream nextPutAll: ' WHERE rownum <= '.
	stream
		nextPutAll:
			(aSession useBinding ifTrue: ['?'] ifFalse: [anInteger printString]).
	command := SQLStringSelectCommand new.
	command setSQLString: stream contents.
	command parameters: (Array with: anInteger).
	command session: aSession.
	^command.! !

!OracleSequence methodsFor: 'public' stamp: ''!
creationString

	^'create sequence ', self qualifiedName.! !

!OracleSequence methodsFor: 'public' stamp: ''!
getSequenceValueFromDatabaseFor: aField in: aRow using: anAccessor 
	aRow at: aField put: (self nextSequenceNumber: anAccessor)! !

!OracleSequence methodsFor: 'initalize-release' stamp: ''!
initialize
	super initialize.
	count := 0.
	sequenceIncrement := 1.
	"This seems to hang on the semaphore in ObjectStudio. Fake around it"
	accessProtect := Dialect isObjectStudio ifTrue: [nil] ifFalse: [Semaphore forMutualExclusion].! !

!OracleSequence methodsFor: 'accessing' stamp: ''!
name: aString
	name := aString copyFrom: 1 to: (aString size min: OraclePlatform new maximumLengthOfTableName).! !

!OracleSequence methodsFor: 'private' stamp: ''!
nextSequenceNumber: anAccessor 
	accessProtect critical: 
			[count := count + 1.
			^(self alwaysGetFromDb or: [count \\ sequenceIncrement = 1]) 
				ifTrue: 
					[count := (anAccessor 
								executeSQLString: self nextSequenceNumberSelectString) first 
								first]
				ifFalse: [count]]! !

!OracleSequence methodsFor: 'private' stamp: ''!
nextSequenceNumberSelectString
	^'SELECT ' , self qualifiedName , '.NEXTVAL FROM DUAL'! !

!OracleSequence methodsFor: 'public' stamp: ''!
postWriteAssignSequenceValueFor: aField in: aRow using: aSession! !

!OracleSequence methodsFor: 'private' stamp: ''!
reserveViaBackupMechanism: anInteger in: aSession
	"The regular mechanism didn't work, (probably because there weren't enough rows in the table), try a different technique. Oracle doesn't repeatedly evaluate nextval if we do it multiple times on a line, so instead guess at a table we know should have lots of rows in Oracle and use it"

	| rows command |
	command := self backupReserveCommand: anInteger in: aSession.
	command blockFactor: anInteger.
	rows := aSession accessor executeCommand: command.
	rows do: [:each | self reservedNumbers add: each first].! !

!OracleSequence methodsFor: 'private' stamp: ''!
reserveViaTableSelect: anInteger in: aSession for: aTable
	"In Oracle, there can be performance issues selecting against our tables, which we don't seem to get if we just use SYS.ALLOBJECTS. I'm suspicious the problem is if we have fragmentation of the file containing the table, but regardless, ALLOBJECTS seems consistently fast, so always use it"
	self reserveViaBackupMechanism: anInteger in: aSession.

"	| rows command |
	command := self
		tableSelectCommandFor: aTable
		in: aSession
		reserving: anInteger.
	command parameters: (Array with: anInteger).
	rows := aSession accessor executeCommand: command.
	rows do: [:each | self reservedNumbers add: each first]."! !

!OracleSequence methodsFor: 'accessing' stamp: ''!
sequenceIncrement
	^sequenceIncrement! !

!OracleSequence methodsFor: 'accessing' stamp: ''!
sequenceIncrement: aNumber 
	sequenceIncrement := aNumber! !

!OracleSequence methodsFor: 'private' stamp: ''!
tableSelectCommandFor: aTable in: aSession reserving: anInteger
	| stream |
	(tableSelectCommand notNil and: [tableSelectCommand parameters includes: anInteger]) ifFalse:
			[stream := WriteStream on: (String new: 200).
			stream nextPutAll: 'select '.
			stream nextPutAll: self qualifiedName.
			stream nextPutAll: '.nextval from '.
			stream nextPutAll: aTable name.
			stream nextPutAll: ' WHERE rownum <= '.
			stream nextPutAll: (aSession useBinding ifTrue: ['?'] ifFalse: [anInteger printString]).
			tableSelectCommand := SQLStringSelectCommand new.
			tableSelectCommand parameters: (Array with: anInteger).
			tableSelectCommand setSQLString: stream contents.
			tableSelectCommand blockFactor: anInteger.
			tableSelectCommand session: aSession].
	^tableSelectCommand.! !

NamedSequence subclass: #PGSequence
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!PGSequence commentStamp: '<historical>' prior: 0!
This implements postgresql sequences using nextVal. The mechanism is similar to Oracle sequences, but we don't have to give it a table to select against when all we want is a bunch of numbers, so the implementation is simpler.!

!PGSequence methodsFor: 'As yet unclassified' stamp: ''!
backupReserveCommand: anInteger in: aSession
	| stream command |
	stream := WriteStream on: (String new: 200).
	stream nextPutAll: 'select '.
	GlorpHelper
		print: [:i | 'nextval(''' , self qualifiedName , ''')']
		on: stream
		for: (1 to: anInteger)
		separatedBy: ','.
	command := SQLStringSelectCommand new.
	command setSQLString: stream contents.
	command parameters: (Array with: anInteger).
	command session: aSession.
	^command.! !

!PGSequence methodsFor: 'SQL' stamp: ''!
getSequenceValueFromDatabaseFor: aDatabaseField in: aDatabaseRow using: anAccessor

	| statement rows pk |

	"this could be made smarter by aggregating keys to fetch into a single sql query - maybe later -tb"
	statement := 'select nextval(''',  self qualifiedName, ''')'.

	rows := anAccessor executeSQLString: statement.
	pk := rows first first.

	aDatabaseRow at: aDatabaseField put: pk.! !

!PGSequence methodsFor: 'SQL' stamp: ''!
postWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: aSession 
	"aDatabaseRow at: aDatabaseField
		put: (aSession accessor 
				executeSQLString: 'SELECT CURRVAL(' , self qualifiedName 	, ')')."! !

!PGSequence methodsFor: 'SQL' stamp: ''!
reserveViaBackupMechanism: anInteger in: aSession
	"The regular mechanism didn't work, (probably because there weren't enough rows in the table), try a different technique."

	| command row |
	command := self backupReserveCommand: anInteger in: aSession.
	row := (aSession accessor executeCommand: command) first.
	self reservedNumbers addAll: row.! !

!PGSequence methodsFor: 'As yet unclassified' stamp: ''!
tableSelectCommandFor: aTable in: aSession reserving: anInteger
	| stream |
	(tableSelectCommand notNil and: [tableSelectCommand parameters includes: anInteger]) ifFalse:
			[stream := WriteStream on: (String new: 200).
			stream nextPutAll: 'select nextval('.
			self qualifiedName printOn: stream.
			stream nextPutAll: ') from ' , aTable name.
			stream nextPutAll: ' limit '.
			stream nextPutAll: (aSession useBinding ifTrue: ['?'] ifFalse: [anInteger printString]).
			tableSelectCommand := SQLStringSelectCommand new.
			tableSelectCommand parameters: (Array with: anInteger).
			tableSelectCommand setSQLString: stream contents.
			tableSelectCommand session: aSession].
	^tableSelectCommand.! !

!PGSequence methodsFor: 'SQL' stamp: ''!
tableToSelectFromFor: aTable
	"Sometimes it's inefficient to select sequence values against our own table and we can use a database-specific system table instead to be much faster. Subclasses can override where this is the case"
	^(DatabaseTable named: 'pg_attribute').! !

NamedSequence subclass: #SQLServerSequence
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!SQLServerSequence commentStamp: '<historical>' prior: 0!
This implements SQL Server sequences, which aren't really sequences at all, but identity columns. That means we can't do more than one at a time, and we have to do an extra select to get the value back.!

!SQLServerSequence methodsFor: 'sequencing' stamp: ''!
getSequenceValueFromDatabaseFor: aField in: aRow using: aSession! !

!SQLServerSequence methodsFor: 'testing' stamp: ''!
isIdentityColumn
	^true.! !

!SQLServerSequence methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor 
	aDatabaseRow at: aDatabaseField
		put: ((anAccessor 
				executeSQLString: 'SELECT @@IDENTITY') first atIndex: 1).! !

!SQLServerSequence methodsFor: 'sequencing' stamp: ''!
reserveSequenceNumbers: anInteger in: aSession for: aTable
	"No real sequences here, just identity columns, which we can't pre-allocate"! !

DatabaseSequence subclass: #NullSequence
	instanceVariableNames: ''
	classVariableNames: 'Singleton'
	poolDictionaries: ''
	category: 'Glorp-Database'!

!NullSequence commentStamp: '<historical>' prior: 0!
This is a sequence that doesn't do anything. It's really just a placeholder.!

!NullSequence classSide methodsFor: 'defaults' stamp: ''!
default

	^self new.! !

!NullSequence classSide methodsFor: 'instance creation' stamp: ''!
new

	Singleton isNil ifTrue: [Singleton := self basicNew].
	^Singleton.! !

!NullSequence methodsFor: 'public' stamp: ''!
getSequenceValueFromDatabaseFor: aField in: aRow using: aSession! !

!NullSequence methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aField in: aRow! !

!NullSequence methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aField in: aRow using: aSession! !

!NullSequence methodsFor: 'sequencing' stamp: ''!
preWriteAssignSequenceValueFor: aField in: aRow! !

DatabaseSequence subclass: #SQLite3Sequence
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!SQLite3Sequence commentStamp: '<historical>' prior: 0!
Sequences for SQLite.!

!SQLite3Sequence methodsFor: 'sequencing' stamp: ''!
getSequenceValueFromDatabaseFor: aField in: aRow using: aSession! !

!SQLite3Sequence methodsFor: 'testing' stamp: ''!
isIdentityColumn
	^true.! !

!SQLite3Sequence methodsFor: 'sequencing' stamp: ''!
name: aString
	"We ignore names"! !

!SQLite3Sequence methodsFor: 'sequencing' stamp: ''!
postWriteAssignSequenceValueFor: aDatabaseField in: aDatabaseRow using: anAccessor
	
	| value |
	value := anAccessor connection isNil ifTrue: [nil] ifFalse: [anAccessor connection lastAutoIncrementValue asInteger].
	aDatabaseRow at: aDatabaseField put: value! !

!SQLite3Sequence methodsFor: 'sequencing' stamp: ''!
reserveSequenceNumbers: anInteger in: aSession for: aTable
	"No real sequences here, just identity columns, which we can't pre-allocate"! !

!SQLite3Sequence methodsFor: 'sequencing' stamp: ''!
schema: aString
	"This is SQLite - we don't have schemas"! !

DatabaseSequence subclass: #TableBasedSequence
	instanceVariableNames: 'sequenceTableName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!TableBasedSequence commentStamp: '<historical>' prior: 0!
This is a sequence that gets its values from a sequence table. It doesn't look like it's fully implemented, or used.

Instance Variables:
	sequenceTableName	<String>	The name of the sequence table to use.

!

!TableBasedSequence classSide methodsFor: 'defaults' stamp: ''!
default

	^self new sequenceTableName: 'SEQUENCE'.! !

!TableBasedSequence methodsFor: 'testing' stamp: ''!
isIdentityColumn
	^true.! !

!TableBasedSequence methodsFor: 'accessing' stamp: ''!
sequenceTableName
	"Private - Answer the value of the receiver's ''sequenceTableName'' instance variable."

	^sequenceTableName! !

!TableBasedSequence methodsFor: 'accessing' stamp: ''!
sequenceTableName: aString

	sequenceTableName := aString! !

GlorpDatabaseValueGenerator subclass: #VersionGenerator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!VersionGenerator commentStamp: '<historical>' prior: 0!
An abstract superclass for generators that generate version numbers (rather than sequence numbers).

Subclasses must implement the following messages:
	sequencing
		nextValueFor:in:

!

VersionGenerator subclass: #AdHocVersionGenerator
	instanceVariableNames: 'block numberOfArguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!AdHocVersionGenerator commentStamp: '<historical>' prior: 0!
This is a version generator that takes an arbitrary block and runs it.

Instance Variables:
	block	<BlockClosure>	The block we evaluate to get the next version.
	numberOfArguments	<Integer>	Generated automatically. How many arguments does block take. If zero, we invoke it without our arguments. If not, we give it the arguments to use.

!

!AdHocVersionGenerator classSide methodsFor: 'instance creation' stamp: ''!
block: aBlock
	^self new block: aBlock.! !

!AdHocVersionGenerator methodsFor: 'accessing' stamp: ''!
block
	^block! !

!AdHocVersionGenerator methodsFor: 'accessing' stamp: ''!
block: aBlock
	block := aBlock.
	numberOfArguments := Dialect argumentCountFor: block.! !

!AdHocVersionGenerator methodsFor: 'sequencing' stamp: ''!
nextValueFor: aField in: aRow
	"Get the next value by evaluating the block. If it doesn't care about the field or row, don't pass them as arguments. If it does, then give it both."
	numberOfArguments = 0 ifTrue: [^block value] ifFalse: [^block value: aField value: aRow].! !

VersionGenerator subclass: #TimestampVersionGenerator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!TimestampVersionGenerator commentStamp: '<historical>' prior: 0!
A generator that puts Timestamp now into fields when it is invoked.!

!TimestampVersionGenerator methodsFor: 'sequencing' stamp: ''!
nextValueFor: aField in: aRow
	"If a platform maps the timestamp class to something else (e.g. SQLite maps timestamps to strings as it has no support for time-related types), then it must be allowed to do so here as well.  For most platforms, the wrapping #toTimestamp:for: call will be a noop."

	^aField type platform
		toTimestamp: Dialect timestampNow
		for: aField type! !

!VersionGenerator methodsFor: 'sequencing' stamp: ''!
nextValueFor: aField in: aRow
	self subclassResponsibility.! !

!VersionGenerator methodsFor: 'sequencing' stamp: ''!
preWriteAssignSequenceValueFor: aField in: aRow using: anAccessor
	"Overwrite a pre-existing value in the field if necessary"
	aRow at: aField forcePut: (self nextValueFor: aField in: aRow).! !

VersionGenerator subclass: #VersionNumberVersionGenerator
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!VersionNumberVersionGenerator commentStamp: '<historical>' prior: 0!
This generates version numbers by adding one to the previous value of the designated field.!

!VersionNumberVersionGenerator methodsFor: 'sequencing' stamp: ''!
nextValueFor: aField in: aRow
	| oldValue |
	oldValue := aRow at: aField ifAbsent: [0].
	^(oldValue isNil ifTrue: [0] ifFalse: [oldValue]) + 1! !

Object subclass: #GlorpDictionaryDescription
	instanceVariableNames: 'keyExpression valueExpression'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!GlorpDictionaryDescription commentStamp: '<historical>' prior: 0!
When querying for a dictionary, it's not enough to tell the query to return its collection in a dictionary. We need to be able to describe both the key and the value. So we would pass an instance of this as the collectionType: to a query or mapping.

Instance Variables:
	keyExpression	<GlorpExpression | Block>	The description of the key
	valueExpression	<GlorpExpression | Block>	The description of the value

!

!GlorpDictionaryDescription methodsFor: 'instance creation' stamp: ''!
copyEmpty: anInteger
	"Masquerade as a instance of a collection :-)"
	^Dictionary new: anInteger.! !

!GlorpDictionaryDescription methodsFor: 'accessing' stamp: ''!
keyExpression
	^keyExpression! !

!GlorpDictionaryDescription methodsFor: 'accessing' stamp: ''!
keyExpression: anExpression
	keyExpression := anExpression! !

!GlorpDictionaryDescription methodsFor: 'instance creation' stamp: ''!
new
	"Masquerade as a class"
	^Dictionary new.! !

!GlorpDictionaryDescription methodsFor: 'instance creation' stamp: ''!
new: anInteger
	"Masquerade as a class"
	^Dictionary new: anInteger.! !

!GlorpDictionaryDescription methodsFor: 'accessing' stamp: ''!
valueExpression
	^valueExpression! !

!GlorpDictionaryDescription methodsFor: 'accessing' stamp: ''!
valueExpression: anExpression
	valueExpression := anExpression! !

Object subclass: #GlorpExpression
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!GlorpExpression commentStamp: '<historical>' prior: 0!
This is the superclass of Glorp expression tree nodes. These are used to represent the where clause of a query, or things that describe relationships between objects and/or represent boolean expressions evaluated against them. They are typically created by converting a block into an expression using asGlorpExpression. Then they are used in generating the SQL string to be sent to the database.

Subclasses must implement the following messages:
	accessing
		canHaveBase
	api
		base
		get:withArguments:
	converting
		asJoin
	navigating
		ultimateBaseExpression
	preparing
		asExpressionJoiningSource:toTarget:
		rebuildOn:startingFrom:withOuterScopeBase:
	printing
		printOnlySelfOn:
		printTreeOn:

!

GlorpExpression subclass: #ConstantExpression
	instanceVariableNames: 'value alias'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!ConstantExpression commentStamp: '<historical>' prior: 0!
This represents a literal constant. So, if object field = 2, this would be the (2) part.

Instance Variables:
	alias	<String>	Sometimes in queries we want to retrieve something specific in a column value. So, e.g. "select name as one, 2 as two from ...." Where we get back columns named 'one' and 'two', where the value of 'two' is always 2. The alias tells us what to alias the column for this as, if we're using this as something to be selected.
	value	<Object>	The literal constant value.

!

!ConstantExpression classSide methodsFor: 'instance creation' stamp: ''!
for: anObject

	^self new value: anObject.! !

!ConstantExpression methodsFor: 'accessing' stamp: ''!
alias
	^alias! !

!ConstantExpression methodsFor: 'accessing' stamp: ''!
alias: anObject
	alias := anObject! !

!ConstantExpression methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."

	^self.! !

!ConstantExpression methodsFor: 'converting' stamp: ''!
asGlorpExpressionForDescriptor: aDescriptor
	"This should only get sent to us if we're the top level of an expression. And since we don't have children, we're also the bottom level. This is only useful for the case where the expression is [:each | true] or [:each | false]."

	^(EmptyExpression on: value)
		rebuildOn: (BaseExpression new descriptor: aDescriptor).! !

!ConstantExpression methodsFor: 'converting' stamp: ''!
asGlorpExpressionOn: aBaseExpression
	^self.! !

!ConstantExpression methodsFor: 'testing' stamp: ''!
canBind
	"Return true if this represents a value that can be bound into a prepared statement"
	^true.! !

!ConstantExpression methodsFor: 'accessing' stamp: ''!
canHaveBase
	"Return true if this type of expression can have a base expression on which other things can be built. Doesn't say whether we actually have a valid one or not."
	^false.! !

!ConstantExpression methodsFor: 'accessing' stamp: ''!
field
	^self.! !

!ConstantExpression methodsFor: 'api' stamp: ''!
get: aSymbol withArguments: anArray
	| functionExpression |
	functionExpression := self getFunction: aSymbol withArguments: anArray.
	functionExpression isNil ifFalse: [^functionExpression].
	^RelationExpression
		named: aSymbol
		basedOn: self
		withArguments: anArray! !

!ConstantExpression methodsFor: 'testing' stamp: ''!
isConstantExpression
	^true.! !

!ConstantExpression methodsFor: 'As yet unclassified' stamp: ''!
mappedFields
	^Array with: self.! !

!ConstantExpression methodsFor: 'printing' stamp: ''!
printOnlySelfOn: aStream

	aStream print: value! !

!ConstantExpression methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	self value glorpPrintSQLOn: aStream.! !

!ConstantExpression methodsFor: 'printing' stamp: ''!
printSelectSQLOn: aStream withParameters: aDictionary
	self value glorpPrintSelectSQLOn: aStream.
	alias notNil ifTrue: [
		aStream 
			nextPutAll: ' AS ';
			nextPutAll: alias].! !

!ConstantExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	aStream print: value! !

!ConstantExpression methodsFor: 'printing' stamp: ''!
printUnqualifiedSQLOn: aStream withParameters: aDictionary
	self printSQLOn: aStream withParameters: aDictionary.! !

!ConstantExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	^self.! !

!ConstantExpression methodsFor: 'As yet unclassified' stamp: ''!
table
	^nil.! !

!ConstantExpression methodsFor: 'accessing' stamp: ''!
ultimateBaseExpression
	^nil.! !

!ConstantExpression methodsFor: 'accessing' stamp: ''!
value
	^value! !

!ConstantExpression methodsFor: 'accessing' stamp: ''!
value: anObject
	value := anObject! !

!ConstantExpression methodsFor: 'accessing' stamp: ''!
valueIn: aDictionary
	^value! !

!ConstantExpression methodsFor: 'mapping' stamp: ''!
valueInBuilder: aBuilder
	^value.! !

!ConstantExpression methodsFor: 'As yet unclassified' stamp: ''!
valueInBuilder: aBuilder as: anExpression
	^value.! !

GlorpExpression subclass: #EmptyExpression
	instanceVariableNames: 'base value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!EmptyExpression commentStamp: '<historical>' prior: 0!
This represents the null expression, but more polymorphically than just sticking nil in the where clause. We also use this as a short-circuiting mechanism if we're given a where clause that is either just literal true or literal false. So if you end up creating a query with a literal false, we will never go to the database, we'll just return an empty collection as the result. This is more useful than you might think.

Instance Variables:
	base	<GlorpExpression>	Our base expression, which we'll use in case someone ANDs or ORs something onto us.
	value	<Boolean| nil > If we are true or false, then this result may end up being short-circuited. If it's nil, then we're just an empty where clause.

!

!EmptyExpression classSide methodsFor: 'instance creation' stamp: ''!
on: aValue
	^self new value: aValue.! !

!EmptyExpression methodsFor: 'api' stamp: ''!
AND: anExpression

	anExpression isNil ifTrue: [^self].
	^anExpression asGlorpExpression.! !

!EmptyExpression methodsFor: 'api' stamp: ''!
OR: anExpression

	anExpression isNil ifTrue: [^self].
	^anExpression! !

!EmptyExpression methodsFor: 'accessing' stamp: ''!
base
	^base.! !

!EmptyExpression methodsFor: 'accessing' stamp: ''!
base: aBaseExpression
	base := aBaseExpression.! !

!EmptyExpression methodsFor: 'testing' stamp: ''!
canHaveBase
	^true.! !

!EmptyExpression methodsFor: 'testing' stamp: ''!
canPrint
	"Return true if we can print into a where clause"
	^false.! !

!EmptyExpression methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Iterate over the expression tree"

	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	base do: aBlock skipping: aSet.
	aBlock value: self.! !

!EmptyExpression methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	value := true.! !

!EmptyExpression methodsFor: 'testing' stamp: ''!
isEmptyExpression
	^true.! !

!EmptyExpression methodsFor: 'accessing' stamp: ''!
isFalse
	^value not! !

!EmptyExpression methodsFor: 'accessing' stamp: ''!
isTrue
	^value! !

!EmptyExpression methodsFor: 'As yet unclassified' stamp: ''!
printSQLOn: aCommand withParameters: aDictionary
	^self.! !

!EmptyExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	aStream nextPutAll: 'empty expression'.! !

!EmptyExpression methodsFor: 'iterating' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	| copy |
	self == anExpression ifTrue: [^aBaseExpression].
	self base == ultimateBaseExpression ifTrue: [^self].
	copy := self copy.
	copy base: aBaseExpression.
	^copy.! !

!EmptyExpression methodsFor: 'As yet unclassified' stamp: ''!
tablesForANSIJoin
	^#().! !

!EmptyExpression methodsFor: 'navigating' stamp: ''!
ultimateBaseExpression

	base isNil ifTrue: [base := BaseExpression new].
	^base ultimateBaseExpression.! !

!EmptyExpression methodsFor: 'accessing' stamp: ''!
value: aValue
	"a value is expected to be nil, true, or false. we treat nil as true"
	value := aValue isNil ifTrue: [true] ifFalse: [aValue].! !

!EmptyExpression methodsFor: 'testing' stamp: ''!
willPrintAsWhereClause
	^false.! !

GlorpExpression subclass: #ExpressionGroup
	instanceVariableNames: 'children'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!ExpressionGroup commentStamp: '<historical>' prior: 0!
This isn't really an expression, in that it can never occur due to parsing. It's a way of grouping several expressions together so that we can process them together, essentially making sure that the iteration methods will loop over all the expressions, but only do each node once, even if it occurs in multiple expressions.  This is used in processing order expressions to figure out what tables and join expressions we need.

Because it is only used in transient ways, it probably doesn't implement all the required operations for normal expression usage.
!

!ExpressionGroup classSide methodsFor: 'instance creation' stamp: ''!
with: anExpression
	^self new add: anExpression.! !

!ExpressionGroup methodsFor: 'accessing' stamp: ''!
add: anExpression
	anExpression isNil ifTrue: [^self].
	children add: anExpression.! !

!ExpressionGroup methodsFor: 'accessing' stamp: ''!
addAll: anExpressionCollection
	anExpressionCollection isNil ifTrue: [^self].
	children addAll: anExpressionCollection.! !

!ExpressionGroup methodsFor: 'accessing' stamp: ''!
children
	^children.! !

!ExpressionGroup methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Iterate over the expression tree. Keep track of who has already been visited, so we don't get trapped in cycles or visit nodes twice."

	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	children do: 	[:each | each do: aBlock skipping: aSet].
	aBlock value: self.! !

!ExpressionGroup methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	children := OrderedCollection new.! !

!ExpressionGroup methodsFor: 'accessing' stamp: ''!
isEmpty
	^self size = 0.! !

!ExpressionGroup methodsFor: 'printing' stamp: ''!
printTreeOn: aStream

	children do: [:each |
		aStream nextPut: $(.
		each printOn: aStream.
		aStream nextPut: $).
		aStream cr].! !

!ExpressionGroup methodsFor: 'accessing' stamp: ''!
queryLevelBaseExpression
	^nil.! !

!ExpressionGroup methodsFor: 'accessing' stamp: ''!
size
	^children size.! !

GlorpExpression subclass: #FieldExpression
	instanceVariableNames: 'field base alias'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!FieldExpression commentStamp: '<historical>' prior: 0!
This is an expression representing a field in a table. It could be used, for example, if we were writing a data level expression.
   getTable: 'foo') getField: 'bar'

Instance Variables:
	alias	<Object>	Sometimes in queries we want to retrieve values in columns other than what they are normally named. So we can give the field an alias, and generate something like "select id, name as somethingElse"
	base	<GlorpExpression>	The expression this is based on, which ought to be something which is associated with the table that contains our field.
	field	<DatabaseField>	The field we represent in the expression.

!

!FieldExpression classSide methodsFor: 'instance creation' stamp: ''!
forField: aField basedOn: anObjectExpression

	^self new field: aField base: anObjectExpression; yourself! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
alias
	^alias! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
alias: aString
	"What column name, if anything, should we print for 'as X' when retrieving"
	alias := aString! !

!FieldExpression methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."
	| newTarget |
	newTarget := (target allControlledTables includes: field table) 
		ifTrue: [target]
		ifFalse: [base asExpressionJoiningSource: source toTarget: target].
	^newTarget getField: field.! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
base

	^base.! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
basicField
	^field.! !

!FieldExpression methodsFor: 'testing' stamp: ''!
canBeUsedForRetrieve
	"Return true if this is a valid argument for a retrieve: clause"
	^true.! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
canHaveBase
	"Return true if this type of expression can have a base expression on which other things can be built. Doesn't say whether we actually have a valid one or not."
	^true.! !

!FieldExpression methodsFor: 'As yet unclassified' stamp: ''!
convertedDbValueOf: anObject
	"We don't do any conversion"
	^anObject! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
converterForStType: aClass
	^self field converterForStType: aClass.! !

!FieldExpression methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Iterate over the expression tree"

	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	base do: aBlock skipping: aSet.
	aBlock value: self.! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
field
	^base translateField: field! !

!FieldExpression methodsFor: 'initializing' stamp: ''!
field: aField base: anObjectExpression

	field := aField.
	base := anObjectExpression.! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
fieldExpression
	^self.! !

!FieldExpression methodsFor: 'api' stamp: ''!
get: aSymbol withArguments: anArray

	| functionExpression |
	functionExpression := self getFunction: aSymbol withArguments: anArray.
	functionExpression isNil ifFalse: [^functionExpression].
	anArray isEmpty ifTrue: [self error: 'Field expressions do not have attributes'].
	^RelationExpression named: aSymbol basedOn: self withArguments: anArray.! !

!FieldExpression methodsFor: 'testing' stamp: ''!
hasField
	^true.! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
impliedSmalltalkType
	^self field impliedSmalltalkType.! !

!FieldExpression methodsFor: 'mapping' stamp: ''!
mappedFields
	^Array with: self field.! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
name
	^field name.! !

!FieldExpression methodsFor: 'printing' stamp: ''!
printOnlySelfOn: aStream 
	base printsTable
		ifTrue: [field printUnqualifiedSQLOn: aStream withParameters: #()]
		ifFalse: [field printSQLOn: aStream withParameters: #()]! !

!FieldExpression methodsFor: 'printing SQL' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	self field printSQLOn: aStream withParameters:aDictionary.! !

!FieldExpression methodsFor: 'printing SQL' stamp: ''!
printSelectSQLOn: aStream withParameters: aDictionary
	self printSQLOn: aStream withParameters:aDictionary.
	alias notNil ifTrue: [aStream nextPutAll: ' AS '; nextPutAll: alias].! !

!FieldExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	base printOn: aStream.
	aStream nextPut: $..
	base printsTable 
		ifTrue: [field printUnqualifiedSQLOn: aStream withParameters: #()]
		ifFalse: [field printSQLOn: aStream withParameters: #()]! !

!FieldExpression methodsFor: 'printing SQL' stamp: ''!
printUnqualifiedSQLOn: aStream withParameters: aDictionary
	self field printUnqualifiedSQLOn: aStream withParameters:aDictionary.! !

!FieldExpression methodsFor: 'navigating' stamp: ''!
queryLevelBaseExpression
	^base queryLevelBaseExpression.! !

!FieldExpression methodsFor: 'As yet unclassified' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	| result |
	self == anExpression ifTrue: [^aBaseExpression].
	result := (base rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression) getField: field.
	result alias: self alias.
	^result.! !

!FieldExpression methodsFor: 'copying' stamp: ''!
setTable: aTable
	"Modify (do not copy) ourselves to use aTable. Presumably aTable is an aliased version of our table"
	field := field withTable: aTable.! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
table
	^self field table.! !

!FieldExpression methodsFor: 'As yet unclassified' stamp: ''!
tableForANSIJoin
	^self field table.! !

!FieldExpression methodsFor: 'preparing' stamp: ''!
tables

	^base tables.! !

!FieldExpression methodsFor: 'preparing' stamp: ''!
tablesToPrint

	^Array with: self table! !

!FieldExpression methodsFor: 'accessing' stamp: ''!
type
	^self field type.! !

!FieldExpression methodsFor: 'navigating' stamp: ''!
ultimateBaseExpression
	^base ultimateBaseExpression.! !

!FieldExpression methodsFor: 'mapping' stamp: ''!
valueInBuilder: anElementBuilder
	^self valueInBuilder: anElementBuilder as: self field.! !

!FieldExpression methodsFor: 'mapping' stamp: ''!
valueInBuilder: anElementBuilder as: anExpression
	^anElementBuilder valueOfField: anExpression.! !

!FieldExpression methodsFor: 'copying' stamp: ''!
withTable: aTable
	"Modify (do not copy) ourselves to use aTable. Presumably aTable is an aliased version of our table"
	field := field withTable: aTable.! !

GlorpExpression subclass: #FunctionExpression
	instanceVariableNames: 'function base alias symbol type'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

FunctionExpression class
	instanceVariableNames: 'functions'!

!FunctionExpression commentStamp: '<historical>' prior: 0!
This represents a database function or other modifier. For example, conversion to upper or lower case, or the ascending/descending modifier in order by clauses. At the moment it is hard-coded to to handle only the descending modifier and does not handle e.g. function arguments, functions that differ between databases, functional syntax ( as opposed to postfix). One would probably define subclasses to handle these cases, but this is the simplest thing that could possibly work for the current functionality.

Instance Variables:
	function	<String>	The string we print to the database
	base <GlorpExpression> The thing we're a function of.
	alias <String> If we're to be selected as a field, what should we be selcted as. Think SELECT t1.A + t1.B AS FRED.
	symbol <Symbol> The original symbol used to create us. Helpful if recreating onto a different platform where the DB string can be different.
	type <GlorpDatabaseType> Optional. If we have a type defined, then we can record it here. Otherwise, we will return the type of the field we are operating on.
	
!

FunctionExpression class
	instanceVariableNames: 'functions'!

FunctionExpression subclass: #DualRoleFunction
	instanceVariableNames: 'arguments whereClauseVersion selectListVersion'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!DualRoleFunction commentStamp: '<historical>' prior: 0!
This is a function that prints in different ways in a field list than it does in a where clause, typically because use of the where clause version isn't supported in field lists.!

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
arguments
	^arguments.! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
arguments: anArray
	arguments := anArray.
	whereClauseVersion arguments: anArray.
	selectListVersion arguments: anArray.! !

!DualRoleFunction methodsFor: 'converting' stamp: ''!
asField
	^self.! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
base: anExpression
	base := anExpression.
	"whereClauseVersion base: anExpression."
	selectListVersion base: anExpression.! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
basicArguments: anArray
	arguments := anArray.! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
basicBase: anExpression
	base := anExpression.! !

!DualRoleFunction methodsFor: 'copying' stamp: ''!
copy
	| new |
	new := super copy.
	"We need to make sure we copy these before the base gets set. This is horribly ugly."
	new selectListVersion: selectListVersion copy.
	new whereClauseVersion: whereClauseVersion copy.
	^new.! !

!DualRoleFunction methodsFor: 'copying' stamp: ''!
createFor: argumentList withBase: anExpression symbol: aSymbol
	| new newBase newArgument newFunction |
	new := super createFor: argumentList withBase: anExpression symbol: aSymbol.
	"For the where clause version, we've got a dualRoleFunction. We can't just rebuild the whole thing, so rebuild the children, and then re-assemble it recreating from the template."
	newBase := new whereClauseVersion base isNil ifTrue: [anExpression] ifFalse: [new whereClauseVersion base rebuildOn: anExpression].
	newArgument := new whereClauseVersion argument rebuildOn: anExpression.
	newFunction := self whereClauseVersion copy.
	newFunction symbol: aSymbol.
	newFunction base: newBase.
	newFunction argument: newArgument.
	new whereClauseVersion: newFunction.
	"Don't try to rebuild the selectListVersion at all, because at least for the motivating case we have, it's just a shell, not a full blown expression, and it will already have been copied."
	^new.! !

!DualRoleFunction methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Iterate over the expression tree. Keep track of who has already been visited, so we don't get trapped in cycles or visit nodes twice."

	(aSet includes: self) ifTrue: [^self].
	super do: aBlock skipping: aSet.
	arguments do: [:each |
		each do: aBlock skipping: aSet].! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
field
	^whereClauseVersion field.! !

!DualRoleFunction methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	whereClauseVersion printSQLOn: aStream withParameters: aDictionary.! !

!DualRoleFunction methodsFor: 'printing' stamp: ''!
printSelectSQLOn: aStream withParameters: aDictionary
	selectListVersion printSelectSQLOn: aStream withParameters: aDictionary.! !

!DualRoleFunction methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	| rebuilt |
	self == anExpression ifTrue: [^aBaseExpression].
	"Rebuilding just one of them will rebuild our whole structure, since it all comes from the same definition."
	rebuilt := whereClauseVersion rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression.
	rebuilt == whereClauseVersion ifTrue: [^self].

	rebuilt arguments: (self arguments collect: [:each | each rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression]).
	^rebuilt.! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
selectListVersion
	^selectListVersion.! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
selectListVersion: aFunctionExpression
	selectListVersion := aFunctionExpression.! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
symbol: aSymbol
	symbol := aSymbol.
	whereClauseVersion symbol: aSymbol.
	selectListVersion symbol: aSymbol.! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
type: aSymbol
	super type: aSymbol.
	whereClauseVersion type: aSymbol.
	selectListVersion type: aSymbol.! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
whereClauseVersion
	^whereClauseVersion.! !

!DualRoleFunction methodsFor: 'accessing' stamp: ''!
whereClauseVersion: aFunctionExpression
	whereClauseVersion := aFunctionExpression.! !

!FunctionExpression classSide methodsFor: 'private' stamp: ''!
basicFunctions
	functions isNil ifTrue: [functions := self createBasicFunctionsFor: nil].
	^functions.! !

!FunctionExpression classSide methodsFor: 'private' stamp: ''!
createBasicFunctionsFor: anObject
	| exists notExists |
	functions := IdentityDictionary new.
	functions
		at: #descending put: (PostfixFunction named: 'DESC');
		at: #distinct put: (PrefixFunction named: 'DISTINCT');
		at: #max put: (PrefixFunction named: 'MAX');
		at: #min put: (PrefixFunction named: 'MIN');
		at: #sum put: (PrefixFunction named: 'SUM');
		at: #avg put: (PrefixFunction named: 'AVG');
		at: #average put: (PrefixFunction named: 'AVG');
		at: #toChar: put: (PrefixFunction named: 'TO_CHAR');
		at: #not put: (PrefixFunction named: 'NOT');
		at: #, put: (InfixFunction named: '||');
		at: #countStar put: (StandaloneFunction named: 'COUNT(*)');
		at: #count put: (PrefixFunction named: 'COUNT');
		at: #between:and: put: (InfixFunction named: #('BETWEEN' 'AND'));
		at: #isNil
			put:
				((RelationExpression new)
						relation: #=;
						rightChild: nil asGlorpExpression);
		at: #notNil
			put:
				((RelationExpression new)
						relation: #<>;
						rightChild: nil asGlorpExpression);
		at: #isNIL
			put:
				((RelationExpression new)
						relation: #=;
						rightChild: nil asGlorpExpression);
		at: #notNIL
			put:
				((RelationExpression new)
						relation: #<>;
						rightChild: nil asGlorpExpression).
	exists := InfixFunction named: 'EXISTS'.
	anObject isNil ifFalse: [exists type: anObject boolean].
	functions at: #exists: put: exists.
	notExists := InfixFunction named: 'NOT EXISTS'.
	anObject isNil ifFalse: [notExists type: anObject boolean].
	functions at: #notExists: put: notExists.

	functions
		at: #asUppercase put: (PrefixFunction named: 'UPPER');
		at: #asLowercase put: (PrefixFunction named: 'LOWER');
		at: #replace:with: put: (PrefixFunction named: 'REPLACE');
		at: #cast: put: ((Cast named: 'CAST') separator: ' AS ');
		at: #copyFrom:to: put: (SubstringFunction named: 'SUBSTR').

	^functions.! !

!FunctionExpression classSide methodsFor: 'instance creation' stamp: ''!
for: aSymbol withArguments: anArray basedOn: anExpression
	| base system |
	base := anExpression ultimateBaseExpression.
	system := base isNil ifTrue: [nil] ifFalse: [base system].	
	^self for: aSymbol withArguments: anArray basedOn: anExpression system: system.! !

!FunctionExpression classSide methodsFor: 'instance creation' stamp: ''!
for: aSymbol withArguments: anArray basedOn: anExpression system: system
	| functionTemplate newFunction |
	functionTemplate := (system isNil
		ifTrue: [self basicFunctions]
		ifFalse: [system platform functions]) at: aSymbol ifAbsent: [^nil].
	newFunction := functionTemplate
		createFor: anArray
		withBase: anExpression
		symbol: aSymbol.
	^newFunction.! !

!FunctionExpression classSide methodsFor: 'private' stamp: ''!
initialize

	self resetFunctions.! !

!FunctionExpression classSide methodsFor: 'private' stamp: ''!
named: aString
	"Used for creating template instances only"

	^self new function: aString.! !

!FunctionExpression classSide methodsFor: 'private' stamp: ''!
resetFunctions
	"self resetFunctions"
	functions := nil.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
alias
	^alias! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
alias: aString
	alias := aString! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
arguments
	^#().! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
arguments: anArray
	^self.! !

!FunctionExpression methodsFor: 'converting' stamp: ''!
asDistinctFieldOn: anExpression
	self isDistinct ifTrue: [^self asGlorpExpressionOn: anExpression].
	^super asDistinctFieldOn: anExpression.! !

!FunctionExpression methodsFor: 'api' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."

	| newBase |
	newBase := base asExpressionJoiningSource: source toTarget: target.
	^self copy function: function; base: newBase.! !

!FunctionExpression methodsFor: 'converting' stamp: ''!
asNonDistinctField
	^self isDistinct ifTrue: [base] ifFalse: [self].! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
base
	^base.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
base: anExpression
	base := anExpression.! !

!FunctionExpression methodsFor: 'testing' stamp: ''!
canBeUsedForRetrieve
	"Return true if this is a valid argument for a retrieve: clause"
	^true.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
canHaveBase
	^true.! !

!FunctionExpression methodsFor: 'iterating' stamp: ''!
convertValueOf: anObject in: aDictionary
	| translated |

	translated := anObject isGlorpExpression 
		ifTrue: [anObject valueIn: aDictionary]
		ifFalse: [anObject].
	translated isGlorpExpression ifTrue: [^translated].
	^self base convertedDbValueOf: translated.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
convertedDbValueOf: anObject
	"Not clear what should happen here, but let's just pass it through for the moment"
	^(self converterForStType: anObject class) convert: anObject toDatabaseRepresentationAs: self type.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
converterForStType: aClass
	^self type notNil 
		ifTrue: [self type converterForStType: aClass]
		ifFalse: [self field converterForStType: aClass].! !

!FunctionExpression methodsFor: 'copying' stamp: ''!
createFor: arguments
	| new |
	new := self copy.
	new arguments: arguments.
	new type: type.
	^new.! !

!FunctionExpression methodsFor: 'copying' stamp: ''!
createFor: arguments withBase: anExpression symbol: aSymbol
	| new |
	new := self copy.
	new base: anExpression.
	new symbol: aSymbol.
	new type: type.
	new arguments: arguments.
	^new.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
defaultType
	"We don't know our type. Return a basic numeric type as being as good a guess as any."
	^base system platform numeric! !

!FunctionExpression methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Iterate over the expression tree. Keep track of who has already been visited, so we don't get trapped in cycles or visit nodes twice."

	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	base do: aBlock skipping: aSet.
	aBlock value: self.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
field

	^base field.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
fieldExpression

	^base fieldExpression.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
function: aString

	function := aString.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
function: aString arguments: anArray

	self function: aString.! !

!FunctionExpression methodsFor: 'api' stamp: ''!
get: aSymbol withArguments: anArray
	| functionExpression |
	functionExpression := self getFunction: aSymbol withArguments: anArray.
	functionExpression isNil ifFalse: [^functionExpression].
	^RelationExpression
		named: aSymbol
		basedOn: self
		withArguments: anArray! !

!FunctionExpression methodsFor: 'converting' stamp: ''!
hasField
	"If we correspond to a single field, return it. This is a bit of a hack having this only here, because for some functions it doesn't work."
	^base hasField.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
impliedSmalltalkType
	type isNil ifFalse: [^type impliedSmalltalkType].
	^self field impliedSmalltalkType.! !

!FunctionExpression methodsFor: 'testing' stamp: ''!
isDistinct
	^function = 'DISTINCT'.! !

!FunctionExpression methodsFor: 'testing' stamp: ''!
isGenerated
	^false.! !

!FunctionExpression methodsFor: 'testing' stamp: ''!
isSameAliasedFieldAs: aFieldOrFunction
	^self == aFieldOrFunction.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
mappedFields
	^Array with: self.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
name
	^function.! !

!FunctionExpression methodsFor: 'printing' stamp: ''!
printOnlySelfOn: aStream
	aStream nextPutAll: function.! !

!FunctionExpression methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	
	self subclassResponsibility.! !

!FunctionExpression methodsFor: 'printing' stamp: ''!
printSelectSQLOn: aStream withParameters: aDictionary
	self printSQLOn: aStream withParameters: aDictionary.
	alias notNil ifTrue: [aStream nextPutAll: ' AS '; nextPutAll: alias].! !

!FunctionExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	aStream 
		nextPutAll: function! !

!FunctionExpression methodsFor: 'testing' stamp: ''!
printsIntoJoins
	"Certain functions should be printed into joins as well as into select lists (e.g. UPPER). However, there are others that don't (at least I don't think so) make any sense in joins, e.g. DISTINCT. Differentiate the two. Right now the only one I can think of that doesn't make sense is distinct. It's not clear to me that I'm not just hacking my way through this"
	^self isDistinct not.! !

!FunctionExpression methodsFor: 'navigating' stamp: ''!
queryLevelBaseExpression

	^base queryLevelBaseExpression.! !

!FunctionExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	| rebuilt |
	self == anExpression ifTrue: [^aBaseExpression].
	rebuilt := (base rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression) getFunction: self symbol withArguments: (self arguments collect: [:each | each rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression]) in: aBaseExpression system.
	rebuilt alias: self alias.
	^rebuilt.! !

!FunctionExpression methodsFor: 'preparing' stamp: ''!
rewriteBooleanEquality: relation
	base := base rewriteBooleanEquality: relation.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
symbol
	^symbol! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
symbol: aSymbol
	symbol := aSymbol! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
table
	(self hasField and: [self field notNil]) ifTrue: [^self field table].
	"If we don't have a field, then maybe we have a relationship mapping or a base, so get the descriptor's primary table. Relevant if we have a function like EXISTS"
	^self base table.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
tablesToPrint
	^#().! !

!FunctionExpression methodsFor: 'compatibility' stamp: ''!
translateField: aField
	^base translateField: aField.! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
type
	| field |
	type isNil ifFalse: [^type].
	field := self field.
	^field isNil ifTrue: [self defaultType] ifFalse: [field type].! !

!FunctionExpression methodsFor: 'accessing' stamp: ''!
type: aGlorpDatabaseType
	type := aGlorpDatabaseType! !

!FunctionExpression methodsFor: 'navigating' stamp: ''!
ultimateBaseExpression

	^base ultimateBaseExpression.! !

!FunctionExpression methodsFor: 'mapping' stamp: ''!
valueInBuilder: anElementBuilder
	^self base valueInBuilder: anElementBuilder as: self.! !

!FunctionExpression methodsFor: 'As yet unclassified' stamp: ''!
valueInBuilder: anElementBuilder as: anExpression
	^self base valueInBuilder: anElementBuilder as: anExpression.! !

!FunctionExpression methodsFor: 'copying' stamp: ''!
withTable: aTable
	"For a function expression, we shouldn't need to do anything here. It will have already been rebuilt on a base that has aliases"
	^self.

"	| copy newBase |
	newBase := BaseExpression new.
	newBase descriptor: base descriptor.
	copy := self rebuildOn: newBase.
	copy fieldExpression setTable: aTable.
	^copy."! !

FunctionExpression subclass: #InfixFunction
	instanceVariableNames: 'argument arguments functionParts'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!InfixFunction commentStamp: '<historical>' prior: 0!
This is a function that is infix. That is, it prints its name in between its arguments. For example, + for string concatenation, || for logical OR. 

Instance Variables:
	arguments	<Collection of GlorpExpression>	Our post-arguments (the first one being the base).
	functionParts <Array of: String> The parts of our name. For a two-argument function, this is just the same as an array containing the name, but for one with more (e.g. BETWEEN AND) it contains the different portions of the name.
!

!InfixFunction classSide methodsFor: 'instance creation' stamp: ''!
named: aStringOrArray
	| instance trimmedName |
	instance := self new.
	aStringOrArray isString
		ifTrue: [instance function: aStringOrArray]
		ifFalse: [
			instance functionParts: aStringOrArray.
			trimmedName := aStringOrArray copyFrom: 2 to: aStringOrArray size.
			instance function: (trimmedName inject: aStringOrArray first into: [:sum :each | sum, ' ', each])].
	^instance.! !

!InfixFunction methodsFor: 'accessing' stamp: ''!
argument
	^arguments first.! !

!InfixFunction methodsFor: 'accessing' stamp: ''!
argument: anObject
	arguments := Array with: anObject.! !

!InfixFunction methodsFor: 'accessing' stamp: ''!
arguments
	^arguments.! !

!InfixFunction methodsFor: 'accessing' stamp: ''!
arguments: anArray
	arguments := anArray.! !

!InfixFunction methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."
	
	| newArguments |
	newArguments := arguments collect: [:each | each asExpressionJoiningSource: source toTarget: target].
	^(super asExpressionJoiningSource: source toTarget: target) arguments: newArguments.! !

!InfixFunction methodsFor: 'converting' stamp: ''!
asField
	"If we correspond to a single field, return it. This is a bit of a hack having this only here, because for some functions it doesn't work."
	^self.! !

!InfixFunction methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Iterate over the expression tree. Keep track of who has already been visited, so we don't get trapped in cycles or visit nodes twice."

	(aSet includes: self) ifTrue: [^self].
	super do: aBlock skipping: aSet.
	arguments do: [:each | each do: aBlock skipping: aSet].! !

!InfixFunction methodsFor: 'accessing' stamp: ''!
functionParts
	functionParts isNil ifTrue: [functionParts := Array with: function].
	^functionParts.! !

!InfixFunction methodsFor: 'accessing' stamp: ''!
functionParts: anArray
	functionParts := anArray.! !

!InfixFunction methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	| translated |
	base printSQLOn: aStream withParameters: aDictionary.
	self printsIntoJoins ifFalse: [^self].
	1 to: self functionParts size do: [:i |
		aStream 
			nextPut: $ ;
			nextPutAll: (self functionParts at: i);
			nextPut: $ .
		i > arguments size ifFalse: [ 
			translated := self convertValueOf: (arguments at: i) in: aDictionary.
			translated isGlorpExpression
				ifTrue: [translated printSQLOn: aStream withParameters: aDictionary]
				ifFalse: [self type print: translated on: aStream]]].! !

!InfixFunction methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	base printOn: aStream.
	aStream
		nextPutAll: ' ', function, ' '.
	arguments printOn: aStream.! !

!InfixFunction methodsFor: 'printing' stamp: ''!
printUnqualifiedSQLOn: aStream withParameters: aDictionary
	base printUnqualifiedSQLOn: aStream withParameters: aDictionary.
	self functionParts with: arguments do: [:functionPart :arg |
		aStream 
			nextPut: $ ;
			nextPutAll: functionPart;
			nextPut: $ .
		arg printUnqualifiedSQLOn: aStream withParameters: aDictionary].! !

!InfixFunction methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	| rebuilt newArgs |
	self == anExpression ifTrue: [^aBaseExpression].
	base == aBaseExpression ifTrue: [^self].
	newArgs := arguments collect: [:each | each rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression].
	rebuilt := (base rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression) get: self symbol withArguments: newArgs.
	rebuilt alias: self alias.
	^rebuilt.! !

FunctionExpression subclass: #PostfixFunction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!PostfixFunction commentStamp: '<historical>' prior: 0!
This represents a function that prints postfix, meaning its name comes after its argument. For example, DESC for a descending sort. That's maybe stretching the definition of a function, but that's how it's implemented for purposes of Glorp SQL definition.
!

!PostfixFunction methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	
	base printSQLOn: aStream withParameters: aDictionary.
	self printsIntoJoins ifFalse: [^self].
	aStream nextPutAll: ' '; nextPutAll: function.! !

!PostfixFunction methodsFor: 'printing' stamp: ''!
printUnqualifiedSQLOn: aStream withParameters: aDictionary
	
	base printUnqualifiedSQLOn: aStream withParameters: aDictionary.
	aStream nextPutAll: ' '; nextPutAll: function.! !

FunctionExpression subclass: #PrefixFunction
	instanceVariableNames: 'arguments separator argumentModifierBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!PrefixFunction commentStamp: '<historical>' prior: 0!
This represents a function which prints prefix, printing its name, followed by the arguments in parentheses. This is the most common sort of function, e.g. MIN(), MAX(), DISTINCT, TO_CHAR 

Instance Variables:
	arguments	<(SequenceableCollection of: GlorpExpression)>	Our arguments (in addition to our base).
	separator	<String>	What we will use to separate the arguments. By default, a comma.

!

PrefixFunction subclass: #Case
	instanceVariableNames: 'argumentTypes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!Case commentStamp: '<historical>' prior: 0!
This represents a function which prints prefix, printing CASE, followed by the argument in parentheses, and then the rest of the MS SQL Server CASE statement, with a single test condition (WHEN) and a default value (ELSE):

CASE (arg1)
	WHEN arg2 THEN arg3 
	ELSE arg4
END AS arg5

For example:

CASE ('SELECT count(*) FROM aTable')
	WHEN '0' THEN 0 
	ELSE 1 
END AS arg3
!

!Case methodsFor: 'As yet unclassified' stamp: ''!
argumentTypes: anArray 
	argumentTypes := anArray.! !

!Case methodsFor: 'initializing' stamp: ''!
initialize
	super initialize.
	function := 'CASE'.
	separator := ' '.
	argumentTypes := OrderedCollection new: 3.! !

!Case methodsFor: 'printing' stamp: ''!
printBasicSQLOn: aStream withParameters: aDictionary forSelect: aBoolean
	"Example: CASE (base?)
		WHEN arg1 THEN arg2 
		 ELSE arg3
	END AS alias"
	
	aStream
		nextPutAll: function;
		nextPut: $(.
	aBoolean
		ifTrue: [base printSelectSQLOn: aStream withParameters: aDictionary]
		ifFalse: [base printSQLOn: aStream withParameters: aDictionary].
	aStream nextPutAll: ')'.
	1 to: arguments size do: [:i |
		| keyword argument argumentType |
		keyword := #(' WHEN ' ' THEN ' ' ELSE ') at: i.
		argument := arguments at: i.
		argumentType := argumentTypes at: i.
		aStream nextPutAll: keyword.
		"Our printing of constants with types is unfortunately tied up with relation expressions, so we shortcircuit some of the logic and just have the types print directly. This will not let us e.g. bind parameters to this function, but we don't expect to be doing that right now."
		argumentType print: argument on: aStream].

	aStream nextPutAll: ' END '.
	alias notNil
		ifTrue: [aStream
				nextPutAll: ' AS ';
				nextPutAll: alias]! !

PrefixFunction subclass: #Cast
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!Cast commentStamp: '<historical>' prior: 0!
This is an attempt to allow a cast to be used on constants, as e.g.
   CAST value: '1-2-3' as: platform date
It doesn't work yet, since printing of constants doesn't like having functions in the way - it conflicts with type conversion, and looks like it'll take a bit of work to fix.!

!Cast classSide methodsFor: 'instance creation' stamp: ''!
value: anObject as: aDatabaseType
	self error: 'this is work in progress, you probably shouldn''t use it yet'.
	(anObject isNil or: [anObject isGlorpExpression and: [anObject isConstantExpression and: [anObject value isNil]]])
		ifTrue: [^anObject].
	^self new base: anObject asGlorpExpression; arguments: (Array with: aDatabaseType asGlorpExpression).! !

!Cast methodsFor: 'initialize-release' stamp: ''!
initialize
	super initialize.
	function := 'CAST'.
	separator := ' AS '.! !

!Cast methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	^super rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression.
"	| rebuilt |
	self == anExpression ifTrue: [^aBaseExpression].
	rebuilt := (base rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression).
	rebuilt := self class value: base as: arguments first.
	rebuilt alias: self alias.
	^rebuilt."! !

!Cast methodsFor: 'accessing' stamp: ''!
type
	^arguments last value.! !

!PrefixFunction methodsFor: 'accessing' stamp: ''!
argumentModifierBlock
	^argumentModifierBlock! !

!PrefixFunction methodsFor: 'accessing' stamp: ''!
argumentModifierBlock: aBlock
	argumentModifierBlock := aBlock! !

!PrefixFunction methodsFor: 'accessing' stamp: ''!
arguments
	^arguments.! !

!PrefixFunction methodsFor: 'accessing' stamp: ''!
arguments: aCollection
	"Our arguments may have already been set to a block, in which case we evaluate it with the incoming collection."
	argumentModifierBlock notNil ifTrue: [^arguments := argumentModifierBlock value: self value: aCollection].
	arguments := aCollection.! !

!PrefixFunction methodsFor: 'converting' stamp: ''!
convertedStValueOf: anObject
	"This assumes that functions that do conversions have already had their effect in the database, and all we're concerned with is the fundamental data type conversion"
	^base convertedStValueOf: anObject.! !

!PrefixFunction methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Iterate over the expression tree. Keep track of who has already been visited, so we don't get trapped in cycles or visit nodes twice."

	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	base do: aBlock skipping: aSet.
	arguments do: [:each | each do: aBlock skipping: aSet].
	aBlock value: self.! !

!PrefixFunction methodsFor: 'initializing' stamp: ''!
initialize
	super initialize.
	arguments := #().
	separator := ','.! !

!PrefixFunction methodsFor: 'printing' stamp: ''!
printBasicSQLOn: aStream withParameters: aDictionary forSelect: aBoolean
	aStream
		nextPutAll: function;
		nextPut: $(.
	aBoolean
		ifTrue: [base printSelectSQLOn: aStream withParameters: aDictionary]
		ifFalse: [base printSQLOn: aStream withParameters: aDictionary].
	arguments
		do:
			[:each | 
			"It's possible there needs to be type conversion here similar to InfixFunction>>printSQLOn:withParameters:, but it's not clear how common that case is, or how best to do it, so leaving it until it becomes a problem."
			aStream nextPutAll: separator.
			aBoolean 
				ifTrue: [each printSelectSQLOn: aStream withParameters: aDictionary]
				ifFalse: [each printSQLOn: aStream withParameters: aDictionary]].
	aStream nextPut: $).
	alias notNil
		ifTrue:
			[aStream
				nextPutAll: ' AS ';
				nextPutAll: alias].! !

!PrefixFunction methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	self printsIntoJoins
		ifTrue: [self printBasicSQLOn: aStream withParameters: aDictionary forSelect: false]
		ifFalse: [base printSQLOn: aStream withParameters: aDictionary].! !

!PrefixFunction methodsFor: 'printing' stamp: ''!
printSelectSQLOn: aStream withParameters: aDictionary
	self printBasicSQLOn: aStream withParameters: aDictionary forSelect: true.! !

!PrefixFunction methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	aStream
		nextPutAll: function, '('.
	base printOn: aStream.
	aStream nextPutAll: ')'.! !

!PrefixFunction methodsFor: 'accessing' stamp: ''!
separator
	^separator! !

!PrefixFunction methodsFor: 'accessing' stamp: ''!
separator: aString
	separator := aString! !

PrefixFunction subclass: #SubstringFunction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!SubstringFunction commentStamp: '<historical>' prior: 0!
This is a function subclass specifically for substring functions. It needs its own separate subclass because it has to be written out differently for different databases, particularly if applied to LOB types.!

!SubstringFunction methodsFor: 'rewriting' stamp: ''!
rewriteForOracleBlobsWith: mainArgument
	"For Oracle blobs, we can't query with a normal substring function, but rather have to use the DBMS_LOB function, whose arguments are backwards. Further, we don't seem to be able to compare to byte-arrays, so turn it into a string."
	| oldBase template startPosition endPosition length |
	(mainArgument isGlorpExpression and: [mainArgument isConstantExpression]) ifFalse: [^self].
	function := 'UTL_RAW.CAST_TO_VARCHAR2'.
	startPosition := arguments first value.
	endPosition := arguments last value.
	length := (endPosition - startPosition + 1).
	arguments := #().
	oldBase := base.
	template := PrefixFunction named: 'DBMS_LOB.SUBSTR'.
	base := template createFor: (Array with: length asGlorpExpression with: startPosition asGlorpExpression) withBase: oldBase symbol: #dbmsLobSubstr.! !

!SubstringFunction methodsFor: 'rewriting' stamp: ''!
rewriteForPostgresqlBase64BlobsWith: mainArgument
	"This is pretty awful. We expect our main argument to be a string, and we're a substring function, but we know that for some types, the string representation comes out differently in the database. In particular, the Postgresql driver stores strings into lobs in Base64. So, if our previous argument = the string size, rewrite it to the database size. This makes it work if we do
    (someObject blob copyFrom: 1 to: 3) = 'abc'
when in fact, the length of that string in the blob will be 4."
	| stream actualArgument base64Length base64Start |
	(mainArgument isGlorpExpression and: [mainArgument isConstantExpression]) ifFalse: [^self].
	self field type = self field platform blob ifFalse: [^self].
	(mainArgument value size \\ 3) = 0 ifFalse: [self error: 'Sorry, but you will need to use a number of characters that base64 encodes exactly'].
	base64Length := (mainArgument value size // 3) * 4.
	arguments at: 2 put: base64Length asGlorpExpression.
	base64Start := ((arguments first value // 3) * 4) + 1.
	arguments at: 1 put: base64Start asGlorpExpression.
	stream := WriteStream on: String new.
	self field type print: mainArgument value on: stream.
	actualArgument := (ReadStream on: stream contents) skip: 1; upTo: $'.
	mainArgument value: actualArgument.! !

!SubstringFunction methodsFor: 'rewriting' stamp: ''!
rewriteFunctionArgumentsWith: mainArgument
	"This is pretty awful. We expect our main argument to be a string, and we're a substring function, but we know that for some types, the string representation comes out differently in the database. In particular, the Postgresql driver stores strings into lobs in Base64. So, if our previous argument = the string size, rewrite it to the database size. This makes it work if we do
    (someObject blob copyFrom: 1 to: 3) = 'abc'
when in fact, the length of that string in the blob will be 4."
	| platform |
	platform := base ultimateBaseExpression system platform.
	(platform class == PostgreSQLPlatform and: [self field type = platform blob]) ifTrue: [^self rewriteForPostgresqlBase64BlobsWith: mainArgument].
	(platform isOraclePlatform and: [self field type = platform blob]) ifTrue: [^self rewriteForOracleBlobsWith: mainArgument].
	"For MySQL and SQL Server, it wants position, length rather than startPosition, endPosition."
	(platform isMySQLPlatform or: [platform isSQLServerPlatform]) ifTrue: [arguments last value: (arguments last value - arguments first value + 1)].! !

!SubstringFunction methodsFor: 'accessing' stamp: ''!
type
	^self ultimateBaseExpression system platform varchar: 255.! !

FunctionExpression subclass: #StandaloneFunction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!StandaloneFunction commentStamp: '<historical>' prior: 0!
This represents a function that takes no arguments and is complete in itself. The obvious example is COUNT(*). Making a special '*' pseudo-field seems even more of a hack than just hard-coding it into a function like this. It's possible that things like nextval(sequence) might be able to fall in this category too.!

!StandaloneFunction methodsFor: 'mapping' stamp: ''!
mappedFields
	^Array with: self.! !

!StandaloneFunction methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	aStream nextPutAll: ' '; nextPutAll: function.! !

!StandaloneFunction methodsFor: 'accessing' stamp: ''!
type
	type isNil ifFalse: [^type].
	^self defaultType.! !

!StandaloneFunction methodsFor: 'mapping' stamp: ''!
valueInBuilder: anElementBuilder
	^anElementBuilder valueOfField: self! !

!StandaloneFunction methodsFor: 'mapping' stamp: ''!
valueInBuilder: anElementBuilder as: anExpression
	^anElementBuilder valueOfField: anExpression.! !

GlorpExpression subclass: #GeneralSubSelectExpression
	instanceVariableNames: 'query base'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!GeneralSubSelectExpression commentStamp: '<historical>' prior: 0!
This represents an arbitrary sub-select. This would normally get created by embedding a query instance inside a query block. For example
   where: [:each | each id > 2 & (each something = aWholeOtherQuery)].
Warning: This can be get very complicated, especially if aWholeOtherQuery refers to variables inside the main query block, creating a correlated subselect.

Instance Variables:
	query	<Query>	The query that we subselect.

!

!GeneralSubSelectExpression classSide methodsFor: 'instance creation' stamp: ''!
basedOn: anExpression
	"The defining characteristic of a general subselect is that it doesn't care what it's based on. The base is just a convenient place to hang an independent expression. So make sure we're built on a true base, which won't print confusing stuff for us"
	^self new base: anExpression ultimateBaseExpression.! !

!GeneralSubSelectExpression methodsFor: 'preparing' stamp: ''!
addBoundExpressionsIn: aCommand to: boundExpressions
	self query session: aCommand query session.
	self query prepare.
	self query addBoundExpressionsIn: aCommand to: boundExpressions.! !

!GeneralSubSelectExpression methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	" *** This method was defined by Glorp.GlorpExpression as a subclass responsibility.
	Replace its body with a proper implementation. *** "

	self error: 'Subclass responsibility stub not reimplemented'! !

!GeneralSubSelectExpression methodsFor: 'converting' stamp: ''!
asJoin
	" *** This method was defined by Glorp.GlorpExpression as a subclass responsibility.
	Replace its body with a proper implementation. *** "

	self error: 'Subclass responsibility stub not reimplemented'! !

!GeneralSubSelectExpression methodsFor: 'api' stamp: ''!
base
	^base! !

!GeneralSubSelectExpression methodsFor: 'accessing' stamp: ''!
base: anExpression
	base := anExpression.! !

!GeneralSubSelectExpression methodsFor: 'accessing' stamp: ''!
canHaveBase
	"Return true if this type of expression can have a base expression on which other things can be built. Doesn't say whether we actually have a valid one or not."
	^true.! !

!GeneralSubSelectExpression methodsFor: 'converting' stamp: ''!
convertedDbValueOf: anObject
	^anObject.! !

!GeneralSubSelectExpression methodsFor: 'accessing' stamp: ''!
descriptor
	^base system descriptorFor: query resultClass.! !

!GeneralSubSelectExpression methodsFor: 'accessing' stamp: ''!
field
	^nil.! !

!GeneralSubSelectExpression methodsFor: 'api' stamp: ''!
get: aSymbol withArguments: anArray 

	| functionExpression |
	functionExpression := self getFunction: aSymbol withArguments: anArray.
	functionExpression isNil ifFalse: [^functionExpression].
	^RelationExpression
		named: aSymbol
		basedOn: self
		withArguments: anArray! !

!GeneralSubSelectExpression methodsFor: 'api' stamp: ''!
getField: aField
	"This is rather a hack on the use of get:, but I think it should be all right because we should never run into one of these in normal expression generation. The issue is that if we do an expression of the form (x = aQuery) where the query returns an object of the same class as X. We want to rewrite that into a primary key comparison for the database, so we want to do a get: against aQuery. But the only thing that seems to sensibly translate into is a retrieve: clause and return the query. That will only work for a single level get:, but I think that a multi-level one, from converting a block, shouldn't ever get to a general sub-select."
	query retrieve: [:each | each getField: aField].! !

!GeneralSubSelectExpression methodsFor: 'api' stamp: ''!
getMapping: aMapping named: aSymbol
	"Suppose we do an expression of the form (x = aQuery) where the query returns an object of the same class as X. We want to rewrite that into a primary key comparison for the database, so we want to do a getMapping:named: against aQuery. But the only thing that seems to sensibly translate into is a retrieve: clause and return the query. That will only work for a single level get:, but I think that a multi-level one, from converting a block, shouldn't ever get to a general sub-select."
	query retrieve: aSymbol.! !

!GeneralSubSelectExpression methodsFor: 'testing' stamp: ''!
isMappable
	"Return true if we're something that can be returned in a select statement."
	^false.! !

!GeneralSubSelectExpression methodsFor: 'accessing' stamp: ''!
mappedFields
	"Let's try this and see if it works returning empty"
	^#().! !

!GeneralSubSelectExpression methodsFor: 'printing' stamp: ''!
printOnlySelfOn: aStream
	" *** This method was defined by Glorp.GlorpExpression as a subclass responsibility.
	Replace its body with a proper implementation. *** "

	self error: 'Subclass responsibility stub not reimplemented'! !

!GeneralSubSelectExpression methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary 
	| command |
	aStream nextPutAll: '('.
	self query session: aStream query session.
	command := self query sqlWith: aDictionary.
	aStream nextPutAll: command sqlString.
	aStream nextPutAll: ')'.! !

!GeneralSubSelectExpression methodsFor: 'printing' stamp: ''!
printSelectSQLOn: aStream withParameters: aDictionary
	self printSQLOn: aStream withParameters: aDictionary.! !

!GeneralSubSelectExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream
	aStream nextPutAll: 'subselect('.
	query printOn: aStream.
	aStream nextPutAll: ')'.! !

!GeneralSubSelectExpression methodsFor: 'accessing' stamp: ''!
query
	^query! !

!GeneralSubSelectExpression methodsFor: 'accessing' stamp: ''!
query: aQuery
	query := aQuery! !

!GeneralSubSelectExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	| rebuiltBase rebuilt |
	self == anExpression ifTrue: [^aBaseExpression].
	rebuiltBase := base rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression.
	rebuilt := self copy.
	rebuilt base: rebuiltBase.
	rebuilt query: (query rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression).
	^rebuilt.
"	^query asGlorpExpressionOn: aBaseExpression."! !

!GeneralSubSelectExpression methodsFor: 'accessing' stamp: ''!
type
	^query resultType.! !

!GeneralSubSelectExpression methodsFor: 'navigating' stamp: ''!
ultimateBaseExpression
	^base ultimateBaseExpression.! !

!GlorpExpression classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
, anExpression
	^self get: #, withArguments: (Array with: anExpression asGlorpExpression).! !

!GlorpExpression methodsFor: 'api' stamp: ''!
AND: anExpression
	"This method doesn't really have to exist, because it would be inferred using operationFor:, but it's included here for efficiency and to make it a little less confusing how relation expression get created. Note that the two expression must already be  built on the same base!!"

	anExpression isNil ifTrue: [^self].
	anExpression == true ifTrue: [^self].
	anExpression == false ifTrue: [^false].
	(anExpression isGlorpExpression and: [anExpression isEmptyExpression]) ifTrue: [^self].
	^RelationExpression named: #AND basedOn: self withArguments: (Array with: anExpression).! !

!GlorpExpression methodsFor: 'api' stamp: ''!
AND: anExpression AND: otherExpression
	"A convenience method to allow easy conjunction of multiple ANDs at the same level"
	^(self AND: anExpression) AND: otherExpression.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
AND: anExpression AND: otherExpression AND: thirdExpression
	"A convenience method to allow easy conjunction of multiple ANDs at the same level"
	^((self AND: anExpression) AND: otherExpression) AND: thirdExpression.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
AND: anExpression AND: otherExpression AND: thirdExpression AND: fourthExpression
	"A convenience method to allow easy conjunction of multiple ANDs at the same level"
	^(((self AND: anExpression) AND: otherExpression) AND: thirdExpression) AND: fourthExpression.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
OR: anExpression
	"This method doesn't really have to exist, because it would be inferred using operationFor:, but it's included here for efficiency and to make it a little less confusing how relation expression get created.  Note that the two expression must already be  built on the same base!!"
	anExpression isNil ifTrue: [^self].	
	anExpression == true ifTrue: [^true].
	anExpression == false ifTrue: [^self].
	(anExpression isGlorpExpression and: [anExpression isEmptyExpression]) ifTrue: [^self].
	^RelationExpression named: #OR basedOn: self withArguments: (Array with: anExpression).! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
addBoundExpressionsIn: aQuery to: boundExpressions! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
addToTargetFields: aCollection
	"If we can provide target fields, add them to the collection"
	^self.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
additionalExpressions

	^#().! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
additionalExpressionsIn: aQuery 
	"Return the collection of additional expressions (representing joins) that this expression tree requires. Because some additional expressions may in turn have their own additional expressions (mapping joins represented by general expressions), repeat until we don't get any more."

	| allExpressions newExpressions allNewExpressions expressionsConsidered |
	allExpressions := ExpressionGroup with: self.
	allExpressions addAll: aQuery ordering.
	allExpressions add: aQuery groupBy.
	allExpressions addAll: aQuery tracing additionalExpressions.
	expressionsConsidered := IdentitySet new.
	allNewExpressions := OrderedCollection new.
	newExpressions := nil.
	[newExpressions := (newExpressions isNil ifTrue: [allExpressions] ifFalse: [newExpressions])
		inject: ExpressionGroup new 
		into: [:sum :each | 
			((expressionsConsidered includes: each) not and: [
				each queryLevelBaseExpression == aQuery baseExpression]) ifTrue: [
					sum addAll: each additionalExpressions.
					expressionsConsidered add: each].
			sum].
	allNewExpressions addAll: newExpressions children.
	newExpressions isEmpty] whileFalse.
	^allNewExpressions.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
allControlledTables
	"Return all the tables that we control, directly or indirectly. Basically, if we are a mapping expression with a link table, return the link table as well as any tables of ours."
	^self tables.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
allRelationsFor: rootExpression do: aBlock andBetweenDo: anotherBlock
	"In any normal relationship, there's only one thing. Just do it"
	aBlock value: rootExpression leftChild value: rootExpression rightChild.! !

!GlorpExpression methodsFor: 'iterating' stamp: ''!
allSatisfy: aBlock
	"Answer true if aBlock answers true for any element of the receiver.
	 An empty collection answers false."

	self do: [:each| (aBlock value: each) ifFalse: [^false]].
	^true! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
allTables

	^self inject: Set new into: [:sum :each | 
		sum addAll: each tables. sum].! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
allTablesToPrint

	^self inject: Set new into: [:sum :each | 
		sum addAll: each tablesToPrint. sum].! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
allTablesToPrintStoppingAt: aSubSelectBaseExpression

	^self inject: Set new into: [:sum :each | 
		each queryLevelBaseExpression == aSubSelectBaseExpression ifTrue: [
			sum addAll: each tablesToPrint]. sum].! !

!GlorpExpression methodsFor: 'iterating' stamp: ''!
anySatisfy: aBlock
	"Answer true if aBlock answers true for any element of the receiver.
	 An empty collection answers false."

	self do: [:each| (aBlock value: each) ifTrue: [^true]].
	^false! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asDistinctFieldOn: anExpression
	^(self asGlorpExpressionOn: anExpression) get: #distinct withArguments: #().! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."

	self subclassResponsibility.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asField
	"If we correspond to a single field, return it"
	^self field.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asGeneralGlorpExpression
	"Convert the result to a general (tree-format) expression, if it's the more limited join expression"
	^self.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asGeneralGlorpExpressionWithFields
	"Convert the result to a general (tree-format) expression, if it's the more limited join expression."
	^self.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
asGlorpExpression

	^self.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asGlorpExpressionForDescriptor: aDescriptor

	self ultimateBaseExpression descriptor: aDescriptor.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asGlorpExpressionOn: aBaseExpression
	aBaseExpression ultimateBaseExpression == self ultimateBaseExpression
		ifTrue: [^self].
	^self rebuildOn: aBaseExpression.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asGlorpExpressionOn: aBaseExpression basedOn: oldBaseExpression withUltimateBase: ultimateBaseExpression
	^self
		rebuildOn: aBaseExpression
		startingFrom: oldBaseExpression
		withOuterScopeBase: ultimateBaseExpression.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asGlorpExpressionOn: aBaseExpression in: aQuery

	^self asGlorpExpressionOn: aBaseExpression.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asGlorpExpressionOn: anExpression withUltimateBase: aBaseExpression
	"Build the expression, knowing that some parts of it may refer to an outer context based on aBaseExpression. This is primarily useful for subselect blocks"
	^self rebuildOn: anExpression startingFrom: self queryLevelBaseExpression withOuterScopeBase: aBaseExpression.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asIndependentJoins
	"If this is an ANDed clause, split it into independent joins"
	^Array with: self.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asJoin
	self subclassResponsibility.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
asNonDistinctField
	^self! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
assignTableAliasesStartingAt: anInteger prefix: aString
	^anInteger.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
base

	self subclassResponsibility.! !

!GlorpExpression methodsFor: 'accessing' stamp: ''!
beOuterJoin
	^self.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
bindingType
	^self type.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
canBeUsedForRetrieve
	"Return true if this is a valid argument for a retrieve: clause"
	^false.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
canBind
	"Return true if this represents a value that can be bound into a prepared statement"
	^false.! !

!GlorpExpression methodsFor: 'accessing' stamp: ''!
canHaveBase
	"Return true if this type of expression can have a base expression on which other things can be built. Doesn't say whether we actually have a valid one or not."
	^self subclassResponsibility.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
canKnit
	"Return true if, when building objects, we can knit the object corresponding to this expression to a related object. Roughly speaking, is this a mapping expression"
	^false.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
canPrint
	"Return true if we can print into a where clause"
	^true.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
canTranslateFields
	"Return true if we are a mapping expression and can thus translate from fields into their aliased version for a particular query."
	^false.! !

!GlorpExpression methodsFor: 'printing' stamp: ''!
className

	^self class name.! !

!GlorpExpression methodsFor: 'iterating' stamp: ''!
collect: aBlock

	| newCollection |
	newCollection := OrderedCollection new.
	self do: [:each | newCollection add: (aBlock value: each)].
	^newCollection.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
condenseBooleanEquality
	^self.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
condenseBooleanEqualityExpressions
	^self.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
condensePrimaryKeyComparison
	^self.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
condensePrimaryKeyComparisons
	^self.! !

!GlorpExpression methodsFor: 'iterating' stamp: ''!
detect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.
	Answer the first element for which aBlock evaluates to true."

	^self detect: aBlock ifNone: [self notFoundError]! !

!GlorpExpression methodsFor: 'iterating' stamp: ''!
detect: aBlock ifNone: exceptionBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument.
	Answer the first element for which aBlock evaluates to true."

	self do: [:each | (aBlock value: each) ifTrue: [^each]].
	^exceptionBlock value! !

!GlorpExpression methodsFor: 'printing' stamp: ''!
displayString

	| stream |
	stream := String new writeStream.
	self printOnlySelfOn: stream.
	^stream contents.! !

!GlorpExpression methodsFor: 'iterating' stamp: ''!
do: aBlock
	"Iterate over the expression tree"

	self do: aBlock skipping: Dialect identitySetClass new.! !

!GlorpExpression methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Iterate over the expression tree. Keep track of who has already been visited, so we don't get trapped in cycles or visit nodes twice."

	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	aBlock value: self.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
equals: anExpression

	^RelationExpression named: #= basedOn: self withArguments: (Array with: anExpression).! !

!GlorpExpression methodsFor: 'accessing' stamp: ''!
field
	^nil.! !

!GlorpExpression methodsFor: 'accessing' stamp: ''!
fieldFromMeOrSubclasses
	"For mapping expression, return the field for this mapping, but if the field is not mapped, and the class uses inheritance, check for subclasses, and use a random one of the subclass mappings. For all others, just return the field."
	^self field.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
generalExpressionPart
	"If we have a mixture of joins and general expressions, return only the general part"
	^self.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
get: aSymbol withArguments: anArray

	self subclassResponsibility.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
getFunction: aSymbol withArguments: anArray

	| expression |
	expression := FunctionExpression for: aSymbol withArguments: anArray basedOn: self.
	expression isNil ifTrue: [^nil].
	^expression.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
getFunction: aSymbol withArguments: anArray in: system

	| expression |
	expression := FunctionExpression for: aSymbol withArguments: anArray basedOn: self system: system.
	expression isNil ifTrue: [^nil].
	^expression.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
hasBindableExpressionsIn: aCommand
	^false.! !

!GlorpExpression methodsFor: 'accessing' stamp: ''!
hasDescriptor

	^false.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
hasImpliedClauses
	"Return true if this implies additional SQL clauses beyond just a single field expression"
	^false.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
hasProvidedAdditionalExpressions
	"Return true if we have already provided additional expressions to some query. Only important when we're preparing a subselect and don't want to bring in joins that are part of the main query"
	^true.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
hasTableAliases
	^false.! !

!GlorpExpression methodsFor: 'initialize' stamp: ''!
in: anExpression

	^RelationExpression named: #IN basedOn: self withArguments: (Array with: anExpression).! !

!GlorpExpression methodsFor: 'initialize' stamp: ''!
initialize! !

!GlorpExpression methodsFor: 'iterating' stamp: ''!
inject: anObject into: aBlock

	| sum |
	sum := anObject.
	self do: [:each | sum := aBlock value: sum value: each].
	^sum! !

!GlorpExpression methodsFor: 'inspecting' stamp: ''!
inspectorHierarchies
	| hierarchy |
	hierarchy := ((Smalltalk at: #Tools ifAbsent: [^#()])
		at: #Trippy ifAbsent: [^#()])
		at: #Hierarchy ifAbsent: [^#()].
	^Array with: (hierarchy
			id: #expression
			label: 'Expression Tree'
			parentBlock: [:each | nil]
			childrenBlock: [:each | each inspectorChildren])! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
inverseJoin
	"We assume that only the parts that are actually joins ought to invert"
	| new |
	new := self asGlorpExpressionOn: BaseExpression new.
	new do: [:each |
		each replaceJoinsWithInverse].
	^new.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
isConstantExpression
	^false.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
isDistinct
	^false.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
isEmptyExpression
	^false.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
isGlorpExpression
	^true.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
isJoin
	^false.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
isMappable
	"Return true if we're something that can be returned in a select statement."
	^true.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
onlyJoinParts
	"Strip out any generic parts, leaving only the things that are actually joins"
	| new |
	new := self asGlorpExpressionOn: BaseExpression new.
	new do: [:each |
		each removeNonJoinParts].
	new := new condenseBooleanEquality.
	new isJoin ifFalse: [new do: [:each | each condenseBooleanEqualityExpressions]].
	^new.! !

!GlorpExpression methodsFor: 'api' stamp: ''!
parameter: aConstantExpression
	"Create a parameter expression with the given name. But note that the name doesn't have to be a string. Database fields, symbols, and integers are all plausible"
	| value |
	value := aConstantExpression class == ParameterExpression ifTrue: [aConstantExpression field] ifFalse: [aConstantExpression value].
	^ParameterExpression forField: value basedOn: self.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
prepareIn: aQuery
	"A little weird, because what we end up asking for additional expressions from, might not be us"

	aQuery adjustWhereClause.
	(aQuery whereClause additionalExpressionsIn: aQuery)
		do:
			[:each | 
			(aQuery shouldExpressionBeJoin: each)
				ifTrue: [aQuery addJoin: each]
				ifFalse: [aQuery privateAnd: each]].! !

!GlorpExpression methodsFor: 'primary keys' stamp: ''!
primaryKeyFromDictionary: aDictionary
	"Given a set of parameters, return a primary key suitable for retrieving our target. We can't do this for general expressions, so indicate failure by returning nil"
	^nil.! !

!GlorpExpression methodsFor: 'printing' stamp: ''!
printOn: aStream 
	self printTreeOn: aStream! !

!GlorpExpression methodsFor: 'printing' stamp: ''!
printOnlySelfOn: aStream
	self subclassResponsibility.! !

!GlorpExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream

	self subclassResponsibility.! !

!GlorpExpression methodsFor: 'accessing' stamp: ''!
printsTable

	^false.! !

!GlorpExpression methodsFor: 'navigating' stamp: ''!
queryLevelBaseExpression
	^self ultimateBaseExpression! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
rebuildOn: aBaseExpression
	^self rebuildOn: aBaseExpression startingFrom: nil.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression
	"Rebuild the expression on the new base. If we encounter anExpression, treat that as the base, and build only from that point up. e.g. if we have a.b.c.d and we rebuildOn: z startingFrom: b, then the result will be z.c.d"
	^self rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: nil.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	"Rebuild the expression on the new base. If we encounter anExpression, treat that as the base, and build only from that point up. e.g. if we have a.b.c.d and we rebuildOn: z startingFrom: b, then the result will be z.c.d. Pieces that aren't based on anExpression should be built onto ultimateBaseExpression instead of aBaseExpression. They represents parts of outer scope trees that are referenced within us.
  Note that anExpression can be nil, but if it is, then ultimateBaseExpression is effectively ignored."
	self subclassResponsibility.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
removeNonJoinParts
	"Remove anything that's not an actual join. Only useful if we have something that has mixed joins and general expressions, where we assume we don't want to use the general ones, because they're not field level."
	^self.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
replaceJoinsWithInverse
	"If we don't know specifically what to do with a kind of expression, assume it doesn't apply and we should be replacing it with nothing."
	^self.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
representsDerivedObject
	^false.! !

!GlorpExpression methodsFor: 'private/accessing' stamp: ''!
requiresDistinct
	^false.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
returnsDictionary
	"Return true if we representa  ampping to a dictionary"
	^false.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
rewriteBooleanEquality: relation
	"Does not apply to general expressions"! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
rewriteEquality
	^self.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
rewriteEqualityExpressions
	^self.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
rewriteFunctionArguments
	"If a function needs to adjust its arguments, do that here. The motivating example is one where a string argument is rewritten into a form that's of a different length, e.g. the Postgresql driver does that to strings in Blobs"
	^self.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
rewriteFunctionArgumentsWith: ourArgument
	"If a function needs to adjust its arguments, do that here. The motivating example is one where a string argument is rewritten into a form that's of a different length, e.g. the Postgresql driver does that to strings in Blobs"
	^self.! !

!GlorpExpression methodsFor: 'iterating' stamp: ''!
select: aBlock

	| newCollection |
	newCollection := OrderedCollection new.
	self do: [:each | (aBlock value: each) ifTrue: [newCollection add: each]].
	^newCollection! !

!GlorpExpression methodsFor: 'debugging' stamp: ''!
sharedPiecesWith: anExpression
	"Return a set of the pieces we share in common with anExpression. Mostly useful for debugging unwanted sharing, or the lack of wanted sharing. Note that this won't dive down into subselects."
	| myPieces shared |
	myPieces := Dialect identitySetClass new.
	shared := Dialect identitySetClass new.
	self do: [:each | myPieces add: each].
	anExpression do: [:each | (myPieces includes: each) ifTrue: [shared add: each]].
	^shared.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
shortCircuitOnNullComparison: anObject
	"Return true if we represent an expression that compares null to a non-nullable value, so we don't ever need to really evaluate it"

	^false.! !

!GlorpExpression methodsFor: 'converting' stamp: ''!
sourceForTarget: aField

	^nil.! !

!GlorpExpression methodsFor: 'As yet unclassified' stamp: ''!
sqlString
	| command |
	command := QuerySelectCommand new
		useBinding: false;
		session: self ultimateBaseExpression descriptor session.
	self printSQLOn: command withParameters: #().
	^command contents.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
tableForANSIJoin
	"Which table will we join to."
	^nil.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
tables

	^#().! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
tablesContainingMe
	"Return a list of tables that contain this object, whether or not we control them. This is mostly the same as tables, but will differ for embedded mappings."
	^self tables.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
tablesToPrint

	^#().! !

!GlorpExpression methodsFor: 'accessing' stamp: ''!
type
	^self field type.! !

!GlorpExpression methodsFor: 'navigating' stamp: ''!
ultimateBaseExpression
	self subclassResponsibility.! !

!GlorpExpression methodsFor: 'preparing' stamp: ''!
validate! !

!GlorpExpression methodsFor: 'accessing' stamp: ''!
valueIn: aDictionary
	"Return the value associated with this expression given the parameters in aDictionary. Only meaningful for ParameterExpressions"

	^self.! !

!GlorpExpression methodsFor: 'testing' stamp: ''!
willPrintAsWhereClause
	^true.! !

GlorpExpression subclass: #ObjectExpression
	instanceVariableNames: 'mappingExpressions requiresDistinct tableAliases fieldAliases'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!ObjectExpression commentStamp: '<historical>' prior: 0!
This is a not-especially well-named superclass for expressions that end up mapping to rows in the database, more or less. That is, it's more or less a superclass for things that have the capacity to alias tables, and to have mapping expressions derived from them. Since, depending on your point of view, everything is an object, the name isn't very descriptive. Or if some things aren't considered objects, TableExpression is a good candidate. But we have the common functionality for managing tables, and for keeping a dictionary of our sub-mapping expressions.

Subclasses must implement the following messages:
	accessing
		system
		table
	fields
		controlsTables

Instance Variables:
	fieldAliases	<Dictionary from: DatabaseField to: DatabaseField>	For each field, the corresponding aliased field for our query. So, e.g. PERSON.NAME will map to t1.NAME. These aren't strings, though, these are fields derived from the aliased table.
	mappingExpressions	<Dictionary from: Symbol to: ( MappingExpression | TableExpression) >	Each attribute that's derived from us will be maintained in this dictionary. This is very important, because Glorp relies critically on these expressions being identical. So "object field == object field".
	requiresDistinct	<Boolean>	Do the fields that we will cause to be added to the SELECT portion of the query need a DISTINCT wrapped around them?
	tableAliases	<Dictionary from: DatabaseTable to: DatabaseTable>	While the query is being prepared, tables will be assigned aliases. This keeps track of the aliases for the tables that we control. As with fieldAliases, note that this stores table copies with their names altered, not strings.

!

ObjectExpression subclass: #BaseExpression
	instanceVariableNames: 'descriptor availableSubSelectNumber'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!BaseExpression commentStamp: '<historical>' prior: 0!
This represents the base object that we're looking for in a query. So if we ask to read: Person, then we will have a base expression corresponding to Person. Even if what we end up retrieving is not part of the Person object, it remains the base, because it's what defines our point of view.

Instance Variables:
	availableSubSelectNumber	<Integer>	As the base of the query, we keep track of how many subselects there are. New subselects will need to have a unique number so they can prefix their tables as e.g. s1t1 when aliasing them, and the number comes from this counter.
	descriptor	<Descriptor>	The descriptor for the class we're looking for.

!

!BaseExpression methodsFor: 'accessing' stamp: ''!
additionalExpressions
	| expressions |
	expressions := OrderedCollection new.
	self descriptor typeMapping addTypeMappingCriteriaTo: expressions in: self.
	expressions addAll: self multipleTableExpressions.
	^expressions! !

!BaseExpression methodsFor: 'preparing' stamp: ''!
additionalExpressionsOn: anExpression
	^#().! !

!BaseExpression methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."

	^source.! !

!BaseExpression methodsFor: 'preparing' stamp: ''!
availableSubSelectNumber
	availableSubSelectNumber := availableSubSelectNumber isNil ifTrue: [1] ifFalse: [availableSubSelectNumber + 1].
	^availableSubSelectNumber.! !

!BaseExpression methodsFor: 'api' stamp: ''!
base

	^nil.! !

!BaseExpression methodsFor: 'testing' stamp: ''!
canBeUsedForRetrieve
	"Return true if this is a valid argument for a retrieve: clause"
	^true.! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
canHaveBase
	"Return true if this type of expression can have a base expression on which other things can be built. We don't have a base, but we *are* a base, so return true"
	^true.! !

!BaseExpression methodsFor: 'testing' stamp: ''!
canPrint
	"Return true if we can print into a where clause"
	^false.! !

!BaseExpression methodsFor: 'printing' stamp: ''!
className

	^'Base'.! !

!BaseExpression methodsFor: 'preparing' stamp: ''!
controlsTables

	^true.! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
descriptor

	^descriptor! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
descriptor: aDescriptor
	descriptor := aDescriptor! !

!BaseExpression methodsFor: 'preparing' stamp: ''!
fieldsForSelectStatement
	^self descriptor mappedFields.! !

!BaseExpression methodsFor: 'api' stamp: ''!
getParameter: aDatabaseField 
	^ParameterExpression forField: aDatabaseField basedOn: self.! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
hasDescriptor

	^self descriptor notNil.! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
mappedFields
	^self descriptor mappedFields collect: [:each | self translateField: each].! !

!BaseExpression methodsFor: 'testing' stamp: ''!
mapsSimpleObject
	"Try to do a fast test for this, to look for the pseudo-descriptors that direct to-many mappings use."
	^self descriptor mappings size = 1 and: [(self descriptor mappingForAttributeNamed: #myValue) notNil].! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
multipleTableExpressions
	^self descriptor multipleTableJoin 
		collect: [:each | each asExpressionJoiningSource: self toTarget: self]! !

!BaseExpression methodsFor: 'printing' stamp: ''!
printOn: aStream 
	aStream
		nextPutAll: self className;
		nextPut: $(.
	self printTreeOn: aStream.
	aStream nextPut: $)! !

!BaseExpression methodsFor: 'printing' stamp: ''!
printOnlySelfOn: aStream 
	self descriptor isNil ifTrue: [aStream nextPutAll: 'Empty Base'. ^self].
	aStream print: self descriptor describedClass.
	self printTableAliasesOn: aStream! !

!BaseExpression methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary 
	"Do nothing. We're not printable"! !

!BaseExpression methodsFor: 'printing' stamp: ''!
printSelectSQLOn: aStream withParameters: aDictionary 
	"Normally, we're not printable, but we might be if we're inverting a direct to-many mapping. In that case, there should only be one mapped field"
	self descriptor mappedFields do: [:each | (self translateField: each) printSQLOn: aStream withParameters: aDictionary].! !

!BaseExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	aStream 
		print: (self descriptor isNil ifTrue: [nil] ifFalse: [self descriptor describedClass])! !

!BaseExpression methodsFor: 'navigating' stamp: ''!
queryLevelBaseExpression

	^self! !

!BaseExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	anExpression isNil ifTrue: [^aBaseExpression].
	anExpression == self ifTrue: [^aBaseExpression].
	^ultimateBaseExpression isNil ifTrue: [self] ifFalse: [ultimateBaseExpression].! !

!BaseExpression methodsFor: 'api' stamp: ''!
referenceClass
	"Return the class that our mapping refers to, if we have a mapping. If we're a base expression, just return our class, and if we're a table expression, just return nil."
	^self descriptor describedClass.! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
system
	self descriptor isNil ifTrue: [^nil].
	^self descriptor system.! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
table

	^self descriptor primaryTable.! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
tables

	^self descriptor isNil ifTrue: [#()] ifFalse: [self descriptor tables].! !

!BaseExpression methodsFor: 'preparing' stamp: ''!
tablesToPrint
	"We derive the base's tables from the fields that are being selected, but make sure that at least the primary table is listed."
	self descriptor tables isEmpty ifTrue: [^#()].
	^Array with: (self aliasedTableFor: self descriptor primaryTable).! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
targetDescriptor
	self halt! !

!BaseExpression methodsFor: 'accessing' stamp: ''!
trueBase
	^self.! !

!BaseExpression methodsFor: 'navigating' stamp: ''!
ultimateBaseExpression

	^self! !

!BaseExpression methodsFor: 'mapping' stamp: ''!
valueInBuilder: anElementBuilder
	"If we ever get here, this is because we're being used as the reverse of a direct to-many mapping, so we expect to have a descriptor with only one mapping, named #myValue"
	| mapping |
	mapping := self descriptor mappingForAttributeNamed: #myValue.
	^mapping valueInBuilder: anElementBuilder as: (self get: #myValue) field.! !

BaseExpression subclass: #JoinBaseExpression
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!JoinBaseExpression commentStamp: '<historical>' prior: 0!
When we use an expression as a join, we tend to write it backwards. We use expressions when the relationship is more complicated than we can express as a single join. And we describe it by giving the target and describing how to use relationships to get from the target object to the source object. We can't do it forwards, because the forward relationship is exactly what we're trying to describe with this join. For example, for a (rather contrived) Address to Job relationship, we might write it as
  [:eachJob | eachJob worker address]
However, this means that when we're trying to rebuild this expression and fit it into a query, it's backwards from most expressions. We know that, so we give it a special kind of base that makes it rebuild itself slightly different, by preferring the target rather than the source.  !

!JoinBaseExpression methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."

	^target.! !

BaseExpression subclass: #SubSelectBaseExpression
	instanceVariableNames: 'trueBase'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!SubSelectBaseExpression commentStamp: '<historical>' prior: 0!
This is a special variation of a base expression which is used when we know we are a subselect. In some circumstances we'll need to operate with respect to the subselect base, as if we were a standalone query. But sometimes we'll need to know the broader context in which we're living. Having a separate subclass helps us distinguish.

Instance Variables:
	trueBase	<MappingExpression>	The base on which we are built. Note that this doesn't mean this is the overall base expression, in fact it's probably a mapping expression, it's just the thing from which this subselect was built. Presumably the ultimate base expression can be found by following that outer context down, but there could be multiple levels of subselects involved.

!

!SubSelectBaseExpression classSide methodsFor: 'instance creation' stamp: ''!
on: aBaseExpression
	^self new trueBase: aBaseExpression.! !

!SubSelectBaseExpression methodsFor: 'accessing' stamp: ''!
additionalExpressions
	"We effectively take over the join for the thing we're based on, because we want it to happen in the subselect, not in the main one."

	^trueBase additionalExpressionsOn: self.! !

!SubSelectBaseExpression methodsFor: 'preparing' stamp: ''!
additionalExpressionsOn: anExpression
	^trueBase additionalExpressionsOn: anExpression.! !

!SubSelectBaseExpression methodsFor: 'accessing' stamp: ''!
base
	^trueBase base.! !

!SubSelectBaseExpression methodsFor: 'printing' stamp: ''!
className

	^'SubSelectBase'.! !

!SubSelectBaseExpression methodsFor: 'converting' stamp: ''!
convertedDbValueOf: anObject

	^self trueBase mapping convertedDbValueOf: anObject.! !

!SubSelectBaseExpression methodsFor: 'accessing' stamp: ''!
descriptor
	^trueBase descriptor.! !

!SubSelectBaseExpression methodsFor: 'accessing' stamp: ''!
field
	^self trueBase mapping field.! !

!SubSelectBaseExpression methodsFor: 'accessing' stamp: ''!
join
	^self trueBase join.! !

!SubSelectBaseExpression methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	"If I'm ever going to print this directly, it's because I'm participating in a direct to-many relationship, so I do have a valid field"
	(self translateField: self field) printSQLOn: aStream withParameters: aDictionary.! !

!SubSelectBaseExpression methodsFor: 'navigating' stamp: ''!
queryLevelBaseExpression

	^self! !

!SubSelectBaseExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	anExpression isNil ifTrue: [^aBaseExpression].
	anExpression == self ifTrue: [^aBaseExpression].
	trueBase := trueBase rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression.
	^self.! !

!SubSelectBaseExpression methodsFor: 'testing' stamp: ''!
representsDerivedObject
	^self trueBase representsDerivedObject! !

!SubSelectBaseExpression methodsFor: 'accessing' stamp: ''!
table
	^self trueBase table.! !

!SubSelectBaseExpression methodsFor: 'accessing' stamp: ''!
trueBase
	^trueBase! !

!SubSelectBaseExpression methodsFor: 'accessing' stamp: ''!
trueBase: aBaseExpression
	trueBase := aBaseExpression.! !

!SubSelectBaseExpression methodsFor: 'navigating' stamp: ''!
ultimateBaseExpression

	^trueBase ultimateBaseExpression! !

!SubSelectBaseExpression methodsFor: 'preparing' stamp: ''!
validate
	super validate.
	trueBase validate.! !

ObjectExpression subclass: #MappingExpression
	instanceVariableNames: 'name base outerJoin mapping phantom'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!MappingExpression commentStamp: '<historical>' prior: 0!
This represents an attribute of some other object. It's called a mapping expression because each attribute has a mapping, and when we come to prepare the query, we'll be using that mapping's information to deduce joins, and to print properly.

Instance Variables:
	base	<BaseExpression | MappingExpression>	The expression we are built on, which is to say the expression describing the object of which we are an attribute.
	mapping	<Mapping>	The mapping we correspond to in the base. So if the base was [:each | each owner city] and we are #name, then we the mapping is for the attribute #name in the descriptor for City.
	name	<Symbol>	The name of our attribute. We will use this to determine the mapping.
	outerJoin	<Boolean>	Are we to interpret the join to our parent object's table as being an outer join or not?
	phantom <Boolean> Is this a phantom mapping - one which is not defined in the descriptor explicitly, but is temporarily created to define a reverse relationship, an alsoFetch of subclasses, or some other transient usage.

!

!MappingExpression classSide methodsFor: 'instance creation' stamp: ''!
named: aSymbol basedOn: anExpression

	^self new
		named: aSymbol
		basedOn: anExpression.! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
additionalExpressions
	^self additionalExpressionsOn: self.! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
additionalExpressionsOn: anExpression
	"We can get hijacked, so pass in the expression to use rather than just using self"
	| exp |
	self mapping isNil ifTrue: [^#()].
	exp := self mapping joinExpressionFor: anExpression.
	outerJoin ifTrue: [exp beOuterJoin].
	^(self multipleTableExpressionsOn: anExpression), (exp isNil ifTrue: [#()] ifFalse: [Array with: exp]).! !

!MappingExpression methodsFor: 'fields' stamp: ''!
aliasedTableFor: aDatabaseTable 
	^self controlsTables 
		ifTrue: [super aliasedTableFor: aDatabaseTable]
		ifFalse: [base aliasedTableFor: (self mapping translateTable: aDatabaseTable)]! !

!MappingExpression methodsFor: 'fields' stamp: ''!
aliasedTableFor: aDatabaseTable ifAbsent: aBlock
	^self controlsTables 
		ifTrue: [super aliasedTableFor: aDatabaseTable ifAbsent: aBlock]
		ifFalse: [base aliasedTableFor: aDatabaseTable ifAbsent: aBlock]! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
allControlledTables
	"Return all the tables that we control, directly or indirectly. Basically, if we are a mapping expression with a link table, return the link table as well as any tables of ours."
	| all join |
	self controlsTables ifFalse: [^#()].
	all := Set new.
	all addAll: self tables.
	mapping := self mapping.
	mapping isNil 
		ifFalse: [
			join := mapping join.
			join isNil ifFalse: [all addAll: join allControlledTables]].
	^all.! !

!MappingExpression methodsFor: 'internal' stamp: ''!
allRelationsFor: rootExpression
	^self mapping allRelationsFor: rootExpression.! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
allRelationsFor: rootExpression do: aBlock andBetweenDo: anotherBlock
	"We might have multiple clauses to print, depending on our mapping"
	self mapping 
		allRelationsFor: rootExpression
		do: aBlock
		andBetweenDo: anotherBlock.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
allSourceFields
	"If we are sending this, then we're treating the expression as a join, for a complex (and probably read-only) relationship. So in that case we want to treat the source fields as the parameters that this expression takes, and specifically the parameters that are based on fields."
	^self inject: OrderedCollection new into: [:sum :each |
		each addToTargetFields: sum.
		sum].! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
allTargetFields
	"If we are sending this, then we're treating the expression as a join, for a complex (and probably read-only) relationship. So in that case we want to treat the target fields as being the source fields of the bottom-most mapping. That is, the fields we'll use to connect to the thing we want are the things that our expression's base (which represents the thing we want) would use to connect to the first level mapping arising from it."
	^self penultimateBaseExpression join allSourceFields.! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."

	| newBase |
	newBase := base asExpressionJoiningSource: source toTarget: target.
	^newBase getMapping: self mapping named: name.! !

!MappingExpression methodsFor: 'api' stamp: ''!
asOuterJoin
	outerJoin := true.! !

!MappingExpression methodsFor: 'api' stamp: ''!
base

	^base.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
bePhantom
	phantom := true.! !

!MappingExpression methodsFor: 'testing' stamp: ''!
canBeUsedForRetrieve
	"Return true if this is a valid argument for a retrieve: clause"
	^self mapping canBeUsedForRetrieve.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
canHaveBase
	"Return true if this type of expression can have a base expression on which other things can be built. Doesn't say whether we actually have a valid one or not."
	^true.! !

!MappingExpression methodsFor: 'testing' stamp: ''!
canKnit
	"Return true if, when building objects, we can knit the object corresponding to this expression to a related object. Roughly speaking, is this a mapping expression"
	^true.! !

!MappingExpression methodsFor: 'testing' stamp: ''!
canPrint
	"Return true if we can print into a where clause"
	^self hasField.! !

!MappingExpression methodsFor: 'fields' stamp: ''!
controlsTables
	| theMapping |
	theMapping := self mappingFromMeOrSubclasses.
	theMapping isNil ifTrue: [^true].
	^theMapping controlsTables! !

!MappingExpression methodsFor: 'converting' stamp: ''!
convertedDbValueOf: anObject

	^self mapping convertedDbValueOf: anObject.! !

!MappingExpression methodsFor: 'converting' stamp: ''!
convertedStValueOf: anObject

	^self mapping convertedStValueOf: anObject.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
descriptor
	| system theMapping |
	system := self system.
	system isNil ifTrue: [^nil].
	theMapping := self mappingFromMeOrSubclasses.
	theMapping isNil ifTrue: [^nil].
	^system descriptorFor: theMapping referenceClass.! !

!MappingExpression methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Iterate over the expression tree"

	| myField |
	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	base do: aBlock skipping: aSet.
	aBlock value: self.
	(self hasField and: [(myField := self field) isGlorpExpression])
		ifTrue: [myField do: aBlock skipping: aSet].! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
field

	self mapping isNil ifTrue: [self error: '"', name, '" is not a mapped property name in ', base descriptor describedClass name].
	self hasDescriptor ifTrue: [
		"If we have a descriptor, then either we're a mapping to a full object, in which case we can't answer this properly, or we're a direct to-many mapping, in which case we can."
		self mapping mapsSimpleObject
			ifTrue: [^self translateField: self mapping field]
			ifFalse: [self error: '"',name, 
				'" is not an attribute that resolves to a field in the mapped tables for ', 
					base descriptor describedClass name]].
	^base translateField: self mapping field.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
fieldFromMeOrSubclasses
	"Return the field for this mapping, but if the field is not mapped here, and the class uses inheritance, check for subclasses, and use a random one of the subclass mappings."
	| subMapping |
	subMapping := self mappingFromMeOrSubclasses.
	^subMapping isNil ifTrue: [self field] ifFalse: [subMapping field].! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
fieldsForSelectStatement
	
	^self mapping fieldsForSelectStatement.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
hasDescriptor
	"Does the object that we describe have its own descriptor"
	| myMapping |
	myMapping := self mapping.
	^myMapping notNil and: [myMapping isRelationship]! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
hasField

	self mapping isNil ifTrue: [^false].
"If we have a descriptor, then either we're a mapping to a full object, in which case we can't answer this properly, or we're a direct to-many mapping, in which case we can."
	self hasDescriptor ifTrue: [
		^self mapping mapsSimpleObject].
	^true.! !

!MappingExpression methodsFor: 'testing' stamp: ''!
hasImpliedClauses
	^self mapping notNil and: [self mapping hasImpliedClauses].! !

!MappingExpression methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	phantom := false.! !

!MappingExpression methodsFor: 'api' stamp: ''!
isEmpty
	"Only useful for mappings that represent collections. This checks to see if the collection has elements by building a not exists subselect."
	
	^self base existenceCheck: self as: #notExists:! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
join
	self mapping isNil ifTrue: [^nil].
	^self mapping join.! !

!MappingExpression methodsFor: 'iterating' stamp: ''!
mappableFieldsDo: aBlock
	"Used when we are emulating a Join"
	self allSourceFields with: self allTargetFields do: [:source :target |
		source isMappable ifTrue: [aBlock value: source value: target]].! !

!MappingExpression methodsFor: 'iterating' stamp: ''!
mappableSourceFieldsDo: aBlock
	"Used when we are emulating a Join"
	self allSourceFields do: [:source |
		source isMappable ifTrue: [aBlock value: source]].! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
mappedFields
	
	self mapping isNil ifTrue: [^#()].
	self mapping isRelationship ifTrue: [self error: '"',name, '" is not an attribute that resolves to a field in the mapped tables for ', base descriptor describedClass name].
	^self mapping mappedFields collect: [:each | base translateField: each].! !

!MappingExpression methodsFor: 'internal' stamp: ''!
mapping
	mapping isNil ifTrue: [
		| descriptor |
		descriptor := self sourceDescriptor.
		descriptor isNil ifTrue: [^nil].
		mapping := descriptor mappingForAttributeNamed: name].
	^mapping.! !

!MappingExpression methodsFor: 'private/initialization' stamp: ''!
mapping: aMapping
	mapping := aMapping.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
mappingFromMeOrSubclasses
	"Return the mapping for me, but if the field is not mapped here, and the class uses inheritance, check for subclasses, and use a random one of the subclass mappings."

	| subclassDescriptors |
	self mapping isNil ifFalse: [^self mapping].
	self sourceDescriptor isNil ifTrue: [^self mapping].
	subclassDescriptors := (self sourceDescriptor typeResolver descriptorsRequiringIndependentQueriesFor: self sourceDescriptor describedClass).
	subclassDescriptors do: [:each | 
		| subMapping |
		subMapping := each mappingForAttributeNamed: name.
		subMapping isNil ifFalse: [^subMapping]].
	"If subclasses didn't work, just fall through to the normal error processing."
	^self mapping.! !

!MappingExpression methodsFor: 'testing' stamp: ''!
mapsSimpleObject
	^self mapping mapsSimpleObject.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
multipleTableExpressions
	^self multipleTableExpressionsOn: self.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
multipleTableExpressionsOn: anExpression
	^self mapping multipleTableExpressionsFor: anExpression.! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
mustReGetMappingIfRebuildingOn: newBase
	"Return true if we need to do a full get: when rebuilding, rather than being able to preserve the mapping that we had before."

	"If we didn't have a mapping yet, then just do a get"
	mapping isNil ifTrue: [^true].
	"If this is a phantom mapping, we must not try to re-get it."
	self phantom ifTrue: [^false].
	"Otherwise, we're a normal mapping, and we don't need to re-get the mapping unless the class is different, probably because of inheritance. But note that in some cases we may not know our descriptor yet."
	self base hasDescriptor ifFalse: [^false].
	self base descriptor isNil ifTrue: [^false].
	^self base descriptor ~= newBase descriptor.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
name

	^name.! !

!MappingExpression methodsFor: 'private/initialization' stamp: ''!
named: aSymbol basedOn: anExpression

	name := aSymbol.
	base := anExpression.
	outerJoin := false.! !

!MappingExpression methodsFor: 'api' stamp: ''!
notEmpty
	"Only useful for mappings that represent collections. This checks to see if the collection has elements by building a not exists subselect."
	
	^self base existenceCheck: self as: #exists:! !

!MappingExpression methodsFor: 'navigating' stamp: ''!
penultimateBaseExpression
	"Return the one from the last base expression in our tree. Useful if we want to find the first mapping involved in the expression"
	self base base isNil ifTrue: [^self].
	^self base penultimateBaseExpression.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
phantom
	^phantom! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
phantom: aBoolean
	phantom := aBoolean! !

!MappingExpression methodsFor: 'printing' stamp: ''!
printOnlySelfOn: aStream

	aStream nextPutAll: name.
	self printTableAliasesOn: aStream.! !

!MappingExpression methodsFor: 'printing SQL' stamp: ''!
printSQLOn: aStream withParameters: aDictionary 
	self hasField ifFalse: [^self].
	self field asField printSQLOn: aStream withParameters: aDictionary.! !

!MappingExpression methodsFor: 'printing SQL' stamp: ''!
printSelectSQLOn: aStream withParameters: aDictionary 
	self hasField ifFalse: [^self].
	self field asField printSelectSQLOn: aStream withParameters: aDictionary.! !

!MappingExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	aStream
		print: base;
		nextPut: $.;
		nextPutAll: (name isString ifTrue: [name] ifFalse: [name printString])! !

!MappingExpression methodsFor: 'printing SQL' stamp: ''!
printUnqualifiedSQLOn: aStream withParameters: aDictionary 
	self field asField printUnqualifiedSQLOn: aStream withParameters: aDictionary.! !

!MappingExpression methodsFor: 'navigating' stamp: ''!
queryLevelBaseExpression

	^base queryLevelBaseExpression.! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	
	| expression newBase |
	self == anExpression ifTrue: [^aBaseExpression].
	newBase := base rebuildOn: aBaseExpression startingFrom: anExpression
		withOuterScopeBase: ultimateBaseExpression.	
	"For phantom mappings we have to pass through our mapping since they can't compute their own. For the normal case, if we have a mapping, and our base class hasn't changed, then just re-use the same mapping as an optimization."
	expression := (mapping isNil or: [self mustReGetMappingIfRebuildingOn: newBase])
		ifTrue: [newBase get: name]
		ifFalse: [newBase getMapping: self mapping named: name].
	outerJoin ifTrue: [expression asOuterJoin].
	expression phantom: self phantom.
	^expression.! !

!MappingExpression methodsFor: 'api' stamp: ''!
referenceClass
	"Return the class that our mapping refers to, if we have a mapping. If we're a base expression, just return our class, and if we're a table expression, just return nil."
	^self mapping referenceClass.! !

!MappingExpression methodsFor: 'testing' stamp: ''!
representsDerivedObject
	^self mapping isRelationship.! !

!MappingExpression methodsFor: 'private/accessing' stamp: ''!
requiresDistinct: aBoolean
	super requiresDistinct: aBoolean.
	base requiresDistinct: aBoolean.! !

!MappingExpression methodsFor: 'testing' stamp: ''!
returnsDictionary
	"Return true if we representa  ampping to a dictionary"
	^self mapping isToManyRelationship and: [self mapping collectionType == Dictionary].! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
rewriteBooleanEquality: relation
	"If we are a boolean that doesn't have a comparison relation on top of us, replace us with an ' = true'"

	"Only rewrite booleans"
	(self mapping notNil and: [self mapping attribute type = Boolean]) ifFalse: [^self].

	"See if we're already part of an equality comparison, don't rewrite twice"
	(relation = #= | (relation = #<>) | (relation = #IN)) ifTrue: [^self].
	"If we're a boolean function, we don't need or want the comparison. e.g. EXISTS. This is a bit hard to test for."
	(self mapping field isGlorpExpression and: [self mapping field field isNil]) ifTrue: [^self].

	^self get: #= withArguments: (Array with: true asGlorpExpression).! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
sourceDescriptor
	^base descriptor.! !

!MappingExpression methodsFor: 'api' stamp: ''!
sqlSelect: anExpression
	"Only useful for mappings that represent collections. This restricts the range of values in a variable. Most useful as part of an aggregation. e.g. 
    read: SomeClass where: [:each | (each collection select: [:eachItem | eachItem attribute = someValue]) sqlCount > 10]"

	"We may be passed an expression corresponding to either a field, or to a full object, depending on the aggregate function. In either case, build backwards from the first full object"
	| objectExpression newMapping previousJoin |
	objectExpression := self mapping isRelationship 
		ifTrue: [self] ifFalse: [self base].
	newMapping := self mapping class new.
	newMapping attributeName: #temporaryAttributeFromSubselect.
	newMapping referenceClass: self mapping referenceClass.
	newMapping descriptor: self mapping descriptor.
	newMapping linkFields: self mapping linkFields.
	newMapping usesLinkTable: self mapping usesLinkTable.
	previousJoin := self mapping join.
	newMapping join: ((anExpression rebuildOn: previousJoin ultimateBaseExpression startingFrom: self) AND: previousJoin asGlorpExpression).

	^objectExpression base getMapping: newMapping named: ( 'a temporary restricted mapping for #select: based on ', self mapping attributeName).! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
system
	^base system.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
table
	self hasDescriptor ifTrue: [^self descriptor primaryTable].
	^self field table.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
tables

	self controlsTables ifFalse: [^#()].
	^self tablesContainingMe.! !

!MappingExpression methodsFor: 'accessing' stamp: ''!
tablesContainingMe
	"Return a list of tables that contain this object, whether or not we control them. This is mostly the same as tables, but will differ for embedded mappings."
	| set |
	self descriptor isNil ifTrue: [^#()].
	set := self descriptor tables asSet.
	^set.! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
tablesToPrint

	self hasDescriptor ifFalse: [^#()].
	^self tables collect: [:each |
		self aliasedTableFor: each].! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
translateField: aDatabaseField
	| translatedField |
	translatedField := (self mapping translateFields: (Array with: aDatabaseField)) first.
	^super translateField: (translatedField isNil ifTrue: [aDatabaseField] ifFalse: [translatedField]).! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
translateFields: anOrderedCollection 
	"Ugh. Unify these mechnisms"
	^super translateFields: (self mapping translateFields: anOrderedCollection).! !

!MappingExpression methodsFor: 'navigating' stamp: ''!
ultimateBaseExpression

	^base ultimateBaseExpression.! !

!MappingExpression methodsFor: 'preparing' stamp: ''!
validate
	self mappingFromMeOrSubclasses isNil ifTrue: [self error: 'no mapping for ', self printString].! !

!MappingExpression methodsFor: 'converting' stamp: ''!
valueInBuilder: anElementBuilder
	^self valueInBuilder: anElementBuilder as: self base.! !

!MappingExpression methodsFor: 'converting' stamp: ''!
valueInBuilder: anElementBuilder as: anExpression
	^self mapping valueIn: anElementBuilder as: anExpression.! !

!MappingExpression methodsFor: 'testing' stamp: ''!
willPrintAsWhereClause
	^self hasField.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
aggregate: anExpression as: aggregateFunctionName

	^self aggregate: anExpression as: aggregateFunctionName where: nil.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
aggregate: anExpression as: aggregateFunctionName where: whereExpression
	"Make a subselect of the expression that is our argument, returning an aggregated value. e.g.
session read: GlorpCustomer where: 	[:each |  
	(each aggregate: eachCust accounts as: #countStar where: [:acct | acct price > 1000]]) = 1].
	This is the most general form of aggregation convenience method, and everything else is implemented in terms of it"

	| q expression next objectExpression retrievalExpression basicExpression |
	"We may be passed an expression corresponding to either a field, or to a full object, depending on the aggregate function. In either case, build backwards from the first full object"
	objectExpression := anExpression mapping isRelationship 
		ifTrue: [anExpression] ifFalse: [anExpression base].
	q := Query readManyOf: objectExpression referenceClass.
	expression := q baseExpression.
	next := objectExpression.
	[next base isNil] whileFalse: [
		expression := next mapping addPhantomMappingExpressionOn: expression.
		next := next base].
	"We have to force construct the equality expression, because we don't want 'self' rebuilt onto the other base"
	expression := RelationExpression new outerJoin: false; relation: #=; leftChild: expression; rightChild: self.
	q whereClause: expression.
	"We need to rebuild the thing we're using to retrieve in the right context. Also, if we're called directly, the symbol may be wrapped as a constantExpression. Unwrap it"
	basicExpression := anExpression rebuildOn: q baseExpression startingFrom: objectExpression.
	retrievalExpression := (basicExpression get: aggregateFunctionName asGlorpExpression value).
	q retrieve: retrievalExpression.
	whereExpression == nil ifFalse: [
		q AND: (whereExpression asGlorpExpressionOn: basicExpression)].
	^q asGlorpExpressionOn: self ultimateBaseExpression.! !

!ObjectExpression methodsFor: 'fields' stamp: ''!
aliasTable: aDatabaseTable to: aString 
	| newTable |
	newTable := aDatabaseTable copy.
	newTable name: aString.
	newTable parent: aDatabaseTable.
	newTable schema: ''.
	self tableAliases at: aDatabaseTable put: newTable! !

!ObjectExpression methodsFor: 'fields' stamp: ''!
aliasedTableFor: aDatabaseTable 
	tableAliases isNil ifTrue: [^aDatabaseTable].
	aDatabaseTable isAliased ifTrue: [^aDatabaseTable].
	^tableAliases 
		at: aDatabaseTable 
		ifAbsent: [
			(self tables includes: aDatabaseTable) 	
				ifTrue: [self error: 'Bug. Expression has only partially aliased tables']
				ifFalse: [
					| tableExpression |
					tableExpression := self mappingExpressions at: aDatabaseTable ifAbsent: [nil].
					tableExpression isNil 
						ifTrue: [aDatabaseTable]
						ifFalse: [tableExpression aliasedTableFor: aDatabaseTable]]].! !

!ObjectExpression methodsFor: 'fields' stamp: ''!
aliasedTableFor: aDatabaseTable  ifAbsent: aBlock
	tableAliases isNil ifTrue: [^aBlock value].
	aDatabaseTable isAliased ifTrue: [^aDatabaseTable].
	^tableAliases at: aDatabaseTable ifAbsent: [aBlock value].! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
anySatisfyDefault: anExpression 
	^(self system isNil or: [self system useJoinsForAnySatisfy])
		ifTrue: [self anySatisfyJoin: anExpression]
		ifFalse: [self anySatisfyExists: anExpression].! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
anySatisfyExists: anExpression 
	| newExpression |
	anExpression isConstantExpression ifTrue: [self error: 'The argument to anySatisfy: must be a block evaluating to a boolean condition.'].
	newExpression := SubSelectExpression 
		named: #anySatisfy:
		basedOn: self
		withArguments: (Array with: anExpression).
	^newExpression.! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
anySatisfyJoin: anExpression 
	| newExpression |
	anExpression isConstantExpression ifTrue: [self error: 'The argument to anySatisfy: must be a block evaluating to a boolean condition.'].
	self base requiresDistinct: true.
	newExpression := CollectionExpression 
		named: #anySatisfy:
		basedOn: self
		withArguments: (Array with: anExpression).
	self base removeMappingExpression: self.
	^newExpression.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
asDate
	"Cast the receiver to be a date"
	^self getFunction: #cast: withArguments: (Array with: self system platform date asGlorpExpression).! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
asTimestamp
	"Cast the receiver to be a timestamp"
	^self getFunction: #cast: withArguments: (Array with: self system platform timestamp asGlorpExpression).! !

!ObjectExpression methodsFor: 'preparing' stamp: ''!
assignTableAliasesStartingAt: anInteger prefix: aString
	| tableNumber prefix |
	self controlsTables ifFalse: [^anInteger].
	self hasTableAliases ifTrue: [^anInteger].
	tableNumber := anInteger.
	prefix := aString isNil ifTrue: ['t'] ifFalse: [aString].
	self tables
		do:
			[:each | 
			self aliasTable: each to: prefix , tableNumber printString.
			tableNumber := tableNumber + 1].
	^tableNumber.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
average: anExpression
	^self aggregate: anExpression as: #average.! !

!ObjectExpression methodsFor: 'testing' stamp: ''!
canTranslateFields
	"Return true if we are a mapping expression and can thus translate from fields into their aliased version for a particular query. More generally, do we respond to translateField:"
	^true.! !

!ObjectExpression methodsFor: 'fields' stamp: ''!
controlsTables

	self subclassResponsibility.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
count: anExpression
	"Make a subselect of the expression that is our argument, returning a count(*). e.g.
session read: GlorpCustomer where: 	[:each |  (each count: [:eachCust | eachCust accounts]) = 1]."
	^self aggregate: anExpression as: #countStar.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
count: anExpression where: filterExpression
	"Make a subselect of the expression that is our argument, returning a count(*). e.g.
session read: GlorpCustomer where: 	[:each |  (each count: [:eachCust | eachCust accounts]) = 1]."
	^self aggregate: anExpression as: #countStar where: filterExpression.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
existenceCheck: anExpression as: functionName
	"Make a subselect of the expression that is our argument, returning it wrapped in either an exists or a not exists, e.g. 
	session read: GlorpCustomer where: 	[:each |  
		each existenceCheck: eachCust accounts as: #exists:].
	This is the most general form of existence checking convenience method, others are implemented in terms of it. It's very similar to aggregate:as:where:, but simpler."

	| q expression next objectExpression |
	"We may be passed an expression corresponding to either a field, or to a full object, depending on the aggregate function. In either case, build backwards from the first full object"
	objectExpression := anExpression mapping isRelationship 
		ifTrue: [anExpression] ifFalse: [anExpression base].
	q := Query readManyOf: objectExpression mapping referenceClass.
	expression := q baseExpression.
	next := objectExpression.
	[next base isNil] whileFalse: [
		expression := next mapping addPhantomMappingExpressionOn: expression.
		next := next base].
	"We have to force construct the equality expression, because we don't want 'self' rebuilt onto the other base"
	expression := RelationExpression new outerJoin: false; relation: #=; leftChild: expression; rightChild: self.
	q whereClause: expression.
	"Just read back a literal constant, we don't actually care about the data"
	q retrieve: ([:each | 'x' ] asGlorpExpression getFunction: #distinct withArguments: #()).
	^self get: functionName withArguments: (Array with: (q asGlorpExpressionOn: self ultimateBaseExpression)).! !

!ObjectExpression methodsFor: 'accessing' stamp: ''!
fieldAliases

	fieldAliases isNil ifTrue: [fieldAliases := IdentityDictionary new].
	^fieldAliases.! !

!ObjectExpression methodsFor: 'fields' stamp: ''!
findFieldNamed: aString
	| fields |
	fields := OrderedCollection new: 1.
	self tables do: [:each |
		| field |
		field := each fieldNamed: aString ifAbsent: [nil].
		field isNil ifFalse: [fields add: field]].
	fields isEmpty ifTrue: [self error: 'No field named ', aString, ' in ', self printString].
	fields size > 1 ifTrue: [self error: 'Ambiguous field ', aString, ' in ', self printString].
	^fields at: 1.! !

!ObjectExpression methodsFor: 'api' stamp: ''!
get: aSymbol 
	"Return the mapping expression corresponding to the named attribute"

	| reallyASymbol functionExpression |
	reallyASymbol := aSymbol asSymbol.
	functionExpression := self getFunction: aSymbol withArguments: #().
	functionExpression isNil ifFalse: [^functionExpression].
	^self mappingExpressions at: reallyASymbol
		ifAbsentPut: [MappingExpression named: reallyASymbol basedOn: self]! !

!ObjectExpression methodsFor: 'api' stamp: ''!
get: aSymbol withArguments: anArray 

	| specialExpression functionExpression |
	functionExpression := self getFunction: aSymbol withArguments: anArray.
	functionExpression isNil ifFalse: [^functionExpression].
	specialExpression := self performSpecialExpression: aSymbol withArguments: anArray.
	specialExpression isNil ifFalse: [^specialExpression].
	anArray isEmpty ifTrue: [^self get: aSymbol].
	^RelationExpression
		named: aSymbol
		basedOn: self
		withArguments: anArray! !

!ObjectExpression methodsFor: 'api' stamp: ''!
getConstant: anObject
	"Allows us to refer to a constant object easily. This is most useful for subselects, where I can already write something like
   [:each | (each count: subQuery) = 5]
but since the subselect part might be very verbose, I might prefer to put the constant value first. I can't write
   [:each | 5 = (each count: subQuery)]
because then the left-hand side expression doesn't refer back to the block argument, and we can't build the expression correctly. So I can write
   [:each | (each getConstant: 5) = (each count: subQuery)]"
	^anObject asGlorpExpressionOn: self.! !

!ObjectExpression methodsFor: 'api' stamp: ''!
getField: aField
	| realField |
	realField := aField isString
		ifTrue: [self findFieldNamed: aField]
		ifFalse: [aField].
	"This might be an expression, most notably a constant expression, in which case it either contains a string or a field. This is awfully ugly"
	realField isGlorpExpression
		ifTrue:
			[realField class == ConstantExpression
				ifTrue:
					[realField := realField value isString
						ifTrue: [realField := self findFieldNamed: realField value]
						ifFalse: [realField]]].
	^self getFieldKnownNotToBeAName: realField.! !

!ObjectExpression methodsFor: 'api' stamp: ''!
getFieldKnownNotToBeAName: aField
	"The public getField method may, for convenience, be passed a string or a constant expression with a string that is a field name. This is a mess, and can be confused with a constant value. This method does not allow that."

	| realField |
	realField := aField.
	realField isGlorpExpression
		ifTrue:
			[realField class == ConstantExpression ifTrue: [^realField].
			realField class == ParameterExpression
				ifTrue: [realField := realField field]].
	^self getFieldThatsAnActualFieldObject: realField.! !

!ObjectExpression methodsFor: 'api' stamp: ''!
getFieldThatsAnActualFieldObject: aField
	"If the field isn't from our table, do a getTable first. If we don't know, assume we can get it directly."

	self controlsTables ifFalse: [^self base getFieldThatsAnActualFieldObject: (self translateField: aField)].
	^((self hasDescriptor not) or: [self tables includes: aField table])
		ifTrue:
			[self mappingExpressions
				at: aField
				ifAbsentPut: [self newFieldExpressionFor: aField]]
		ifFalse: [(self getTable: aField table) getField: aField].! !

!ObjectExpression methodsFor: 'api' stamp: ''!
getMapping: aMapping named: aString
	"Get a mapping expression given the actual mapping (this is most useful for an anonymous mapping that isn't named or in our descriptor). Only if you really know what you're doing!!. Short circuit looking up the mapping, checking for functions, and so on."

	^self mappingExpressions at: aString
		ifAbsentPut: [ | expression |
			expression := MappingExpression named: aString basedOn: self.
			expression mapping: aMapping.
			expression].! !

!ObjectExpression methodsFor: 'api' stamp: ''!
getParameter: aDatabaseField 
	^ParameterExpression forField: aDatabaseField basedOn: self.! !

!ObjectExpression methodsFor: 'api' stamp: ''!
getTable: aTable
	"This can take a string, a constantExpression containing a string, or a table object"

	| realTable |
	realTable := aTable isString
		ifTrue: [self system tableNamed: aTable]
		ifFalse: [aTable].	"This might be an expression, most notably a constant expression, in which case it either contains a string or a field"
	realTable isGlorpExpression
		ifTrue:
			[realTable value isString
				ifTrue: [realTable := self system tableNamed: realTable value]
				ifFalse: [realTable := realTable value]].
	^self mappingExpressions
		at: realTable
		ifAbsentPut: [TableExpression forTable: realTable basedOn: self].! !

!ObjectExpression methodsFor: 'testing' stamp: ''!
hasField
	^false.! !

!ObjectExpression methodsFor: 'testing' stamp: ''!
hasProvidedAdditionalExpressions
	"Return true if we have already provided additional expressions to some query. Only important when we're preparing a subselect and don't want to bring in joins that are part of the main query"
	^self hasTableAliases.! !

!ObjectExpression methodsFor: 'testing' stamp: ''!
hasTableAliases
	^tableAliases notNil! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
includes: anObject
	"Allow us to specify includes: as a shortcut for (anySatisfy: [:each | each = anObject]). Rather than just call anySatisfy:, we need to go through the get: mechanism, both to avoid the anySatisfy: method on expressions that treats them like a collection, and to make sure we get the default mechanism for anySatisfy: (either joins or exists)"
	^self get: #anySatisfy: withArguments: (Array with: ([:each | each = anObject] asGlorpExpressionOn: self)).! !

!ObjectExpression methodsFor: 'initialize' stamp: ''!
initialize

	super initialize.
	requiresDistinct := false.! !

!ObjectExpression methodsFor: 'accessing' stamp: ''!
mappingExpressions
	mappingExpressions isNil ifTrue: [mappingExpressions := IdentityDictionary new].
	^mappingExpressions.! !

!ObjectExpression methodsFor: 'accessing' stamp: ''!
mappingExpressions: anObject
	mappingExpressions := anObject! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
max: anExpression
	^self aggregate: anExpression as: #max.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
min: anExpression
	^self aggregate: anExpression as: #min.! !

!ObjectExpression methodsFor: 'fields' stamp: ''!
newFieldExpressionFor: aField

	^FieldExpression forField: aField basedOn: self! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
noneSatisfy: anExpression 
	| newExpression |
	newExpression := SubSelectExpression 
		named: #noneSatisfy:
		basedOn: self
		withArguments: (Array with: anExpression).
	^newExpression.! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
noneSatisfyExists: anExpression 
	| newExpression |
	newExpression := SubSelectExpression 
		named: #noneSatisfy:
		basedOn: self
		withArguments: (Array with: anExpression).
	^newExpression.! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
oneArgumentSpecialSelectors
	^#(#anySatisfyDefault: #anySatisfyJoin: #anySatisfyExists: #getTable: #getField: #fieldNamed: #parameter: #noneSatisfy: #noneSatisfyExists: #getConstant: #count: #sum: #min: #max: #average: #sqlSelect: #includes:)! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
performSpecialExpression: aSymbol withArguments: anArray
	| symbolToPerform |
	symbolToPerform := self specialSymbolToPerformFor: aSymbol.
	(self zeroArgumentSpecialSelectors includes: symbolToPerform) ifTrue: [^self perform: symbolToPerform].
	(self oneArgumentSpecialSelectors includes: symbolToPerform) ifTrue: [^self perform: symbolToPerform with: anArray first ].
	(self twoArgumentSpecialSelectors includes: symbolToPerform) ifTrue: [^self perform: symbolToPerform with: (anArray at: 1) with: (anArray at: 2)].
	^nil.! !

!ObjectExpression methodsFor: 'printing' stamp: ''!
printTableAliasesOn: aStream 
	self hasTableAliases 
		ifTrue: 
			[aStream nextPutAll: ' '.
			tableAliases keysAndValuesDo: [:eachKey :eachValue | 
				aStream nextPutAll: eachKey name, '->', eachValue name , ' ']]! !

!ObjectExpression methodsFor: 'api' stamp: ''!
referenceClass
	"Return the class that our mapping refers to, if we have a mapping. If we're a base expression, just return our class, and if we're a table expression, just return nil."
	^self subclassResponsibility.! !

!ObjectExpression methodsFor: 'private/accessing' stamp: ''!
removeMappingExpression: anExpression
	"Private. Normally you would never do this, but in the case of an anySatisfy: or allSatisfy: we want to have each of them as distinct joins, so we will remove the entry from the mappingExpression of the base, making sure that relationship will not be used for anything else. Since any/allSatisfy: is the only valid use of a collection relationship, we don't have to worry about whether it was used for something else earlier."

	self mappingExpressions removeKey: anExpression name.! !

!ObjectExpression methodsFor: 'private/accessing' stamp: ''!
requiresDistinct
	^requiresDistinct! !

!ObjectExpression methodsFor: 'private/accessing' stamp: ''!
requiresDistinct: aBoolean 
	requiresDistinct := aBoolean! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
specialSymbolToPerformFor: aSymbol
	"Avoid polymorphism between things you can send to an expression inside a query block and things you can do to an expression to iterate over an expression true"
	aSymbol == #anySatisfy: ifTrue: [^#anySatisfyDefault:].
	aSymbol == #select: ifTrue: [^#sqlSelect:].
	^aSymbol.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
sqlAverage
	"Make a subselect of the expression that is our argument, returning an avg(). e.g.
session read: GlorpCustomer where: 	[:each |  each accounts balance sqlAverage > 100]."
	^self queryLevelBaseExpression average: self.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
sqlAvg
	"Make a subselect of the expression that is our argument, returning an avg(). e.g.
session read: GlorpCustomer where: 	[:each |  each accounts balance sqlAvg > 100]."
	^self queryLevelBaseExpression average: self.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
sqlCount
	"Make a subselect of the expression that is our argument, returning a count(). e.g.
session read: GlorpCustomer where: 	[:each |  each accounts sqlCount = 1]."
	^self queryLevelBaseExpression count: self.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
sqlCountWhere: filterExpression
	"Make a subselect of the expression that is our argument, returning a count(). e.g.
session read: GlorpCustomer where: 	[:each |  each accounts sqlCount = 1]."
	^self queryLevelBaseExpression count: self where: filterExpression.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
sqlMax
	"Make a subselect of the expression that is our argument, returning a max(). e.g.
session read: GlorpCustomer where: 	[:each |  each accounts balance sqlMax = 1]."
	^self queryLevelBaseExpression max: self.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
sqlMin
	"Make a subselect of the expression that is our argument, returning a minimum. e.g.
session read: GlorpCustomer where: 	[:each |  each accounts balance sqlMin = 1]."
	^self queryLevelBaseExpression min: self.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
sqlSum
	"Make a subselect of the expression that is our argument, returning a count(*). e.g.
session read: GlorpCustomer where: 	[:each |  each accounts balance sqlSum = 1]."
	^self queryLevelBaseExpression sum: self.! !

!ObjectExpression methodsFor: 'api-aggregates' stamp: ''!
sum: anExpression
	"Make a subselect of the expression that is our argument, returning a sum. e.g.
session read: GlorpCustomer where: 	[:each |  (each sum: [:eachCust | eachCust accounts id]) > 5]."
	^self aggregate: anExpression as: #sum.! !

!ObjectExpression methodsFor: 'accessing' stamp: ''!
system
	self subclassResponsibility.! !

!ObjectExpression methodsFor: 'accessing' stamp: ''!
table
	self subclassResponsibility.! !

!ObjectExpression methodsFor: 'accessing' stamp: ''!
tableAliases
	tableAliases isNil ifTrue: [
		tableAliases := IdentityDictionary new: 3].
	^tableAliases! !

!ObjectExpression methodsFor: 'fields' stamp: ''!
translateField: aDatabaseField
	| newTable rebuilt |
	rebuilt := aDatabaseField isGlorpExpression
		ifTrue: [aDatabaseField asGlorpExpressionOn: self withUltimateBase: self ultimateBaseExpression]
		ifFalse: [aDatabaseField].
	newTable := self aliasedTableFor: rebuilt table.
	newTable == rebuilt table ifTrue: [^rebuilt].
	^self fieldAliases at: aDatabaseField ifAbsentPut: [rebuilt withTable: newTable].! !

!ObjectExpression methodsFor: 'fields' stamp: ''!
translateFields: anOrderedCollection 
	^anOrderedCollection collect: [:each | self translateField: each]! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
twoArgumentSpecialSelectors
	^#(#aggregate:as:)! !

!ObjectExpression methodsFor: 'internal' stamp: ''!
zeroArgumentSpecialSelectors
	^#(#asOuterJoin #sqlCount #sqlSum #sqlMin #sqlMax #sqlAverage #sqlAvg #isEmpty #notEmpty #asDate #asTimestamp #bePhantom).! !

ObjectExpression subclass: #TableExpression
	instanceVariableNames: 'table base'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!TableExpression commentStamp: '<historical>' prior: 0!
This is an expression representing a table. Typically we refer to things in terms of objects which control tables, but sometimes you want to refer directly to a table that's not mapped, either in the process of building up a more OO representing (e.g. referring to a link table) or because we just want to do something more complicated that's not mapped, or incompletely mapped.

Instance Variables:
	base	<BaseExpression | MappingExpression>	The expression on which we are built.
	table	<DatabaseTable>	The table that we represent.

!

!TableExpression classSide methodsFor: 'instance creation' stamp: ''!
forTable: aDatabaseTable basedOn: aBaseExpression

	^self new table: aDatabaseTable base: aBaseExpression; yourself! !

!TableExpression methodsFor: 'preparing' stamp: ''!
aliasedTableFor: aDatabaseTable 
	^self controlsTables 
		ifTrue: [super aliasedTableFor: aDatabaseTable]
		ifFalse: [base aliasedTableFor: aDatabaseTable]! !

!TableExpression methodsFor: 'preparing' stamp: ''!
aliasedTableFor: aDatabaseTable ifAbsent: aBlock
	^self controlsTables 
		ifTrue: [super aliasedTableFor: aDatabaseTable ifAbsent: aBlock]
		ifFalse: [base aliasedTableFor: aDatabaseTable ifAbsent: aBlock]! !

!TableExpression methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."

	^(source tablesContainingMe includes: table) ifTrue: [source] ifFalse: [target getTable: table].! !

!TableExpression methodsFor: 'accessing' stamp: ''!
base

	^base.! !

!TableExpression methodsFor: 'accessing' stamp: ''!
canHaveBase
	"Return true if this type of expression can have a base expression on which other things can be built. Doesn't say whether we actually have a valid one or not."
	^true.! !

!TableExpression methodsFor: 'testing' stamp: ''!
canPrint
	"Return true if we can print into a where clause"
	^false.! !

!TableExpression methodsFor: 'preparing' stamp: ''!
controlsTables
	"We can end up with a table expression built on top of a base that has the same table. If so, we don't count as controlling that table"

	base isNil ifTrue: [^true].
	base hasDescriptor ifFalse: [^true].
	^(base descriptor tables includes: table) not.! !

!TableExpression methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Iterate over the expression tree"

	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	base do: aBlock skipping: aSet.
	aBlock value: self.! !

!TableExpression methodsFor: 'api' stamp: ''!
getFieldThatsAnActualFieldObject: aField

	^self mappingExpressions at: aField ifAbsentPut: [self newFieldExpressionFor: aField].! !

!TableExpression methodsFor: 'api' stamp: ''!
getTable: aTable
	self error: 'getting a table from a table expression doesn''t make sense'.! !

!TableExpression methodsFor: 'printing' stamp: ''!
printOnlySelfOn: aStream

	table printSQLOn: aStream withParameters: #().
	self printTableAliasesOn: aStream.! !

!TableExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	base printOn: aStream.
	aStream nextPut: $..
	table printSQLOn: aStream withParameters: #()! !

!TableExpression methodsFor: 'accessing' stamp: ''!
printsTable

	^true.! !

!TableExpression methodsFor: 'accessing' stamp: ''!
queryLevelBaseExpression
	^base queryLevelBaseExpression.! !

!TableExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	self == anExpression ifTrue: [^aBaseExpression].
	^(base rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression) getTable: table.! !

!TableExpression methodsFor: 'api' stamp: ''!
referenceClass
	"Return the class that our mapping refers to, if we have a mapping. If we're a base expression, just return our class, and if we're a table expression, just return nil."
	^nil.! !

!TableExpression methodsFor: 'accessing' stamp: ''!
table

	^table.! !

!TableExpression methodsFor: 'initialize/release' stamp: ''!
table: aDatabaseTable base: aBaseExpression

	table := aDatabaseTable.
	base := aBaseExpression.! !

!TableExpression methodsFor: 'accessing' stamp: ''!
tables

	^Array with: table.! !

!TableExpression methodsFor: 'preparing' stamp: ''!
tablesToPrint

	self controlsTables ifFalse: [^#()].
	^Array with: (self aliasedTableFor: table).! !

!TableExpression methodsFor: 'accessing' stamp: ''!
ultimateBaseExpression
	^base ultimateBaseExpression.! !

GlorpExpression subclass: #ParameterExpression
	instanceVariableNames: 'base lookupKey'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!ParameterExpression commentStamp: '<historical>' prior: 0!
This is an expression representing a placeholder for a parameter value. So, e.g. object attribute = <parameter for field OWNER_ID>. This is most frequently used for the queries in proxies, although it can be used anywhere. The parameter can be identified by either a field, a name, or an index, which can get fairly ugly.

Instance Variables:
	base	<GlorpExpression>	Our base expression.
	lookupKey	<DatabaseField | SomethingElse>	The thing based on which we look up our parameter value in the array or dictionary of parameters that's passed to the query. So it can be number (for positional usage), a field, or something else.

. !

!ParameterExpression classSide methodsFor: 'instance creation' stamp: ''!
forField: aField basedOn: anObjectExpression

	^self new field: aField base: anObjectExpression; yourself! !

!ParameterExpression methodsFor: 'preparing' stamp: ''!
addToTargetFields: aCollection
	"If we can provide target fields, add them to the collection"
	lookupKey class == DatabaseField ifTrue: [
		aCollection add: lookupKey].! !

!ParameterExpression methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."

	^source getField: lookupKey.! !

!ParameterExpression methodsFor: 'accessing' stamp: ''!
base

	^base! !

!ParameterExpression methodsFor: 'testing' stamp: ''!
canBind
	"Return true if this represents a value that can be bound into a prepared statement"
	^true.! !

!ParameterExpression methodsFor: 'accessing' stamp: ''!
canHaveBase
	"Return true if this type of expression can have a base expression on which other things can be built. Doesn't say whether we actually have a valid one or not."
	^true.! !

!ParameterExpression methodsFor: 'printing' stamp: ''!
convertedDbValueOf: anObject
	"We don't do any conversion"
	^anObject! !

!ParameterExpression methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet

	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	base do: aBlock skipping: aSet.
	aBlock value: self.! !

!ParameterExpression methodsFor: 'accessing' stamp: ''!
field
	^lookupKey! !

!ParameterExpression methodsFor: 'initialize/release' stamp: ''!
field: aDatabaseField base: aBaseExpression 
	lookupKey := aDatabaseField.
	base := aBaseExpression! !

!ParameterExpression methodsFor: 'api' stamp: ''!
get: aSelector withArguments: anArray
	| functionExpression |
	functionExpression := self getFunction: aSelector withArguments: anArray.
	functionExpression isNil ifFalse: [^functionExpression].
	^RelationExpression
		named: aSelector
		basedOn: self
		withArguments: anArray! !

!ParameterExpression methodsFor: 'api' stamp: ''!
getMapping: aMapping named: aSymbol
	"Get an expression given the actual mapping. Only if you really know what you're doing!! In this case, what we're probably doing is rewriting a parameter that's a mapped object into a set of parameters, one for each of its primary keys."
	| expression |
	expression := nil.
	aMapping mappedFields do: [:each |
		expression := (each asGlorpExpressionOn: self) AND: expression].
	^expression.! !

!ParameterExpression methodsFor: 'printing' stamp: ''!
printOn: aStream 
	aStream nextPutAll: 'Parameter('.
	self printTreeOn: aStream.
	aStream nextPut: $)! !

!ParameterExpression methodsFor: 'printing' stamp: ''!
printOnlySelfOn: aStream

	lookupKey printSQLOn: aStream withParameters: #()! !

!ParameterExpression methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	self field type print: (self valueIn: aDictionary) on: aStream.! !

!ParameterExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	[lookupKey printSQLOn: aStream withParameters: #()] on: Dialect error do: [:ex | lookupKey printOn: aStream].! !

!ParameterExpression methodsFor: 'navigating' stamp: ''!
queryLevelBaseExpression
	^base queryLevelBaseExpression.! !

!ParameterExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	self == anExpression ifTrue: [^aBaseExpression].
	self base == ultimateBaseExpression ifTrue: [^self].
	^aBaseExpression getParameter: lookupKey.! !

!ParameterExpression methodsFor: 'navigating' stamp: ''!
ultimateBaseExpression
	^base ultimateBaseExpression.! !

!ParameterExpression methodsFor: 'printing' stamp: ''!
valueIn: aDictionary 
	^aDictionary at: lookupKey.! !

GlorpExpression subclass: #RelationExpression
	instanceVariableNames: 'relation leftChild rightChild outerJoin'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!RelationExpression commentStamp: '<historical>' prior: 0!
This represents a relationship between two sub-expressions, that is something like = or AND. So in object attribute = field, we're the equals node. We have exactly two children, and mostly just print ourselves in between them. However there is some other interesting logic in there about splitting up comparisons between objects into comparisons between keys, and even just printing can get fairly complicated.

Instance Variables:
	leftChild	<GlorpExpression>	The left child.
	rightChild	<GlorpExpression>	The right child.
	outerJoin	<Boolean>	Is this an outer join (only makes sense if we are an equality comparison.
	relation	<String>	The string or symbol that we print for our relationship.

!

RelationExpression subclass: #CollectionExpression
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!CollectionExpression commentStamp: '<historical>' prior: 0!
This represents expressions on collection objects taking a block, which at the moment means just anySatisfy:

We treat this as a relation, but with the special properties that when we convert the right hand side into an expression we assume it's a block and give it a base which is the left-hand side. Also, we don't print this relation when printing SQL, we just print the right hand side.!

!CollectionExpression methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Because our argument is a block, and the thing passed to it is our left child, we can assume that our left child will be referenced via that block. So we'll ignore it. This will help when condensing out primary key expressions, because otherwise us iterating over them forces them to stay around."
	| |
	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	rightChild do: aBlock skipping: aSet.
	aBlock value: self.! !

!CollectionExpression methodsFor: 'printing SQL' stamp: ''!
printSQLOn: aStream withParameters: aDictionary 
	"Don't print the left child or ourselves, just the expression that is the right side.
e.g. aPerson addresses anySatisfy: [:each | each city='Ottawa'] prints as
where (address.city = 'Ottawa')
The relation 'aPerson addresses' will ensure that the join gets printed"
	rightChild printSQLOn: aStream withParameters: aDictionary.! !

!RelationExpression classSide methodsFor: 'instance creation' stamp: ''!
named: aSymbol basedOn: anExpression withArguments: anArray

	^self new
		named: aSymbol
		basedOn: anExpression
		withArguments: anArray.! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
addBoundExpressionsIn: aCommand to: boundExpressions
	"If our right-child can be used for binding, add it to the list. We need to do this at this level because the expressions themselves don't know what type they'll be matched against"

	| translated |
	rightChild canBind ifFalse: [^self].
	(boundExpressions includes: self) ifTrue: [^self].
	translated := self convertValueOf: rightChild in: aCommand parameters.
	(self useBindingFor: translated to: leftChild type in: aCommand) ifTrue: [boundExpressions add: self].! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
additionalExpressions

	^#().! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
allSourceFields
	"If we are sending this, then we're treating the expression as a join, for a complex (and probably read-only) relationship. So in that case we want to treat the source fields as the parameters that this expression takes, and specifically the parameters that are based on fields."
	^self inject: OrderedCollection new into: [:sum :each |
		each addToTargetFields: sum.
		sum].! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
allTargetFields
	"If we are sending this, then we're treating the expression as a join, for a complex (and probably read-only) relationship. So in that case we want to treat the target fields as being the source fields of the bottom-most mapping. That is, the fields we'll use to connect to the thing we want are the things that our expression's base (which represents the thing we want) would use to connect to the first level mapping arising from it. If we have a relation, that's a bit of a mess, but guess that we want the left child's information."
	^self leftChild allTargetFields.! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."

	| left right |
	left := leftChild asExpressionJoiningSource: source toTarget: target.
	right := rightChild asExpressionJoiningSource: source toTarget: target.
	^self class new 
		relation: relation;
		leftChild: left;
		rightChild: right.! !

!RelationExpression methodsFor: 'converting' stamp: ''!
asGeneralGlorpExpression
	"Convert the result to a general (tree-format) expression, if it's the more limited join expression."
	| result |
	result := self copy.
	result leftChild: leftChild asGeneralGlorpExpression.
	result rightChild: rightChild asGeneralGlorpExpression.
	^result.! !

!RelationExpression methodsFor: 'converting' stamp: ''!
asGeneralGlorpExpressionWithFields
	"Convert the result to a general (tree-format) expression, if it's the more limited join expression."
	| result |
	result := self copy.
	result leftChild: leftChild asGeneralGlorpExpressionWithFields.
	result rightChild: rightChild asGeneralGlorpExpressionWithFields.
	^result.! !

!RelationExpression methodsFor: 'converting' stamp: ''!
asIndependentJoins
	| |
	"If this is an ANDed clause, referring to two different tables split it into independent joins"
	relation == #AND ifFalse: [^Array with: self].
	leftChild tablesForANSIJoin asSortedCollection = rightChild tablesForANSIJoin asSortedCollection
		ifTrue: [^Array with: self].
	^(Array with: leftChild with: rightChild) inject: OrderedCollection new into: [:sum :each |
		sum addAll: each asIndependentJoins. sum].! !

!RelationExpression methodsFor: 'converting' stamp: ''!
asJoin
	| leftField rightField |
	leftField := self leftChild field.
	rightField := self rightChild field.
	(leftField notNil and: [rightField notNil]) ifTrue: [^Join from: leftField to: rightField].
	^leftChild asJoin, rightChild asJoin.! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
beOuterJoin
	outerJoin := true.
	leftChild beOuterJoin.
	rightChild beOuterJoin.! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
bindingIn: aCommand

	^self convertValueOf: rightChild in: aCommand parameters.! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
bindingType
	"If we're to be used for binding, return the field type to be bound to"

	^self leftChild bindingType.! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
canHaveBase
	"Return true if this type of expression can have a base expression on which other things can be built. Doesn't say whether we actually have a valid one or not."
	^true.! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
canUseBinding
	"Return true if we can use binding for our right child's value"
	^self expectsCollectionArgument not.! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
condenseBooleanEquality
	(relation = #AND | (relation = #OR)) ifFalse: [^self].
	(leftChild isGlorpExpression and: [leftChild isConstantExpression])
		ifTrue: [
			(relation = #AND and: [leftChild value = true]) ifTrue: [^rightChild].
			(relation = #AND and: [leftChild value = false]) ifTrue: [^leftChild].
			(relation = #OR and: [leftChild value = true]) ifTrue: [^leftChild].
			(relation = #OR and: [leftChild value = false]) ifTrue: [^rightChild]].
	(rightChild isGlorpExpression and: [rightChild isConstantExpression])
		ifTrue: [
			(relation = #AND and: [rightChild value = true]) ifTrue: [^leftChild].
			(relation = #AND and: [rightChild value = false]) ifTrue: [^rightChild].
			(relation = #OR and: [rightChild value = true]) ifTrue: [^rightChild].
			(relation = #OR and: [rightChild value = false]) ifTrue: [^leftChild]].

	^self.! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
condenseBooleanEqualityExpressions
	leftChild := leftChild condenseBooleanEquality.
	rightChild := rightChild condenseBooleanEquality.! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
condensePrimaryKeyComparison
	"This tries to condense expressions of the form
         something relatedObject id = x
      with 
          something foreignKeyToRelatedObjectId = x, saving a join (and making the thing work at all, in the case where x = nil"

	| fieldBeingCompared baseObjectExpression join sourceField |
	(relation = #= or: [relation = #<>]) ifFalse: [^self].
	self leftChild mappedFields size = 1 ifFalse: [^self].
	fieldBeingCompared := self leftChild field.
	fieldBeingCompared class == DatabaseField ifFalse: [^self].
	fieldBeingCompared isPrimaryKey ifFalse: [^self].
	baseObjectExpression := self leftChild base.
	baseObjectExpression representsDerivedObject ifFalse: [^self].
	"If we have an expression for a join, don't even try."
	baseObjectExpression join isJoin ifFalse: [^self].
	join := baseObjectExpression join asGlorpExpressionOn: baseObjectExpression base.
	sourceField := join sourceForTarget: fieldBeingCompared.
	sourceField isNil ifTrue: [^self].

	"OK, we've got a match, replace ourselves with the shortcut"
	^(sourceField isGlorpExpression and: [sourceField isConstantExpression])
		ifTrue: [self rightChild get: relation withArguments: (Array with: sourceField)]
		ifFalse: [(self leftChild base base getField: sourceField) get: relation withArguments: (Array with: self rightChild)].! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
condensePrimaryKeyComparisons
	"Because we're replacing the expression, we need to do it at the parent and replace the previous child"
	leftChild := leftChild condensePrimaryKeyComparison.
	rightChild := rightChild condensePrimaryKeyComparison.! !

!RelationExpression methodsFor: 'iterating' stamp: ''!
convertValueOf: anObject in: aDictionary
	| translated convertedValue |

	translated := anObject isGlorpExpression 
		ifTrue: [anObject valueIn: aDictionary]
		ifFalse: [anObject].
	translated isGlorpExpression ifTrue: [^translated].
	convertedValue := self expectsCollectionArgument 
		ifTrue: [translated collect: [:each | self leftChild convertedDbValueOf: each]]
		ifFalse: [self leftChild convertedDbValueOf: translated].
	^convertedValue.! !

!RelationExpression methodsFor: 'copying' stamp: ''!
createFor: arguments withBase: anExpression symbol: aSymbol
	"Create a copy of our expression with the given parameters. Used in function creation, for 'functions' like isNil that really just turn into relations."

	^self class
		named: relation
		basedOn: anExpression
		withArguments: (Array with: rightChild).! !

!RelationExpression methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	| clauses |
	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	leftChild hasImpliedClauses
		ifTrue:
			[clauses := leftChild allRelationsFor: self.
			clauses do: [:each | each do: aBlock skipping: aSet]]
		ifFalse:
			[leftChild do: aBlock skipping: aSet.
			rightChild do: aBlock skipping: aSet.
			aBlock value: self.].! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
expectsCollectionArgument
	^self relationsWithCollectionArguments includes: relation.! !

!RelationExpression methodsFor: 'converting' stamp: ''!
generalExpressionPart
	| left right |
	left := leftChild generalExpressionPart.
	right := rightChild generalExpressionPart.
	(left isNil and: [right isNil]) ifTrue: [^nil].
	left isNil ifTrue: [^right].
	right isNil ifTrue: [^left].
	^self.! !

!RelationExpression methodsFor: 'api' stamp: ''!
get: aSymbol withArguments: anArray
	"We treat NOT as a function, so we have to check for functions here"
	| functionExpression |
	functionExpression := self getFunction: aSymbol withArguments: anArray.
	functionExpression isNil ifFalse: [^functionExpression].
	^anArray size = 1
		ifTrue: [RelationExpression named: aSymbol basedOn: self withArguments: anArray]
		ifFalse: ["This is either an error, or something defined as a selector on relation expressions"
			self perform: aSymbol withArguments: anArray.].! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
hasBindableExpressionsIn: aCommand
	"Return true if our right-child can be used for binding. We need to do this at this level because the expressions themselves don't know what type they'll be matched against"

	| translated |
	rightChild canBind ifFalse: [^false].
	translated := self convertValueOf: rightChild in: aCommand parameters.
	^self useBindingFor: translated to: leftChild field type in: aCommand.! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
isOuterJoin
	outerJoin isNil ifTrue: [outerJoin := false].
	^outerJoin.! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
leftChild
	^leftChild! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
leftChild: anExpression 
	leftChild := anExpression! !

!RelationExpression methodsFor: 'iterating' stamp: ''!
mappableSourceFieldsDo: aBlock
	"Used when we are emulating a Join"
	self allSourceFields do: [:source |
		source isMappable ifTrue: [aBlock value: source]].! !

!RelationExpression methodsFor: 'private/initializing' stamp: ''!
named: aSymbol basedOn: anExpression withArguments: anArray

	| base right rightBase |
	outerJoin := false.
	relation := self operationFor: aSymbol.
	leftChild := anExpression.

	"The only time we don't expect anExpression to have a base is if it's a constant, in which case the other side should be a variable expression and thus have a base."
	base := anExpression canHaveBase ifTrue: [anExpression ultimateBaseExpression] ifFalse: [anArray first ultimateBaseExpression].
	right := anArray first.
	right isGlorpExpression ifFalse: [rightChild := right asGlorpExpressionOn: base. ^self].
	rightBase := right ultimateBaseExpression.
	rightChild := rightBase ~= base
		ifTrue: [
			right asGlorpExpressionOn: ((rightBase notNil and: [rightBase hasDescriptor]) ifTrue: ["Correlated subselect" rightBase] ifFalse: [base])]
		ifFalse: [right].! !

!RelationExpression methodsFor: 'private/initializing' stamp: ''!
operationFor: aSymbol
	"Simple translation of operators"

	aSymbol == #AND: ifTrue: [^#AND].
	aSymbol == #& ifTrue: [^#AND].
	aSymbol == #OR: ifTrue: [^#OR].
	aSymbol == #| ifTrue: [^#OR].
	aSymbol == #~= ifTrue: [^#<>].
	 aSymbol == #like: ifTrue: [^#LIKE].
	 aSymbol == #ilike: ifTrue: [^#ILIKE].		"Case-insensitive variant of LIKE. Only supported on PostgreSQL at the moment"
	 aSymbol == #in: ifTrue: [^#IN].
	 aSymbol == #notIn: ifTrue: [^#'NOT IN'].

	^aSymbol.! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
outerJoin
	outerJoin isNil ifTrue: [outerJoin := false].
	^outerJoin.! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
outerJoin: aBoolean
	outerJoin := aBoolean.! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
primaryKeyFromDictionary: aDictionary
	"Given a set of parameters, return a primary key suitable for retrieving our target. Do this only if the expression is for a primary key, and has no other conditions than the primary key one"

	| left right field |
	relation = #AND ifTrue: [
		left := leftChild primaryKeyFromDictionary: aDictionary.
		right := rightChild primaryKeyFromDictionary: aDictionary.
		^(left notNil and: [right notNil]) ifTrue: [Array with: left with: right] ifFalse: [nil]].
	relation = #= ifFalse: [^nil].
	field := leftChild fieldFromMeOrSubclasses.
	field isNil ifTrue: [^nil].
	field isGlorpExpression ifTrue: [^nil].
	^(field isPrimaryKey)
		ifTrue: [rightChild valueIn: aDictionary]
		ifFalse: [nil].! !

!RelationExpression methodsFor: 'printing SQL' stamp: ''!
printBasicSQLOn: aCommand withParameters: aDictionary 
	aCommand platform parenthesizeCommandsInCompoundQuery ifTrue: [aCommand nextPut: $(].
	leftChild printSQLOn: aCommand withParameters: aDictionary.
	self 
		printComparisonTo: rightChild
		withParameters: aDictionary
		on: aCommand.
	self printOracleOuterJoinOn: aCommand.
	aCommand platform parenthesizeCommandsInCompoundQuery ifTrue: [aCommand nextPut: $)]! !

!RelationExpression methodsFor: 'printing SQL' stamp: ''!
printComparisonTo: value withParameters: aDictionary on: aStream
	"Horribly convoluted logic to handle the cases where the value might be a constant, an expression that results in a value (constant or parameter) or a regular expression, with the caveat that any value that turns out to be null has to be printed with IS NULL rather than = NULL."

	| translated |
	translated := self convertValueOf: value in: aDictionary.
	translated isGlorpExpression
		ifTrue:
			[
			translated canPrint ifTrue: [
				self printRelationOn: aStream.
				translated
					printSQLOn: aStream
					withParameters: aDictionary]]
		ifFalse:
			[self
				printSimpleValueComparisonTo: translated
				on: aStream].! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
printForANSIJoinTo: table on: aCommand
	"Print ourselves as table JOIN otherTable USING (criteria). Return the table we joined"

	self outerJoin
		ifTrue: [aCommand nextPutAll: ' LEFT OUTER JOIN ']
		ifFalse: [aCommand nextPutAll: ' INNER JOIN '].
	aCommand nextPutAll: table sqlTableName.
	aCommand nextPutAll: ' ON '.
	self printSQLOn: aCommand withParameters: aCommand parameters.
	^table.! !

!RelationExpression methodsFor: 'printing SQL' stamp: ''!
printMicrosoftOuterJoinOn: aCommand
	self isOuterJoin ifFalse: [^self].
	aCommand platform useMicrosoftOuterJoins ifTrue: [
		aCommand nextPutAll: '*'].! !

!RelationExpression methodsFor: 'printing' stamp: ''!
printOnlySelfOn: aStream

	aStream nextPutAll: relation.! !

!RelationExpression methodsFor: 'printing SQL' stamp: ''!
printOracleOuterJoinOn: aCommand
	self isOuterJoin ifFalse: [^self].
	(self leftChild class == self class and: [self leftChild isOuterJoin]) ifTrue: [^self].
	(self rightChild class == self class and: [self rightChild isOuterJoin]) ifTrue: [^self].
	aCommand platform useOracleOuterJoins ifTrue: [
		aCommand nextPutAll: ' (+) '].! !

!RelationExpression methodsFor: 'printing SQL' stamp: ''!
printRelationOn: aStream
	aStream space.
	self printMicrosoftOuterJoinOn: aStream.
	aStream 
		nextPutAll: self relation;
		space.! !

!RelationExpression methodsFor: 'printing SQL' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	self leftChild hasImpliedClauses
		ifTrue:
			[| impliedClauses |
			impliedClauses := self leftChild allRelationsFor: self.
			impliedClauses outerJoin: self outerJoin.
			impliedClauses
				printSQLOn: aStream
				withParameters: aDictionary]
		ifFalse:
			[self
				printBasicSQLOn: aStream
				withParameters: aDictionary].! !

!RelationExpression methodsFor: 'printing SQL' stamp: ''!
printSimpleValueComparisonTo: value on: aStream
	value isNil
		ifTrue: [self printWithNullOn: aStream]
		ifFalse:
			[self printRelationOn: aStream.
			self printValue: value on: aStream].! !

!RelationExpression methodsFor: 'printing' stamp: ''!
printTreeOn: aStream 
	aStream
		print: leftChild;
		space;
		nextPutAll: relation;
		space;
		print: rightChild! !

!RelationExpression methodsFor: 'printing SQL' stamp: ''!
printValue: value on: aCommand
	| type |
	type := [self leftChild type] on: Dialect error do: [:ex | self error: 'Invalid comparison operation: ', self printString].
	(self useBindingFor: value to: type in: aCommand) ifTrue: [^aCommand nextPutAll: '?'].
	self expectsCollectionArgument
		ifTrue:
			[type
				printCollection: value
				on: aCommand]
		ifFalse:
			[type
				print: value
				on: aCommand].! !

!RelationExpression methodsFor: 'printing SQL' stamp: ''!
printWithNullOn: aStream
	aStream nextPutAll: ' IS '.
	self relation = #<> ifTrue: [
		aStream nextPutAll: 'NOT '].
	aStream nextPutAll: 'NULL'.! !

!RelationExpression methodsFor: 'navigating' stamp: ''!
queryLevelBaseExpression

	^leftChild canHaveBase ifTrue: [leftChild queryLevelBaseExpression] ifFalse: [rightChild queryLevelBaseExpression].! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	| expression |
	self == anExpression ifTrue: [^aBaseExpression].
	expression := (leftChild rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression)
		get: relation
		withArguments:
			(Array
				with:
					(self rightChildToRebuild rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression)).
	self isOuterJoin ifTrue: [expression beOuterJoin].
	^expression.! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
relation

	^relation.! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
relation: aSymbol

	relation := aSymbol.! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
relationsWithCollectionArguments
	^#(#IN #'NOT IN')! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
removeNonJoinParts
	"Remove anything that's not an actual join. Only useful if we have something that has mixed joins and general expressions, where we assume we don't want to use the general ones, because they're not field level. Also only meaningful to remove this for ANDs and ORs, not sub-components."
	(relation = #AND or: [relation = #OR]) ifFalse: [^self].
	leftChild := leftChild isJoin ifTrue: [leftChild] ifFalse: [true asGlorpExpression].
	rightChild := rightChild isJoin ifTrue: [rightChild] ifFalse: [true asGlorpExpression].! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
replaceJoinsWithInverse
	"Replace any actual Join objects we have with their inverse. Only useful if we have something that has mixed joins and general expressions, where we assume we don't want to mess with the general ones, because they're not field-level"
	leftChild isJoin ifTrue: [leftChild := leftChild inverseJoin].
	rightChild isJoin ifTrue: [rightChild := rightChild inverseJoin].! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
rewriteEquality
	| keyEquality descriptor describedClass |
	leftChild := leftChild rewriteBooleanEquality: relation.
	leftChild hasDescriptor ifFalse: [^self].
	"A slightly convoluted way of checking if this is a pseudo-descriptor for a primitive type used in something like a DirectToManyMapping, in which case we can't rewrite. But if it's a pseudo-descriptor for a dictionary mapping where we actually use the mapping as the key rather than a class, then we're ok. It'd be nice if the code for this were more expressive."
	descriptor := leftChild descriptor.
	describedClass := descriptor describedClass.
	(describedClass ~~ Association and: [(descriptor session descriptorFor: describedClass) isNil]) ifTrue: [^self].

	(relation = #= | (relation = #<>) | (relation = #IN)) ifFalse: [^self].
	(relation = #IN and: [leftChild descriptor primaryTable primaryKeyFields size > 1]) ifTrue: [
		(GlorpInvalidExpressionError signal: 'Cannot convert an object comparison to an IN comparison when the table uses a composite primary key')].
	keyEquality := leftChild descriptor
		primaryKeyExpressionFor: rightChild
		basedOn: leftChild
		relation: relation.
	^keyEquality.! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
rewriteEqualityExpressions
	leftChild := leftChild rewriteEquality.
	rightChild := rightChild rewriteEquality.! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
rewriteFunctionArguments
	"If a function needs to adjust its arguments, do that here. The motivating example is one where a string argument is rewritten into a form that's of a different length, e.g. the Postgresql driver does that to strings in Blobs."
	leftChild rewriteFunctionArgumentsWith: rightChild.! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
rightChild
	^rightChild! !

!RelationExpression methodsFor: 'accessing' stamp: ''!
rightChild: anExpression
	rightChild := anExpression! !

!RelationExpression methodsFor: 'preparing' stamp: ''!
rightChildToRebuild
	^rightChild.! !

!RelationExpression methodsFor: 'testing' stamp: ''!
shortCircuitOnNullComparison: parameters
	"Return true if we represent an expression that compares null to a non-nullable value, so we don't ever need to really evaluate it.This could be more sophisticated and test if any part of an ANDed and not NOTed clause does such a comparison. But this will do for the most important cases"

	| value |
	value := rightChild valueIn: parameters.
	^(relation == #=
		and: [value isNil and: [leftChild field isNullable not]])! !

!RelationExpression methodsFor: 'converting' stamp: ''!
sourceForTarget: aField

	| left |
	relation = #= ifFalse: [
		left := leftChild sourceForTarget: aField.
		left isNil ifFalse: [^left].
		^rightChild sourceForTarget: aField].
	rightChild field == aField ifTrue: [^leftChild field].
	leftChild field == aField ifTrue: [^rightChild field].
	^nil.! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
tableToJoinFrom: aTableCollection
	"Return a table we could legitimately use at this point to write as an ANSI join - meaning that all but one must be  already used and thus available, and one that isn't. aTableCollection lists the used ones"
	| possibleTables |
	possibleTables := self tablesForANSIJoin asArray
		select: [:each | (aTableCollection includes: each) not].
	^possibleTables size = 1 ifTrue: [possibleTables first] ifFalse: [nil].! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
tablesForANSIJoin
	"Which tables will we join. Assumes this is a single-level join"
	^self inject: Set new into: [:sum :each |
		each tableForANSIJoin isNil ifFalse: [sum add: each tableForANSIJoin].
		sum].! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
targetKeys
	"Assume that part of us is a join, and the rest doesn't count"
	^self inject: OrderedCollection new into: [:sum :each |
		each isJoin ifTrue: [sum addAll: each targetKeys].
		sum].! !

!RelationExpression methodsFor: 'navigating' stamp: ''!
ultimateBaseExpression

	^leftChild canHaveBase ifTrue: [leftChild ultimateBaseExpression] ifFalse: [rightChild ultimateBaseExpression].! !

!RelationExpression methodsFor: 'As yet unclassified' stamp: ''!
useBindingFor: aValue to: aType in: aCommand
	"Return true if we can use binding for our right child's value, in the context of this command"
	aCommand useBinding ifFalse: [^false].
	self expectsCollectionArgument ifTrue: [^false].
	^aCommand canBind: aValue to: aType.! !

RelationExpression subclass: #SubSelectExpression
	instanceVariableNames: 'query localBase'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!SubSelectExpression commentStamp: '<historical>' prior: 0!
This is a special sort of relation expression used when printing subselects that are generated out of anySatisfy: or noneSatisfy: using an EXISTS/NOT EXISTS clause.

Instance Variables:
	localBase	<SubSelectBaseExpression>	The base expression that our sub-query is built on, as opposed to the overall base expression of query.
	query	<SimpleQuery>	The sub-query we are responsible for.

!

!SubSelectExpression methodsFor: 'preparing' stamp: ''!
addBoundExpressionsIn: aCommand to: boundExpressions
	self query prepare.
	self query addBoundExpressionsIn: aCommand to: boundExpressions.! !

!SubSelectExpression methodsFor: 'preparing' stamp: ''!
condensePrimaryKeyComparison
	"Relies on this having been called after the children are condensed. If the child has had its reference to our base condensed away, then we aren't necessary, the comparison is at the same level and no longer needs a subselect"
	^leftChild queryLevelBaseExpression == rightChild queryLevelBaseExpression ifTrue: [rightChild] ifFalse: [self].! !

!SubSelectExpression methodsFor: 'accessing' stamp: ''!
descriptor
	^self leftChild descriptor.! !

!SubSelectExpression methodsFor: 'accessing' stamp: ''!
initializeQuery
	query := Query
		returningManyOf: localBase descriptor describedClass
		where: rightChild.
	query tablePrefix: ('s', localBase ultimateBaseExpression availableSubSelectNumber printString, 't').
"	query baseExpression: localBase."
	query session: localBase descriptor session.
	localBase descriptor tables do: [:eachTable |
		eachTable primaryKeyFields do: [:each | query retrieve: [:base | base getField: each]]].! !

!SubSelectExpression methodsFor: 'internal' stamp: ''!
named: aSymbol basedOn: anExpression withArguments: anArray

	| right |
	outerJoin := false.
	relation := self operationFor: aSymbol.
	leftChild := anExpression base.

	localBase := SubSelectBaseExpression on: anExpression.
	right := anArray first.
	rightChild := right rebuildOn: localBase startingFrom: anExpression.! !

!SubSelectExpression methodsFor: 'printing SQL' stamp: ''!
printSQLOn: aStream withParameters: aDictionary 
	| command toPrint |
	toPrint := relation = #noneSatisfy: ifTrue: ['NOT EXISTS ('] ifFalse: [' EXISTS ('].
	aStream nextPutAll: toPrint.
	command := self query sqlWith: aDictionary.
	aStream nextPutAll: command sqlString.
	aStream nextPutAll: ')'.! !

!SubSelectExpression methodsFor: 'accessing' stamp: ''!
query
	query isNil ifTrue: [self initializeQuery].
	^query.! !

!SubSelectExpression methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	| expression left |
	self == anExpression ifTrue: [^aBaseExpression].
	"We've swallowed our left child, so we have to unswallow it before rebuilding"
	left := localBase trueBase rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression.
	expression := left 
		get: relation
		withArguments:
			(Array
				with:
					(self rightChildToRebuild rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression)).
	self isOuterJoin ifTrue: [expression beOuterJoin].
	^expression.! !

Object subclass: #GlorpExpressionsPackage
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Licence'!

!GlorpExpressionsPackage commentStamp: '<historical>' prior: 0!
This class represents package information for the GlorpExpressions package, for systems that don't have things like package properties to store things like copyright and license information.

LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!

!GlorpExpressionsPackage classSide methodsFor: 'instance creation' stamp: ''!
LICENSE
"LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA"! !

Object subclass: #GlorpExtensionsPackage
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Licence'!

!GlorpExtensionsPackage commentStamp: '<historical>' prior: 0!
This class represents package information for the GlorpExtensions package, for systems that don't have things like package properties to store things like copyright and license information.

LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!

!GlorpExtensionsPackage classSide methodsFor: 'instance creation' stamp: ''!
LICENSE
"LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA"! !

Object subclass: #GlorpHelper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!GlorpHelper commentStamp: '<historical>' prior: 0!
This is a helper class for methods that I'd like to be able to add as extensions, but don't want to for fear of conflicting with dialects that have them, or people who implement them for their own projects, and any other miscellaneous extensions. One extension is that this class is in the Smalltalk namespace, so is universally accessible, and it provides a way for class extensions that don't have visibility of the Glorp namespace to access the Dialect class.!

!GlorpHelper classSide methodsFor: 'classes' stamp: ''!
dialect
	"Provide a way to refer to the dialect class from an extension or class that doesn't include the Glorp namespace, while still avoiding explicit namespace imports (which confuse other dialects when the 'pool dictionaries' don't exist)."
	| glorpNamespace |
	glorpNamespace := Smalltalk at: #Glorp ifAbsent: [nil].
	^((glorpNamespace isNil ifTrue: [Smalltalk] ifFalse: [glorpNamespace]) at: #Dialect ifAbsent: [Smalltalk at: #Dialect]).! !

!GlorpHelper classSide methodsFor: 'helpers' stamp: ''!
do: aBlock for: aCollection separatedBy: separatorBlock

	| array |
	array := aCollection asArray.
	1 to: array size do: [:i |
		| each |
		each := array at: i.
		aBlock value: each.
		i = array size ifFalse: [
			separatorBlock value]].! !

!GlorpHelper classSide methodsFor: 'classes' stamp: ''!
glorpConstantExpressionClass
	^self dialect glorpConstantExpressionClass! !

!GlorpHelper classSide methodsFor: 'helpers' stamp: ''!
print: printBlock on: stream for: aCollection separatedBy: separatorString

	| array |
	array := aCollection asArray.
	1 to: array size do: [:index |
		stream nextPutAll: (printBlock value: (array at: index)).
		index == array size ifFalse: [
			stream nextPutAll: separatorString]].! !

!GlorpHelper classSide methodsFor: 'helpers' stamp: ''!
separate: aCollection by: aOneArgumentBlock

	^aCollection inject: Dictionary new into: [:dict :each |
		| val |
		val := aOneArgumentBlock value: each.
		(dict at: val ifAbsentPut: [OrderedCollection new]) add: each].! !

Object subclass: #GlorpMappingsPackage
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Licence'!

!GlorpMappingsPackage commentStamp: '<historical>' prior: 0!
This class represents package information for the GlorpMappings package, for systems that don't have things like package properties to store things like copyright and license information.

LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!

!GlorpMappingsPackage classSide methodsFor: 'instance creation' stamp: ''!
LICENSE
"LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA"! !

Object subclass: #GlorpMiscPackage
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Licence'!

!GlorpMiscPackage commentStamp: '<historical>' prior: 0!
This class represents package information for the GlorpMisc package, for systems that don't have things like package properties to store things like copyright and license information.

LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!

!GlorpMiscPackage classSide methodsFor: 'instance creation' stamp: ''!
LICENSE
"LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA"! !

Object subclass: #GlorpPreparedStatement
	instanceVariableNames: 'signature statement busy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!GlorpPreparedStatement commentStamp: '<historical>' prior: 0!
This represents a prepared statement in the database. 

Instance Variables:
	busy	<Boolean>	Are we currently in use - that is to say, are we executing and/or holding an active cursor.
	signature	<String>	A string that we can use to match up to commands and see if we are appropriate for the command to use. The signature of a command is its SQL String plus a string representing the types of its parameters.
	statement	<Object>	The actual statement object. Because we don't really know what type this is, and it varies by dialect, we allow the accessor to extract it from us and do the real work.

!

!GlorpPreparedStatement classSide methodsFor: 'instance creation' stamp: ''!
new
	"Answer a newly created and initialized instance."

	^super new initialize! !

!GlorpPreparedStatement methodsFor: 'accessing' stamp: ''!
busy
	^busy.! !

!GlorpPreparedStatement methodsFor: 'accessing' stamp: ''!
busy: aBoolean
	busy := aBoolean.! !

!GlorpPreparedStatement methodsFor: 'As yet unclassified' stamp: ''!
glorpNoticeOfExpiryIn: aSession
	self release.! !

!GlorpPreparedStatement methodsFor: 'initialize-release' stamp: ''!
initialize
	busy := false.! !

!GlorpPreparedStatement methodsFor: 'As yet unclassified' stamp: ''!
release
	super release.
	statement isNil ifFalse: [
		| stmt |
		stmt := statement.
		statement := nil.
		stmt disconnect].! !

!GlorpPreparedStatement methodsFor: 'accessing' stamp: ''!
signature
	^signature! !

!GlorpPreparedStatement methodsFor: 'accessing' stamp: ''!
signature: aString
	signature := aString! !

!GlorpPreparedStatement methodsFor: 'accessing' stamp: ''!
statement
	^statement! !

!GlorpPreparedStatement methodsFor: 'accessing' stamp: ''!
statement: aStatementHandle
	statement := aStatementHandle! !

Object subclass: #GlorpQueries
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Licence'!

!GlorpQueries commentStamp: '<historical>' prior: 0!
This class represents package information for the GlorpQueries package, for systems that don't have things like package properties to store things like copyright and license information.

LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!

Object subclass: #GlorpRowSorter
	instanceVariableNames: 'childSelector queue sorted current items childCache'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!GlorpRowSorter commentStamp: '<historical>' prior: 0!
This sorts a set of rows to determine the order in which they should be inserted into the database. It does this using a breadth first topological sort. What that means is that instead of traversing the graph depth first, as is normal for topological, we traverse it breadth first. This gives us the desirable property of keeping the child rows together as much as possible. Since the initial input will already have been sorted by the static table ordering that is implied by the foreign key constraints, this will also keep rows from the same table together, unless the sort algorithm finds a reason the order needs to be changed. This maximizes the opportunities for using optimizations like Oracle array binding, which want to write groups of similar rows all at once. Note that we can't just use the table order directly, because there may be row-level issues. That is, rows may need to be written before other rows in the same table, or even if there is a table constraint, it may only be relevant for a small number of rows.

Instance Variables:
	childCache	<IdentityDictionary from: Object to: Collection>	The calculation of children may be expensive, and may have to be done more than once for an item, so we cache the information here.
	childSelector	<Symbol | BlockClosure>	The zero-argument selector which we perform on an item, or the one-argument block which we evaluate in order to get the children.
	current	<DatabaseRow | GlorpBreadthFirstTopologicalSortItem>	The item we are currently examining.
	items	<ReadStream>	A stream on the items we are intended to sort.
	queue	<OrderedCollection>	This maintains the breadth first order. So, when we encounter an item, we add its children to the end of the queue. The Sorter processes the items in order off the queue, resulting in a breadth first traversal.
	sorted	<OrderedCollection>	The resulting collection, in sorted order.

!

!GlorpRowSorter classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize.! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
childrenOf: anItem
	^childCache at: anItem ifAbsentPut: [
		childSelector isSymbol
			ifTrue: [anItem perform: childSelector]
			ifFalse: [childSelector value: anItem]].! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
currentChildren
	^self childrenOf: current.! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
currentItemIsDone
	^current status = #done.! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
currentItemIsInProgress
	^current status = #visitedOnce.! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
currentItemIsNew
	^current status = #notVisited.! !

!GlorpRowSorter methodsFor: 'accessing' stamp: ''!
getChildrenVia: aSelector
	childSelector := aSelector.! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
hasSomethingToProcess
	self next.
	^current notNil.! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
initialVisit
	| children |
	children := self currentChildren select: [:each  | each status == #notVisited].
	children do: [:each | each parent: current].
	queue addAll: children.
	current status: #visitedOnce.
	children isEmpty ifTrue: [queue add: current].! !

!GlorpRowSorter methodsFor: 'initializing' stamp: ''!
initialize
	queue := OrderedCollection new: 100.
	sorted := OrderedCollection new: 100.
	childCache := IdentityDictionary new: 100.! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
next
	queue isEmpty ifTrue: [
		| candidate |
		candidate := items detect: [:each | each status == #notVisited] ifNone: [nil].
		candidate isNil ifTrue: [
			"We're probably at the end, but check again for part-way done nodes, which can happen if there are heavy cycles"
			items reset. 
			candidate := items detect: [:each | each status == #visitedOnce] ifNone: [nil]].
		candidate isNil ifFalse: [queue add: candidate]].
	current := queue isEmpty ifTrue: [nil] ifFalse: [queue removeFirst].! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
processCurrent
	self currentItemIsNew ifTrue: [^self initialVisit].
	self currentItemIsInProgress ifTrue: [^self secondVisit].
	self currentItemIsDone ifTrue: [^self].
	self error: 'Unknown item status'.! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
secondVisit
	sorted add: current.
	current status: #done.
	current parent isNil
		ifFalse:
			[((self childrenOf: current parent)
				allSatisfy: [:each | each status == #done])
				ifTrue: [queue add: current parent].
			current parent: nil].! !

!GlorpRowSorter methodsFor: 'sorting' stamp: ''!
sort: aCollection
	items := aCollection readStream.
	aCollection do: [:each | each status: #notVisited].
	[self hasSomethingToProcess] whileTrue: [self processCurrent].
	^sorted.! !

Object subclass: #GlorpSession
	instanceVariableNames: 'system currentUnitOfWork cache accessor applicationData reaper errorAction modificationTracker'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!GlorpSession commentStamp: '<historical>' prior: 0!
This is the main class for communicating between the user and Glorp. It acts as a broker for reading queries, and as the main point for registering objects for modification.

Instance Variables:

	system	<DescriptorSystem>	The definition of our tables, classes, and the relationships between them.
	currentUnitOfWork	<UnitOfWork | nil>	If we have an active unit of work, this will hold it. All modifications are tracked through the unit of work.
	cache	<CacheManager>	Our cache of objects that we've already read.
	accessor	<DatabaseAccessor>	Our connection to the database.
	application	<?>	application-specific data
	reaper <TimedProxyReaper> If we are using TimedProxy for some of our relationships, this holds the reaper which will expire them on timeout.
	errorAction <Block> What block should we evaluate when there's an error.
	modificationTracker <GlorpVWWriteBarrierTracker> VisualWorks-specific. This holds the GlorpVWWriteBarrierTracker in use for this session. This uses immutability to keep track of objects which are changed.
!

!GlorpSession classSide methodsFor: 'instance creation' stamp: ''!
forSystem: aSystem 
	^self new system: aSystem! !

!GlorpSession classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
accessor
	^accessor! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
accessor: aDatabaseAccessor 
	accessor := aDatabaseAccessor.
	system isNil ifFalse: [system platform: (accessor platform)].! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
addToReaper: aProxy

	aProxy class == TimedProxy ifFalse: [^self].
	self reaper add: aProxy.! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
applicationData
	^applicationData! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
applicationData: anObject	
	applicationData := anObject! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
beginTransaction
	accessor beginTransaction.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
beginUnitOfWork

	self hasUnitOfWork ifTrue: [self error: 'Cannot nest units of work yet'].
	currentUnitOfWork := UnitOfWork new.
	currentUnitOfWork session: self.! !

!GlorpSession methodsFor: 'caching' stamp: ''!
cacheAt: aKey  forClass: aClass ifNone: failureBlock

	^cache lookupClass: aClass key: aKey ifAbsent: failureBlock.! !

!GlorpSession methodsFor: 'caching' stamp: ''!
cacheAt: keyObject put: valueObject 
	^cache at: keyObject insert: valueObject! !

!GlorpSession methodsFor: 'caching' stamp: ''!
cacheContainsObject: anObject
	| descriptor key |
	(descriptor := self descriptorFor: anObject) isNil ifTrue: [^false].
	(key := descriptor primaryKeyFor: anObject) isNil ifTrue: [^false].
	^[self cacheContainsObject: anObject key: key] on: DuplicatePrimaryKeyException do: [:ex | false]! !

!GlorpSession methodsFor: 'caching' stamp: ''!
cacheContainsObject: anObject key: aKey 
	"Just test containment, don't return the result or trigger anything due to expiration"
	aKey isNil ifTrue: [^false].
	^cache
		containsObject: anObject
		key: aKey.! !

!GlorpSession methodsFor: 'caching' stamp: ''!
cacheContainsObjectForClass: aClass key: aKey 
	"Just test containment, don't return the result or trigger anything due to expiration"
	aKey isNil ifTrue: [^false].
	^cache
		containsObjectForClass: aClass
		key: aKey.! !

!GlorpSession methodsFor: 'caching' stamp: ''!
cacheContainsObjectForRow: aDatabaseRow
	^self 
		cacheContainsObjectForClass: aDatabaseRow owner class
		key: aDatabaseRow primaryKey! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
cacheFor: anObject
	^cache cacheFor: anObject.! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
cacheForClass: aClass
	^cache cacheForClass: aClass.! !

!GlorpSession methodsFor: 'caching' stamp: ''!
cacheLookupForClass: aClass key: aKey 
	^self 
		cacheAt: aKey
		forClass: aClass
		ifNone: [nil]! !

!GlorpSession methodsFor: 'caching' stamp: ''!
cacheLookupObjectForRow: aDatabaseRow
	^self 
		cacheLookupForClass: aDatabaseRow owner class
		key: aDatabaseRow primaryKey! !

!GlorpSession methodsFor: 'caching' stamp: ''!
cacheRemoveObject: anObject
	| key |
	key := (self descriptorFor: anObject) primaryKeyFor: anObject.
	cache removeClass: anObject class key: key ifAbsent: [].! !

!GlorpSession methodsFor: 'As yet unclassified' stamp: ''!
commandForHomogeneousRows: aCollectionOfDatabaseRows
	"Create the command for this set of rows. These are assumed to be homogeneous, in that they are all for the same table, and are all the same type of operation (i.e. all insert or all update)"
	| representativeRow |
	representativeRow := aCollectionOfDatabaseRows first.
	representativeRow forDeletion ifTrue: [self error: 'invalid way of handling deletes (this should never happen)'].
	^(self shouldInsert: representativeRow)
		ifTrue: [InsertCommand forRows: aCollectionOfDatabaseRows useBinding: self useBinding session: self]
		ifFalse: [UpdateCommand forRows: aCollectionOfDatabaseRows useBinding: self useBinding session: self].! !

!GlorpSession methodsFor: 'internal/writing' stamp: ''!
commandForRow: aDatabaseRow 
	
	(aDatabaseRow forDeletion) ifTrue: [^DeleteCommand forRow: aDatabaseRow useBinding: self useBinding session: self].
	^(self shouldInsert: aDatabaseRow)
		ifTrue: [^InsertCommand forRow: aDatabaseRow useBinding: self useBinding session: self]
		ifFalse: [^UpdateCommand forRow: aDatabaseRow useBinding: self useBinding session: self].! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
commitTransaction
	accessor commitTransaction! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
commitUnitOfWork
	[self isInTransaction
		ifTrue: [currentUnitOfWork commit]
		ifFalse: [self inTransactionDo: [currentUnitOfWork commit]]] 
	ifCurtailed: [self rollbackUnitOfWork].
	currentUnitOfWork := nil.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
commitUnitOfWorkAndContinue
	"Commit the current unit of work, but then keep going with the same set of registered objects, with their state updated to reflect current values."
	| registeredObjects |	
	currentUnitOfWork isNil ifTrue: [^self error: 'Not in unit of work'].
	registeredObjects := currentUnitOfWork registeredObjects.
	currentUnitOfWork deletedObjects do: [:each | registeredObjects remove: each].
	self commitUnitOfWork.
	self beginUnitOfWork.
	self registerAll: registeredObjects.! !

!GlorpSession methodsFor: 'copying' stamp: ''!
copy

	^self shallowCopy postCopy.! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
count: aClass
	^self execute: (SimpleQuery count: aClass).! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
count: aClass where: anExpression
	^self execute: (SimpleQuery count: aClass where: anExpression).! !

!GlorpSession methodsFor: 'internal/writing' stamp: ''!
createDeleteRowsFor: anObject in: rowMap
	"Create records for rows that require deletion"

	(self descriptorFor: anObject) createDeleteRowsFor: anObject in: rowMap.! !

!GlorpSession methodsFor: 'internal/writing' stamp: ''!
createRowsFor: anObject in: rowMap
	| descriptor |
	(descriptor := self descriptorFor: anObject class) isNil
		ifFalse: [descriptor createRowsFor: anObject in: rowMap].! !

!GlorpSession methodsFor: 'table creation' stamp: ''!
createTables
	"Create the tables for our descriptor system. This will fail if the tables already exist. Use recreateTables for that."

	self createTables: system allTables.! !

!GlorpSession methodsFor: 'table creation' stamp: ''!
createTables: aCollection
	"Create the tables for our descriptor system. This will fail if the tables already exist. Use recreateTables for that."

	| sequences |
	system platform areSequencesExplicitlyCreated
		ifTrue:
			[sequences := self system allSequencesFor: aCollection.
			sequences do: [:each | accessor createSequence: each ifError: [:ex | self halt]]].
	aCollection do: [:each | accessor createTable: each ifError: [:ex | self halt]].
	aCollection do: [:each | accessor createTableIndexes: each ifError: [:ex | self halt]].
	self platform supportsConstraints ifTrue: [aCollection do: [:each | accessor createTableFKConstraints: each ifError: [:ex | self halt]]].! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
currentLogin
	^accessor currentLogin! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
currentServerUTCTimestampOrNil
	"This is kind of ugly, but it is Smalltalk Dialect neutral"

	| sqlString result |
	sqlString := [self platform sqlTextCurrentServerUTCTimestamp]
		on: Dialect error
		do:  [:exception | exception return: nil].
	sqlString == nil ifTrue: [^nil].
	result := [self accessor executeSQLString: sqlString]
		on: Dialect error
		do:  [:exception | exception return: nil].
	^result == nil
		ifTrue: [nil]
		ifFalse: 
			[(result first isEmpty or: [(result first first isKindOf: Dialect timestampClass) not])
				ifTrue: [nil]
				ifFalse: [result first first]]! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
delete: anObject
	| realObject |
	"Get the real object, instantiating if necessary"
	realObject := anObject glorpRealObject.
	self hasUnitOfWork
		ifTrue: [currentUnitOfWork delete: realObject]
		ifFalse:
			[self inUnitOfWorkDo: [
				currentUnitOfWork delete: realObject]].! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
deleteAll: aCollection

 	aCollection do: [:each | self delete: each].! !

!GlorpSession methodsFor: 'api' stamp: ''!
descriptorFor: aClassOrInstance

	^system descriptorFor: aClassOrInstance.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
doDDLOperation: aBlock
	"Execute DDL SQL - something which creates/drops/modifies tables or other schema information. The main reason for this method is that some databases require a transaction, while others don't require it, and still others don't allow it. Try to do the right thing."
	
	^self accessor doDDLOperation: aBlock.! !

!GlorpSession methodsFor: 'initialize' stamp: ''!
dontUseModificationTracker
	modificationTracker release.
	modificationTracker := nil.! !

!GlorpSession methodsFor: 'table creation' stamp: ''!
dropTables
	"Drop the tables for our descriptor system, and their associated sequences."

	self dropTables: system allTables.! !

!GlorpSession methodsFor: 'table creation' stamp: 'am From VisualWorksÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂ¬ÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂÃÂ NonCommercial, 7.4.1 of May 30, 2006 on November 20, 2007 at 12:04:18'!
dropTables: aCollection
	"Drop the tables for our descriptor system, and their associated sequences."

	| sequences |
	accessor dropTables: ((TableSorter for: aCollection) sort reversed).
	system platform areSequencesExplicitlyCreated ifTrue: [
		sequences := system allSequencesFor: aCollection.
		accessor dropSequences: sequences].! !

!GlorpSession methodsFor: 'private' stamp: ''!
embeddedObjectsFor: anObject
	"Return all the objects embedded into this one"
	| embedded descriptor |
	descriptor := self descriptorFor: anObject.
	descriptor isNil ifTrue: [^#()].
	embedded := OrderedCollection new: 3.
	descriptor mappings do: [:each |
		(each class == EmbeddedValueOneToOneMapping) ifTrue: [
			embedded add: (each getValueFrom: anObject)]].
	^embedded.! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
errorAction
	^errorAction! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
errorAction: aThreeArgumentBlockExceptionSessionCommand
	errorAction := aThreeArgumentBlockExceptionSessionCommand.! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
execute: aQuery
	^aQuery executeWithParameters: #() in: self.! !

!GlorpSession methodsFor: 'private' stamp: ''!
expiredInstanceOf: aClass key: key

	^cache expiredInstanceOf: aClass key: key.! !

!GlorpSession methodsFor: 'read/write' stamp: ''!
filterDeletionFrom: anObject

	self hasUnitOfWork ifFalse: [^anObject].
	currentUnitOfWork shouldFilterDeletions ifFalse: [^anObject].
	(currentUnitOfWork willDelete: anObject)
		ifTrue: [^nil].
	^anObject.! !

!GlorpSession methodsFor: 'api' stamp: ''!
forceSaveOf: anObject 
	"This is the same as registering an object, but it tells us to disregard whether/how the object has changed and to save all of its fields. This is useful if we need to register an object after changes have occurred, and can be used to implement something more like an object by object save facility. Still somewhat experimental."
	| realObject |
	realObject := self realObjectFor: anObject ifNone: [^self].
	
	self inUnitOfWorkDo: [currentUnitOfWork registerAsNew: realObject].! !

!GlorpSession methodsFor: 'read/write' stamp: ' 22/6/07 12:54'!
groupWrite: aCollectionOfDatabaseRows
	| command |
	aCollectionOfDatabaseRows
		do: [:each | each preWriteAssignSequencesUsing: self accessor].
	command := self commandForHomogeneousRows: aCollectionOfDatabaseRows.
	accessor executeCommand: command.
	^command succeeded
		ifTrue: [true]
		ifFalse:
			[((GlorpWriteFailure new)
				command: command;
				object: (aCollectionOfDatabaseRows collect: [:each | each owner]))
				signal].! !

!GlorpSession methodsFor: 'api' stamp: ''!
hasDescriptorFor: aClass

	^system hasDescriptorFor: aClass.! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
hasExpired: anObject 
	^cache hasExpired: anObject.! !

!GlorpSession methodsFor: 'caching' stamp: ''!
hasExpired: aClass key: key
	^cache hasExpired: aClass key: key.! !

!GlorpSession methodsFor: 'caching' stamp: ''!
hasObjectExpiredOfClass: aClass withKey: key
	^cache hasObjectExpiredOfClass: aClass withKey: key.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
hasUnitOfWork

	^currentUnitOfWork notNil.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
inTransactionDo: aBlock
	"Execute aBlock in a database transaction. If one is already in progress, just continue. If not, start a transaction."
	^self accessor inTransactionDo: aBlock.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
inUnitOfWorkDo: aBlock
	| result myUnitOfWork |
	[
	myUnitOfWork := self hasUnitOfWork not.
	myUnitOfWork ifTrue: [self beginUnitOfWork].
	result := aBlock numArgs = 1 ifTrue: [aBlock value: self] ifFalse: [aBlock value].
	myUnitOfWork ifTrue: [
		self commitUnitOfWork]]
		ifCurtailed:
			[myUnitOfWork ifTrue: [self rollbackUnitOfWork]].
	^result.! !

!GlorpSession methodsFor: 'initialize' stamp: ''!
initialize
	
	self initializeCache.! !

!GlorpSession methodsFor: 'initialize' stamp: ''!
initializeCache

	cache release.
	cache := CacheManager forSession: self.! !

!GlorpSession methodsFor: 'initialize' stamp: ''!
initializeReaper

	self releaseReaper.
	reaper := TimedProxyReaper new.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
isInTransaction
	^accessor isInTransaction.! !

!GlorpSession methodsFor: 'login' stamp: ''!
isLoggedIn
	^accessor isLoggedIn.! !

!GlorpSession methodsFor: 'testing' stamp: ''!
isNew: anObject 
	"When registering, do we need to add this object to the collection of new objects? New objects are treated specially when computing what needs to be written, since we don't have their previous state"

	| key descriptor |
	(currentUnitOfWork notNil and: [currentUnitOfWork isRegistered: anObject]) ifTrue: [^false].
	descriptor := self descriptorFor: anObject.
	descriptor isNil ifTrue: [^false].
	"For embedded values we assume that they are not new. This appears to work. I can't really justify it."
	self needsWork: 'cross your fingers'.
	descriptor mapsPrimaryKeys ifFalse: [^false].

	key := descriptor primaryKeyFor: anObject.
	key isNil ifTrue: [^true].
	"If the cache contains the object, but the existing entry is due to be deleted, then count this entry as a new one being added with the same primary key (ick) as the old one"
	^[(self cacheContainsObject: anObject key: key) not]
		on: DuplicatePrimaryKeyException
		do: [:ex |
			(currentUnitOfWork notNil and: [currentUnitOfWork willDelete: ex existingObject]) 
				ifTrue: [
					self cacheRemoveObject: ex existingObject.
					ex return: true]
				ifFalse: [ex pass]].! !

!GlorpSession methodsFor: 'caching' stamp: ''!
isRegistered: anObject

	currentUnitOfWork isNil ifTrue: [^false].
	^currentUnitOfWork isRegistered: anObject.! !

!GlorpSession methodsFor: 'testing' stamp: ''!
isUninstantiatedProxy: anObject
	^anObject isGlorpProxy and: [anObject isInstantiated not].! !

!GlorpSession methodsFor: 'login' stamp: ''!
login
	self isLoggedIn ifTrue: [^nil].
	^self loginIfError: [:ex | ex pass]! !

!GlorpSession methodsFor: 'login' stamp: ''!
loginIfError: aBlock
	| result |
	result := self accessor loginIfError: aBlock.
	system platform characterEncoding: accessor encoding.
	^result.! !

!GlorpSession methodsFor: 'login' stamp: ''!
logout
	^self accessor logout.! !

!GlorpSession methodsFor: 'caching' stamp: ''!
lookupRootClassFor: aClass
	| descriptor |
	descriptor := self system descriptorFor: aClass.
	^descriptor notNil
			ifTrue: [ descriptor typeMappingRootDescriptor describedClass ]
			ifFalse: [ aClass ]! !

!GlorpSession methodsFor: 'private' stamp: ''!
markAsCurrentOfClass: aClass key: key

	cache markAsCurrentOfClass: aClass key: key.! !

!GlorpSession methodsFor: 'table creation' stamp: ''!
migrateFromSystem: from toSystem: to
	| fromTables toTables toDelete toAdd toModify newVersion |
	fromTables := from allTables copy.
	toTables := to allTables copy.
	toDelete := fromTables reject: [:each | 
		toTables anySatisfy: [:otherTable | otherTable sqlString sameAs: each sqlString]].
	toAdd := toTables reject: [:each | 
		fromTables anySatisfy: [:otherTable | otherTable sqlString sameAs: each sqlString]].
	toModify := fromTables reject: [:each | (toDelete includes: each) or: [toAdd includes: each]].
	self doDDLOperation: [
		self createTables: toAdd.
		self dropTables: toDelete.
		toModify do: [:each | 
			newVersion := to tableNamed: each sqlString.
			self updateTableFrom: each to: newVersion]].! !

!GlorpSession methodsFor: 'table creation' stamp: ''!
migrateFromVersion: fromVersionNumber toVersion: toVersionNumber 
	self halt.
	"Get the class name of our system, strip off version number, replace with to and from, create systems for them."! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
modificationTracker
	"If we killed off the tracker due to a session reset, then bring it back again."
	modificationTracker == #needed ifTrue: [self useModificationTracker].
	^modificationTracker! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
modificationTracker: aGlorpModificationTrackingPolicy
	modificationTracker := aGlorpModificationTrackingPolicy! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
modify: anObject in: aBlock
	"Evaluate aBlock inside a unit of work. Start a database transaction at the beginning and commit it at the end. If we don't terminate normally, roll everything back. Register anObject before starting. Explicitly circumvent the registration of  the block result at the end, because that can cause problems if we're doing things like delete and re-insert and end up returning e.g. the deleted object"
	| result |
	self transact: [
		self register: anObject.
		result := aBlock value.
		nil].
	^result.! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
platform
	^self system platform.! !

!GlorpSession methodsFor: 'copying' stamp: ''!
postCopy

	super postCopy.
	self initializeCache.
	currentUnitOfWork := nil.
	accessor := accessor copy.! !

!GlorpSession methodsFor: 'private' stamp: ''!
privateGetCache

	^cache.! !

!GlorpSession methodsFor: 'private' stamp: ''!
privateGetCurrentUnitOfWork

	^currentUnitOfWork.! !

!GlorpSession methodsFor: 'private' stamp: ''!
privateRegisterAsNew: anObject 
	"Register the object as one which is not already in cache. Private!! Normally just use the register: operation."
	currentUnitOfWork isNil ifTrue: [^self].
	currentUnitOfWork registerAsNew: anObject.! !

!GlorpSession methodsFor: 'private' stamp: ''!
privateRegisterAsOld: anObject 
	"Register the object as something we already read from the database, skipping the isNew: test. Private!! Normally you would just use register:"
	| realObject |
	currentUnitOfWork isNil ifTrue: [^self].
	realObject := self realObjectFor: anObject ifNone: [^self].

	currentUnitOfWork register: realObject! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
read: aClass
	^self execute: (SimpleQuery returningManyOf: aClass).! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
read: aClass limit: anInteger
	^self execute: (SimpleQuery returningManyOf: aClass limit: anInteger).! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
read: aClass orderBy: aCriteria
	| query |
	query := SimpleQuery returningManyOf: aClass.
	query orderBy: aCriteria.
	^self execute: query.! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
read: aClass where: aBlock
	^self execute: (SimpleQuery returningManyOf: aClass where: aBlock).! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
read: aClass where: aBlock limit: anInteger
	^self
		execute:
			(SimpleQuery
				returningManyOf: aClass
				where: aBlock
				limit: anInteger).! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
readManyOf: aClass
	^self execute: (SimpleQuery returningManyOf: aClass).! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
readManyOf: aClass limit: anInteger
	^self execute: (SimpleQuery returningManyOf: aClass limit: anInteger).! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
readManyOf: aClass orderBy: aCriteria
	| query |
	query := SimpleQuery returningManyOf: aClass.
	query orderBy: aCriteria.
	^self execute: query.! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
readManyOf: aClass where: aBlock
	^self execute: (SimpleQuery returningManyOf: aClass where: aBlock).! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
readManyOf: aClass where: aBlock limit: anInteger
	^self
		execute:
			(SimpleQuery
				returningManyOf: aClass
				where: aBlock
				limit: anInteger).! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
readOneOf: aClass
	^self execute: (SimpleQuery returningOneOf: aClass).! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
readOneOf: aClass where: aBlock
	^self execute: (Query returningOneOf: aClass where: aBlock).! !

!GlorpSession methodsFor: 'private' stamp: ''!
realObjectFor: anObject 
	"If this is a proxy, return the contents (if available). Otherwise, return nil"
	^self realObjectFor: anObject ifNone: [nil].! !

!GlorpSession methodsFor: 'private' stamp: ''!
realObjectFor: anObject ifNone: aBlock
	"If this is a proxy, return the contents (if available). Otherwise, evalute the block"
	^anObject isGlorpProxy 
		ifTrue: [anObject isInstantiated ifTrue: [anObject getValue] ifFalse: [aBlock value]]
		ifFalse: [anObject]! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
reaper

	reaper isNil ifTrue: [self initializeReaper].
	^reaper.! !

!GlorpSession methodsFor: 'table creation' stamp: ''!
recreateTables
	"Drop and re-create the tables for our descriptor system, and their associated sequences."

	self recreateTables: system allTables.! !

!GlorpSession methodsFor: 'table creation' stamp: ''!
recreateTables: aCollection
	"Drop the tables for our descriptor system, and their associated sequences."

	| sequences |
	accessor dropTables: (TableSorter for: aCollection) sort reverse.
	system platform areSequencesExplicitlyCreated
		ifTrue:
			[
			sequences := self system allSequencesFor: aCollection.
			accessor dropSequences: sequences.
			sequences
				do: [:each | accessor createSequence: each ifError: [:ex | self halt]]].
	aCollection do: [:each | accessor createTable: each ifError: [:ex | self halt]].
	aCollection do: [:each | accessor createTableIndexes: each ifError: [:ex | self halt]].
	self platform supportsConstraints ifTrue: [aCollection do: [:each | accessor createTableFKConstraints: each ifError: [:ex | self halt]]].! !

!GlorpSession methodsFor: 'table creation' stamp: ''!
recreateTables: aCollection ifError: aBlock
	"Drop the tables for our descriptor system, and their associated sequences."

	(TableSorter for: aCollection) sort reverse
		do: [ :table | accessor dropTable: table ifError: [ :ex | ex "so we can insert a breakpoint"] ].

	system platform areSequencesExplicitlyCreated
		ifTrue: [(self system allSequencesFor: aCollection)
					do: [:each | 
						accessor dropSequence: each ifAbsent: [:ex | ex "so we can insert a breakpoint" ].
						accessor createSequence: each ifError: [:ex | ^aBlock value: ex ]]].

	aCollection do: [:each | accessor createTable: each ifError: [:ex | ^aBlock value: ex]].
	aCollection do: [:each | accessor createTableIndexes: each ifError: [:ex | ^aBlock value: ex ]].
	self platform supportsConstraints ifTrue: [aCollection do: [:each | accessor createTableFKConstraints: each ifError: [:ex | ^aBlock value: ex]]].! !

!GlorpSession methodsFor: 'table creation' stamp: ''!
recreateTablesIfError: aBlock
	"Drop and re-create the tables for our descriptor system, and their associated sequences."

	self recreateTables: system allTables ifError: aBlock .! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
refresh: anObject
	| exp query realObject descriptor |
	realObject := self realObjectFor: anObject ifNone: [^self].
	descriptor := self descriptorFor: realObject.
	descriptor isNil
		ifTrue: [self error: 'Cannot refresh an object with no descriptor'].
	exp := descriptor
		primaryKeyExpressionFor: realObject
		basedOn: BaseExpression new
		relation: #=.
	query := Query returningOneOf: realObject class where: exp.
	query shouldRefresh: true.
	^self execute: query.! !

!GlorpSession methodsFor: 'api' stamp: ''!
register: anObject 
	"Register the object with our current unit of work. We need to do this *before* making changes to the objects."
	| realObject |
	currentUnitOfWork isNil ifTrue: [^self].
	realObject := self realObjectFor: anObject ifNone: [^self].

	(self isNew: realObject) 
		ifTrue: [currentUnitOfWork registerAsNew: realObject]
		ifFalse: [currentUnitOfWork register: realObject]! !

!GlorpSession methodsFor: 'api' stamp: ''!
registerAll: aCollection
	aCollection do: [:ea | self register: ea].! !

!GlorpSession methodsFor: 'api' stamp: ''!
registerForDelete: anObject 
	"Register the object with our current unit of work, with intent to delete it."
	| realObject |
	currentUnitOfWork isNil ifTrue: [^self].
	realObject := self realObjectFor: anObject ifNone: [^self].

	currentUnitOfWork registerForDelete: realObject.! !

!GlorpSession methodsFor: 'initialize' stamp: ''!
releaseReaper

	reaper release.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
requireUnitOfWork
	"Start a unit of work if there isn't one already."
	self hasUnitOfWork ifFalse: [self beginUnitOfWork].! !

!GlorpSession methodsFor: 'initialize' stamp: ''!
reset
	self initializeCache.
	self releaseReaper.
	accessor reset.
	currentUnitOfWork := nil.
	modificationTracker release.
	modificationTracker isNil ifFalse: [modificationTracker := #needed].! !

!GlorpSession methodsFor: 'initialize' stamp: ''!
resetPermissions
	accessor isNil ifFalse: [accessor resetPermissions].! !

!GlorpSession methodsFor: 'testing' stamp: ''!
reusePreparedStatements
	^accessor reusePreparedStatements! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
reusePreparedStatements: aBoolean
	accessor reusePreparedStatements: aBoolean.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
rollbackTransaction
	accessor execute: [ accessor rollbackTransaction].! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
rollbackUnitOfWork

	currentUnitOfWork isNil ifTrue: [^self].
	currentUnitOfWork abort.
	currentUnitOfWork := nil.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
rollbackUnitOfWorkFor: anObject
	currentUnitOfWork abort: anObject! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
save: anObject 
	"Make sure anObject will be written to the database by starting a unit of work if we aren't already in one, and registering the object. This is in most ways the same as just registering an object, but it tells us to disregard whether/how the object has changed and to save all of its fields. This is useful if we need to register an object *after* changes have occurred, and can be used to implement something more like an object by object save facility. Still somewhat experimental."
	| realObject |
	realObject := self realObjectFor: anObject ifNone: [^self].
	
	self hasUnitOfWork
		ifTrue:	[currentUnitOfWork registerAsNew: realObject]
		ifFalse:
			[self inUnitOfWorkDo: [
				currentUnitOfWork registerAsNew: realObject]].! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
saveAndContinue
	"A shorter synonym"
	^self commitUnitOfWorkAndContinue.! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
schema

	^ self accessor schema! !

!GlorpSession methodsFor: 'events' stamp: ''!
sendPostFetchEventTo: anObject
	^anObject glorpPostFetch: self.! !

!GlorpSession methodsFor: 'events' stamp: ''!
sendPostFetchValidateEventTo: anObject
	^anObject glorpPostFetchValidate: self.! !

!GlorpSession methodsFor: 'events' stamp: ''!
sendPostWriteEventTo: anObject
	anObject glorpPostWrite: self.! !

!GlorpSession methodsFor: 'events' stamp: ''!
sendPreWriteEventTo: anObject
	anObject glorpPreWrite: self.! !

!GlorpSession methodsFor: 'internal/writing' stamp: ''!
shouldInsert: aDatabaseRow 
	(currentUnitOfWork notNil and: [currentUnitOfWork forcesUpdateFor: aDatabaseRow]) ifTrue: [^false].
	^(self cacheContainsObjectForRow: aDatabaseRow) not.! !

!GlorpSession methodsFor: 'initialize' stamp: ''!
shouldUseModificationTracker: aBoolean
	aBoolean
		ifTrue: [modificationTracker isNil ifTrue: [self useModificationTracker]]
		ifFalse: [modificationTracker := nil].! !

!GlorpSession methodsFor: 'accessing' stamp: ''!
system
	^system! !

!GlorpSession methodsFor: 'api' stamp: ''!
system: aSystem 
	aSystem session: self.
	accessor isNil ifFalse: [aSystem platform: (accessor currentLogin database)].
	system := aSystem.
	system validate.! !

!GlorpSession methodsFor: 'internal/writing' stamp: ''!
tablesInCommitOrder

	^system tablesInCommitOrder.! !

!GlorpSession methodsFor: 'api/transactions' stamp: ''!
transact: aBlock
	"Evaluate aBlock inside a unit of work. Start a database transaction at the beginning and commit it at the end. If we don' terminate normally, roll everything back.  This might more consistently be  called inUnitOfWorkWithTransactionDo:, but that's too verbose"
	| alreadyInTransaction alreadyInUnitOfWork result |
	[[
	alreadyInUnitOfWork := false.
	alreadyInTransaction := self isInTransaction.
	alreadyInTransaction ifFalse: [self beginTransaction].
	alreadyInUnitOfWork := self hasUnitOfWork.
	alreadyInUnitOfWork ifFalse: [self beginUnitOfWork].
	result := aBlock numArgs = 1 ifTrue: [aBlock value: self] ifFalse: [aBlock value].
	self register: result.
	alreadyInUnitOfWork ifFalse: [self commitUnitOfWork].
	alreadyInTransaction ifFalse: [self commitTransaction]]
		ifCurtailed:
			[alreadyInUnitOfWork ifFalse: [self rollbackUnitOfWork]]]
			ifCurtailed: [alreadyInTransaction ifFalse: [self rollbackTransaction]].
	^result.! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
undelete: anObject
	self undelete: anObject ifNotDeleted: [self error: 'Trying to undelete something that was not deleted'].! !

!GlorpSession methodsFor: 'api/queries' stamp: ''!
undelete: anObject ifNotDeleted: aBlock
	| realObject |
	"Get the real object, instantiating if necessary"
	realObject := anObject glorpRealObject.
	self hasUnitOfWork
		ifTrue: [currentUnitOfWork undelete: realObject ifNotDeleted: aBlock]
		ifFalse:
			[self error: 'Can only undelete while the unit of work is in progress'].! !

!GlorpSession methodsFor: 'table creation' stamp: ''!
updateTableFrom: aTable to: newTable
	"Update the given table based on the metadata."
	
	| newFields deletedFields modifiedFields newFKConstraints deletedFKConstraints newUniqueConstraints deletedUniqueConstraints |
	newFKConstraints := newTable foreignKeyConstraints reject: 
		[:each | aTable hasForeignKeyConstraintMatching: each].
	deletedFKConstraints := aTable foreignKeyConstraints reject: 
		[:each | newTable hasForeignKeyConstraintMatching: each].
	newUniqueConstraints := newTable uniqueConstraints reject: 
		[:each | aTable hasUniqueConstraintMatching: each].
	deletedUniqueConstraints := aTable uniqueConstraints reject: 
		[:each | newTable hasUniqueConstraintMatching: each].
	"So far, we just handle adding and deleting fields, not changing them."
	newFields := newTable fields reject: 
		[:each | aTable hasFieldNamed: each name].
	deletedFields := aTable fields reject: 
		[:each | newTable hasFieldNamed: each name].
	modifiedFields := newTable fields reject: 
		[:each | newFields includes: each].
	modifiedFields := modifiedFields reject: 
		[:each | (aTable fieldNamed: each name) isTheSameAs: each].	
	"Cheap version, for modified fields, drop and re-add them."
	newFields addAll: modifiedFields.
	deletedFields addAll: modifiedFields.
	deletedFKConstraints do: [:each | accessor dropConstraint: each].
	deletedUniqueConstraints do: [:each | accessor dropConstraint: each].
	accessor dropFields: deletedFields.
	accessor addFields: newFields.
	newFKConstraints do: [:each | accessor addConstraint: each].
	newUniqueConstraints do: [:each | accessor addConstraint: each]! !

!GlorpSession methodsFor: 'testing' stamp: ''!
useBinding
	self accessor isNil ifTrue: [^false].
	^self accessor useBinding.! !

!GlorpSession methodsFor: 'initialize' stamp: ''!
useModificationTracker
	Dialect isVisualWorks ifFalse: [^self].
	modificationTracker := (Smalltalk
		at: #'Glorp.GlorpVWWriteBarrierTracker'
		ifAbsent: [nil]) ifNotNil: [:class | class new].
	modificationTracker session: self.! !

!GlorpSession methodsFor: 'virtual collections' stamp: ''!
virtualCollectionOf: aClass
	^GlorpVirtualCollection on: aClass in: self.! !

!GlorpSession methodsFor: 'read/write' stamp: ''!
writeHomogeneousRows: aCollectionOfDatabaseRows
	"Write out a collection of rows belonging to a single table"

	| representativeRow succeeded maximumSize start |

	aCollectionOfDatabaseRows isEmpty ifTrue: [^true].
	representativeRow := aCollectionOfDatabaseRows first.
	representativeRow shouldBeWritten ifFalse: [^true].
	succeeded := true.
	maximumSize := self system maximumSizeToGroupWriteFor: aCollectionOfDatabaseRows.
	start := 1.
	[start > aCollectionOfDatabaseRows size] whileFalse: [
		| end |
		end := (start + maximumSize - 1) min: aCollectionOfDatabaseRows size.
		succeeded := succeeded
			& (self writeHomogeneousRows: aCollectionOfDatabaseRows from: start to: end).
		start := end + 1].
	^succeeded.! !

!GlorpSession methodsFor: 'read/write' stamp: ''!
writeHomogeneousRows: aCollectionOfDatabaseRows from: startIndex to: endIndex
	"Write out the specified range from a collection of rows belonging to a single table"

	| succeeded |
	succeeded := true.
	^(self system 	
			canUseGroupedWritesFor: (aCollectionOfDatabaseRows at: startIndex) table)
		ifTrue: [self groupWrite: (aCollectionOfDatabaseRows copyFrom: startIndex to: endIndex)]
		ifFalse:
			[startIndex to: endIndex do: [:i |
				 |eachRow rowSucceeded |
				eachRow := aCollectionOfDatabaseRows at: i.
				rowSucceeded := self writeRow: eachRow.
				succeeded := succeeded & rowSucceeded].
			succeeded].! !

!GlorpSession methodsFor: 'read/write' stamp: ' 22/6/07 12:54'!
writeRow: aDatabaseRow
	| command success |
	aDatabaseRow shouldBeWritten ifFalse: [^true].
	aDatabaseRow preWriteAssignSequencesUsing: self accessor.
	command := self commandForRow: aDatabaseRow.
	accessor executeCommand: command.
	success := true.
	command succeeded
		ifFalse:
			[success := (GlorpWriteFailure new)
				command: command;
				object: (Array with: aDatabaseRow owner);
				signal].
	^success.! !

Object subclass: #GlorpUnitOfWorkPackage
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Licence'!

!GlorpUnitOfWorkPackage commentStamp: '<historical>' prior: 0!
This class represents package information for the GlorpUnitOfWork package, for systems that don't have things like package properties to store things like copyright and license information.

LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
!

!GlorpUnitOfWorkPackage classSide methodsFor: 'instance creation' stamp: ''!
LICENSE
"LGPL(S)

The code is Copyright (C) 2000-2010  Alan Knight

This is a Smalltalk library, licensed under the LGPL, and distributed WITHOUT
ANY WARRANTY (see below).  However, it is not obvious how some of
the terms and concepts in the license should be sensibly applied to Smalltalk
The various Smalltalk image models often do not directly correspond to 
the ideas of programs and libraries that are linked to form executables. 
We, the authors, would like to clarify our interpretation of the LGPL as it
applies to Smalltalk, and what we permit you to do with this code. We
will sometimes use the term LGPL(S) to distinguish the use of the LGPL 
with these clarifications.

This code is intended to be usable as a library, without the intention to 
restrict the license of the program that uses it. Thus, you may use this library
the same way you would normally use any other Smalltalk library. That is, 
you may load it into an image or otherwise make it available for use in whatever
ways are appropriate to the Smalltalk implementation you are using (including,
but not limited to file-in, linking a shared library, or loading a binary 
representation such as a parcel,  BOSS file, image segment or image component).
You may write code that uses the library, including subclassing library classes
and having the library call back to your code using blocks, the #perform: 
mechanism, or similar mechanisms. You may use this code,  including creating
and distributing packaged images, libraries, development images and executables
that include the library code. We do not consider these activities sufficient
to make the entire program/image/executable or any portion of it beyond the
library itself be considered a derivative work, and in any case we place no
restrictions on what licenses you may use for code that uses this library in
these or similar manners.

If, however, you modify the class definitions or methods of the library, we do
consider those modified sections a derivative work, and as such they are fully subject
to the restrictions described in the LGPL. Basically, you'll need to release the
source of any modifications.

If you require any further clarification, we'd be happy to provide it. You can contact
the authors c/o Alan Knight, knight@acm.org.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA"! !

Object subclass: #Join
	instanceVariableNames: 'sources targets base outerJoin owningMapping'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!Join commentStamp: '<historical>' prior: 0!
This is a specialized variety of expression that is more constrained and is used for defining relationships. It has two main purposes
 - ease of construction: Relationships are normally defined by field to field equality expressions (my foreign key field = his primary key field). These are more tedious to create via block expressions, so this provides a simpler syntax.
 - constrained semantics. These define both read and write for the relationship, so fully general expressions won't work (most notably, relations other than equality are hard to write). Using a primaryKeyExpression ensures that we satisfy these constraints.

I'm not completely sure this class is a good idea. It makes for an annoying assymetry between different kinds of expressions. This is especially notable now that we allow sources to be constants. It's possible that all we need is an expression constructor that generates real expressions, but with more convenient syntax and ensuring that the constraints are met.

Note that although these are typically fk=pk, it's allowed to be the other way around -- i.e. our object-level relationships can be the opposite of the way the fk's "point" in the database.

Instance Variables:
	base	<BaseExpression>	The base on which we are built. Mostly used if we want to convert this into a real expression.
	sources	<SequenceableCollection of: (DatabaseField | ConstantExpression)> The source fields (typically the foreign keys)
	targets	<SequenceableCollection of: DatabaseField> The target fields (typically the targets of the foreign keys)

!

!Join classSide methodsFor: 'instance creation' stamp: ''!
from: aField to: anotherField

	^self new
		addSource: aField target: anotherField.! !

!Join classSide methodsFor: 'instance creation' stamp: ''!
from: from1Field to: to1Field
from: from2Field to: to2Field

	^self new
		addSource: from1Field target: to1Field;
		addSource: from2Field target: to2Field.! !

!Join classSide methodsFor: 'instance creation' stamp: ''!
from: from1Field to: to1Field
from: from2Field to: to2Field
from: from3Field to: to3Field

	^self new
		addSource: from1Field target: to1Field;
		addSource: from2Field target: to2Field;
		addSource: from3Field target: to3Field.! !

!Join classSide methodsFor: 'instance creation' stamp: ''!
from: from1Field to: to1Field
from: from2Field to: to2Field
from: from3Field to: to3Field
from: from4Field to: to4Field

	^self new
		addSource: from1Field target: to1Field;
		addSource: from2Field target: to2Field;
		addSource: from3Field target: to3Field;		
		addSource: from4Field target: to4Field! !

!Join classSide methodsFor: 'instance creation' stamp: ''!
from: from1Field to: to1Field
from: from2Field to: to2Field
from: from3Field to: to3Field
from: from4Field to: to4Field
from: from5Field to: to5Field

	^self new
		addSource: from1Field target: to1Field;
		addSource: from2Field target: to2Field;
		addSource: from3Field target: to3Field;		
		addSource: from4Field target: to4Field;
		addSource: from5Field target: to5Field.! !

!Join classSide methodsFor: 'instance creation' stamp: ''!
fromAll: sourceFields toAll: targetFields

	^self new
		sourceFields: sourceFields targetFields: targetFields.! !

!Join classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!Join methodsFor: 'converting' stamp: ''!
, aJoin
	| newJoin |
	aJoin isNil ifTrue: [^self].
	aJoin == self ifTrue: [^self].
	newJoin := (self class fromAll: (sources, aJoin allSourceFields) toAll: (targets, aJoin targetKeys)) removeDuplicates.
	"If either one of the contributors is an outer join, make the result be an outer join"
	newJoin outerJoin: (self outerJoin | aJoin outerJoin).
	^newJoin! !

!Join methodsFor: 'comparing' stamp: ''!
= aJoin
	aJoin class == self class ifFalse: [^false].
	aJoin allSourceFields = sources ifFalse: [^false].
	aJoin allTargetFields = targets ifFalse: [^false].
	^true.! !

!Join methodsFor: 'api' stamp: ''!
addSource: aField target: anotherField 
	| value targetValue |
	value := (self isConstant: aField) 
				ifTrue: [ConstantExpression for: aField]
				ifFalse: [aField].
	sources add: value.
	targetValue := anotherField class == SimpleQuery
		ifTrue: [anotherField asGlorpExpressionOn: base] 
		ifFalse: [anotherField].
	(self isConstant: targetValue) ifTrue: [
		self error: 'You are attempting to set a constant value as the target of a relationship. I suspect you want to set it on the source instead. Either that or you''re trying to use a field name where a field object is required.'].
	targets add: targetValue! !

!Join methodsFor: 'preparing' stamp: ''!
addToTargetFields: aCollection
	"If we can provide target fields, add them to the collection"
	aCollection addAll: targets.! !

!Join methodsFor: 'preparing' stamp: ''!
additionalExpressions
	^#()! !

!Join methodsFor: 'preparing' stamp: ''!
additionalExpressionsIn: aQuery 
	^#()! !

!Join methodsFor: 'accessing' stamp: ''!
allControlledTables

	^self allTables.! !

!Join methodsFor: 'accessing' stamp: ''!
allSourceFields

	^sources! !

!Join methodsFor: 'accessing' stamp: ''!
allTables

	^(targets collect: [:each | each table]) asSet.! !

!Join methodsFor: 'preparing' stamp: ''!
allTablesToPrint

	^targets inject: Set new into: [:sum :each | 
		sum add: each table. sum].! !

!Join methodsFor: 'accessing' stamp: ''!
allTargetFields

	^targets! !

!Join methodsFor: 'converting' stamp: ''!
asExpressionJoiningSource: source toTarget: target
	"Create a general expression which represents this relationship where the values of the targets (which are normally parameters) are supplied out of the context provided by 'target' and the source fields are referencing things out of the context of source. Produces something suitable for ANDing into an expression when doing a join
   Example: If we had CUSTOMER.ADDRESS_ID = ADDRESS.ID as a parameter, and we want to AND this into an expression [:customer | customer address street = 'Main'] then we have customer as a base, and we get 
	(customer.ADDRESS.STREET = 'Main') AND (customer.CUSTOMER.ADDRESS_ID = customer.ADDRESS.ID)
	The primary key expression for the relationship has been translated into field references into the customer and address tables in a particular context."

	| sourceFieldExpression targetFieldExpression completeExpression relationExpression |
	completeExpression := EmptyExpression new base: base.
	sources with: targets
		do: 	[:sourceField :targetField | 
			sourceFieldExpression := source getFieldKnownNotToBeAName: sourceField.
			targetFieldExpression := target getFieldKnownNotToBeAName: targetField.
			relationExpression := sourceFieldExpression equals: targetFieldExpression.
			completeExpression := relationExpression AND: completeExpression].
	outerJoin ifTrue: [completeExpression beOuterJoin].
	^completeExpression! !

!Join methodsFor: 'converting' stamp: ''!
asGeneralGlorpExpression
	"Convert this to a 'normal' expression representing the same information"

	| main clause |
	main := EmptyExpression new base: base.
	sources with: targets
		do: 
			[:eachSource :eachTarget | 
			| srcExp targetExp |
			srcExp := self sourceExpressionFor: eachSource.
			targetExp := self targetExpressionFor: eachTarget.
			"Reversing the order is important because the source is the parameter, and sql won't accept '27 = FOO' "
			clause := targetExp equals: srcExp.
			main := main AND: clause].
	outerJoin ifTrue: [main beOuterJoin].
	^main! !

!Join methodsFor: 'converting' stamp: ''!
asGeneralGlorpExpressionWithFields
	"The normal conversion of a join is to an expression equating fields and parameters. Instead, generate one with field to field comparisons"

	| main clause |
	main := nil.
	sources with: targets
		do: 
			[:eachSource :eachTarget | 
			| srcExp targetExp |
			srcExp := self targetExpressionFor: eachSource.
			targetExp := self targetExpressionFor: eachTarget.
			clause := srcExp equals: targetExp.
			main := main == nil ifTrue: [clause] ifFalse: [main AND: clause]].
	outerJoin ifTrue: [main beOuterJoin].
	^main! !

!Join methodsFor: 'api' stamp: ''!
asGlorpExpression

	^self.! !

!Join methodsFor: 'api' stamp: ''!
asGlorpExpressionForDescriptor: aDescriptor

	base descriptor: aDescriptor.! !

!Join methodsFor: 'api' stamp: ''!
asGlorpExpressionOn: aBaseExpression
	| copy |
	copy := self copy.
	copy base: aBaseExpression.
	copy outerJoin: outerJoin.
	^copy.! !

!Join methodsFor: 'api' stamp: ''!
asOuterJoin
	^self copy beOuterJoin.! !

!Join methodsFor: 'accessing' stamp: ''!
base
	^base.! !

!Join methodsFor: 'accessing' stamp: ''!
base: aBaseExpression
	base := aBaseExpression.! !

!Join methodsFor: 'api' stamp: ''!
beOuterJoin
	outerJoin := true.! !

!Join methodsFor: 'preparing' stamp: ''!
condenseBooleanEquality
	^self.! !

!Join methodsFor: 'preparing' stamp: ''!
condenseBooleanEqualityExpressions
	^self.! !

!Join methodsFor: 'iterating' stamp: ''!
do: aBlock skipping: aSet
	"Joins act as if they don't have children, and just evaluate for themselves"
	| |
	(aSet includes: self) ifTrue: [^self].
	aSet add: self.
	aBlock value: self.! !

!Join methodsFor: 'iterating' stamp: ''!
fieldsDo: aBlock

	sources with: targets do: aBlock.! !

!Join methodsFor: 'accessing' stamp: ''!
from: aField to: anotherField

	^self addSource: aField target: anotherField.! !

!Join methodsFor: 'converting' stamp: ''!
generalExpressionPart
	^nil.! !

!Join methodsFor: 'accessing' stamp: ''!
hasDescriptor

	^false.! !

!Join methodsFor: 'initialize' stamp: ''!
initialize

	sources := OrderedCollection new: 2.
	targets := OrderedCollection new: 2.
	base := BaseExpression new.
	outerJoin := false.! !

!Join methodsFor: 'converting' stamp: ''!
inverseJoin
	| inverse |
	inverse := self class new
		base: base;
		sourceFields: targets
		targetFields: sources.
	inverse owningMapping: self owningMapping.
	^inverse.! !

!Join methodsFor: 'converting' stamp: ''!
isConstant: aTarget
	"The target can be either a constant (which gets turned into a ConstantExpression) or (usually) a DatabaseField, representing a parameter to the query"
	(aTarget class == DatabaseField) ifTrue: [^false].
	(aTarget isGlorpExpression and: [aTarget isConstantExpression not]) ifTrue: [^false].
	^true.! !

!Join methodsFor: 'testing' stamp: ''!
isEmptyExpression
	^sources isEmpty.! !

!Join methodsFor: 'testing' stamp: ''!
isGlorpExpression
	^true.! !

!Join methodsFor: 'testing' stamp: ''!
isJoin
	^true.! !

!Join methodsFor: 'testing' stamp: ''!
isOuterJoin
	^outerJoin.! !

!Join methodsFor: 'api' stamp: ''!
mapFromSource: sourceObject andTarget: targetObject intoRowsIn: aRowMap
	(sourceObject isNil or: [targetObject isNil]) ifTrue: [^self].
	1
		to: sources size
		do:
			[:i | 
			| eachSourceField eachTargetField |
			eachSourceField := sources at: i.
			eachTargetField := targets at: i.
			(self isConstant: eachSourceField)
				ifTrue: [aRowMap at: eachTargetField key: targetObject put: eachSourceField value]
				ifFalse:
					[aRowMap
						unifyField: eachSourceField
						key: sourceObject
						withField: eachTargetField
						key: targetObject]].! !

!Join methodsFor: 'iterating' stamp: ''!
mappableFieldsDo: aBlock

	sources with: targets do: [:source :target |
		source isMappable ifTrue: [aBlock value: source value: target]].! !

!Join methodsFor: 'iterating' stamp: ''!
mappableSourceFieldsDo: aBlock

	sources do: [:source  |
		source isMappable ifTrue: [aBlock value: source]].! !

!Join methodsFor: 'accessing' stamp: ''!
numberOfParameters
	^sources size! !

!Join methodsFor: 'preparing' stamp: ''!
onlyJoinParts
	"Strip out any generic parts, leaving only the things that are actually joins"
	
	^self.! !

!Join methodsFor: 'accessing' stamp: ''!
outerJoin
	^outerJoin.! !

!Join methodsFor: 'accessing' stamp: ''!
outerJoin: aBoolean
	outerJoin := aBoolean.! !

!Join methodsFor: 'debugging' stamp: ''!
owningMapping
	"Return the mapping (if there is one) that is responsible for. Not needed for the code, but might be a useful debugging aid if trying to figure out where a @#%#@% join came from"
	^owningMapping.! !

!Join methodsFor: 'debugging' stamp: ''!
owningMapping: aMapping
	owningMapping := aMapping.! !

!Join methodsFor: 'preparing' stamp: ''!
prepareIn: aQuery 
	"Do nothing."
	aQuery whereClause: (self asGeneralGlorpExpression).
	aQuery whereClause prepareIn: aQuery.! !

!Join methodsFor: 'primary keys' stamp: ''!
primaryKeyFromDictionary: aDictionary
	"Given a set of parameters, return a primary key suitable for retrieving our target. Return either a value for the key, nil for no key found, or an array for compound keys."
	
	sources size = 1 ifTrue: [^aDictionary at: sources first ifAbsent: [nil]].
	^sources collect: [:each | aDictionary at: each ifAbsent: [nil]].! !

!Join methodsFor: 'printing' stamp: ''!
printOn: aStream

	sources with: targets do: [:source :target |
		aStream nextPut: $(.
		source printSQLOn: aStream withParameters: #().
		aStream nextPutAll: ' = '.
		target printSQLOn: aStream withParameters: #().
		aStream nextPutAll: ') ']! !

!Join methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary 
	1 to: sources size
		do: 
			[:i | 
			| eachTarget eachSource sourceValue |
			eachTarget := targets at: i.
			eachSource := sources at: i.
			eachTarget printSQLOn: aStream withParameters: aDictionary.
			sourceValue := (self isConstant: eachSource) ifTrue: [eachSource value] ifFalse: [aDictionary at: eachSource].
			sourceValue isNil 
				ifTrue: [aStream nextPutAll: ' IS NULL ']
				ifFalse: 
					[aStream nextPutAll: ' = '.
					sourceValue printOn: aStream].
			i = targets size ifFalse: [aStream nextPutAll: ' AND ']]! !

!Join methodsFor: 'accessing' stamp: ''!
queryLevelBaseExpression
	^self ultimateBaseExpression.! !

!Join methodsFor: 'As yet unclassified' stamp: ''!
rebuildOn: aGlorpMappingExpression 
	^self asGlorpExpressionOn: aGlorpMappingExpression.! !

!Join methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	^self asGlorpExpressionOn: aBaseExpression.! !

!Join methodsFor: 'converting' stamp: ''!
removeDuplicates
	| existing newSources newTargets |
	existing := Set new.
	newSources := OrderedCollection new: sources size.
	newTargets := OrderedCollection new: targets size.
	sources
		with: targets
		do:
			[:eachSource :eachTarget | 
			| association |
			association := eachSource -> eachTarget.
			(existing includes: association)
				ifFalse:
					[existing add: association.
					newSources add: eachSource.
					newTargets add: eachTarget]].
	sources := newSources.
	targets := newTargets.! !

!Join methodsFor: 'preparing' stamp: ''!
removeNonJoinParts
	"Remove anything that's not an actual join. Only useful if we have something that has mixed joins and general expressions, where we assume we don't want to use the general ones, because they're not field level."
	^self.! !

!Join methodsFor: 'preparing' stamp: ''!
replaceJoinsWithInverse
	"Replace any actual Join objects we have with their inverse. Only useful if we have something that has mixed joins and general expressions, where we assume we don't want to mess with the general ones, because they're not field-level. As actual joins, we ignore this, because it will be done at the parent level"
	^self.! !

!Join methodsFor: 'converting' stamp: ''!
replacing: aTable withEquivalenceFrom: aJoin
	"Create a new join in which references to fields in aTable have been replaced with the equivalent based on a Join (we assume that aTable references are in the sources of aJoin). If the field doesn't occur in the translation, just remove it"
	| newSources newTargets newSource |
	newSources := OrderedCollection new.
	newTargets := OrderedCollection new.
	self fieldsDo: [:eachSource :eachTarget |
		(eachSource isGlorpExpression not and: [eachSource table = aTable])
			ifTrue: [
				newSource := aJoin targetForSource: eachSource.
				newSource isNil ifFalse: [
					newSources add: newSource.
					newTargets add: eachTarget]]].
	^self class fromAll: newSources toAll: newTargets.! !

!Join methodsFor: 'preparing' stamp: ''!
rewriteEquality
	^self.! !

!Join methodsFor: 'testing' stamp: ''!
shortCircuitOnNullComparison: parameters
	"Return true if we represent an expression that compares null to a non-nullable value (in any of our clauses), so we don't ever need to really evaluate it"

	self fieldsDo: [:source :target |
		(target isNullable not and: [(parameters at: source) isNil]) ifTrue: [^true]].
	^false.! !

!Join methodsFor: 'converting' stamp: ''!
sourceExpressionFor: source
	
	^((self isConstant: source) or: [source isMappable not]) "This might be a subselect"
		ifTrue: [source]
		ifFalse: [base getParameter: source].! !

!Join methodsFor: 'initialize' stamp: ''!
sourceFields: sourceFields targetFields: targetFields
	sources := sourceFields.
	targets := targetFields.! !

!Join methodsFor: 'preparing' stamp: ''!
sourceForTarget: aField

	| index |
	index := targets indexOf: aField.
	index = 0 ifTrue: [^nil].
	^sources at: index.! !

!Join methodsFor: 'converting' stamp: ''!
targetExpressionFor: eachTarget 
	^(self isConstant: eachTarget)
		ifTrue: [eachTarget]
		ifFalse: [
			eachTarget isGlorpExpression 
			ifTrue: [eachTarget] 
			ifFalse: [(base getTable: eachTarget table) getField: eachTarget]].! !

!Join methodsFor: 'preparing' stamp: ''!
targetForSource: aField

	| index |
	index := sources indexOf: aField.
	index = 0 ifTrue: [^nil].
	^targets at: index.! !

!Join methodsFor: 'accessing' stamp: ''!
targetKeys

	^targets.! !

!Join methodsFor: 'accessing' stamp: ''!
ultimateBaseExpression

	^base ultimateBaseExpression.! !

Object subclass: #JoinPrinter
	instanceVariableNames: 'joinsToProcess availableTables query'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!JoinPrinter commentStamp: '<historical>' prior: 0!
This is a helper class used for printing collections of joins in the <expletive deleted> ANSI join syntax. This syntax wants us to embed all of the joins into the initial list of tables, so we have to come up with an order for printing that makes sure we get all the joins in properly. This is more complicated than you might think.

Instance Variables:
	availableTables	<(Collection of: DatabaseTable)>	Which are the tables that we plan to print.
	joinsToProcess	<(Collection of: Join)>	Which joins do we have to put in between those tables somehow or other.
	query	<Query>	What is the query in which we are printing all of this.

!

!JoinPrinter classSide methodsFor: 'instance creation' stamp: ''!
for: aQuery
	^self new query: aQuery.! !

!JoinPrinter methodsFor: 'accessing' stamp: ''!
allTables
	^query tablesToPrint.! !

!JoinPrinter methodsFor: 'printing' stamp: ''!
nextJoin
	^joinsToProcess detect: [:eachJoinExpression |
		eachJoinExpression tablesForANSIJoin anySatisfy: [:eachTable |
			availableTables includes: eachTable]].! !

!JoinPrinter methodsFor: 'printing' stamp: ''!
printJoinsOn: aCommand
	"This is pretty hideous. Figure out an order in which to print these stupid joins. As a bonus, they may be circular, so we may have to collapse joins together. And if we collapse, we can't print as we go because we don't know how many parentheses to add, so store them up and print them all at the end"
	| allJoins |
	joinsToProcess := query joins copy.
	availableTables := Set with: self rootTable.
	allJoins := OrderedCollection new.
	[joinsToProcess isEmpty] whileFalse: [
		| next nextTable allRelatedJoins fullJoin|
		next := self nextJoin.
		joinsToProcess remove: next.
		nextTable := next tableToJoinFrom: availableTables.
		allRelatedJoins := joinsToProcess select: [:each | (each tableToJoinFrom: availableTables) = nextTable].
		fullJoin := allRelatedJoins inject: next into: [:sum :each | sum AND: each].
		joinsToProcess removeAll: allRelatedJoins.
		allJoins add: (fullJoin->nextTable).
		availableTables addAll: (next tablesForANSIJoin)].
	allJoins size timesRepeat: [aCommand platform parenthesizeCommandsInCompoundQuery ifTrue: [aCommand nextPut: $(]].
	aCommand nextPutAll: self rootTable sqlTableName.
	allJoins do: [:each | 
		each key printForANSIJoinTo: each value on: aCommand.
		aCommand platform parenthesizeCommandsInCompoundQuery ifTrue: [aCommand nextPut: $)]].
	self printLeftoverTablesOn: aCommand.! !

!JoinPrinter methodsFor: 'printing' stamp: ''!
printLeftoverTablesOn: aCommand
	"Now there might be leftover tables whose joins were implied directly by the where clause"
	| leftOverTables |
	leftOverTables := self allTables asSet copy.
	availableTables do: [:each | leftOverTables remove: each ifAbsent: []].
	leftOverTables isEmpty ifFalse: [aCommand nextPutAll: ', '].
	GlorpHelper
		print: [:each | each sqlTableName]
		on: aCommand
		for: leftOverTables
		separatedBy: ', '.! !

!JoinPrinter methodsFor: 'accessing' stamp: ''!
query: aQuery
	query := aQuery.	
	self setUp.! !

!JoinPrinter methodsFor: 'accessing' stamp: ''!
rootTable
	"Pick a table to start with"
	^self allTables first.! !

!JoinPrinter methodsFor: 'initializing' stamp: ''!
setUp
	availableTables := Set new.! !

Object subclass: #Login
	instanceVariableNames: 'encodingStrategy database username password connectString name schema databaseName host port'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Database'!

!Login commentStamp: '<historical>' prior: 0!
Enough information to log in to a particular database.

Instance Variables:
	connectString	<String>	The string that will be passed to the driver level to tell us what to connect to. This might be e.g. an ODBC data source name, or it might be a host name plus socket/database name for Postgresql.
	database	<DatabasePlatform>	The platform describing what type of database this is. Used here mostly to indicate how we need to talk to the underlying connection layer, which may be different by database.
	name	<String>	A name we can use to describe this. No significance to the database.
	password	<String>	The connect password.
	schema	<String>	What should our default schema be. Used mostly internally, to set the schema of the tables we are using.
	username	<String>	The connect username.

!

!Login classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize.! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:12'!
= aLogin 
	^self class == aLogin class and: 
			[self name = aLogin name and: 
					[self database class = aLogin database class and: 
							[self username = aLogin username and: 
									[self password = aLogin password 
										and: [self connectString = aLogin connectString]]]]]! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:12'!
connectString
	^connectString! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:12'!
connectString: aString 
	connectString := aString! !

!Login methodsFor: 'accessing' stamp: ''!
database
	^database! !

!Login methodsFor: 'accessing' stamp: ''!
database: aDatabasePlatform 
	database := aDatabasePlatform! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:13'!
databaseName
	"connect string looks like hostname:portNumber_db-name.
	:portNumber is optional"
	
	| str portIndex index dbIndex |
	
	databaseName == nil
		ifFalse: [^ databaseName].
	databaseName == nil
		ifTrue: [str := self connectString.
			portIndex := index := str indexOf: $:.
			dbIndex := str indexOf: $_.
			index = 0
				ifTrue: [index := dbIndex].
			databaseName := str copyFrom: dbIndex + 1 to: str size.
			^ databaseName]! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:13'!
encodingStrategy
	^encodingStrategy! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:13'!
encodingStrategy: anEncodingStrategy
	encodingStrategy := anEncodingStrategy.! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:13'!
hash
	^self name hash + self database class hash + self username hash 
		+ self password hash + self connectString hash! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:13'!
host
	"connect string looks like hostname:portNumber_db-name.
	:portNumber is optional"
	| str portIndex index dbIndex |
	host == nil
		ifFalse: [^ host].
	host == nil
		ifTrue: [str := self connectString.
			portIndex := index := str indexOf: $:.
			dbIndex := str indexOf: $_.
			index = 0
				ifTrue: [index := dbIndex].
			host := str copyFrom: 1 to: index - 1.
			^ host]! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:13'!
host: aHost
	host := aHost.! !

!Login methodsFor: 'initialize-release' stamp: ''!
initialize
	secure := true.! !

!Login methodsFor: 'accessing' stamp: ''!
name
	name isNil ifTrue: [^self connectString] ifFalse: [^name].! !

!Login methodsFor: 'accessing' stamp: ''!
name: aString
	name := aString.! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:14'!
password
	^password! !

!Login methodsFor: 'accessing' stamp: ''!
password: aString 
	password := aString! !

!Login methodsFor: 'accessing' stamp: ''!
platform
	^self database.! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:14'!
port
	"connect string looks like hostname:portNumber_db-name.
	:portNumber is optional"
	| str portIndex index dbIndex |
	port == nil
		ifFalse: [^ port].
	port == nil
		ifTrue: [str := self connectString.
			portIndex := index := str indexOf: $:.
			dbIndex := str indexOf: $_.
			index = 0
				ifTrue: [index := dbIndex].
			port := portIndex = 0
						ifTrue: [self platform defaultPort asString]
						ifFalse: [(str copyFrom: portIndex + 1 to: dbIndex - 1) asString].
			^ port]! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:14'!
port: aPort
	port := aPort.! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:14'!
printOn: aStream

	aStream nextPutAll: 'a Login('.
	database printOn: aStream.
	aStream nextPutAll: ', '.
	username printOn: aStream.
	aStream nextPutAll: ', '.
	password printOn: aStream.
	aStream nextPutAll: ', '.
	connectString printOn: aStream.
	aStream nextPutAll: ')'.! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:15'!
schema
	^schema! !

!Login methodsFor: 'accessing' stamp: ''!
schema: aString
	schema := aString.! !

!Login methodsFor: 'accessing' stamp: 'GuillermoPolito 4/17/2011 18:15'!
sessionForDescriptorSystem: aDescriptorSystemClass
	"Create a session for this login, using aDescriptorSystemClass for its metadata"
	^aDescriptorSystemClass sessionForLogin: self.! !

!Login methodsFor: 'accessing' stamp: ''!
username
	^username! !

!Login methodsFor: 'accessing' stamp: ''!
username: aString 
	username := aString! !

Object subclass: #Mapping
	instanceVariableNames: 'descriptor attribute debugRead debugWrite'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!Mapping commentStamp: '<historical>' prior: 0!
This is the abstract superclass for all Glorp mappings. A mapping defines how one particular instance variable will be mapped into and out of the database. There is considerable variation among the subclasses.

Subclasses must implement the following messages:
	accessing
		allTables
	mapping
		mapFromObject:intoRowsIn:
		mapObject:inElementBuilder:
		referencedIndependentObjectsFrom:
		trace:context:
	testing
		controlsTables
		isRelationship
		isStoredInSameTable
		mappedFields

Instance Variables:
	attribute	<GlorpAttributeModel>	The description of the instance variable we map.
	canRead	<Boolean>	Is this mapping allowed to read.
	canWrite	<Boolean>	Is this mapping allowed to write.
	debugRead	<Boolean>	For debugging purposes. If true, we will halt when this mapping is about to read..
	debugWrite	<Boolean>	For debugging purposes. If true, we will halt when this mapping is about to write.
	descriptor	<Descriptor>	The descriptor that contains this mapping.

!

Mapping subclass: #AdHocMapping
	instanceVariableNames: 'fromDbMappingBlock toDbMappingBlock mappedFields shouldProxy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!AdHocMapping commentStamp: '<historical>' prior: 0!
AdHocMapping is a configurable sort of mapping, done via two blocks. It's not easy to use, but gives you very direct access to the underlying mechanisms, so it should be possible to write mappings that are otherwise very complicated or impossible.

Instance Variables:
	fromDbMappingBlock	<BlockClosure>	The block we invoke when reading from the database. It's given arguments of the "row" we got from the database (note that this doesn't necessarily correspond to a single row, but is all the fields we selected in the query) , the elementBuilder associated with our containing object, and the "context", the expression from the query for what we're reading from the database. Typical usage is to use the context to translate a field into the appropriate aliased field from the query, and use the element builder to get the value of that field. e.g. elementBuilder valueOfField: (context translateField: someField) in: row
	mappedFields	<SequenceableCollection of: DatabaseField>	The list of fields we work with. We need to specify this so that Glorp knows that it needs to e.g. includes these fields in the list of those selected.
	shouldProxy	<Boolean>	Should we create a proxy for our target objects when we read an object containing this mapping.
	toDbMappingBlock	<BlockClosure>	The block we invoke when writing to the database. It takes as arguments the rowmap being created, the target object we are writing, and a dictionary of the rows we already have associated with this object. The block is expected to update the rowmap with the new information.

!

!AdHocMapping classSide methodsFor: 'instance creation' stamp: ''!
forAttribute: aSymbol fromDb: fromBlock toDb: toBlock mappingFields: aFieldCollection 
	^super new 
		setAttribute: aSymbol
		fromDb: fromBlock
		toDb: toBlock
		mappingFields: aFieldCollection! !

!AdHocMapping methodsFor: 'printing SQL' stamp: ''!
allRelationsFor: rootExpression
	"We may have multiple relationships."
	| tables rows result rightChild rightHandSide |
	tables := self mappedTables.
	rows := Dictionary new.
	tables do: [:each | rows at: each put: (DatabaseRow newForTable: each)].
	rightChild := rootExpression rightChild.
	rightHandSide := rightChild isConstantExpression ifTrue: [rightChild value] ifFalse: [rightChild].
	toDbMappingBlock value: rows value: rightHandSide value: Dictionary new. 
	result := nil.
	rows do: [:eachRow |
		| table |
		table := rootExpression leftChild base getTable: eachRow table.
		eachRow fieldsAndValidValuesDo: [:eachField :eachValue |  | newExp |
			newExp := (table getField: eachField) get: rootExpression relation withArguments: (Array with: eachValue).
			result := newExp AND: result.]].
	^result.! !

!AdHocMapping methodsFor: 'testing' stamp: ''!
controlsTables
	^false! !

!AdHocMapping methodsFor: 'printing SQL' stamp: ''!
convertedDbValueOf: anObject
	| tables rows |
	tables := self mappedTables.
	rows := Dictionary new.
	tables do: [:each | rows at: each put: (DatabaseRow newForTable: each)].
	toDbMappingBlock value: rows value: anObject.
	rows keysAndValuesDo: [:eachTable :eachRow |
		^eachRow at: mappedFields first].! !

!AdHocMapping methodsFor: 'mapping' stamp: ''!
field
	| fields |
	fields := self mappedFields.
	fields size = 1 ifFalse: [self error: 'This mapping does not correspond to exactly one field'].
	^fields first.! !

!AdHocMapping methodsFor: 'initialize/release' stamp: ''!
forAttribute: aSymbol fromDb: fromBlock toDb: toBlock mappingFields: aFieldCollection
	self
		setAttribute: aSymbol
		fromDb: fromBlock
		toDb: toBlock
		mappingFields: aFieldCollection.! !

!AdHocMapping methodsFor: 'testing' stamp: ''!
hasImpliedClauses
	"We may imply more than one clause, or a clause which is different from the one directly implied by the relationship"
	^true! !

!AdHocMapping methodsFor: 'initialize/release' stamp: ''!
initialize

	super initialize.
	self shouldProxy: false.! !

!AdHocMapping methodsFor: 'testing' stamp: ''!
isRelationship
	"True when the mapping associates different persistent classes."

	^false! !

!AdHocMapping methodsFor: 'testing' stamp: ''!
isStoredInSameTable
	^true.! !

!AdHocMapping methodsFor: 'public' stamp: ''!
mapFromObject: anObject intoRowsIn: aRowMap

	| target sourceRows targetRows valueDescriptor |
	self canWrite ifFalse: [^self].
	target := self getValueFrom: anObject.
	target := self session realObjectFor: target.
	target := aRowMap mementoFor: target.
	sourceRows := Dictionary new: 3.
	targetRows := Dictionary new: 3.
	descriptor tables do: [:each |
		sourceRows at: each put: (aRowMap findOrAddRowForTable: each withKey: anObject)].
	valueDescriptor := self system descriptorFor: target.
	valueDescriptor isNil ifFalse: [
		valueDescriptor tables do: [:each |
			targetRows at: each put: (aRowMap findOrAddRowForTable: each withKey: target)]].
	toDbMappingBlock value: sourceRows value: target value: targetRows.! !

!AdHocMapping methodsFor: 'public' stamp: ''!
mapObject: anObject inElementBuilder: anElementBuilder

	| readBlock row |
	row := anElementBuilder row.
	readBlock := [fromDbMappingBlock value: row value: anElementBuilder value: BaseExpression new].
	self shouldProxy 
		ifTrue: [self setValueIn: anObject to: (AdHocProxy readBlock: readBlock)]
		ifFalse: [self setValueIn: anObject to: readBlock value]! !

!AdHocMapping methodsFor: 'mapping' stamp: ''!
mappedFields

	^mappedFields.! !

!AdHocMapping methodsFor: 'mapping' stamp: ''!
referencedIndependentObjectsFrom: anObject
	"Assumes that the only object this might affect is our primary attribute. That's probably valid. I think."
	| object otherDescriptor |
	object := self getValueFrom: anObject.
	(object isGlorpProxy and: [object isInstantiated not]) ifTrue: [^#()].
	otherDescriptor := self system descriptorFor: object.
	^otherDescriptor isNil ifTrue: [#()] ifFalse: [Array with: object].! !

!AdHocMapping methodsFor: 'initialize/release' stamp: ''!
setAttribute: aSymbol fromDb: fromBlock toDb: toBlock mappingFields: aFieldCollection
 
	self attributeName: aSymbol.
	fromDbMappingBlock := fromBlock.
	toDbMappingBlock := toBlock.
	mappedFields := aFieldCollection.! !

!AdHocMapping methodsFor: 'accessing' stamp: ''!
shouldProxy

	^shouldProxy.! !

!AdHocMapping methodsFor: 'accessing' stamp: ''!
shouldProxy: aBoolean

	shouldProxy := aBoolean.! !

!AdHocMapping methodsFor: 'mapping' stamp: ''!
trace: aTracing context: anExpression
	^self.! !

!AdHocMapping methodsFor: 'mapping' stamp: ''!
valueIn: anElementBuilder as: anExpression
	^fromDbMappingBlock
		value: anElementBuilder row
		value: anElementBuilder
		value: anExpression.! !

Mapping subclass: #ConditionalMapping
	instanceVariableNames: 'conditionalField conditionalMethod cases otherwiseCase conditionalFieldMapping'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!ConditionalMapping commentStamp: '<historical>' prior: 0!
This is a mapping that does different things depending on some testable condition. Depending on whether we're reading or writing, we are likely to test different things. This is useful if a field can be interpreted in more than one way. The API is rather messy.

Instance Variables:
	cases	<SequenceableCollection of: (Association)>	Used on read - a list of associations representing the cases we test. The key is either a one-argument block to be evaluated or a one-argument selector to be performed by the descriptorSystem, returning a boolean, and the value is the mapping to use if it returns true. 
	conditionalField	<DatabaseField>	The field whose value will be used as the argument to the blocks/selectors in cases.
	otherwiseCase	<Mapping>	If none of the other cases evaluate to true on read, use this mapping.

	conditionalMethod	<Block | Symbol>	Used on write - a zero-argument selector or one-argument block. The selector will be sent to the object being written, the block will be evaluated with that object as the argument. This is expected to return the object that will be written via the appropriate mapping.
	conditionalFieldMapping	<Mapping>	Used on write - the mapping used to write out the value in the conditional field.

!

!ConditionalMapping methodsFor: 'mapping' stamp: ''!
applicableMappingForConditionalValue: rowValue
	cases
		do:
			[:each | 
			((Dialect isBlock: each key)
				ifTrue: [each key value: rowValue]
				ifFalse: [self descriptor system perform: each key with: rowValue])
				ifTrue: [^each value]].
	^otherwiseCase.! !

!ConditionalMapping methodsFor: 'mapping' stamp: ''!
applicableMappingForObject: anObject
	| conditionalValue |
	conditionalValue := self conditionalValueFor: anObject.
	^self applicableMappingForConditionalValue: conditionalValue.! !

!ConditionalMapping methodsFor: 'mapping' stamp: ''!
applicableMappingForRow: anArray in: anElementBuilder
	"If a conditional field has been specified, then pass in the extracted value. Otherwise pass in the whole row"
	| rowValue |
	rowValue := conditionalField isNil
		ifFalse: [anElementBuilder valueOfField: conditionalField in: anArray]
		ifTrue: [anArray].
	^self applicableMappingForConditionalValue: rowValue.! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
attributeName
	^self representativeMapping attributeName.! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
cases
	^cases.! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
casesDo: aBlock
	"Iterate over all of our cases"
	cases do: [:each | aBlock value: each value].
	aBlock value: otherwiseCase.! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
conditionalField: aField

	(conditionalField notNil and: [conditionalField ~~ aField]) ifTrue: [self error: 'You can only test one field in a conditional mapping'].
	conditionalField := aField.! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
conditionalFieldMapping
	^conditionalFieldMapping! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
conditionalFieldMapping: aMapping
	"This is a write-only mapping for the conditional field value, which writes out the result of performing the conditional method"
	conditionalFieldMapping := aMapping.
	descriptor isNil ifFalse: [aMapping descriptor: descriptor].
	(descriptor mappings includes: aMapping) ifTrue: [self error: 'Mapping is part of a conditional and part of the main descriptor. Be sure to send newMapping: to the conditional, not to the descriptor'].! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
conditionalMethod: aSymbol
	(conditionalMethod notNil and: [conditionalMethod ~~ aSymbol]) ifTrue: [self error: 'You can only test one attribute in a conditional mapping'].
	conditionalMethod := aSymbol.! !

!ConditionalMapping methodsFor: 'mapping' stamp: ''!
conditionalValueFor: anObject
	^(Dialect isBlock: conditionalMethod)
		ifTrue: [conditionalMethod value: anObject]
		ifFalse: [anObject perform: conditionalMethod].! !

!ConditionalMapping methodsFor: 'testing' stamp: ''!
controlsTables
	^self cases first value controlsTables.! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
descriptor: aDescriptor

	super descriptor: aDescriptor.
	cases do: [:each | each value descriptor: aDescriptor].
	otherwiseCase isNil ifFalse: [otherwiseCase descriptor: aDescriptor].
	conditionalFieldMapping isNil ifFalse: [conditionalFieldMapping descriptor: aDescriptor].! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
field
	^self mappedFields asArray first.! !

!ConditionalMapping methodsFor: 'conditions' stamp: ''!
forField: aField attribute: attributeBlock if: testBlock useMapping: aMapping
	self conditionalField: aField.
	self conditionalMethod: attributeBlock.
	self if: testBlock then: aMapping.! !

!ConditionalMapping methodsFor: 'testing' stamp: ''!
hasField
	^self mappedFields size = 1.! !

!ConditionalMapping methodsFor: 'conditions' stamp: ''!
if: conditionSelector then: aMapping

	cases add: (Association key: conditionSelector value: aMapping).
	descriptor isNil ifFalse: [aMapping descriptor: descriptor].
	(descriptor mappings includes: aMapping) ifTrue: [self error: 'Mapping is part of a conditional and part of the main descriptor. Be sure to send newMapping: to the conditional, not to the descriptor'].! !

!ConditionalMapping methodsFor: 'initialize/release' stamp: ''!
initialize
	
	super initialize.
	cases := OrderedCollection new.! !

!ConditionalMapping methodsFor: 'testing' stamp: ''!
isRelationship

	^false! !

!ConditionalMapping methodsFor: 'preparing' stamp: ''!
joinExpressionFor: anExpression

	^self representativeMapping joinExpressionFor: anExpression.! !

!ConditionalMapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject intoRowsIn: aRowMap 
	self canWrite ifFalse: [^self].
	(self applicableMappingForObject: anObject) 
		mapFromObject: anObject
		intoRowsIn: aRowMap.
	conditionalFieldMapping isNil ifTrue: [^self].
	conditionalFieldMapping 
		mapFromObject: (self conditionalValueFor: anObject)
		intoRowsIn: aRowMap! !

!ConditionalMapping methodsFor: 'mapping' stamp: ''!
mapObject: anObject inElementBuilder: anElementBuilder
	(self applicableMappingForRow: anElementBuilder row in: anElementBuilder)
		mapObject: anObject
		inElementBuilder: anElementBuilder.! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
mappedFields

	| all |
	all := Set new.
	conditionalFieldMapping isNil 
		ifTrue: [all add: conditionalField]
		ifFalse: [all addAll: conditionalFieldMapping mappedFields].
	cases do: [:each |
		all addAll: each value mappedFields].
	^all.! !

!ConditionalMapping methodsFor: 'mapping' stamp: ''!
newMapping: aMappingClass
	"Implement this so that we can create our sub-mappings without having them added to the parent descriptor as independent mappings"
	| mapping |
	mapping := aMappingClass new.
	mapping descriptor: self descriptor.
	descriptor resetMappedFields.
	^mapping.! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
otherwise
	^otherwiseCase.! !

!ConditionalMapping methodsFor: 'conditions' stamp: ''!
otherwise: aMapping

	otherwiseCase := aMapping.
	descriptor isNil ifFalse: [otherwiseCase descriptor: descriptor].
	(descriptor mappings includes: aMapping) ifTrue: [self error: 'Mapping is part of a conditional and part of the main descriptor. Be sure to send newMapping: to the conditional, not to the descriptor'].! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
referenceClass
	^self representativeMapping referenceClass.! !

!ConditionalMapping methodsFor: 'mapping' stamp: ''!
referencedIndependentObjectsFrom: anObject

	| allReferencedObjects |
	allReferencedObjects := OrderedCollection new.
	cases do: [:each | 
		allReferencedObjects addAll: (each value referencedIndependentObjectsFrom: anObject)].
	^allReferencedObjects.! !

!ConditionalMapping methodsFor: 'accessing' stamp: ''!
representativeMapping
	"At some points we may need to have this mapping behave some way without knowing yet which of its cases is satisfied. We assume that there is some basic level of similarity so that we can use one mapping as a representative to discover, e.g. what the reference class is. This works all right for cases where we have a mapping or a constant. It wouldn't work very well if we have a one to one mapping to one class in case (a) but a many-many to an entirely different class in the other"
	^cases first value.! !

!ConditionalMapping methodsFor: 'conditions' stamp: ''!
trace: aTracing context: anExpression
	"To make a join, we need to look at all of our possible cases"

	conditionalFieldMapping isNil ifFalse: [
		conditionalFieldMapping trace: aTracing context: anExpression].
	cases do: [:each |
		each value trace: aTracing context: anExpression].! !

!ConditionalMapping methodsFor: 'validation' stamp: ''!
validateAttribute
	cases do: [:each | each value validateAttribute].
	otherwiseCase validateAttribute.
	conditionalFieldMapping isNil ifFalse: [conditionalFieldMapping validateAttribute].! !

Mapping subclass: #ConstantMapping
	instanceVariableNames: 'constantValue valueIsSession'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!ConstantMapping commentStamp: '<historical>' prior: 0!
Sometimes you just want a constant value to be set, either in the row, the object or both. And sometimes you just want a non-mapping (e.g. with a ConditionalMapping where one
of the conditions means "this isn't mapped"). This mapping represents these situations.
It also handles the special case where it's useful to have access to the session inside a
domain object, by allowing you to map it to an instance variable.

So far only the case of mapping to an inst var is implemented.

Instance Variables:
!

!ConstantMapping methodsFor: 'accessing' stamp: ''!
constantValue
	^constantValue.! !

!ConstantMapping methodsFor: 'accessing' stamp: ''!
constantValue: anObject

	constantValue := anObject.! !

!ConstantMapping methodsFor: 'accessing' stamp: ''!
constantValueIn: aSession

	^valueIsSession
		ifTrue: [aSession]
		ifFalse: [constantValue].! !

!ConstantMapping methodsFor: 'accessing' stamp: ''!
constantValueIsSession

	valueIsSession := true.! !

!ConstantMapping methodsFor: 'testing' stamp: ''!
controlsTables
	"Return true if this type of method 'owns' the tables it's associated with, and expression nodes using this mapping should alias those tables where necessary"

	^false! !

!ConstantMapping methodsFor: 'api' stamp: ''!
getValueFrom: anObject

	^constantValue.! !

!ConstantMapping methodsFor: 'initialize/release' stamp: ''!
initialize

	super initialize.
	valueIsSession := false.! !

!ConstantMapping methodsFor: 'testing' stamp: ''!
isRelationship
	^false! !

!ConstantMapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject intoRowsIn: aRowMap! !

!ConstantMapping methodsFor: 'mapping' stamp: ''!
mapObject: anObject inElementBuilder: anElementBuilder
	| value |
	value := anElementBuilder isNil
		ifTrue: [constantValue]
		ifFalse: [self constantValueIn: anElementBuilder session].
	self setValueIn: anObject to: value.! !

!ConstantMapping methodsFor: 'accessing' stamp: ''!
mappedFields
	"Return a collection of fields that this mapping will write into any of the containing object's rows"

	^#().! !

!ConstantMapping methodsFor: 'mapping' stamp: ''!
referencedIndependentObjectsFrom: anObject

	^#().! !

!ConstantMapping methodsFor: 'mapping' stamp: ''!
trace: aTracing context: anExpression

	^self.! !

!ConstantMapping methodsFor: 'validation' stamp: ''!
validateAttribute
	^self.! !

Mapping subclass: #DirectMapping
	instanceVariableNames: 'field converter shouldProxy query'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!DirectMapping commentStamp: '<historical>' prior: 0!
This is the simplest type of mapping, it maps directly between an instance variable that holds some "primitive" type, and a field in a database table.

Instance Variables:
	converter	<DatabaseConverter>	The converter used to transform the data into database form and back again. This will be automatically computed based on the types defined, if possible, but for more complex conversions (e.g. booleans into strings in the database) a converter can be explicitly specified.
	field	<DatabaseField | FunctionExpression>	The field to write to. One complication is that (for reading purposes) the field may be a function. So we might have a mapping that reads from the TO_UPPER of some field, or even an aggregate function like MAX.

!

!DirectMapping classSide methodsFor: 'instance creation' stamp: ''!
from: attributeName to: field

	^self from: attributeName
		type: field impliedSmalltalkType
		to: field! !

!DirectMapping classSide methodsFor: 'instance creation' stamp: ''!
from: attributeName type: aClass to: field

	^self new
		attributeName: attributeName;
		field: field;
		type: aClass.! !

!DirectMapping classSide methodsFor: 'instance creation' stamp: ''!
fromPseudoVariable: attributeName to: field

	^self fromPseudoVariable: attributeName
		type: field impliedSmalltalkType
		to: field! !

!DirectMapping classSide methodsFor: 'instance creation' stamp: ''!
fromPseudoVariable: attributeName type: aClass to: field

	^self new
		attributeName: attributeName;
		field: field;
		beForPseudoVariable.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
buildQuery
	"We just want to read our main object back again. So we create a join from our primary key fields to themselves. One side will remain field references, the other side will turn into parameters when it's converted to a where clause."
	| whereClause |
	whereClause := Join new.
	descriptor primaryTable primaryKeyFields do: [:each |
		whereClause addSource: each target: each].
	query := Query returningOneOf: descriptor describedClass where: whereClause.
	query retrieve: self attribute name.
	query session: self session.
	^query.! !

!DirectMapping methodsFor: 'testing' stamp: ''!
controlsTables
	"Return true if this type of method 'owns' the tables it's associated with, and expression nodes using this mapping should alias those tables where necessary"

	^false! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
convertValueToDatabaseForm: aValue
	
	converter isNil ifTrue: [^aValue].
	^converter convertedDbValueOf: aValue.! !

!DirectMapping methodsFor: 'private' stamp: ''!
convertedDbValueOf: stValue 
	^converter isNil
		ifTrue: [stValue]
		ifFalse: [converter convert: stValue toDatabaseRepresentationAs: self field type]! !

!DirectMapping methodsFor: 'private' stamp: ''!
convertedStValueOf: dbValue 
	^(converter isNil)
		ifTrue: [dbValue]
		ifFalse: [converter convert: dbValue fromDatabaseRepresentationAs: self field type]! !

!DirectMapping methodsFor: 'accessing' stamp: ''!
converter
	^converter.! !

!DirectMapping methodsFor: 'accessing' stamp: ''!
converter: aDatabaseConverter
	converter := aDatabaseConverter.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
expressionFor: anObject basedOn: anExpression relation: aSymbol
	"Return our expression using the object's values. e.g. if this was a direct mapping from id->ID and the object had id: 3, then return TABLE.ID=3"

	| value |
	value := anObject isNil
		ifTrue: [nil]
		ifFalse:
			[anObject isGlorpExpression
				ifTrue: [anObject getMapping: self named: self attributeName]
				ifFalse: [anObject glorpIsCollection
					ifTrue: [anObject collect: [:each | attribute getValueFrom: each]]
					ifFalse: [attribute getValueFrom: anObject]]].
	^(anExpression get: self attribute name) get: aSymbol withArguments: (Array with: value).! !

!DirectMapping methodsFor: 'accessing' stamp: ''!
field
	^field! !

!DirectMapping methodsFor: 'accessing' stamp: ''!
field: aDatabaseFieldOrExpression
	field := (Dialect isBlock: aDatabaseFieldOrExpression)
		ifTrue: [
			self beReadOnly.
			aDatabaseFieldOrExpression asGlorpExpressionOn: (BaseExpression new descriptor: descriptor)]
		ifFalse: [aDatabaseFieldOrExpression].! !

!DirectMapping methodsFor: 'accessing' stamp: ''!
fieldsForDescriptorSelectStatement
	"Return a collection of fields that this mapping will read from a row when the containing object is read, as opposed to when we're specifically told to use this field. Normally it amounts to the same thing, but if a direct mapping is proxied, we don't want its field when the parent is read"
	^self shouldProxy ifTrue: [self descriptor primaryTable primaryKeyFields] ifFalse: [super fieldsForDescriptorSelectStatement].! !

!DirectMapping methodsFor: 'initialize-release' stamp: ''!
from: anAttributeName to: aField

	| type |
	self attributeName: anAttributeName.
	self field: (descriptor fieldFor: aField).
	type := self attribute type isNil ifTrue: [field impliedSmalltalkType] ifFalse: [self attribute type].
	self type: type.! !

!DirectMapping methodsFor: 'initialize-release' stamp: ''!
from: anAttributeName type: aClass to: aField

	self
		attributeName: anAttributeName;
		field: aField;
		type: aClass.! !

!DirectMapping methodsFor: 'initialize-release' stamp: ''!
fromPseudoVariable: anAttributeName to: aField

	^self fromPseudoVariable: anAttributeName
		type: aField impliedSmalltalkType
		to: aField! !

!DirectMapping methodsFor: 'initialize-release' stamp: ''!
fromPseudoVariable: anAttributeName type: aClass to: aField

	self from: anAttributeName type: aClass to: aField.
	self beForPseudoVariable.! !

!DirectMapping methodsFor: 'initialize-release' stamp: ''!
initialize
	super initialize.
	shouldProxy := false.! !

!DirectMapping methodsFor: 'testing' stamp: ''!
isRelationship
	"True when the mapping associates different persistent classes."

	^false! !

!DirectMapping methodsFor: 'testing' stamp: ''!
isStoredInSameTable

	^true! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject intoRowsIn: aRowMap

	^self mapFromObject: anObject toTarget: anObject puttingRowsIn: aRowMap.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject toTarget: target puttingRowsIn: aRowMap

	| dbValue value row |
	self canWrite ifFalse: [^self].
	value := self getValueFrom: anObject.
	value := self session realObjectFor: value ifNone: [^self].
	dbValue := self convertedDbValueOf: value.
	row := aRowMap findOrAddRowForTable: self field table withKey: target.
	row at: field put: dbValue.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
mapObject: anObject inElementBuilder: anElementBuilder
	| value |
	self shouldProxy ifTrue: [
		^self setValueIn: anObject to: ((self newProxyIn: anElementBuilder)
						session: descriptor session;
						parameters: (self parametersIn: anElementBuilder);
						glorpOwner: anObject)].
	value := self valueInBuilder: anElementBuilder.
	self setValueIn: anObject to: value.! !

!DirectMapping methodsFor: 'testing' stamp: ''!
mappedFields
	"Return a collection of fields that this mapping will write into any of the containing object's rows"

	^Array with: self field! !

!DirectMapping methodsFor: 'testing' stamp: ''!
mapsSimpleObject
	^true.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
parametersIn: anElementBuilder
	| parameters |
	parameters := IdentityDictionary new.
	descriptor primaryTable primaryKeyFields do: [:each | parameters at: each put: (anElementBuilder valueOfField: each)].
	^parameters.! !

!DirectMapping methodsFor: 'printing SQL' stamp: ''!
printSQLOn: aStream withParameters: aDictionary
	self field printSQLOn: aStream withParameters:aDictionary.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
query

	query isNil ifTrue: [self buildQuery].
	^query.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
queryFor: anObjectBuilder
	"If we have a specific query we should use in this case, rather than our general one, construct it. Right now that only applies if we're using filtered reads"
	^self query.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
readBackNewRowInformationFor: anObject fromRowsIn: aRowMap 
	| value row |
	field isGenerated ifFalse: [^self].
	self canWrite ifFalse: [^false].
	row := aRowMap findOrAddRowForTable: self field table withKey: anObject.
	value := self convertedStValueOf: (row at: field ifAbsent: [^self]).
	attribute setValueIn: anObject to: value! !

!DirectMapping methodsFor: 'enumerating' stamp: ''!
referencedIndependentObjectsFrom: anObject

	^#().! !

!DirectMapping methodsFor: 'initialize-release' stamp: ''!
setConverterBetween: aClass and: aDbField 
	aClass isNil ifTrue: [^self].
	converter := descriptor ifNotNil: 
					[descriptor system converterBetweenStType: aClass andField: aDbField]
				ifNil: [aDbField converterForStType: aClass]! !

!DirectMapping methodsFor: 'accessing' stamp: ''!
shouldProxy

	^shouldProxy.! !

!DirectMapping methodsFor: 'accessing' stamp: ''!
shouldProxy: aBoolean

	shouldProxy := aBoolean.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
trace: aTracing context: anExpression
	^self.! !

!DirectMapping methodsFor: 'accessing' stamp: ''!
type: aClass
	self attribute type: aClass.
	field isNil ifFalse: [
		self setConverterBetween: aClass and: field].! !

!DirectMapping methodsFor: 'validation' stamp: ''!
validate
	super validate.
	(field isGlorpExpression and: [self readOnly not]) ifTrue: [self error: 'Mappings to functions cannot be written'].! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
valueIn: anElementBuilder as: anExpression
	| dbValue valueToGet |
	valueToGet := anExpression canTranslateFields 
		ifTrue: [anExpression translateField: field] 
		ifFalse: [anExpression].
	dbValue := anElementBuilder
		valueOfField: valueToGet.
	^self convertedStValueOf: dbValue.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
valueInBuilder: anElementBuilder
	^self valueInBuilder: anElementBuilder as: field.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
valueInBuilder: anElementBuilder as: anExpression
	| dbValue |
	dbValue := anElementBuilder valueOfField: anExpression.
	^self convertedStValueOf: dbValue.! !

!DirectMapping methodsFor: 'mapping' stamp: ''!
valueOfField: aField fromObject: anObject
	field = aField ifFalse: [self error: 'Mapping doesn''t describe field'].
	^self convertedDbValueOf: (self getValueFrom: anObject).! !

!Mapping classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!Mapping methodsFor: 'printing SQL' stamp: ''!
allRelationsFor: rootExpression do: aBlock andBetweenDo: anotherBlock
	"Normal mappings just operate on a single expression"
	aBlock value: rootExpression leftChild value: rootExpression rightChild.! !

!Mapping methodsFor: 'accessing' stamp: ''!
allTables

	self subclassResponsibility.! !

!Mapping methodsFor: 'mapping' stamp: ''!
applicableMappingForObject: anObject 
	"For polymorphism with conditional mappings"
	^self! !

!Mapping methodsFor: 'accessing' stamp: ''!
attribute
	attribute isNil ifTrue: [attribute := GlorpAttributeModel new].
	^attribute.! !

!Mapping methodsFor: 'accessing' stamp: ''!
attributeAccessor
	^self attribute.! !

!Mapping methodsFor: 'accessing' stamp: ''!
attributeName
	^self attribute name.! !

!Mapping methodsFor: 'accessing' stamp: ''!
attributeName: aSymbol
	| oldAttribute symbolName |
	symbolName := aSymbol asSymbol.
	oldAttribute := attribute.
	attribute := self classModel attributeNamed: symbolName.
	oldAttribute isNil
		ifTrue: [self initializeAttribute]
		ifFalse: [attribute copyFrom: oldAttribute].! !

!Mapping methodsFor: 'api' stamp: ''!
beExclusive
	self error: 'Only relationship mappings can be exclusive'.! !

!Mapping methodsFor: 'accessing' stamp: ''!
beForPseudoVariable
	self attribute beForPseudoVariable.! !

!Mapping methodsFor: 'accessing' stamp: ''!
beReadOnly
	self canWrite: false.! !

!Mapping methodsFor: 'testing' stamp: ''!
canBeUsedForRetrieve
	"Return true if this is a valid argument for a retrieve: clause"
	self isRelationship ifFalse: [^true].
	^self isToManyRelationship not.! !

!Mapping methodsFor: 'accessing' stamp: ''!
canRead
	^self attribute canRead.! !

!Mapping methodsFor: 'accessing' stamp: ''!
canWrite
	^self attribute canWrite.! !

!Mapping methodsFor: 'accessing' stamp: ''!
canWrite: aBoolean
	self attribute canWrite: aBoolean.! !

!Mapping methodsFor: 'accessing' stamp: ''!
classModel
	"If we don't have one, just make a temporary one. We'll fix it later"
	descriptor isNil ifTrue: [^GlorpClassModel new].
	^self descriptor classModel.! !

!Mapping methodsFor: 'testing' stamp: ''!
controlsTables
	"Return true if this type of mapping 'owns' the tables it's associated with, and expression nodes using this mapping should alias those tables where necessary"

	self subclassResponsibility! !

!Mapping methodsFor: 'private' stamp: ''!
convertedDbValueOf: stValue 
	"Mappings in general don't know how to do type conversion. Subclasses that deal with database primitive types should override."
	^stValue! !

!Mapping methodsFor: 'initialize/release' stamp: ''!
copyAttributeAttributes
	"We have an attributemodel that was created before we actually had a descriptor. It's possible that it had properties set. Make sure we copy them. If necessary, initialize things that require it."
	| oldAttribute |
	oldAttribute := attribute.
	attribute := self classModel attributeNamed: oldAttribute name type: oldAttribute type collectionType: oldAttribute collectionType.
	attribute copyFrom: oldAttribute.
	self initializeAttribute.! !

!Mapping methodsFor: 'mapping' stamp: ''!
createDeleteRowFor: anObject in: aRowMap
	"Only applies to mappings that write into link tables"! !

!Mapping methodsFor: 'debugging' stamp: ''!
debugRead
	^debugRead! !

!Mapping methodsFor: 'debugging' stamp: ''!
debugRead: aBoolean
	debugRead := aBoolean! !

!Mapping methodsFor: 'debugging' stamp: ''!
debugWrite
	^debugWrite! !

!Mapping methodsFor: 'debugging' stamp: ''!
debugWrite: aBoolean
	debugWrite := aBoolean! !

!Mapping methodsFor: 'accessing' stamp: ''!
descriptor
	^descriptor! !

!Mapping methodsFor: 'accessing' stamp: ''!
descriptor: aDescriptor
	descriptor := aDescriptor.
	attribute isNil ifFalse: [self copyAttributeAttributes].! !

!Mapping methodsFor: 'mapping' stamp: ''!
expressionFor: anObject
	"Return an expression representing the value of the object. This can be nil, an object value or values, an expression, or a collection of expressions (for a composite key, if we're passed an expression)"

	anObject isNil ifTrue: [^#(nil)].
	anObject isGlorpExpression ifFalse: [
		^self mappedFields collect: [:each | 	
			self valueOfField: each fromObject: anObject]].
	^self mappedFields
		collect: [:each | (anObject getField: each)].! !

!Mapping methodsFor: 'mapping' stamp: ''!
expressionFor: anObject basedOn: anExpression relation: aSymbol
	"Return our expression using the object's values. e.g. if this was a direct mapping from id->ID and the object had id: 3, then return TABLE.ID=3. Used when rewriting object=object into field=field"

	| myValue result |
	myValue := self expressionFor: anObject.
	result := nil.
	myValue with: self join allTargetFields do: [:eachValue :eachField |
		| source |
		source := anExpression get: self attribute name.
		source hasDescriptor ifTrue: [source := source getField: eachField].
		result := (source get: aSymbol withArguments: (Array with: eachValue)) AND: result].
	^result.! !

!Mapping methodsFor: 'accessing' stamp: ''!
fieldsForDescriptorSelectStatement
	"Return a collection of fields that this mapping will read from a row when the containing object is read, as opposed to when we're specifically told to use this field. Normally it amounts to the same thing."
	^self mappedFields.! !

!Mapping methodsFor: 'accessing' stamp: ''!
fieldsForSelectStatement
	"Return a collection of fields that this mapping will read from a row"
	self canRead ifFalse: [^#()].
	^self mappedFields! !

!Mapping methodsFor: 'public' stamp: ''!
getValueFrom: anObject

	^self attribute getValueFrom: anObject.! !

!Mapping methodsFor: 'testing' stamp: ''!
hasAttribute
	"Return true if this maps an attribute, which is normally true"
	^true! !

!Mapping methodsFor: 'testing' stamp: ''!
hasImpliedClauses
	"Return true if this implies multiple sql clauses"
	^false.! !

!Mapping methodsFor: 'testing' stamp: ''!
hasValueThatCanProvideField: aField from: anObject
	"Return true if we have enough information in ourselves to populate our fields. The only time this won't be true is if we're either holding an uninstantiated proxy or have a to-many relationship mapping with an empty collection."
	^true.! !

!Mapping methodsFor: 'testing' stamp: ''!
includesSubFieldsInSelectStatement
	^false! !

!Mapping methodsFor: 'initialize/release' stamp: ''!
initialize
	debugRead := false.
	debugWrite := false.! !

!Mapping methodsFor: 'initialize/release' stamp: ''!
initializeAttribute
	"This is where we can initialize aspects of the attribute, if required. The motivating example is initializing the collection type of the attribute if it wasn't set. We need it to have some type, but we don't want to initialize it while we're doing basic initialization and aren't yet properly attached to a descriptor or class model"! !

!Mapping methodsFor: 'api' stamp: ''!
isExclusive
	^false.! !

!Mapping methodsFor: 'api' stamp: ''!
isExclusive: aBoolean
	aBoolean ifTrue: [self error: 'Only relationship mappings can be exclusive'].! !

!Mapping methodsFor: 'accessing' stamp: ''!
isForPseudoVariable
	^self attribute isForPseudoVariable or: [self attributeAccessor isValid not].! !

!Mapping methodsFor: 'testing' stamp: ''!
isRelationship
	"True when the mapping associates different persistent classes."

	^self subclassResponsibility! !

!Mapping methodsFor: 'testing' stamp: ''!
isStoredInSameTable
	"True when the mapping is between two objects that occupy the same table, e.g. an embedded mapping."

	^self subclassResponsibility! !

!Mapping methodsFor: 'testing' stamp: ''!
isToManyRelationship
	^false.! !

!Mapping methodsFor: 'testing' stamp: ''!
isTypeMapping
	^false! !

!Mapping methodsFor: 'preparing' stamp: ''!
joinExpressionFor: anExpression

	^nil.! !

!Mapping methodsFor: 'iterating' stamp: ''!
loopOverAttributeOf: anObject doing: aTwoArgumentBlock
	"Loop over the target, which we extract from anObject, and iterate over it evaluating the block. For things that aren't collections, iterating means just run once. Note that the argument is expected to be an instance of our descriptor, as opposed to the attribute method, which expects the target object"
	| target |
	target := self getValueFrom: anObject.
	self attribute loopOver: target doing: aTwoArgumentBlock.! !

!Mapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject intoRowsIn: aRowMap

	self subclassResponsibility.! !

!Mapping methodsFor: 'mapping' stamp: ''!
mapObject: anObject inElementBuilder: anObject1
	self subclassResponsibility.! !

!Mapping methodsFor: 'testing' stamp: ''!
mappedFields
	self subclassResponsibility! !

!Mapping methodsFor: 'mapping' stamp: ''!
mappedTables
	^(self mappedFields collect: [:each | each table]) asSet.! !

!Mapping methodsFor: 'testing' stamp: ''!
mapsPrimaryKeys
	^(self mappedFields detect: [:each | each isPrimaryKey] ifNone: [nil]) notNil.! !

!Mapping methodsFor: 'testing' stamp: ''!
mapsSimpleObject
	"Return true if we know that we map something without a descriptor. For some mappings, like ad hoc or conditional, we don't really know, so we default to false"
	^false.! !

!Mapping methodsFor: 'preparing' stamp: ''!
multipleTableExpressionsFor: anExpression

	^#().! !

!Mapping methodsFor: 'proxies' stamp: ''!
newProxyIn: anObjectBuilder 
	| proxy |
	proxy := Proxy new.
	proxy query: (self queryFor: anObjectBuilder).
	proxy glorpProxyAction: self proxyAction.
	^proxy! !

!Mapping methodsFor: 'public' stamp: ''!
printOn: aStream

	super printOn: aStream.
	aStream 
		nextPutAll: '(';
		nextPutAll: (attribute isNil ifTrue: [''] ifFalse: [self attributeName]) ;
		nextPutAll: ')'.! !

!Mapping methodsFor: 'proxies' stamp: ''!
proxyAction
	^nil.! !

!Mapping methodsFor: 'mapping' stamp: ''!
readBackNewRowInformationFor: anObject fromRowsIn: aRowMap 
	"
	self subclassResponsibility. ?"! !

!Mapping methodsFor: 'accessing' stamp: ''!
readOnly
	^self canWrite not.! !

!Mapping methodsFor: 'accessing' stamp: ''!
readOnly: aBoolean
	self canWrite: aBoolean not.! !

!Mapping methodsFor: 'accessing' stamp: ''!
referenceClass
	^self attribute type.! !

!Mapping methodsFor: 'mapping' stamp: ''!
referencedIndependentObjectsFrom: anObject

	self subclassResponsibility.! !

!Mapping methodsFor: 'accessing' stamp: ''!
session
	
	^self descriptor session.! !

!Mapping methodsFor: 'public' stamp: ''!
setValueIn: anObject to: aValue
	self canRead ifFalse: [^self].
	self attribute setValueIn: anObject to: aValue.! !

!Mapping methodsFor: 'debugging' stamp: ''!
stopIfDebuggingRead
	debugRead ifTrue: [self halt].! !

!Mapping methodsFor: 'debugging' stamp: ''!
stopIfDebuggingWrite
	debugWrite ifTrue: [self halt].! !

!Mapping methodsFor: 'accessing' stamp: ''!
system
	
	^self descriptor system.! !

!Mapping methodsFor: 'mapping' stamp: ''!
trace: aTracing context: anExpression

	self subclassResponsibility.! !

!Mapping methodsFor: 'mapping' stamp: ''!
translateFields: anOrderedCollection 
	"Normal mappings don't translate"
	^anOrderedCollection.! !

!Mapping methodsFor: 'mapping' stamp: ''!
translateTable: aTable
	"If we have a field translation, return the translated (i.e. real) table to use. This ideally allows us to use either the real table in which something is embedded, or the imaginary table"
	^aTable.! !

!Mapping methodsFor: 'validation' stamp: ''!
validate
	self validateAttribute! !

!Mapping methodsFor: 'validation' stamp: ''!
validateAttribute
	self canWrite ifTrue: [self attribute validateCanWrite].
	self canRead ifTrue: [self attribute validateCanRead]! !

!Mapping methodsFor: 'mapping' stamp: ''!
valueOfField: aField fromObject: anObject
	self error: 'Can''t get value of field'.! !

Mapping subclass: #RelationshipMapping
	instanceVariableNames: 'query join usesLinkTable shouldProxy linkTableJoin targetTableJoin reverseJoin shouldUseFilteredRead rowMapKeyConstructorBlock controlsTables isExclusive linkFields proxyAction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!RelationshipMapping commentStamp: '<historical>' prior: 0!
This is an abstract superclass for relationships between one mapped object and another (or a collection of others).

Instance Variables:
	join	<Join>	The join between our table and the target. Note that if we use a link table, this will be a join between our table and the link table.
	targetTableJoin	<Join>	If we're using a link table, it may be desirable to also join to attributes of the main table. This optional extra join allows that.

	query	<SimpleQuery>	The query that will return the other side of the relationship, given parameters describing our side.
	linkFields	<Collection of: DatabaseField>	When there's a link table, we need to know the other side of the relationship, from link table to target. Usually we can figure it out based on the target table and the foreign key constraints, but sometimes (e.g. if there are multiple relationships to the same table) we need to be told which fields are the ones that connect to this table. Again that's FROM the LINK table TO the TARGET table, not the entry in the link table that we'd connect  to.
	reverseJoin	<Join>	caches the other end of the relationship, as above.
	rowMapKeyConstructorBlock	<BlockClosure>	Usually row map keys are constructed automatically, but it's possible that in weird cases we  might need to construct complicated ones (e.g. if there are 3 things that need to participate instead of just two, as in Store's use of the same link table and the same row to indicate the relationship from package to class and package to metaclass).
	shouldProxy	<Boolean>	Should we proxy or directly construct the referenced object
	shouldUseFilteredRead	<Boolean>	should we use a filtered read.
	usesLinkTable	<Boolean>	Do we use a link table, or is it a direct join to the target.
	isExclusive <Boolean> Is this an exclusive relationship - so that when the parent object is deleted, the child or children will also be deleted.

!

RelationshipMapping subclass: #OneToOneMapping
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!OneToOneMapping commentStamp: '<historical>' prior: 0!
This represents a simple one to one relationship, e.g. from Person to Address. Assuming each person has only one address. Note that this has no methods, and only really exists for backward compatibility purposes, since a to-one relationship is the default for RelationshipMapping.!

OneToOneMapping subclass: #EmbeddedValueOneToOneMapping
	instanceVariableNames: 'fieldTranslation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!EmbeddedValueOneToOneMapping commentStamp: '<historical>' prior: 0!
This represents a one-to-one mapping in which the referenced object is stored as part of the same table as the containing object.!

!EmbeddedValueOneToOneMapping methodsFor: 'testing' stamp: ''!
controlsTables
	"Return true if this type of method 'owns' the tables it's associated with, and expression nodes using this mapping should alias those tables where necessary"

	^false! !

!EmbeddedValueOneToOneMapping methodsFor: 'transformations' stamp: ''!
defaultTransformationExpressionFor: aDescriptor 
	"If there's no transformation, get all the mapped fields from the other descriptor and construct a transformation of each onto itself. This lets us unify the fields in my row with the fields in its row"

	| fields transform |
	fields := IdentitySet new.
	aDescriptor mappings do: [:each | fields addAll: each mappedFields].
	transform := Join new.
	fields do: [:each | transform addSource: each target: each].
	^transform! !

!EmbeddedValueOneToOneMapping methodsFor: 'accessing' stamp: ''!
fieldTranslation
	^fieldTranslation! !

!EmbeddedValueOneToOneMapping methodsFor: 'accessing' stamp: ''!
fieldTranslation: aPrimaryKeyExpression 
	fieldTranslation := aPrimaryKeyExpression! !

!EmbeddedValueOneToOneMapping methodsFor: 'internal' stamp: ''!
fieldsForSelectStatement
	"Return a collection of fields that this mapping will read from a row"
	"Return nothing, because our sub-objects will take care of adding their own fields, translated correctly through us."
	^#().! !

!EmbeddedValueOneToOneMapping methodsFor: 'accessing' stamp: ''!
hasFieldTranslation

	^fieldTranslation notNil.! !

!EmbeddedValueOneToOneMapping methodsFor: 'transformations' stamp: ''!
hasTransformation

	^false.! !

!EmbeddedValueOneToOneMapping methodsFor: 'testing' stamp: ''!
isStoredInSameTable
	^true! !

!EmbeddedValueOneToOneMapping methodsFor: 'accessing' stamp: ''!
join
	^nil.! !

!EmbeddedValueOneToOneMapping methodsFor: 'accessing' stamp: ''!
join: anExpression
	self error: 'Embedded mappings do not use joins'.! !

!EmbeddedValueOneToOneMapping methodsFor: 'preparing' stamp: ''!
joinExpressionFor: targetExpression
	"We're looking for the object represented by this mapping, and we know the object represented by its source. Use our mapping criteria to construct a join that traverses that instance of this relationship.
	Embedded values never induce a join."

	^nil.! !

!EmbeddedValueOneToOneMapping methodsFor: 'mapping' stamp: ''!
knitFromBuilder: sourceBuilder toBuilder: targetBuilder 
	sourceBuilder requiresPopulating ifFalse: [^self].
	super knitFromBuilder: sourceBuilder toBuilder: targetBuilder! !

!EmbeddedValueOneToOneMapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject toTarget: target puttingRowsIn: aRowMap 

	self transformationExpression 
		mapFromSource: anObject
		andTarget: target
		intoRowsIn: aRowMap.

	self transformationExpression allTables do: [:eachTable |	
		aRowMap rowForTable: eachTable withKey: target ifPresent: [  :each |
			each shouldBeWritten: false]].! !

!EmbeddedValueOneToOneMapping methodsFor: 'mapping' stamp: ''!
mapObject: anObject inElementBuilder: anElementBuilder
	| myTraceNode myBuilder |
	"If the object already has a value in my slot, then this it got a cache hit, the embedded value was carried along for the ride, and we don't need to assign anything"
	(self getValueFrom: anObject) isNil ifFalse: [^self].	"Otherwise, we need to look up the trace node that corresponds to this mapping, and get its instance"
	myTraceNode := anElementBuilder expression get: self attributeName.
	myBuilder := anElementBuilder query elementBuilderFor: myTraceNode.
	self setValueIn: anObject to: myBuilder instance.! !

!EmbeddedValueOneToOneMapping methodsFor: 'internal' stamp: ''!
mappedFields
	"Return a collection of fields that this mapping will write into any of the containing object's rows"

	fieldTranslation isNil ifFalse: [^fieldTranslation allSourceFields].
	^self referenceDescriptor mappedFields.! !

!EmbeddedValueOneToOneMapping methodsFor: 'mapping' stamp: ''!
reverseMapping
	^self class new
		attributeName: #temporaryReverseMappingForEmbeddedMapping;
		referenceClass: self descriptor describedClass;
		descriptor: self descriptor.! !

!EmbeddedValueOneToOneMapping methodsFor: 'testing' stamp: ''!
shouldProxy

	^false.! !

!EmbeddedValueOneToOneMapping methodsFor: 'transformations' stamp: ''!
transformationExpression

	^self hasFieldTranslation
		ifTrue: [fieldTranslation]
		ifFalse: [fieldTranslation := self defaultTransformationExpressionFor: (self referenceDescriptor)].! !

!EmbeddedValueOneToOneMapping methodsFor: 'mapping' stamp: ''!
translateFields: anOrderedCollection 
	fieldTranslation isNil ifTrue: [^anOrderedCollection].
	^anOrderedCollection collect: [:each |
		fieldTranslation sourceForTarget: each].! !

!EmbeddedValueOneToOneMapping methodsFor: 'mapping' stamp: ''!
translateTable: aTable
	"If we have a field translation, return the translated (i.e. real) table to use. This ideally allows us to use either the real table in which something is embedded, or the imaginary table"

	fieldTranslation isNil ifTrue: [^aTable].
	1 to: fieldTranslation allTargetFields size
		do:
			[:i | 
			(fieldTranslation allTargetFields at: i) table == aTable
				ifTrue: [^(fieldTranslation allSourceFields at: i) table]].
	^aTable.! !

!EmbeddedValueOneToOneMapping methodsFor: 'mapping' stamp: ''!
valueOfField: aField fromObject: anObject ifNone: aBlock
	| targetField target |
	targetField := self transformationExpression targetForSource: aField.
	targetField isNil ifTrue: [^aBlock value].
	target := self getValueFrom: anObject.
	^self referenceDescriptor valueOfField: targetField fromObject: target.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
addPhantomMappingExpressionOn: base 
	| phantomMapping phantomMappingName phantomExpression |
	phantomMapping := self reverseMapping.
	phantomMappingName := 'glorp_internal_' , self attributeName , '_to_' 
				, self descriptor describedClass name.
	phantomExpression := base getMapping: phantomMapping
				named: phantomMappingName.
	phantomExpression bePhantom.
	^phantomExpression! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
allTables

	(self join) isNil ifTrue: [^#()].
	^(self join) allTables.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
basicReferenceClass
	^self attribute type.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
basicReferenceDescriptor
	^self system descriptorFor: self basicReferenceClass.! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
beExclusive
	"Make this an exclusive relationship - so that when the parent object is deleted, the child or children will also be deleted"
	isExclusive := true.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
buildGeneralQueryIn: anObjectBuilder
	"For filtered queries. We need to build this in terms of the parent query"

	| base parentCriteria fullCriteria phantomExpression newQuery |
	base := BaseExpression new.
	base descriptor: self referenceDescriptor.
	phantomExpression := self addPhantomMappingExpressionOn: base.
	parentCriteria := anObjectBuilder query whereClause.
	fullCriteria := parentCriteria
		rebuildOn: phantomExpression.
	newQuery := Query returningManyOf: self referenceClass.
	"newQuery collectionType: self collectionType."
	fullCriteria isEmptyExpression ifFalse: [newQuery where: fullCriteria].
	self
		setGeneralQueryRetrievalIn: newQuery
		basedOn: base
		phantom: phantomExpression.
	^newQuery.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
buildQuery
	| toMany |
	toMany := self attribute isCollectionAttribute.
	query := toMany
		ifTrue: [Query read: self referenceClass where: self join]
		ifFalse: [Query readOneOf: self referenceClass where: self join].
	query session: self session.
	query baseExpression descriptor: (self session descriptorFor: self referenceClass).
	query collectionType: self collectionType.
	self usesLinkTable ifTrue: [
		query AND: (self expressionFromLinkToReferenceTableWithBase: query baseExpression)].
	^query.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
buildRowMapKeyForSource: source target: target index: index
	^(RowMapKey new)
		key1: target yourSelf;
		key2: source yourSelf.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
collectionType
	^nil.! !

!RelationshipMapping methodsFor: 'private/expressions' stamp: ''!
constraints
	"Return a list of the constraints from our link table. If the linkField is set, filter the constraints to those that include that field. "
	| referenceKeys linkTable referenceTables allConstraints relevantConstraints |
	referenceKeys := (self join) targetKeys asOrderedCollection.
	linkTable := referenceKeys first table.
	"If we haven't been told the relevant link table fields, assume we can find them by looking at all the ones that aren't the ones from our source to the link, and all the rest will be from the link to the target"
	allConstraints := linkTable foreignKeyConstraints.
	relevantConstraints := linkFields isNil 
		ifTrue: [allConstraints reject: [:each |  each sourceFields allSatisfy: [:eachSourceField |
			(referenceKeys includes: eachSourceField)]]]
		ifFalse: [allConstraints select: [:each | each sourceFields anySatisfy: [:eachSourceField |
			linkFields includes: eachSourceField]]].

	"Validate that we can handle this case"
	referenceTables := (relevantConstraints collect: [:each | each targetTable]) asSet.
	referenceTables size > 1 ifTrue: [self error: 'Cannot handle this general a case. Try hinting using relevantLinkTableFields:'].
	referenceTables size = 0 ifTrue: [self error: 'No tables found. Did you set up foreign key references in the table definitions?'].

	^relevantConstraints.! !

!RelationshipMapping methodsFor: 'initializing' stamp: ''!
constraintsFrom: oneTableSet to: anotherTableSet
	^(self oneWayConstraintsFrom: oneTableSet to: anotherTableSet), 
		(self oneWayConstraintsFrom: anotherTableSet to: oneTableSet)! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
constructRowMapKeyAs: aBlock
	"Give us the opportunity to construct a custom row map key. This is useful if you need to force two relationships to share a link table entry"
	rowMapKeyConstructorBlock := aBlock.! !

!RelationshipMapping methodsFor: 'testing' stamp: ''!
controlsTables
	"Return true if this type of method 'owns' the tables it's associated with, and expression nodes using this mapping should alias those tables where necessary"

	^controlsTables! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
controlsTables: aBoolean
	"We can indicate that this relationship does not actually control its tables. This is going to be true if the parent controls the same tables and coexists with us. Think StoreVersionlessPundle and Pundle"
	controlsTables := aBoolean.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
createDeleteRowFor: anObject in: aRowMap
	"This is like firing the mapping, but instead of creating the data, we just need to check if there are link table entries, and if so, make sure they are deleted"
	| target mementoizedTarget|
	self canWrite ifFalse: [^self].
	self usesLinkTable ifFalse: [^self].
	target := self getValueFrom: anObject.
	target := self session realObjectFor: target ifNone: [^self].
	(self isValidTarget: target) ifTrue: [
		mementoizedTarget := aRowMap collectionMementoFor: target.
		self createDeleteRowForLinkFrom: anObject toTarget: mementoizedTarget in: aRowMap].! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
createDeleteRowForLinkFrom: anObject toTarget: target in: aRowMap
	| linkTable |
	self usesLinkTable ifFalse: [^false].
	linkTable := self linkTableJoin allTables asArray first.
	self attribute
		loopOver: target
		doing:
			[:index :eachTarget | 
			| intermediate |
			intermediate := self
				rowMapKeyForSource: anObject
				target: eachTarget
				index: index.
			aRowMap
				rowForTable: linkTable
				withKey: intermediate
				ifPresent:
					[:linkRow | linkRow isEmpty ifFalse: [linkRow forDeletion: true]]].! !

!RelationshipMapping methodsFor: 'private/expressions' stamp: ''!
expressionFromLinkToReferenceTableWithBase: base

	| reverse |
	reverse := self reverseJoin asGlorpExpressionOn: base.
	^reverse asGeneralGlorpExpressionWithFields.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
extendedJoin
	"In order to do a many-to-many read we need more information than just the write, we need to know 
	the relationship to the other table. Construct that based on the table information"

	| generalMappingCriteria base |
	self usesLinkTable ifFalse: [^self join].
	generalMappingCriteria := self join asGeneralGlorpExpressionWithFields.
	base := generalMappingCriteria ultimateBaseExpression.
	^(self expressionFromLinkToReferenceTableWithBase: base) AND: generalMappingCriteria.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
extendedJoinFromSource: sourceExpression toTarget: targetExpression
	"In order to do a many-to-many read we need more information than just the write, we need to know 
	the relationship to the other table. Construct that based on the table information"

	| generalMappingCriteria |
	self usesLinkTable 
		ifFalse: [^self join asExpressionJoiningSource: sourceExpression toTarget: targetExpression].

	generalMappingCriteria := self join asExpressionJoiningSource: sourceExpression toTarget: targetExpression.
	^(self expressionFromLinkToReferenceTableWithBase: targetExpression) AND: generalMappingCriteria.! !

!RelationshipMapping methodsFor: 'testing' stamp: ''!
hasValueThatCanProvideField: aField from: anObject
	"Return true if we have enough information in ourselves to populate our fields. The only time this won't be true is if we're either holding an uninstantiated proxy or have a to-many relationship mapping with an empty collection. Assumes that aField is one that we map."
	| value |
	value := self getValueFrom: anObject.
	value isGlorpProxy ifFalse: [^value notNil].
	value isInstantiated ifTrue: [^true].
	^value parameters includesKey: aField.! !

!RelationshipMapping methodsFor: 'initializing' stamp: ''!
initialize

	super initialize.
	shouldProxy := true.
	shouldUseFilteredRead := false.
	usesLinkTable := false.
	controlsTables := true.
	isExclusive := false.! !

!RelationshipMapping methodsFor: 'initializing' stamp: ''!
initializeJoin
	"We haven't been given a join. There might be a compound one, or else see if we can make an educated guess at one"
	linkTableJoin notNil ifTrue: [join := linkTableJoin, targetTableJoin. ^self].
	self initializeJoinTo: self relatedTables.! !

!RelationshipMapping methodsFor: 'initializing' stamp: ''!
initializeJoinTo: otherTables
	"We haven't been giving a join. See if we can make an educated guess at one by looking for tables with constraints to us. If linkFields is set, it tells us the source field for the relationship between the link table and the target, so in looking for our own relationship to the link table, we exclude that."
	| toConstraints fromConstraints |
	self descriptor tables isEmpty ifTrue: [^self join: Join new].
	(self referenceDescriptor tables isEmpty and: [self referenceDescriptor typeResolver usesInheritance]) ifTrue: [^self join: Join new].
	toConstraints := self oneWayConstraintsFrom: self descriptor tables to: otherTables.
	linkFields isNil ifFalse: [toConstraints := toConstraints reject: [:eachConstraint | eachConstraint targetFields anySatisfy: [:eachField | linkFields includes: eachField]]].
	fromConstraints :=  self oneWayConstraintsFrom: otherTables to: self descriptor tables.
	linkFields isNil ifFalse: [fromConstraints := fromConstraints reject: [:eachConstraint | eachConstraint sourceFields anySatisfy: [:eachField | linkFields includes: eachField]]].

	self join: (self system joinFor: self toTables: otherTables fromConstraints: fromConstraints toConstraints: toConstraints).! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
isExclusive
	"Return whether this an exclusive relationship - so that when the parent object is deleted, the child or children will also be deleted"
	^isExclusive.! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
isExclusive: aBoolean
	"Set whether this an exclusive relationship - so that when the parent object is deleted, the child or children will also be deleted"
	isExclusive := aBoolean.! !

!RelationshipMapping methodsFor: 'testing' stamp: ''!
isRelationship
	"True when the mapping associates different persistent classes."

	^true! !

!RelationshipMapping methodsFor: 'testing' stamp: ''!
isStoredInSameTable
	"Normally, a relationship mapping isn't stored in the same table, but if we're part of a dictionary mapping and the table is mapped to an association, then we want to be treated that way. The test is a horrible hack"
	^self classModel describedClass == Association.! !

!RelationshipMapping methodsFor: 'testing' stamp: ''!
isToManyRelationship
	^self attribute isCollectionAttribute.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
isValidTarget: anObject

	^anObject isGlorpProxy
		ifTrue: [anObject isInstantiated]
		ifFalse: [true].! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
join
	join isNil ifTrue: [self initializeJoin].
	^join.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
join: aJoin
	"We allow the join to be specified either as a Join (the normal case) or as an expression. We expect an expression specifying the relationship in terms of other relationships. So, e.g. to specify the join from InsurancePlan to PolicyHolder we might specify
[:eachPolicyHolder | eachPolicyHolder employer availablePlans].  Note that the relationship is described from the point of view of the target, and is just a list of mappings. Then we take this internally and transform it into an equality expression we can use to define the relationship."
	| base basicExpression |
	(Dialect isBlock: aJoin) 
		ifTrue: [
			self beReadOnly.
			base := JoinBaseExpression new.
			base descriptor: (self system descriptorFor: self referenceClass).
			basicExpression := aJoin asGlorpExpressionOn: base.
			"It doesn't matter what name we give the parameter, because it will be rewritten into a primary key comparison anyway, that will use the fields as 'names'. Also, it's important that we assign into join here, because otherwise the rewriteEquality operation will try and figure out a join for us, since it's still nil."
			join := basicExpression get: #= withArguments: (
				Array with: (ParameterExpression forField: 'sourceObject' basedOn: basicExpression join)).
			join := join rewriteEquality]
		ifFalse: [join := aJoin].
	[join owningMapping: self] on: MessageNotUnderstood do: [:ex | ex | ex return].! !

!RelationshipMapping methodsFor: 'preparing' stamp: ''!
joinExpressionFor: targetExpression
	"We're looking for the object represented by this mapping, and we know the object represented by its source. Use our mapping criteria to construct a join that traverses that instance of this relationship"

	| sourceExpression |
	(self controlsTables not and: [self classModel describedClass == Association]) ifTrue: [^nil].
	sourceExpression := targetExpression base.
	^self extendedJoinFromSource: sourceExpression toTarget: targetExpression.! !

!RelationshipMapping methodsFor: 'private/expressions' stamp: ''!
joinFromLinkToReferenceTable
	"Generate the inverse mapping expression, i.e. the one connecting the link table to the reference table, using the foreign key constraints. We assume that if there's a general part, it applies only to the original join, and not to the reverse join that we're generating."

	| expression generalPart |
	expression := Join new.
	self constraints do: [:eachConstraint | 
		eachConstraint sourceFields with: eachConstraint targetFields do: 
			[:eachSource :eachTarget |
				expression addSource: eachSource target: eachTarget]].
	generalPart := self join generalExpressionPart.
	^generalPart isNil ifTrue: [expression] ifFalse: [generalPart AND: expression].! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
knitFromBuilder: sourceBuilder toBuilder: targetBuilder 
	self canRead ifFalse: [^self].
	self attribute knit: sourceBuilder instance to: targetBuilder instance! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
linkField: aDatabaseField
	linkFields := Array with: aDatabaseField.! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
linkFields
	^linkFields.! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
linkFields: aCollection
	"Set the fields that are foreign keys *from* the link table (this is only used if we have one) to the *target* table. This is used as a hint in the case where the link table is ambiguous, and is easier than specifying the full join."
	linkFields := aCollection.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
linkTableJoin
	linkTableJoin isNil ifTrue: [^self join].
	^linkTableJoin! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
linkTableJoin: aJoin
	"Set the join from the link table to the target table. Normally this is calculated automatically (see reverseJoin), but it can be set manually if that's not possible (for example, because you want to include a constant value in the join, that won't be included in the foreign key constraints). See also relevantLinkTableFields if you need to give a hint."
	linkTableJoin := aJoin.
	[aJoin owningMapping: self] on: MessageNotUnderstood do: [:ex | ex | ex return].! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject intoRowsIn: aRowMap 
	"Our target is a collection. The tricky bit is that if we're building rows into a RowMapForMementos, then the collection we contain isn't the one we want to use. We want the old version. Ask the row map to give it to us. If it's a normal row map, we'll just get the same thing back" 
	| target mementoizedTarget|
	self canWrite ifFalse: [^self].
	target := self getValueFrom: anObject.
	target := self session realObjectFor: target ifNone: [^self].
	(self isValidTarget: target) ifTrue: [
		mementoizedTarget := aRowMap collectionMementoFor: target.
		self mapFromObject: anObject toTarget: mementoizedTarget puttingRowsIn: aRowMap].! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject toTarget: target puttingRowsIn: aRowMap
	self attribute
		loopOver: target
		doing:
			[:index :eachTarget | 
			| intermediate |
			intermediate := self
				rowMapKeyForSource: anObject
				target: eachTarget
				index: index.
			self
				mapOneFromObject: anObject
				toTarget: eachTarget
				intermediate: intermediate
				atPosition: index
				puttingRowsIn: aRowMap].! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
mapObject: anObject inElementBuilder: anElementBuilder
	| value session parameters |
	session := descriptor session.
	parameters := self parametersIn: anElementBuilder.
	value := self shouldProxy
		ifTrue:
			[| proxy |
			proxy := (self newProxyIn: anElementBuilder)
				session: session;
				parameters: parameters;
				glorpOwner: anObject.
			"If we know in advance the proxy would return nil, or an empty collection, just put that in rather than the proxy"
			(proxy query shortCircuitEmptyReturn: parameters)
				ifTrue: [proxy getValue]
				ifFalse: [proxy]]
		ifFalse:
			[(self queryFor: anElementBuilder)
				executeWithParameters: parameters
				in: session].
	self setValueIn: anObject to: value.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
mapOneFromObject: sourceObject toTarget: targetObject intermediate: aRowMapKey atPosition: index puttingRowsIn: aRowMap
	self linkTableJoin
		mapFromSource: sourceObject
		andTarget: aRowMapKey
		intoRowsIn: aRowMap.
	self targetTableJoin isNil
		ifFalse:
			[self targetTableJoin
				mapFromSource: sourceObject
				andTarget: (self mapsSimpleObject ifTrue: [aRowMapKey] ifFalse: [targetObject])
				intoRowsIn: aRowMap].
	self usesLinkTable
		ifTrue:
			[self reverseJoin
				mapFromSource: aRowMapKey
				andTarget: (self mapsSimpleObject ifTrue: [aRowMapKey] ifFalse: [targetObject])
				intoRowsIn: aRowMap].
	self
		writeExtraInformationInto: aRowMap
		atKey: aRowMapKey
		forSourceObject: sourceObject
		targetObject: targetObject
		atIndex: index.! !

!RelationshipMapping methodsFor: 'testing' stamp: ''!
mappedFields
	"Return a collection of fields that this mapping will write into any of the containing object's rows"

	^self join allSourceFields select: [:each | each isMappable and: [(each isGlorpExpression not or: [each isConstantExpression not])]].! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
mappingCriteria
	^self join.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
mappingCriteria: aJoin
	self join: aJoin.! !

!RelationshipMapping methodsFor: 'preparing' stamp: ''!
multipleTableExpressionsFor: anExpression
	^self referenceDescriptor multipleTableJoin
		collect: [:each | each asExpressionJoiningSource: anExpression toTarget: anExpression]! !

!RelationshipMapping methodsFor: 'testing' stamp: ''!
needsSpecialRowMapKey
	"Return true if we have to use a special key object for the row map, rather than just our target"
	^self usesLinkTable.! !

!RelationshipMapping methodsFor: 'initializing' stamp: ''!
oneWayConstraintsFrom: oneTableSet to: anotherTableSet
	| matches |
	matches := OrderedCollection new: 3.
	oneTableSet do: [:eachFromTable |
		 anotherTableSet do: [:eachToTable |
			matches addAll:  (eachFromTable foreignKeyConstraints
					select: [:eachConstraint | eachConstraint refersToTable: eachToTable])]].
	^matches.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
parametersIn: anElementBuilder
	| parameters |
	self shouldUseFilteredRead ifTrue: [^anElementBuilder command parameters].
	parameters := IdentityDictionary new.
	self join
		mappableSourceFieldsDo:
			[:eachSource | parameters at: eachSource put: (anElementBuilder valueOfField: eachSource)].
	^parameters.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
proxyAction
	"Return an action each proxy should take when it's instantiated. This is a two-argument block
		[:owner :target | "
	^proxyAction.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
proxyAction: aBlock
	"Set an action each proxy should take when it's instantiated. This is a two-argument block
		[:owner :target | "
	proxyAction := aBlock.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
query

	query isNil ifTrue: [self buildQuery].
	^query.! !

!RelationshipMapping methodsFor: 'proxies' stamp: ''!
queryFor: anObjectBuilder 
	"If we have a specific query we should use in this case, rather than our general one, construct it. Right now that only applies if we're using filtered reads"
	| basicQuery collection primaryKey newQuery cachedQuery filterBlock |
	self shouldUseFilteredRead 
		ifFalse: [^self query].
	cachedQuery := anObjectBuilder cachedQueryFor: self.
	cachedQuery isNil 
		ifTrue: 
			[basicQuery := self buildGeneralQueryIn: anObjectBuilder.
			self query tracing alsoFetchExpressions 
				do: [:each | basicQuery alsoFetch: each].
			collection := GlorpVirtualCollection new setQuery: basicQuery]
		ifFalse: [collection := cachedQuery collection].
	primaryKey := anObjectBuilder primaryKey.
	filterBlock := self descriptor keyFields size = 1 
				ifTrue: [[:objects :parameters | (objects at: 1) = primaryKey]]
				ifFalse: 
					[
					[:objects :parameters | 
					(1 to: primaryKey size) inject: true
						into: [:sum :i | sum and: [(primaryKey at: i) = (objects at: i)]]]].
	newQuery := FilteringQuery on: collection filter: filterBlock.
	newQuery collectBlock: [:each | each last].
	newQuery readsOneObject: self attribute isCollectionAttribute not.
	newQuery collectionType: self collectionType.
	anObjectBuilder cachedQueryFor: self is: newQuery.
	^newQuery! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
referenceClass: aClass
	self attribute type: aClass.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
referenceDescriptor

	^self system descriptorFor: self referenceClass.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
referenceTableFieldFor: aField
	^self usesLinkTable
		ifTrue: [nil]
		ifFalse: [self join targetForSource: aField].! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
referencedIndependentObjectsFrom: anObject

	| value |
	self canWrite ifFalse: [^#()].
	value := self getValueFrom: anObject.
	value == nil ifTrue: [^#()].
	^Array with: value.! !

!RelationshipMapping methodsFor: 'initializing' stamp: ''!
relatedTables
	| otherTables possibleLinkTables |
	otherTables := self basicReferenceDescriptor tables.
	usesLinkTable ifFalse: [^otherTables].

	linkFields isNil ifFalse: [^(linkFields collect: [:each | each table]) asSet asArray].
	possibleLinkTables := self system allTables
		select:
			[:each | 
			(each hasConstraintsToAnyOf: self descriptor tables)
				and: [each hasConstraintsToAnyOf: otherTables]].
	possibleLinkTables size ~= 1
		ifTrue: [CannotAutomaticallyDetermineJoin new
			tables: possibleLinkTables;
			signal: 'Cannot automatically determine link table..'].
	^possibleLinkTables.! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
relevantLinkTableFields
	^linkFields.! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
relevantLinkTableFields: aCollection
	linkFields := aCollection.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
reverseJoin

	reverseJoin isNil ifTrue: [reverseJoin := self joinFromLinkToReferenceTable].
	^reverseJoin.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
reverseJoin: aJoin
	"Set the reverse join - the join from the link table to the target table. Normally this will be calculated based on the foreign key constraints, but in rare cases (e.g. a foreign key that may be to more than one table) you may want to set it explicitly"

	reverseJoin := aJoin.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
reverseMapping
	"Construct the backwards mapping that corresponds to us. If it's using a link table, then we are at least potentially a toMany mapping. Otherwise, we can point backwards only to a single object."
	| mapping |
	^self usesLinkTable
		ifTrue: 
			[mapping := (ManyToManyMapping new)
				attributeName: #temporaryAttributeFromSubselect;
				referenceClass: self descriptor describedClass;
				join: self reverseJoin inverseJoin;
				descriptor: self descriptor.
			mapping reverseJoin: self join inverseJoin onlyJoinParts]
		ifFalse: 
			[(OneToOneMapping new)
				attributeName: #temporaryAttributeFromSubselect;
				referenceClass: self descriptor describedClass;
				join: self join inverseJoin;
				descriptor: self descriptor]! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
rowMapKeyForSource: source target: target index: index
	"Create the row map key. Make sure we're holding onto the real objects, so proxies don't time out on us"

	self needsSpecialRowMapKey ifFalse: [^target].
	(target yourSelf isNil or: [source yourSelf isNil]) ifTrue: [^nil].
	^rowMapKeyConstructorBlock isNil
		ifTrue:
			[self
				buildRowMapKeyForSource: source
				target: target
				index: index]
		ifFalse:
			[rowMapKeyConstructorBlock value: source yourSelf value: target yourSelf].! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
setGeneralQueryRetrievalIn: newQuery basedOn: base phantom: phantomExpression
	self descriptor keyFields
		do: [:eachField | newQuery retrieve: (phantomExpression getField: eachField)].
	newQuery retrieve: base.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
shouldProxy

	^shouldProxy.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
shouldProxy: aBoolean

	shouldProxy := aBoolean.! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
shouldUseFilteredRead
	^shouldUseFilteredRead.! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
shouldUseFilteredRead: aBoolean
	shouldUseFilteredRead := aBoolean.! !

!RelationshipMapping methodsFor: 'private/expressions' stamp: ''!
singleJoinFromLinkToReferenceTableForConstraints: constraints
	"Generate the inverse mapping expression, i.e. the one connecting the link table to the reference table, using the foreign key constraints. We assume that if there's a general part, it applies only to the original join, and not to the reverse join that we're generating."

	| expression generalPart |
	expression := Join new.
	constraints do: [:eachConstraint | 
		eachConstraint sourceFields with: eachConstraint targetFields do: 
			[:eachSource :eachTarget |
				expression addSource: eachSource target: eachTarget]].
	generalPart := self join generalExpressionPart.
	^generalPart isNil ifTrue: [expression] ifFalse: [generalPart AND: expression].! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
targetTableJoin
	^targetTableJoin! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
targetTableJoin: aJoin
	"Set the join from the source table to the target table. Note that is is only required if you are using a link table, and in addition to the link table join need to connect from the source table directly to the target table as well.  This is *NOT* supposed to be a join from the link table to the target table. That's reverseJoin, and is calculated automatically. See relevantLinkTableFields if you need to give a hint, or linkTableJoin: if you want to set it explicitly."
	targetTableJoin := aJoin.
	aJoin owningMapping: self.! !

!RelationshipMapping methodsFor: 'validation' stamp: ''!
targetTables
	^((self usesLinkTable ifTrue: [self reverseJoin] ifFalse: [self join]) allTargetFields collect: [:each | each table]) asSet.! !

!RelationshipMapping methodsFor: 'processing' stamp: ''!
trace: aTracing context: anExpression
	"Currently we don't trace relationships across tables, so all we do here
is accumulate the list of embedded mappings"
	| newContext |
	(aTracing tracesThrough: self) ifFalse: [^self].
	newContext := anExpression get: self attributeName.
	aTracing addExpression: newContext.
	self referenceDescriptor trace: aTracing context: newContext.! !

!RelationshipMapping methodsFor: 'api' stamp: ''!
useFilteredRead
	shouldUseFilteredRead := true.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
useLinkTable
	usesLinkTable := true.! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
usesLinkTable
	^usesLinkTable! !

!RelationshipMapping methodsFor: 'accessing' stamp: ''!
usesLinkTable: anObject
	usesLinkTable := anObject! !

!RelationshipMapping methodsFor: 'validation' stamp: ''!
validate
	| targetJoinSourceFields targetJoinTargetTables |
	super validate.
	self join isNil ifTrue: [^self].
	targetJoinSourceFields := targetTableJoin isNil ifTrue: [#()] ifFalse: [targetTableJoin allSourceFields].
	targetJoinTargetTables :=  targetTableJoin isNil ifTrue: [#()] ifFalse: [targetTableJoin allTargetFields collect: [:each | each table]].
	(self join allSourceFields asArray, targetJoinSourceFields)
		do:
			[:each | 
			(each class == DatabaseField or: [each isKindOf: FunctionExpression]) 
				ifTrue: [(self descriptor tables includes: each table)
					ifFalse: [self error: 'Joins should go from the table(s) associated with this descriptor to the other table(s)']]].
	(self targetTables asArray, targetJoinTargetTables)
		do:
			[:each | 
			(self referenceDescriptor allTables includes: each)
				ifFalse:
					[self
						error:
							'Joins should go to the table(s) associated with the reference class of the mapping']].

	join allSourceFields asSet size = join allSourceFields size ifFalse: [self error: 'The same field occurs twice in the join. This might be legal, but is probably a cut and paste error.'].
	join allTargetFields asSet size = join allTargetFields size ifFalse: [self error: 'The same field occurs twice in the join. This might be legal, but is probably a cut and paste error.'].

	(self isExclusive and: [self canWrite not]) ifTrue: [self error: 'It is not meaningful to have an exclusive relationship that is read-only.'].! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
valueOfField: aField fromObject: anObject
	^self valueOfField: aField fromObject: anObject ifNone: [self error: 'Cannot find value'].! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
valueOfField: aField fromObject: anObject ifNone: aBlock
	"Return the value of a given field in our target object. Used in extracting primary keys when they are mapped via relationships"
	| targetField target |
	targetField := self referenceTableFieldFor: aField.
	targetField isNil ifTrue: [^aBlock value].
	target := self getValueFrom: anObject.
	"If it's a proxy and uninstantiated, we have to get the value from the proxy"
	(target isGlorpProxy and: [target isInstantiated not])
		ifTrue: [^target parameters at: aField].
	"If the object is instantiated, then we prefer to get the value from the object, because if it's non-primary key then the proxy may not know it, or it may vary from what the proxy has. Well, that seems unlikely, but try this anyway"
	self attribute 	
		loopOver: target 
		doing: [:index :eachTarget | | value |
			value := self referenceDescriptor valueOfField: targetField fromObject: eachTarget ifNone: [self].
			value == self ifFalse: [^value]].
	"We couldn't get it from the object, probably because it's unmapped. See if the proxy has it after all"
	target isGlorpProxy ifTrue: [^target parameters at: aField].
	"We couldn't get the value at all. This may mean that the object is new, so it has no proxies holding the value. Return nil and hope it's new and that it will all get figured out"
	^aBlock value.! !

!RelationshipMapping methodsFor: 'mapping' stamp: ''!
writeExtraInformationInto: aRowMap atKey: rowMapKey forSourceObject: sourceObject targetObject: targetObject  atIndex: index! !

RelationshipMapping subclass: #ToManyMapping
	instanceVariableNames: 'orderBy shouldWriteTheOrderField'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!ToManyMapping commentStamp: '<historical>' prior: 0!
This is a superclass for mappings which represent a relationship between one object and a collection of others. It isn't an abstract superclass, you can use this class directly, and you will get the same behaviour as if you were using OneToManyMapping. Or if you set useLinkTable to true, you'll get the same behaviour as a ManyToManyMapping.

Instance Variables:
	orderBy	<(SequenceableCollection of: (GlorpExpression | BlockClosure))>	By what criteria should we order the collection of items when we are reading.
	shouldWriteTheOrderField	<Boolean>	On write, when we write an item in the collection, should we also write the index at which it occurs in the collection into the database. This lets us preserve ordering if we just used a SequenceableCollection in memory. If this is true, we assume that the first item in orderBy will give us the field to write the index into.

!

ToManyMapping subclass: #ConditionalToManyMapping
	instanceVariableNames: 'conditional'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!ConditionalToManyMapping commentStamp: '<historical>' prior: 0!
This is a conditional mapping, but one where the condition is applied to a relationship with a collection, and applies to individual elements of the collection. So, rather than making a single choice overall of which mapping to apply, we loop over the collection and make the choice separately for each element.

This is experimental, and probably suffers from lots of limitations.!

!ConditionalToManyMapping methodsFor: 'mapping' stamp: ''!
applicableMappingForObject: anObject
	"We can actually be called here with our parent object, but the contexts in which that happens we basically just want to do nothing, so return our conditional mapping."
	[self attribute checkTargetType: anObject] on: GlorpInvalidTypeError do: [:ex | ^self conditional].
	^self conditional applicableMappingForObject: anObject.! !

!ConditionalToManyMapping methodsFor: 'mapping' stamp: ''!
applicableMappingForRow: anArray in: anElementBuilder
	^self conditional applicableMappingForRow: anArray in: anElementBuilder.! !

!ConditionalToManyMapping methodsFor: 'mapping' stamp: ''!
buildQuery
	query := CompositeQuery new.
	self conditional casesDo: [:each |
		query addQuery: each query].
	^query! !

!ConditionalToManyMapping methodsFor: 'accessing' stamp: ''!
conditional
	conditional isNil ifTrue: [conditional := self newMapping: ConditionalMapping].
	^conditional.! !

!ConditionalToManyMapping methodsFor: 'accessing' stamp: ''!
conditional: anObject
	conditional := anObject! !

!ConditionalToManyMapping methodsFor: 'conditions' stamp: ''!
conditionalField: aField
	self conditional conditionalField: aField.! !

!ConditionalToManyMapping methodsFor: 'conditions' stamp: ''!
conditionalMethod: aSymbol
	self conditional conditionalMethod: aSymbol.! !

!ConditionalToManyMapping methodsFor: 'conditions' stamp: ''!
forField: aField attribute: attributeBlock if: testBlock useMapping: aMapping
	self conditionalField: aField.
	self conditionalMethod: attributeBlock.
	self if: testBlock then: aMapping.! !

!ConditionalToManyMapping methodsFor: 'conditions' stamp: ''!
if: conditionSelector then: aMapping
	self conditional if: conditionSelector then: aMapping.! !

!ConditionalToManyMapping methodsFor: 'mapping' stamp: ''!
mapOneFromObject: sourceObject toTarget: targetObject intermediate: aRowMapKey atPosition: index puttingRowsIn: aRowMap
	
	(self applicableMappingForObject: targetObject) mapOneFromObject:
			sourceObject toTarget: targetObject intermediate: aRowMapKey atPosition:
			index puttingRowsIn: aRowMap.! !

!ConditionalToManyMapping methodsFor: 'accessing' stamp: ''!
mappedFields
	^self conditional mappedFields! !

!ConditionalToManyMapping methodsFor: 'accessing' stamp: ''!
newMapping: aMappingClass
	"Implement this so that we can create our sub-mappings without having them added to the parent descriptor as independent mappings"
	| mapping |
	mapping := aMappingClass new.
	mapping descriptor: self descriptor.
	descriptor resetMappedFields.
	^mapping.! !

!ConditionalToManyMapping methodsFor: 'conditions' stamp: ''!
otherwise: aMapping
	self conditional otherwise: aMapping.! !

!ConditionalToManyMapping methodsFor: 'mapping' stamp: ''!
parametersIn: anElementBuilder
	| parameters |
	self shouldUseFilteredRead ifTrue: [^anElementBuilder command parameters].
	parameters := IdentityDictionary new.
	self conditional casesDo: [:each |
		each join mappableSourceFieldsDo:
			[:eachSource | parameters at: eachSource put: (anElementBuilder valueOfField: eachSource)]].
	^parameters.! !

!ConditionalToManyMapping methodsFor: 'mapping' stamp: ''!
rowMapKeyForSource: source target: target index: index
	^(self applicableMappingForObject: target) rowMapKeyForSource: source target: target index: index.! !

ToManyMapping subclass: #DictionaryMapping
	instanceVariableNames: 'keyMapping valueMapping associationDescriptor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!DictionaryMapping commentStamp: '<historical>' prior: 0!
This allows us to map a dictionary into tables. It should, I hope, support arbitrarily either mapping to single fields, or to objects, for either the key or the value. At the moment, it at least passes the tests for the value->object and value->value cases.

Instance Variables:
	keyMapping	<DatabaseMapping>	The mapping to use to read/write the key. It should be for an attribute named #key.
	valueMapping	<DatabaseMapping>	The mapping to use to read/write the value. It should be for an attribute named #value. Note that this is optional, and isn't a full blown mapping, because if we have a relationship, we will normally just use this mapping to describe it directly. If we have a direct mapping for the value, then we have this odd situation where we use the Join from this mapping, but the direct mapping to control the read/write of the variable.
	associationDescriptor	<Descriptor>	We manipulate the key/value pairs as associations in some places, and have a mock descriptor for association, specific to this mapping, which is used to get the key/value.


!

DictionaryMapping subclass: #BasicDictionaryMapping
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!BasicDictionaryMapping commentStamp: '<historical>' prior: 0!
This class is obsolete, but exists only so that people using the old name will not have to change their code.!

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
associationClassModel
	| model |
	model := GlorpClassModel for: Association.
	model newAttributeNamed: #key.
	model newAttributeNamed: #value.
	^model.! !

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
associationDescriptor
	associationDescriptor isNil ifTrue: [
		associationDescriptor := Descriptor new.
		associationDescriptor system: self system.
		associationDescriptor classModel: self associationClassModel].
	^associationDescriptor.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
basicReferenceDescriptor
	^self system descriptorFor: self basicReferenceClass.! !

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
buildQuery
	| keyExpression valueExpression |
	query := super buildQuery.
	keyExpression := self keyExpression.
	valueExpression := self valueExpression.
	query retrieveDictionaryWithKey: keyExpression value: valueExpression.
	^query.! !

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
buildRowMapKeyForSource: source target: target index: index
	^self mapsSimpleObject 
		ifTrue: [
			MultipleRowMapKey
				with: target yourSelf
				with: source yourSelf
				with: index]
		ifFalse: [super buildRowMapKeyForSource: source target: target index: index].! !

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
chooseBaseForMapping: aMapping from: base and: phantomExpression
	aMapping isNil ifTrue: [^base].
	^(aMapping mappedFields
		allSatisfy: [:each | base tables includes: each table])
		ifTrue: [base]
		ifFalse: [phantomExpression].! !

!DictionaryMapping methodsFor: 'initializing' stamp: ''!
defaultCollectionType
	^Dictionary.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
field
	^self valueField.! !

!DictionaryMapping methodsFor: 'testing' stamp: ''!
glorpCanHaveDescriptorDirectly
	"We allow instances of ourselves to have a descriptor, which normally only classes can do. This is so that we can have multiple different descriptors in use for e.g. String, depending on the particular usage."
"	^self mapsSimpleObject."
	^true.! !

!DictionaryMapping methodsFor: 'initializing' stamp: ''!
initialize
	super initialize.
	usesLinkTable := true.! !

!DictionaryMapping methodsFor: 'initializing' stamp: ''!
initializeValueMapping
	| des |
	self descriptor system privateAndBasicAddDescriptorFor: self as: self associationDescriptor.
	(self associationDescriptor classModel attributeNamed: #key) type: self attribute keyType.
	(self associationDescriptor classModel attributeNamed: #value) type: self attribute type.
	des := self referenceDescriptor.
	(des isNil or: [des tables isEmpty]) ifTrue: [self initializeValueMappingJoin].! !

!DictionaryMapping methodsFor: 'initializing' stamp: ''!
initializeValueMappingJoin
	| tables |
	"We don't know which of these will come first in the descriptor, so call the other initialization if it looks like we have enough information, and it hasn't been done"
	self referenceDescriptor isNil ifTrue: [self initializeValueMapping].
	"Have we already done this?"
	self referenceDescriptor tables isEmpty ifFalse: [^self].

	tables := self targetTables asArray.
	tables do: [:each | self referenceDescriptor addTable: each].
	(self referenceDescriptor newMapping: DirectMapping) 
		from: #key to: self keyField.
	self mapsSimpleObject 
		ifTrue: [
			(self referenceDescriptor newMapping: DirectMapping) 
			from: #value to: self valueField]
		ifFalse: [
			(self referenceDescriptor newMapping: OneToOneMapping)
				attributeName: #value;	
				linkFields: self linkFields;
				usesLinkTable: self usesLinkTable;
				join: self join;
				controlsTables: false].! !

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
keyExpression
	"Return the expression to use for our key. The key can be built using data from either the target object, or from the link table, so build the expression on one or the other, as appropriate"
	 | base intermediate |
	base := BaseExpression new.
	intermediate := (self descriptor tables includes: keyMapping mappedFields first table)
		ifTrue: [base] 
		ifFalse: [base getTable: keyMapping mappedFields first table].
	^intermediate getMapping: keyMapping named: #key.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
keyField
	^keyMapping field.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
keyField: aDatabaseField
	keyMapping := self newMapping: DirectMapping.
	keyMapping attributeName: #key.
	keyMapping from: #key to: aDatabaseField.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
keyMapping
	^keyMapping! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
keyMapping: aMapping
	keyMapping := aMapping! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
mappedFields
	"Return a collection of fields that this mapping will write into any of the containing object's rows"

	| fields |
	fields := super mappedFields copy.
	fields addAll: keyMapping mappedFields.
	valueMapping isNil ifFalse: [
		fields addAll: valueMapping mappedFields].
	^fields.! !

!DictionaryMapping methodsFor: 'testing' stamp: ''!
mapsSimpleObject
	^valueMapping notNil and: [valueMapping mapsSimpleObject].! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
name
	"This can get called when we're masquering as a class, for purposes of indexing descriptors, so just print something useful"
	^self printString.! !

!DictionaryMapping methodsFor: 'testing' stamp: ''!
needsSpecialRowMapKey
	^super needsSpecialRowMapKey or: [
		self mapsSimpleObject].! !

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
newMapping: aMappingClass
	"Implement this so that we can create our sub-mappings without having them added to the parent descriptor as independent mappings"
	| mapping |
	mapping := aMappingClass new.
	mapping descriptor: self associationDescriptor.
	descriptor resetMappedFields.
	^mapping.! !

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
pseudoDescriptor
	"Return a pretend descriptor for our reference class. Since that class is a primitive, it won't have a real descriptor, and it may vary by usage anyway"
	| pseudoDescriptor |
	pseudoDescriptor := Descriptor new.
	pseudoDescriptor system: descriptor system.
	^pseudoDescriptor.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
referenceClass
	"Return ourselves as the reference class, as a sort of a proxy for what we really want, which is a primitive, and varies according to the relationship anyway. Note that we have stuffed in a descriptor in the system, with ourselves (this instance) as the class."
"	^self mapsSimpleObject ifTrue: [self] ifFalse: [super referenceClass]."
	^self.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
referenceDescriptor
	| ref |
	ref := super referenceDescriptor.
	^ref isNil 
		ifTrue: [
			"This means we haven't yet added our fake descriptor, and are in initialization, so just return the descriptor for the 'real' reference class"
			self system descriptorFor: self basicReferenceClass]
		ifFalse: [ref].! !

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
setGeneralQueryRetrievalIn: newQuery basedOn: base phantom: phantomExpression
	| keyExpression valueExpression collType keyExpressionBase valueExpressionBase |
	self descriptor primaryKeyMappings
		do:
			[:eachMapping | newQuery retrieve: (phantomExpression get: eachMapping attributeName)].
	keyExpressionBase := self
		chooseBaseForMapping: keyMapping
		from: base
		and: phantomExpression.
	keyExpression := self keyExpression asGlorpExpressionOn: keyExpressionBase.
	valueExpressionBase := self
		chooseBaseForMapping: valueMapping
		from: base
		and: phantomExpression.
	valueExpression := self valueExpression
		asGlorpExpressionOn: valueExpressionBase.
	collType := (GlorpDictionaryDescription new)
		keyExpression: keyExpression;
		valueExpression: valueExpression.
	newQuery collectionType: collType.
	newQuery retrieve: base.! !

!DictionaryMapping methodsFor: 'validation' stamp: ''!
validate
	self initializeValueMapping.
	super validate.
	self attribute isDictionaryAttribute
		ifFalse: [self error: 'Dictionary Mapping that does not have a dictionary type in the attribute'].

	"The reason you can't do this, is that we end up throwing away the primary key of the target table when we read in the objects, and we have no way to properly do deletes/updates without that information"
	(self usesLinkTable and: [self mapsSimpleObject]) ifTrue: [self error: 'Dictionary mappings to simple objects cannot use a link table'].! !

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
valueExpression
	 | base intermediate |
	valueMapping isNil ifTrue: [ ^[:each | each value] ].
	base := BaseExpression new.
	intermediate := (self descriptor tables includes: valueMapping mappedFields first table)
		ifTrue: [base] 
		ifFalse: [base getTable: valueMapping mappedFields first table].
	^intermediate getMapping: valueMapping named: #value.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
valueField
	^valueMapping field.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
valueField: aDatabaseField
	| mapping |
	mapping := self newMapping: DirectMapping.
	mapping attributeName: #value.
	mapping from: #value to: aDatabaseField.
	self valueMapping: mapping.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
valueMapping
	^valueMapping.! !

!DictionaryMapping methodsFor: 'accessing' stamp: ''!
valueMapping: aMapping
	valueMapping := aMapping.! !

!DictionaryMapping methodsFor: 'mapping' stamp: ''!
writeExtraInformationInto: aRowMap atKey: rowMapKey forSourceObject: sourceObject targetObject: targetObject atIndex: eachKey
	"Write out any additional information that goes in the link table or else the target table. That means our key, if necessary. If we have a specific value mapping, rather than just using the normal attributes of this mapping, then write that as well. That normally happens when our value is simple"
	
	"We need to know if the key mapping is trying to write to the link table, or to the target. If it's to the link table, then we always use the row map key. If it's to the target table, then we need to use whatever is appropriate for the target table, which is a row map key if we have a simple target, and otherwise it's just the target object"
	| target |
	target := (keyMapping mappedTables includes: self linkTable)
		ifTrue: [rowMapKey]
		ifFalse: [self mapsSimpleObject ifTrue: [rowMapKey] ifFalse: [targetObject]].
	keyMapping
		mapFromObject: eachKey -> nil
		toTarget: target
		puttingRowsIn: aRowMap.

	valueMapping isNil ifTrue: [^self].
	"I'm not sure this will ever be relevant. If the value mapping goes to the link table, then does that mean the key is really the value? And if so, why not just consider the other one to be the link?"
	target := (valueMapping mappedTables includes: self linkTable)
		ifTrue: [rowMapKey]
		ifFalse: [self mapsSimpleObject ifTrue: [rowMapKey] ifFalse: [targetObject]].
	valueMapping
		mapFromObject: nil -> targetObject
		toTarget: target
		puttingRowsIn: aRowMap.! !

ToManyMapping subclass: #DirectToManyMapping
	instanceVariableNames: 'resultField'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!DirectToManyMapping commentStamp: '<historical>' prior: 0!
This is a to-many mapping where the class in the collection is a simple type, not a compound object. This requires fiddling with the descriptors, to allow each such mapping to define a different descriptor for the target class. This should be regarded as something of a prototype. It seems to work, at least for basic operations, but it may well have problems.

Instance Variables:
	resultField	<DatabaseField>	The field in the target table in which we have our results.

!

!DirectToManyMapping methodsFor: 'accessing' stamp: ''!
attributeName: anAttributeName
	^self attributeName: anAttributeName referenceClass: nil.! !

!DirectToManyMapping methodsFor: 'accessing' stamp: ''!
attributeName: anAttributeName referenceClass: aClass
	"In addition to setting the attribute, use that information to further configure our faked-up descriptor. We can pass in a reference class here, although this is only really useful if we don't already have an explicit classModel."
	| model |
	super attributeName: anAttributeName.
	aClass isNil ifFalse: [attribute type: aClass].
	model := GlorpClassModel for: super referenceClass.
	"We add a mapping for our value. We don't normally use this, because we expect people to refer directly to the attribute, but it means, for example, that the field shows up in our mapped fields without us having to fake it up. And you could write a query referring to anAttribute myValue if you wanted"
	model newAttributeNamed: #myValue type: self attribute type.
	self referenceDescriptor classModel: model.! !

!DirectToManyMapping methodsFor: 'accessing' stamp: ''!
buildQuery
	query := super buildQuery.
	query retrieve: [:x | x getField: resultField].
	^query.! !

!DirectToManyMapping methodsFor: 'mapping' stamp: ''!
buildRowMapKeyForSource: source target: target index: index
	"What should this do??? We want to distinguish the entries in e.g. # (1 2 3 3 4 3 5). Or do we? They aren't distinguishable in the DB. Does that mean we always want to write the order field? Or what? What if this is a set in memory. Then we really can't have two entries"
	^MultipleRowMapKey
		with: target yourSelf
		with: source yourSelf
		with: index.! !

!DirectToManyMapping methodsFor: 'converting' stamp: ''!
convertedDbValueOf: anObject
	"It may be that we need to introduce a converter here, the same as with a regular direct mapping"
	^anObject.! !

!DirectToManyMapping methodsFor: 'accessing' stamp: ''!
descriptor: aDescriptor
	super descriptor: aDescriptor.
	aDescriptor system privateAndBasicAddDescriptorFor: self as: self pseudoDescriptor.! !

!DirectToManyMapping methodsFor: 'accessing' stamp: ''!
field
	^resultField.! !

!DirectToManyMapping methodsFor: 'testing' stamp: ''!
glorpCanHaveDescriptorDirectly
	"We allow instances of ourselves to have a descriptor, which normally only classes can do. This is so that we can have multiple different descriptors in use for e.g. String, depending on the particular usage."
	^true.! !

!DirectToManyMapping methodsFor: 'accessing' stamp: ''!
join: aJoin
	"In addition to setting up the join, use that information to configure our faked up descriptor"
	| tables |
	super join: aJoin.
	tables := self targetTables asArray.
	tables size > 1
		ifTrue: [self error: 'Cannot have multiple tables for a direct toMany'].
	self referenceDescriptor table: tables first.
	"We add a mapping for our value. We don't normally use this, because we expect people to refer directly to the attribute, but it means, for example, that the field shows up in our mapped fields without us having to fake it up. And you could write a query referring to anAttribute myValue if you wanted"
	(self referenceDescriptor newMapping: DirectMapping) 
		from: #myValue to: resultField.! !

!DirectToManyMapping methodsFor: 'testing' stamp: ''!
mapsSimpleObject
	^true.! !

!DirectToManyMapping methodsFor: 'testing' stamp: ''!
needsSpecialRowMapKey
	"Return true if we have to use a special key object for the row map, rather than just our target. We always do, because our target may not have identity."

	^true.! !

!DirectToManyMapping methodsFor: 'mapping' stamp: ''!
pseudoDescriptor
	"Return a pretend descriptor for our reference class. Since that class is a primitive, it won't have a real descriptor, and it may vary by usage anyway"
	| pseudoDescriptor |
	pseudoDescriptor := Descriptor new.
	pseudoDescriptor system: descriptor system.
	^pseudoDescriptor.! !

!DirectToManyMapping methodsFor: 'accessing' stamp: ''!
referenceClass
	"Return ourselves as the reference class, as a sort of a proxy for what we really want, which is a primitive, and varies according to the relationship anyway. Note that we have stuffed in a descriptor in the system, with ourselves (this instance) as the class."
	^self.! !

!DirectToManyMapping methodsFor: 'accessing' stamp: ''!
resultField
	^resultField.! !

!DirectToManyMapping methodsFor: 'accessing' stamp: ''!
resultField: aField
	resultField := aField.! !

!DirectToManyMapping methodsFor: 'validation' stamp: ''!
validate
	super validate.
	"We have to write the order field. Otherwise, we get confused e.g. on deletion, because we don't have a way of telling the rows in the rowmap apart when the indexes have changed"
	(self readOnly not and: [shouldWriteTheOrderField not]) ifTrue: [self error: 'To-many mappings to simple types must write an order field so that we have some way to tell the rows apart'].

	"If we're for a pseudo-variable, we're neither reading nor writing, so we don''t need the order."
	self isForPseudoVariable ifTrue: [^self].
	orderBy isNil ifTrue: [self error: 'To-many mappings to simple types must specify and write an order field so that we have some way to tell the rows apart'].! !

!DirectToManyMapping methodsFor: 'accessing' stamp: ''!
writeExtraInformationInto: aRowMap atKey: rowMapKey forSourceObject: sourceObject targetObject: targetObject atIndex: index
	super
		writeExtraInformationInto: aRowMap
		atKey: rowMapKey
		forSourceObject: sourceObject
		targetObject: targetObject
		atIndex: index.
	aRowMap
		at: self resultField
		key: rowMapKey
		put: targetObject.! !

ToManyMapping subclass: #ManyToManyMapping
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!ManyToManyMapping commentStamp: '<historical>' prior: 0!
This is a to-many relationship in which we use a link table to the collection, rather than a foreign key in the items of the collection pointing back to the "owner". Using this class is exactly equivalent to just using a ToManyMapping and setting the useLinkTable field to true.!

!ManyToManyMapping methodsFor: 'initializing' stamp: ''!
initialize
	super initialize.
	usesLinkTable := true.! !

ToManyMapping subclass: #OneToManyMapping
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!OneToManyMapping commentStamp: '<historical>' prior: 0!
This represents a mapping from one object to a collection of others. This class is really only here for backward compatibility. Using ToManyMapping directly will give exactly the same behaviour.

!

ToManyMapping subclass: #SQLite3PragmaMapping
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!SQLite3PragmaMapping methodsFor: 'mapping' stamp: ''!
buildQuery
	| toMany |
	toMany := self attribute isCollectionAttribute.
	query := toMany
		ifTrue: [SQLite3PragmaQuery read: self referenceClass where: self join]
		ifFalse: [SQLite3PragmaQuery readOneOf: self referenceClass where: self join].
	query session: self session.
	query baseExpression descriptor: (self session descriptorFor: self referenceClass).
	query collectionType: self collectionType.
	self usesLinkTable ifTrue: [
		query AND: (self expressionFromLinkToReferenceTableWithBase: query baseExpression)].
	orderBy isNil ifFalse: [orderBy do: [:each | query orderBy: each]].
	^query.! !

!SQLite3PragmaMapping methodsFor: 'initialize-release' stamp: ''!
initialize
	"Initialize a newly created instance. This method must answer the receiver."

	super initialize.
	" *** Replace this comment with the appropriate initialization code *** "
	^self! !

!SQLite3PragmaMapping methodsFor: 'public' stamp: ''!
setValueIn: anObject to: aValue
	"This method serves as a hook to do special work for the collection objects, #fields and #foreignKeysData.
	It connects the objects to their owning tables (a SQLite3DatabaseTable object), and also resets the field
	type for a primary key field whose owning table has 'AUTOINCREMENT' declared for its pkey field. Such
	a field has type #serial. It would be nicer if that info were available elsewhere, but right now, this is where
	the table and field first meet (and where that info can now be shared)."

	self canRead ifFalse: [^self].
	self attribute setValueIn: anObject to: aValue.
	self attribute name = #fields
		ifTrue: 
			[aValue do: 
					[:fld |
					fld table: anObject.
					fld isPrimaryKey
						ifTrue: [anObject hasAutoIncrement ifTrue: [fld type: self descriptor system platform serial]]]].
	self attribute name = #foreignKeysData
		ifTrue: [aValue do: [:fld | fld sourceTableName: anObject name]]! !

!ToManyMapping methodsFor: 'mapping' stamp: ''!
buildGeneralQueryIn: anObjectBuilder
	| newQuery |
	newQuery := super buildGeneralQueryIn: anObjectBuilder.
	orderBy isNil ifFalse: [orderBy do: [:each | newQuery orderBy: each]].
	^newQuery.! !

!ToManyMapping methodsFor: 'mapping' stamp: ''!
buildQuery
	query := super buildQuery.
	orderBy isNil ifFalse: [orderBy do: [:each | query orderBy: each]].
	^query.! !

!ToManyMapping methodsFor: 'api' stamp: ''!
collectionType
	^self attribute collectionType.! !

!ToManyMapping methodsFor: 'api' stamp: ''!
collectionType: aClass
	self attribute collectionType: aClass.! !

!ToManyMapping methodsFor: 'initialize/release' stamp: ''!
defaultCollectionType
	^OrderedCollection.! !

!ToManyMapping methodsFor: 'testing' stamp: ''!
hasValueThatCanProvideField: aField from: anObject
	"Return true if we have enough information in ourselves to populate our fields. The only time this won't be true is if we're either holding an uninstantiated proxy or have a to-many relationship mapping with an empty collection. Inline the superclass implementation to avoid fetching the value twice."
	| value |
	(super hasValueThatCanProvideField: aField from: anObject) ifFalse: [^false].
	value := self getValueFrom: anObject.
	^value isEmpty not.! !

!ToManyMapping methodsFor: 'initialize/release' stamp: ''!
initialize
	super initialize.
	shouldWriteTheOrderField := false.! !

!ToManyMapping methodsFor: 'initialize/release' stamp: ''!
initializeAttribute
	"This is where we can initialize aspects of the attribute, if required. The motivating example is initializing the collection type of the attribute if it wasn't set. We need it to have some type, but we don't want to initialize it while we're doing basic initialization and aren't yet properly attached to a descriptor or class model"
	super initializeAttribute.
	self attribute collectionType isNil
		ifTrue: [self attribute collectionType: self defaultCollectionType].! !

!ToManyMapping methodsFor: 'mapping' stamp: ''!
keyForOrderingFrom: anObject to: target
	self usesLinkTable ifFalse: [^target].
	^self orderField table == self linkTable ifTrue: [target] ifFalse: [anObject].! !

!ToManyMapping methodsFor: 'private/expressions' stamp: ''!
linkTable
	| referenceKeys |
	self usesLinkTable ifFalse: [^nil].
	referenceKeys := (self join) targetKeys asOrderedCollection.
	^referenceKeys first table.! !

!ToManyMapping methodsFor: 'api' stamp: ''!
orderBy
	^orderBy.! !

!ToManyMapping methodsFor: 'api' stamp: ''!
orderBy: aBlockOrSelector
	orderBy isNil ifTrue: [orderBy := OrderedCollection new].
	orderBy add: aBlockOrSelector.! !

!ToManyMapping methodsFor: 'private/expressions' stamp: ''!
orderField
	| exp |
	exp := Query new expressionBlockFor: orderBy first.
	^(exp asGlorpExpressionOn: (BaseExpression new descriptor: self referenceDescriptor)) field.! !

!ToManyMapping methodsFor: 'api' stamp: ''!
shouldWriteTheOrderField
	^shouldWriteTheOrderField! !

!ToManyMapping methodsFor: 'api' stamp: ''!
shouldWriteTheOrderField: aBoolean
	shouldWriteTheOrderField := aBoolean.! !

!ToManyMapping methodsFor: 'validation' stamp: ''!
validate

	| targetClassModel |
	super validate.
	(self collectionType = GlorpCursoredStream and: [self shouldUseFilteredRead]) ifTrue: [self error: 'Filtered mappings can cause duplicates and so cannot return cursors'].

	(shouldWriteTheOrderField and: [orderBy isNil]) ifTrue: [self error: 'Mapping tries to write the collection order to a field, but has no order by field specified'].
	
	orderBy isNil ifTrue: [^self].
	targetClassModel := self system classModelFor: attribute type.

	"Note that this will only validate when the orderBy is a symbol - not a block, expression, or field"
	orderBy do: [:attributeName |
		attributeName isSymbol ifTrue: [
			(targetClassModel hasAttributeNamed: attributeName) ifFalse: [
				self error: attribute printString, ' is attempting to order by #', attributeName, ' which is not a mapped attribute of ', targetClassModel describedClass name]]].! !

!ToManyMapping methodsFor: 'mapping' stamp: ''!
writeExtraInformationInto: aRowMap atKey: rowMapKey forSourceObject: sourceObject targetObject: targetObject atIndex: index
	shouldWriteTheOrderField
		ifTrue:
			[| keyForOrdering |
			keyForOrdering := self keyForOrderingFrom: sourceObject to: rowMapKey.
			aRowMap
				at: self orderField
				key: keyForOrdering
				put: index].! !

!ToManyMapping methodsFor: 'api' stamp: ''!
writeTheOrderField
	shouldWriteTheOrderField := true.! !

Mapping subclass: #TypeMapping
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!TypeMapping commentStamp: '<historical>' prior: 0!
TypeMappings are used to implement inheritance. They do not map to particular instance variables, but instead define things about the way the class is read. The default type mapping is IdentityTypeMapping, which does nothing. To use inheritance, substitute either a FilteredTypeMapping or a HorizontalTypeMapping. TypeMappings become part of a TypeResolver, which keeps the information for a class and its subclasses, each of which will have a typeMapping specific to that class.

Subclasses must implement the following messages:
	testing
		isAbstract

!

TypeMapping subclass: #FilteredTypeMapping
	instanceVariableNames: 'field key keyDictionary'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!FilteredTypeMapping commentStamp: '<historical>' prior: 0!
FilteredTypeMapping assumes that all objects are stored in a single table, and there's a field in that table that identifies which subclass it is.

Instance Variables:
	field	<DatabaseField>	The field which identifies the subclass.
	key	<Object>	The value we expect to find in that field for this particular subclass
	keyDictionary	<Dictionary>	A dictionary for us and all of our subclasses from key values to subclass.

!

!FilteredTypeMapping classSide methodsFor: 'instance creation' stamp: ''!
to: field keyedBy: key
	^self new
		field: field
		keyedBy: key! !

!FilteredTypeMapping methodsFor: 'mapping' stamp: ''!
addTypeMappingCriteriaTo: collection in: base
	| singleRightValue r l |
	singleRightValue := self keys size = 1.
	r := ConstantExpression for: (singleRightValue
											ifTrue: [self keys asArray first]
											ifFalse: [self keys]).
	l := FieldExpression forField: self field basedOn: base.
	collection add: (singleRightValue ifTrue: [l equals: r] ifFalse: [l in: r]).! !

!FilteredTypeMapping methodsFor: 'initialize-release' stamp: ''!
buildKeyDictionary
	| subclassDescriptor |
	keyDictionary := Dictionary new.
	key isNil ifFalse: [keyDictionary at: key put: descriptor describedClass].
	descriptor describedClass glorpAllSubclasses do: 
			[:each | 
			subclassDescriptor := descriptor system descriptorFor: each.
			(subclassDescriptor isNil 
				or: [subclassDescriptor typeMapping keyedBy isNil]) 
					ifFalse: 
						[keyDictionary at: subclassDescriptor typeMapping keyedBy
							put: subclassDescriptor describedClass]]! !

!FilteredTypeMapping methodsFor: 'types' stamp: ''!
describedConcreteClassFor: aRow withBuilder: builder

	^self keyDictionary at: (aRow atIndex: (builder translateFieldPosition: field))! !

!FilteredTypeMapping methodsFor: 'accessing' stamp: ''!
field
	"If we don't have this, look it up from one of our sub-mappings"
	| concrete |
	field isNil ifTrue: [
		concrete := self descriptor typeResolver concreteMembers detect: [:each | each typeMapping field notNil].
		field := concrete typeMapping field].
	^field.! !

!FilteredTypeMapping methodsFor: 'initialize-release' stamp: ''!
field: aField keyedBy: aKey
	field := aField.
	key := aKey.! !

!FilteredTypeMapping methodsFor: 'accessing' stamp: ''!
keyDictionary
	keyDictionary isNil ifTrue: [ self buildKeyDictionary ].
	^keyDictionary! !

!FilteredTypeMapping methodsFor: 'accessing' stamp: ''!
keyedBy
	^key! !

!FilteredTypeMapping methodsFor: 'accessing' stamp: ''!
keyedBy: aKey
	key := aKey! !

!FilteredTypeMapping methodsFor: 'types' stamp: ''!
keys
	^self keyDictionary keys! !

!FilteredTypeMapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject intoRowsIn: aRowMap

	| row |
	self canWrite ifFalse: [^self].
	row := aRowMap findOrAddRowForTable: self field table withKey: anObject.
	row at: field put: key.! !

!FilteredTypeMapping methodsFor: 'accessing' stamp: ''!
mappedFields
	"Return a collection of fields that this mapping will write into any of the containing object's rows"

	^Array with: self field.! !

!FilteredTypeMapping methodsFor: 'mapping' stamp: ''!
trace: aTracing context: anExpression
	^self.! !

TypeMapping subclass: #HorizontalTypeMapping
	instanceVariableNames: 'mappedClass isAbstract'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!HorizontalTypeMapping commentStamp: '<historical>' prior: 0!
HorizontalTypeMapping assumes that all objects are stored in different tables. Only leaf classes in the inheritance hierarchy can be mapped. The descriptors for the classes themselves have the necessary information, our only job is to be involved in building up the query. Queries can be done either by splitting up into a series of separate queries, or as one horribly complex query with union all clauses.

Instance Variables:
	isAbstract	<Boolean>	Is this an abstract superclass.
	mappedClass	<Class>	The class that we are mapping.

!

!HorizontalTypeMapping classSide methodsFor: 'instance creation' stamp: ''!
forClass: aClass
	^self new
		mappedClass: aClass;
		yourself! !

!HorizontalTypeMapping methodsFor: 'mapping' stamp: ''!
allDescribedConcreteClasses
	| col |
	col := OrderedCollection new
		add: self describedClass;
		addAll: self describedClass glorpAllSubclasses;
		yourself.
	self needsWork: 'This belongs in someone else''s responsibility'.
	^col select: [:each |
		(self system descriptorFor: each) typeMapping isAbstract not ].! !

!HorizontalTypeMapping methodsFor: 'accessing' stamp: ''!
beAbstract
	isAbstract := true! !

!HorizontalTypeMapping methodsFor: 'accessing' stamp: ''!
isAbstract
	^isAbstract isNil ifTrue: [ isAbstract := false ] ifFalse: [ isAbstract ]! !

!HorizontalTypeMapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject intoRowsIn: aRowMap

	"do nothing"! !

!HorizontalTypeMapping methodsFor: 'initializing' stamp: ''!
mappedClass: aClass
	mappedClass := aClass! !

!HorizontalTypeMapping methodsFor: 'mapping' stamp: ''!
mappedFields
	^#()! !

!HorizontalTypeMapping methodsFor: 'mapping' stamp: ''!
trace: aTracing context: anExpression

	^self! !

TypeMapping subclass: #IdentityTypeMapping
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!IdentityTypeMapping commentStamp: '<historical>' prior: 0!
This is the default type mapping for classes that aren't using inheritance. It implements all of the protocol as no-ops.!

!IdentityTypeMapping methodsFor: 'testing' stamp: ''!
canBeTypeMappingParent
	^false! !

!IdentityTypeMapping methodsFor: 'testing' stamp: ''!
isAbstract
	^false! !

!IdentityTypeMapping methodsFor: 'testing' stamp: ''!
isTypeMappingRoot
	^true! !

!IdentityTypeMapping methodsFor: 'mapping' stamp: ''!
mappedFields
	^#()! !

!IdentityTypeMapping methodsFor: 'mapping' stamp: ''!
trace: aTracing context: anExpression
	"do nothing"! !

!TypeMapping methodsFor: 'mapping' stamp: ''!
addTypeMappingCriteriaTo: collection in: expression
	^self! !

!TypeMapping methodsFor: 'mapping' stamp: ''!
allDescribedConcreteClasses

	^Array with: (self describedClass)! !

!TypeMapping methodsFor: 'testing' stamp: ''!
canBeTypeMappingParent
	^true! !

!TypeMapping methodsFor: 'mapping' stamp: ''!
describedClass
	^self descriptor describedClass! !

!TypeMapping methodsFor: 'mapping' stamp: ''!
describedConcreteClassFor: aRow withBuilder: builder

	^self mappedClass! !

!TypeMapping methodsFor: 'mapping' stamp: ''!
descriptorForSuperclass
	^self system descriptorFor: self mappedClass superclass! !

!TypeMapping methodsFor: 'testing' stamp: ''!
hasAttribute
	"Return true if this maps an attribute, which is normally true"
	^false.! !

!TypeMapping methodsFor: 'testing' stamp: ''!
isAbstract
	self subclassResponsibility! !

!TypeMapping methodsFor: 'testing' stamp: ''!
isRelationship

	^false.! !

!TypeMapping methodsFor: 'testing' stamp: ''!
isTypeMapping
	^true! !

!TypeMapping methodsFor: 'testing' stamp: ''!
isTypeMappingRoot
	| superClassDescriptor |
	superClassDescriptor := self descriptorForSuperclass.
	^superClassDescriptor isNil or: [ superClassDescriptor typeMapping canBeTypeMappingParent not ]! !

!TypeMapping methodsFor: 'mapping' stamp: ''!
mapFromObject: anObject intoRowsIn: aRowMap
	"do  nothing"! !

!TypeMapping methodsFor: 'mapping' stamp: ''!
mapObject: anObject inElementBuilder: anElementBuilder
	"do  nothing"! !

!TypeMapping methodsFor: 'accessing' stamp: ''!
mappedClass
	^self descriptor describedClass! !

!TypeMapping methodsFor: 'mapping' stamp: ''!
referencedIndependentObjectsFrom: anObject
	^#().! !

!TypeMapping methodsFor: 'mapping' stamp: ''!
typeMappingRoot
	^self isTypeMappingRoot 
		ifTrue: [self mappedClass]
		ifFalse: [self descriptorForSuperclass typeMapping typeMappingRoot]! !

!TypeMapping methodsFor: 'validation' stamp: ''!
validateAttribute
	" TypeMapping doesn't use the attribute, validation would therefore fail. Probably should introduce AttributeMapping as sibling to TypeMapping "! !

Object subclass: #MultipleRowMapKey
	instanceVariableNames: 'keys'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-UnitOfWork'!

!MultipleRowMapKey commentStamp: '<historical>' prior: 0!
This is a special (and rarely needed) form of row map key that allows an arbitrary number of objects to participate in it.!

!MultipleRowMapKey classSide methodsFor: 'instance creation' stamp: ''!
new
	"Answer a newly created and initialized instance."

	^super new initialize! !

!MultipleRowMapKey classSide methodsFor: 'instance creation' stamp: ''!
with: key1 with: key2 with: key3
	(key1 isNil or: [key2 isNil or: [key3 isNil]]) ifTrue: [self error: 'Nil key in row map, probably indicates an invalid object structure'].
	^self new 
		addKey: key1;
		addKey: key2;
		addKey: key3;
		yourself.! !

!MultipleRowMapKey classSide methodsFor: 'instance creation' stamp: ''!
withAll: aCollection
	
	^self new 
		addAllKeys: aCollection;
		yourself.! !

!MultipleRowMapKey methodsFor: 'comparing' stamp: ''!
= aRowMapKey

	aRowMapKey class == self class ifFalse: [^false].
	aRowMapKey keys size = keys size ifFalse: [^false].
	"We can't rely on the order being the same because of the obscure but real case of two different related objects with the same identityHash"
	aRowMapKey keys do: [:each |
		(keys includes: each) ifFalse: [^false]].
	^true.! !

!MultipleRowMapKey methodsFor: 'accessing' stamp: ''!
addAllKeys: aKeyCollection
	keys addAll: aKeyCollection.! !

!MultipleRowMapKey methodsFor: 'accessing' stamp: ''!
addKey: aKey
	keys add: aKey.! !

!MultipleRowMapKey methodsFor: 'adjusting' stamp: ''!
adjustIn: aCorrespondenceMap
	"Adjust ourselves to refer to the copies rather than the originals"
	| adjusted |
	adjusted := self copy.
	adjusted keys: (keys collect: [:each | aCorrespondenceMap at: each ifAbsent: [each]]).
	^adjusted.! !

!MultipleRowMapKey methodsFor: 'comparing' stamp: ''!
hash
	"Note: the two things with the same identity hash problem is OK here because if they do then the hashes are (by definition) the same, and the order doesn't matter"
	| increment hash |
	increment := 14 // keys size.
	hash := 0.
	keys doWithIndex: [:each :index |
		hash := hash bitXor: (each identityHash bitShift: increment * index)].
	^hash.! !

!MultipleRowMapKey methodsFor: 'initialize-release' stamp: ''!
initialize
	keys := (SortedCollection new: 4) sortBlock: self sortBlock.! !

!MultipleRowMapKey methodsFor: 'accessing' stamp: ''!
keys
	^keys.! !

!MultipleRowMapKey methodsFor: 'accessing' stamp: ''!
keys: aCollection
	keys := (aCollection asSortedCollection: self sortBlock).! !

!MultipleRowMapKey methodsFor: 'printing' stamp: ''!
printOn: aStream
	aStream nextPutAll: 'KEY('.
	self hash printOn: aStream.
	aStream nextPutAll: '):'.
	keys printOn: aStream.! !

!MultipleRowMapKey methodsFor: 'adjusting' stamp: ''!
reverseAdjustIn: aRowMapForMementos
	"Adjust ourselves to refer to the originals rather than the copies"
	| adjusted |
	adjusted := self copy.
	adjusted keys: (keys collect: [:each | 
		aRowMapForMementos reversedCorrespondenceMap at: each ifAbsent: [each]]).
	^adjusted.! !

!MultipleRowMapKey methodsFor: 'initialize-release' stamp: ''!
sortBlock
	^[:a :b | a identityHash <= b identityHash].! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
asGlorpExpression
	^GlorpHelper glorpConstantExpressionClass for: self.! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
asGlorpExpressionOn: anExpression

	^self asGlorpExpression.! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
asGlorpExpressionOn: anExpression basedOn: aMessageArchiver withUltimateBase: aBaseExpression

	^self asGlorpExpression.! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
glorpBasicSize
	^self basicSize.! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
glorpCanHaveDescriptorDirectly
	^self isBehavior.! !

!Object methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:17'!
glorpCopyIn: aDictionary
	"For non-storeForGlorp objects, just copy normally"
	^self copy.! !

!Object methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:17'!
glorpEndOfIndexedInstVars
	"Return the index at which indexed instance variables start. The same as basicSize for most dialects, but funny for ObjectStudio"

	^self basicSize.! !

!Object methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:17'!
glorpIsBlock
	^false.! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
glorpIsCollection

	^false.! !

!Object methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:17'!
glorpPostFetch: aSession
	"This is a selector that objects can implement in order to receive notification when Glorp has read them."! !

!Object methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:17'!
glorpPostFetchValidate: aSession
	"This allows us to do post-read notification of the objects. Note that if this method explicitly returns a false, then we will treat that as meaning that the object is invalid and should not be read. Yes, this is kind of a hack."! !

!Object methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:17'!
glorpPostWrite: aSession! !

!Object methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:18'!
glorpPreWrite: aSession
	"This is a selector that objects can implement in order to receive notification when Glorp is about to write them. Only objects that have pending modifications will receive this notice. However, note that because of that, the data to be written will have already been calculated before this method is called, which makes it difficult to use this to update the method. If you want to put in audit information (e.g. the time at which the object was written, the associated username) then you probably want to use a generator associated with the field instead. See, for example GlorpRecordWithUpdate in the test suite."! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
glorpPrintSQLOn: aCommand

	self printOn: aCommand.! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
glorpPrintSelectSQLOn: aCommand

	self glorpPrintSQLOn: aCommand.! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
glorpRealObject
	"Return the real object. Will force instantiation of a proxy, even on compilers which inline yourself"
	^self.! !

!Object methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:18'!
glorpStartOfIndexedInstVars
	"Return the index at which indexed instance variables start. 1 for most dialects, but funny for ObjectStudio"
	^1.! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
isGlorpExpression
	^false.! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
isGlorpProxy
	^false.! !

!Object methodsFor: '*Glorp-Core' stamp: 'GuillermoPolito 4/26/2011 03:19'!
needsWork: aString

	^self.! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
todo
	"marker"! !

!Object methodsFor: '*Glorp-Core' stamp: ''!
yourSelf
	"Return the real object. Will force instantiation of a proxy, even on compilers which inline yourself"
	^self.! !

Object subclass: #ObjectTransaction
	instanceVariableNames: 'undoMap collectionMementos'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-UnitOfWork'!

!ObjectTransaction commentStamp: '<historical>' prior: 0!
An ObjectTransaction knows how to remember the state of objects and revert them back to that state later on. It does this by making a *shallow* copy of the registered objects and everything connected to them, and then putting that into an identity dictionary keyed by the originals.

If you have to undo, you push the state from the shallow copies back into the originals.

Yes, that works, and it's all you have to do. It even handles collections become:ing different sizes.

This is fairly independent of GLORP. You could use this mechanism in general, if you provided your own mechanism for figuring out what to register, or even just uncommented the one in here.

Instance Variables:
	undoMap	<IdentityDictionary>	 The dictionary of originals->copies.

!

!ObjectTransaction classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize! !

!ObjectTransaction methodsFor: 'begin/commit/abort' stamp: ''!
abort
	undoMap keysAndValuesDo: [:original :copy | self restoreStateOf: original toThatOf: copy]! !

!ObjectTransaction methodsFor: 'begin/commit/abort' stamp: ''!
abort: original
	| copy |
	copy := undoMap at: original ifAbsent: [^self].
	self restoreStateOf: original toThatOf: copy! !

!ObjectTransaction methodsFor: 'begin/commit/abort' stamp: ''!
begin
	self initializeUndoMap! !

!ObjectTransaction methodsFor: 'accessing' stamp: ''!
collectionMementoFor: anObject
	anObject isNil ifTrue: [^#()].
	^collectionMementos at: anObject ifAbsent: [self mementoFor: anObject].! !

!ObjectTransaction methodsFor: 'accessing' stamp: ''!
collectionMementos
	^collectionMementos.! !

!ObjectTransaction methodsFor: 'begin/commit/abort' stamp: ''!
commit
	self initializeUndoMap! !

!ObjectTransaction methodsFor: 'initializing' stamp: ''!
initialize
	self initializeUndoMap.! !

!ObjectTransaction methodsFor: 'initializing' stamp: ''!
initializeUndoMap

	undoMap := IdentityDictionary new: 1000.
	collectionMementos := IdentityDictionary new: 200.! !

!ObjectTransaction methodsFor: 'registering' stamp: ''!
isRegistered: anObject 
	"Note: We can never have a situation where a proxy is registered but its contents aren't, so we don't have to worry about that ambiguous case."
	| realObject |
	realObject := self realObjectFor: anObject ifNone: [^false].
	^undoMap includesKey: realObject.! !

!ObjectTransaction methodsFor: 'private/restoring' stamp: ''!
isShapeOf: original differentThanThatOf: copy

	^original class ~~ copy class or: [original glorpBasicSize ~= copy glorpBasicSize]! !

!ObjectTransaction methodsFor: 'accessing' stamp: ''!
mementoFor: anObject
	^undoMap at: anObject.! !

!ObjectTransaction methodsFor: 'registering' stamp: ''!
needsCollectionMementoFor: aCollection
	"We can avoid the need for a collection copy if the collection has no internal structure - that is, if we'll be able to loop over a simple shallow copy. This is dialect and collection class dependent"
	^Dialect needsCollectionMementoFor: aCollection.! !

!ObjectTransaction methodsFor: 'registering' stamp: ''!
realObjectFor: anObject 
	"If this is a proxy, return the contents (if available). Otherwise, return nil"
	^self realObjectFor: anObject ifNone: [nil].! !

!ObjectTransaction methodsFor: 'registering' stamp: ''!
realObjectFor: anObject ifNone: aBlock
	"If this is a proxy, return the contents (if available). Otherwise, evaluate the block"
	^anObject isGlorpProxy 
		ifTrue: [anObject isInstantiated ifTrue: [anObject getValue] ifFalse: [aBlock value]]
		ifFalse: [anObject]! !

!ObjectTransaction methodsFor: 'registering' stamp: ''!
register: anObject 
	"Make anObject be a member of the current transaction. Return the object if registered, or nil otherwise"

	| copy realObject |
	(self requiresRegistrationFor: anObject) ifFalse: [^nil].
	realObject := self realObjectFor: anObject ifNone: [^nil].
	copy := self shallowCopyOf: realObject ifNotNeeded: [^nil].
	undoMap at: realObject put: copy.
	self registerTransientInternalsOfCollection: realObject.
	^realObject! !

!ObjectTransaction methodsFor: 'registering' stamp: ''!
registerTransientInternalsOfCollection: aCollection
	"If this is a collection, then we may need to register any internal structures it has, e.g. an internal array. This is implementation dependent for the collection. We will also explicitly exclude strings"

	aCollection glorpIsCollection ifFalse: [^self].
	aCollection class isBits ifTrue: [^self].
	(self needsCollectionMementoFor: aCollection)
		ifTrue: [collectionMementos at: aCollection put: aCollection copy].
	aCollection glorpRegisterCollectionInternalsIn: self.

"For this to work, the dialect must not copy collections so shallowly that internals are reused identically.
In VW, Dictionary copy calls postCopy, which copies the associations or value collection (but not the
keys).  Without this, mementos added to the original would also be added to the copy.  In VA, sending
copy to a Dictionary does a shallow copy.  In VA, inline the call of copy to shallowCopy and send the
postCopy call to it explicitly to get the behaviour we need without changing Dictionary copy behaviour
in VA, which might have side-effects.  This feature's tests have been altered to check this copy safety."! !

!ObjectTransaction methodsFor: 'registering' stamp: ''!
registeredObjectsDo: aBlock
	"Iterate over all our objects. Note that this will include objects without descriptors. Be sure we're iterating over a copy of the keys, because this will add objects to the undoMap. Allow the block to return a collection which we will then add to the list of things to process, until there's nothing further"

	| newAdditions previousVersion |
	newAdditions := undoMap keys.
	previousVersion := newAdditions.
	[newAdditions isEmpty]
		whileFalse:
			[newAdditions do: aBlock.
			newAdditions := OrderedCollection new.
			undoMap
				keysDo:
					[:eachKey | (previousVersion includes: eachKey) ifFalse: [newAdditions add: eachKey]].
			newAdditions isEmpty ifFalse: [previousVersion := undoMap keys]].! !

!ObjectTransaction methodsFor: 'registering' stamp: ''!
requiresRegistrationFor: anObject

	| realObject |
	realObject := self realObjectFor: anObject ifNone: [^false].
	^(self isRegistered: realObject) not.! !

!ObjectTransaction methodsFor: 'private/restoring' stamp: ''!
restoreIndexedInstanceVariablesOf: original toThoseOf: copy

	copy glorpStartOfIndexedInstVars to: copy glorpEndOfIndexedInstVars do: [:index |
		original isString 
			ifTrue: [original at: index put: (copy at: index)]
			ifFalse: [original basicAt: index put: (copy basicAt: index)]].! !

!ObjectTransaction methodsFor: 'private/restoring' stamp: ''!
restoreNamedInstanceVariablesOf: original toThoseOf: copy

	1 to: copy class instSize do: [:index |
		original instVarAt: index put: (copy instVarAt: index)]! !

!ObjectTransaction methodsFor: 'private/restoring' stamp: ''!
restoreShapeOf: original toThatOf: copy

	| newOriginal |
	(copy class isBits or: [copy class isVariable])
		ifTrue: [newOriginal := copy class basicNew: copy glorpBasicSize]
		ifFalse: [newOriginal := copy class basicNew].
	original become: newOriginal.! !

!ObjectTransaction methodsFor: 'private/restoring' stamp: ''!
restoreStateOf: original toThatOf: copy 

	(self isShapeOf: original differentThanThatOf: copy) 
		ifTrue: [self restoreShapeOf: original toThatOf: copy].
	self restoreNamedInstanceVariablesOf: original toThoseOf: copy.
	self restoreIndexedInstanceVariablesOf: original toThoseOf: copy! !

!ObjectTransaction methodsFor: 'private/registering' stamp: ''!
shallowCopyOf: anObject ifNotNeeded: aBlock

	| copy |
	copy := anObject shallowCopy.
	^copy == anObject 
		ifTrue: [aBlock value]
		ifFalse: [copy]! !

!ObjectTransaction methodsFor: 'accessing' stamp: ''!
undoMap
	^undoMap! !

Object subclass: #PersistentObject
	instanceVariableNames: 'session'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!PersistentObject commentStamp: '<historical>' prior: 0!
This is an abstract superclass for Glorp persistent objects. Things don't have to subclass from this in order to be persistent, but if they do, they will inherit some methods which may be useful.!

!PersistentObject classSide methodsFor: 'accessing' stamp: ''!
classModel
	^self descriptor classModel! !

!PersistentObject classSide methodsFor: 'finding objects' stamp: ''!
count
	"Count the number of objects in the database"
	"Person count"

	^ self getSession count: self! !

!PersistentObject classSide methodsFor: 'accessing' stamp: ''!
descriptor
	^self getSession descriptorFor: self! !

!PersistentObject classSide methodsFor: 'finding objects' stamp: ''!
find: primaryKey
	"Query the database for an object whose primary key = the given value."
	"Person find: 1"

	| session |
	session := self getSession.
	^session execute: (self query: primaryKey).! !

!PersistentObject classSide methodsFor: 'finding objects' stamp: ''!
findAll
	"Query the database for all objects of this class."
	"Person findAll"
	^self getSession read: self.! !

!PersistentObject classSide methodsFor: 'finding objects' stamp: ''!
findWhere: aBlock
	"Query the database for all objects of this class satisfying aBlock. The block has to be a query block, meaning you can't execute arbitrary code, but rather it will be analyzed and used to generate SQL."
	"Person findWhere: [:each | (each name = 'John Doe') & (each address houseNum = 1000)]"

	^self getSession read: self where: aBlock.! !

!PersistentObject classSide methodsFor: 'session' stamp: ''!
getSession
	"Return the session to be used for the current query. In general, we don't know how to find our session. Callers can catch the CannotFindSession exception and resume it, or subclasses can implement this method to find the session using mechanisms like process environment, a global, or whatever else they have available."
	^CannotFindSession raise.! !

!PersistentObject classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize.! !

!PersistentObject classSide methodsFor: 'finding objects' stamp: ''!
query: primaryKey
	"Return a query object that will query the database for an object whose primary key = the given value. If you want to just return the object, use find: instead. The advantage of this is that you can use this to get a basic query, but then modify it before executing it."
	"Person query: 1"

	| pkAttribute session |
	session := self getSession.
	pkAttribute := (session system descriptorFor: self) primaryKeyMappings first attribute name.
	^Query readOneOf: self where: [:each | (each perform: pkAttribute) = primaryKey].! !

!PersistentObject classSide methodsFor: 'finding objects' stamp: ''!
queryAll
	"Return a query object that will query the database for all objects of this class. If you want to just return the object, use find: instead. The advantage of this is that you can use this to get a basic query, but then modify it before executing it."
	"Person queryAll"
	^Query read: self.! !

!PersistentObject classSide methodsFor: 'finding objects' stamp: ''!
queryWhere: aBlock
	"Return a Query object that will query the database for all objects of this class satisfying aBlock. The block has to be a query block, meaning you can't execute arbitrary code, but rather it will be analyzed and used to generate SQL. If you want to just return the objects, use findWhere: instead. The advantage of this is that you can use this to get a basic query, but then modify it before executing it, e.g. using AND: or OR:."
	"Person queryWhere: [:each | (each name = 'John Doe') & (each address houseNum = 1000)]"

	^Query read: self where: aBlock.! !

!PersistentObject classSide methodsFor: 'accessing' stamp: ''!
table
	^self descriptor ifNotNil: [:descriptor | descriptor table]! !

!PersistentObject methodsFor: 'comparing' stamp: ''!
= anObject
	"Implement equality to compare primary keys if they are set. Make sure that if we have proxies, we check through them."

	| descriptor myKey |
	self class = anObject yourSelf class ifFalse: [^false].
	descriptor := self descriptor.	

	"If the primary keys aren't equal, then we know for sure we are unequal. Otherwise we have to check further."
	myKey := descriptor primaryKeyCollectionFor: self.
	myKey = (descriptor primaryKeyCollectionFor: anObject)
		ifFalse: [^false].

	"If the primary key contains nils, then fall back to the default equality comparison. If they don't, and are equal, then the objects are equal."
	^(myKey includes: nil) 
		ifTrue: [super = anObject yourSelf]
		ifFalse: [true].! !

!PersistentObject methodsFor: 'actions' stamp: ''!
bePersistent
	"Queue our object to be saved to the database upon commit."

	self session
		requireUnitOfWork;
		register: self! !

!PersistentObject methodsFor: 'accessing' stamp: ''!
classModel
	^self descriptor classModel! !

!PersistentObject methodsFor: 'actions' stamp: ''!
commitUnitOfWork
	"Commit the entire unit of work; this may include objects other than myself."

	self session hasUnitOfWork ifFalse: [^self].
	self session commitUnitOfWork! !

!PersistentObject methodsFor: 'actions' stamp: ''!
delete
	"Queue our object to be removed from the database on commit."

	self session delete: self! !

!PersistentObject methodsFor: 'accessing' stamp: ''!
descriptor
	^self session descriptorFor: self class! !

!PersistentObject methodsFor: 'finding objects' stamp: ''!
find
	"This is a query by example facility. We take whichever fields are filled in the object and read back any objects in the database that match."
	
	| instances |
	instances := GlorpVirtualCollection on: self class in: self session.
	self descriptor mappings do: [:eachMapping | 
		instances := self findMapping: eachMapping in: instances].
	^instances.! !

!PersistentObject methodsFor: 'private' stamp: ''!
findMapping: aMapping in: aVirtualCollection
	
	aMapping hasAttribute not ifTrue: [^aVirtualCollection].
	aMapping attribute name = #session ifTrue: [^aVirtualCollection].
	^(aMapping attribute getValueFrom: self)
		ifNil: [aVirtualCollection]
		ifNotNil: [:value | aVirtualCollection select: [:each | self matchAttribute: (each get: aMapping attribute name) to: value]].! !

!PersistentObject methodsFor: 'comparing' stamp: ''!
hash
	"Implement hash to compare primary keys if they are set. Make sure that if we have proxies, we check through them."

	| myKey |

	"If the primary key contains nils, then fall back to the default hash. If they don't, hash the collection."
	myKey := self descriptor primaryKeyCollectionFor: self.
	^(myKey includes: nil) 
		ifTrue: [super hash]
		ifFalse: [myKey hash]! !

!PersistentObject methodsFor: 'private' stamp: ''!
matchAttribute: anExpression to: value
	"We match expressions differently depending on what type they are."
	
	^(value isString and: [value includes: $%]) 
		ifTrue: [anExpression like: value] 
		ifFalse: [anExpression = value].! !

!PersistentObject methodsFor: 'actions' stamp: ''!
refresh
	"Refresh our object from the database and rollback any changes registered in the unit of work."

	self rollback.
	self session refresh: self! !

!PersistentObject methodsFor: 'actions' stamp: ''!
rollback
	"Rollback our entries in the queue so that we are not involved in the commit."

	self session hasUnitOfWork ifFalse: [^self].
	self session rollbackUnitOfWorkFor: self! !

!PersistentObject methodsFor: 'actions' stamp: ''!
rollbackUnitOfWork
	"Rollback the entire unit of work; this may include objects other than myself."

	self session hasUnitOfWork ifFalse: [^self].
	self session rollbackUnitOfWork! !

!PersistentObject methodsFor: 'accessing' stamp: ''!
session
	session isNil ifTrue: [session := self class getSession].
	^session.! !

!PersistentObject methodsFor: 'accessing' stamp: ''!
session: aGlorpSession
	session := aGlorpSession! !

Object subclass: #Query
	instanceVariableNames: 'session prepared expectedRows collectionType ordering maximumLobSizeToRetrieveDirectly readsOneObject groupBy'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!Query commentStamp: '<historical>' prior: 0!
This is an abstract superclass for Glorp queries.

Subclasses must implement the following messages:
	accessing
		readsOneObject
	executing
		executeWithParameters:in:

Instance Variables:
	collectionType	<Class>	What sort of collection should this query return its results in.
	expectedRows	<Integer>	Roughly how many rows should we expect to get back. Doesn't have to be right, but can be used to set buffer sizes appropriately in the database.
	ordering	<(SequenceableCollection of: (GlorpExpression | Block | Symbol))>	If we order the results, what are the ordering criteria. Either an expression, but one whose root ends up evaluating to a field, or a block or symbol that can be converted to same.
	prepared	<Boolean>	Has this query already been prepared (at the Glorp level), this says nothing about prepared statements for the database, which will be separate objects.
	session	<GlorpSession>	What session should this query execute in. Can be nil until we start to prepare the query.
!

Query subclass: #AbstractReadQuery
	instanceVariableNames: 'resultClass whereClause base limit offset proxyType shouldRefresh tracing'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!AbstractReadQuery commentStamp: '<historical>' prior: 0!
An abstract superclass for Glorp queries. Most query types will descend from this, but a couple are specialized.

Subclasses must implement the following messages:
	executing
		rowsFromDatabaseWithParameters:

Instance Variables:
	base	<BaseExpression>	In Glorp expressions it's very important that expressions we're trying to compose together are built on the same base, which represents the object that we're talking about in the query. This holds the base expression that we're expecting to use in this query.
	limit	<Integer>	Some databases support limiting the number of rows retrieved. If we can, we'll use this variable to set that.
	offset	<Integer>	Some databases support starting from a given point in the collection of retrievable rows, often used in combination with limit. So, e.g. get me the rows, sorted by name, starting at the 1012th and giving me 40 of them.
	proxyType	<Class | nil>	If this is nil, do not use proxies. Otherwise, use this class to create proxies for relationships. At the moment, likely to be either Proxy or TimedProxy.
	readsOneObject	<Boolean>	Do we return only one object, or a collection. That is, is this a readManyOf: or a readOneOf:
	resultClass	<Behavior>	What kind of objects are we looking for.
	shouldRefresh	<Boolean>	If the result is already in cache, should we refresh it according to what the database says, or just return it as is.
	tracing	<Tracing>	The tracing. This describes the graph of objects we will actually retrieve.
	whereClause	<GlorpExpression>	What is the criteria on which we will filter the results.

!

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
newQuery
	^SimpleQuery new.! !

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
read: aClass limit: anInteger
	^self returningManyOf: aClass limit: anInteger.! !

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
read: aClass where: anExpression limit: anInteger
	^self
		returningManyOf: aClass
		where: anExpression
		limit: anInteger.! !

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
readManyOf: aClass limit: anInteger
	^self returningManyOf: aClass limit: anInteger.! !

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
readManyOf: aClass where: anExpression limit: anInteger
	^self
		returningManyOf: aClass
		where: anExpression
		limit: anInteger.! !

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
returningManyOf: aClass
	^self returningManyOf: aClass where: nil.! !

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
returningManyOf: aClass limit: anInteger
	^self
		returningManyOf: aClass
		where: nil
		limit: anInteger.! !

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
returningManyOf: aClass where: anExpression
	^self returningManyOf: aClass where: anExpression limit: nil.! !

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
returningManyOf: aClass where: anExpression limit: anInteger
	^(self newQuery)
		limit: anInteger;
		initResultClass: aClass
			whereClause: anExpression
			singleObject: false;
		yourself.! !

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
returningOneOf: aClass
	^self returningOneOf: aClass where: nil.! !

!AbstractReadQuery classSide methodsFor: 'instance creation' stamp: ''!
returningOneOf: aClass where: anExpression
	^self newQuery
		initResultClass: aClass
		whereClause: anExpression
		singleObject: true.! !

!AbstractReadQuery methodsFor: 'boolean operations' stamp: ''!
AND: anExpression
	"Allow you to send AND: or OR: directly to a query to build up a query dynamically without needing to mess with the criteria explicitly"

	whereClause := (anExpression
		asGlorpExpressionOn: base) AND: ((whereClause asGlorpExpressionOn: base) asGeneralGlorpExpression).! !

!AbstractReadQuery methodsFor: 'boolean operations' stamp: ''!
OR: anExpression
	"Allow you to send AND: or OR: directly to a query to build up a query dynamically without needing to mess with the criteria explicitly"

	whereClause := (anExpression asGlorpExpressionOn: base)
		OR: (whereClause asGlorpExpressionOn: base) asGeneralGlorpExpression.! !

!AbstractReadQuery methodsFor: 'preparing' stamp: ''!
adjustWhereClause
	^self.! !

!AbstractReadQuery methodsFor: 'specifying retrievals' stamp: ''!
alsoFetch: anExpression
	self tracing alsoFetch: (self expressionBlockFor: anExpression).! !

!AbstractReadQuery methodsFor: 'converting' stamp: ''!
asFullQuery
	^self.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
baseExpression
	^base.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
baseExpression: aBaseExpression
	base := aBaseExpression.! !

!AbstractReadQuery methodsFor: 'caching' stamp: ''!
checkCacheWithParameters: aDictionary ifPresent: aBlock
	| primaryKey result filteredResult |
	readsOneObject ifFalse: [^nil].
	self shouldRefresh ifTrue: [^nil].
	"Check if we have retrieval expressions for something other than the base. If so, we're not actually reading this object, or at least not just this object"
	self tracing retrievalExpressions do: [:each | each base isNil ifFalse: [^self]].
	primaryKey := self primaryKeyFrom: aDictionary.
	"We can't figure out the primary key"
	primaryKey isNil ifTrue: [^nil].
	"The primary key is invalid (i.e. has nulls in it). So we've found the entry, and we know it's nil"
	primaryKey = #() ifTrue: [^aBlock value: nil].
	"If it's expired, make sure we do the read but still refresh"
  	(session hasExpired: resultClass key: primaryKey) ifTrue: [
		self shouldRefresh: true.
		^nil].
	result := session 
		cacheAt: primaryKey
		forClass: resultClass
		ifNone: [nil].
	filteredResult := session filterDeletionFrom: result.
	^filteredResult isNil ifTrue: [nil] ifFalse: [
		(session system shouldResetExpiryTimeForObjectOfClass: filteredResult class readFrom: self)
			ifTrue: [self session markAsCurrentOfClass: filteredResult class key: primaryKey].
		aBlock value: filteredResult].! !

!AbstractReadQuery methodsFor: 'validation' stamp: ''!
checkValidity
	resultClass glorpCanHaveDescriptorDirectly ifFalse: [self error: 'resultClass must be a class'].
	self descriptor isNil ifTrue: [self error: 'No descriptor for ', self resultClass name].
	(ordering notNil and: [self descriptor supportsOrdering not]) ifTrue: [self error: 'The descriptor for ', self resultClass name, ' does not support ordering in queries'].! !

!AbstractReadQuery methodsFor: 'copying' stamp: ''!
copyAttributesTo: newQuery
	"We're creating a new query similar to ourselves, but not a direct copy. Copy over the relevant attributes."

	super copyAttributesTo: newQuery.
	newQuery limit: self limit.
	newQuery offset: self offset.
	newQuery proxyType: self proxyType.
	newQuery shouldRefresh: self shouldRefresh.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
defaultTracing
	| defaultTracing |
	defaultTracing := Tracing new.
	defaultTracing base: base.
	^defaultTracing.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
descriptor
	^session descriptorFor: resultClass.! !

!AbstractReadQuery methodsFor: 'boolean operations' stamp: ''!
except: aQuery
	^self minus: aQuery.! !

!AbstractReadQuery methodsFor: 'executing' stamp: ''!
executeWithParameters: parameterArray in: aSession 
	| |
	session := aSession.
	self requiresFullQuery 
		ifTrue: [^self asFullQuery executeWithParameters: parameterArray in: aSession].
	self checkValidity.
	self setUpExpressions.
	self checkCacheWithParameters: parameterArray ifPresent: [:hit | session register: hit. ^hit].
	self setupTracing.
	^self 
		readFromDatabaseWithParameters: parameterArray.! !

!AbstractReadQuery methodsFor: 'testing' stamp: ''!
hasLimit
	^self limit notNil.! !

!AbstractReadQuery methodsFor: 'testing' stamp: ''!
hasOffset
	^self offset notNil and: [self offset > 0].! !

!AbstractReadQuery methodsFor: 'testing' stamp: ''!
hasTracing
	"Return true if we've given this query a tracing already"
	^false.! !

!AbstractReadQuery methodsFor: 'initialize' stamp: ''!
initResultClass: aClass whereClause: theCriteria singleObject: aBoolean 
	resultClass := aClass.
	self whereClause: ((theCriteria isNil or: [theCriteria = true or: [theCriteria = false]]) ifTrue: [EmptyExpression on: theCriteria] ifFalse: [theCriteria]).
	readsOneObject := aBoolean.
	aBoolean ifTrue: [self limit: 1].! !

!AbstractReadQuery methodsFor: 'initialize' stamp: ''!
initialize
	
	super initialize.
	proxyType := nil.
	shouldRefresh := false.! !

!AbstractReadQuery methodsFor: 'boolean operations' stamp: ''!
intersect: aQuery
	aQuery shortCircuitEmptyReturn ifTrue: [^aQuery].
	self shortCircuitEmptyReturn ifTrue: [^self].
	^CompoundQuery newIntersect addQuery: self; addQuery: aQuery.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
limit
	^limit.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
limit: anInteger
	limit := anInteger.! !

!AbstractReadQuery methodsFor: 'testing' stamp: ''!
limitMakesSense
	^self hasLimit not or: [self limit > 0].! !

!AbstractReadQuery methodsFor: 'boolean operations' stamp: ''!
minus: aQuery
	aQuery shortCircuitEmptyReturn ifTrue: [^self].
	self shortCircuitEmptyReturn ifTrue: [^aQuery].
	^CompoundQuery newMinus addQuery: self; addQuery: aQuery.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
offset
	^offset! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
offset: anObject
	offset := anObject.! !

!AbstractReadQuery methodsFor: 'copying' stamp: ''!
postCopy
	super postCopy.
	base := BaseExpression new.
	whereClause := whereClause asGlorpExpressionOn: base.
	session isNil ifFalse: [self setUpExpressions].! !

!AbstractReadQuery methodsFor: 'preparing' stamp: ''!
prepare
	prepared ifTrue: [^self].
	self setupTracing.
	prepared := true.! !

!AbstractReadQuery methodsFor: 'caching' stamp: ''!
primaryKeyFrom: aDictionary
	"Construct a primary key from the given parameters."

	self prepare.
	(aDictionary
		anySatisfy:
			[:each | self session system shouldForeignKeyValueBeTreatedAsNull: each])
		ifTrue: [^#()].
	^self whereClause primaryKeyFromDictionary: aDictionary.! !

!AbstractReadQuery methodsFor: 'private' stamp: ''!
privateAnd: anExpression
	"AND these two together without checking for common bases, rebuilding, etc. Everything had better be right"
	whereClause := whereClause AND: anExpression.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
proxyType
	"Answer the type of proxies the receiver should build when instantiating objects:
		nil - Do not build proxies.
		TimedProxy - Build TimedProxies.
		Proxy - Build Proxies."

	^proxyType! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
proxyType: aClass
	
	proxyType := aClass.! !

!AbstractReadQuery methodsFor: 'executing' stamp: ''!
readFromDatabaseWithParameters: anArray
	| answerStream result |
	answerStream := self rowsFromDatabaseWithParameters: anArray.
	^readsOneObject
		ifTrue: [result := answerStream next. answerStream release. result]
		ifFalse: [self resultCollectionFor: answerStream].! !

!AbstractReadQuery methodsFor: 'testing' stamp: ''!
requiresFullQuery
	^false.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
resultClass
	^resultClass! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
resultType
	"Return the database type that we are expecting to return from this query. This only makes sense if we're returning a single field. If we return a collection of types, just return nil here. If we return a mapped object, the #type message will cause an exception."
	self tracing retrievalExpressions size = 1 ifFalse: [^nil].
	^self tracing retrievalExpressions asArray first type.! !

!AbstractReadQuery methodsFor: 'specifying retrievals' stamp: ''!
retrieve: anExpression
	self tracing retrieve: (self expressionBlockFor: anExpression).! !

!AbstractReadQuery methodsFor: 'specifying retrievals' stamp: ''!
retrieveMax: anExpression
	self tracing retrieve: ((self expressionBlockFor: anExpression) asGlorpExpression get: #max).! !

!AbstractReadQuery methodsFor: 'specifying retrievals' stamp: ''!
retrieveMin: anExpression
	self tracing retrieve: ((self expressionBlockFor: anExpression) asGlorpExpression get: #min).! !

!AbstractReadQuery methodsFor: 'specifying retrievals' stamp: ''!
retrieveSum: anExpression
	self tracing retrieve: ((self expressionBlockFor: anExpression) asGlorpExpression get: #sum).! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
returnProxies
	^proxyType notNil! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
returnProxies: aBoolean
	
	aBoolean ifTrue: [proxyType := Proxy] ifFalse: [proxyType := nil].! !

!AbstractReadQuery methodsFor: 'testing' stamp: ''!
returnTimedProxies
	^proxyType == TimedProxy! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
returnTimedProxies: aBoolean

	aBoolean ifTrue: [proxyType := TimedProxy] ifFalse: [proxyType := nil]! !

!AbstractReadQuery methodsFor: 'executing' stamp: ''!
rowsFromDatabaseWithParameters: anArray
	self subclassResponsibility.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
setUpBase
	whereClause isGlorpExpression ifTrue: [base := whereClause queryLevelBaseExpression] ifFalse: [base := BaseExpression new].! !

!AbstractReadQuery methodsFor: 'executing' stamp: ''!
setUpExpressions
	(whereClause isGlorpExpression and: [whereClause queryLevelBaseExpression notNil])
		ifTrue:
			[base := whereClause queryLevelBaseExpression.
			base hasDescriptor
				ifFalse: [base descriptor: (session descriptorFor: resultClass)]]
		ifFalse:
			[base isNil ifTrue: [base := BaseExpression new].
			base descriptor: (session descriptorFor: resultClass).
			whereClause := whereClause asGlorpExpressionOn: base in: self].
	ordering isNil
		ifFalse:
			[ordering := ordering
				collect:
					[:each | (self expressionBlockFor: each) asGlorpExpressionOn: base in: self]].
	groupBy isNil ifFalse: [groupBy := (self expressionBlockFor: groupBy) asGlorpExpressionOn: base in: self].
	tracing isNil ifFalse: [tracing updateBase: base].
	base descriptor typeResolver setUpExpressionsForQuery: self.! !

!AbstractReadQuery methodsFor: 'tracing' stamp: ''!
setupTracing
	self setUpExpressions.
	self tracing setup.
	self tracing additionalExpressions do: [:each |
		each hasDescriptor ifTrue: [
			each descriptor trace: self tracing context: each].
			each validate].! !

!AbstractReadQuery methodsFor: 'testing' stamp: ''!
shortCircuitEmptyReturn
	"If we have a literal false for criteria, we never need to go to the database.
	We also don't need to go to database if limit doesnt make sense in terms of knowing beforehand that no results will be returned"

	(whereClause class == EmptyExpression and: [whereClause isFalse]) ifTrue: [^true].
	self limitMakesSense ifFalse: [^true].
	^false.! !

!AbstractReadQuery methodsFor: 'testing' stamp: ''!
shortCircuitEmptyReturn: parameters
	"If we have a literal false for criteria, we never need to go to the database.
	We also don't need to go to database if limit doesnt make sense in terms of knowing beforehand that no results will be returned"

	(whereClause class == EmptyExpression and: [whereClause isFalse])
		ifTrue: [^true].
	self limitMakesSense ifFalse: [^true].
	self prepare.
	(whereClause shortCircuitOnNullComparison: parameters) ifTrue: [^true].
	^false.! !

!AbstractReadQuery methodsFor: 'preparing' stamp: ''!
shouldExpressionBeJoin: anExpression
	"Return true if the expression should be a join in our context (which means it refers only to tables we have in our from list) or if it should be ANDed into the where clause). Only relevant if we're using ANSI joins"
	^anExpression
		allSatisfy:
			[:eachNode | 
			| queryBase |
			queryBase := eachNode queryLevelBaseExpression.
			queryBase isNil or: [	queryBase = self baseExpression]].! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
shouldRefresh
	^shouldRefresh! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
shouldRefresh: aBoolean
	shouldRefresh := aBoolean! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
tracing
	tracing isNil ifTrue: [tracing := self defaultTracing].
	^tracing.! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
tracing: aTracing
	tracing := aTracing.
	tracing updateBase: base.
	tracing setup.! !

!AbstractReadQuery methodsFor: 'boolean operations' stamp: ''!
unionAll: aQuery
	aQuery shortCircuitEmptyReturn ifTrue: [^self].
	self shortCircuitEmptyReturn ifTrue: [^aQuery].
	^CompoundQuery newUnionAll addQuery: self; addQuery: aQuery.! !

!AbstractReadQuery methodsFor: 'executing' stamp: ''!
validateWhereClause
	resultClass glorpCanHaveDescriptorDirectly ifFalse: [self error: 'Invalid result class for query'].
	whereClause isJoin ifFalse: [
		whereClause do: [:each | each validate]].
	tracing isNil ifTrue: [^self].
	tracing allTracings do: [:eachTracing |
		eachTracing do: [:eachExpressionNode | eachExpressionNode validate]].! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
whereClause
	^whereClause! !

!AbstractReadQuery methodsFor: 'accessing' stamp: ''!
whereClause: anExpression
	whereClause := anExpression.
	self setUpBase.! !

AbstractReadQuery subclass: #ReadQuery
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!ReadQuery commentStamp: '<historical>' prior: 0!
This represents a general read query. By general we mean that it might require more than one trip to the database. It computes a "tracing" indicating which groups of objects can be read simultaneously, then constructs a group of corresponding SimpleQuery instances and executes them.!

!ReadQuery methodsFor: 'converting' stamp: ''!
asQueryWithNullsForOtherClasses: aClass
	"Construct a specialized form of the query that's set up to retrieve instances of aClass, but with a bunch of empty fields that get populated with nulls, so we can retrieve several different classes at the same go."

	| newQuery nullField fieldsToIgnore markerExpression fieldsMinusTheConstant pseudoMappingName newBase tablesToIgnore |
	newQuery := self asSimpleQueryFor: aClass.	"We need to rebuild the where clause to refer to one of the subclass pseudo-mappings, because that's what we're retrieving. We don't want it against the actual class we're querying for, because we need the fields to be exactly the same for each of the sub-queries."
	pseudoMappingName := self descriptor typeResolver
		pseudoAttributeNameForClass: aClass.
	newBase := newQuery baseExpression get: pseudoMappingName.
	newBase bePhantom.
	newQuery where: (newQuery whereClause rebuildOn: newBase).	
	"We don't want it to try and retrieve the parent class. Instead, we alsoFetch: each of the subclasses, and insert a special builder that will select from among the valid results. To prevent it from trying to do anything with the main class itself, remove that from the tracing, and replace it with a special constant marker that we know will generate only one field, that we can remove later."
	newQuery tracing retrievalExpressions remove: newQuery tracing base.
	newQuery tracing allTracings remove: newQuery tracing base.
	markerExpression := ConstantExpression for: 'ignore me'.
	newQuery tracing retrievalExpressions add: markerExpression.
	newQuery tracing allTracings add: markerExpression.
	newQuery prepare.
	newQuery
		privateReplaceBuilderWith:
			(SelectingElementBuilder new expression: markerExpression).	
	"For each retrieval that isn't our class, change the fields being returned to nulls, and wrap a cast around them so that SQL doesn't complain. First, we figure out which fields those are."
	fieldsToIgnore := IdentityDictionary new.
	newQuery tracing alsoFetchExpressions
		do:
			[:each | 
			each mapping referenceClass = aClass
				ifFalse:
					[each fieldAliases
						do: [:eachFieldAlias | fieldsToIgnore at: eachFieldAlias put: each]]].	
	"Now we remove the constant field we know we caused to be generated earlier, and replace the others that we don't want with nulls"
	fieldsMinusTheConstant := newQuery fields
		copyFrom: 1
		to: newQuery fields size - 1.
	newQuery
		privateSetFields:
			(fieldsMinusTheConstant
				collect:
					[:eachField | 
					(fieldsToIgnore includesKey: eachField)
						ifFalse: [eachField]
						ifTrue:
							[nullField := self nullFieldFor: eachField.
							nullField]]).	
	"Supress all tables except the one(s) we actually want."
	newQuery prepare.
	tablesToIgnore := (fieldsToIgnore keys collect: [:each | each table]) asSet asArray.
	tablesToIgnore := tablesToIgnore, (newQuery baseExpression tablesToPrint).
	newQuery tablesToPrint: (newQuery tablesToPrint reject: [:each | tablesToIgnore includes: each]) asArray.	
	^newQuery.! !

!ReadQuery methodsFor: 'converting' stamp: ''!
asSimpleQueryFor: aClass
	| newQuery newCriteria newBase |
	"Rebuild the expression, because this means a full query is being split into multiple sub-queries, e.g. for an inheritance read. The expression may get prepared differently in each case (e.g. table aliases), so we can't share"
	newBase := BaseExpression new.
	newBase descriptor: (session system descriptorFor: aClass).
	newCriteria := whereClause rebuildOn: newBase startingFrom: newBase withOuterScopeBase: newBase.
	newQuery := SimpleQuery new
		initResultClass: aClass
		whereClause: newCriteria
		singleObject: readsOneObject.
	self copyAttributesTo: newQuery.
	newQuery session: session.
	newQuery setOrdering: ordering.
	newQuery groupBy: groupBy.
	newQuery setUpExpressions.
	newQuery tracing: tracing copy.
	^newQuery.! !

!ReadQuery methodsFor: 'executing' stamp: ''!
executeSubQueries: subQueries withParameters: anArray
	^self shouldUseSingleQuery
		ifTrue: [self executeSubQueriesCombined: subQueries withParameters: anArray]
		ifFalse: [self executeSubQueriesIndividually: subQueries withParameters: anArray].! !

!ReadQuery methodsFor: 'executing' stamp: ''!
executeSubQueriesCombined: subQueries withParameters: anArray
	"Execute out sub queries as a single query with a UNION ALL between the sub-queries. Then select from the results the one which has a value for that row. Used for horizontal inheritance queries."
	| combinedQuery |
	combinedQuery := CompoundQuery newUnionAll.
	self copyAttributesTo: combinedQuery.
	combinedQuery requiresDistinct: false.
	subQueries 	do: 	[:eachQuery | 
		combinedQuery addQuery: eachQuery].
	^combinedQuery executeWithParameters: anArray in: session.! !

!ReadQuery methodsFor: 'executing' stamp: ''!
executeSubQueriesIndividually: subQueries withParameters: anArray
	| result |
	result := OrderedCollection new.
	subQueries 	do: 	[:eachQuery | 
		| individualResult |
		individualResult := eachQuery readFromDatabaseWithParameters: anArray.
		eachQuery readsOneObject
			ifTrue: [individualResult isNil ifFalse: [result add: individualResult]]
			ifFalse: [result addAll: individualResult]].
	^self readsOneObject ifTrue: [result first] ifFalse: [result].! !

!ReadQuery methodsFor: 'converting' stamp: ''!
nullFieldFor: eachField
	self platform requiresCastsForNullsInUnions ifFalse: [^(ConstantExpression for: nil) alias: eachField table name, 'nullField', eachField position printString].
	^FunctionExpression
		for: #cast:
		withArguments: (Array with: (ConstantExpression for: eachField type))
		basedOn: (ConstantExpression for: nil).! !

!ReadQuery methodsFor: 'executing' stamp: ''!
readFromDatabaseWithParameters: anArray 
	| subQueries |
	subQueries := self descriptor classesRequiringIndependentQueries 
		collect: [:eachClass | 
			self shouldUseSingleQuery 
				ifTrue: [self asQueryWithNullsForOtherClasses: eachClass]
				ifFalse: [self asSimpleQueryFor: eachClass]].
	^self executeSubQueries: subQueries withParameters: anArray.! !

!ReadQuery methodsFor: 'accessing' stamp: ''!
shouldUseSingleQuery
	^self descriptor typeResolver shouldUseSingleQuery.! !

AbstractReadQuery subclass: #SimpleQuery
	instanceVariableNames: 'builders fields joins tablePrefix customCommand tablesToPrint'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!SimpleQuery commentStamp: '<historical>' prior: 0!
This is a query that is directly executable. A single query might be more than we can do in a single database read, so we might have to break it down into simple queries. But at the moment we just break anything down into an equivalent single query.

Instance Variables:
	builders	<OrderedCollection of: ElementBuilder)>	The builders that will assemble the object from the row that this query returns.
	fields	<OrderedCollection of: DatabaseField>	The fields being selected.
	joins	<OrderedCollection of: Join> While preparing the query, we will need to build up a list of joins implied by the relationships in the where clause, ordering, and so forth. This acts as storage for those.
	tablePrefix	<String>	If we are part of a subselect, we want to prefix all of our table aliases with some known, common prefix, so that we won't have our own t1 which would get confused with the parent query's t1. The normal convention is something like 's1' for subselect 1, so we get s1t1, and so forth.
	command <DatabaseCommand> We can tell the query to use a specific command instead of generating one appropriately. Note that if this is set, it overrides most of the other sql generation.

!

SimpleQuery subclass: #SQLite3PragmaQuery
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!SQLite3PragmaQuery commentStamp: '<historical>' prior: 0!
This is a query intended to execute a SQLite3 PRAGMA command. The command behaves like a stored function that takes a parameter and returns a result set. The meta information schema for SQLite is not available in the standard views, but it can be eventually deduced by executing particular pragma commands.
!

!SQLite3PragmaQuery classSide methodsFor: 'instance creation' stamp: ''!
newQuery
	^SQLite3PragmaQuery new
	"^super newQuery"! !

!SQLite3PragmaQuery classSide methodsFor: 'instance creation' stamp: ''!
returningManyOf: aClass where: anExpression
	"Backward-compatibility, since we changed the class name."
	^self returningManyOf: aClass where: anExpression limit: nil! !

!SQLite3PragmaQuery methodsFor: 'sql generation' stamp: ''!
printWhereClauseOn: aCommand

	self hasEmptyWhereClause
		ifFalse: 
			[whereClause rightChild printSQLOn: aCommand withParameters: aCommand parameters]! !

!SQLite3PragmaQuery methodsFor: 'sql generation' stamp: ''!
sqlWith: aDictionary
	| result |
	self prepare.
	customCommand isNil ifFalse: [
		result := customCommand copy.
		result query: self; parameters: aDictionary; useBinding: session useBinding; session: session.
		^result].
	^SQLite3PragmaCommand
		forQuery: self
		parameters: aDictionary
		useBinding: session useBinding
		session: session.! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
addAllOrderingsForBuilder: aBuilder
	
	| expression fieldExpressions |
	expression := aBuilder expression.
	fieldExpressions := expression descriptor primaryTable primaryKeyFields collect: [:each | expression getField: each].
	fieldExpressions do: [:each | self addSingleOrderingForBuilder: each].! !

!SimpleQuery methodsFor: 'accessing' stamp: ''!
addBoundExpressionsIn: aCommand to: aCollection
	self joins , self fields, (Array with: self whereClause)
		do:
			[:eachBigExpression | 
			eachBigExpression isGlorpExpression
				ifTrue: [eachBigExpression do:
					[:eachIndividualExpressionNode | 
					eachIndividualExpressionNode
						addBoundExpressionsIn: aCommand
						to: aCollection]]].! !

!SimpleQuery methodsFor: 'fields' stamp: ''!
addFields: aliasedFields returningTranslationForFields: originalFields distinct: isDistinct
	"The query has computed a set of fields the way the mappings see them, which are then transformed to account for field aliasing in embedded mappings. Add those to our collection, and set up the translation which knows which fields are at which index in the resulting row. If necessary, note that those fields are selected as distinct"

	| translation |
	translation := IdentityDictionary new.
	aliasedFields
		with: originalFields
		do:
			[:aliased :original | 
			| position fieldToAdd|
			position := self findPositionofAliasedField: aliased.
			position = 0
				ifTrue:  
					[fieldToAdd := isDistinct ifTrue: [aliased asDistinctFieldOn: self baseExpression] ifFalse: [aliased].
					fields add: fieldToAdd .
					position := fields size].
			translation at: original put: position].
	^translation.! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
addJoin: anExpression
	joins addAll: anExpression asIndependentJoins.! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
addOrderingForToManyJoins
	"If we have a join to a to-many relation, then we need for all the joined bits to be grouped together in order in the result set"
	| problemBuilders baseBuilders |
	problemBuilders := builders select: [:each | each canCauseDuplicateRows].
	baseBuilders := problemBuilders
		collect:
			[:each | 
			builders
				detect: [:eachBuilder | eachBuilder expression == each expression base]
				ifNone: [nil]].
	baseBuilders do: [:each | 
		each isNil ifFalse: [self addAllOrderingsForBuilder: each]].! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
addSingleOrderingForBuilder: aFieldExpression
	"Add the ordering to the query, as long as we don't already have an equivalent ordering there."	

	ordering isNil ifTrue: [^self orderBy: aFieldExpression].
	ordering 
		detect: [:eachExistingOrdering | 
			eachExistingOrdering = aFieldExpression 
				or: [eachExistingOrdering field = aFieldExpression field and: [eachExistingOrdering base = aFieldExpression base]]]
		ifNone: [self orderBy: aFieldExpression].! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
adjustWhereClause
	"Make any adjustments to the where clause. This includes adjusting direct object equality expressions to be primary key comparisons, removing primary key comparisons where we can compare a foreign key, and condensing boolean constants out of expressions."

	whereClause := whereClause rewriteEquality. "Special because otherwise we'd be throwing away the top-level one"
	whereClause := whereClause rewriteBooleanEquality: nil.
	whereClause do: [:each | each rewriteEqualityExpressions].
	whereClause do: [:each | each condensePrimaryKeyComparisons].
	whereClause := whereClause condensePrimaryKeyComparison.
	whereClause := whereClause condenseBooleanEquality.
	whereClause do: [:each | each condenseBooleanEqualityExpressions].
	(whereClause isConstantExpression and: [whereClause value = true])
		ifTrue: [
			whereClause := EmptyExpression on: whereClause value.
			whereClause base: base.].
	whereClause do: [:each | each rewriteFunctionArguments].! !

!SimpleQuery methodsFor: 'converting' stamp: ''!
asFullQuery
	| newQuery |
	newQuery := ReadQuery new
		initResultClass: resultClass whereClause: whereClause singleObject: readsOneObject.
	self copyAttributesTo: newQuery.
	newQuery setOrdering: ordering.
	newQuery groupBy: groupBy.
	^newQuery.! !

!SimpleQuery methodsFor: 'converting' stamp: ''!
asGlorpExpression
	^self asGlorpExpressionOn: BaseExpression new.! !

!SimpleQuery methodsFor: 'converting' stamp: ''!
asGlorpExpressionOn: anExpression
	"If we're being converted into an expression, that means we're a subselect inside an expression block. We don't actually have any relationship to anExpression, so create our own base and make a pseudo-mapping with no join to describe the relationship. Don't even give it a string-name, just use ourselves as the key"
	
	| expression localBase phantomMapping phantomExpression |
	phantomMapping := OneToOneMapping new
		referenceClass: self resultClass;
		attributeName: #temporaryAttributeFromSubselect;
		join: Join new;
		descriptor: (anExpression system descriptorFor: self resultClass).
	phantomExpression := anExpression getMapping: phantomMapping named: self.
	expression := GeneralSubSelectExpression 
		basedOn: phantomExpression.
	expression query: self.
	localBase := SubSelectBaseExpression on: phantomExpression.
	self baseExpression: localBase.
	self tablePrefix: ('s', anExpression ultimateBaseExpression availableSubSelectNumber printString, 't').
	^expression.! !

!SimpleQuery methodsFor: 'converting' stamp: ''!
asGlorpExpressionOn: anExpression basedOn: aMessageArchiver withUltimateBase: aBaseExpression
	^self asGlorpExpressionOn: aBaseExpression.! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
assignTableAliases
	| tableNumber allExpressions |
	whereClause isJoin ifTrue: [^self].
	tableNumber := 1.
	allExpressions := ExpressionGroup with: whereClause.
	allExpressions addAll: ordering.
	allExpressions add: groupBy.
	allExpressions addAll: joins.
	builders do: [:each | allExpressions add: each expression].
	allExpressions
		do:
			[:each | tableNumber := each assignTableAliasesStartingAt: tableNumber prefix: self tablePrefix].
	"We've aliased tables, joins may now become independent"
	joins := joins inject: OrderedCollection new into: [:sum :each | (sum addAll: each asIndependentJoins). sum].! !

!SimpleQuery methodsFor: 'converting' stamp: ''!
basicRebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	"Rebuild the query, which is part of a subselect expression, into the new context. The only thing that really matters here is that we get correlated subselect pieces inside the query right. The rest is independent of our main expression"
	prepared := false.
	builders := nil.
	base := SubSelectBaseExpression on: (base trueBase rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression).
	whereClause := whereClause asGlorpExpressionOn: base withUltimateBase: ultimateBaseExpression.
	joins := joins collect: [:each | each asGlorpExpressionOn: base withUltimateBase: ultimateBaseExpression].
	ordering isNil ifFalse: [ordering := ordering collect: [:each | each asGlorpExpressionOn: base withUltimateBase: ultimateBaseExpression]].
	groupBy isNil ifFalse: [groupBy := groupBy asGlorpExpressionOn: base withUltimateBase: ultimateBaseExpression].
"	fields := fields collect: [:each | each isGlorpExpression 
		ifTrue: [each asGlorpExpressionOn: base withUltimateBase: ultimateBaseExpression]
		ifFalse: [each]]."
	fields := OrderedCollection new.! !

!SimpleQuery methodsFor: 'accessing' stamp: ''!
builders

	^builders.! !

!SimpleQuery methodsFor: 'accessing' stamp: ''!
command: aDatabaseCommand
	"Explicitly set the database command we will use. This is normally used in order to provide a SQLStringSelectCommand, and specify the sql for the query."
	customCommand := aDatabaseCommand.! !

!SimpleQuery methodsFor: 'executing' stamp: ''!
computeFields
	builders 
		do: [:each | self computeFieldsFor: each]! !

!SimpleQuery methodsFor: 'executing' stamp: ''!
computeFieldsFor: anElementBuilder 
	| translatedFields |
	translatedFields := self 
		addFields: anElementBuilder fieldsForSelectStatement
		returningTranslationForFields: anElementBuilder fieldsFromMyPerspective
		distinct: anElementBuilder requiresDistinct.
	anElementBuilder fieldTranslations: translatedFields.! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
computeTablesToPrint
	| allTables stoppingPoint |
	allTables := Set new.
	fields do: [:each | allTables addAll: each tablesToPrint].
	stoppingPoint := base.
	allTables
		addAll:
			(stoppingPoint isNil
				ifTrue: [whereClause allTablesToPrint]
				ifFalse: [whereClause allTablesToPrintStoppingAt: stoppingPoint]).
	joins
		do:
			[:eachJoin | 
			allTables
				addAll:
					(stoppingPoint isNil
						ifTrue: [eachJoin allTablesToPrint]
						ifFalse: [eachJoin allTablesToPrintStoppingAt: stoppingPoint])].
	ordering isNil
		ifFalse: [ordering do: [:each | 
			each field isNil ifFalse: [allTables add: each field table]]].
	(groupBy notNil and: [groupBy field notNil])
		ifTrue: [allTables add: groupBy field table].
	self tracing allTracings
		do: [:each | allTables addAll: each allTablesToPrint].
	^allTables asSortedCollection.! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
copyOfBuilders 
	"Copy the builders, but make sure that subBuilders also get matched up"
	| newBuilders |
	newBuilders := builders collect: [:each | each copy].
	newBuilders do: [:each | each reconcileWithBuilders: builders andCopy: newBuilders].
	^newBuilders.! !

!SimpleQuery methodsFor: 'accessing' stamp: ''!
elementBuilderFor: anExpression 
	^builders detect: [:each | each expression == anExpression] ifNone: [nil].! !

!SimpleQuery methodsFor: 'accessing' stamp: ''!
fields
	fields isNil ifTrue: [self prepare].
	^fields! !

!SimpleQuery methodsFor: 'fields' stamp: ''!
findPositionofAliasedField: aField
	1 to: fields size do: [:i |
		| existingField |
		existingField := fields at: i.
		(existingField asField isSameAliasedFieldAs: aField)
			ifTrue: [^i]].
	^0.! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
fixJoins
	| pseudoJoins realJoins |
	pseudoJoins := joins select: [:each | each tablesForANSIJoin size < 2].
	pseudoJoins do: [:each | whereClause := each AND: whereClause].
	realJoins :=  joins select: [:each| each tablesForANSIJoin size >= 2].
	joins := realJoins.! !

!SimpleQuery methodsFor: 'testing' stamp: ''!
hasEmptyWhereClause
	"If we have regular where clause entries, or if we have joins that aren't going to be printed in the tables portion, then we're not empty"
	whereClause willPrintAsWhereClause ifTrue: [^false].
	self useANSIJoins ifTrue: [^true].
	^joins isEmpty.! !

!SimpleQuery methodsFor: 'testing' stamp: ''!
hasTracing
	"Return true if we've given this query a tracing already"
	^builders notNil.! !

!SimpleQuery methodsFor: 'initialize' stamp: ''!
initResultClass: aClass whereClause: anExpression singleObject: aBoolean 
	super 
		initResultClass: aClass
		whereClause: anExpression
		singleObject: aBoolean.
	prepared := false.
	fields := OrderedCollection new.
	joins := OrderedCollection new.! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
isPrepared
	^prepared.! !

!SimpleQuery methodsFor: 'accessing' stamp: ''!
joins
	^joins.! !

!SimpleQuery methodsFor: 'testing' stamp: ''!
mightHaveDuplicateRows

	^builders anySatisfy: [:each | each canCauseDuplicateRows].! !

!SimpleQuery methodsFor: 'copying' stamp: ''!
postCopy
	super postCopy.
	fields := OrderedCollection new.
	joins := OrderedCollection new.
	builders := nil.
	tablesToPrint := nil.! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
prepare
	prepared ifTrue: [^self].
	self setupTracing.
	whereClause prepareIn: self.
	self assignTableAliases.
	self fixJoins.
	self computeFields.
	prepared := true.! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printANSITablesOn: aCommand
	"Print ourselves using the JOIN... USING syntax. Note that we have to put the joins in the right order because we're not allowed to refer to tables not mentioned yet. Great syntax. Reminds me of Pascal. And so easy to deal with."

	| printer |
	printer := JoinPrinter for: self.
	printer printJoinsOn: aCommand.! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printJoinsOn: aCommand
	| noLeadIn|
	self platform supportsANSIJoins ifTrue: [^self].
	joins isEmpty ifTrue: [^self].
	noLeadIn := whereClause isEmptyExpression.
	noLeadIn ifFalse: [aCommand nextPutAll: ' AND ('].
	GlorpHelper 
			do: [:each | each printSQLOn: aCommand withParameters: aCommand parameters]
			for: joins
			separatedBy: [aCommand nextPutAll: ' AND '].
	noLeadIn ifFalse: [aCommand nextPut: $)].! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printNormalTablesOn: aCommand
	self printNormalTablesOn: aCommand excluding: #().! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printNormalTablesOn: aCommand excluding: aCollection
	aCollection do: [:each | self tablesToPrint remove: each].
	GlorpHelper
		print: [:table | table sqlTableName]
		on: aCommand
		for: self tablesToPrint
		separatedBy: ', '.
	^self tablesToPrint.! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printOffsetOn: aCommand
	self hasOffset ifFalse: [^self].
	self platform printOffset: self offset on: aCommand.! !

!SimpleQuery methodsFor: 'printing' stamp: ''!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: ' for '.
	aStream nextPutAll: (resultClass isNil ifTrue: ['nil'] ifFalse: [resultClass name]).! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printPostLimitOn: aCommand
	self hasLimit ifFalse: [^self].
	self platform printPostLimit: self limit on: aCommand.! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printPreLimitOn: aCommand
	self hasLimit ifFalse: [^self].
	self platform printPreLimit: self limit on: aCommand.! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printSelectFields: aCollection on: stream 
	| fieldsToPrint |
	(aCollection anySatisfy: [:each | each isDistinct]) 
		ifTrue: [
			fieldsToPrint := aCollection collect: [:each | each asField].
			stream nextPutAll: 'DISTINCT ']
		ifFalse: [fieldsToPrint := aCollection].

	"For SQL Server, the TOP has to come after the distinct, but before the fields"
	self printPreLimitOn: stream.
			
	GlorpHelper 
		do: [:field | field printSelectSQLOn: stream withParameters: nil. ]
		for: fieldsToPrint
		separatedBy: [stream nextPutAll: ', '].! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printSelectFieldsOn: stream 
     self printSelectFields: fields on: stream! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printTablesOn: aCommand
	aCommand
		cr;
		nextPutAll: ' FROM '.
	self useANSIJoins
		ifTrue: [self printANSITablesOn: aCommand]
		ifFalse: [self printNormalTablesOn: aCommand].! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
printWhereClauseOn: aCommand
	self hasEmptyWhereClause
		ifFalse:
			[aCommand
				cr;
				nextPutAll: ' WHERE '.
			whereClause printSQLOn: aCommand withParameters: aCommand parameters].! !

!SimpleQuery methodsFor: 'private' stamp: ''!
privateAddBuilder: anElementBuilder

	builders := builders copyWith: anElementBuilder.! !

!SimpleQuery methodsFor: 'private' stamp: ''!
privateReplaceBuilderWith: anElementBuilder

	| oldBuilder |
	oldBuilder := self elementBuilderFor: anElementBuilder expression.
	oldBuilder isNil ifTrue: [self error: 'No builder found to replace.'].
	builders at: (builders indexOf: oldBuilder) put: anElementBuilder.! !

!SimpleQuery methodsFor: 'private' stamp: ''!
privateSetFields: aCollection
	fields := aCollection.! !

!SimpleQuery methodsFor: 'As yet unclassified' stamp: ''!
queries
	"A hack so you can write things that manipulate both this and compound queries the same way"
	^Array with: self.! !

!SimpleQuery methodsFor: 'converting' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	"Rebuild the query, which is part of a subselect expression, into the new context. Don't do it if it's not necessary. But if it is necessary, also the copy the query. Otherwise two mappings using the same subselect query will mess each other up."
	self == anExpression ifTrue: [^aBaseExpression].
	self baseExpression ultimateBaseExpression == ultimateBaseExpression ifTrue: [^self].
	^self shallowCopy basicRebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression.! !

!SimpleQuery methodsFor: 'testing' stamp: ''!
requiresFullQuery
	| descriptor independent |
	descriptor := self descriptor.
	"This isn't even a valid query."
	descriptor isNil ifTrue: [^false].
	independent := descriptor classesRequiringIndependentQueries.
	^(independent size ~= 1) or: [(independent at: 1) ~= descriptor describedClass].! !

!SimpleQuery methodsFor: 'executing' stamp: ''!
rowsFromDatabaseWithParameters: anArray

	| alreadyTriedAction command |
	(self shortCircuitEmptyReturn: anArray) ifTrue: [^session accessor cursorFor: (ReadStream on: #()) command: (self sqlWith: anArray)].
	alreadyTriedAction := false.
	command := self sqlWith: anArray.
	^[session accessor
		executeCommand: command
		returnCursor: true] 
	on: GlorpDatabaseReadError 
	do: [:ex | 
		alreadyTriedAction ifTrue: [ex pass] ifFalse: [alreadyTriedAction := true].
		session errorAction isNil 
			ifTrue: [ex pass] 
			ifFalse: [
				session errorAction value: ex value: session value: self.	
				ex retry]].! !

!SimpleQuery methodsFor: 'executing' stamp: ''!
setUpExpressions
	super setUpExpressions.
	self validateWhereClause.! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
setupTracing
	| |
	builders isNil ifFalse: [^self].	"Already been done"
	super setupTracing.
	builders := tracing allTracings asArray
		inject: #()
		into:
			[:sum :each | sum , (ElementBuilder for: each in: self) withAllSubBuilders].
	self addOrderingForToManyJoins.! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
signature
	session useBinding ifFalse: [^''].
	^self sqlWith: Dictionary new.! !

!SimpleQuery methodsFor: 'accessing' stamp: ''!
sql: aString
	"Specify the sql string to use for this query, instead of having it be generated. Note that doing this will override the normal sql generation mechanisms."
	self command: (SQLStringSelectCommand new setSQLString: aString).! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
sqlString
	"Mostly for debugging"
	^(self sqlWith: #()) sqlString.! !

!SimpleQuery methodsFor: 'As yet unclassified' stamp: ''!
sqlWith: aDictionary
	| result |
	self prepare.
	customCommand isNil ifFalse: [
		result := customCommand copy.
		result query: self; parameters: aDictionary; useBinding: session useBinding; session: session.
		^result].
	^QuerySelectCommand
		forQuery: self
		parameters: aDictionary
		useBinding: session useBinding
		session: session.! !

!SimpleQuery methodsFor: 'accessing' stamp: ''!
tablePrefix
	^tablePrefix! !

!SimpleQuery methodsFor: 'accessing' stamp: ''!
tablePrefix: aString
	tablePrefix := aString.! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
tablesToPrint
	tablesToPrint isNil ifTrue: [tablesToPrint := self computeTablesToPrint].
	^tablesToPrint.! !

!SimpleQuery methodsFor: 'sql generation' stamp: ''!
tablesToPrint: aCollection
	tablesToPrint := aCollection.! !

!SimpleQuery methodsFor: 'preparing' stamp: ''!
traceExpressionInContextFor: anExpression
	^anExpression rebuildOn: base.! !

!SimpleQuery methodsFor: 'testing' stamp: ''!
useANSIJoins
	^self session platform supportsANSIJoins.! !

!SimpleQuery methodsFor: 'accessing' stamp: ''!
where: anExpression
	whereClause := anExpression.
	self setUpBase.! !

Query subclass: #CompositeQuery
	instanceVariableNames: 'queries'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!CompositeQuery commentStamp: '<historical>' prior: 0!
This is a query that contains several sub-queries, and executes them all in sequence, returning the concatenation of all the results. This is different from a CompoundQuery, which groups multiple queries together according to operations like UNION ALL, or INTERSECT. This just puts together several different queries and executes them separately, combining the results in Smalltalk. CompoundQuery is more powerful, and can take advantage of database features, but this is more flexible.!

!CompositeQuery methodsFor: 'accessing' stamp: ''!
addQuery: aQuery
	queries add: aQuery.! !

!CompositeQuery methodsFor: 'accessing' stamp: ''!
baseExpression
	^queries first baseExpression.! !

!CompositeQuery methodsFor: 'accessing' stamp: ''!
collectionType
	^self primaryQuery collectionType.! !

!CompositeQuery methodsFor: 'preparing' stamp: ''!
copyOfBuilders 
	^self primaryQuery copyOfBuilders.! !

!CompositeQuery methodsFor: 'executing' stamp: ''!
executeWithParameters: parameterArray in: aSession
	| results |
	results := self collectionType new writeStream.
	queries do: [:each |
		results nextPutAll: (each executeWithParameters: parameterArray in: aSession)].
	^results contents.! !

!CompositeQuery methodsFor: 'initialize' stamp: ''!
initialize
	super initialize.
	queries := OrderedCollection new.! !

!CompositeQuery methodsFor: 'copying' stamp: ''!
postCopy

	super postCopy.
	queries := queries collect: [:each | each copy].! !

!CompositeQuery methodsFor: 'preparing' stamp: ''!
prepare
	queries do: [:each | each session: session].
	queries do: [:each | each prepare].! !

!CompositeQuery methodsFor: 'accessing' stamp: ''!
primaryQuery
	^queries first.! !

!CompositeQuery methodsFor: 'accessing' stamp: ''!
proxyType
	^self primaryQuery proxyType.! !

!CompositeQuery methodsFor: 'accessing' stamp: ''!
proxyType: aClass	
	"We currently ignore proxy types and use the ones from our main query."
	^self.! !

Query subclass: #CompoundQuery
	instanceVariableNames: 'queries operation primaryQuery specialBuilders requiresDistinct'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!CompoundQuery commentStamp: '<historical>' prior: 0!
This is a query that has more than one subquery, combined by expressions like UNION ALL or INTERSECTION. It doesn't do all that much except print its subqueries and combine them according to the operation.

Instance Variables:
	operation	<String>	The name of the operation we are using to combine the queries.
	primaryQuery	<Query>	The first query that we contain. This determines what objects we actually return. All of the queries should match up in terms of what they return, or the database is unlikely to accept this as a valid query.
	queries	<(SequenceableCollection of: SimpleQuery)>	The queries we contain.
	specialBuilders <Array of: ElementBuilder> If we want to do something to modify our result, we can add our own set of builders here, which will be added to those of the internal queries. Motivating example is doing a horizontal inheritance query, where we can end up with one of N different objects in the row, and we just want to return that one directly.

!

!CompoundQuery classSide methodsFor: 'instance creation' stamp: ''!
newIntersect
	^self new operation: #INTERSECT.! !

!CompoundQuery classSide methodsFor: 'instance creation' stamp: ''!
newMinus
	^self new operation: #EXCEPT.! !

!CompoundQuery classSide methodsFor: 'instance creation' stamp: ''!
newUnionAll
	^self new operation: #'UNION ALL'.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
addQuery: aQuery
	queries isEmpty ifTrue: [primaryQuery := aQuery].
	queries add: aQuery.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
baseExpression
	^queries first baseExpression.! !

!CompoundQuery methodsFor: 'preparing' stamp: ''!
copyOfBuilders 
	^primaryQuery copyOfBuilders.! !

!CompoundQuery methodsFor: 'boolean operations' stamp: ''!
except: aQuery
	^self minus: aQuery.! !

!CompoundQuery methodsFor: 'executing' stamp: ''!
executeWithParameters: parameterArray in: aSession
	session := aSession.
	^self 
		readFromDatabaseWithParameters: parameterArray.! !

!CompoundQuery methodsFor: 'initializing' stamp: ''!
initialize
	super initialize.
	requiresDistinct := true.
	queries := OrderedCollection new: 3.! !

!CompoundQuery methodsFor: 'testing' stamp: ''!
isMinus
	^operation = 'EXCEPT'.! !

!CompoundQuery methodsFor: 'testing' stamp: ''!
isUnionAll
	^operation = #'UNION ALL'.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
joins
	^queries inject: #() into: [:sum :each | sum, each joins].! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
limit
	^primaryQuery limit.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
limit: anInteger
	"We currently ignore limit sets and use the one from our contained queries."
	^self.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
mightHaveDuplicateRows
	^primaryQuery mightHaveDuplicateRows.! !

!CompoundQuery methodsFor: 'boolean operations' stamp: ''!
minus: aQuery
	aQuery shortCircuitEmptyReturn ifTrue: [^self].
	^CompoundQuery newMinus addQuery: self; addQuery: aQuery.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
offset
	^primaryQuery offset.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
offset: anInteger
	"We currently ignore offset sets and use the one from our contained queries."
	^self.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
operation: aString
	operation := aString.! !

!CompoundQuery methodsFor: 'copying' stamp: ''!
postCopy

	super postCopy.
	queries := queries collect: [:each | each copy].
	primaryQuery := queries first.! !

!CompoundQuery methodsFor: 'preparing' stamp: ''!
prepare
	| |
	queries do: [:each | each session: session].
	queries do: [:each | each prepare].
	ordering isNil ifFalse: [
		ordering := ordering collect: [:each |
			(self expressionBlockFor: each) asGlorpExpressionOn: queries first baseExpression]].
	groupBy isNil ifFalse: [groupBy := (self expressionBlockFor: groupBy) asGlorpExpressionOn: queries first baseExpression].! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
primaryQuery: aQuery
	primaryQuery := aQuery.! !

!CompoundQuery methodsFor: 'printing' stamp: ''!
printCompoundSelectAliasesOn: aStream
	"We only want one table. This is a somewhat convoluted way of getting 't1' uniformly appended"
	aStream nextPutAll: ' '.
	self session platform supportsAliasInCompoundQuery ifTrue: [
		primaryQuery tablesToPrint first printSQLOn: aStream withParameters: nil].! !

!CompoundQuery methodsFor: 'printing' stamp: ''!
printOperationOn: aCommand
	aCommand space.
	aCommand nextPutAll: (self platform compoundOperationFor: operation).
	aCommand space.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
proxyType
	^primaryQuery proxyType.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
proxyType: aClass	
	"We currently ignore proxy types and use the ones from our main query."
	^self.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
queries
	^queries! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
queries: aCollectionOfQueries
	queries := aCollectionOfQueries! !

!CompoundQuery methodsFor: 'executing' stamp: ''!
readFromDatabaseWithParameters: anArray
	| answerStream result |
	answerStream := self rowsFromDatabaseWithParameters: anArray.
	^self readsOneObject
		ifTrue: [result := answerStream next. answerStream release. result]
		ifFalse: [self resultCollectionFor: answerStream].! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
requiresDistinct
	^requiresDistinct! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
requiresDistinct: aBoolean
	requiresDistinct := aBoolean.! !

!CompoundQuery methodsFor: 'As yet unclassified' stamp: ''!
resultClass
	^primaryQuery resultClass.! !

!CompoundQuery methodsFor: 'sql generation' stamp: ''!
rewriteExcept
	"Rewrite any EXCEPT operation we have into a NOT EXISTS test with a correlated subselect. Only a CompoundQuery would have an EXCEPT"
	
	| |
	operation == #EXCEPT ifFalse: [^self].
	queries := queries collect: [:each | each rewriteExcept].
	^self rewriteSetOperationUsing: #notExists:! !

!CompoundQuery methodsFor: 'sql generation' stamp: ''!
rewriteIntersect
	"Rewrite any INTERSECT operation we have into an EXISTS test for a correlated subselect. Only a CompoundQuery would have an intersect"
	
	| |
	operation == #INTERSECT ifFalse: [^self].
	queries := queries collect: [:each | each rewriteIntersect].
	^self rewriteSetOperationUsing: #exists:! !

!CompoundQuery methodsFor: 'sql generation' stamp: ''!
rewriteSetOperationUsing: aSelector
	"Rewrite any INTERSECT OR EXCEPT operation we have into a correlated EXISTS or NOT EXISTS test, respectively. We take the sub-query, and AND into it a constraint that its base expression is equal to the first queries base expression, and then do either an exists: or a notExists:, depending on aSelector."
	
	| newQuery |
	newQuery := queries first copy.
	2 to: queries size do: [:i | | q expression mainExpression |
		q := (queries at: i) copy.
		expression := RelationExpression new outerJoin: false; relation: #=; leftChild: newQuery baseExpression; rightChild: q baseExpression.
		mainExpression := RelationExpression new outerJoin: false; relation: #AND; leftChild: q whereClause; rightChild: expression.
		q whereClause: mainExpression.
		newQuery AND: [:each | each perform: aSelector with: q]].
	^newQuery.! !

!CompoundQuery methodsFor: 'executing' stamp: ''!
rowsFromDatabaseWithParameters: anArray
	^session accessor
		executeCommand: (self sqlWith: anArray)
		returnCursor: true.! !

!CompoundQuery methodsFor: 'As yet unclassified' stamp: ''!
shortCircuitEmptyReturn
	^false.! !

!CompoundQuery methodsFor: 'As yet unclassified' stamp: ''!
shortCircuitEmptyReturn: parameters
	^false.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
shouldRefresh
	^primaryQuery shouldRefresh.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
shouldRefresh: aBoolean
	"We currently ignore sets of this, and use the value from our primary query."
	^self.! !

!CompoundQuery methodsFor: 'printing' stamp: ''!
sqlString
	"Mostly for debugging"
	^(self sqlWith: #()) sqlString.! !

!CompoundQuery methodsFor: 'executing' stamp: ''!
sqlWith: aDictionary

	self prepare.
	session platform queryWithUnsupportedOperationsEliminatedFrom: self do: [:newQuery | 
		^newQuery
			prepare;
			sqlWith: aDictionary].
	^CompoundSelectCommand
		forQuery: self
		parameters: aDictionary
		useBinding: session useBinding
		session: session.! !

!CompoundQuery methodsFor: 'printing' stamp: ''!
tablesToPrint
	^self queries first tablesToPrint.! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
tracing
	^primaryQuery tracing.! !

!CompoundQuery methodsFor: 'boolean operations' stamp: ''!
unionAll: aQuery
	aQuery shortCircuitEmptyReturn ifTrue: [^self].
	^self isUnionAll
		ifTrue:
			[queries add: aQuery.
			self]
		ifFalse:
			[(CompoundQuery newUnionAll)
				addQuery: self;
				addQuery: aQuery].! !

!CompoundQuery methodsFor: 'accessing' stamp: ''!
whereClause
	| allCriteria |
	allCriteria := queries inject: #() into: [:sum :each | sum, each whereClause].
	^ExpressionGroup new addAll: allCriteria.! !

Query subclass: #FilteringQuery
	instanceVariableNames: 'collection filter collectBlock'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!FilteringQuery commentStamp: '<historical>' prior: 0!
This is a query that expects all the objects it might return to already be in memory, and in a collection from which it can pick them out by evaluating a block. This is used to implement filtered reads on a mapping. When one of the mappings fires, it reads all the relevant results, and so each instance  using that mapping can get the results out of memory without going to the database again.

Instance Variables:
	collectBlock	<Block>	A one-argument block we evaluate to actually collect up the results. That is, internally we are likely to have a more complex representation in the collection, e.g. an array containing various extra information plus the objects. This gets the objects out of the collection.
	collection	<(GlorpVirtualCollection of: Object)>	The collection on which we filter. Typically a virtual collection, at least for filtered reads, but can also handle a normal collection.
	filter	<Block>	The two-argument block we use to filter. The first argument is the entry from the collection, the second is the parameters to the query that were used.
	readsOneObject	<Boolean>	A duplicate of readsOneObject in AbstractReadyQuery. Do we return a collection, or a single result.

!

!FilteringQuery classSide methodsFor: 'instance creation' stamp: ''!
on: aCollection filter: aBlock
	^self new
		collection: aCollection;
 		filter: aBlock.! !

!FilteringQuery methodsFor: 'accessing' stamp: ''!
collectBlock
	^collectBlock! !

!FilteringQuery methodsFor: 'accessing' stamp: ''!
collectBlock: aBlock
	collectBlock:= aBlock.! !

!FilteringQuery methodsFor: 'accessing' stamp: ''!
collection
	^collection! !

!FilteringQuery methodsFor: 'accessing' stamp: ''!
collection: aCollection
	aCollection class == GlorpVirtualCollection ifTrue: [collectionType := aCollection getQuery collectionType].
	collection := aCollection.! !

!FilteringQuery methodsFor: 'executing' stamp: ''!
collectionTypesAgree: aCollection
	^self collectionType new class == aCollection class! !

!FilteringQuery methodsFor: 'executing' stamp: ''!
executeWithParameters: parameterArray in: aSession
	| filtered result |
	session := aSession.
	collection class == GlorpVirtualCollection
		ifTrue:
			[collection setSession: aSession.
			collection parameters: parameterArray.
			collection := collection realObjects].
	filtered := self filterResultsForParameters: parameterArray.
	result := self transformsValues
		ifTrue: [filtered collect: [:each | self valueFor: each]]
		ifFalse: [filtered].
	result := (self collectionTypesAgree: result) ifTrue: [result] ifFalse: [self resultCollectionFor: result readStream].
	^self readsOneObject
		ifTrue: [result detect: [:each | true] ifNone: [nil]]
		ifFalse: [result].! !

!FilteringQuery methodsFor: 'accessing' stamp: ''!
filter
	^filter! !

!FilteringQuery methodsFor: 'accessing' stamp: ''!
filter: aBlock
	filter := aBlock! !

!FilteringQuery methodsFor: 'executing' stamp: ''!
filterResultsForParameters: parameterArray
	| passing seen |
	seen := Dialect newIdentitySet.
	passing := collection
		select:
			[:each | 
			| passes new value |
			passes := filter value: each value: parameterArray.
			value := self valueFor: each.
			passes
				ifTrue:
					[new := (seen includes: value) not.
					new ifTrue: [seen add: value]].
			passes and: [new]].
	^passing.! !

!FilteringQuery methodsFor: 'accessing' stamp: ''!
resultClass
	collection class == GlorpVirtualCollection ifTrue: [^collection getQuery resultClass].
	^(collection detect: [:each | true] ifNone: [nil]) class! !

!FilteringQuery methodsFor: 'testing' stamp: ''!
shortCircuitEmptyReturn: parameters
	^false.! !

!FilteringQuery methodsFor: 'executing' stamp: ''!
transformsValues
	^collectBlock notNil.! !

!FilteringQuery methodsFor: 'executing' stamp: ''!
valueFor: anObject
	collectBlock isNil ifTrue: [^anObject].
	^collectBlock value: anObject.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
count: aClass
	^self count: aClass where: nil.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
count: aClass where: anExpression
	| query |
	query := self returningOneOf: aClass where: anExpression.
	query retrieve: #countStar.
	^query.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
read: aClass
	"readManyOf: and returningManyOf: are synonyms. This now seems more natural to me, to be consistent with session API"
	^self returningManyOf: aClass where: nil.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
read: aClass where: anExpression
	^self returningManyOf: aClass where: anExpression.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
readManyOf: aClass
	"readManyOf: and returningManyOf: are synonyms. This now seems more natural to me, to be consistent with session API"
	^self returningManyOf: aClass where: nil.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
readManyOf: aClass where: anExpression
	^self returningManyOf: aClass where: anExpression.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
readOneOf: aClass
	^self returningOneOf: aClass.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
readOneOf: aClass where: anExpression
	^self returningOneOf: aClass where: anExpression.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
returningManyOf: aClass
	^self returningManyOf: aClass where: nil.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
returningManyOf: aClass where: anExpression
	"Backward-compatibility, since we changed the class name."
	^SimpleQuery returningManyOf: aClass where: anExpression.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
returningOneOf: aClass
	"Backward-compatibility, since we changed the class name."
	^SimpleQuery returningOneOf: aClass.! !

!Query classSide methodsFor: 'instance creation' stamp: ''!
returningOneOf: aClass where: anExpression
	"Backward-compatibility, since we changed the class name."
	^SimpleQuery returningOneOf: aClass where: anExpression.! !

!Query methodsFor: 'accessing' stamp: ''!
collectionType
	"Note that queries default the collection type to array, while mappings default to OrderedCollection. I think it makes sense"
	collectionType isNil ifTrue: [collectionType := Array].
	^collectionType.! !

!Query methodsFor: 'accessing' stamp: ''!
collectionType: aClass
	collectionType := aClass.! !

!Query methodsFor: 'copying' stamp: ''!
copy
	^self shallowCopy postCopy.! !

!Query methodsFor: 'copying' stamp: ''!
copyAttributesTo: newQuery
	"We're creating a new query similar to ourselves, ut not a direct copy. Copy over the relevant attributes. Note that we don't copy over the ordering, because that's less universally appropriate. Caller beware."

	newQuery expectedRows: self expectedRows.
	newQuery collectionType: self collectionType.
	newQuery readsOneObject: self readsOneObject.
	newQuery maximumLobSizeToRetrieveDirectly: self maximumLobSizeToRetrieveDirectly.! !

!Query methodsFor: 'executing' stamp: ''!
executeIn: aSession

	^self executeWithParameters:#() in: aSession.! !

!Query methodsFor: 'executing' stamp: ''!
executeWithParameters: parameterArray in: aSession 
	self subclassResponsibility.! !

!Query methodsFor: 'accessing' stamp: ''!
expectedRows
	"How many rows do we think it's likely this query will bring back. Used for tweaking things like block factor"
	^expectedRows isNil
		ifTrue: [expectedRows := self readsOneObject ifTrue: [1] ifFalse: [100]]
		ifFalse: [expectedRows].! !

!Query methodsFor: 'accessing' stamp: ''!
expectedRows: anInteger
	"How many rows do we think it's likely this query will bring back. Used for tweaking things like block factor"
	expectedRows := anInteger! !

!Query methodsFor: 'ordering' stamp: ''!
expressionBlockFor: anOrderingCriteria
	"Allow us to use symbols interchangeably with simple blocks for ordering, so 
	#firstName is equivalent to [:each | each firstName]. Also, allow chains of symbols, so #(owner firstName)"

	anOrderingCriteria isGlorpExpression ifTrue: [^anOrderingCriteria].
	anOrderingCriteria glorpIsBlock ifTrue: [^anOrderingCriteria].

	anOrderingCriteria isSymbol 
		ifTrue: [^[:each | each perform: anOrderingCriteria]].
	
	"otherwise, we assume it's a collection of symbols, the only other valid case"
	anOrderingCriteria do: [:each | each isSymbol ifFalse: [self error: 'invalid ordering criteria']].
	^[:each | anOrderingCriteria inject: each into: [:sum :eachExpression |
		sum perform: eachExpression]].! !

!Query methodsFor: 'ordering' stamp: ''!
groupBy
	^groupBy.! !

!Query methodsFor: 'ordering' stamp: ''!
groupBy: aBlock
	groupBy := aBlock.! !

!Query methodsFor: 'initialize' stamp: ''!
initialize

	prepared := false.
	maximumLobSizeToRetrieveDirectly := 32768.! !

!Query methodsFor: 'accessing' stamp: ''!
maximumLobSizeToRetrieveDirectly
	^maximumLobSizeToRetrieveDirectly! !

!Query methodsFor: 'accessing' stamp: ''!
maximumLobSizeToRetrieveDirectly: anInteger
	maximumLobSizeToRetrieveDirectly := anInteger.! !

!Query methodsFor: 'ordering' stamp: ''!
orderBy: aBlock
	ordering isNil 
		ifTrue: [ordering := Array with: aBlock]
		ifFalse: [ordering := ordering , (Array with: aBlock)].! !

!Query methodsFor: 'ordering' stamp: ''!
ordering
	^ordering.! !

!Query methodsFor: 'accessing' stamp: ''!
platform
	^session system platform.! !

!Query methodsFor: 'copying' stamp: ''!
postCopy

	prepared := false.! !

!Query methodsFor: 'sql generation' stamp: ''!
printCompoundSelectAliasesOn: aStream
	^self.! !

!Query methodsFor: 'sql generation' stamp: ''!
printGroupByOn: aStream
	groupBy isNil ifTrue: [^self].
	aStream nextPutAll: ' GROUP BY '.
	groupBy printSQLOn: aStream withParameters: nil.! !

!Query methodsFor: 'printing' stamp: ''!
printOn: aStream
	"I hate the printing with namespaces in vw"
      | title |
        title := self class name.
        aStream nextPutAll:
                ((title at: 1) isVowel ifTrue: ['an '] ifFalse: ['a ']).
        aStream nextPutAll: title! !

!Query methodsFor: 'sql generation' stamp: ''!
printOrderingOn: aStream
	ordering isNil ifTrue: [^self].
	aStream nextPutAll: ' ORDER BY '.
	GlorpHelper
		do: [:each | each printSQLOn: aStream withParameters: nil]
		for: ordering
		separatedBy: [aStream nextPutAll: ', '].! !

!Query methodsFor: 'accessing' stamp: ''!
readsOneObject
	readsOneObject isNil ifTrue: [^false].
	^readsOneObject.! !

!Query methodsFor: 'accessing' stamp: ''!
readsOneObject: aBoolean 
	readsOneObject := aBoolean! !

!Query methodsFor: 'executing' stamp: ''!
resultCollectionFor: aCursoredStream
	| results finalAnswer |
	collectionType isNil ifTrue: [^aCursoredStream upToEnd].
	collectionType == aCursoredStream class ifTrue: [^aCursoredStream].
	results := (aCursoredStream newCollectionOfSize: self expectedRows for: collectionType) writeStream.
	results nextPutAll: aCursoredStream upToEnd.
	finalAnswer := results contents.
	session register: finalAnswer.
	^finalAnswer.! !

!Query methodsFor: 'specifying retrievals' stamp: ''!
retrieveDictionaryWithKey: keyExpression value: valueExpression

	self collectionType: (GlorpDictionaryDescription new keyExpression: keyExpression; valueExpression: valueExpression).! !

!Query methodsFor: 'specifying retrievals' stamp: ''!
returnsDictionary

	^collectionType class == GlorpDictionaryDescription.! !

!Query methodsFor: 'sql generation' stamp: ''!
rewriteExcept
	"Rewrite any EXCEPT operation we have into a NOT EXISTS tests. Only a CompoundQuery would have such an operation"! !

!Query methodsFor: 'sql generation' stamp: ''!
rewriteIntersect
	"Rewrite any INTERSECT operation we have into an EXISTS tests. Only a CompoundQuery would have such an operation"! !

!Query methodsFor: 'accessing' stamp: ''!
session
	^session! !

!Query methodsFor: 'accessing' stamp: ''!
session: aSession 
	session := aSession! !

!Query methodsFor: 'ordering' stamp: ''!
setOrdering: aCollection
	ordering := aCollection.! !

!Query methodsFor: 'testing' stamp: ''!
shortCircuitEmptyReturn: parameters
	^false.! !

Object subclass: #RowMap
	instanceVariableNames: 'rowDictionary hasDeletions'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-UnitOfWork'!

!RowMap commentStamp: '<historical>' prior: 0!
This is the central class used for writing. We build up a collection of rows that are to be written, indexed by the objects that "own" them. While any object can contribute values to a row (for example, in a relationship, the value of a field that is a foreign key may come from the primary key of another object) it is a central assumption that each row is owned by a single object. Once we know all the rows, we can compute which ones have changed from their previous database values (see RowMapForMementos) and which fields within them have changed, and then we can write them out.

Instance Variables:
	hasDeletions	<Boolean>	Do we have any deletions. Deletions are marked by a flag in the row, but it's useful to know quickly if we have any or not.
	rowDictionary	<(Dictionary from: DatabaseTable to: (Dictionary from: Object to: DatabaseRow)>	The rows, indexed by their table, then by the owning object.

!

!RowMap classSide methodsFor: 'instance creation' stamp: ''!
forRows: aCollection
	"Convenience - construct a row map from a given set of rows"
	| map |
	map := self new.
	aCollection do: [:each | map addRow: each].
	^map.! !

!RowMap classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!RowMap classSide methodsFor: 'instance creation' stamp: ''!
new: size

	^super new initialize: size.! !

!RowMap methodsFor: 'lookup' stamp: ''!
addRow: aRow
	^self addRow: aRow forTable: aRow table withKey: aRow owner.! !

!RowMap methodsFor: 'lookup' stamp: ''!
addRow: aRow forTable: aTable withKey: aKey 
	| submap |
	aKey isGlorpProxy ifTrue: [self halt].
	submap := self subMapForTable: aTable withKey: aKey.
	submap at: aKey put: aRow.
	^aRow.! !

!RowMap methodsFor: 'set operations' stamp: ''!
additiveDifferencesFrom: aRowMap into: differencesMap
	"Add everything which is in us, but not in aRowMap into differencesMap"
	| newRow |
	self objectsAndRowsDo: [:object :row |
		| correspondingRow |
		correspondingRow := aRowMap
			rowForTable: row table
			withKey: object
			ifAbsent: [DatabaseRow new].
		(row equals: correspondingRow)
			ifFalse:
				[newRow := differencesMap
					addRow: (row withAllFieldsIn: correspondingRow)
					forTable: row table
					withKey: object.
				newRow forDeletion: row forDeletion]].! !

!RowMap methodsFor: 'private/mapping' stamp: ''!
adjustForMementos: objects
	^self.! !

!RowMap methodsFor: 'api' stamp: ''!
at: aField key: anObject put: aValue
	(self findOrAddRowForTable: aField table withKey: anObject) at: aField put: aValue.! !

!RowMap methodsFor: 'private/deletion' stamp: ''!
collapseMatchingInsertsAndDeletes
	^rowDictionary
		inject: Dialect identitySetClass new
		into:
			[:sum :eachTablesRowList | 
			sum addAll: (self collapseMatchingInsertsAndDeletesFor: eachTablesRowList).
			sum].! !

!RowMap methodsFor: 'private/deletion' stamp: ''!
collapseMatchingInsertsAndDeletesFor: eachTablesRows
	"Find any inserts and deletes for these rows which have the same primary key. We expect these to be link table rows, because for rows that map to objects, object identity should ensure this never happens. Force the insert to be an update instead (by returning it to the calling method, which will force that to happen) and remove the deletion"
	| deletions updates matchedDeletions deletion nonGeneratedFieldSize |
	(self isSuitableForCollapsing: eachTablesRows) ifFalse: [^#()].
	deletions := Dictionary new: 50.
	updates := OrderedCollection new.
	eachTablesRows do: [:each | each forDeletion ifTrue: [
		deletions at: (each primaryKey) put: each]].
	deletions isEmpty ifTrue: [^#()].
	matchedDeletions := Dialect identitySetClass new: deletions size.

	nonGeneratedFieldSize := (eachTablesRows detect: [:each | true]) nonGeneratedFields size.
	eachTablesRows do: [:each |
		| pk |
		pk := each primaryKey.
		((deletions includesKey: pk) and: [each forDeletion not]) ifTrue: [
			each numberOfFields = nonGeneratedFieldSize ifFalse: [self error: 'link table entry requiring update is only partially populated'].
			deletion := deletions at: pk.
			(matchedDeletions includes: deletion)
				ifFalse: [
					"Now here's a very special case. If there are no non-primary key fields, then we collapse this down to a no-op. We've deleted and re-inserted with exactly the same information. However, we still leave it in the list of updated rows, because we want it to be put into the session cache if it's not already there at the end."
					each table primaryKeyFields size = each fields size
						ifTrue: [self privateDeleteRow: each].
					updates add: each.
					matchedDeletions add: deletion.
					self privateDeleteRow: deletion]]].
	^updates! !

!RowMap methodsFor: 'private/mapping' stamp: ''!
collectionMementoFor: aCollection

	^aCollection.! !

!RowMap methodsFor: 'private/deletion' stamp: ''!
compress
	"Remove all deleted rows and rebuild"
	| newDictionary |
	hasDeletions ifFalse: [^self].
	Transcript cr; show: 'compressing'.
	hasDeletions := false.
	newDictionary := IdentityDictionary new: rowDictionary size.
	rowDictionary keysAndValuesDo: [:eachTable :eachRowDict |
		| newRowDict |
		newRowDict := IdentityDictionary new: eachRowDict size.
		eachRowDict keysAndValuesDo: [:eachObject :eachRow |
			(self isDeleted: eachRow) ifFalse: [
				newRowDict at: eachObject put: eachRow]].
		newDictionary at: eachTable put: newRowDict].
	rowDictionary := newDictionary.! !

!RowMap methodsFor: 'lookup' stamp: ''!
deleteRowForTable: aTable withKey: aKey ifAbsent: aBlock
	"Don't actually delete, the fixup of the hash tables is too expensive. Mark invisible instead, overloading the status variable that we know we're not using until later"

	| row |
	row := (self subMapForTable: aTable) at: aKey ifAbsent: aBlock.
	row isNil ifFalse: [self privateDeleteRow: row].! !

!RowMap methodsFor: 'private/mapping' stamp: ''!
dictionaryClassRequiredForKey: anObject

	^(self isRowMapKey: anObject)
		ifTrue: [Dictionary]
		ifFalse: [IdentityDictionary].! !

!RowMap methodsFor: 'set operations' stamp: ''!
differenceFrom: aMementoRowMap
	| differencesMap |
	differencesMap := RowMap new.
	self	additiveDifferencesFrom: aMementoRowMap into: differencesMap.
	self	subtractiveDifferencesFrom: aMementoRowMap into: differencesMap.
	^differencesMap.! !

!RowMap methodsFor: 'api' stamp: ''!
findOrAddRowForTable: aTable withKey: aKey
	| submap row |
	submap := self subMapForTable: aTable withKey: aKey.
	row := submap
		at: aKey
		ifAbsentPut: [DatabaseRow newForTable: aTable withOwner: aKey].
	^(self isDeleted: row)
		ifTrue:
			[| newRow |
			newRow := DatabaseRow newForTable: aTable withOwner: aKey.
			submap at: aKey put: newRow.
			newRow]
		ifFalse: [row].! !

!RowMap methodsFor: 'lookup' stamp: ''!
includesRowForTable: aTable withKey: aKey

	| row |
	row := (self subMapForTable: aTable ifAbsent: [^false]) at: aKey ifAbsent: [^false].
	^(self isDeleted: row) not.! !

!RowMap methodsFor: 'initialize/release' stamp: ''!
initialize
      self initialize: 17! !

!RowMap methodsFor: 'initialize/release' stamp: ''!
initialize: size
	rowDictionary := IdentityDictionary new: size.
	hasDeletions := false.! !

!RowMap methodsFor: 'private/deletion' stamp: ''!
isDeleted: row
	^row status == #invisible.! !

!RowMap methodsFor: 'testing' stamp: ''!
isEmpty
	self compress.
	self rowsDo: [:each | (self isDeleted: each) ifFalse: [^false]].
	^true.! !

!RowMap methodsFor: 'api' stamp: ''!
isRowMapKey: anObject
	^anObject class == RowMapKey or: [anObject class == MultipleRowMapKey].! !

!RowMap methodsFor: 'private/deletion' stamp: ''!
isSuitableForCollapsing: eachTablesRows
	"Return true if this set of rows are suitable for collapsing the inserts and deletes together. It's suitable if this is a link table, and if there is a primary key defined which we can use to decide if the rows match or not"
	eachTablesRows isEmpty ifTrue: [^false].
	eachTablesRows
		keysAndValuesDo:
			[:eachKey :eachRow | 
				"(self isRowMapKey: eachKey) ifFalse: [^false]."
				eachRow table primaryKeyFields isEmpty ifTrue: [^false].
				^true ].! !

!RowMap methodsFor: 'iterating' stamp: ''!
keysAndValuesDo: aBlock 
	self tables do: [:each | 
		(self subMapForTable: each) keysAndValuesDo: aBlock].! !

!RowMap methodsFor: 'api' stamp: ''!
mementoFor: anObject
	^anObject.! !

!RowMap methodsFor: 'testing' stamp: ''!
notEmpty
	^self isEmpty not! !

!RowMap methodsFor: 'counting' stamp: ''!
numberOfEntries
	self compress.
	^rowDictionary inject: 0 into: [:sum :each | sum + each size]! !

!RowMap methodsFor: 'counting' stamp: ''!
numberOfEntriesForTable: aTable
	self compress.
	^(self subMapForTable: aTable) size.! !

!RowMap methodsFor: 'iterating' stamp: ''!
objects
	| objects |
	objects := IdentitySet new.
	self tables do: [:each | objects addAll: (self subMapForTable: each) keys].
	^objects! !

!RowMap methodsFor: 'iterating' stamp: ''!
objectsAndRowsDo: aTwoArgumentBlock
	rowDictionary
		do:
			[:eachObjectToRowDictionary | 
			eachObjectToRowDictionary
				keysAndValuesDo:
					[:key :value | 
					(self isDeleted: value)
						ifFalse: [aTwoArgumentBlock value: key value: value]]].! !

!RowMap methodsFor: 'iterating' stamp: ''!
objectsAndRowsForTable: aTable do: aTwoArgumentBlock
	^(self subMapForTable: aTable)
		keysAndValuesDo:
			[:key :value | 
			(self isDeleted: value)
				ifFalse: [aTwoArgumentBlock value: key value: value]].! !

!RowMap methodsFor: 'iterating' stamp: ''!
objectsDo: aBlock
	self tables
		do:
			[:each | 
			(self subMapForTable: each)
				keysAndValuesDo:
					[:key :value | (self isDeleted: value) ifFalse: [aBlock value: key]]].! !

!RowMap methodsFor: 'private/deletion' stamp: ''!
privateDeleteRow: row
	row status: #invisible.
	hasDeletions := true.! !

!RowMap methodsFor: 'private/mapping' stamp: ''!
reverseLookup: anObject

	^anObject.! !

!RowMap methodsFor: 'lookup' stamp: ''!
rowForTable: aTable withKey: aKey 
	^self rowForTable: aTable withKey: aKey ifAbsent: [self error: 'missing element'].! !

!RowMap methodsFor: 'lookup' stamp: ''!
rowForTable: aTable withKey: aKey ifAbsent: aBlock
	| row |
	row := (self subMapForTable: aTable) at: aKey ifAbsent: aBlock.
	^(self isDeleted: row) ifTrue: [aBlock value] ifFalse: [row].! !

!RowMap methodsFor: 'lookup' stamp: ''!
rowForTable: aTable withKey: aKey ifPresent: aBlock
	^(self subMapForTable: aTable)
		at: aKey
		ifPresent:
			[:row | (self isDeleted: row) ifTrue: [nil] ifFalse: [aBlock value: row]].! !

!RowMap methodsFor: 'iterating' stamp: ''!
rowsDo: aBlock 
	self tables do: [:each | self rowsForTable: each do: aBlock]! !

!RowMap methodsFor: 'private/mapping' stamp: ''!
rowsForKey: aKey
	"Return a collection of all rows for any table which are keyed by aKey"

	| rowsForKey |
	rowsForKey := OrderedCollection new: 5.
	rowDictionary
		do:
			[:each | 
			| row |
			row := each at: aKey ifAbsent: [nil].
			(row notNil and: [(self isDeleted: row) not]) ifTrue: [rowsForKey add: row]].
	^rowsForKey.! !

!RowMap methodsFor: 'iterating' stamp: ''!
rowsForTable: aTable do: aBlock
	^(self subMapForTable: aTable)
		keysAndValuesDo:
			[:key :value | (self isDeleted: value) ifFalse: [aBlock value: value]].! !

!RowMap methodsFor: 'private/mapping' stamp: ''!
subMapForTable: aTable
	
	^self subMapForTable: aTable withKey: nil.! !

!RowMap methodsFor: 'private/mapping' stamp: ''!
subMapForTable: aTable ifAbsent: aBlock 
	^rowDictionary at: aTable ifAbsent: aBlock! !

!RowMap methodsFor: 'private/mapping' stamp: ''!
subMapForTable: aTable withKey: anObject 
	anObject isGlorpProxy ifTrue: [self halt].
	^rowDictionary at: aTable
		ifAbsentPut: [(self dictionaryClassRequiredForKey: anObject) new]! !

!RowMap methodsFor: 'set operations' stamp: ''!
subtractiveDifferencesFrom: aRowMap into: differencesMap
	"Figure out which things are in aRowMap but not in us. These should be flagged as delete rows. Since we have no further use for aRowMap after this, we can compute this destructively by removing everything that's in us"

	self objectsAndRowsDo: [:object :row |
		aRowMap
			deleteRowForTable: row table
			withKey: object
			ifAbsent: []].
	aRowMap
		objectsAndRowsDo:
			[:object :row | 
			| adjustedObject |
			adjustedObject := (aRowMap reverseLookup: object).
			row forDeletion: true.
			differencesMap
				addRow: row
				forTable: row table
				withKey: adjustedObject].! !

!RowMap methodsFor: 'private/mapping' stamp: ''!
tables

	^rowDictionary keys.! !

!RowMap methodsFor: 'api' stamp: ''!
unifyField: eachSourceField key: sourceObject withField: eachTargetField key: targetObject
	| sourceRow targetRow |
	sourceRow := self findOrAddRowForTable: eachSourceField table withKey: sourceObject.
	targetRow := self findOrAddRowForTable: eachTargetField table withKey: targetObject.
	sourceRow unifyField: eachSourceField withField: eachTargetField inRow: targetRow.! !

!RowMap methodsFor: 'api' stamp: ''!
unifyField: sourceField key: sourceObject withField: targetField key: targetObject transformationForward: forwardTransformation inverse: inverseTransformation
	| sourceRow targetRow existingWrapper newWrapper |
	sourceRow := self findOrAddRowForTable: sourceField table withKey: sourceObject.
	targetRow := self findOrAddRowForTable: targetField table withKey: targetObject.
	sourceRow unifyField: sourceField withField: targetField inRow: targetRow.
	existingWrapper := targetRow wrapperAt: targetField.
	(existingWrapper class == TransformingFieldValueWrapper and: [
		existingWrapper forwardTransformation == forwardTransformation and: [
			existingWrapper inverseTransformation == inverseTransformation]])
				ifTrue: [^self].
	newWrapper := (TransformingFieldValueWrapper new)
		innerWrapper: existingWrapper;
		forwardTransformation: forwardTransformation;
		inverseTransformation: inverseTransformation.
	targetRow wrapperAt: targetField put: newWrapper.! !

RowMap subclass: #RowMapForMementos
	instanceVariableNames: 'correspondenceMap reversedCorrespondenceMap collectionMementos'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-UnitOfWork'!

!RowMapForMementos commentStamp: '<historical>' prior: 0!
This is a specialized version of RowMap for creating rowmaps out of the mementos in the undo/correspondence map. When doing partial writes we create a rowmap for the current state of the objects, then a rowmap for the original state, and difference the two.

The tricky part is that the mementos refer back to the original objects, so when we establish unification constraints between rows, they would establish them to original objects. This is wrong, and not trivial to debug.

So this rowmap keeps the correspondence map and knows that it has to compensate and get the memento for any related objects.
!

!RowMapForMementos classSide methodsFor: 'instance creation' stamp: ''!
withCorrespondenceMap: aDictionary

	^self new
		correspondenceMap: aDictionary.! !

!RowMapForMementos methodsFor: 'api' stamp: ''!
adjustForMementos: objects
	"We may be being passed row map keys that refer to originals when they should refer to mementos. Fix.This is horribly ugly."

	1 to: objects size do: [:i | | possibleRowMapKey |
		possibleRowMapKey := objects at: i.
		(self isRowMapKey: possibleRowMapKey) ifTrue: [
			objects at: i put: (possibleRowMapKey adjustIn: correspondenceMap)]].
	^objects.! !

!RowMapForMementos methodsFor: 'private/mapping' stamp: ''!
collectionMementoFor: aCollection

	aCollection glorpIsCollection ifFalse: [^aCollection].
	^collectionMementos at: aCollection ifAbsent: [correspondenceMap at: aCollection].! !

!RowMapForMementos methodsFor: 'accessing' stamp: ''!
collectionMementos: aDictionary
	collectionMementos := aDictionary.! !

!RowMapForMementos methodsFor: 'accessing' stamp: ''!
correspondenceMap
	^correspondenceMap! !

!RowMapForMementos methodsFor: 'accessing' stamp: ''!
correspondenceMap: anObject
	correspondenceMap := anObject! !

!RowMapForMementos methodsFor: 'api' stamp: ''!
deleteRowForTable: aTable withKey: aKey ifAbsent: aBlock
	^super deleteRowForTable: aTable withKey: (self mementoFor: aKey) ifAbsent: aBlock.! !

!RowMapForMementos methodsFor: 'api' stamp: ''!
findOrAddRowForTable: aTable withKey: aKey 
	^super findOrAddRowForTable: aTable withKey: (self mementoFor: aKey).! !

!RowMapForMementos methodsFor: 'api' stamp: ''!
mementoFor: aKey
	^(self isRowMapKey: aKey)
		ifTrue: [aKey adjustIn: correspondenceMap]
		ifFalse: [correspondenceMap at: aKey ifAbsent: [aKey]].! !

!RowMapForMementos methodsFor: 'private/mapping' stamp: ''!
originalObjectFor: anObject

	^correspondenceMap at: anObject.! !

!RowMapForMementos methodsFor: 'private/mapping' stamp: ''!
reverseLookup: anObject

      (self isRowMapKey: anObject) ifTrue: [
            ^anObject reverseAdjustIn: self].
      ^ self reversedCorrespondenceMap at: anObject.! !

!RowMapForMementos methodsFor: 'api' stamp: ''!
reversedCorrespondenceMap
	(reversedCorrespondenceMap isNil
		or: [reversedCorrespondenceMap size ~= correspondenceMap size])
		ifTrue:
			[reversedCorrespondenceMap := IdentityDictionary
				new: correspondenceMap size.
			correspondenceMap
				keysAndValuesDo: [:k :v | reversedCorrespondenceMap at: v put: k]].
	^reversedCorrespondenceMap.! !

!RowMapForMementos methodsFor: 'api' stamp: ''!
rowForTable: aTable withKey: aKey ifAbsent: aBlock
	^super
		rowForTable: aTable
		withKey: (self mementoFor: aKey)
		ifAbsent: aBlock.! !

Object subclass: #RowMapKey
	instanceVariableNames: 'key1 key2'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-UnitOfWork'!

!RowMapKey commentStamp: '<historical>' prior: 0!
This class serves as a key for a dictionary containing two sub-keys, where we want to be able to look up based on the identity of both sub-keys paired together. This is used primarily for many-to-many mappings indexing into rowmaps, where we want to key the row by the identity of the object that determines it, but there are two of them.

Instance Variables:

key1	<Object>	One sub-key.
key2	<Object>	The other sub-key.'!

!RowMapKey methodsFor: 'comparing' stamp: ''!
= aRowMapKey

	aRowMapKey class == self class ifFalse: [^false].
	^(key1 == aRowMapKey key1 and: [key2 == aRowMapKey key2]) or: [
		key2 == aRowMapKey key1 and: [key1 == aRowMapKey key2]].! !

!RowMapKey methodsFor: 'adjusting' stamp: ''!
adjustIn: aCorrespondenceMap
	"Adjust ourselves to refer to the copies rather than the originals"
	| adjusted |
	adjusted := self copy.
	adjusted key1: (aCorrespondenceMap at: key1 ifAbsent: [key1]).
	adjusted key2: (aCorrespondenceMap at: key2 ifAbsent: [key2]).
	^adjusted.! !

!RowMapKey methodsFor: 'comparing' stamp: ''!
hash
	| hash1 hash2 hash1Temp longHash |
	hash1 := key1 identityHash.
	hash2 := key2 identityHash.
	hash1 > hash2 ifTrue: [
		hash1Temp := hash1.
		hash1 := hash2.
		hash2 := hash1Temp].
	longHash := (hash1 bitShift: 8) bitXor: hash2.
	^Dialect fitHashIntoSmallInteger: longHash.! !

!RowMapKey methodsFor: 'accessing' stamp: ''!
key1
	^key1! !

!RowMapKey methodsFor: 'accessing' stamp: ''!
key1: anObject
	key1 := anObject! !

!RowMapKey methodsFor: 'accessing' stamp: ''!
key2
	^key2! !

!RowMapKey methodsFor: 'accessing' stamp: ''!
key2: anObject
	key2 := anObject! !

!RowMapKey methodsFor: 'adjusting' stamp: ''!
reverseAdjustIn: aRowMapForMementos
	"Adjust ourselves to refer to the copies rather than the originals"
	| adjusted |
	adjusted := self copy.
	adjusted key1: (aRowMapForMementos reversedCorrespondenceMap at: key1 ifAbsent: [key1]).
	adjusted key2: (aRowMapForMementos reversedCorrespondenceMap at: key2 ifAbsent: [key2]).
	^adjusted.! !

Object subclass: #TableSorter
	instanceVariableNames: 'orderedTables tables visitedTables'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-UnitOfWork'!

!TableSorter commentStamp: '<historical>' prior: 0!
This implements a topological sort for tables, based on the foreign key constraints between them.

Instance Variables:
	orderedTables	<Collection of: DatabaseTable>	The resulting tables in sorted order.
	tables	<(Collection of: (DatabaseTable))>	The input tables.
	visitedTables	<IdentitySet of: DatabaseTable>	The tables we have already visited during the sort.

!

!TableSorter classSide methodsFor: 'instance creation' stamp: ''!
for: tables

	| sorter |
	sorter := self new.
	tables do: [:each |
		sorter addTable: each].
	^sorter.! !

!TableSorter classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!TableSorter methodsFor: 'accessing' stamp: ''!
addTable: aTable 
	tables add: aTable! !

!TableSorter methodsFor: 'accessing' stamp: ''!
hasBeenVisited: aTable

	^visitedTables includes: aTable.! !

!TableSorter methodsFor: 'initializing' stamp: ''!
initialize
	tables := OrderedCollection new: 100.
	visitedTables := IdentitySet new: 100.! !

!TableSorter methodsFor: 'accessing' stamp: ''!
markVisited: aTable

	visitedTables add: aTable.! !

!TableSorter methodsFor: 'sorting' stamp: ''!
sort
	orderedTables := OrderedCollection new: tables size.
	tables do: [:each | self visit: each].
	^orderedTables select: [:each | tables includes: each].! !

!TableSorter methodsFor: 'sorting' stamp: ''!
visit: aTable 
	"The essential bit of topological sort. Visit each node in post-order, traversing dependencies, based on foreign key constraints to database-generated fields. "

	(self hasBeenVisited: aTable) ifTrue: [^self].
	self markVisited: aTable.
	self visitDependentTablesFor: aTable.
	orderedTables add: aTable! !

!TableSorter methodsFor: 'sorting' stamp: ''!
visitDependentTablesFor: aTable 
	aTable foreignKeyConstraints do: [:eachConstraint | 
          self visit: eachConstraint targetTable.]! !

Object subclass: #TimedProxyReaper
	instanceVariableNames: 'weakValueDictionary keysSortedByExpiryTime indexOffset mutex reaperThread delay maximumWork nextId timeout done startTime'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!TimedProxyReaper commentStamp: '<historical>' prior: 0!
Overview:

A TimedProxyReaper collects TimedProxies and, using a background process, uninstantiates when they are due to expire. You can control the number of seconds between each invocation of the background process, with the message #delay:, and you can control the amount of work the background process will do each time it runs with the message #maximumWork:.

API Methods:
!

!TimedProxyReaper classSide methodsFor: 'As yet unclassified' stamp: ''!
delay: seconds
	^self new delay: seconds! !

!TimedProxyReaper classSide methodsFor: 'As yet unclassified' stamp: ''!
delay: seconds maximumWork: milliseconds
	^self new delay: seconds; maximumWork: milliseconds! !

!TimedProxyReaper classSide methodsFor: 'As yet unclassified' stamp: ''!
new
	^super new initialize! !

!TimedProxyReaper methodsFor: 'caching' stamp: ''!
add: aTimedProxy
	[mutex wait.
	self noWaitAdd: aTimedProxy] ensure: [mutex signal].
	self restartIfNecessary.! !

!TimedProxyReaper methodsFor: 'caching' stamp: ''!
at: id ifAbsent: errorBlock
	| result |
	[mutex wait.
	result := weakValueDictionary at: id ifAbsent: [^errorBlock value]]
		ensure: [mutex signal].
	^result.! !

!TimedProxyReaper methodsFor: 'accessing' stamp: ''!
delay
	"Answer the delay, in seconds, between runs of the reaper process."

	^delay! !

!TimedProxyReaper methodsFor: 'accessing' stamp: ''!
delay: seconds
	"Set delay between runs of the reaper process in seconds."

	delay := seconds! !

!TimedProxyReaper methodsFor: 'reaping' stamp: ''!
finish
"	Transcript cr; show: 'Done', (Time millisecondClockValue // 1000) printString."
	done := true.! !

!TimedProxyReaper methodsFor: 'reaping' stamp: ''!
forkReaperThread

	| block |
	block := [
		[done] whileFalse: [
			(Delay forSeconds: self delay) wait.
			self reapExpiredObjects]].
	reaperThread := block forkAt: Processor userBackgroundPriority! !

!TimedProxyReaper methodsFor: 'caching' stamp: ''!
includes: aTimedProxy

	self at: aTimedProxy glorpId ifAbsent: [^false].
	^true! !

!TimedProxyReaper methodsFor: 'initializing' stamp: ''!
initialize

	done := false.
	self delay: 5. "seconds"
	self maximumWork: 100. "milliseconds"
	nextId := 1.
	weakValueDictionary := Dialect weakValueDictionaryClass new.
	weakValueDictionary manager: self.

	"Note: The sort used by keysSortedByExpiryTime _must_ be stable."
	keysSortedByExpiryTime := SortedCollection sortBlock: [:associationA :associationB | associationA value "timeToDie" <= associationB value].

	indexOffset := 0.
	mutex := Semaphore forMutualExclusion.
	self forkReaperThread! !

!TimedProxyReaper methodsFor: 'accessing' stamp: ''!
maximumWork
	"Answer the maximum amount of time the reaper thread will run for before yielding."

	^maximumWork! !

!TimedProxyReaper methodsFor: 'accessing' stamp: ''!
maximumWork: milliseconds
	"Set the maximum amount of time the reaper thread will run for before yielding."

	maximumWork := milliseconds! !

!TimedProxyReaper methodsFor: 'reaping' stamp: ''!
mournKeyOf: anEphemeron
	"Don't really need to do anything."

	anEphemeron value uninstantiate.! !

!TimedProxyReaper methodsFor: 'accessing' stamp: ''!
nextSequenceNumber

	| id |
	id := nextId.
	nextId := nextId + 1.
	^id! !

!TimedProxyReaper methodsFor: 'caching' stamp: ''!
noWaitAdd: aTimedProxy

	| association id |
	aTimedProxy glorpId isNil ifTrue: [aTimedProxy glorpId: self nextSequenceNumber].
	id := aTimedProxy glorpId.	
	association := Association key: id value: aTimedProxy glorpTimeToDie.
	weakValueDictionary at: id put: aTimedProxy.
	aTimedProxy glorpAssociation: association.
	keysSortedByExpiryTime add: association.
	aTimedProxy glorpReaper: self.! !

!TimedProxyReaper methodsFor: 'reaping' stamp: ''!
privateReapObjects
	"Private - Go through all of the objects (sorted by time) and remove those which have expired from the receiver's weakValueDictionary.
	Answer the number of objects reaped."

	| numberReaped timesUp |
	startTime := Time millisecondClockValue.
	numberReaped := 0.
	timesUp := false.
	[(Delay forMilliseconds: 100) wait.
	timesUp := true] forkAt: Processor activePriorityPlusOne.
	keysSortedByExpiryTime
		do:
			[:association | 
			done | timesUp ifTrue: [^numberReaped].
			(self shouldReap: association)
				ifTrue:
					[| timedProxy |
					numberReaped := numberReaped + 1.
					association key isNil
						ifFalse:
							["Been touched in the meantime. Just skip over it, but don't reap"
							timedProxy := weakValueDictionary
								removeKey: association key
								ifAbsent: [nil].
							timedProxy class == TimedProxy ifTrue: [timedProxy uninstantiate]]]
				ifFalse: [^numberReaped]].
	^numberReaped.! !

!TimedProxyReaper methodsFor: 'reaping' stamp: ''!
reapExpiredObjects
	| numberReaped |
	[mutex wait.
	keysSortedByExpiryTime isEmpty ifTrue: [self finish].
	done ifTrue: [^self].
	numberReaped := self privateReapObjects.	"If the cache is empty, shut down"
	numberReaped timesRepeat: [keysSortedByExpiryTime removeFirst].
	indexOffset := indexOffset + numberReaped] ensure: [mutex signal].! !

!TimedProxyReaper methodsFor: 'initializing' stamp: ''!
release

	done := true.
	maximumWork := 0.
	reaperThread notNil ifTrue: [reaperThread terminate].
	reaperThread := nil.! !

!TimedProxyReaper methodsFor: 'caching' stamp: ''!
remove: aTimedProxy ifAbsent: errorBlock

	^self removeKey: aTimedProxy glorpId ifAbsent: errorBlock.! !

!TimedProxyReaper methodsFor: 'caching' stamp: ''!
removeKey: key ifAbsent: errorBlock
	"Note: The key is NOT removed from keysSortedByExpiryTime. The reaper
	will eventually reap the given key, and ignore it since it is already gone.
	Removing objects from the middle of keysSortedByExpiryTime will screw
	up the indexing of objects that are registered later. Don't do it."

	| result |
	[mutex wait.
	result := weakValueDictionary removeKey: key ifAbsent: [^errorBlock value]]
		ensure: [mutex signal].
	^result.! !

!TimedProxyReaper methodsFor: 'reaping' stamp: ''!
restartIfNecessary
	[mutex wait.
	done
		ifTrue:
			["Transcript cr; show: 'Restarting', (Time millisecondClockValue // 1000) printString."
			done := false.
			self forkReaperThread]] ensure: [mutex signal].! !

!TimedProxyReaper methodsFor: 'reaping' stamp: ''!
shouldReap: anAssociation
	"Check that either we've passed the start time, or that there's been a wraparound in the millisecond value and we're now larger than it by some ridiculous margin, which we'll call 24 hours."

	| expireTime milliseconds shouldReap |
	anAssociation key isNil ifTrue: [^true].  "It's been touched. Skip over but don't reap, handled elsewhere"
	expireTime := anAssociation value.
	milliseconds := "24 * 60 * 60 * 1000" 86400000.
	shouldReap := expireTime <= startTime or: [expireTime - milliseconds > startTime].
	^shouldReap.! !

!TimedProxyReaper methodsFor: 'caching' stamp: ''!
touch: aTimedProxy
	[| association |
	mutex wait.
	association := aTimedProxy glorpAssociation.
	association isNil ifTrue: [^self].
	association key: nil.	"This indicates that the key should not be reaped at this time."
	self noWaitAdd: aTimedProxy] ensure: [mutex signal].! !

Object subclass: #Tracing
	instanceVariableNames: 'base allTracings retrievalExpressions alsoFetchExpressions query'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Queries'!

!Tracing commentStamp: '<historical>' prior: 0!
A tracing is a collection of expressions representing the graph of other objects which
are to be read at the same time as the root object.

Instance Variables:

	base	<Expression>	The base expression representing the root object. Same as the parameter to the query block
	allTracings	<Collection of: Expression>	The expressions representing each of the associated objects. e.g. base accounts, base amount serviceCharge .
	alsoFetchExpressions	<(Collection of: GlorpExpression)>	Objects to also retrieve, but not included in the result set, just knitted together with the other related objects.
	retrievalExpressions	<(Collection of: GlorpExpression)>	Objects to also retrieve, and to include in teh result set

!

!Tracing classSide methodsFor: 'instance creation' stamp: ''!
for: aQuery
	^self new base: aQuery whereClause ultimateBaseExpression.! !

!Tracing classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!Tracing methodsFor: 'accessing' stamp: ''!
addExpression: anExpression
	self addExpression: anExpression andDo: [:ignore |].! !

!Tracing methodsFor: 'accessing' stamp: ''!
addExpression: anExpression andDo: aBlock
	| exp |
	exp := anExpression asGlorpExpressionOn: self base in: query.
	(allTracings includes: exp) ifFalse: [
		allTracings add: exp.
		aBlock value: exp].! !

!Tracing methodsFor: 'accessing' stamp: ''!
additionalExpressions
	| all |
	alsoFetchExpressions isEmpty ifTrue: [^retrievalExpressions].
	all := OrderedCollection new.
	all addAll: self retrievalExpressions.
	all addAll: self alsoFetchExpressions.
	^all.! !

!Tracing methodsFor: 'accessing' stamp: ''!
allTracings

	^allTracings.! !

!Tracing methodsFor: 'api' stamp: ''!
alsoFetch: anExpression 
	"Add the expression as something which will be explicitly retrieved and knit together with other results, but NOT included in the result list"

	self addExpression: anExpression andDo: [:exp |
		alsoFetchExpressions add: exp].! !

!Tracing methodsFor: 'accessing' stamp: ''!
alsoFetchExpressions
	^alsoFetchExpressions! !

!Tracing methodsFor: 'accessing' stamp: ''!
base
	^query isNil ifTrue: [base] ifFalse: [query baseExpression].! !

!Tracing methodsFor: 'accessing' stamp: ''!
base: anExpression 
	base := anExpression! !

!Tracing methodsFor: 'copying' stamp: ''!
copy
	^self shallowCopy postCopy.! !

!Tracing methodsFor: 'initialize' stamp: ''!
initialize

	base := BaseExpression new.
	allTracings := OrderedCollection new: 2.
	retrievalExpressions := Set new: 3.
	alsoFetchExpressions := Set new: 3.! !

!Tracing methodsFor: 'copying' stamp: ''!
postCopy
	allTracings := allTracings copy.
	retrievalExpressions := retrievalExpressions copy.
	alsoFetchExpressions := alsoFetchExpressions copy.! !

!Tracing methodsFor: 'accessing' stamp: ''!
retrievalExpressions
	^retrievalExpressions.! !

!Tracing methodsFor: 'api' stamp: ''!
retrieve: anExpression 
	"Add the expression as something which will be explicitly retrieved and knit together with other results, and included in the result list"
	self addExpression: anExpression andDo: [:exp |
		retrievalExpressions add: exp].! !

!Tracing methodsFor: 'setup' stamp: ''!
setup
	"We have been put into a query. If we aren't to trace anything else, trace the base"
	retrievalExpressions isEmpty ifTrue: [
		allTracings addFirst: base.
		retrievalExpressions add: base].! !

!Tracing methodsFor: 'querying' stamp: ''!
tracesThrough: aMapping

	^aMapping isStoredInSameTable.! !

!Tracing methodsFor: 'setup' stamp: ''!
updateBase: aBaseExpression
	"Make sure we have the same base as the query"
	| transformed |
	transformed := IdentityDictionary new.
	base == aBaseExpression ifTrue: [^self].
	base := aBaseExpression.
	allTracings := allTracings collect: [:each | | new | new := each asGlorpExpressionOn: base.
		transformed at: each put: new.
		new].
	retrievalExpressions := retrievalExpressions collect: [:each | 
		transformed at: each].
	alsoFetchExpressions := alsoFetchExpressions collect: [:each | 
		transformed at: each].! !

Object subclass: #TypeResolver
	instanceVariableNames: 'members system'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!TypeResolver commentStamp: '<historical>' prior: 0!
TypeResolvers are used to implement inheritance. Each descriptor has a type resolver, which can be asked for the types associated with that descriptor. By default we get an IdentityTypeResolver, which just gives us a single class. With inheritance we can get different types.

Subclasses must implement the following messages:
	accessing
		classesRequiringIndependentQueriesFor:
		describedConcreteClassFor:withBuilder:descriptor:
	type resolution
		typeMappingRootDescriptor

Instance Variables:
	members	<(SequenceableCollection of: (Descriptor))>	The descriptors of the classes that this resolver applies to.
	system	<DescriptorSystem>	The system in which the members live.

!

TypeResolver subclass: #BasicTypeResolver
	instanceVariableNames: 'concreteMembers subclassDescriptorsBuilt rootDescriptor rootClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!BasicTypeResolver commentStamp: '<historical>' prior: 0!
This is an abstract superclass for TypeResolvers that do inheritance in some form.

Instance Variables:
	concreteMembers	<(Collection of: (Descriptor))>	Which of our members can actually be instantiated, as opposed to abstract superclasses.
	rootClass	<Behavior>	Which is the root class of the inheritance hierarchy we represent.
	rootDescriptor	<Descriptor>	The descriptor for the root class
	subclassDescriptorsBuilt	<Boolean>	Have we asked the system for the descriptors for all of the root's subclasses yet?

!

!BasicTypeResolver methodsFor: 'private' stamp: ''!
allDescribedConcreteClasses
	self subclassDescriptorsBuilt ifFalse: [self forceSubclassDescriptorLoads].
	^self concreteMembers collect: [ :each | each describedClass ]! !

!BasicTypeResolver methodsFor: 'accessing' stamp: ''!
concreteMembers
	^concreteMembers isNil 
		ifTrue: [ concreteMembers := OrderedCollection new ]
		ifFalse: [ concreteMembers]! !

!BasicTypeResolver methodsFor: 'private' stamp: ''!
forceSubclassDescriptorLoads
	self rootClass allSubclassesDo: [ :each | self system descriptorFor: each ].
	subclassDescriptorsBuilt := true! !

!BasicTypeResolver methodsFor: 'registering' stamp: ''!
register: aDescriptor abstract: abstract

	super register: aDescriptor abstract: abstract.
	abstract ifFalse: [ self concreteMembers add: aDescriptor ]! !

!BasicTypeResolver methodsFor: 'accessing' stamp: ''!
rootClass
	^rootClass! !

!BasicTypeResolver methodsFor: 'accessing' stamp: ''!
rootClass: anObject
	rootClass := anObject! !

!BasicTypeResolver methodsFor: 'accessing' stamp: ''!
rootDescriptor
	^rootDescriptor isNil
		ifTrue: [ rootDescriptor := self system descriptorFor: self rootClass ]
		ifFalse: [ rootDescriptor ]! !

!BasicTypeResolver methodsFor: 'other' stamp: ''!
rootDescriptor: anObject
	rootDescriptor := anObject! !

!BasicTypeResolver methodsFor: 'accessing' stamp: ''!
subclassDescriptorsBuilt
	^subclassDescriptorsBuilt isNil
		ifTrue: [ subclassDescriptorsBuilt := false ]
		ifFalse: [ subclassDescriptorsBuilt ]! !

!BasicTypeResolver methodsFor: 'type resolution' stamp: ''!
typeMappingRootDescriptor
	^self rootDescriptor! !

!BasicTypeResolver methodsFor: 'testing' stamp: ''!
usesInheritance
	^true.! !

!BasicTypeResolver methodsFor: 'validation' stamp: ''!
validate
	| tables sequences |
	super validate.
	tables := self concreteMembers collect: [:each | each primaryTable].
	sequences := tables inject: #() into: [:sum :each | sum, ((each primaryKeyFields) collect: [:eachField | eachField type sequence])].
	sequences asSet size > 1 ifTrue: [self error: 'When using inheritance, all tables must use the same sequence. Otherwise, the primary key references will be ambiguous.'].! !

BasicTypeResolver subclass: #FilteredTypeResolver
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!FilteredTypeResolver commentStamp: '<historical>' prior: 0!
This is a type resolver for inheritance where all of the classes are represented in a single table, with some sort of identifying field for which subclass they are. The table would have the union of all possible fields for all classes. It's called filtering, because for any particular class, we filter out those instances from all those represented in the table.!

!FilteredTypeResolver classSide methodsFor: 'instance creation' stamp: ''!
forRootClass: aClass
	^self new
		rootClass: aClass;
		yourself.! !

!FilteredTypeResolver methodsFor: 'accessing' stamp: ''!
allTablesFor: aDescriptor
	"Return all tables used by this descriptor or its subclasses. Since all our subclasses use a single table, this is the same as for a normal descriptor."
	^aDescriptor tables.! !

!FilteredTypeResolver methodsFor: 'type resolving' stamp: ''!
askConcreteMembersForConcreteClassFor: row withBuilder: builder
	concreteMembers do: [:each |
		each describedConcreteClassFor: row withBuilder: builder].! !

!FilteredTypeResolver methodsFor: 'accessing' stamp: ''!
classesRequiringIndependentQueriesFor: aClass
	^Array with: aClass! !

!FilteredTypeResolver methodsFor: 'type resolving' stamp: ''!
describedConcreteClassFor: row withBuilder: builder descriptor: aDescriptor
	^aDescriptor typeMapping describedConcreteClassFor: row withBuilder: builder! !

!FilteredTypeResolver methodsFor: 'type resolving' stamp: ''!
fieldsForSelectStatement
	"Return fields that are needed in a select statement - i.e. return all inherited fields that are part of the tables we are already selecting for this object, but not in the main descriptor"
	| fields rootFields |
	fields := OrderedCollection new.
	rootFields := self rootDescriptor mappedFields asSet.
	self concreteMembers do: [:each |
		each == self rootDescriptor ifFalse: [
			each mappedFields do: [:eachSubField |
				(rootFields includes: eachSubField) ifFalse: [
					fields add: eachSubField]]]].
	^fields.! !

!FilteredTypeResolver methodsFor: 'type resolving' stamp: ''!
isAbstract: aDescriptor
	^aDescriptor == self rootDescriptor! !

!FilteredTypeResolver methodsFor: 'registering' stamp: ''!
register: aDescriptor abstract: aBoolean
	super register: aDescriptor abstract: aBoolean.
	aBoolean
		ifTrue:
			[| fakeTypeMapping |
			fakeTypeMapping := FilteredTypeMapping to: nil keyedBy: nil.
			fakeTypeMapping attribute name: #filteredTypeResolverFilterMapping.
			fakeTypeMapping readOnly: true.
			aDescriptor typeMapping: fakeTypeMapping].! !

!FilteredTypeResolver methodsFor: 'registering' stamp: ''!
register: aDescriptor keyedBy: aKey field: aField
	self register: aDescriptor abstract: false.
	aDescriptor typeMapping: (FilteredTypeMapping to: aField keyedBy: aKey).! !

BasicTypeResolver subclass: #HorizontalTypeResolver
	instanceVariableNames: 'shouldUseSingleQuery'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!HorizontalTypeResolver commentStamp: '<historical>' prior: 0!
This implements inheritance where each concrete class is represented in its own table. Each table will duplicate the fields that are in common between the concrete classes.!

!HorizontalTypeResolver classSide methodsFor: 'instance creation' stamp: ''!
forRootClass: aClass
	^self new
		rootClass: aClass;
		yourself.! !

!HorizontalTypeResolver methodsFor: 'accessing' stamp: ''!
allTablesFor: aDescriptor
	"Return all tables used by this descriptor or its subclasses."
	^(self descriptorsRequiringIndependentQueriesFor: aDescriptor describedClass) inject: OrderedCollection new into: [:sum :eachDescriptor |
		sum addAll: ((eachDescriptor == aDescriptor) ifTrue: [eachDescriptor tables] ifFalse: [eachDescriptor allTables]).
		sum].! !

!HorizontalTypeResolver methodsFor: 'accessing' stamp: ''!
classesRequiringIndependentQueriesFor: aClass
	^(self descriptorsRequiringIndependentQueriesFor: aClass) collect: [:each |
		each describedClass].! !

!HorizontalTypeResolver methodsFor: 'accessing' stamp: ''!
describedConcreteClassFor: row withBuilder: builder descriptor: aDescriptor
	^aDescriptor describedClass! !

!HorizontalTypeResolver methodsFor: 'accessing' stamp: ''!
descriptorsRequiringIndependentQueriesFor: aClass
	^self concreteMembers select: [ :each | each describedClass includesBehavior: aClass ].! !

!HorizontalTypeResolver methodsFor: 'testing' stamp: ''!
isTypeMappingRoot: aDescriptor
	^aDescriptor == rootDescriptor! !

!HorizontalTypeResolver methodsFor: 'query setup' stamp: ''!
mappingForSubclass: aDescriptor connectTo: existingSubclassMapping
	"Create a pseudo-mapping for the subclass."
	| attributeName mapping newJoin |
	attributeName := self pseudoAttributeNameForClass: aDescriptor describedClass.
	mapping := self rootDescriptor mappingForAttributeNamed: attributeName.
	mapping isNil ifTrue: [
		rootDescriptor classModel newAttributeNamed: attributeName type: aDescriptor describedClass.
		mapping := (rootDescriptor newMapping: OneToOneMapping)
			attributeName: attributeName.
		mapping beForPseudoVariable.
		existingSubclassMapping isNil ifFalse: [
			newJoin := Join new.
			mapping join: newJoin]].
	^mapping.! !

!HorizontalTypeResolver methodsFor: 'query setup' stamp: ''!
pseudoAttributeNameForClass: aClass
	^('subclassPseudoVariable', aClass name) asSymbol.! !

!HorizontalTypeResolver methodsFor: 'query setup' stamp: ''!
setUpExpressionsForQuery: aQuery
	| mapping firstMapping allDescriptors |
	(self concreteMembers includes: aQuery resultClass) ifTrue: [^self].
	self shouldUseSingleQuery ifFalse: [^self].
	firstMapping := nil.
	allDescriptors := self
		descriptorsRequiringIndependentQueriesFor: aQuery resultClass.
	allDescriptors size = 1 ifTrue: [^self].
	allDescriptors
		do:
			[:eachSubclassDescriptor | 
			mapping := self
				mappingForSubclass: eachSubclassDescriptor
				connectTo: firstMapping.
			firstMapping isNil ifTrue: [firstMapping := mapping].
			aQuery
				alsoFetch:
					[:object | (object getMapping: mapping named: mapping attribute name) bePhantom asOuterJoin]].! !

!HorizontalTypeResolver methodsFor: 'accessing' stamp: ''!
shouldUseSingleQuery
	"Return true if we should attempt to combine queries for multiple subclasses in disjoint tables into one."

	shouldUseSingleQuery isNil ifTrue: [shouldUseSingleQuery := false].
	^shouldUseSingleQuery.! !

!HorizontalTypeResolver methodsFor: 'testing' stamp: ''!
supportsOrderingFor: aDescriptor
	concreteMembers detect: [:each | each == aDescriptor] ifNone: [^false].
	^true.! !

!HorizontalTypeResolver methodsFor: 'accessing' stamp: ''!
useSingleQuery: aBoolean
	shouldUseSingleQuery := aBoolean! !

TypeResolver subclass: #IdentityTypeResolver
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Mappings'!

!IdentityTypeResolver commentStamp: '<historical>' prior: 0!
This is a trivial type resolver (and the default one) that just uses a single class, with no inheritance.!

!IdentityTypeResolver methodsFor: 'accessing' stamp: ''!
allTablesFor: aDescriptor
	"Return all tables used by this descriptor or its subclasses. For a non-inheritance descriptor, this is the same as tables."
	^aDescriptor tables.! !

!IdentityTypeResolver methodsFor: 'type resolution' stamp: ''!
classesRequiringIndependentQueriesFor: aClass
	^Array with: aClass! !

!IdentityTypeResolver methodsFor: 'accessing' stamp: ''!
describedConcreteClassFor: aRow withBuilder: builder descriptor: aDescriptor
	^aDescriptor describedClass! !

!IdentityTypeResolver methodsFor: 'As yet unclassified' stamp: ''!
descriptorsRequiringIndependentQueriesFor: aClass
	^members.! !

!IdentityTypeResolver methodsFor: 'accessing' stamp: ''!
typeMappingRootDescriptor
	^members first! !

!TypeResolver methodsFor: 'accessing' stamp: ''!
addMember: aDescriptor
	members isNil ifTrue: [ members := OrderedCollection new].
	members add: aDescriptor! !

!TypeResolver methodsFor: 'accessing' stamp: ''!
allTablesFor: aDescriptor
	"Return all tables used by this descriptor or its subclasses. For a non-inheritance descriptor, this is the same as tables."
	self subclassResponsibility.! !

!TypeResolver methodsFor: 'accessing' stamp: ''!
classesRequiringIndependentQueriesFor: aClass
	self subclassResponsibility! !

!TypeResolver methodsFor: 'other' stamp: ''!
describedClasses
	^members collect: [:each | each describedClass ].! !

!TypeResolver methodsFor: 'accessing' stamp: ''!
describedConcreteClassFor: row withBuilder: builder descriptor: aDescriptor
	self subclassResponsibility! !

!TypeResolver methodsFor: 'type resolution' stamp: ''!
fieldsForSelectStatement
	"Return fields that are needed in a select statement - i.e. return all inherited fields that are part of the tables we are already selecting for this object"
	^#().! !

!TypeResolver methodsFor: 'registering' stamp: ''!
register: aDescriptor
	^self register: aDescriptor abstract: false! !

!TypeResolver methodsFor: 'registering' stamp: ''!
register: aDescriptor abstract: abstract

	self system: aDescriptor system.
	self addMember: aDescriptor.
	aDescriptor typeResolver: self! !

!TypeResolver methodsFor: 'query setup' stamp: ''!
setUpExpressionsForQuery: aQuery
	"A hook to let us do any modifications to the query that we need."
	^self.! !

!TypeResolver methodsFor: 'accessing' stamp: ''!
shouldUseSingleQuery
	"Return true if we should attempt to combine queries for multiple subclasses in disjoint tables into one."
	^false.! !

!TypeResolver methodsFor: 'testing' stamp: ''!
supportsOrderingFor: aDescriptor
	^true.! !

!TypeResolver methodsFor: 'accessing' stamp: ''!
system
	^system! !

!TypeResolver methodsFor: 'accessing' stamp: ''!
system: anObject
	system := anObject! !

!TypeResolver methodsFor: 'type resolution' stamp: ''!
typeMappingRootDescriptor
	self subclassResponsibility! !

!TypeResolver methodsFor: 'testing' stamp: ''!
usesInheritance
	^false.! !

!TypeResolver methodsFor: 'validation' stamp: ''!
validate
	^self.! !

Object subclass: #UnitOfWork
	instanceVariableNames: 'session transaction deletedObjects newObjects rowMap commitPlan deletePlan rowsByTable succeeded linkRowsForUpdate commitPhase deleting'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-UnitOfWork'!

!UnitOfWork commentStamp: '<historical>' prior: 0!
A UnitOfWork keeps track of objects which might potentially be modified and lets you roll them back or commit the changes into the database.

Instance Variables:
	newObjects	<IdentitySet of: Object>	The objects registered with us as being new (i.e. not existing already in the database)
	session	<Session>	The session in which this is all taking place.
	transaction	<ObjectTransaction>	Keeps track of the original object state so that we can revert it.
	rowMap	<RowMap>	A holder for the rows when we are writing out changes.
	commitPlan	<(OrderedCollection of: DatabaseRow)>	The list of rows to be written, in order. Constructed by topological sorting the contents of the row map.
	calculatingRows <Boolean> Are we in the process of computing the rows to be written. If so, we should not be doing reads and registrations - it's probably an error in the mapping code by accidentally forcing a read.

!

!UnitOfWork classSide methodsFor: 'instance creation' stamp: ''!
new

	^super new initialize.! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
abort
	self reinitialize.! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
abort: anObject
	transaction abort: anObject! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
addObject: eachObject toCacheKeyedBy: key 

	| tracker |	
	self session cacheAt: key put: eachObject.
	tracker := self session modificationTracker.
	tracker isNil ifFalse: [tracker track: eachObject]! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
addToCommitPlan: aRow
	aRow shouldBeWritten ifFalse: [^self].
	commitPlan add: aRow.! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
addToDeletePlan: aRow
	
	deletePlan add: aRow.! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ' 22/6/07 12:54'!
assignSequenceNumbers
	commitPlan do: [:each | each preWriteAssignSequencesUsing: self session accessor].! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
begin
	self reinitialize.! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
buildCommitPlan
	| tablesInCommitOrder |
	commitPlan := OrderedCollection new.
	deletePlan := OrderedCollection new.
	tablesInCommitOrder := session tablesInCommitOrder.
	tablesInCommitOrder do: 
			[:eachTable | 
			self rowsForTable: eachTable
				do: 
					[:eachRow | 
					eachRow forDeletion 
						ifTrue: [self addToDeletePlan: eachRow]
						ifFalse: [self addToCommitPlan: eachRow]]].
	self sortCommitPlanRows.! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
commit
	[self preCommit.
	self writeRows] ifCurtailed: [succeeded := false].
	succeeded
		ifTrue: [self postCommit]
		ifFalse:
			[GlorpTransactionFailure signal].! !

!UnitOfWork methodsFor: 'deletion' stamp: ''!
computeCollectionDeletes
	"If we have objects that are part of exclusive collections, and they have been removed from those collections, then they must be deleted. Note that we do this *before* we register the transitive closure at the end of the unit of work. That means we won't know about any new objects yet, but we don't care because new objects won't be deleted in this way. And that way if we delete and then add another object with (ick) the same primary key, we shouldn't have a clash because the delete will already have been issued."

	self registeredObjects do: [:each |
		| descriptor |
		(descriptor := session descriptorFor: each class) isNil
			ifFalse: [descriptor mappings do: [:eachMapping |
				self computeCollectionDeletesFor: eachMapping inObject: each]]].! !

!UnitOfWork methodsFor: 'deletion' stamp: ''!
computeCollectionDeletesFor: aMapping inObject: anObject
	"Figure out which, if any, dependent objects we should delete based on changes to the corresponding collection in anObject"
	| collection memento collectionMemento |
	aMapping isExclusive ifFalse: [^self].
	aMapping attribute isCollectionAttribute ifFalse: [^self].
	collection := aMapping getValueFrom: anObject.
	collection := self session realObjectFor: collection ifNone: [^self].
	(aMapping isValidTarget: collection) ifFalse: [^self].
	memento := transaction mementoFor: anObject.
	collectionMemento := transaction collectionMementoFor: (self session realObjectFor: (aMapping getValueFrom: memento)).
	"OK, now we have the state of the original collection and the new collection, we can look for differences."
	collectionMemento do: [:each |
		| matching |
		(each isGlorpProxy not or: [each isInstantiated]) ifTrue: [
			matching := collection 
				detect: [:eachElement | 
					(eachElement isGlorpProxy not or: [eachElement isInstantiated])
						and: [each yourSelf == eachElement yourSelf]]
				ifNone: [nil].
			matching == nil ifTrue: [session delete: each]
		]
	].! !

!UnitOfWork methodsFor: 'accessing' stamp: ''!
correspondenceMap
	^transaction undoMap! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
createMementoRowMapFor: objects 
	"Create a rowmap for the objects whose state was already known. We subtract this from the rowmap of all known objects to get the rows that need to be written. New objects are also registered, so we only generate rows here for non-new objects"

	| localRowMap |
	localRowMap := RowMapForMementos 
				withCorrespondenceMap: self correspondenceMap.
	localRowMap collectionMementos: transaction collectionMementos.
	objects keysAndValuesDo: 
			[:original :memento | 
			(self newObjects includes: original) 
				ifFalse: [session createRowsFor: memento in: localRowMap]].
	^localRowMap! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
createRowMapFor: objects
	| localRowMap |
	localRowMap := RowMap new: objects size.
	objects do: [:each | session createRowsFor: each in: localRowMap].
	^localRowMap! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
createRows
	self createRowsForPartialWrites.! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
createRowsForCompleteWrites
	"reference implementation. not called from anywhere"

	self registeredObjectsDo: [:eachObject | session createRowsFor: eachObject in: rowMap].
	deletedObjects do: [:eachObject | session createDeleteRowsFor: eachObject in: rowMap].! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
createRowsForPartialWrites
	| registeredObjectsRowMap mementoObjectsRowMap |
	registeredObjectsRowMap := self createRowMapFor: self registeredObjects.
	deletedObjects 
		do: [:eachObject | session createDeleteRowsFor: eachObject in: registeredObjectsRowMap].
	mementoObjectsRowMap := self createMementoRowMapFor: self mementoObjects.
	rowMap := registeredObjectsRowMap differenceFrom: mementoObjectsRowMap.! !

!UnitOfWork methodsFor: 'deletion' stamp: ''!
delete: anObject
	(self isRegistered: anObject) ifFalse: [session registerForDelete: anObject].
	deletedObjects add: anObject.
	(session embeddedObjectsFor: anObject) do: [:each | self delete: each].! !

!UnitOfWork methodsFor: 'private' stamp: ''!
deletedObjects
	^deletedObjects.! !

!UnitOfWork methodsFor: 'testing' stamp: ''!
forcesUpdateFor: aDatabaseRow
	^linkRowsForUpdate notNil and: [linkRowsForUpdate includes: aDatabaseRow].! !

!UnitOfWork methodsFor: 'deletion' stamp: ''!
hasPendingDeletions

	^deletedObjects isEmpty not.! !

!UnitOfWork methodsFor: 'initializing' stamp: ''!
initialize

	transaction := ObjectTransaction new.
	self reinitialize! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
instantiateIfRequiredFor: anObject mapping: eachMapping
	"Sometimes we have to instantiate the targets if they weren't. Specifically, if there's a relationship where the target has a foreign key to us. e.g. if X has a 1-many to Y, and we don't instantiate the collection of Y, but then replace it with some other collection. The Y's keys have to be updated, so we need to make sure they're read. There are other cases, too"

	| originalTarget targetObject mapping instantiated |
	mapping := eachMapping applicableMappingForObject: anObject.
	mapping isRelationship ifFalse: [^false].
	mapping readOnly ifTrue: [^false].
	targetObject := mapping getValueFrom: anObject.
	(self isNewObject: anObject)
		ifTrue:
			[^self instantiateIfRequiredForNewObjectReferenceTo: targetObject].
	instantiated := (self willDelete: anObject)
		ifTrue:
			[self instantiateIfRequiredForDeletedObjectReferencing: targetObject from: anObject in: eachMapping]
		ifFalse: [false].
	originalTarget := self originalValueFor: anObject mapping: mapping.
	^instantiated or: [self
		instantiateIfRequiredForReferenceChangedFrom: originalTarget
		to: targetObject].! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
instantiateIfRequiredForDeletedObjectReferencing: targetObject from: anObject in: aMapping
	"If we are deleted, then we need to have at least one mapping for each of our primary keys instantiated. If we're lucky, there's a direct mapping, and it's easy. Note that this only looks for the primary table. This might not be correct, but it's not clear, and for secondary tables there's a multiple table join that can be used without needing a mapping. And I'm not prepared to think about secondary tables whose keys are mapped only as relationships. In addition, if we use a link table, then we need to instantiate the reference, because we'll need to have enough information to delete the link table entry/entries."
	| mappingsToUseForPrimaryKeys |
	aMapping mapsPrimaryKeys ifFalse: [^false].	
	aMapping usesLinkTable ifTrue: [
		"Make sure we get the non-filtered version, and return false so that if the reference has changed, we'll also get the original version."
		self notFilteringDeletesDo: [targetObject yourSelf]. ^false].
	mappingsToUseForPrimaryKeys := aMapping descriptor primaryKeyMappingsFor: aMapping descriptor primaryTable.
	"For any of the fields we map, is there anyone else who can get us the value. If not, instantiate us. It might be better to optimize this a bit to look for e.g. one-one mappings before to-many."
	aMapping mappedFields do: [:each |
		| mappings |
		mappings := mappingsToUseForPrimaryKeys at: each ifAbsent: [nil].
		mappings isNil ifFalse: [
			(mappings allSatisfy: [:eachMapping | 
				(eachMapping hasValueThatCanProvideField: each from: anObject) not])
					ifTrue: [
						targetObject yourSelf.
						^true]]].
	^false.! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
instantiateIfRequiredForNewObjectReferenceTo: targetObject
	"If it's a new object, then we always need to instantiate"
	targetObject isGlorpProxy ifFalse: [^false].
	targetObject getValue.
	^true.! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
instantiateIfRequiredForReferenceChangedFrom: originalTarget to: newTarget
	"For an existing relationship, we need to instantiate if the object we're referring to changed, and either was or is now a proxy. Actually we might get away without if it used to be, but now isn't, but better safe than sorry."

	| wasProxy  |
	originalTarget == newTarget ifTrue: [^false].
	wasProxy := false.
	self notFilteringDeletesDo: [
		originalTarget isGlorpProxy ifTrue: [wasProxy := true. originalTarget getValue].
		newTarget isGlorpProxy ifTrue: [wasProxy := true. newTarget getValue]].
	^wasProxy.! !

!UnitOfWork methodsFor: 'testing' stamp: ''!
isNewObject: anObject 
	^self newObjects includes: anObject.! !

!UnitOfWork methodsFor: 'registering' stamp: ''!
isRegistered: anObject

	^transaction isRegistered: anObject.! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
mementoObjects
	"Warning: Excessive cleverness!!!!!! The mementoObjects we want to iterate over are the values in the correspondenceMap dictionary. We were getting the values and returning them, but if all we need to do is iterate, then the dictionary itself works fine"
	^self correspondenceMap.! !

!UnitOfWork methodsFor: 'accessing' stamp: ''!
newObjects
	newObjects isNil ifTrue: [newObjects := IdentitySet new].
	^newObjects! !

!UnitOfWork methodsFor: 'private' stamp: ''!
notFilteringDeletesDo: aBlock
	"Execute aBlock, with filtering of deletes turned off. This is useful if we are in the process of deleting an object. Suppose we are deleting an object that has a link table relationship, but we have already deleted one of the linked objects. We need to make sure we also delete the link table entry, which means we need to know about that linked object, so we don't want Glorp filtering it out for us."

	| oldCommitPhase |
	[oldCommitPhase := commitPhase.
	commitPhase := #registering.
	aBlock value] ensure: [commitPhase := oldCommitPhase].! !

!UnitOfWork methodsFor: 'accessing' stamp: ''!
numberOfRows
	^commitPlan size + deletePlan size.! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
originalValueFor: anObject mapping: eachMapping
	| memento |
	memento := transaction undoMap at: anObject.
	^eachMapping getValueFrom: memento.! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
postCommit
	self sendPostWriteNotification.
	self updateSessionCache.
	self validateRowsPostWrite.! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
preCommit
	commitPhase := #registering.
	self registerTransitiveClosure.
	commitPhase := #building.
	self createRows.
	linkRowsForUpdate := rowMap collapseMatchingInsertsAndDeletes.
	self buildCommitPlan.
	commitPhase := nil.
	self sendPreWriteNotification.
	self validateRows.! !

!UnitOfWork methodsFor: 'private' stamp: ''!
privateGetRowMap

	^rowMap! !

!UnitOfWork methodsFor: 'private' stamp: ''!
privateGetTransaction

	^transaction.! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
propagateDeletes
	"If we have objects which are to be deleted, and they have exclusive relationships to other objects, then we must also delete those other objects."
	| deletionQueue |
	deletionQueue := OrderedCollection new: deletedObjects size.
	deletionQueue addAll: deletedObjects.
	[deletionQueue isEmpty]
		whileFalse:
			[| each |
			each := deletionQueue removeFirst.
			(self session descriptorFor: each) mappings
				do:
					[:eachMapping | 
					eachMapping isExclusive
						ifTrue:
							[eachMapping
								loopOverAttributeOf: each
								doing:
									[:index :subObject | 
									| realSubObject |
									realSubObject := subObject yourSelf.
									(realSubObject notNil and: [(deletedObjects includes: realSubObject) not])
										ifTrue:
											[session delete: subObject.
											deletionQueue add: realSubObject]]]]].! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
readBackNewRowInformation
	| changedObjects |
	changedObjects := rowMap objects.
	changedObjects
		do: 
			[:each | 
			| descriptor |
			descriptor := session descriptorFor: each.
			descriptor isNil ifFalse: [descriptor readBackNewRowInformationFor: each in: rowMap]]! !

!UnitOfWork methodsFor: 'registering' stamp: ''!
register: anObject 

	| realObject |
	commitPhase == #building ifTrue: [self halt]. "Should not happen. Probably indicates that we're triggering proxies while firing the mappings"
	realObject := transaction register: anObject.
	self registerTransitiveClosureFrom: realObject.! !

!UnitOfWork methodsFor: 'private' stamp: ''!
registerAsNew: anObject 
	anObject isNil ifTrue: [^nil].
	commitPhase == #building ifTrue: [self halt]. "Should not happen. Probably indicates that we're triggering proxies while firing the mappings"
	self newObjects add: anObject.
	self register: anObject.
	^anObject! !

!UnitOfWork methodsFor: 'registering' stamp: ''!
registerForDelete: anObject 
	"Register anObject with intent to delete it."
	[deleting add: anObject.
	self register: anObject] ensure: [deleting remove: anObject ifAbsent: []].! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
registerTransitiveClosure
	"Look for new objects reachable from currently registered objects, and propagate deletions"

	self computeCollectionDeletes.
	self
		registeredObjectsDo:
			[:eachObject | self registerTransitiveClosureFrom: eachObject].
	self propagateDeletes.! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
registerTransitiveClosureFrom: anObject

	| descriptor |
	anObject glorpIsCollection ifTrue: [
		anObject do: [:each | session register: each].
		^self].

	descriptor := session descriptorFor: anObject class.
	descriptor isNil ifTrue: [^self].
	descriptor mappings do: [:eachMapping |
			self instantiateIfRequiredFor: anObject mapping: eachMapping].
	descriptor referencedIndependentObjectsFrom: anObject do: [:eachObject |
		session register: eachObject].! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
registeredObjects
	^self correspondenceMap keys! !

!UnitOfWork methodsFor: 'enumerating' stamp: ''!
registeredObjectsDo: aBlock

	transaction registeredObjectsDo: [:each |
		(each glorpIsCollection or: [session hasDescriptorFor: each]) ifTrue: [
			aBlock value: each]].! !

!UnitOfWork methodsFor: 'initializing' stamp: ''!
reinitialize

	rowMap := RowMap new.
	commitPhase := nil.
	deletedObjects := IdentitySet new.
	deleting := IdentitySet new.
	commitPlan := OrderedCollection new.
	deletePlan := OrderedCollection new.
	transaction abort.! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
relatedRowsFor: each
	| rows extraRows |
	rows := each relatedRowsIn: rowMap.
	extraRows := self session system additionalRelatedRowsFor: each in: rowMap.
	extraRows isNil ifFalse: [rows addAll: extraRows].
	^rows select: [:eachRelatedRow | eachRelatedRow status notNil].! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
reserveSequenceNumbers
	rowsByTable := commitPlan glorpGroupedBy: [:each | each table].
	rowsByTable keysAndValuesDo: [:eachTable :rows | 
		eachTable sequences do: [:eachSequence | 
			eachSequence 
				reserveSequenceNumbers: (
					rows inject: 0 into: [:sum :each | 
						(each needsValueFromSequence: eachSequence) 
							ifTrue: [sum + 1] ifFalse: [sum]])
					in: self session
					for: eachTable]]! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
rollback
	self abort.! !

!UnitOfWork methodsFor: 'enumerating' stamp: ''!
rowsForTable: aTable do: aBlock

	rowMap rowsForTable: aTable do: aBlock.! !

!UnitOfWork methodsFor: 'accessing' stamp: ''!
rowsToWrite
	^commitPlan.! !

!UnitOfWork methodsFor: 'private' stamp: ''!
sendPostWriteNotification
	self 
		registeredObjectsDo: [:eachObject | session sendPostWriteEventTo: eachObject]! !

!UnitOfWork methodsFor: 'private' stamp: ''!
sendPreWriteNotification
	self 
		registeredObjectsDo: [:eachObject | session sendPreWriteEventTo: eachObject]! !

!UnitOfWork methodsFor: 'accessing' stamp: ''!
session
	^session! !

!UnitOfWork methodsFor: 'accessing' stamp: ''!
session: aGlorpSession
	session := aGlorpSession! !

!UnitOfWork methodsFor: 'testing' stamp: ''!
shouldFilterDeletions
	"Should we filter deleted objects from the results of queries. We don't do this if we're in the process of registering the transitive closure, because any fetches we do there are in order to get information we need for the commit (e.g. primary keys). If we filtered the deleted objects, then we may not be able to figure out primary keys for the objects to be deleted"
	^commitPhase ~~ #registering.! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
sortCommitPlanRows
	| sorter |
	sorter := GlorpRowSorter new
		getChildrenVia: [:each | self relatedRowsFor: each].
	commitPlan := sorter sort: commitPlan.
	sorter := GlorpRowSorter new
		getChildrenVia: [:each | self relatedRowsFor: each].
	deletePlan := sorter sort: deletePlan.! !

!UnitOfWork methodsFor: 'accessing' stamp: ''!
system
	^session system.! !

!UnitOfWork methodsFor: 'deletion' stamp: ''!
undelete: anObject ifNotDeleted: aBlock
	deletedObjects remove: anObject ifAbsent: aBlock.
	(session embeddedObjectsFor: anObject) do: [:each | self undelete: each].! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
updateSessionCache
	rowMap
		keysAndValuesDo: [:eachObject :eachRow | 
			eachRow shouldBeWritten ifTrue: [
				self updateSessionCacheFor: eachObject withRow: eachRow]].
	deletedObjects do: [:each |
		session cacheRemoveObject: each].
	linkRowsForUpdate do: [:eachRow |
		eachRow shouldBeWritten ifTrue: [
			self updateSessionCacheFor: eachRow owner withRow: eachRow]].! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
updateSessionCacheFor: anObject withRow: aRow 
	| key |
	(rowMap isRowMapKey: anObject) ifTrue: [^self].  "Not cachable"
	key := aRow primaryKey.
	key isNil ifTrue: [^self].
	(session cacheContainsObjectForClass: anObject class key: key) 
		ifFalse: [self addObject: anObject toCacheKeyedBy: key].! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
validateRows
	"Perform basic validation. Always test for equal named but non-identical tables, a sign of a malformed  system or other loss of identity. Also delegate to the descriptorSystem, which may have useful application-specific validations to do."
	
	| tables tableNames |
	tables := Set new.
	rowMap rowsDo: [:each |
		tables add: each table].
	tableNames := tables collect: [:each | each qualifiedName].
	tables asSet size = tableNames asSet size ifFalse: [self error: 'multiple table objects with the same name'].
	self system validateRows: self.! !

!UnitOfWork methodsFor: 'begin/commit/abort' stamp: ''!
validateRowsPostWrite
	"Perform basic validation after writing is done. It's better to validate beforehand, but some things are easier to catch afterwards"
	
	self system validateRowsPostWrite: self.! !

!UnitOfWork methodsFor: 'deletion' stamp: ''!
willDelete: anObject
	"Return true if we will delete anObject when we commit."	
	^(deletedObjects includes: anObject) or: [deleting includes: anObject].! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
writeRows

	succeeded := true.
	self reserveSequenceNumbers.
	self assignSequenceNumbers.
	self session tablesInCommitOrder do: [:eachTable | self writeRowsForTable: eachTable].
	deletePlan reverseDo: [:eachRow | | rowSucceeded |
		rowSucceeded := session writeRow: eachRow.
		succeeded := succeeded & rowSucceeded].
	self readBackNewRowInformation.! !

!UnitOfWork methodsFor: 'private/mapping' stamp: ''!
writeRowsForTable: aTable

	| rows rowsByOperation |

	rows := rowsByTable at: aTable ifAbsent: [#()].
	^rows isEmpty ifFalse: [
		rowsByOperation := rows glorpGroupedBy: [:eachRow | session shouldInsert: eachRow].
		rowsByOperation do: [:eachListOfHomogeneousRows | 
			succeeded := succeeded & (session writeHomogeneousRows: eachListOfHomogeneousRows)]].! !

Object subclass: #VariableJoin
	instanceVariableNames: 'joins'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Expressions'!

!VariableJoin commentStamp: '<historical>' prior: 0!
This represents a join that could be to multiple different tables, e.g. when using horizontal inheritance.

Instance Variables:
	joins	<Dictionary from: Class to: Join>	description of joins

!

!VariableJoin classSide methodsFor: 'instance creation' stamp: ''!
new
	^super new initialize.! !

!VariableJoin methodsFor: 'accessing' stamp: ''!
addJoinFor: aClass as: aJoin
	joins at: aClass put: aJoin.! !

!VariableJoin methodsFor: 'accessing' stamp: ''!
allSourceFields
	"We assume that variable joins vary in the targets, not the in sources."
	^self anyJoin allSourceFields! !

!VariableJoin methodsFor: 'accessing' stamp: ''!
allTargetFields
	^joins inject: OrderedCollection new into: [:sum :each | sum addAll: each allTargetFields. sum].! !

!VariableJoin methodsFor: 'accessing' stamp: ''!
anyJoin
	^joins detect: [:each | true].! !

!VariableJoin methodsFor: 'converting' stamp: ''!
asGeneralGlorpExpression
	"We know that we represent querying for an abstract class, so we don't want to convert ourselves to a join form, but rather just to a source mapping form. There are some assumptions in here, but it should work for the basic cases anyway."
	| base expression anyJoin descriptor |
	anyJoin := self anyJoin.
	base := anyJoin base.
	descriptor := base system descriptorForTable: anyJoin allTargetFields first table.
	expression := EmptyExpression new base: base.
	anyJoin mappableFieldsDo: [:source :target |
		| mapping thingToCompare |
		mapping := descriptor mappingForField: target.
		thingToCompare := mapping isNil
			ifTrue: [base getField: target]
			ifFalse: [base get: mapping attributeName].
		expression := expression AND: (thingToCompare get: #= withArguments: (Array with: (source asGlorpExpressionOn: base)))].
	^expression.! !

!VariableJoin methodsFor: 'converting' stamp: ''!
asGeneralGlorpExpressionWithFields
	"The normal conversion of a join is to an expression equating fields and parameters. Instead, generate one with field to field comparisons. "
	| base expression mappingName anyJoin descriptor |
	anyJoin := self anyJoin.
	base := anyJoin base.
	descriptor := base system descriptorForTable: anyJoin allTargetFields first table.
	expression := EmptyExpression new base: base.
	anyJoin mappableFieldsDo: [:source :target |
		mappingName := (descriptor mappingForField: target) attributeName.
		expression := expression AND: ((base get: mappingName) get: #= withArguments: (Array with: (self targetExpressionFor: source)))].
	^expression.! !

!VariableJoin methodsFor: 'converting' stamp: ''!
asGlorpExpressionOn: aBaseExpression
	| copy |
	copy := self copy.
	copy joins: (copy joins collect: [:each | each asGlorpExpressionOn: aBaseExpression]).
	^copy.! !

!VariableJoin methodsFor: 'accessing' stamp: ''!
generalExpressionPart
	^nil.! !

!VariableJoin methodsFor: 'initialize-release' stamp: ''!
initialize
	joins := IdentityDictionary new.! !

!VariableJoin methodsFor: 'converting' stamp: ''!
isConstant: aTarget
	"The target can be either a constant (which gets turned into a ConstantExpression) or (usually) a DatabaseField, representing a parameter to the query"
	(aTarget class == DatabaseField) ifTrue: [^false].
	(aTarget isGlorpExpression and: [aTarget isConstantExpression not]) ifTrue: [^false].
	^true.! !

!VariableJoin methodsFor: 'testing' stamp: ''!
isGlorpExpression
	^true.! !

!VariableJoin methodsFor: 'testing' stamp: ''!
isJoin
	^true.! !

!VariableJoin methodsFor: 'accessing' stamp: ''!
joins
	^joins.! !

!VariableJoin methodsFor: 'accessing' stamp: ''!
joins: aCollection
	joins := aCollection.! !

!VariableJoin methodsFor: 'api' stamp: ''!
mapFromSource: sourceObject andTarget: targetObject intoRowsIn: aRowMap
	(sourceObject isNil or: [targetObject isNil]) ifTrue: [^self].
	^(joins at: targetObject class)
		mapFromSource: sourceObject
		andTarget: targetObject
		intoRowsIn: aRowMap.! !

!VariableJoin methodsFor: 'accessing' stamp: ''!
mappableSourceFieldsDo: aBlock

	self anyJoin mappableSourceFieldsDo: aBlock.! !

!VariableJoin methodsFor: 'preparing' stamp: ''!
prepareIn: aQuery! !

!VariableJoin methodsFor: 'As yet unclassified' stamp: ''!
primaryKeyFromDictionary: anIdentityDictionary 
	^self anyJoin primaryKeyFromDictionary: anIdentityDictionary.! !

!VariableJoin methodsFor: 'printing' stamp: ''!
printOn: aStream
	[aStream query] on: Error do: [:ex | ^super printOn: aStream].
	(joins at: aStream query resultClass) printOn: aStream.! !

!VariableJoin methodsFor: 'printing' stamp: ''!
printSQLOn: aStream withParameters: aDictionary 
	[aStream query] on: Error do: [:ex | ^super printOn: aStream].
	(joins at: aStream query resultClass) printSQLOn: aStream withParameters: aDictionary.! !

!VariableJoin methodsFor: 'accessing' stamp: ''!
queryLevelBaseExpression
	^self ultimateBaseExpression.! !

!VariableJoin methodsFor: 'preparing' stamp: ''!
rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression
	^(self joins at: aBaseExpression descriptor describedClass) 
		rebuildOn: aBaseExpression startingFrom: anExpression withOuterScopeBase: ultimateBaseExpression! !

!VariableJoin methodsFor: 'As yet unclassified' stamp: ''!
shortCircuitOnNullComparison: anIdentityDictionary 
	joins do: [:each |
		(each shortCircuitOnNullComparison: anIdentityDictionary)
			ifFalse: [^false]].
	^true.! !

!VariableJoin methodsFor: 'converting' stamp: ''!
targetExpressionFor: eachTarget 
	^(self isConstant: eachTarget)
		ifTrue: [eachTarget]
		ifFalse: [
			eachTarget isGlorpExpression 
			ifTrue: [eachTarget] 
			ifFalse: [(self anyJoin base getTable: eachTarget table) getField: eachTarget]].! !

!VariableJoin methodsFor: 'accessing' stamp: ''!
ultimateBaseExpression

	^self anyJoin base.! !

!SortedCollection methodsFor: '*Glorp-Core' stamp: ''!
writeStream
	^AddingWriteStream on: self.! !

!IdentityDictionary methodsFor: '*Glorp-Core' stamp: ''!
glorpAtIdentity: aKey ifAbsent: aBlock
	^self at: aKey ifAbsent: aBlock.! !

Error subclass: #GlorpError
	instanceVariableNames: ''
	classVariableNames: 'Announcements'
	poolDictionaries: ''
	category: 'Glorp-Core'!

!GlorpError commentStamp: '<historical>' prior: 0!
This is an abstract superclass for GLORP exceptions.
!

GlorpError subclass: #CannotAutomaticallyDetermineJoin
	instanceVariableNames: 'constraints tables'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!CannotAutomaticallyDetermineJoin commentStamp: '<historical>' prior: 0!
When we create descriptors, Glorp will try to determine the join to use between two tables based on foreign key constraints, if it hasn't been specified explicitly. If we can't find any foreign key constraints, or if we find too many, and can't decide automatically which one to use, we raise this exception.

This is an exception raised when creating 
Instance Variables:
	constraints	<Array of: ForeignKeyConstraint>	The list of constraints we found.
	tables	<Collection of: DatabaseTable>	The two tables we were trying to find a join between.

!

!CannotAutomaticallyDetermineJoin methodsFor: 'accessing' stamp: ''!
constraints
	constraints isNil ifTrue: [constraints := #()].
	^constraints! !

!CannotAutomaticallyDetermineJoin methodsFor: 'accessing' stamp: ''!
constraints: aCollection
	constraints := aCollection! !

!CannotAutomaticallyDetermineJoin methodsFor: 'accessing' stamp: ''!
defaultMessageText
	^'Cannot calculate the join automatically, not enough information.'.! !

!CannotAutomaticallyDetermineJoin methodsFor: 'accessing' stamp: ''!
tables
	^tables! !

!CannotAutomaticallyDetermineJoin methodsFor: 'accessing' stamp: ''!
tables: databaseTables
	tables := databaseTables! !

GlorpError subclass: #CannotFindSession
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!CannotFindSession commentStamp: '<historical>' prior: 0!
This is an error that we raise when trying to determine what session to use for an operation, and we don't have enough information to figure it out.
!

!CannotFindSession classSide methodsFor: 'testing' stamp: ''!
mayResume
	^true.! !

!CannotFindSession methodsFor: 'printing' stamp: ''!
defaultMessageText
	^'Cannot find the right session to use - send the #session: message first.'.! !

GlorpError subclass: #DuplicatePrimaryKeyException
	instanceVariableNames: 'existingObject newObject'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!DuplicatePrimaryKeyException commentStamp: '<historical>' prior: 0!
This is an exception we raise when we detect two different objects with the same primary key in our cache.
!

!DuplicatePrimaryKeyException classSide methodsFor: 'instance creation' stamp: ''!
new: newObject existing: existingObject
	^self new
		newObject: newObject;
		existingObject: existingObject;
		yourself.! !

!DuplicatePrimaryKeyException methodsFor: 'accessing' stamp: ''!
defaultMessageText
	^'Duplicate objects with same primary key'.! !

!DuplicatePrimaryKeyException methodsFor: 'accessing' stamp: ''!
existingObject
	^existingObject! !

!DuplicatePrimaryKeyException methodsFor: 'accessing' stamp: ''!
existingObject: anObject
	existingObject := anObject! !

!DuplicatePrimaryKeyException methodsFor: 'accessing' stamp: ''!
newObject
	^newObject! !

!DuplicatePrimaryKeyException methodsFor: 'accessing' stamp: ''!
newObject: anObject
	newObject := anObject! !

GlorpError subclass: #GlorpDatabaseReadError
	instanceVariableNames: 'command databaseError accessor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!GlorpDatabaseReadError commentStamp: '<historical>' prior: 0!
This is the error we raise when we encounter an exception inside Glorp's database read code. This is uniform across dialects, so it can be reliably caught. We wrap the underlying exception.

Instance Variables:
	accessor	<DatabaseAccessor>	The accessor we were executing against.
	command	<DatabaseCommand>	The command we were executing
	databaseError	<Exception>	The exception that the system raised.

!

!GlorpDatabaseReadError classSide methodsFor: 'initialize-release' stamp: ''!
mayResume
	^true.! !

!GlorpDatabaseReadError methodsFor: 'accessing' stamp: ''!
accessor
	^accessor! !

!GlorpDatabaseReadError methodsFor: 'accessing' stamp: ''!
accessor: aDatabaseAccessor
	accessor := aDatabaseAccessor! !

!GlorpDatabaseReadError methodsFor: 'accessing' stamp: ''!
command
	^command! !

!GlorpDatabaseReadError methodsFor: 'accessing' stamp: ''!
command: aDatabaseCommand
	command := aDatabaseCommand! !

!GlorpDatabaseReadError methodsFor: 'accessing' stamp: ''!
databaseError
	^databaseError! !

!GlorpDatabaseReadError methodsFor: 'accessing' stamp: ''!
databaseError: aDatabaseErrorObject
	databaseError := aDatabaseErrorObject! !

!GlorpDatabaseReadError methodsFor: 'accessing' stamp: ''!
session
	^command session.! !

GlorpError subclass: #GlorpDatabaseWriteError
	instanceVariableNames: 'command databaseError accessor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Core'!

!GlorpDatabaseWriteError commentStamp: '<historical>' prior: 0!
This is the error we raise when we encounter an exception inside Glorp's database write code. This is uniform across dialects, so it can be reliably caught. We wrap the underlying exception.

Instance Variables:
	accessor	<DatabaseAccessor>	The accessor we were executing against.
	command	<DatabaseCommand>	The command we were executing
	databaseError	<Exception>	The exception that the system raised.

!

!GlorpDatabaseWriteError methodsFor: 'accessing' stamp: ''!
accessor
	^accessor! !

!GlorpDatabaseWriteError methodsFor: 'accessing' stamp: ''!
accessor: anObject
	accessor := anObject! !

!GlorpDatabaseWriteError methodsFor: 'accessing' stamp: ''!
command
	^command! !

!GlorpDatabaseWriteError methodsFor: 'accessing' stamp: ''!
command: aDatabaseCommand
	command := aDatabaseCommand! !

!GlorpDatabaseWriteError methodsFor: 'accessing' stamp: ''!
databaseError
	^databaseError! !

!GlorpDatabaseWriteError methodsFor: 'accessing' stamp: ''!
databaseError: aDatabaseErrorObject
	databaseError := aDatabaseErrorObject! !

!GlorpDatabaseWriteError methodsFor: 'initialize-release' stamp: ''!
initialize
	"Initialize a newly created instance. This method must answer the receiver."

	super initialize.
	" *** Edit the following to properly initialize instance variables ***"
	command := nil.
	databaseError := nil.
	" *** And replace this comment with additional initialization code *** "
	^self! !

!GlorpDatabaseWriteError methodsFor: 'accessing' stamp: ''!
session
	^command session.! !

!GlorpError classSide methodsFor: 'exception compatibility' stamp: ''!
ancestorOf: anExceptionThingy
	"VA Compatibility"
	^self exceptionalEvent handles: anExceptionThingy! !

!GlorpError classSide methodsFor: 'exception compatibility' stamp: ''!
description
	"For ObjectStudio"
	^super description.! !

!GlorpError classSide methodsFor: 'testing' stamp: ''!
mayResume
	"This is a VisualWorks convention. Reimplement this here so that it will also be found in other dialects that don't have it in the superclass."
	^false! !

!GlorpError classSide methodsFor: 'ANSI' stamp: ''!
signal
	^self new signal.! !

!GlorpError classSide methodsFor: 'ANSI' stamp: ''!
signal: aString
	^self new signal: aString.! !

!GlorpError methodsFor: 'testing' stamp: ''!
ancestorOf: anExceptionThingy
	"VA Compatibility"
	^self handles: anExceptionThingy! !

!GlorpError methodsFor: 'accessing' stamp: ''!
databaseError
"rh 8/1/2004 21:26 for #getMessageText"

	^self class name! !

!GlorpError methodsFor: 'printing' stamp: ''!
description
	"Make sure that we've populated our messageText before calling the superclass implementation."
	self getMessageText.
	^super description.! !

!GlorpError methodsFor: 'accessing' stamp: 'GuillermoPolito 4/16/2011 23:03'!
getMessageText
	"This is horribly ugly. Not only are dialects inconsistent, but behaviour between types of exceptions is inconsistent within dialects"
	
	Dialect isVisualWorks ifTrue: [^self getVisualWorksMessageText].
	Dialect isVisualAge ifTrue: [^self getVisualAgeMessageText].
	(Dialect isSqueak or: [Dialect isPharo]) ifTrue: [^self databaseError].
	^self printString.! !

!GlorpError methodsFor: 'accessing' stamp: ''!
getVisualAgeMessageText
	
	| tempTag |
	tempTag := self basicTag.
	^tempTag isNil ifTrue: [self printString] ifFalse: [tempTag errorText].! !

!GlorpError methodsFor: 'accessing' stamp: ''!
getVisualWorksMessageText
	"In VisualWorks, we often get arrays of ExternalDatabaseExceptions, whose parameter is an array of ExternalDatabaseError objects. Or we might just get an exception, or we might get nil or something else entirely, in which case we fall back to a default error message."
	| dbError |
	self databaseError isNil ifTrue: [^messageText := self defaultMessageText].
	dbError := self databaseError.
	dbError glorpIsCollection ifTrue: [dbError := dbError first].
	messageText := [dbError dbmsErrorString] on: Error do: [:ex | 
		ex return: ([dbError messageText] on: Error do: [:anotherEx |
			anotherEx return: self defaultMessageText])].
	^messageText.! !

!GlorpError methodsFor: 'testing' stamp: ''!
isResumable
	^self class mayResume.! !

!GlorpError methodsFor: 'accessing' stamp: ''!
messageText
	(messageText isNil or: [messageText isEmpty]) ifTrue: [
		messageText := self getMessageText].
	^messageText.! !

!GlorpError methodsFor: 'accessing' stamp: ''!
session
	"We don't know what our session is, return nil by default."
	^nil.! !

!GlorpError methodsFor: 'ANSI' stamp: ''!
signal
	"VisualWorks has non-ANSI semantics for this method. Re-implement to do the standard thing"
	^Dialect isVisualWorks ifTrue: [self raise] ifFalse: [super signal].! !

!GlorpError methodsFor: 'ANSI' stamp: ''!
signal: aString
	"VisualWorks has non-ANSI semantics for this method. Re-implement to do the standard thing"

	^Dialect isVisualWorks
		ifTrue:
			[self messageText: aString.
			self raise]
		ifFalse: [super signal: aString].! !

GlorpError subclass: #GlorpIllegalCommand
	instanceVariableNames: 'command'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!GlorpIllegalCommand commentStamp: '<historical>' prior: 0!
This is the exception we raise if we try to execute a command for which permission has been denied at the Glorp level. e.g. we try to insert while denyWrites is active.

Instance Variables:
	command	<DatabaseCommand>	The command we tried to execute.

!

!GlorpIllegalCommand methodsFor: 'accessing' stamp: ''!
command
	^command! !

!GlorpIllegalCommand methodsFor: 'accessing' stamp: ''!
command: aDatabaseCommand
	command := aDatabaseCommand! !

!GlorpIllegalCommand methodsFor: 'accessing' stamp: ''!
defaultMessageText
	^'Illegal Command'! !

!GlorpIllegalCommand methodsFor: 'accessing' stamp: ''!
session
	^command session.! !

GlorpError subclass: #GlorpInvalidExpressionError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!GlorpInvalidExpressionError commentStamp: '<historical>' prior: 0!
This represents an invalid expression condition in a query.!

GlorpError subclass: #GlorpInvalidTypeError
	instanceVariableNames: 'object expectedType'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!GlorpInvalidTypeError commentStamp: '<historical>' prior: 0!
When mapping, if we encounter an object in a relationship that doesn't match the type we're expecting from the ClassModel/AttributeModel then we raise this exception. 

Instance Variables:
	expectedType	<GlorpClassModel>		The classModel for the type we were expecting.
	object	<Object>	The object that contained the invalid reference.

!

!GlorpInvalidTypeError methodsFor: 'accessing' stamp: ''!
defaultMessageText
	^'Object being written does not match the type of the attribute: ''', self object printString,  ''' ~= ''' , self expectedType printString, ''''.! !

!GlorpInvalidTypeError methodsFor: 'accessing' stamp: ''!
expectedType
	^expectedType! !

!GlorpInvalidTypeError methodsFor: 'accessing' stamp: ''!
expectedType: aGlorpClassModel
	expectedType := aGlorpClassModel! !

!GlorpInvalidTypeError methodsFor: 'accessing' stamp: ''!
object
	^object! !

!GlorpInvalidTypeError methodsFor: 'accessing' stamp: ''!
object: anObject
	object := anObject! !

GlorpError subclass: #GlorpTransactionFailure
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!GlorpTransactionFailure commentStamp: '<historical>' prior: 0!
This is the exception we raise if we encounter a database error trying to commit a transaction.
!

!GlorpTransactionFailure methodsFor: 'accessing' stamp: ''!
defaultMessageText
	^'Optimistic Lock Conflict'.! !

GlorpError subclass: #GlorpWriteFailure
	instanceVariableNames: 'command object'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!GlorpWriteFailure commentStamp: '<historical>' prior: 0!
This is the error we raise if we encounter an error trying to write some rows. This typically indicates an optimistic lock failure, although it might also happen if we violated a constraint.

Instance Variables:
	command	<DatabaseCommand>	The command we were trying to execute.
	object	<Object>	The object or objects we were trying to write.

!

!GlorpWriteFailure classSide methodsFor: 'testing' stamp: ''!
mayResume
	^true.! !

!GlorpWriteFailure methodsFor: 'accessing' stamp: ''!
command
	^command! !

!GlorpWriteFailure methodsFor: 'accessing' stamp: ''!
command: anObject
	command := anObject! !

!GlorpWriteFailure methodsFor: 'accessing' stamp: ''!
defaultMessageText
	^'Database write failed'.! !

!GlorpWriteFailure methodsFor: 'accessing' stamp: ''!
object
	^object! !

!GlorpWriteFailure methodsFor: 'accessing' stamp: ''!
object: anObject
	object := anObject! !

!GlorpWriteFailure methodsFor: 'accessing' stamp: ''!
session
	^self command session.! !

GlorpError subclass: #ValidationError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!ValidationError commentStamp: '<historical>' prior: 0!
ValidationError is a superclass for all errors that are raise while validating a descriptor. The exception may resume.

!

ValidationError subclass: #InvalidSessionSetup
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!InvalidSessionSetup commentStamp: '<historical>' prior: 0!
InvalidSessionSetup will be raised the descriptor system we are validating is different from the current session descriptor.

!

ValidationError subclass: #MultipleMappingsForInstanceVariable
	instanceVariableNames: 'attributeNames readableMappings'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!MultipleMappingsForInstanceVariable commentStamp: '<historical>' prior: 0!
This error raises when validating a descriptor if there is more than one mapping to the same instance variable.

Instance Variables:
	attributeNames	<Collection of String>	the list of attributes referencing the same instance variable
	readableMappings	<Collection of Mapping>	the list of mappings referencing the same instance variable

!

!MultipleMappingsForInstanceVariable methodsFor: 'accessing' stamp: ''!
attributeNames
	^attributeNames! !

!MultipleMappingsForInstanceVariable methodsFor: 'accessing' stamp: ''!
attributeNames: anObject
	attributeNames := anObject! !

!MultipleMappingsForInstanceVariable methodsFor: 'accessing' stamp: ''!
messageText
	^'Multiple readable mappings for one instance variable:<1s>' expandMacrosWith: self readableMappings  printString! !

!MultipleMappingsForInstanceVariable methodsFor: 'accessing' stamp: ''!
readableMappings
	^readableMappings! !

!MultipleMappingsForInstanceVariable methodsFor: 'accessing' stamp: ''!
readableMappings: anObject
	readableMappings := anObject! !

ValidationError subclass: #NoTableFound
	instanceVariableNames: 'expectedTables'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Glorp-Misc'!

!NoTableFound commentStamp: '<historical>' prior: 0!
NoTableFound is raise if there is possible mismatch between an active record and database table names

Instance Variables:
	expectedTables	<OrderedCollection>	 expected table names to map the model

!

!NoTableFound methodsFor: 'accessing' stamp: ''!
expectedTables
	^expectedTables! !

!NoTableFound methodsFor: 'accessing' stamp: ''!
expectedTables: aCollection
	expectedTables := aCollection! !

!NoTableFound methodsFor: 'accessing' stamp: ''!
messageText
	self model isString ifTrue: [^'No table <1s> found' expandMacrosWith: self model].
	^'No table found for <1s>, expected to see one of  <2s>' expandMacrosWith: self model name with: self expectedTables printString! !

!NoTableFound methodsFor: 'printing' stamp: ''!
printOn: aStream
	aStream nextPutAll: self messageText! !

!ValidationError classSide methodsFor: 'testing' stamp: ''!
mayResume

	^true! !

!ValidationError methodsFor: 'printing' stamp: ''!
description
	^self messageText! !

!ValidationError methodsFor: 'accessing' stamp: ''!
model
	^self parameter! !

!ValidationError methodsFor: 'accessing' stamp: ''!
model: aClass
	self parameter: aClass! !
FunctionExpression initialize!
